// Copyright (C) 2009 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;
using System.Diagnostics;
using System.Linq;
using System.IO;

namespace Px2Pg
{
	public class MainClass
	{
		public static int Main(string[] args)
		{
			if (args.Length != 2) {
				ImprimirUso();
				return 1;
			}
			
			var directorioSistema = args[0];
			var archivoSalida = args[1];
			
			// Chequeo que el directorio del sistema exista
			if (!Directory.Exists(directorioSistema)) {
				Console.WriteLine("El directorio del sistema no existe.");
				return 1;
			}
			
			// Creo un directorio temporal para trabajar
			var tempDir = Path.GetTempPath() +
				Path.DirectorySeparatorChar.ToString() +
				"stratos";
			
			Console.WriteLine("Directorio temporal: " + tempDir);
			
			if (Directory.Exists(tempDir))
				Directory.Delete(tempDir, true);
			
			Directory.CreateDirectory(tempDir);
			
			// Elimino el archivo de salida
			if (File.Exists(archivoSalida))
				File.Delete(archivoSalida);
			
			/* Declaro una variable que me indique la fecha de última modificación
			 * de la carpeta del sistema de mosaicos pividori menos 30 días. O sea
			 * que lo que hago es ver cuál es la fecha de última modificación de
			 * la carpeta del sistema que se pasó como argumento. Siendo esta X,
			 * calculo otra que sea Y = X - 30 días. Así busco los archivos .db
			 * que hayan sido modificados tal que su fecha de modificación sea:
			 * Z > Y (o sea, que se haya modificado dentro de los últimos 30 días
			 * anteriores). */
			
			var ultimaFechaModificacionDirectorioSistema =
				Directory.GetLastWriteTime(directorioSistema);
			var fechaLimiteModificacion =
				ultimaFechaModificacionDirectorioSistema - TimeSpan.FromDays(30);
			
			string[] forzarInclusion = new string[] {
				"Trubro.db",
				"Tsubrub.db"
			};
			
			// Obtengo una lista de los archivos del sistema de MP que me interesan
			var archivosParadox =
				from f in Directory.GetFiles(directorioSistema, "*", SearchOption.AllDirectories)
					let fi = new FileInfo(f)
					where
						(!fi.Name.Contains(" ") && // El nombre del archivo no debe tener espacios en blanco
					 	fi.LastWriteTime > fechaLimiteModificacion &&
					 	(fi.Length / 1024.0) >= 10.0 && // El archivo debe ser mayor a 10 kilobytes
						fi.Name.ToLower().EndsWith(".db")) ||
						forzarInclusion.Contains<string>(fi.Name)
					select fi;
			
			Console.WriteLine("Archivos Paradox a procesar: " + archivosParadox.Count<FileInfo>());
//			foreach (var f in archivosParadox)
//				Console.Write(f.Name + " ");
//			Console.WriteLine();
			
			// Le agrego un BEGIN
			File.AppendAllText(archivoSalida, "BEGIN;\n");
			
			File.AppendAllText(archivoSalida, "SET search_path TO 'viejo';\n");
			
			// Por cada archivo lo convierto ejecutando pxsqldump
			Process proc;
			ProcessStartInfo startInfo;
			
			foreach (var fi in archivosParadox) {
				Console.WriteLine("Procesando archivo '" + fi.Name + "'");
				proc = new Process();
				
				Console.WriteLine("  Proceso: pxsqldump -d pgsql -f '" + fi.FullName + "'");
				startInfo = new ProcessStartInfo("pxsqldump", "-d pgsql -f '" + fi.FullName + "'");
				startInfo.UseShellExecute = false;
				startInfo.RedirectStandardOutput = true;
				startInfo.StandardOutputEncoding = Encoding.GetEncoding("ISO-8859-1");
				proc.StartInfo = startInfo;
				
				// Ejecuto el proceso y espero a que termine
				Console.Write("    Ejecutando y esperando...");
				proc.Start();
				
				Console.WriteLine(" OK");
				
				// Escribo toda la salida de pxsqldump a un archivo
				// en el directorio temporal
				var archivoS = tempDir +
					Path.DirectorySeparatorChar +
					fi.Name + "-stratos" + PxSqlFixer.numeroTabla;
				
				Console.WriteLine("    Escribiendo archivo de salida: " + archivoS);
				
				File.WriteAllText(archivoS,
				                  proc.StandardOutput.ReadToEnd(),
				                  proc.StandardOutput.CurrentEncoding);
				                  
				
				// Ahora lo proceso con oem2ansi. El archivo de salida tiene
				// el mismo nombre con un agregado ".ansi" (así funciona oem2ansi)
				proc = new Process();
				startInfo = new ProcessStartInfo("oem2ansi.bin", "file='" + archivoS + "' oemtab=spanish.oem");
				startInfo.UseShellExecute = false;
				startInfo.RedirectStandardOutput = true;
				startInfo.StandardOutputEncoding = Encoding.GetEncoding("ISO-8859-1");
				proc.StartInfo = startInfo;
				
				Console.WriteLine("    Ejecutando oem2ansi...");
				proc.Start();
				proc.WaitForExit();
				
				// Reemplazo los nombres de las tablas con pxsql_fixer
				Console.WriteLine("    Ejecutando pxsql_fixer...");
				var archivoProc = archivoS + ".procesado";
				PxSqlFixer.Convertir(archivoS + ".ansi", archivoProc);
				
				Console.WriteLine("    Agregando el resultado al archivo '" + archivoSalida + "'");
				File.AppendAllText(archivoSalida, File.ReadAllText(archivoProc));
			}
			
			// Le agrego un COMMIT
			File.AppendAllText(archivoSalida, "COMMIT;\n");
			
			return 0;
		}
		
		private static void ImprimirUso()
		{
			Console.WriteLine("px2pg: Buscar archivos de Paradox (*.db) " +
			                  "y genera un archivo .sql final para importar " +
			                  "en una base de datos PostgreSQL.");
			
			Console.WriteLine("px2pg DIRECTORIO_SISTEMA_MOSAICOS_PIVIDORI ARCHIVO_SQL_OUTPUT");
		}
	}
}