// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Configuration;

using Ice;
using log4net;

using StratosServer.Controladores;
using Stratos.Servicios;
using Stratos.Common.Excepciones;

namespace ConsultasWeb
{
	public class StratosServerConexion : Ice.Application
	{
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);
		
		private const string HOST_STRATOS_SERVER = "hostStratosServer";
		private const string PUERTO_STRATOS_SERVER = "puertoStratosServer";
		private const string STRATOS_SERVER_TIMEOUT = "stratosServerTimeout";
		
		private string host, puerto, timeout;
		
		private static StratosServerConexion instancia;
		
		private static ControladorTerminalesPrx controladorTerminales;
		
		public ControladorProductosPrx controladorProductos;
		
		private DatosTerminal datosTerminal = new DatosTerminal {
			identificador = "ClienteWeb"
		};
		
		private StratosServerConexion()
		{
			host = ConfigurationManager.AppSettings[HOST_STRATOS_SERVER];
			puerto = ConfigurationManager.AppSettings[PUERTO_STRATOS_SERVER];
			timeout = ConfigurationManager.AppSettings[STRATOS_SERVER_TIMEOUT];
		}
		
		public static StratosServerConexion Inicializar()
		{
			StratosServerConexion con = new StratosServerConexion();
			con.main(new string[0]);
			
			return con;
		}
		
		public override int run (string[] args)
		{
			shutdownOnInterrupt();
			ConectarServidor();
			
			return 0;
		}
		
		private void InicializarControladorTerminales()
		{
			// Obtengo el proxy al ControladorTerminales
			Ice.ObjectPrx proxyObj =
				Ice.Util.initialize()
					.stringToProxy("ControladorTerminales:tcp -h " +
					               host +
					               " -p " + puerto +
					               " -t " + timeout);
			
			controladorTerminales =
				ControladorTerminalesPrxHelper.checkedCast(proxyObj);
			
			// Verifico si el proxy para el Controlador de Terminales es válido
		    if (controladorTerminales == null)
			{
				string mensaje = "El proxy para el Controlador de Terminales es inválido";
		        logger.Error(mensaje);
				throw new System.Exception(mensaje);
			}
		}
		
		public void ConectarServidor()
		{
			logger.Debug("Conectando con el servidor");
			
			InicializarControladorTerminales();
			
			// Me conecto al servidor
			// FIXME: podría arrojar excepciones
			try {
				logger.Debug("Conectando con el servidor");
				controladorTerminales.Conectar(datosTerminal);
			}
			catch (ConexionException ex) {
				logger.Error("Error al intentar conectar con el servidor: " + ex.mensaje);
				
				logger.Debug("Reintentando conexión al servidor");
				
				// FIXME: Podría haber un tipo de excepción especial para esto
				// Intento desconectarme del servidor (el error puede deberse a ello) y
				// reconectarme.
				
				try
				{
					/* Me desconecto del servidor que ha rechazado mi conexión */
					this.DesconectarServidor();
					
					controladorTerminales.Conectar(datosTerminal);
					
					logger.Debug("Reconexión exitosa.");
				}
				catch(System.Exception)
				{
					string mensaje = "La reconexión ha fallado";
					throw new System.Exception(mensaje);
				}
			}
			
			Ice.ObjectPrx tempProxyObj =
				controladorTerminales.ObtenerProxyControlador("ControladorProductos");
			
			try
			{
		    	controladorProductos =
					ControladorProductosPrxHelper.checkedCast(tempProxyObj);
			}
			catch (Ice.ConnectionRefusedException)
			{
				string mensaje = "Conexión rechazada al validar proxy: ControladorProductos";
				throw new System.Exception(mensaje);
			}
			
			if (controladorProductos == null)
			{
				string mensaje = "El proxy para el Controlador de Terminales es inválido: " +
					"ControladorProductos";
				throw new System.Exception(mensaje);
			}
		}
		
		public void DesconectarServidor()
		{
			logger.Debug("Intentando desconectar con el servidor");
			
			// Verifico si el proxy para el Controlador de Terminales es válido
		    if (controladorTerminales == null)
			{
				string mensaje = "El proxy para el Controlador de Terminales es inválido";
		        logger.Error(mensaje);
				throw new System.Exception(mensaje);
			}
			
			logger.Debug("Desconectando del servidor");
			// Intento desconectarme del servidor
			controladorTerminales.Desconectar(new DatosTerminal {
				identificador = "ClienteWeb"
			});
			logger.Debug("Desconexión realizada correctamente");
		}
		
		public bool VerificarConexion()
		{
			if (controladorTerminales != null && controladorProductos != null)
			{
				try
				{
					controladorTerminales.ice_ping();
					controladorProductos.ice_ping();
				}
				catch (System.Exception)
				{
					return false;
				}
			}
			else
			{
				return false;
			}
			
			return true;
		}
	}
}

