// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using System.Data;

using Castle.ActiveRecord;

using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Migrador
{
	public class MapeoClientes : IMapeoTabla
	{
		private static MapeoClientes instancia;
		
		private MapeoClientes()
		{
		}
		
		public string MensajeAntesDeComenzar {
			get {
				return "- Poner la clave primaria a PrimaryKeyType.Assigned";
			}
		}
		
		public string MensajeDespuesDeTerminar {
			get {
				return "- Poner la clave Primaria a PrimaryKeyType.Native\n" +
					"- Hacer que la secuencia tenga el siguiente número de cliente disponible.";
			}
		}
		
		public string DescripcionMapeo {
			get {
				return "Tabla de Clientes";
			}
		}
		
		public string Tabla {
			get {
				return "viejo.\"Cliente.db-stratos2\"";
			}
		}
		
		public string ConsultaSql {
			get {
				return "select * from " + this.Tabla +
					" order by \"Cod_Cli\"::numeric asc ";
			}
		}
		
		public static Cliente LeerClienteDesdeDataReader(IDataReader reader)
		{
			DataGetter dg = new DataGetter(reader);
			
			Cliente cliente = new Cliente();
			
			cliente.PorcentajeBonificacion = Convert.ToDecimal(dg.GetDouble("Por_Bon"));
			
			string codigoCliente = dg.GetString("Cod_Cli");
			if (codigoCliente != null && Stratos.Common.StratosUtil.IsLong(codigoCliente))
				cliente.Codigo = Convert.ToInt32(codigoCliente);
			
			cliente.Cuit = new CUIT(dg.GetString("Num_Cuit"));
			cliente.Domicilio.Direccion = dg.GetString("Dir_Cli");
			cliente.Domicilio.Localidad = dg.GetString("Loc_Cli");
			cliente.Domicilio.CodigoPostal = dg.GetString("Cod_Pos");
			cliente.Domicilio.Provincia = GetProvinciaDto(dg.GetString("Cod_Pcia"));
			// Las notas las omito, porque tienen basura
			//cliente.Notas = dg.GetString("Notas");
			cliente.RazonSocial = dg.GetString("Nom_Cli");
			
			string tipoIva = dg.GetString("Tip_Iva");
			if (tipoIva.Equals("E"))
				cliente.TipoResponsable = TipoResponsable.Exento;
			else if (tipoIva.Equals("C"))
				cliente.TipoResponsable = TipoResponsable.ConsumidorFinal;
			else if (tipoIva.Equals("I"))
				cliente.TipoResponsable = TipoResponsable.ResponsableInscripto;
			
			if (dg.ColumnaExiste("Tel_Cli"))
				cliente.Telefono = dg.GetString("Tel_Cli");
			
			return cliente;
		}
		
		public void Transformar(IDataReader reader)
		{
			DataGetter dg = new DataGetter(reader);
			
			Cliente cliente;

			while (reader.Read()) {
				using (SessionScope sessionScope = new SessionScope()) {
					cliente = LeerClienteDesdeDataReader(reader);
					
					Console.WriteLine("Procesando cliente: " + cliente.RazonSocial);
					
					// Verifico si el cliente es inválido
					if (!cliente.IsValid()) {
						
						Console.WriteLine("  Corrijiendo cliente...");
						
						// Itero por cada propiedad inválida y agrego una nota sobre la misma
						PropertyInfo pi;
						PropertyInfo[] propiedades =
							cliente.PropertiesValidationErrorMessages.Keys
								.Cast<PropertyInfo>().ToArray();
						
						for (int i=0; i<cliente.PropertiesValidationErrorMessages.Count; i++) {
							pi = propiedades[i];
							
							Console.WriteLine("    ERROR: Propiedad inválida: '" + pi.Name + "' (" + pi.GetValue(cliente, null) + ")");
							cliente.Notas += pi.Name + " inválido (" +
								pi.GetValue(cliente, null) + ");\n";
							
							Console.WriteLine("      Mensaje: " + cliente.ValidationErrorMessages[i]);
						}
						
						Cliente.ModoMigracion();
					}
					
					try {
						cliente.Create();
						Console.WriteLine("  ¡Cliente convertido correctamente!");
					}
					catch(ActiveRecordValidationException ex) {
						Console.WriteLine("  No se pudo corregir. Propiedades inválidas:");
						
						// Itero por cada propiedad inválida y la imprimo en consola
//						foreach (KeyValuePair<PropertyInfo,string[]> kvp in ex.PropertiesValidationErrorMessages)
//							Console.WriteLine("    " + kvp.Key + " (" + kvp.Value[0] + ")");
						
//						Console.WriteLine();
//						Console.WriteLine("Presione una tecla para continuar...");
//						Console.ReadLine();
					}
					finally {
						Cliente.ModoNormal();
					}
					
					Console.WriteLine();
				}
			}
		}
		
		private static ProvinciaDto? GetProvinciaDto(string prov)
		{
			if (prov == null || prov.Trim().Equals(String.Empty))
				return ProvinciaDto.NoEspecificada;
			
			ProvinciaDto? provinciaDto = null;
			
			if (prov.Equals("BA"))
				provinciaDto = ProvinciaDto.BuenosAires;
			else if (prov.Equals("CB"))
				provinciaDto = ProvinciaDto.Cordoba;
			else if (prov.Equals("CH"))
				provinciaDto = ProvinciaDto.Chaco;
			else if (prov.Equals("CO"))
				provinciaDto = ProvinciaDto.Corrientes;
			else if (prov.Equals("CU"))
				provinciaDto = ProvinciaDto.Chubut;
			else if (prov.Equals("ER"))
				provinciaDto = ProvinciaDto.EntreRios;
			else if (prov.Equals("FR"))
				provinciaDto = ProvinciaDto.Formosa;
			else if (prov.Equals("MS"))
				provinciaDto = ProvinciaDto.Misiones;
			else if (prov.Equals("MZ"))
				provinciaDto = ProvinciaDto.Mendoza;
			else if (prov.Equals("NQ"))
				provinciaDto = ProvinciaDto.Neuquen;
			else if (prov.Equals("SE"))
				provinciaDto = ProvinciaDto.SantiagoDelEstero;
			else if (prov.Equals("SF"))
				provinciaDto = ProvinciaDto.SantaFe;
			else if (prov.Equals("ST"))
				provinciaDto = ProvinciaDto.Salta;
			else
				provinciaDto = ProvinciaDto.NoEspecificada;
			
			return provinciaDto;
		}
		
		public void RealizarLimpieza(IDbConnection con)
		{
			IDbCommand comando = con.CreateCommand();
			
			comando.CommandText =
				"update " + this.Tabla + " " +
				"set \"Num_Cuit\" = NULL " +
				"where trim(both \"Num_Cuit\") = '';"
				+
				"update " + this.Tabla + " " +
				"set \"Cod_Pcia\" = NULL " +
				"where trim(both \"Cod_Pcia\") = '';"
				+
				"update " + this.Tabla + " " +
				"set \"Num_Cuit\" = regexp_replace(\"Num_Cuit\",'[^0-9]', '', 'g');";
			
			comando.ExecuteNonQuery();
		}
		
		public static MapeoClientes Instancia {
			get {
				if (instancia == null)
					instancia = new MapeoClientes();
				
				return instancia;
			}
		}
	}
}
