// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using System.Data;

using Castle.ActiveRecord;

using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Migrador
{
	public class MapeoProveedores : IMapeoTabla
	{
		private static MapeoProveedores instancia;
		
		private MapeoProveedores()
		{
		}
		
		public string MensajeAntesDeComenzar {
			get {
				return "- Poner la clave primaria a PrimaryKeyType.Assigned";
			}
		}
		
		public string MensajeDespuesDeTerminar {
			get {
				return "- Poner la clave Primaria a PrimaryKeyType.Native\n" +
					"- Hacer que la secuencia tenga el siguiente número de proveedor disponible.";
			}
		}
		
		public string DescripcionMapeo {
			get {
				return "Tabla de Proveedores";
			}
		}
		
		public string Tabla {
			get {
				return "viejo.\"Prove.db-stratos8\"";
			}
		}
		
		public string ConsultaSql {
			get {
				return "select * from " + this.Tabla +
					" order by \"Cod_Cli\"::numeric asc ";
			}
		}
		
		public static Proveedor LeerProveedorDesdeDataReader(IDataReader reader)
		{
			return MapeoProveedores.LeerProveedorDesdeDataReader(reader, "Cod_Cli", "Nom_Cli", "Tip_Iva", "Num_Cuit");
		}
		
		public static Proveedor LeerProveedorDesdeDataReader(IDataReader reader,
		                                                     string columnaCodigo,
		                                                     string columnaRazonSocial,
		                                                     string columnaTipoIva,
		                                                     string columnaCuit)
		{
			DataGetter dg = new DataGetter(reader);
			
			Proveedor proveedor = new Proveedor();
			
			string codigo = dg.GetString(columnaCodigo);
			Console.WriteLine("   Código leído del proveedor: {0}", codigo);
			
			if (codigo != null && Stratos.Common.StratosUtil.IsLong(codigo))
				proveedor.Codigo = Convert.ToInt32(codigo);
			
			proveedor.Cuit = new CUIT(dg.GetString(columnaCuit));
			proveedor.Domicilio.Direccion = dg.GetString("Dir_Cli");
			proveedor.Domicilio.Localidad = dg.GetString("Loc_Cli");
			proveedor.Domicilio.CodigoPostal = dg.GetString("Cod_Pos");
			proveedor.Domicilio.Provincia = GetProvinciaDto(dg.GetString("Cod_Pcia"));
			// Las notas las omito, porque tienen basura
			//cliente.Notas = dg.GetString("Notas");
			proveedor.RazonSocial = dg.GetString(columnaRazonSocial);
			
			string tipoIva = dg.GetString(columnaTipoIva);
			if (tipoIva.Equals("E"))
				proveedor.TipoResponsable = TipoResponsable.Exento;
			else if (tipoIva.Equals("C"))
				proveedor.TipoResponsable = TipoResponsable.ConsumidorFinal;
			else if (tipoIva.Equals("I"))
				proveedor.TipoResponsable = TipoResponsable.ResponsableInscripto;
			
			if (dg.ColumnaExiste("Tel_Cli"))
				proveedor.Telefono = dg.GetString("Tel_Cli");
			
			return proveedor;
		}
		
		public void Transformar(IDataReader reader)
		{
			DataGetter dg = new DataGetter(reader);
			
			Proveedor proveedor;

			while (reader.Read()) {
				using (SessionScope sessionScope = new SessionScope()) {
					proveedor = LeerProveedorDesdeDataReader(reader);
					
					Console.WriteLine("Procesando proveedor: " + proveedor.RazonSocial);
					
					PropertyInfo[] propiedadesInvalidas = null;
					
					// Verifico si el cliente es inválido
					if (!proveedor.IsValid()) {
						
						Console.WriteLine("  Corrijiendo...");
						
						// Itero por cada propiedad inválida y agrego una nota sobre la misma
						PropertyInfo pi;
						propiedadesInvalidas =
							proveedor.PropertiesValidationErrorMessages.Keys
								.Cast<PropertyInfo>().ToArray();
						
						for (int i=0; i<proveedor.PropertiesValidationErrorMessages.Count; i++) {
							pi = propiedadesInvalidas[i];
							
							Console.WriteLine("    ERROR: Propiedad inválida: '" + pi.Name + "' (" + pi.GetValue(proveedor, null) + ")");
							proveedor.Notas += pi.Name + " inválido (" +
								pi.GetValue(proveedor, null) + ");\n";
							
							Console.WriteLine("      Mensaje: " + proveedor.ValidationErrorMessages[i]);
						}
						
						Proveedor.ModoMigracion();
					}
					
					// Si se trata de un error con el cuit, continuo con el siguiente proveedor
					if (propiedadesInvalidas != null && propiedadesInvalidas.Any(prop => prop.Name == "Cuit")) {
						Console.WriteLine("   Continuo con el siguiente...");
						continue;
					}
					
					try {
						proveedor.Create();
						Console.WriteLine("  ¡Proveedor convertido correctamente!");
					}
					catch(ActiveRecordValidationException ex) {
						Console.WriteLine("  NO SE PUDO CORREGIR !!.");
					}
					finally {
						Proveedor.ModoNormal();
					}
					
					Console.WriteLine();
				}
			}
		}
		
		private static ProvinciaDto? GetProvinciaDto(string prov)
		{
			if (prov == null || prov.Trim().Equals(String.Empty))
				return ProvinciaDto.NoEspecificada;
			
			ProvinciaDto? provinciaDto = null;
			
			if (prov.Equals("BA"))
				provinciaDto = ProvinciaDto.BuenosAires;
			else if (prov.Equals("CB"))
				provinciaDto = ProvinciaDto.Cordoba;
			else if (prov.Equals("CH"))
				provinciaDto = ProvinciaDto.Chaco;
			else if (prov.Equals("CO"))
				provinciaDto = ProvinciaDto.Corrientes;
			else if (prov.Equals("CU"))
				provinciaDto = ProvinciaDto.Chubut;
			else if (prov.Equals("ER"))
				provinciaDto = ProvinciaDto.EntreRios;
			else if (prov.Equals("FR"))
				provinciaDto = ProvinciaDto.Formosa;
			else if (prov.Equals("MS"))
				provinciaDto = ProvinciaDto.Misiones;
			else if (prov.Equals("MZ"))
				provinciaDto = ProvinciaDto.Mendoza;
			else if (prov.Equals("NQ"))
				provinciaDto = ProvinciaDto.Neuquen;
			else if (prov.Equals("SE"))
				provinciaDto = ProvinciaDto.SantiagoDelEstero;
			else if (prov.Equals("SF"))
				provinciaDto = ProvinciaDto.SantaFe;
			else if (prov.Equals("ST"))
				provinciaDto = ProvinciaDto.Salta;
			else
				provinciaDto = ProvinciaDto.NoEspecificada;
			
			return provinciaDto;
		}
		
		public void RealizarLimpieza(IDbConnection con)
		{
			IDbCommand comando = con.CreateCommand();
			
			comando.CommandText =
				"update " + this.Tabla + " " +
				"set \"Num_Cuit\" = NULL " +
				"where trim(both \"Num_Cuit\") = '';"
				+
				"update " + this.Tabla + " " +
				"set \"Cod_Pcia\" = NULL " +
				"where trim(both \"Cod_Pcia\") = '';"
				+
				"update " + this.Tabla + " " +
				"set \"Num_Cuit\" = regexp_replace(\"Num_Cuit\",'[^0-9]', '', 'g');";
			
			comando.ExecuteNonQuery();
		}
		
		public static MapeoProveedores Instancia {
			get {
				if (instancia == null)
					instancia = new MapeoProveedores();
				
				return instancia;
			}
		}
	}
}
