// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Configuration;

using Mono.Addins;
using log4net;

namespace Stratos.Common
{
	/// <summary>
	/// Clase genérica para gestionar la configuración. Recibe un tipo genérico
	/// que indica la sección principal del archivo de configuración.
	/// Se encarga de inicializar, modificar, obtener y guardar configuraciones.
	/// </summary>
	public abstract class GestorConfiguracionGenerico<TConfigSection>
		where TConfigSection : StratosConfigurationSection, new()
	{
		private readonly ILog logger;
		
		private Configuration configuracion;
		
		public GestorConfiguracionGenerico()
		{
			this.logger = LogManager.GetLogger(this.GetType());
		}
		
		protected abstract string NombreSeccionPrincipal {
			get;
		}
		
		/// <summary>
		/// Nombre del archivo de configuración.
		/// </summary>
		protected abstract string ArchivoConfiguracion {
			get;
		}
		/// <summary>
		/// Path para los addins que se suscriben a un evento que avisa cuando
		/// la configuración ha sido guardada/cambiada.
		/// </summary>
		protected abstract string PathAddinsConfiguracionCambiada {
			get;
		}
		
		/// <summary>
		/// Obtiene la sección principal de configuración.
		/// </summary>
		public TConfigSection StratosConfigurationSection
		{
			get {
				/* FIXME: El método GetSection puede arrojar excepciones. En parece no arrojarlas,
				 * pero sí en Windows. Por ejemplo, si un campo está marcada como obligatorio
				 * (IsRequired) y no está, arroja una excepción. Hay que ver todos los métodos
				 * GetSection. */
				TConfigSection instancia =
					configuracion.GetSection(this.NombreSeccionPrincipal) as TConfigSection;
				
				return instancia;
			}
		}
		
		/// <summary>
		/// Inicializa la configuración. Crea el archivo si no existe, y agrega todas
		/// las secciones personalizadas incluidas en addins.
		/// </summary>
		public void Inicializar()
		{
			// Armo el path del archivo de configuracion
			string pathArchivoConfiguracion =
				System.IO.Directory.GetCurrentDirectory() +
					System.IO.Path.DirectorySeparatorChar + this.ArchivoConfiguracion;
			
			logger.Debug("El archivo de configuracion es: '" + pathArchivoConfiguracion + "'");
			
			// Creo el elemento Configuration con el path del archivo de configuracion
			ExeConfigurationFileMap configMap = new ExeConfigurationFileMap();
			configMap.ExeConfigFilename = pathArchivoConfiguracion;

			TConfigSection seccionStratos = null;
			
			try {
				this.configuracion =
					ConfigurationManager
						.OpenMappedExeConfiguration(configMap,
						                            ConfigurationUserLevel.None);
			
				// Creo la sección principal de configuracion si es que no existe
				logger.DebugFormat("Obteniendo seccion '{0}'", this.NombreSeccionPrincipal);
				
				seccionStratos = this.StratosConfigurationSection;
			}
			catch (ConfigurationErrorsException ex) {
				
				String mensaje = "Error al intentar obtener la sección de " +
					"configuración de Stratos. Verifique que el archivo " +
					"de configuración sea correcto.";
				
				this.logger.Error(mensaje, ex);
				
				throw new ErrorCriticoException(mensaje, ex);
			}
			catch (System.Xml.XmlException ex) {
				
				String mensaje = "Error al intentar obtener la sección de " +
					"configuración de Stratos. Si desea iniciar el servidor con las " +
					"opciones por defecto, elimine el archivo de configuración e inicie " +
					"nuevamente.";
				
				this.logger.Error(mensaje, ex);
				
				throw new ErrorCriticoException(mensaje, ex);
			}
			
			if (seccionStratos == null) {
				logger.Debug("La seccion 'StratosConfigSection' no existe. Creando una...");
				seccionStratos = new TConfigSection();
				
				configuracion.Sections.Add(this.NombreSeccionPrincipal,
				                           seccionStratos);
			}
			
			// Agrego las instancias de las secciones personalizadas a la seccion principal
			logger.Debug("Agregando las instancias de las secciones personalizadas a la seccion principal");
			
			ConfigurationElement seccionNueva;
			string nombreSeccion;
			
			foreach (object[] cs in seccionStratos.Secciones) {
				nombreSeccion = cs[0].ToString();
				
				logger.Debug("Procesando instancia de seccion: " + nombreSeccion);
				seccionNueva = seccionStratos.GetConfigurationElement(nombreSeccion);
				
				if (seccionNueva == null) {
					logger.Debug("No existe instancia para la seccion: " + nombreSeccion);
					logger.Debug("Agregando instancia para la seccion: " + nombreSeccion);
					
					seccionNueva = (ConfigurationElement) cs[1];
					seccionStratos.Add(nombreSeccion, seccionNueva);
				}
			}
		}
		
		public ConfigurationElement GetConfigurationElement(string elementName)
		{
			TConfigSection instancia = this.StratosConfigurationSection;
			
			return (ConfigurationElement) instancia.GetConfigurationElement(elementName);
		}
		
		private void Save(ConfigurationSaveMode saveMode)
		{
			configuracion.Save(saveMode);
			ConfigurationManager.RefreshSection(this.NombreSeccionPrincipal);
		}
		
		public void Save()
		{
			Save(ConfigurationSaveMode.Modified);
			
			/* Aviso a todos los addins suscriptos que ha sido guardada (y posiblemente
			 * modificada) la configuración */
			foreach (IConfiguracionCambiada configCambiada in
			         AddinManager.GetExtensionObjects(this.PathAddinsConfiguracionCambiada))
				configCambiada.ConfiguracionCambiada();
		}
	}
}
