// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Configuration;

using log4net;
using Mono.Addins;

namespace Stratos.Common
{
	/// <summary>
	/// Sección de configuración genérica.
	/// </summary>
	public abstract class StratosConfigurationSection : ConfigurationSection
	{
		#region Campos
		private readonly ILog logger;
		
		private static ConfigurationPropertyCollection propiedades;
		private static List<object[]> secciones;
		#endregion
		
		#region Constructores
		public StratosConfigurationSection ()
		{
			this.logger = LogManager.GetLogger(this.GetType());
		}
		#endregion
		
		#region Propiedades
		/// <summary>
		/// Path para encontrar los addins suscriptos para agregar secciones
		/// personalizadas.
		/// </summary>
		protected abstract string PathAddins {
			get;
		}
		
		/// <summary>
		/// Obtiene las propiedades personalizadas agregadas a la sección principal
		/// de configuración.
		/// </summary>
		protected override ConfigurationPropertyCollection Properties {
			get {
				// Primero inicializo los addins
				this.InicializarLecturaAddins();
				
				return propiedades;
			}
		}
		
		/// <summary>
		/// Obtiene una lista de arrays que poseen el nombre y la instancia de
		/// las secciones personalizadas.
		/// </summary>
		public List<object[]> Secciones {
			get {
				// Primero inicializo los addins
				this.InicializarLecturaAddins();
				
				return secciones;
			}
		}
		#endregion
		
		#region Métodos
		/// <summary>
		/// Inicializa las secciones y propiedades.
		/// </summary>
		private void InicializarLecturaAddins()
		{
			if (propiedades != null && secciones != null)
				return;
			
			/* Cargo todas las secciones personalizadas de los addins.
			 * En este punto, primero cargo en la seccion principal (StratosConfigSection) las propiedades
			 * con los nombres y tipos de las secciones personalizadas. Luego, mas abajo, agrego las
			 * instancias de estas secciones. Es asi porque primero deben estar las ConfigurationProperty */
			secciones = new List<object[]>();
			
			// Estas son las propiedades para la sección principal que voy a ir cargando
			// dinámicamente según los addins registrados.
			propiedades = new ConfigurationPropertyCollection();
			
			logger.Debug("Configurando la seccion principal con secciones personalizadas (propiedades)");
			foreach (ConfigurationElementNode sectionNode in
			         AddinManager.GetExtensionNodes (this.PathAddins)) {
				
				ConfigurationElement configurationSection =
					sectionNode.GetConfigurationElement() as ConfigurationElement;
				
				if (configurationSection != null) {
					
					logger.Debug("Agregando propiedad para la seccion: " + sectionNode.nombre);
					ConfigurationProperty property =
						new ConfigurationProperty(sectionNode.nombre,
						                          configurationSection.GetType());
					
					propiedades.Add(property);
					
					// Almaceno datos sobre la sección (nombre y la instancia)
					secciones.Add(new object[2] {sectionNode.nombre, configurationSection});
				}
				else
					logger.Debug("Error: '" + configurationSection.ToString() +
					             "' no es del tipo ConfigurationElement");
			}
		}
		
		public ConfigurationElement GetConfigurationElement(string elementName)
		{
			return (ConfigurationElement) this[elementName];
		}
		
		public void Add(string sectionName, ConfigurationElement section)
		{
			this[sectionName] = section;
		}
		#endregion
	}
}
