// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

using Commons.Collections;
using NVelocity;
using NVelocity.Runtime;
using NVelocity.App;
using System.IO;

namespace Stratos.Common
{
	public class GeneradorDocumento
	{
		#region Campos
		protected string archivoTemplate;
		protected string sistemaOperativo;
		
		protected Dictionary<string,object> objetos;
		#endregion
		
		#region Constructores
		public GeneradorDocumento (string archivoTemplate, string nombreObjeto, object objeto)
			: this (archivoTemplate)
		{
			this.AgregarObjeto(nombreObjeto, objeto);
		}
		
		/// <summary>
		/// Permite especificar el nombre del archivo template, el directorio relativo para
		/// WebKit y el sistema operativo donde se renderizará el HTML resultante.
		/// </summary>
		/// <param name="archivoTemplate">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="directorio">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="sistemaOperativo">
		/// A <see cref="System.String"/>
		/// </param>
		public GeneradorDocumento(string archivoTemplate, string sistemaOperativo)
		{
			this.objetos = new Dictionary<string, object>();
			
			this.archivoTemplate = archivoTemplate;
			
			// Agrego el objeto 'sistema_operativo'
			this.sistemaOperativo = sistemaOperativo;
			this.AgregarObjeto("sistema_operativo", this.sistemaOperativo);
		}
		
		/// <summary>
		/// Permite especificar el nombre del archivo template, el directorio relativo para
		/// WebKit. El sistema operativo donde se renderizará, se supone que será el mismo
		/// donde se generará el archivo HTML.
		/// </summary>
		/// <param name="archivoTemplate">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="directorio">
		/// A <see cref="System.String"/>
		/// </param>
		public GeneradorDocumento(string archivoTemplate)
			: this(archivoTemplate, Environment.OSVersion.ToString())
		{
		}
		#endregion
		
		#region Métodos
		/// <summary>
		/// Agrega un objeto al contexto de NVelocity, para que pueda ser
		/// visible desde el template.
		/// </summary>
		/// <param name="nombreObjecto">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="objeto">
		/// A <see cref="System.Object"/>
		/// </param>
		public void AgregarObjeto (string nombreObjeto, object objeto)
		{
			if (nombreObjeto != null && objeto != null)
				this.objetos[nombreObjeto] = objeto;
		}
		
		public string CrearHTML()
		{
			VelocityEngine velocityEngine = new VelocityEngine();
			
			// Configuro el VelocityEngine inicializándolo con algunas propiedades
			ExtendedProperties props = new ExtendedProperties();
			props.AddProperty(RuntimeConstants.INPUT_ENCODING, "UTF-8");
			props.AddProperty(RuntimeConstants.OUTPUT_ENCODING, "UTF-8");
			props.AddProperty(RuntimeConstants.FILE_RESOURCE_LOADER_PATH,
			                  StratosUtil.DirectorioAplicacion());
			
			velocityEngine.Init(props);
			
			// Verifico que el archivo de template exista.
			if (!File.Exists(this.archivoTemplate))
				throw new ExcepcionStratosException("El archivo de template no existe. Creelo antes", true);
			
			// Cargo el template para la impresión del Presupuesto
			Template template = velocityEngine.GetTemplate(this.archivoTemplate);
			
			/* Creo un contexto y seteo las variables del mismo, para que puedan ser utilizadas
			 * desde el template.
			 * Como NVelocity no puede acceder a los campos públicos (fields) de los objetos Dto,
			 * entonces utilizo una clase wrapper para los items, llamada ItemProducto. Esta clase
			 * recibe un ItemProductoDto, y lo envuelve permitiendo acceder a los campos del mismo
			 * via propiedades. */
			VelocityContext velocityContext = new VelocityContext();
			
			foreach (KeyValuePair<string,object> kvp in this.objetos)
				velocityContext.Put(kvp.Key, kvp.Value);
			
			/* Creo un objeto StringWriter para que NVelocity coloque allí el resultado de mergear
			 * el template con los datos colocados en el VelocityContext. La variable 'writer' tendrá
			 * al final el resultado. */
			StringWriter writer = new StringWriter();
			
			template.Merge(velocityContext, writer);
			
			writer.Flush();
			
			return writer.GetStringBuilder().ToString();
		}
		#endregion
	}
}