// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Globalization;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;

using Stratos.ModeloDto;

namespace Stratos.Common
{
	public static class StratosUtil
	{
		#region Campos
		private static NumberFormatInfo numberFormat =
			(NumberFormatInfo) CultureInfo.CurrentCulture.NumberFormat.Clone();
		
		private static NumberStyles numberStyles =
			NumberStyles.Number ^ NumberStyles.AllowThousands;
		#endregion
		
		#region Constructores
		static StratosUtil()
		{
			// El separador de decimales es el punto
			numberFormat.NumberDecimalSeparator = ".";
			numberFormat.CurrencyDecimalSeparator = ".";
			numberFormat.PercentDecimalSeparator = ".";
			
			// Separador de miles
			numberFormat.NumberGroupSeparator = String.Empty;
			numberFormat.CurrencyGroupSeparator = String.Empty;
			numberFormat.PercentGroupSeparator = String.Empty;
		}
		#endregion
		
		/// <summary>
		/// Obtiene el directorio actual de la aplicación
		/// </summary>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		public static string DirectorioAplicacion()
		{
			return Path.GetDirectoryName(System.Reflection.Assembly
			                             .GetExecutingAssembly().Location);
		}
		
		#region Métodos de chequeo de tipos
		public static bool IsDouble(string cadena)
		{
			double temp;
			
			// Reemplazo las comas por puntos
			string valorProcesado = ProcesarNumeroCadena(cadena);
			
			return double.TryParse(valorProcesado, numberStyles,
			                       numberFormat, out temp);
		}
		
		public static bool IsLong(string cadena)
		{
			long temp;
			
			return long.TryParse(cadena, out temp);
		}
		
		public static bool IsInt(string cadena)
		{
			int temp;
			
			return int.TryParse(cadena, out temp);
		}
		
		public static bool IsBool(string cadena)
		{
			bool temp;
			
			return bool.TryParse(cadena, out temp);
		}
		
		public static bool IsDecimal(string cadena)
		{
			decimal temp;
			
			// Reemplazo las comas por puntos
			string valorProcesado = ProcesarNumeroCadena(cadena);
			
			return decimal.TryParse(valorProcesado, numberStyles,
			                        numberFormat, out temp);
		}
		#endregion
		
		#region Métodos de parseo
		/// <summary>
		/// Dado un número en formato de cadena (como por ejemplo "123,23")
		/// procesa el mismo para prepararlo para ser parseado.
		/// </summary>
		/// <param name="numeroCadena">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		private static string ProcesarNumeroCadena(string numeroCadena)
		{
			return numeroCadena.Replace(",", ".");
		}
		
		/// <summary>
		/// Parsea una cadena y la convierte en double. Interpreta los puntos
		/// y comas como separadores de decimales.
		/// </summary>
		/// <param name="valor">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Double"/>
		/// </returns>
		public static double ParseDouble(string valor, int decimales)
		{
			// Reemplazo las comas por puntos
			string valorProcesado = ProcesarNumeroCadena(valor);
			
			try {
				return RoundDouble(double.Parse(valorProcesado,
				                                numberStyles,
				                                numberFormat),
				                   decimales);
			}
			catch (System.FormatException ex) {
				throw new NumeroInvalidoException("El valor '" + valor + "' no puede " +
				                                  "ser interpretado como número");
			}
		}
		
		public static decimal ParseDecimal(string valor, int decimales)
		{
			// Reemplazo las comas por puntos
			string valorProcesado = ProcesarNumeroCadena(valor);
			
			try {
				return RoundDecimal(decimal.Parse(valorProcesado,
				                                  numberStyles,
				                                  numberFormat),
				                   decimales);
			}
			catch (System.FormatException ex) {
				throw new NumeroInvalidoException("El valor '" + valor + "' no puede " +
				                                  "ser interpretado como número");
			}
		}
		#endregion
		
		#region Métodos de redondeo
		public static double RoundDouble(double valor, int decimales)
		{
			decimal valorDecimal = Convert.ToDecimal(valor);
			decimal resultado = StratosUtil.RoundDecimal(valorDecimal, decimales);
			
			return Convert.ToDouble(resultado);
		}
		
		public static decimal RoundDecimal(decimal valor, int decimales)
		{
			return Math.Round(valor, decimales, MidpointRounding.AwayFromZero);
		}
		#endregion
		
		#region Métodos de conversión
		public static string DoubleToString(double valor)
		{
			return valor.ToString("N2", numberFormat);
		}
		
		public static string DecimalToString(decimal valor)
		{
			return valor.ToString("N2", numberFormat);
		}
		
		/// <summary>
		/// Dada una unidad de medida de un producto, devuelve la versión
		/// reducida en forma de cadena.
		/// </summary>
		/// <param name="unidadProducto">
		/// A <see cref="UnidadProducto"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		public static string UnidadResumida(UnidadProducto unidadProducto)
		{
			switch (unidadProducto) {
				case UnidadProducto.Bolsa:
					return "Bl";
				case UnidadProducto.Kilogramo:
					return "Kg";
				case UnidadProducto.Litro:
					return "Lt";
				case UnidadProducto.Metro:
					return "Mt";
				case UnidadProducto.MetroCuadrado:
					return "M2";
				case UnidadProducto.Tonelada:
					return "Tn";
				case UnidadProducto.Unidad:
					return "U";
				default:
					return String.Empty;
			}
		}
		#endregion
		
		#region Otros métodos
		public static string SplitCamelString(string camel)
		{
			return Regex.Replace(Regex.Replace(camel,
			                                   @"(\P{Ll})(\P{Ll}\p{Ll})", "$1 $2" ),
			                     @"(\p{Ll})(\P{Ll})", "$1 $2" );
		}
		#endregion
	}
}
