// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;

using AutoMapper;

using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Esta clase tiene centralizados todos los Maps de AutoMapper que transforman un objeto
	/// persistente a uno DTO.
	/// </summary>
	public static class BaseDtoMappings
	{
		#region Items
		public static IMappingExpression<TSource, TDestination>
			ItemProductoDtoMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : IItemProducto
				where TDestination : ItemProductoDto
		{
			mappingExpression
				.ForMember(dest => dest.productoNotas,
				           opt => opt.MapFrom(src => src.ProductoAsignado != null ? src.ProductoAsignado.Notas : String.Empty));
			
			return mappingExpression;
		}
		#endregion
		
		#region DocumentoGenericoDto
		public static IMappingExpression<TSource, TDestination>
			DocumentoGenericoDtoMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : IDocumentoGenerico
				where TDestination : DocumentoGenericoDto
		{
			mappingExpression
				.ForMember(dest => dest.fechaTicks,
				       opt => opt.MapFrom(src => src.Fecha.Ticks));
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			DocumentoItemsDtoMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : IDocumentoItems
				where TDestination : DocumentoItemsDto
		{
			IList items = null;
			
			mappingExpression
				/* Antes de realizar el map, si la propiedad de items no está inicializada, entonces
				 * la quito temporalmente (para evitar que al accederse en la transformación hacia Dto
				 * sea inicializada por NHibernate, y por lo tanto que acceda a la base de datos).
				 * De esta forma gano en velocidad.
				 * Por defecto, la propiedad Items (y puede ser cualquier otra también), sino está
				 * inicializada, se supone que no se requieren en el cliente, y por lo tanto no se
				 * mapean. */
				.BeforeMap(delegate (TSource documentoFuente, TDestination documentoDestino) {
					if (!NHibernate.NHibernateUtil.IsInitialized(documentoFuente.ItemsGenericos)) {
						items = documentoFuente.ItemsGenericos;
						documentoFuente.ItemsGenericos = null;
					}
				})
				/* Luego del mapeo, vuelvo a colocar la lista de items en su lugar (hago todo esto ya
				 * que no se trata de una lista cualquiera, sino de una lista especial de NHibernate). */
				.AfterMap(delegate (TSource documentoFuente, TDestination documentoDestino) {
					if (documentoFuente.ItemsGenericos == null)
						documentoFuente.ItemsGenericos = items;
				});
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			DocumentoVentaDtoMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : IDocumentoVenta
				where TDestination : DocumentoVentaDto
		{
			mappingExpression
				.DocumentoItemsDtoMap()
				.DocumentoGenericoDtoMap();
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			FacturaGenericaDtoMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : IFacturaGenerica
				where TDestination : FacturaGenericaDto
		{
			mappingExpression
				.DocumentoVentaDtoMap()
				.ForMember(dest => dest.condVenta,
					       opt => opt.MapFrom(src => src.CondicionVenta));
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			FacturaDtoMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : Factura
				where TDestination : FacturaDto
		{
			mappingExpression
				.FacturaGenericaDtoMap()
				.AfterMap(delegate (TSource factura, TDestination facturaDto) {
					
					// Mapeo el cliente a la versión Dto
					ClienteDto clienteDto = facturaDto.cliente;
					
					// Luego copio al mismo los datos fijos
					clienteDto.razonSocial = factura.ClienteRazonSocial;
					clienteDto.tipoRes = factura.ClienteTipoResponsable;
					
					if (factura.ClienteCuit.HasValue)
						clienteDto.cuit = (new CUIT(factura.ClienteCuit.Value)).ValorFormateado;
					else
						clienteDto.cuit = String.Empty;
				});
			
			return mappingExpression;
		}
		#endregion
	}
}
