// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using AutoMapper;
using Castle.ActiveRecord;
using NHibernate;
using NHibernate.Criterion;

using Stratos.Common;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Esta clase estática tiene maps genéricos de AutoMapper para pasar clases DTO
	/// a clases persistentes.
	/// </summary>
	public static class BaseMappings
	{
		#region Items
		public static IMappingExpression<TSource, TDestination>
			IItemGenericoMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : ItemGenericoDto
				where TDestination : IItemGenerico
		{
			mappingExpression
				.ForMember(dest => dest.Id,
					       opt => opt.Ignore())
				.IActiveRecordValidationBaseMap();
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			IItemProductoMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : ItemProductoDto
				where TDestination : IItemProducto
		{
			mappingExpression
				.IItemGenericoMap()
				.AfterMap(delegate (TSource itemProductoDto, TDestination itemProducto) {
					Producto productoAsociado = Producto.TryFind(itemProductoDto.productoId);
					itemProducto.AsignarProducto(productoAsociado);
				});
			
			return mappingExpression;
		}
		#endregion
		
		#region DocumentoGenerico
		public static IMappingExpression<TSource, TDestination>
			IDocumentoGenericoMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : DocumentoGenericoDto
				where TDestination : IDocumentoGenerico
		{
			mappingExpression
				.IActiveRecordValidationBaseMap()
				.ForMember(dest => dest.Persona,
					       opt => opt.Ignore())
				.ForMember(dest => dest.Fecha,
					       opt => opt.MapFrom(src => new DateTime(src.fechaTicks)));
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			IDocumentoItems<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : DocumentoItemsDto
				where TDestination : IDocumentoGenerico, IDocumentoItems
		{
			mappingExpression
				.IDocumentoGenericoMap()
				.ForMember(dest => dest.ItemsGenericos,
					       opt => opt.Ignore());
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			IDocumentoCompraventaMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : DocumentoItemsDto
				where TDestination : IDocumentoCompraventa
		{
			mappingExpression
				.IDocumentoItems()
				.ForMember(dest => dest.ImporteFinal,
				           opt => opt.Ignore());
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			IDocumentoVentaMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : DocumentoVentaDto
				where TDestination : IDocumentoVenta
		{
			mappingExpression
				.IDocumentoCompraventaMap()
				.ForMember(dest => dest.Cliente,
					       opt => opt.MapFrom(src => Cliente.TryFind(src.cliente != null ? src.cliente.codigo : -1)));
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			IDocumentoCompraMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : DocumentoCompraDto
				where TDestination : IDocumentoCompra
		{
			mappingExpression
				.IDocumentoCompraventaMap()
				.ForMember(dest => dest.Proveedor,
					       opt => opt.MapFrom(src => Proveedor.TryFind(src.proveedor != null ? src.proveedor.codigo : -1)));
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			IFacturaGenericaMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : FacturaGenericaDto
				where TDestination : IFacturaGenerica
		{
			mappingExpression
				.IDocumentoVentaMap()
				.ForMember(dest => dest.CondicionVenta,
					       opt => opt.MapFrom(src => new CondicionVentaFacturaGenerica
					                              (src.condVenta.condPago, src.condVenta.plazo)));
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			IFacturaMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : FacturaDto
				where TDestination : Factura
		{
			mappingExpression
				.IFacturaGenericaMap()
				.ForMember(dest => dest.ClienteRazonSocial,
					       opt => opt.Ignore())
				.ForMember(dest => dest.ClienteTipoResponsable,
					       opt => opt.Ignore())
				.ForMember(dest => dest.ClienteCuit,
					       opt => opt.Ignore())
				.ForMember(dest => dest.TipoAbreviado,
					       opt => opt.Ignore())
				.ForMember(dest => dest.PreNotasCreditoAsociadas,
					       opt => opt.Ignore())
				.ForMember(dest => dest.NotasCreditoAsociadas,
					       opt => opt.Ignore());
			
			return mappingExpression;
		}
		#endregion
		
		#region FacturaCompra
		public static IMappingExpression<TSource, TDestination>
			FacturaCompraMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : FacturaCompraDto
				where TDestination : IFacturaCompra
		{
			mappingExpression
				.IDocumentoCompraMap()
				.ForMember(dest => dest.FechaImputacion,
					       opt => opt.MapFrom(src => new DateTime(src.fechaImputacionTicks)))
				.ForMember(dest => dest.ProveedorRazonSocial,
					       opt => opt.MapFrom(src => src.proveedor != null ? src.proveedor.razonSocial : null))
				.ForMember(dest => dest.ProveedorTipoResponsable,
					       opt => opt.MapFrom(src => src.proveedor != null ? src.proveedor.tipoRes : TipoResponsable.ResponsableInscripto))
				.ForMember(dest => dest.ProveedorCuit,
					       opt => opt.MapFrom(delegate (TSource facturaCompraDto) {
								long cuit = 0L;
								
								ProveedorDto proveedorDto = facturaCompraDto.proveedor;
								
								// Parseo el cuit enviado por el cliente en formato string
								if (proveedorDto != null && !String.IsNullOrEmpty(proveedorDto.cuit)) {
									string cuitStringFormateado =
										proveedorDto.cuit.Replace("-", "");
									
									/* Si luego de quitar los guiones al cuit formateado enviado por
									 * el cliente queda un numero, entonces lo tomo como un cuit. */
									if (StratosUtil.IsLong(cuitStringFormateado))
										cuit = long.Parse(cuitStringFormateado);
								}
								
								return cuit;
							}))
				.ForMember(dest => dest.TipoAbreviado,
					       opt => opt.Ignore());
			
			return mappingExpression;
		}
		#endregion
		
		#region Otros maps
		public static IMappingExpression<TSource, TDestination>
			IActiveRecordValidationBaseMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : class
				where TDestination : IActiveRecordValidationBase
		{
			mappingExpression
				.ForMember(dest => dest.ValidationErrorMessages,
				           opt => opt.Ignore())
				.ForMember(dest => dest.PropertiesValidationErrorMessages,
				           opt => opt.Ignore());
			
			return mappingExpression;
		}
		
		public static IMappingExpression<TSource, TDestination>
			FromPreFactura_ToFacturaMap<TSource, TDestination>(this IMappingExpression<TSource, TDestination> mappingExpression)
				where TSource : PreFactura
				where TDestination : IFactura
		{
			mappingExpression
				.IActiveRecordValidationBaseMap()
				.ForMember(dest => dest.ClienteRazonSocial,
					       opt => opt.Ignore())
				.ForMember(dest => dest.ClienteTipoResponsable,
					       opt => opt.Ignore())
				.ForMember(dest => dest.ClienteCuit,
					       opt => opt.Ignore())
				.ForMember(dest => dest.ItemsGenericos,
					       opt => opt.Ignore())
				.ForMember(dest => dest.Id,
				           opt => opt.Ignore())
				.ForMember(dest => dest.TipoAbreviado,
					       opt => opt.Ignore())
				.ForMember(dest => dest.Valida,
				           opt => opt.Ignore())
				.ForMember(dest => dest.FacturaManual,
				           opt => opt.Ignore())
				.ForMember(dest => dest.Anulada,
				           opt => opt.Ignore())
				.ForMember(dest => dest.Codigo,
				           opt => opt.Ignore())
				.ForMember(dest => dest.Numero,
				           opt => opt.Ignore())
				.ForMember(dest => ((IDocumentoCompraventa) dest).ImporteFinal,
				           opt => opt.Ignore())
				.ForMember(dest => ((IDocumentoImportes) dest).ImporteFinal,
				           opt => opt.Ignore());
			
			return mappingExpression;
		}
		#endregion
	}
}
