// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using Stratos.ModeloObjetos.MetaCampos;

using Castle.ActiveRecord;
using AutoMapper;
using NHibernate.Criterion;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Permite realizar conversiones entre objetos persistentes y objetos DTO.
	/// </summary>
	public static class ConvertidorDto
	{
		public static IMappingEngine MapperEngine {
			get { return Mapper.Engine; }
		}
		
		#region Mapeos
		static ConvertidorDto()
		{
			Mapper.CreateMap<System.Double, System.Decimal>().ConvertUsing(Convert.ToDecimal);
			
			Mapper.CreateMap<System.Decimal, System.Double>().ConvertUsing(Convert.ToDouble);
			
			Mapper.CreateMap<Nullable<long>, long>()
				.ConvertUsing(nullLong => nullLong.HasValue ? nullLong.Value : 0L);
			
			#region Domicilio
			// Domicilio -> DomicilioDto
			Mapper.CreateMap<Domicilio, DomicilioDto>();
			
			// DomicilioDto -> Domicilio
			Mapper.CreateMap<DomicilioDto, Domicilio>();
			#endregion
			
			#region Cliente
			// Cliente -> ClienteDto
			Mapper.CreateMap<Cliente, ClienteDto>()
				.ForMember(dest => dest.tipoRes,
				           opt => opt.MapFrom(src => src.TipoResponsable))
				.ForMember(dest => dest.porcentajeBonificacion,
					       opt => opt.MapFrom(src => (double)src.PorcentajeBonificacion))
				.ForMember(dest => dest.cuit,
				           opt => opt.MapFrom(src => src.Cuit.ValorFormateado))
				.ForMember(dest => dest.condicionPagoPorDefecto,
					       opt => opt.MapFrom(src =>
					                              src.ObtenerMetaCampo(MetaCamposKeys.ClienteCondicionVentaPorDefecto) == null ? 
					                              		null : new CondicionPago[] {((CondicionPago)int.Parse(src.ObtenerMetaCampo(MetaCamposKeys.ClienteCondicionVentaPorDefecto)))} ));
				
			
			// ClienteDto -> Cliente
			Mapper.CreateMap<ClienteDto, Cliente>()
				.ForMember(dest => dest.AltaAutomatica,
				           opt => opt.Ignore())
				.ForMember(dest => dest.Valido,
				           opt => opt.Ignore())
				.ForMember(dest => dest.TipoResponsable,
				           opt => opt.MapFrom(src => src.tipoRes))
				.ForMember(dest => dest.PorcentajeBonificacion,
					       opt => opt.MapFrom(src => (decimal)src.porcentajeBonificacion))
				.ForMember(dest => dest.Cuit,
					       opt => opt.MapFrom(src =>
								src.cuit != null &&
									src.cuit.Trim() != String.Empty ?
					                              new CUIT(src.cuit) : new CUIT()))
				.ForMember(dest => dest.ValidationErrorMessages,
				           opt => opt.Ignore())
				.ForMember(dest => dest.PropertiesValidationErrorMessages,
				           opt => opt.Ignore())
				.ForMember(dest => dest.MetaCampos,
					       opt => opt.Ignore())
				.AfterMap(delegate (ClienteDto src, Cliente dest) {
						
						// Si existe la condicion de venta por defecto, la creo o actualizo y sino la elimino
						if ((src.condicionPagoPorDefecto != null) && (src.condicionPagoPorDefecto.Length > 0))
						{		
							dest.SetearMetaCampo(MetaCamposKeys.ClienteCondicionVentaPorDefecto, ((int)src.condicionPagoPorDefecto[0]).ToString());
						}
						else
						{
							dest.EliminarMetaCampo(MetaCamposKeys.ClienteCondicionVentaPorDefecto);
						}
				});
			#endregion

			#region Proveedor
			// Proveedor -> ProveedorDto
			Mapper.CreateMap<Proveedor, ProveedorDto>()
				.ForMember(dest => dest.tipoRes,
				           opt => opt.MapFrom(src => src.TipoResponsable))
				.ForMember(dest => dest.cuit,
				           opt => opt.MapFrom(src => src.Cuit.ValorFormateado));
			
			// ProveedorDto -> Proveedor
			Mapper.CreateMap<ProveedorDto, Proveedor>()
				.ForMember(dest => dest.AltaAutomatica,
				           opt => opt.Ignore())
				.ForMember(dest => dest.TipoResponsable,
				           opt => opt.MapFrom(src => src.tipoRes))
				.ForMember(dest => dest.Cuit,
					       opt => opt.MapFrom(src =>
								src.cuit != null &&
									src.cuit.Trim() != String.Empty ?
					                              new CUIT(src.cuit) : new CUIT()))
				.ForMember(dest => dest.ValidationErrorMessages,
				           opt => opt.Ignore())
				.ForMember(dest => dest.PropertiesValidationErrorMessages,
				           opt => opt.Ignore());
			#endregion
			
			#region Rubro
			// Rubro -> RubroDto
			Mapper.CreateMap<Rubro, RubroDto>()
				.ForMember(dest => dest.rubroPadre,
					       opt => opt.MapFrom(src => src.RubroPadre));
			
			// RubroDto -> Rubro
			Mapper.CreateMap<RubroDto, Rubro>()
				.ForMember(dest => dest.RubroPadre,
					       opt => opt.MapFrom(src => src.rubroPadre))
				.ForMember(dest => dest.ValidationErrorMessages,
				           opt => opt.Ignore())
				.ForMember(dest => dest.PropertiesValidationErrorMessages,
				           opt => opt.Ignore());
			#endregion
			
			#region Producto
			// Producto -> ProductoDto
			Mapper.CreateMap<Producto, ProductoDto>()
				.ForMember(dest => dest.precioCosto,
					       opt => opt.MapFrom(src => (double)src.PrecioCosto))
				.ForMember(dest => dest.importeSinIva,
					       opt => opt.MapFrom(src => (double)src.ImporteSinIva))
				.ForMember(dest => dest.importeVenta,
					       opt => opt.MapFrom(src => (double)src.ImporteVenta))
				.ForMember(dest => dest.porcentajeIva,
					       opt => opt.MapFrom(src => (double)src.PorcentajeIva));
			
			// ProductoDto -> Producto
			Mapper.CreateMap<ProductoDto, Producto>()
				.ForMember(dest => dest.Valido,
					       opt => opt.Ignore())
				.ForMember(dest => dest.PrecioCosto,
					       opt => opt.MapFrom(src => (decimal)src.precioCosto))
				.ForMember(dest => dest.ImporteSinIva,
					       opt => opt.MapFrom(src => (decimal)src.importeSinIva))
				.ForMember(dest => dest.PorcentajeIva,
					       opt => opt.MapFrom(src => (decimal)src.porcentajeIva))
				.ForMember(dest => dest.ImporteVenta,
					       opt => opt.Ignore())
				.ForMember(dest => dest.ValidationErrorMessages,
				           opt => opt.Ignore())
				.ForMember(dest => dest.PropertiesValidationErrorMessages,
				           opt => opt.Ignore());
			#endregion
			
			#region CondicionVenta
			// CondicionVenta -> CondicionVentaDto
			Mapper.CreateMap<CondicionVenta, CondicionVentaDto>()
				.ForMember(dest => dest.condPago,
					       opt => opt.MapFrom(src => src.Condicion));
			
			// CondicionVentaDto -> CondicionVenta
			Mapper.CreateMap<CondicionVentaDto, CondicionVenta>()
				.ForMember(dest => dest.Condicion,
					       opt => opt.MapFrom(src => src.condPago));
			#endregion
			
			#region Items
			#region ToDto
			// IItemProducto -> ItemProductoDto
			Mapper.CreateMap<IItemProducto, ItemProductoDto>()
				.ItemProductoDtoMap();
			
			// ItemFactura -> ItemGenericoDto/ItemProductoDto
			Mapper.CreateMap<ItemFactura, ItemGenericoDto>()
				.ConvertUsing(item => Mapper.Map<IItemProducto,ItemProductoDto>(item));
			
			Mapper.CreateMap<ItemFactura, ItemGenericoDto>()
				.ConvertUsing(item => Mapper.Map<IItemProducto,ItemProductoDto>(item));
			
			// ItemPreFactura -> ItemGenericoDto/ItemProductoDto
			Mapper.CreateMap<ItemPreFactura, ItemGenericoDto>()
				.ConvertUsing(item => Mapper.Map<IItemProducto,ItemProductoDto>(item));
			
			// ItemFacturaCompra -> ItemGenericoDto/ItemProductoDto
			Mapper.CreateMap<ItemFacturaCompra, ItemGenericoDto>()
				.ConvertUsing(item => Mapper.Map<IItemProducto,ItemProductoDto>(item));
			
			// ItemFacturaGasto -> ItemGenericoDto/ItemProductoDto
			Mapper.CreateMap<ItemFacturaCompra, ItemGenericoDto>()
				.ConvertUsing(item => Mapper.Map<IItemProducto,ItemProductoDto>(item));
			
			// ItemPresupuesto -> ItemGenericoDto/ItemProductoDto
			Mapper.CreateMap<ItemPresupuesto, ItemGenericoDto>()
				.ConvertUsing(item => Mapper.Map<IItemProducto,ItemProductoDto>(item));
			
			// ItemRecibo -> ItemReciboDto
			Mapper.CreateMap<ItemRecibo, ItemReciboDto>()
				.ForMember(dest => dest.medioPag,
					       opt => opt.MapFrom(src => src.MedioPago));
			
			Mapper.CreateMap<ItemRecibo, ItemGenericoDto>()
				.ConvertUsing(item => Mapper.Map<ItemRecibo, ItemReciboDto>(item));
			#endregion
			
			#region FromDto
			Mapper.CreateMap<ItemProductoDto, ItemFactura>()
				.IItemProductoMap();
			Mapper.CreateMap<ItemGenericoDto, ItemFactura>()
				.ConvertUsing(itemDto =>
				              Mapper.Map<ItemProductoDto,ItemFactura>((ItemProductoDto)itemDto));
			
			Mapper.CreateMap<ItemProductoDto, ItemPreFactura>()
				.IItemProductoMap();
			Mapper.CreateMap<ItemGenericoDto, ItemPreFactura>()
				.ConvertUsing(itemDto =>
				              Mapper.Map<ItemProductoDto,ItemPreFactura>((ItemProductoDto)itemDto));
			
			Mapper.CreateMap<ItemProductoDto, ItemFacturaCompra>()
				.IItemProductoMap();
			Mapper.CreateMap<ItemGenericoDto, ItemFacturaCompra>()
				.ConvertUsing(itemDto =>
				              Mapper.Map<ItemProductoDto,ItemFacturaCompra>((ItemProductoDto)itemDto));
			
			Mapper.CreateMap<ItemProductoDto, ItemFacturaCompra>()
				.IItemProductoMap();
			Mapper.CreateMap<ItemGenericoDto, ItemFacturaCompra>()
				.ConvertUsing(itemDto =>
				              Mapper.Map<ItemProductoDto,ItemFacturaCompra>((ItemProductoDto)itemDto));
			
			Mapper.CreateMap<ItemProductoDto, ItemPresupuesto>()
				.IItemProductoMap();
			Mapper.CreateMap<ItemGenericoDto, ItemPresupuesto>()
				.ConvertUsing(itemDto =>
				              Mapper.Map<ItemProductoDto,ItemPresupuesto>((ItemProductoDto)itemDto));
			
			// ItemReciboDto -> ItemRecibo
			Mapper.CreateMap<ItemReciboDto, ItemRecibo>()
				.IItemGenericoMap()
				/* Ya que no es posible pasar valores nulos con ZeroC Ice, al realizar el mapping, si
				 * el medio de pago es efectivo, coloco el valor nulo en la propiedad Cheque. */
				.ForMember(dest => dest.Cheque,
					       opt => opt.MapFrom(src => src.medioPag == MedioPago.Cheque ? src.cheque : null))
				.ForMember(dest => dest.MedioPago,
					       opt => opt.MapFrom(src => src.medioPag));
			
			Mapper.CreateMap<ItemGenericoDto, ItemRecibo>()
				.ConvertUsing(itemDto =>
				              Mapper.Map<ItemReciboDto, ItemRecibo>((ItemReciboDto)itemDto));
			#endregion
			#endregion
			
			#region DocumentoGenerico
			#region ToDto
			Mapper.CreateMap<Factura, FacturaDto>()
				.FacturaDtoMap()
				.Include<NotaCredito, NotaCreditoDto>();
			
			Mapper.CreateMap<PreFactura, PreFacturaDto>()
				.FacturaGenericaDtoMap()
				.Include<PreNotaCredito, PreNotaCreditoDto>();
			
			Mapper.CreateMap<NotaCredito, NotaCreditoDto>()
				.FacturaDtoMap();
			
			Mapper.CreateMap<PreNotaCredito, PreNotaCreditoDto>()
				.FacturaGenericaDtoMap();
			
			Mapper.CreateMap<Presupuesto, PresupuestoDto>()
				.DocumentoItemsDtoMap();
			
			Mapper.CreateMap<FacturaCompra, FacturaCompraDto>()
				.DocumentoItemsDtoMap()
				.AfterMap(delegate (FacturaCompra factura, FacturaCompraDto facturaDto) {
					
					// Si la factura de compra no tiene asignado un proveedor, creo uno
					if (facturaDto.proveedor == null)
						facturaDto.proveedor = new ProveedorDto();
					
					// Mapeo el proveedor a la versión Dto
					ProveedorDto proveedorDto = facturaDto.proveedor;
					
					// Luego copio al mismo los datos fijos
					proveedorDto.razonSocial = factura.ProveedorRazonSocial;
					proveedorDto.tipoRes = factura.ProveedorTipoResponsable;
					
					// Si el proveedor no tiene domicilio, entonces siempre creo un domicilio falso
					// y coloca la provincia como NoEspecificada
					if (proveedorDto.domicilio == null) {
						proveedorDto.domicilio = new DomicilioDto();
						proveedorDto.domicilio.provincia = ProvinciaDto.NoEspecificada;
					}
					
					if (factura.ProveedorCuit > 0)
						proveedorDto.cuit = (new CUIT(factura.ProveedorCuit)).ValorFormateado;
					else
						proveedorDto.cuit = String.Empty;
				});
			
			Mapper.CreateMap<Recibo, ReciboDto>();
			#endregion
			
			#region FromDto
			Mapper.CreateMap<FacturaDto, Factura>()
				.IFacturaMap()
				/* Para las colecciones de NHibernate, siempre hay que ignorarlas en el mapping,
				 * ya que no puedo dejar a Automapper crear la instancia de la colección, porque
				 * es un tipo especial de NHibernate. */
				.ForMember(dest => dest.Items,
					       opt => opt.Ignore())
				.Include<NotaCreditoDto, NotaCredito>()
				.AfterMap(delegate (FacturaDto docDto, Factura doc) {
					ColeccionMapper<ItemGenericoDto,ItemFactura> itemsMapper =
						new ColeccionMapper<ItemGenericoDto,ItemFactura>(docDto.items, doc.Items);
					
					itemsMapper.Mapear();
				});
			
			Mapper.CreateMap<PreFacturaDto, PreFactura>()
				.IFacturaGenericaMap()
				/* Para las colecciones de NHibernate, siempre hay que ignorarlas en el mapping,
				 * ya que no puedo dejar a Automapper crear la instancia de la colección, porque
				 * es un tipo especial de NHibernate. */
				.ForMember(dest => dest.Items,
					       opt => opt.Ignore())
				.Include<PreNotaCreditoDto, PreNotaCredito>()
				// Mapeo los items:
				.AfterMap(delegate (PreFacturaDto docDto, PreFactura doc) {
					ColeccionMapper<ItemGenericoDto,ItemPreFactura> itemsMapper =
						new ColeccionMapper<ItemGenericoDto,ItemPreFactura>(docDto.items, doc.Items);
					
					itemsMapper.Mapear();
				});
			
			Mapper.CreateMap<NotaCreditoDto, NotaCredito>()
				.IFacturaMap()
				/* Para las colecciones de NHibernate, siempre hay que ignorarlas en el mapping,
				 * ya que no puedo dejar a Automapper crear la instancia de la colección, porque
				 * es un tipo especial de NHibernate. */
				.ForMember(dest => dest.Items,
					       opt => opt.Ignore())
				.ForMember(dest => dest.FacturaAsociada,
					       opt => opt.MapFrom(src => Factura.TryFind(src.facturaAsociada != null ? src.facturaAsociada.id : -1)))
				// Mapeo los items:
				.AfterMap(delegate (NotaCreditoDto docDto, NotaCredito doc) {
					ColeccionMapper<ItemGenericoDto,ItemFactura> itemsMapper =
						new ColeccionMapper<ItemGenericoDto,ItemFactura>(docDto.items, doc.Items);
					
					itemsMapper.Mapear();
				});
			
			Mapper.CreateMap<PreNotaCreditoDto, PreNotaCredito>()
				.IFacturaGenericaMap()
				/* Para las colecciones de NHibernate, siempre hay que ignorarlas en el mapping,
				 * ya que no puedo dejar a Automapper crear la instancia de la colección, porque
				 * es un tipo especial de NHibernate. */
				.ForMember(dest => dest.Items,
					       opt => opt.Ignore())
				.ForMember(dest => dest.FacturaAsociada,
					       opt => opt.MapFrom(src => Factura.TryFind(src.facturaAsociada != null ? src.facturaAsociada.id : -1)))
				// Mapeo los items:
				.AfterMap(delegate (PreNotaCreditoDto docDto, PreNotaCredito doc) {
					ColeccionMapper<ItemGenericoDto,ItemPreFactura> itemsMapper =
						new ColeccionMapper<ItemGenericoDto,ItemPreFactura>(docDto.items, doc.Items);
					
					itemsMapper.Mapear();
				});
			
			Mapper.CreateMap<PresupuestoDto, Presupuesto>()
				.IDocumentoVentaMap()
				/* Para las colecciones de NHibernate, siempre hay que ignorarlas en el mapping,
				 * ya que no puedo dejar a Automapper crear la instancia de la colección, porque
				 * es un tipo especial de NHibernate. */
				.ForMember(dest => dest.Items,
					       opt => opt.Ignore())
				// Mapeo los items:
				.AfterMap(delegate (PresupuestoDto docDto, Presupuesto doc) {
					ColeccionMapper<ItemGenericoDto,ItemPresupuesto> itemsMapper =
						new ColeccionMapper<ItemGenericoDto,ItemPresupuesto>(docDto.items, doc.Items);
					
					itemsMapper.Mapear();
				});
			
			Mapper.CreateMap<FacturaCompraDto, FacturaCompra>()
				.FacturaCompraMap()
				/* Para las colecciones de NHibernate, siempre hay que ignorarlas en el mapping,
				 * ya que no puedo dejar a Automapper crear la instancia de la colección, porque
				 * es un tipo especial de NHibernate. */
				.ForMember(dest => dest.Items,
					       opt => opt.Ignore())
				// Mapeo los items:
				.AfterMap(delegate (FacturaCompraDto docDto, FacturaCompra doc) {
					ColeccionMapper<ItemGenericoDto,ItemFacturaCompra> itemsMapper =
						new ColeccionMapper<ItemGenericoDto,ItemFacturaCompra>(docDto.items, doc.Items);
					
					itemsMapper.Mapear();
				});	
			
			Mapper.CreateMap<ReciboDto, Recibo>()
				.IDocumentoItems()
				/* Para las colecciones de NHibernate, siempre hay que ignorarlas en el mapping,
				 * ya que no puedo dejar a Automapper crear la instancia de la colección, porque
				 * es un tipo especial de NHibernate. */
				.ForMember(dest => dest.Items,
					       opt => opt.Ignore())
				.ForMember(dest => dest.Cliente,
					       opt => opt.MapFrom(src => Cliente.TryFind(src.cliente != null ? src.cliente.codigo : -1)))
				// Mapeo los items:
				.AfterMap(delegate (ReciboDto docDto, Recibo doc) {
					ColeccionMapper<ItemGenericoDto,ItemRecibo> itemsMapper =
						new ColeccionMapper<ItemGenericoDto,ItemRecibo>(docDto.items, doc.Items);
					
					itemsMapper.Mapear();
				});
			#endregion
			#endregion
			
			#region DocumentoCredito
			Mapper.CreateMap<DocumentoCredito, DocumentoCreditoDto>()
				.ForMember(dest => dest.tipoDoc,
					       opt => opt.MapFrom(src => src.TipoDocumento));
			#endregion
			
			#region DocumentoDebito
			Mapper.CreateMap<DocumentoDebito, DocumentoDebitoDto>()
				.ForMember(dest => dest.tipoDoc,
					       opt => opt.MapFrom(src => src.TipoDocumento));
			#endregion
			
			#region Otros mapeos
			// ItemPreFactura --> ItemFactura
			Mapper.CreateMap<ItemPreFactura, ItemFactura>()
				.IActiveRecordValidationBaseMap()
				.ForMember(dest => dest.Id,
				           opt => opt.Ignore());
			
			// ItemFactura --> ItemPreFactura
			Mapper.CreateMap<ItemFactura, ItemPreFactura>()
				.IActiveRecordValidationBaseMap()
				.ForMember(dest => dest.Id,
				           opt => opt.Ignore());
			
			// PreNotaCredito --> NotaCredito
			Mapper.CreateMap<PreNotaCredito, NotaCredito>()
				.FromPreFactura_ToFacturaMap()
				.ForMember(dest => dest.Historica,
					       opt => opt.Ignore())
				.ForMember(dest => dest.PreNotasCreditoAsociadas,
					       opt => opt.Ignore())
				.ForMember(dest => dest.NotasCreditoAsociadas,
					       opt => opt.Ignore());
			
			// PreFactura --> Factura
			Mapper.CreateMap<PreFactura, Factura>()
				.FromPreFactura_ToFacturaMap()
				.ForMember(dest => dest.Historica,
					       opt => opt.Ignore())
				.ForMember(dest => dest.PreNotasCreditoAsociadas,
					       opt => opt.Ignore())
				.ForMember(dest => dest.NotasCreditoAsociadas,
					       opt => opt.Ignore())
				.Include<PreNotaCredito, NotaCredito>();
			
			// Factura --> PreFactura
			Mapper.CreateMap<Factura, PreFactura>()
				.IActiveRecordValidationBaseMap()
				.ForMember(dest => dest.Fecha,
					       opt => opt.MapFrom(src => DateTime.Today))
				.ForMember(dest => dest.ItemsGenericos,
					       opt => opt.Ignore());
			#endregion
			
			#region Cheque
			Mapper.CreateMap<Cheque, ChequeDto>()
				.ForMember(dest => dest.fechaEmisionTicks,
					       opt => opt.MapFrom(src => src.FechaEmision.Ticks))
				.ForMember(dest => dest.fechaVencimientoTicks,
					       opt => opt.MapFrom(src => src.FechaVencimiento.Ticks));
			
			Mapper.CreateMap<ChequeDto, Cheque>()
				.IActiveRecordValidationBaseMap()
				.ConstructUsing(chequeDto => chequeDto.id > 0 ? Cheque.TryFind(chequeDto.id) : new Cheque())
				.ForMember(dest => dest.FechaEmision,
					       opt => opt.MapFrom(src => new DateTime(src.fechaEmisionTicks)))
				.ForMember(dest => dest.FechaVencimiento,
					       opt => opt.MapFrom(src => new DateTime(src.fechaVencimientoTicks)));
			#endregion
		}
		#endregion
		
		#region ToDto
		/// <summary>
		/// Convierte un objeto Cliente (persistente) en un objeto ClienteDto.
		/// </summary>
		/// <param name="cliente">
		/// A <see cref="Cliente"/>
		/// </param>
		/// <returns>
		/// A <see cref="ClienteDto"/>
		/// </returns>
		public static ClienteDto ToDto(Cliente cliente)
		{
			return Mapper.Map<Cliente, ClienteDto>(cliente);
		}
		
		/// <summary>
		/// Convierte un array de objetos Cliente en un array de objetos ClienteDto.
		/// </summary>
		/// <param name="clientes">
		/// A <see cref="Cliente"/>
		/// </param>
		/// <returns>
		/// A <see cref="ClienteDto"/>
		/// </returns>
		public static ClienteDto[] ToDto(Cliente[] clientes)
		{
			return Mapper.Map<Cliente[], ClienteDto[]>(clientes);
		}

		/// <summary>
		/// Convierte un objeto Proveedor (persistente) en un objeto ProveedorDto.
		/// </summary>
		/// <param name="proveedor">
		/// A <see cref="Proveedor"/>
		/// </param>
		/// <returns>
		/// A <see cref="ProveedorDto"/>
		/// </returns>
		public static ProveedorDto ToDto(Proveedor proveedor)
		{
			return Mapper.Map<Proveedor, ProveedorDto>(proveedor);
		}
		
		/// <summary>
		/// Convierte un array de objetos Proveedor en un array de objetos ProveedorDto.
		/// </summary>
		/// <param name="proveedores">
		/// A <see cref="Proveedor"/>
		/// </param>
		/// <returns>
		/// A <see cref="ProveedorDto"/>
		/// </returns>
		public static ProveedorDto[] ToDto(Proveedor[] proveedores)
		{
			return Mapper.Map<Proveedor[], ProveedorDto[]>(proveedores);
		}
		
		/// <summary>
		/// Convierte un objeto Producto (persistente) en un objeto ProductoDto.
		/// </summary>
		/// <param name="producto">
		/// A <see cref="Producto"/>
		/// </param>
		/// <returns>
		/// A <see cref="ProductoDto"/>
		/// </returns>
		public static ProductoDto ToDto(Producto producto)
		{
			return Mapper.Map<Producto, ProductoDto>(producto);
		}
		
		/// <summary>
		/// Convierte un array de objetos Producto en un array de objetos ProductoDto.
		/// </summary>
		/// <param name="productos">
		/// A <see cref="Producto"/>
		/// </param>
		/// <returns>
		/// A <see cref="ProductoDto"/>
		/// </returns>
		public static ProductoDto[] ToDto(Producto[] productos)
		{
			return Mapper.Map<Producto[], ProductoDto[]>(productos);
		}
		
		/// <summary>
		/// Convierte un objeto Presupuesto (persistente) en un objeto PresupuestoDto.
		/// </summary>
		/// <param name="presupuesto">
		/// A <see cref="Presupuesto"/>
		/// </param>
		/// <returns>
		/// A <see cref="PresupuestoDto"/>
		/// </returns>
		public static PresupuestoDto ToDto(Presupuesto presupuesto)
		{
			return Mapper.Map<Presupuesto, PresupuestoDto>(presupuesto);
		}
		
		/// <summary>
		/// Convierte un array de objetos Presupuesto en un array de objetos PresupuestoDto.
		/// </summary>
		/// <param name="presupuestos">
		/// A <see cref="Presupuesto"/>
		/// </param>
		/// <returns>
		/// A <see cref="PresupuestoDto"/>
		/// </returns>
		public static PresupuestoDto[] ToDto(Presupuesto[] presupuestos)
		{
			return Mapper.Map<Presupuesto[], PresupuestoDto[]>(presupuestos);
		}
		
		/// <summary>
		/// Convierte un objeto PreFactura (persistente) en un objeto PreFacturaDto.
		/// </summary>
		/// <param name="preFactura">
		/// A <see cref="PreFactura"/>
		/// </param>
		/// <returns>
		/// A <see cref="PreFacturaDto"/>
		/// </returns>
		public static PreFacturaDto ToDto(PreFactura preFactura)
		{
			return Mapper.Map<PreFactura, PreFacturaDto>(preFactura);
		}
		
		/// <summary>
		/// Convierte un array de objetos PreFactura en un array de objetos PreFacturaDto.
		/// </summary>
		/// <param name="preFacturas">
		/// A <see cref="PreFactura"/>
		/// </param>
		/// <returns>
		/// A <see cref="PreFacturaDto"/>
		/// </returns>
		public static PreFacturaDto[] ToDto(PreFactura[] preFacturas)
		{
			return Mapper.Map<PreFactura[], PreFacturaDto[]>(preFacturas);
		}
		
		/// <summary>
		/// Convierte un objeto Factura (persistente) en un objeto FacturaDto.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="Factura"/>
		/// </param>
		/// <returns>
		/// A <see cref="FacturaDto"/>
		/// </returns>
		public static FacturaDto ToDto(Factura factura)
		{
			return Mapper.Map<Factura, FacturaDto>(factura);
		}
		
		/// <summary>
		/// Convierte un array de objetos Factura en un array de objetos FacturaDto.
		/// </summary>
		/// <param name="facturas">
		/// A <see cref="Factura"/>
		/// </param>
		/// <returns>
		/// A <see cref="FacturaDto"/>
		/// </returns>
		public static FacturaDto[] ToDto(Factura[] facturas)
		{
			return Mapper.Map<Factura[], FacturaDto[]>(facturas);
		}
		
		public static NotaCreditoDto ToDto(NotaCredito notaCredito)
		{
			return Mapper.Map<NotaCredito, NotaCreditoDto>(notaCredito);
		}
		
		/// <summary>
		/// Convierte un array de objetos NotaCredito a un array de objetos NotaCreditoDto
		/// </summary>
		/// <param name="notasCredito">
		/// A <see cref="NotaCredito[]"/>
		/// </param>
		/// <returns>
		/// A <see cref="NotaCreditoDto[]"/>
		/// </returns>
		public static NotaCreditoDto[] ToDto(NotaCredito[] notasCredito)
		{
			return Mapper.Map<NotaCredito[], NotaCreditoDto[]>(notasCredito);
		}
		
		public static FacturaGenericaDto ToDto(IFacturaGenerica factura)
		{
			if (factura is Factura)
				return ToDto(factura as Factura);
			else
				return ToDto(factura as PreFactura);
		}

		/// <summary>
		/// Convierte un objeto FacturaCompra (persistente) en un objeto FacturaCompraDto.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="FacturaProveedor"/>
		/// </param>
		/// <returns>
		/// A <see cref="FacturaProveedorDto"/>
		/// </returns>
		public static FacturaCompraDto ToDto(FacturaCompra factura)
		{
			return Mapper.Map<FacturaCompra, FacturaCompraDto>(factura);
		}
		
		/// <summary>
		/// Convierte un array de objetos FacturaProveedor en un array de objetos FacturaProveedorDto.
		/// </summary>
		/// <param name="facturas">
		/// A <see cref="FacturaProveedor"/>
		/// </param>
		/// <returns>
		/// A <see cref="FacturaProveedorDto"/>
		/// </returns>
		public static FacturaCompraDto[] ToDto(FacturaCompra[] facturas)
		{
			return Mapper.Map<FacturaCompra[], FacturaCompraDto[]>(facturas);
		}
		
		/// <summary>
		/// Convierte un objeto Rubro (persistente) en un objeto RubroDto.
		/// </summary>
		/// <param name="rubro">
		/// A <see cref="Rubro"/>
		/// </param>
		/// <returns>
		/// A <see cref="RubroDto"/>
		/// </returns>
		public static RubroDto ToDto(Rubro rubro)
		{
			return Mapper.Map<Rubro, RubroDto>(rubro);
		}
		
		/// <summary>
		/// Convierte un array de objetos Rubro en un array de objetos RubroDto.
		/// </summary>
		/// <param name="rubros">
		/// A <see cref="Rubro"/>
		/// </param>
		/// <returns>
		/// A <see cref="RubroDto"/>
		/// </returns>
		public static RubroDto[] ToDto(Rubro[] rubros)
		{
			return Mapper.Map<Rubro[], RubroDto[]>(rubros);
		}
		
		public static ChequeDto[] ToDto(Cheque[] cheques)
		{
			return Mapper.Map<Cheque[], ChequeDto[]>(cheques);
		}
		
		public static ChequeDto ToDto(Cheque cheque)
		{
			return Mapper.Map<Cheque, ChequeDto>(cheque);
		}
		
		public static ReciboDto ToDto(Recibo recibo)
		{
			return Mapper.Map<Recibo,ReciboDto>(recibo);
		}
		
		public static ReciboDto[] ToDto(Recibo[] recibos)
		{
			return Mapper.Map<Recibo[],ReciboDto[]>(recibos);
		}
		
		/// <summary>
		/// Este es un método genérico para construir objetos DTO.
		/// </summary>
		/// <param name="objeto">
		/// A <see cref="T"/>
		/// </param>
		/// <returns>
		/// A <see cref="R"/>
		/// </returns>
		public static R ToDto<T,R> (T objeto)
		{
			return Mapper.Map<T,R>(objeto);
		}
		
		/// <summary>
		/// Método genérico para construir arrays de objetos DTO.
		/// </summary>
		/// <param name="objeto">
		/// A <see cref="T"/>
		/// </param>
		/// <returns>
		/// A <see cref="R"/>
		/// </returns>
		public static R[] ToDto<T,R> (T[] objeto)
		{
			return Mapper.Map<T[],R[]>(objeto);
		}
		#endregion
		
		#region FromDto
		/// <summary>
		/// Convierte un objeto ClienteDto en un objeto Cliente (persistente).
		/// </summary>
		/// <param name="cliente">
		/// A <see cref="ClienteDto"/>
		/// </param>
		/// <returns>
		/// A <see cref="Cliente"/>
		/// </returns>
		public static Cliente FromDto(ClienteDto cliente)
		{
			// Primero traigo el objeto desde ActiveRecord
			Cliente clienteAlmacenado = ActiveRecordBase<Cliente>.TryFind(cliente.codigo);
			
			// Luego realizo el mapeo, llevando todos los cambios al objeto almacenado
			return Mapper.Map<ClienteDto,Cliente>(cliente, clienteAlmacenado);
		}

		/// <summary>
		/// Convierte un objeto ProveedorDto en un objeto Proveedor (persistente).
		/// </summary>
		/// <param name="proveedor">
		/// A <see cref="ProveedorDto"/>
		/// </param>
		/// <returns>
		/// A <see cref="Proveedor"/>
		/// </returns>
		public static Proveedor FromDto(ProveedorDto proveedor)
		{
			// Primero traigo el objeto desde ActiveRecord
			Proveedor proveedorAlmacenado = ActiveRecordBase<Proveedor>.TryFind(proveedor.codigo);
			
			// Luego realizo el mapeo, llevando todos los cambios al objeto almacenado
			return Mapper.Map<ProveedorDto,Proveedor>(proveedor, proveedorAlmacenado);
		}
		
		/// <summary>
		/// Convierte un objeto ProductoDto en un objeto Producto (persistente).
		/// </summary>
		/// <param name="producto">
		/// A <see cref="ProductoDto"/>
		/// </param>
		/// <returns>
		/// A <see cref="Producto"/>
		/// </returns>
		public static Producto FromDto(ProductoDto producto)
		{
			// Primero traigo el objeto desde ActiveRecord
			Producto productoAlmacenado = ActiveRecordBase<Producto>.TryFind(producto.id);
			
			// Luego realizo el mapeo, llevando todos los cambios al objeto almacenado
			return Mapper.Map<ProductoDto,Producto>(producto, productoAlmacenado);
		}
		
		/// <summary>
		/// Convierte un objeto RubroDto en un objeto Rubro (persistente).
		/// </summary>
		/// <param name="rubro">
		/// A <see cref="RubroDto"/>
		/// </param>
		/// <returns>
		/// A <see cref="Rubro"/>
		/// </returns>
		public static Rubro FromDto(RubroDto rubro)
		{
			// Primero traigo el objeto desde ActiveRecord
			Rubro rubroAlmacenado = ActiveRecordBase<Rubro>.TryFind(rubro.codigo);
			
			// Luego realizo el mapeo, llevando todos los cambios al objeto almacenado
			return Mapper.Map<RubroDto,Rubro>(rubro, rubroAlmacenado);
		}
		
		public static Cheque FromDto(ChequeDto cheque)
		{
			Cheque chequeAlmacenado = ActiveRecordBase<Cheque>.TryFind(cheque.id);
			
			return Mapper.Map<ChequeDto,Cheque>(cheque, chequeAlmacenado);
		}
		
		/// <summary>
		/// Permite pasar un objeto PreFactura a uno Factura. Se utiliza cuando se imprime una
		/// PreFactura, y ésta se debe almacenar como Factura si la impresión es correcta.
		/// </summary>
		/// <param name="preFatura">
		/// A <see cref="PreFactura"/>
		/// </param>
		/// <returns>
		/// A <see cref="Factura"/>
		/// </returns>
		public static Factura ToFactura(PreFactura preFatura)
		{
			return Mapper.Map<PreFactura, Factura>(preFatura);
		}
		
		public static PreFactura ToPreFactura(Factura factura)
		{
			return Mapper.Map<Factura, PreFactura>(factura);
		}

		public static NotaCredito ToNotaCredito(PreNotaCredito preNotaCredito)
		{
			return Mapper.Map<PreNotaCredito, NotaCredito>(preNotaCredito);
		}
		
		/// <summary>
		/// Premite pasar un documento (Factura, PreFactura, Presupuesto, etc) DTO a un
		/// objeto ActiveRecordBase.
		/// </summary>
		/// <param name="documento">
		/// A <see cref="T"/>
		/// </param>
		/// <returns>
		/// A <see cref="R"/>
		/// </returns>
		public static R FromDto<T,R> (T documento)
			where R : IDocumentoGenerico
			where T : DocumentoGenericoDto
		{
			// Primero traigo el objeto desde ActiveRecord
			R documentoAlmacenado = ActiveRecordBase<R>.TryFind(documento.id);
			
			// Luego realizo el mapeo, llevando todos los cambios al objeto almacenado
			return Mapper.Map<T,R>(documento, documentoAlmacenado);
		}
		
		/// <summary>
		/// Permite pasar un array de documentos DTO a uno de objetos ActiveRecordBase.
		/// </summary>
		/// <param name="documentos">
		/// A <see cref="T"/>
		/// </param>
		/// <returns>
		/// A <see cref="R"/>
		/// </returns>
		public static R[] FromDto<T,R> (T[] documentos)
			where R : IDocumentoGenerico
			where T : DocumentoGenericoDto
		{
			R[] documentosAlmacenados = new R[documentos.Length];
			
			for (int i=0; i<documentos.Length; i++)
				documentosAlmacenados[i] = FromDto<T, R>(documentos[i]);
			
			return documentosAlmacenados;
		}
		#endregion
	}
}
