// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;

using Castle.ActiveRecord;
using Castle.Components.Validator;

using Stratos.Common;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Representa un documento de venta que tiene importes, bonificaciones e
	/// items de productos.
	/// </summary>
	public abstract class DocumentoVenta<T> : DocumentoCompraventa<T>, IDocumentoVenta
		where T : class
	{
		#region Campos
		// Importe igual a la suma de los importes de los items
		protected decimal importeBruto;
		
		protected decimal porcentajeBonificacion;
		protected decimal importeBonificacion;
		
		// Resta entre importe bruto y bonificación
		protected decimal importeNeto;
		#endregion
		
		#region Constructores
		public DocumentoVenta () : base()
		{
		}
		#endregion
		
		#region Propiedades
		public virtual Cliente Cliente {
			get { return (Cliente)base.Persona; }
			set { base.Persona = value; }
		}
		
		public override int LimiteItems {
			get { return 32; }
		}
		
		[Property(NotNull=true)]
		public virtual decimal ImporteBruto {
    		get { return this.importeBruto; }
			set { this.importeBruto = StratosUtil.RoundDecimal(value, 2); }
    	}
		
		[Property(NotNull=true)]
		[ValidateRange(RangeValidationType.Decimal, -50, 50,
		               "El porcentaje de bonificación es inválido. No puede ser superior al (+/-) 50%")]
		public virtual decimal PorcentajeBonificacion {
			get { return this.porcentajeBonificacion; }
			set { this.porcentajeBonificacion = StratosUtil.RoundDecimal(value, 2); }
		}
		
		[Property(NotNull=true)]
		public virtual decimal ImporteBonificacion {
			get { return this.importeBonificacion; }
			set { this.importeBonificacion = StratosUtil.RoundDecimal(value, 2); }
		}
		
		[Property(NotNull=true)]
		public virtual decimal ImporteNeto {
			get { return this.importeNeto; }
			set { this.importeNeto = StratosUtil.RoundDecimal(value, 2); }
		}
		#endregion
		
		#region Métodos de validación
		/// <summary>
		/// Valida que los importes del DocumentoGenerico sean correctos.
		/// </summary>
		[ValidateSelf]
		public virtual void ValidarImportes(ErrorSummary errorSummary)
		{
			// Valido el Importe Bruto
			if (this.importeBruto <= 0.00m)
				errorSummary.RegisterErrorMessage("ImporteBruto",
				                                  "El Importe bruto no puede ser menor o igual a cero");
			
			// Sumo los importes de los items
			decimal sumaItems = 0.00m;
			
			// Sumo todos los items
			foreach (IItemProducto i in this.ItemsGenericos)
				sumaItems += i.Importe;
			
			if (this.importeBruto != StratosUtil.RoundDecimal(sumaItems, 2))
				errorSummary
					.RegisterErrorMessage("ImporteBruto",
					                      "El Importe bruto es incorrecto");
			
			// Valido el Importe de Bonificación
			if (this.importeBonificacion !=
			    	StratosUtil.RoundDecimal(this.importeBruto * (this.porcentajeBonificacion / 100m), 2))
				errorSummary
					.RegisterErrorMessage("ImporteBonificacion",
					                      "El Importe de bonificación es incorrecto");
			
			// Valido el Importe Neto
			if (this.importeNeto != (this.importeBruto - this.importeBonificacion))
				errorSummary
					.RegisterErrorMessage("ImporteNeto",
					                      "El Importe neto es incorrecto");
			
			// El importe neto no puede superar el millón de pesos
			if (this.importeBruto >= 1000000.0m)
				errorSummary
					.RegisterErrorMessage("ImporteBruto",
					                      "El importe bruto no puede superar el millón de pesos. " +
					                      "Divida el documento.");
		}
		#endregion
	}
}
