// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Text;
using System.Collections;
using System.Collections.Generic;

using NHibernate.Criterion;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Queries;
using Castle.Components.Validator;

using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Esta clase representa una Factura común que ha sido impresa.
	/// </summary>
	[ActiveRecord(DiscriminatorColumn="notacredito",
	              DiscriminatorType="Int32",
	              DiscriminatorValue="0")]
	public class Factura : FacturaGenerica<Factura>, IFactura
	{
		#region Campos
		protected long numero;
		
		// Los datos del cliente son fijos en la factura
		protected string clienteRazonSocial;
		protected TipoResponsable clienteTipoResponsable;
		protected long? clienteCuit;
		
		protected bool valida;
		
		protected IList<ItemFactura> items;
		
		protected bool anulada;
		protected bool facturaManual;
		
		protected bool historica;
		
		protected IList<PreNotaCredito> preNotasCreditoAsociadas;
		protected IList<NotaCredito> notasCreditoAsociadas;
		#endregion
		
		#region Constructores
		public Factura() : base()
		{
			// Por defecto la factura no es histórica
			this.historica = false;
			
			this.items = new List<ItemFactura>();
			
			this.valida = true;
			this.anulada = false;
			this.facturaManual = false;
			
			this.preNotasCreditoAsociadas = new List<PreNotaCredito>();
			this.notasCreditoAsociadas = new List<NotaCredito>();
		}
		#endregion
		
		#region Propiedades nuevas
		[Property(UniqueKey="facturaUnica", NotNull=true)]
    	public override FacturaTipoVenta Tipo {
    		get { return base.Tipo; }
    		set { base.Tipo = value; }
    	}
		
		[Property(UniqueKey="facturaUnica", NotNull=true)]
		[ValidateRange(1, 9999, "El código de la factura es invalido")]
    	public override int Codigo {
    		get { return this.codigo; }
    		set { this.codigo = value; }
    	}
		
		[Property(UniqueKey="facturaUnica", NotNull=true)]
		[ValidateRange(RangeValidationType.Decimal, 1, long.MaxValue,
		               "El número de la factura es invalido")]
		public long Numero {
    		get { return this.numero; }
    		set { this.numero = value; }
    	}
		
		//// <value>
		/// Esta propiedad de la Factura indica si la misma es una Factura válida. Esto se utiliza
		/// únicamente cuando se anula una factura para sincronizar los números de la misma. Por
		/// ejemplo, si se rompe una Factura real al enviarla a imprimir, entonces se debe reimprimir
		/// la misma, pero en el sistema se da de alta una Factura inválida y se anula para lograr
		/// esta sincronización.
		/// </value>
		[Property(NotNull=true)]
    	public bool Valida {
    		get { return this.valida; }
    		set { this.valida = value; }
    	}
		
		[Property(NotNull=true)]
		public bool Anulada {
			get { return this.anulada; }
			set { this.anulada = value; }
		}
		
		[Property(NotNull=true)]
		public bool FacturaManual {
			get { return this.facturaManual; }
			set { this.facturaManual = value; }
		}
		
		/// <summary>
		/// Esta propiedad indica si la factura puede anularse o no.
		/// Cuando la factura fue utilizada para generar el importe
		/// IVA Ventas ya no puede ser anulada, entonces esta
		/// propiedad debe ser 'true'.
		/// </summary>
		[Property(NotNull=true)]
		public bool Historica {
			get { return this.historica; }
			set { this.historica = value; }
		}
		
		[HasMany(Inverse=true, Cascade=ManyRelationCascadeEnum.Delete)]
		public IList<PreNotaCredito> PreNotasCreditoAsociadas {
			get { return this.preNotasCreditoAsociadas; }
			set { this.preNotasCreditoAsociadas = value; }
		}
		
		[HasMany(Inverse=true, Cascade=ManyRelationCascadeEnum.Delete)]
		public IList<NotaCredito> NotasCreditoAsociadas {
			get { return this.notasCreditoAsociadas; }
			set { this.notasCreditoAsociadas = value; }
		}
		
		public virtual string TipoAbreviado {
			get { return "FA"; }
		}
		#endregion
		
		#region Propiedades sobreescritas
		[PrimaryKey(typeof(TableHiLoGeneratorWithMultipleColumns))]
		public override long Id {
			get { return base.Id; }
			set { base.Id = value; }
		}
		
		#region Cliente
		[Property(NotNull=true)]
		public string ClienteRazonSocial {
			get { return this.clienteRazonSocial; }
			set { this.clienteRazonSocial = value; }
		}
		
		[Property(NotNull=true)]
		public TipoResponsable ClienteTipoResponsable {
			get { return this.clienteTipoResponsable; }
			set { this.clienteTipoResponsable = value; }
		}
		
		[Property(NotNull=false)]
		public long? ClienteCuit {
			get { return this.clienteCuit; }
			set { this.clienteCuit = value; }
		}
		#endregion
		
		[HasMany(Table="ItemFactura",
		         ColumnKey="FacturaId",
		         Lazy=true,
		         Cascade=ManyRelationCascadeEnum.AllDeleteOrphan)]
		public IList<ItemFactura> Items {
			get { return this.items; }
			set { this.items = value; }
		}
		
		public override IList ItemsGenericos {
			get { return (IList) this.items; }
			set { this.items = (IList<ItemFactura>) value; }
		}
		#endregion
		
		#region Métodos validadores
		[ValidateSelf(ExecutionOrder=10, RunWhen=RunWhen.Insert)]
		public virtual void ValidarClavePrimariaUnica(ErrorSummary errorSummary)
		{
			using (SessionScope session = new SessionScope(FlushAction.Never)) {
				/* Valido únicamente si no estoy actualizando el objeto (ya que se encontraría
				 * él mismo en la base de datos).
				 * Realizo una consulta para saber si ya existe en la base de datos otra Factura del
				 * mismo tipo, código y número. */
				CountQuery consultaCount =
					new CountQuery(typeof(Factura),
					               @"Tipo = :tipo
										and Codigo = :codigo
										and Numero = :numero");
				
				consultaCount.SetParameter("tipo", this.Tipo);
				consultaCount.SetParameter("codigo", this.Codigo);
				consultaCount.SetParameter("numero", this.Numero);
				
				int count = (int) ActiveRecordMediator.ExecuteQuery(consultaCount);
				
				// Si exsten, entonces invalido la propiedad Numero
				if (count > 0){
					string facturaTipo = this.Tipo == FacturaTipoVenta.FacturaA ? "A" : "B";
					
					errorSummary
						.RegisterErrorMessage("Numero",
						                      "Ya existe una factura de tipo '" + facturaTipo +
						                      "' con código '" + this.Codigo +
						                      "' con el número '" + this.Numero + "'");
				}
			}
		}
		
		[ValidateSelf(RunWhen=RunWhen.Update)]
		public virtual void ValidarFacturaHistoricaInmodificable(ErrorSummary errorSummary)
		{
			// Uso una sessión nueva para obtener los datos de la base de datos no modificados
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				// Obtengo esta misma factura, pero la guardada en la base de datos
				Factura facturaGuardada = Factura.TryFind(this.Id);
				
				if (facturaGuardada != null && facturaGuardada.Historica)
					errorSummary
						.RegisterErrorMessage("Historica",
						                      "La factura no se puede modificar porque es una factura histórica.");
			}
		}
		
		[ValidateSelf(ExecutionOrder=20, RunWhen=RunWhen.Insert)]
		public virtual void ValidarNumeroConsecutivo(ErrorSummary errorSummary)
		{
			// Si se trata de una Factura de sincronización, no valido el número consecutivo
			if (!this.Valida)
				return;
			
			using (SessionScope session = new SessionScope(FlushAction.Never)) {
				/* Consulto la última Factura creada del mismo tipo y código que la instancia actual,
				 * para conocer cuál es su número. Esta factura que se está creando debe tener
				 * como número el consecutivo al mismo. */
				ScalarQuery<Nullable<long>> consultaUltimoNumero =
					new ScalarQuery<Nullable<long>>(typeof(Factura),
					                          @"select max(f.Numero)
												from Factura f
												where f.Tipo = :tipo
													and f.Codigo = :codigo");
				
				consultaUltimoNumero.SetParameter("tipo", this.Tipo);
				consultaUltimoNumero.SetParameter("codigo", this.Codigo);
				
				long? ultimoNumero = consultaUltimoNumero.Execute();
				
				// Realizo la validación solo si la consulta anterior obtuvo resultados
				if (ultimoNumero.HasValue && this.Numero != ultimoNumero + 1) {
					string facturaTipo = this.Tipo == FacturaTipoVenta.FacturaA ? "A" : "B";
					
					errorSummary
						.RegisterErrorMessage("Numero",
						                      "El número de la Factura no es consecutivo con la última " +
						                      "Factura creada del mismo tipo y código. El último número " +
						                      "de Factura '" + facturaTipo + "' con código '" + this.Codigo +
						                      "' es '" + ultimoNumero + "'");
				}
			}
		}
		
		public override void ValidarFecha (ErrorSummary errorSummary)
		{
			// Si se trata de una Factura de sincronización, no valido su fecha
			if (!this.Valida)
				return;
			
			DateTime? fechaMinima, fechaMaxima;
			
			using (SessionScope session = new SessionScope(FlushAction.Never)) {
				/* Consulta la fecha mínima permitida: corresponde a la factura del mismo tipo y
				 * código, con un número inferior a la actual y que no está anulada. */
				ScalarQuery<DateTime?> consultaFechaMinima =
					new ScalarQuery<DateTime?>(typeof(Factura),
					                          @"select max(f.Fecha)
												from Factura f
												where f.Tipo = :tipo
													and f.Codigo = :codigo
													and f.Numero < :numero
													and f.Anulada = cast(false,boolean)");
				
				// Seteo los parámetros de la consulta
				consultaFechaMinima.SetParameter("tipo", this.Tipo);
				consultaFechaMinima.SetParameter("codigo", this.Codigo);
				consultaFechaMinima.SetParameter("numero", this.Numero);
				
				/* Consulta la fecha máxima permitida: corresponde a la factura del
				 * mismo tipo y código con un número superior a la actual y que no está
				 * anulada. */
				ScalarQuery<DateTime?> consultaFechaMaxima =
					new ScalarQuery<DateTime?>(typeof(Factura),
					                          @"select min(f.Fecha)
												from Factura f
												where f.Tipo = :tipo
													and f.Codigo = :codigo
													and f.Numero > :numero
													and f.Anulada = cast(false,boolean)");
				
				// Seteo los parámetros de la consulta
				consultaFechaMaxima.SetParameter("tipo", this.Tipo);
				consultaFechaMaxima.SetParameter("codigo", this.Codigo);
				consultaFechaMaxima.SetParameter("numero", this.Numero);
				
				fechaMinima = consultaFechaMinima.Execute();
				fechaMaxima = consultaFechaMaxima.Execute();
			}
			
			/* Si no hay fecha mínima (porque no hay facturas anteriores por ejemplo), por defecto
			 * la misma será igual a un mes anterior a la fecha actual. */
			if (!fechaMinima.HasValue)
				fechaMinima = DateTime.Today - TimeSpan.FromDays(30);
			
			/* Si no hay fecha máxima (porque no hay factuas posteriores), por defecto la misma
			 * es el día de hoy. */
			if (!fechaMaxima.HasValue)
				fechaMaxima = DateTime.Today;
			
			if (this.Fecha.Date < fechaMinima.Value.Date) {
			    errorSummary
					.RegisterErrorMessage("Fecha", "La fecha de la Factura es inválida. No puede ser " +
					                      "inferior a la fecha de la Factura anterior del mismo código y tipo " +
					                      "(" + fechaMinima.Value.ToShortDateString() + ")");
			}
			else if (this.Fecha.Date > fechaMaxima.Value.Date)
				errorSummary
					.RegisterErrorMessage("Fecha", "La fecha de la Factura es inválida. No puede ser " +
					                      "superior ni a la fecha de la Factura siguiente del mismo código y tipo, ni al día de hoy.");
		}
		#endregion
		
		#region Eventos internos
		/// <summary>
		/// Se ejecuta cuando la factura ha sido anulada y antes
		/// de guardarse estos cambios en la base de datos.
		/// </summary>
		protected virtual void OnFacturaAnulada()
		{
			// Elimino el débito asociado a la factura.
			DocumentoDebito.EliminarDebito(this);
		}
		
		/// <summary>
		/// Se ejecuta cuando la factura ha sido desanulada y antes
		/// de guardarse estos cambios en la base de datos.
		/// </summary>
		protected virtual void OnFacturaDesanulada()
		{
			// Creo un documento de débito asociado a la factura.
			DocumentoDebito.CrearDebito(this);
		}
		
		/// <summary>
		/// Se ejecuta cuando la factura está a punto de crearse.
		/// </summary>
		protected virtual void OnFacturaCreada()
		{
			// Si la condición de pago de la factura es distinta de
			// contado, entonces creo un documento de débito asociado
			// a la misma y al cliente.
			if (this.CondicionVenta.Condicion != CondicionPago.Contado)
				DocumentoDebito.CrearDebito(this);
		}
		
		/// <summary>
		/// Se ejecuta cuando la factura está a punto de eliminarse.
		/// </summary>
		protected virtual void OnFacturaEliminada()
		{
			// Elimino el débito asociado a esta factura que se está por eliminar
			DocumentoDebito.EliminarDebito(this);
		}
		#endregion
		
		#region Métodos de ActiveRecord
		protected override void BeforeDelete (IDictionary adapter)
		{
			base.BeforeDelete (adapter);
			
			// Una factura histórica no puede ser eliminada
			bool facturaHistorica = (bool) adapter["Historica"];
			
			if (facturaHistorica)
				throw new Stratos.Common.Excepciones.
					FacturaImposibleDeEliminarException("Una factura histórica no puede ser eliminada");
			
			// Disparo el evento
			this.OnFacturaEliminada();
		}
		
		protected override bool BeforeSave (IDictionary state)
		{
			bool retval = base.BeforeSave(state);
			
			// Copio los datos fijos en la factura: la razón social del cliente,
			// el tipo de responsabilidad del mismo, y el cuit. Si éstos cambian
			// luego, se matienen fijos en la factura.
			Cliente clienteAsignado = (Cliente) state["Cliente"];
			
			if (clienteAsignado != null) {
				state["ClienteRazonSocial"] = clienteAsignado.RazonSocial;
				state["ClienteTipoResponsable"] = clienteAsignado.TipoResponsable;
				
				if (clienteAsignado.Cuit != null)
					state["ClienteCuit"] = clienteAsignado.Cuit.ValorNumerico;
			}
			
			// Disparo el evento
			this.OnFacturaCreada();
			
			return retval;
		}
		
		protected override bool OnFlushDirty (object id, IDictionary previousState, IDictionary currentState, NHibernate.Type.IType[] types)
		{
			bool retval = base.OnFlushDirty (id, previousState, currentState, types);
			
			// Me fijo si la propiedad Anulada ha cambiado
			if (previousState["Anulada"] != currentState["Anulada"]) {
			    
				// Disparo el evento correspondiente dependiendo de si la factura
				// está ahora anulada o no.
				if ((bool) currentState["Anulada"])
					this.OnFacturaAnulada();
				else
					this.OnFacturaDesanulada();
			}
			
			return retval;
		}
		#endregion
    }
}
