// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Castle.ActiveRecord;
using Castle.ActiveRecord.Queries;
using Castle.Components.Validator;
using NHibernate.Criterion;

using Stratos.Common;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	[ActiveRecord]
	public class FacturaCompra : DocumentoCompra<FacturaCompra>, IFacturaCompra
	{
		#region Campos
		protected static bool migrandoDatos = false;
		protected FacturaTipoCompra tipo;
		
		protected int codigo;		
		protected long numero;
		
		protected DateTime fechaImputacion;
		
		protected IList<ItemFacturaCompra> items;
		
		protected decimal importeGravado;
		protected decimal importeExento;
		protected decimal importeIva;
		
		protected decimal percepcionIB;
		protected decimal percepcionIva;
		protected decimal impuestoCombustible;
		
		protected decimal importeTotal;
		
		// Los datos del proveedor son fijos en la factura
		protected string proveedorRazonSocial;
		protected TipoResponsable proveedorTipoResponsable;
		protected long proveedorCuit;
		
		protected bool historica;
		#endregion
		
		#region Constructores
		public FacturaCompra () : base ()
		{
			this.items = new List<ItemFacturaCompra>();
			this.historica = false;
		}
		#endregion
		
		#region Propiedades
		[PrimaryKey(typeof(TableHiLoGeneratorWithMultipleColumns))]
		public override long Id {
			get { return base.Id; }
			set { base.Id = value; }
		}
		
		[Property(NotNull=true)]
		public virtual FacturaTipoCompra Tipo {
			get { return this.tipo; }
			set { this.tipo = value; }
		}
		
		[Property(UniqueKey="facturaCompraUnica", NotNull=true)]
		[ValidateRange(0, 9999, "El código de la factura es inválido")]
		public virtual int Codigo {
			get { return this.codigo; }
			set { this.codigo = value; }
		}
		
		[Property(UniqueKey="facturaCompraUnica", NotNull=true)]
		[ValidateRange(RangeValidationType.Decimal, 1, long.MaxValue,
		               "El número de la factura es invalido")]
    	public virtual long Numero {
    		get { return this.numero; }
    		set { this.numero = value; }
    	}
		
		#region Proveedor
		[BelongsTo("proveedorId",NotNull=false)]
		public override Proveedor Proveedor {
			get { return base.Proveedor; }
			set {
				base.Proveedor = value;
				
				// Asigno los datos fijos del proveedor
				if (this.Proveedor != null) {
					this.ProveedorRazonSocial = this.Proveedor.RazonSocial;
					this.ProveedorTipoResponsable = this.Proveedor.TipoResponsable;
					
					if (this.Proveedor.Cuit != null && this.Proveedor.Cuit.ValorNumerico.HasValue)
						this.ProveedorCuit = this.Proveedor.Cuit.ValorNumerico.Value;
				}
			}
		}
		
		[Property(NotNull=true)]
		public string ProveedorRazonSocial {
			get { return this.proveedorRazonSocial; }
			set { this.proveedorRazonSocial = value; }
		}
		
		[Property(NotNull=true)]
		public TipoResponsable ProveedorTipoResponsable {
			get { return this.proveedorTipoResponsable; }
			set { this.proveedorTipoResponsable = value; }
		}
		
		[Property(UniqueKey="facturaCompraUnica", NotNull=true)]
		public long ProveedorCuit {
			get { return this.proveedorCuit; }
			set { this.proveedorCuit = value; }
		}
		#endregion
		
		[HasMany(Table="ItemFacturaCompra",
		         ColumnKey="FacturaCompraId",
		         Lazy=true,
		         Cascade=ManyRelationCascadeEnum.AllDeleteOrphan)]
		public IList<ItemFacturaCompra> Items {
			get { return this.items; }
			set { this.items = value; }
		}
		
		public override IList ItemsGenericos {
			get { return (IList) this.items; }
			set { this.items = (IList<ItemFacturaCompra>) value; }
		}
		
		[Property(NotNull=true)] 
		public override DateTime Fecha {
			get { return base.Fecha; }
			set { base.Fecha = value; }
		}
		
		[Property(NotNull=true)] 
		public DateTime FechaImputacion {
			get { return this.fechaImputacion; }
			set { this.fechaImputacion = value; }
		}
		
		[Property(NotNull=true)]
		public decimal PercepcionIva {
			get { return this.percepcionIva; }
			set { this.percepcionIva = value; }
		}
		
		[Property(NotNull=true)]
		public decimal ImporteIva {
			get { return this.importeIva; }
			set { this.importeIva = value; }
		}
		
		[Property(NotNull=true)]
		public decimal PercepcionIB {
			get { return this.percepcionIB; }
			set { this.percepcionIB = value;}
		}
		
		[Property(NotNull=true)]
		public decimal ImpuestoCombustible {
			get { return this.impuestoCombustible; }
			set { this.impuestoCombustible = value;}
		}
		
		[Property(NotNull=true)]
		public decimal ImporteGravado {
			get { return this.importeGravado; }
			set { this.importeGravado = value; }
		}
		
		[Property(NotNull=true)]
		public decimal ImporteExento {
			get { return this.importeExento; }
			set { this.importeExento = value; }
		}
		
		[Property(NotNull=true)]
		public decimal ImporteTotal {
			get { return this.importeTotal; }
			set { this.importeTotal = value; }
		}
		
		public override decimal ImporteFinal {
			get { return this.importeTotal; }
		}
		
		public string TipoAbreviado {
			get { return "FA"; }
		}
		
		[Property(NotNull=true)]
		public bool Historica {
			get { return this.historica; }
			set { this.historica = value; }
		}
		
		public override int LimiteItems {
			get { return 50; }
		}
		#endregion
		
		#region Métodos Validadores
		[ValidateSelf(ExecutionOrder=10)]
		public virtual void ValidarClavePrimariaUnica(ErrorSummary errorSummary)
		{
			using (SessionScope session = new SessionScope(FlushAction.Never)) {
				/* Realizo una consulta para saber si ya existe en la base de datos otra Factura del
				 * mismo proveedor, código y número. */
				CountQuery consultaCount =
					new CountQuery(typeof(FacturaCompra),
					               @"Codigo = :codigo
										and Numero = :numero
										and ProveedorCuit = :cuitProveedor
										and Id <> :idFactura");
				
				consultaCount.SetParameter("codigo", this.Codigo);
				consultaCount.SetParameter("numero", this.Numero);
				consultaCount.SetParameter("cuitProveedor", this.ProveedorCuit);
				consultaCount.SetParameter("idFactura", this.Id);
				
				int count = (int) ActiveRecordMediator.ExecuteQuery(consultaCount);
				
				// Si existen, entonces invalido la propiedad Numero
				if (count > 0)
					errorSummary
						.RegisterErrorMessage("Numero",
						                      "Ya existe una factura del proveedor con cuit '" + this.ProveedorCuit +
						                      "' con código '" + this.Codigo +
						                      "' con el número '" + this.Numero + "'");
			}
		}
		
		/// <summary>
		/// Valida que el proveedor tenga un IVA consistente con el tipo de factura.
		/// </summary>
		[ValidateSelf]
		public virtual void ValidarProveedor(ErrorSummary errorSummary)
		{
			if (String.IsNullOrEmpty(this.ProveedorRazonSocial))
				errorSummary
						.RegisterErrorMessage("ProveedorRazonSocial",
						                      "El Proveedor de una factura de compra debe tener " +
						                      "asignada la razón social.");
			
			if (this.ProveedorCuit <= 0)
				errorSummary
						.RegisterErrorMessage("ProveedorCuit",
						                      "El Proveedor de una factura de compra debe tener " +
						                      "asignado un número de Cuit.");
			
			// Factura A
			if (this.Tipo == FacturaTipoCompra.FacturaA) {
				if (this.ProveedorTipoResponsable == TipoResponsable.ConsumidorFinal ||
				    	this.ProveedorTipoResponsable == TipoResponsable.Exento ||
				    	this.ProveedorTipoResponsable == TipoResponsable.Monotributo)
					errorSummary
						.RegisterErrorMessage("ProveedorTipoResponsable",
						                      "Una Factura A sólo puede haber sido emitida " +
						                      "por un vendedor " +
						                      "Responsable Inscripto o Responsable No " +
						                      "Inscripto");
			}
			
			// Factura C
			else if (this.Tipo == FacturaTipoCompra.FacturaC) {
				if (this.ProveedorTipoResponsable != TipoResponsable.Monotributo)
					errorSummary
						.RegisterErrorMessage("ProveedorTipoResponsable",
						                      "Una Factura C sólo puede haber sido emitida " +
						                      "por un vendedor Monotributo");
			}
		}
		
		/// <summary>
		/// Valida que los importes de la factura sean correctos.
		/// </summary>
		[ValidateSelf]
		public void ValidarImportes(ErrorSummary errorSummary)
		{
			if (migrandoDatos)
				return;
			
			// Valido el Gravado
			if (this.ImporteGravado <= 0.00m)
				errorSummary.RegisterErrorMessage("ImporteGravado",
				                                  "El importe 'Gravado' es incorrecto. " +
				                                  "Debe ser mayor a cero.");
			
			// Valido si es menor o igual a cero
			if (this.ImporteIva <= 0.00m)
				errorSummary.RegisterErrorMessage("ImporteIva",
				                                  "El Importe de IVA no puede ser igual a cero");
			
			// Valido el Exento
			if (this.ImporteExento < 0.00m)
				errorSummary.RegisterErrorMessage("ImporteExento",
				                                  "El importe 'Exento' es incorrecto. " +
				                                  "Debe ser mayor a cero.");
			
			// Valido PercepcionIB
			if (this.PercepcionIB < 0.00m)
				errorSummary.RegisterErrorMessage("PercepcionIB",
				                                  "El importe 'PercepcionIB' es incorrecto. " +
				                                  "Debe ser mayor a cero.");
			
			// Valido PercepcionIVA
			if (this.PercepcionIva < 0.00m)
				errorSummary.RegisterErrorMessage("PercepcionIva",
				                                  "El importe 'PercepcionIva' es incorrecto. " +
				                                  "Debe ser mayor a cero.");
			// Valido ImpuestoCombustible
			if (this.ImpuestoCombustible < 0.00m)
				errorSummary.RegisterErrorMessage("ImpuestoCombustible",
				                                  "El importe 'ImpuestoCombustible' es incorrecto. " +
				                                  "Debe ser mayor a cero.");
			
			decimal totalCalculado = this.ImporteGravado + this.ImporteExento + this.ImporteIva + this.PercepcionIB +
				this.PercepcionIva + this.ImpuestoCombustible;
			
			// Valido ImporteTotal
			if (this.ImporteTotal != StratosUtil.RoundDecimal(totalCalculado, 2))
				errorSummary.RegisterErrorMessage("ImporteTotal", "El Importe total es incorrecto");
		}

		/// <summary>
		/// Valida la Fecha de creación de la factura.
		/// </summary>
		[ValidateSelf]
		public void ValidarFecha(ErrorSummary errorSummary)
		{
			if (migrandoDatos)
				return;
			
			// FIXME: Ver cual es la fecha mínima correcta. Por ahora lo dejo en
			// 60 días anteriores al día de hoy.
//			DateTime fechaMinima = DateTime.Today.AddDays(-60);
			DateTime fechaMaxima = DateTime.Today;
			
//			if (this.Fecha.Date < fechaMinima.Date)
//				errorSummary
//					.RegisterErrorMessage("Fecha", "La Fecha de la factura es inválida." +
//							" Debe ser ingresada a lo sumo 60 días después de su creación.");
			
			if (this.Fecha.Date > fechaMaxima.Date)
				errorSummary
					.RegisterErrorMessage("Fecha", "La Fecha de la factura es inválida. No puede ser " +
					                      "posterior a la fecha actual.");
		}
		
		/// <summary>
		/// Valida la fecha de imputación de la factura.
		/// </summary>
		[ValidateSelf]
		public void ValidarFechaImputacion(ErrorSummary errorSummary)
		{
			if (migrandoDatos)
				return;
			
			DateTime fechaMinima = this.Fecha;
			DateTime fechaMaxima = DateTime.Today;
			
			if (this.FechaImputacion.Date < fechaMinima.Date)
				errorSummary
					.RegisterErrorMessage("FechaImputacion", "La Fecha de Imputacion de la factura es inválida." +
							"No puede ser anterior a la fecha de comprobante de la factura.");
			
			else if (this.FechaImputacion.Date > fechaMaxima.Date)
				errorSummary
					.RegisterErrorMessage("FechaImputacion", "La Fecha de Imputacion de la factura es inválida." +
							"No puede ser posterior a la fecha actual.");
		}
		
		[ValidateSelf(RunWhen=RunWhen.Update)]
		public virtual void ValidarFacturaHistoricaInmodificable(ErrorSummary errorSummary)
		{
			// Uso una sessión nueva para obtener los datos de la base de datos no modificados
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				// Obtengo esta misma factura, pero la guardada en la base de datos
				FacturaCompra facturaGuardada = FacturaCompra.TryFind(this.Id);
				
				if (facturaGuardada != null && facturaGuardada.Historica)
					errorSummary
						.RegisterErrorMessage("Historica",
						                      "La factura no se puede modificar porque es una factura histórica.");
			}
		}
		#endregion
		
		#region Métodos de migración desde sistema viejo
		/// <summary>
		/// Este método activa el modo de migración de facturas del sistema viejo
		/// al sistema nuevo. Este modo hace que se pueda crear facturas con items donde el precio
		/// unitario del mismo difiere al precio del producto.
		/// Ver método OnNotValid()
		/// </summary>
		public static void ModoMigracion()
		{
			migrandoDatos = true;
		}
		
		/// <summary>
		/// Vuelve al modo normal, y sí valida el precio unitario de los items de las facturas.
		/// </summary>
		public static void ModoNormal()
		{
			migrandoDatos = false;
		}
		#endregion
		
		#region Métodos de ActiveRecord
		/// <summary>
		/// Setea las propiedades ProveedorRazonSocial, ProveedorTipoResponsable y
		/// ProveedorCuit justo antes de guardarse en la base de datos.
		/// </summary>
		protected override bool BeforeSave (IDictionary state)
		{
			bool retval = base.BeforeSave(state);
			
			// Copio los datos fijos en la factura: la razón social del proveedor,
			// el tipo de responsabilidad del mismo, y el cuit. Si éstos cambian
			// luego, se matienen fijos en la factura.
			Proveedor proveedorAsignado = (Proveedor) state["Proveedor"];
			
			if (proveedorAsignado != null && proveedorAsignado.Codigo > 0) {
				state["ProveedorRazonSocial"] = proveedorAsignado.RazonSocial;
				state["ProveedorTipoResponsable"] = proveedorAsignado.TipoResponsable;
				
				if (proveedorAsignado.Cuit != null && proveedorAsignado.Cuit.ValorNumerico.HasValue)
					state["ProveedorCuit"] = proveedorAsignado.Cuit.ValorNumerico.Value;
			}
			
			return true;
		}
		
		/// <summary>
		/// Antes de borrar una factura de la base de datos, chequea que esta no esté marcada
		/// como histórica. En caso de estarlo, lanza una excepción.
		/// </summary>
		protected override void BeforeDelete (IDictionary adapter)
		{
			// Una factura histórica no puede ser eliminada
			bool facturaHistorica = (bool) adapter["Historica"];
			
			// Si la factura es Histórica, lanzo una exepción
			if (facturaHistorica)
				throw new Stratos.Common.Excepciones.
					FacturaImposibleDeEliminarException("Una factura histórica no puede ser eliminada");
		}
		#endregion
	}
}
