// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Reflection;
using System.Diagnostics;
using System.Collections;
using System.Collections.Generic;

using Castle.ActiveRecord;
using Castle.Components.Validator;
using Castle.ActiveRecord.Queries;

using Stratos.Common;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Factura genérica que posee todos los campos y métodos comunes de una Factura y
	/// una PreFactura. No genera una tabla en la base de datos.
	/// </summary>
	public abstract class FacturaGenerica<T> : DocumentoVenta<T>, IFacturaGenerica
		where T : class
	{
		#region Campos
		protected static bool migrandoDatos = false;
		
		protected int codigo;
		private FacturaTipoVenta tipo;
		
		private CondicionVentaFacturaGenerica condicionVenta;
		
		// FIXME: Por ahora usamos un string para los remitos
		//protected IList<Remito> remitos; TODO:Para usar cuando implementémos remitos
		private string remitos;
		
    	private decimal importeIva;
		private decimal importeTotal;
		#endregion
		
		#region Constructores
		public FacturaGenerica()
		{
			this.condicionVenta = new CondicionVentaFacturaGenerica();
			this.tipo = FacturaTipoVenta.FacturaA;
		}
		#endregion
		
		#region Propiedades nuevas
		public abstract int Codigo {
    		get;
    		set;
    	}
		
		[Nested]
		public CondicionVentaFacturaGenerica CondicionVenta {
			get { return this.condicionVenta; }
			set { this.condicionVenta = value; }
		}
		
    	public virtual FacturaTipoVenta Tipo {
    		get { return this.tipo; }
    		set { this.tipo = value; }
    	}
		
		[BelongsTo("clienteId",NotNull=true,Lazy=FetchWhen.Immediate,Fetch=FetchEnum.Join)]
		[ValidateNonEmpty("La factura debe tener asociado un cliente")]
		public override Cliente Cliente {
			get { return base.Cliente; }
			set { base.Cliente = value; }
		}
		
		[Property(NotNull=false)]
		[ValidateLength(0, 300,
		                "Los remitos no pueden tener mas de 300 caracteres")]
		public virtual string Remitos {
			get { return this.remitos; }
			set { this.remitos = value; }
		}
		
		[Property(NotNull=true)]
		public virtual decimal ImporteIva {
			get { return this.importeIva; }
			set { this.importeIva = StratosUtil.RoundDecimal(value, 2); }
		}
		
		[Property(NotNull=true)]
		public virtual decimal ImporteTotal {
			get { return this.importeTotal; }
			set { this.importeTotal = StratosUtil.RoundDecimal(value, 2); }
		}
		#endregion
		
		#region Propiedades sobreescritas
		/* Para una PreFactura, se valida la fecha teniendo en cuenta los objetos Factura. De
		 * esta forma la fecha mínima permitida para una PreFactura es la última fecha de
		 * la Factura del mismo tipo. */
		[Property(NotNull=true)]
		public override DateTime Fecha {
			get { return base.Fecha; }
			set { base.Fecha = value; }
		}
		
		public override decimal ImporteFinal {
			get { return this.ImporteTotal; }
		}
		#endregion
		
		#region Métodos validadores
		[ValidateSelf]
		public abstract void ValidarFecha(ErrorSummary errorSummary);
		
		/// <summary>
		/// Valida que el cliente tenga un IVA consistente con el tipo de factura.
		/// </summary>
		[ValidateSelf]
		public void ValidarCliente(ErrorSummary errorSummary)
		{
			if (this.Cliente == null)
				return;
			
			// Factura A
			if (this.Tipo == FacturaTipoVenta.FacturaA) {
				if (this.Cliente.TipoResponsable == TipoResponsable.ConsumidorFinal ||
				    	this.Cliente.TipoResponsable == TipoResponsable.Exento ||
				    	this.Cliente.TipoResponsable == TipoResponsable.Monotributo)
					errorSummary
						.RegisterErrorMessage("Cliente",
						                      "Una Factura A sólo puede tener clientes " +
						                      "Responsables Inscriptos o Responsables No " +
						                      "Inscriptos");
			}
			
			// Factura B
			else {
				if (this.Cliente.TipoResponsable == TipoResponsable.ResponsableInscripto ||
				    	this.Cliente.TipoResponsable == TipoResponsable.ResponsableNoInscripto)
					errorSummary
						.RegisterErrorMessage("Cliente",
						                      "Una Factura B sólo puede tener clientes " +
						                      "Consumidores Finales, Exentos o Monotributo");
			}
		}
		
		/// <summary>
		/// Valida que los precios unitarios sean iguales a los precios del producto
		/// (con o sin iva, de acuerdo al tipo de factura).
		/// </summary>
		/// <param name="errorSummary">
		/// A <see cref="ErrorSummary"/>
		/// </param>
		[ValidateSelf]
		public void ValidarPrecioUnitarioItems(ErrorSummary errorSummary)
		{
			// Si estoy migrando desde el sistema viejo, ignoro esta validación
			if (migrandoDatos)
				return;
			
			foreach(IItemProducto item in this.ItemsGenericos) {
				// Solo valido el precio unitario para productos no abstractos
				if (item.ProductoAbstracto)
					continue;
				
				/* Si se trata de una Factura A y el Precio unitario del item es distinto
				 * al precio del producto sin IVA, entonces el mismo es incorrecto. */
				if (this.Tipo == FacturaTipoVenta.FacturaA &&
				    	item.PrecioUnitario != item.ProductoImporteSinIva)
					errorSummary
						.RegisterErrorMessage("Items",
						                      "El precio unitario del item con descripción '" +
						                      item.Descripcion + "' es incorrecto. El mismo debe ser " +
						                      "igual al precio del producto sin IVA");
				
				/* De la misma forma, para una Factura B, el precio unitario del item debe ser
				 * igual al precio del producto con IVA. */
				else if (this.Tipo == FacturaTipoVenta.FacturaB &&
				    	item.PrecioUnitario != item.ProductoImporteVenta)
					errorSummary
						.RegisterErrorMessage("Items",
						                      "El precio unitario del item con descripción '" +
						                      item.Descripcion + "' es incorrecto. El mismo debe ser " +
						                      "igual al precio del producto con IVA");
			}
		}
		
		/// <summary>
		/// Valida que los importes de la factura sean correctos.
		/// </summary>
		[ValidateSelf]
		public override void ValidarImportes(ErrorSummary errorSummary)
		{
			base.ValidarImportes(errorSummary);

			// TODO: Ver como se calcula el IVA total de una factura, por ahora deja 21 %
			// Valido ImporteIva
			if (this.Tipo == FacturaTipoVenta.FacturaA) {
				// Valido si es menor o igual a cero
				if (this.ImporteIva <= 0.00m)
					errorSummary.RegisterErrorMessage("ImporteIva",
					                                  "El Importe de IVA no puede ser igual a cero en una " +
					                                  "Factura A");
				
				// FIXME: Leer el siguiente comentario:
				/* Valido si coincide con el cálculo del Importe neto por el IVA. Por ahora tomo
				 * el IVA del producto del primer item, pero hay que ver bien cómo es esto. */
				if (this.ImporteIva != StratosUtil.RoundDecimal((this.ImporteNeto * 0.21m), 2))
					errorSummary.RegisterErrorMessage("ImporteIva",
					                                  "El Importe de IVA es incorrecto");
			}
			else if (this.Tipo == FacturaTipoVenta.FacturaB) {
				if (this.ImporteIva != 0)
					errorSummary.RegisterErrorMessage("ImporteIva",
					                                  "El Importe de IVA es inválido: en una Factura B " +
					                                  "éste debe ser igual a cero");
			}
			
			// Valido ImporteTotal
			if (this.ImporteTotal != (this.ImporteIva + this.ImporteNeto))
				errorSummary.RegisterErrorMessage("ImporteTotal", "El Importe total es incorrecto");
		}
		#endregion
		
		#region Métodos de migración desde sistema viejo
		/// <summary>
		/// Este método activa el modo de migración de facturas del sistema viejo
		/// al sistema nuevo. Este modo hace que se pueda crear facturas con items donde el precio
		/// unitario del mismo difiere al precio del producto.
		/// Ver método OnNotValid()
		/// </summary>
		public static void ModoMigracionFacturas()
		{
			migrandoDatos = true;
		}
		
		/// <summary>
		/// Vuelve al modo normal, y sí valida el precio unitario de los items de las facturas.
		/// </summary>
		public static void ModoNormal()
		{
			migrandoDatos = false;
		}
		#endregion
	}
	
	/// <summary>
	/// Es la misma clase que CondicionVenta, pero agrega validaciones acordes
	/// para un objeto de tipo Factura.
	/// </summary>
	public class CondicionVentaFacturaGenerica : CondicionVenta {
		
		public CondicionVentaFacturaGenerica()
			: base()
		{
		}
		
		public CondicionVentaFacturaGenerica(CondicionPago condicion, int plazo)
			: base(condicion, plazo)
		{
		}
		
		[Property(NotNull=true)]
		public override CondicionPago Condicion {
			get { return base.Condicion; }
			set { base.Condicion = value; }
		}
		
		[Property(NotNull=false)]
		[ValidateRange(0, int.MaxValue, "El plazo de la condición de venta es inválido")]
		public override int Plazo {
			get { return base.Plazo; }
			set { base.Plazo = value; }
		}
	}
}
