// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using NHibernate.Criterion;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Queries;
using Castle.Components.Validator;

using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Representa una factura que no ha sido impresa aún. A diferencia de la clase Factura,
	/// es mucho menos restrictiva con las validaciones.
	/// </summary>
	[ActiveRecord(DiscriminatorColumn="notacredito",
	              DiscriminatorType="Int32",
	              DiscriminatorValue="0")]
	public class PreFactura : FacturaGenerica<PreFactura>
	{
		#region Campos
		protected IList<ItemPreFactura> items;
		#endregion
		
		#region Constructores
		public PreFactura() : base()
		{
			this.items = new List<ItemPreFactura>();
		}
		#endregion
		
		#region Propiedades sobreescritas
		[PrimaryKey(typeof(TableHiLoGeneratorWithMultipleColumns))]
		public override long Id {
			get { return base.Id; }
			set { base.Id = value; }
		}
		
		[Property(NotNull=true)]
		[ValidateRange(1, 9999, "El código de la factura es invalido")]
    	public override int Codigo {
    		get { return this.codigo; }
    		set { this.codigo = value; }
    	}
		
		[Property(NotNull=true)]
		public override FacturaTipoVenta Tipo {
    		get { return base.Tipo; }
    		set { base.Tipo = value; }
    	}
		
		[HasMany(Table="ItemPreFactura",
		         ColumnKey="PreFacturaId",
		         Lazy=true,
		         Cascade=ManyRelationCascadeEnum.AllDeleteOrphan)]
		public IList<ItemPreFactura> Items {
			get { return this.items; }
			set { this.items = value; }
		}
		
		public override IList ItemsGenericos {
			get { return (IList) this.items; }
			set { this.items = (IList<ItemPreFactura>) value; }
		}
		#endregion
		
		#region Métodos de validación
		public override void ValidarFecha (ErrorSummary errorSummary)
		{
			DateTime ultimaFecha = DateTime.Today - TimeSpan.FromDays(30);
			
			DateTime? fechaMinima, fechaMaxima = DateTime.Today;
			
			using (SessionScope session = new SessionScope(FlushAction.Never)) {
				/* Obtengo la fecha de la última factura realizada según el tipo
				 * (si lo tiene seteado). Esa fecha será la mínima permitida */
				// FIXME: En el where también tiene que estar el codigo de la factura
				ScalarQuery<DateTime?> consultaUltimaFecha =
					new ScalarQuery<DateTime?>(typeof(Factura),
					                           @"select max(f.Fecha)
										 		from Factura f
									 			where f.Tipo = :tipo
								 					and f.Codigo = :codigo
													and f.Anulada = cast(false,boolean)");
				
				consultaUltimaFecha.SetParameter("tipo", this.Tipo);
				
				// Seteo el código en la consulta
				consultaUltimaFecha.SetParameter("codigo", this.Codigo);
				
				fechaMinima = consultaUltimaFecha.Execute();
			}
			
			/* Si no hay fecha mínima (porque no hay facturas anteriores por ejemplo), por defecto
			 * la misma será igual a un mes anterior a la fecha actual. */
			if (!fechaMinima.HasValue)
				fechaMinima = DateTime.Today - TimeSpan.FromDays(30);
			
			if (this.Fecha.Date < fechaMinima.Value.Date) {
			    errorSummary
					.RegisterErrorMessage("Fecha", "La fecha de la PreFactura es inválida. No puede ser " +
					                      "inferior a la última Factura impresa del mismo código y tipo " +
					                      "(" + fechaMinima.Value.ToShortDateString() + ")");
			}
			else if (this.Fecha.Date > fechaMaxima.Value.Date)
				errorSummary
					.RegisterErrorMessage("Fecha", "La fecha de la PreFactura es inválida. No puede ser " +
					                      "superior al día de hoy.");
		}
		#endregion
	}
}
