// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Castle.ActiveRecord;
using Castle.Components.Validator;

using Stratos.Common;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Item de un DocumentoGenerico. Todas las subclases de DocumentoGenerico
	/// deben tener también una subclase de ésta, ItemDocumentoGenerico, que
	/// representa un item asociado a un Producto para dicha clase.
	/// No genera una tabla en la base de datos.
	/// </summary>
	public abstract class ItemProducto<T> : ItemGenerico<T>, IItemProducto
		where T : class
	{
		#region Campos
		/// <summary>
		/// Esta instancia de Producto no es almacenada con el item. Simplemente
		/// se utiliza en el mapeo de ItemProducto a ItemProductoDto, para
		/// sacar algunos datos del producto que la versión Dto tiene y la
		/// versión persistente no.
		/// </summary>
		private Producto productoAsignado;
		
		private string descripcion;
		
		private decimal cantidad;
		private decimal porcentajeBonificacion;
		private decimal precioUnitario;
		private decimal importe;
		
		#region Producto asociado
		private long productoId;
		private string productoCodigo;
		private bool productoAbstracto;
		private UnidadProducto productoUnidad;
		private long productoRubroId;
		
		// La descripción sólo se copia para los productos abstractos
		private string productoDescripcion;
		
		private decimal productoPrecioCosto;
		private decimal productoImporteSinIva;
		private decimal productoImporteVenta;
		#endregion
		#endregion
		
		#region Constructores
		public ItemProducto()
		{
		}
		
		public ItemProducto(Producto producto, decimal cantidad) : base()
		{
			this.descripcion = producto.Descripcion;
			this.cantidad = cantidad;
			this.porcentajeBonificacion = 0.00m;
			this.precioUnitario = 0.00m;
			this.importe = 0.00m;
			
			this.AsignarProducto(producto);
		}
		#endregion
		
		#region Propiedades
		public Producto ProductoAsignado {
			get {
				if (this.productoAsignado == null)
					this.productoAsignado = Producto.TryFind(this.productoId);
				
				return this.productoAsignado;
			}
		}
		
		#region Propiedades específicas del item
		[Property(NotNull=true)]
		[ValidateNonEmpty("La descripción de un item no puede estar vacía")]
		public virtual string Descripcion {
			get { return this.descripcion; }
			set { this.descripcion = value; }
		}
		
		[Property(NotNull=true)]
		public virtual decimal Cantidad {
			get { return this.cantidad; }
			set { this.cantidad = value; }
		}
		
		[Property(NotNull=true)]
		[ValidateRange(RangeValidationType.Decimal,
		               -50, 50, "La bonificación del item no puede superar el (+/-) 50%")]
		public virtual decimal PorcentajeBonificacion {
			get { return this.porcentajeBonificacion; }
			set { this.porcentajeBonificacion = value; }
		}
		
		[Property(NotNull=true)]
		public virtual decimal PrecioUnitario {
			get { return this.precioUnitario; }
			set { this.precioUnitario = value; }
		}
		
		[Property(NotNull=true)]
		public virtual decimal Importe {
			get { return this.importe; }
			set {
				// Al importe final del item siempre lo redondeo a dos decimales.
				this.importe = StratosUtil.RoundDecimal(value, 2);
			}
		}
		#endregion
		
		#region Propiedades del producto asociado
		[Property(NotNull=true)]
		public long ProductoId {
			get { return this.productoId; }
			set { this.productoId = value; }
		}
		
		[Property(NotNull=true)]
		[ValidateNonEmpty("El código de producto en un item no puede estar vacío")]
		public string ProductoCodigo {
			get { return this.productoCodigo; }
			set { this.productoCodigo = value; }
		}
		
		[Property(NotNull=true)]
		public bool ProductoAbstracto {
			get { return this.productoAbstracto; }
			set { this.productoAbstracto = value; }
		}
		
		[Property(NotNull=true)]
		public UnidadProducto ProductoUnidad {
			get { return this.productoUnidad; }
			set { this.productoUnidad = value; }
		}
		
		[Property(NotNull=true)]
		public long ProductoRubroId {
			get { return this.productoRubroId; }
			set { this.productoRubroId = value; }
		}
		
		[Property(NotNull=false)]
		public string ProductoDescripcion {
			get { return this.productoDescripcion; }
			set { this.productoDescripcion = value; }
		}
		
		[Property(NotNull=true)]
		public decimal ProductoPrecioCosto {
			get { return this.productoPrecioCosto; }
			set { this.productoPrecioCosto = StratosUtil.RoundDecimal(value, 2); }
		}
		
		[Property(NotNull=true)]
		public decimal ProductoImporteSinIva {
			get { return this.productoImporteSinIva; }
			set { this.productoImporteSinIva = StratosUtil.RoundDecimal(value, 2); }
		}
		
		[Property(NotNull=true)]
		public decimal ProductoImporteVenta {
			get { return this.productoImporteVenta; }
			set { this.productoImporteVenta = StratosUtil.RoundDecimal(value, 2); }
		}
		#endregion
		#endregion
		
		#region Métodos validadores
		[ValidateSelf]
		public virtual void ValidarItem(ErrorSummary errorSummary)
		{
			#region Precio costo
			if (!this.ProductoAbstracto && this.ProductoPrecioCosto <= 0.00m)
				errorSummary.RegisterErrorMessage("ProductoPrecioCosto",
				                                  "El precio de costo del producto en un item no puede ser menor " +
				                                  "o igual a cero");
			#endregion
			
			#region Precio unitario
			// Valido que el precio unitario no sea menor o igual a cero
			if (this.PrecioUnitario <= 0.00m)
				errorSummary.RegisterErrorMessage("PrecioUnitario",
				                                  "El precio unitario de un producto no puede ser menor " +
				                                  "o igual a cero");
			#endregion
			
			#region Cantidad
			if (this.Cantidad <= 0.00m)
				errorSummary
					.RegisterErrorMessage("Cantidad",
					                      "La cantidad especificada debe ser mayor a cero.");
			
			if (this.Cantidad >= 10000.0m)
				errorSummary
					.RegisterErrorMessage("Cantidad",
					                      "La cantidad especificada no puede ser superior a " +
					                      (9999.99m).ToString("N2"));
			#endregion
			
			#region Producto asociado
			if (this.ProductoId <= 0)
				errorSummary.RegisterErrorMessage("ProductoId",
				                                  "Un item debe tener un producto asociado.");
			#endregion
			
			#region Importes producto
			// Valido que el precio unitario no sea menor o igual a cero
			if (!this.ProductoAbstracto && this.ProductoImporteSinIva <= 0.00m)
				errorSummary.RegisterErrorMessage("ProductoImporteSinIva",
				                                  "El importe sin IVA del producto asociado al item no puede ser menor " +
				                                  "o igual a cero");
			
			if (!this.ProductoAbstracto && this.ProductoImporteVenta <= 0.00m)
				errorSummary.RegisterErrorMessage("ProductoImporteVenta",
				                                  "El importe de venta del producto asociado al item no puede ser menor " +
				                                  "o igual a cero");
			#endregion
		}
		#endregion
		
		[ValidateSelf]
		public virtual void ValidarImporteItem(ErrorSummary errorSummary)
		{
			// Calculo el precio unitario, restándole la bonificación especificada
			// para el item de producto.
			decimal precioUnitarioMenosBonificacion =
				this.PrecioUnitario -
					(this.PrecioUnitario * (this.PorcentajeBonificacion / 100.0m));
			
			// Calculo el importe final
			decimal importeFinal =
				StratosUtil.RoundDecimal(precioUnitarioMenosBonificacion * this.Cantidad, 2);
			
			// Verifico que el importe final calculado sea igual que el especificado en el item,
			// con un margen de error de +- 0.01 pesos (por redondeo)
			if (! (this.Importe >= importeFinal - 0.01m && this.Importe <= importeFinal + 0.01m))
				errorSummary.RegisterErrorMessage("Importe",
				                                  "El importe del item con descripción '" +
				                                  this.Descripcion + "' es incorrecto.");
		}
		
		#region Otros métodos
		/// <summary>
		/// Asigna al item los datos del producto.
		/// </summary>
		/// <param name="producto">
		/// A <see cref="Producto"/>
		/// </param>
		public void AsignarProducto(Producto producto)
		{
			if (producto == null)
				return;
			
			this.productoAsignado = producto;
			
			// Solo asigno los campos que no fueron asignados
			
			this.productoId = producto.Id;
			this.productoCodigo = producto.Codigo;
			this.productoAbstracto = producto.ProductoAbstracto;
			this.productoUnidad = producto.Unidad;
			
			if (producto.Rubro != null)
				this.productoRubroId = producto.Rubro.Codigo;
			
			// Solo almaceno la descripción si el producto es abstracto, ya
			// que la voy a necesitar para realizar validaciones sobre la
			// descirpción del item. Si no es abstracto el producto, entonces
			// la descripción no puede cambiar, y no hace falta copiarla en
			// el campo 'productoDescripcion', solo en 'descripcion'.
			if (producto.ProductoAbstracto)
				this.productoDescripcion = producto.Descripcion;
			else
				this.productoDescripcion = null;
			
			this.productoPrecioCosto = producto.PrecioCosto;
			
			// Solo cambio los importes si estos no fueron asignados
			if (this.productoImporteSinIva <= 0.00m)
				this.productoImporteSinIva = producto.ImporteSinIva;
			
			if (this.productoImporteVenta <= 0.00m)
				this.productoImporteVenta = producto.ImporteVenta;
		}
		#endregion
	}
}
