// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;

using NHibernate.Criterion;
using Castle.ActiveRecord;
using Castle.Components.Validator;


namespace Stratos.ModeloObjetos
{
	[ActiveRecord]
	public class Presupuesto : DocumentoVenta<Presupuesto>, IVersionable
	{
		#region Campos
		private string razonSocial;
		private string telefono;
		private string correoElectronico;
		
		protected IList<ItemPresupuesto> items;
		#endregion
		
		#region Constructores
		public Presupuesto()
		{
			this.items = new List<ItemPresupuesto>();
		}
		#endregion
		
		#region Propiedades
		[PrimaryKey(PrimaryKeyType.Native, SequenceName="presupuesto_sequence", UnsavedValue="0")]
		public override long Id {
			get { return base.Id; }
			set { base.Id = value; }
		}
		
		[Property(NotNull=true)]
		public override DateTime Fecha {
			get { return base.Fecha; }
			set { base.Fecha = value; }
		}
		
		[BelongsTo("clienteId", NotNull=false,Lazy=FetchWhen.Immediate,Fetch=FetchEnum.Join)]
		public override Cliente Cliente {
			get { return base.Cliente; }
			set {
				base.Cliente = value;
				
				// Si el cliente no es nulo, entonces la razón social no puede variar al mismo
				if (value != null)
					this.RazonSocial = value.RazonSocial;
			}
		}
		
		[Property(NotNull=false)]
		public string RazonSocial {
			get { return this.razonSocial; }
			set { this.razonSocial = value; }
		}
		
		[Property(NotNull=false)]
		[ValidateLength(0, 30, "El teléfono no puede tener una longitud mayor a los 30 caracteres")]
    	public string Telefono {
    		get { return this.telefono; }
    		set { this.telefono = value; }
    	}
		
		[Property(NotNull=false)]
		[ValidateEmail("El correo electrónico es inválido")]
		public string CorreoElectronico {
			get { return this.correoElectronico; }
			set { this.correoElectronico = value; }
		}
		
		[HasMany(Table="ItemPresupuesto",
		         ColumnKey="PresupuestoId",
		         Lazy=true,
		         Cascade=ManyRelationCascadeEnum.AllDeleteOrphan)]
		public IList<ItemPresupuesto> Items {
			get { return this.items; }
			set { this.items = value; }
		}
		
		public override IList ItemsGenericos {
			get { return (IList) this.items; }
			set { this.items = (IList<ItemPresupuesto>) value; }
		}
		
		public override decimal ImporteFinal {
			get { return this.ImporteNeto; }
		}
		
		public override int LimiteItems {
			get { return 30; }
		}
		#endregion
		
		#region Métodos validadores
		[ValidateSelf]
		public void ValidarFecha(ErrorSummary errorSummary)
		{
			// Verifico que la fecha del presupuesto no sea posterior a la actual
			if (this.Fecha.Date > DateTime.Today.Date)
				errorSummary.RegisterErrorMessage("Fecha",
				                                  "La fecha del presupuesto no puede ser posterior a " +
				                                  "la fecha de hoy (" +
				                                  DateTime.Today.Date.ToShortDateString() + ")");
			
			// Verifico que la fecha del presupuesto no sea menor a la actual menos 30 días
			if (this.Fecha.Date < (DateTime.Today.Date - TimeSpan.FromDays(90)))
				errorSummary.RegisterErrorMessage("Fecha",
				                                  "La fecha del presupuesto no puede ser inferior a " +
				                                  "90 días de la fecha actual.");
		}
		#endregion
	}
}
