// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Castle.Components.Validator;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using Castle.ActiveRecord.Queries;
using NHibernate.Criterion;

using Stratos.Common;
using Stratos.Common.Excepciones;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	[ActiveRecord]
	public class Recibo : DocumentoGenerico<Recibo>, IDocumentoItems
	{
		#region Campos
		protected long numero;
		
		protected IList<ItemRecibo> items;
		
		protected string concepto;
		protected decimal importe;
		#endregion
		
		#region Constructores
		public Recibo() : base()
		{
			this.items = new List<ItemRecibo>();
		}
		#endregion
		
		#region Propiedades
		[PrimaryKey(typeof(TableHiLoGeneratorWithMultipleColumns))]
		public override long Id {
			get { return base.Id; }
			set { base.Id = value; }
		}
		
		[Property(NotNull=true)]
		public override DateTime Fecha {
			get { return base.Fecha; }
			set { base.Fecha = value; }
		}
		
		[Property(NotNull=true,Unique=true)]
		public long Numero {
			get { return this.numero; }
			set { this.numero = value; }
		}
		
		[BelongsTo("clienteId", NotNull=true, Lazy=FetchWhen.Immediate,Fetch=FetchEnum.Join)]
		public Cliente Cliente {
			get { return (Cliente)base.Persona; }
			set { base.Persona = value; }
		}
		
		[HasMany(Table="ItemRecibo",
		         ColumnKey="ReciboId",
		         Lazy=true,
		         Cascade=ManyRelationCascadeEnum.AllDeleteOrphan)]
		public IList<ItemRecibo> Items {
			get { return this.items; }
			set { this.items = value; }
		}
		
		public IList ItemsGenericos {
			get { return (IList) this.items; }
			set { this.items = (IList<ItemRecibo>) value; }
		}
		
		[Property(NotNull=true)]
		[ValidateNonEmpty("El recibo debe tener asignado un concepto")]
		public string Concepto {
			get { return this.concepto; }
			set { this.concepto = value; }
		}
		
		[Property(NotNull=true)]
		public decimal Importe {
			get { return this.importe; }
			set { this.importe = StratosUtil.RoundDecimal(value, 2); }
		}
		#endregion
		
		#region Métodos de validación
		[ValidateSelf]
		public virtual void ValidarItems(ErrorSummary errorSummary)
		{
			// Verifico que la cantidad de items no sea cero
			if (this.items.Count == 0)
				errorSummary.RegisterErrorMessage("Items",
				                                  "El recibo debe tener al menos un item.");
			
			// Si el item corresponde a un cheque, este debe estar asignado
			// Si corresponde a efectivo, no debe tener asignado un cheque
			foreach (ItemRecibo item in this.items) {
				if (item.MedioPago == MedioPago.Cheque && item.Cheque == null)
					errorSummary.RegisterErrorMessage("Items",
					                                  "Falta asignar el cheque a un item.");
				else if (item.MedioPago == MedioPago.Efectivo && item.Cheque != null)
					errorSummary.RegisterErrorMessage("Items",
					                                  "Un item de efectivo tiene asignado un cheque.");
			}
		}
		
		[ValidateSelf]
		public virtual void ValidarImportes(ErrorSummary errorSummary)
		{
			// El importe del recibo no puede ser menor o igual a cero
			if (this.importe <= 0.00m)
				errorSummary.RegisterErrorMessage("Importe",
				                                  "El importe del recibo no puede ser igual o menor a cero.");
			
			// El importe debe ser la suma de los importes de los items
			decimal sumaImporteItems = this.Items.Sum(item => item.Importe);
			
			if (this.importe != StratosUtil.RoundDecimal(sumaImporteItems, 2))
				errorSummary.RegisterErrorMessage("Importe",
				                                  "El importe del recibo no coincide con la suma del importe " +
				                                  "de los items.");
		}
		
		[ValidateSelf]
		public virtual void ValidarCliente(ErrorSummary errorSummary)
		{
			/* El cliente del recibo debe ser el mismo cliente que todas las facturas
			 * asociadas. */
			
			if (this.Cliente == null) {
				errorSummary.RegisterErrorMessage("Cliente",
				                                  "Un recibo debe tener asociado un cliente");
				return;
			}
		}
		#endregion
		
		#region Métodos ActiveRecord
		protected override bool BeforeSave (IDictionary state)
		{
			base.BeforeSave (state);
			
			// Asigno el siguiente número al Recibo que se está creando
			state["Numero"] = ReciboIDService.SiguienteNumero;
			
			// Creo un objeto DocumentoCredito correspondiente al recibo
			// que se está creando
			DocumentoCredito.CrearCredito(this);
			
			return true;
		}
		#endregion
	}
}
