// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Castle.ActiveRecord;
using Castle.Components.Validator;

namespace Stratos.ModeloObjetos
{
	public class CUIT
	{
		#region Campos
		public const string CUIT_INVALIDO = "[CUIT inválido]";
		private long? cuit;
		#endregion
		
		#region Constructores
		public CUIT()
		{
			this.cuit = null;
		}
		
		public CUIT(long cuit) : this()
		{
			this.cuit = cuit;
		}
		
		public CUIT(string cuit) : this()
		{
			try {
				if (cuit != null && cuit != String.Empty)
					this.cuit = Convert.ToInt64(cuit.Replace("-", ""));
			}
			catch (FormatException) {
				/* Si hay un problema al parsear el cuit en cadena, entonces le asigno
				 * un valor nulo. */
				this.cuit = -1;
			}
		}
		#endregion
		
		#region Propiedades
		public string ValorFormateado {
			get {
				// Si el cuit es inválido
				if (!this.cuit.HasValue)
					return String.Empty;
				
				else if (this.cuit.Value < 0)
					return CUIT_INVALIDO;
				
				else if (this.cuit.Value.ToString().Length < 5)
					return this.cuit.Value.ToString();
				
				string cuitString = this.cuit.ToString();
				int cuitTamanio = cuitString.Length;
				
				string valorFormateado = cuitString.Substring(0, 2) + "-" +
					cuitString.Substring(2, cuitTamanio - 3) + "-" +
						cuitString.Substring(cuitTamanio - 1,1);
				
				return valorFormateado;
			}
		}
		
		[Property("cuit",Unique=true,NotNull=false)]
		public long? ValorNumerico {
			get { return this.cuit; }
			set { this.cuit = value; }
		}
		#endregion
		
		#region Otros métodos
		public bool CuitValido {
			get {
				if (Cliente.MigrandoDatos || Proveedor.MigrandoDatos)
					return true;
				
				if (!this.cuit.HasValue)
					return false;
				
				if (this.cuit.Value < 20000000000 || this.cuit.Value > 34000000000)
					return false;
				
				string cuit = this.cuit.Value.ToString();
				string cabecera = cuit.Substring(0, 2);
				
//				// TODO ver si hay otros numeros de cabeceras
//				if (!(cabecera.Equals("20") || cabecera.Equals("23") ||
//				      cabecera.Equals("24") || cabecera.Equals("27") ||
//				      cabecera.Equals("30") || cabecera.Equals("33")))
//					return false;
				
				int digitoValido =
					this.CalcularDigitoCuit(this.ValorFormateado.Replace("-", ""));
				
				if (digitoValido == Char.GetNumericValue(cuit[10]))
					return true;
				
				return false;
			}
		}
		
		/// <summary>
		/// Calcula el dígito verificador dado un CUIT completo o sin él.
		/// </summary>
		/// <param name="cuit">El CUIT como String sin guiones</param>
		/// <returns>El valor del dígito verificador calculado.</returns>
		private int CalcularDigitoCuit(string cuit)
		{
			int[] mult = new[] { 5, 4, 3, 2, 7, 6, 5, 4, 3, 2 };
			char[] nums = cuit.ToCharArray();
			int total = 0;
			
			for (int i = 0; i < mult.Length; i++)
				total += int.Parse(nums[i].ToString()) * mult[i];
			
			var resto = total % 11;
			return resto == 0 ? 0 : resto == 1 ? 9 : 11 - resto;
		}
		
		public override string ToString ()
		{
			return this.ValorFormateado;
		}
		#endregion
	}
}
