// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using NHibernate.Criterion;
using Castle.ActiveRecord;
using Castle.Components.Validator;

using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Almacena información sobre las deudas del cliente (débito)
	/// </summary>
	[ActiveRecord]
	public class DocumentoCredito : ActiveRecordValidationBase<DocumentoCredito>, IActiveRecordValidationBase
	{
		#region Campos
		private long id;
		
		private Cliente cliente;
		
		private TipoDocumento tipoDocumento;
		private long idDocumento;
		private int codigoNotaCredito;
		private long numeroNotaCredito;
		
		private decimal importeTotal;
		private decimal importePendiente;
		#endregion
		
		#region Constructores
		public DocumentoCredito ()
		{
		}
		#endregion
		
		#region Propiedades
		[PrimaryKey(typeof(TableHiLoGeneratorWithMultipleColumns))]
		public long Id {
			get { return this.id; }
			set { this.id = value; }
		}
		
		[BelongsTo("clienteId",NotNull=true,Lazy=FetchWhen.Immediate,Fetch=FetchEnum.Join)]
		[ValidateNonEmpty("El documento de débito debe tener asociado un cliente")]
		public Cliente Cliente {
			get { return this.cliente; }
			set { this.cliente = value; }
		}
		
		[Property(UniqueKey="clienteCreditoUnico", NotNull=true)]
		public TipoDocumento TipoDocumento {
			get { return this.tipoDocumento; }
			set { this.tipoDocumento = value; }
		}
		
		[Property(UniqueKey="clienteCreditoUnico", NotNull=true)]
		public long IdDocumento {
			get { return this.idDocumento; }
			set { this.idDocumento = value; }
		}
		
		[Property(NotNull=true)]
		public int CodigoNotaCredito {
			get { return this.codigoNotaCredito; }
			set { this.codigoNotaCredito = value; }
		}
		
		[Property(NotNull=true)]
		public long NumeroNotaCredito {
			get { return this.numeroNotaCredito; }
			set { this.numeroNotaCredito = value; }
		}
		
		[Property(NotNull=true)]
		public decimal ImporteTotal {
			get { return this.importeTotal; }
			set { this.importeTotal = value; }
		}
		
		[Property(NotNull=true)]
		public decimal ImportePendiente {
			get { return this.importePendiente; }
			set { this.importePendiente = value; }
		}
		#endregion
		
		#region Métodos
		protected static bool DocumentoCreditoAsociadoExiste(TipoDocumento tipoDocumento, long idDocumento)
		{
			return DocumentoCredito
				.Exists(Restrictions.Eq("TipoDocumento", tipoDocumento),
				        Restrictions.Eq("IdDocumento", idDocumento));
		}
		
		/// <summary>
		/// Crea un Documento de crédito a partir de una Nota de crédito.
		/// </summary>
		/// <param name="notaCredito">
		/// A <see cref="NotaCredito"/>
		/// </param>
		public static void CrearCredito(NotaCredito notaCredito)
		{
			if (DocumentoCredito.DocumentoCreditoAsociadoExiste(TipoDocumento.NotaCredito, notaCredito.Id))
				return;
			
			DocumentoCredito credito = new DocumentoCredito();
			
			credito.Cliente = notaCredito.Cliente;
			
			credito.IdDocumento = notaCredito.Id;
			credito.TipoDocumento = TipoDocumento.NotaCredito;
			credito.CodigoNotaCredito = notaCredito.Codigo;
			credito.NumeroNotaCredito = notaCredito.Numero;
			
			credito.ImporteTotal = notaCredito.ImporteTotal;
			credito.ImportePendiente = notaCredito.ImporteTotal;
			
			credito.Create();
		}
		
		public static void CrearCredito(Recibo recibo)
		{
			DocumentoCredito credito = new DocumentoCredito();
			
			credito.Cliente = recibo.Cliente;
			
			credito.IdDocumento = recibo.Id;
			credito.TipoDocumento = TipoDocumento.Recibo;
			
			credito.ImporteTotal = recibo.Importe;
			credito.ImportePendiente = recibo.Importe;
			
			credito.Create();
		}
		
		public static void EliminarCredito(NotaCredito notaCredito)
		{
			DocumentoCredito docCred =
				DocumentoCredito.FindOne(Restrictions.Eq("TipoDocumento", TipoDocumento.NotaCredito),
				                         Restrictions.Eq("IdDocumento", notaCredito.Id));
			
			if (docCred != null)
				docCred.Delete();
		}
		#endregion
	}
}
