// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using NHibernate.Criterion;
using Castle.ActiveRecord;
using Castle.Components.Validator;

using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	/// <summary>
	/// Almacena información sobre las deudas del cliente (débito)
	/// </summary>
	[ActiveRecord]
	public class DocumentoDebito : ActiveRecordValidationBase<DocumentoDebito>, IActiveRecordValidationBase
	{
		#region Campos
		private long id;
		
		private Cliente cliente;
		
		private TipoDocumento tipoDocumento;
		private long idDocumento;
		private int codigoFactura;
		private long numeroFactura;
		private DateTime fechaVencimiento;
		
		private decimal importeTotal;
		private decimal importePendiente;
		#endregion
		
		#region Constructores
		public DocumentoDebito ()
		{
		}
		#endregion
		
		#region Propiedades
		[PrimaryKey(typeof(TableHiLoGeneratorWithMultipleColumns))]
		public long Id {
			get { return this.id; }
			set { this.id = value; }
		}
		
		[BelongsTo("clienteId",NotNull=true,Lazy=FetchWhen.Immediate,Fetch=FetchEnum.Join)]
		[ValidateNonEmpty("El documento de débito debe tener asociado un cliente")]
		public Cliente Cliente {
			get { return this.cliente; }
			set { this.cliente = value; }
		}
		
		[Property(UniqueKey="clienteDebitoUnico", NotNull=true)]
		public TipoDocumento TipoDocumento {
			get { return this.tipoDocumento; }
			set { this.tipoDocumento = value; }
		}
		
		[Property(UniqueKey="clienteDebitoUnico", NotNull=true)]
		public long IdDocumento {
			get { return this.idDocumento; }
			set { this.idDocumento = value; }
		}
		
		[Property(NotNull=true)]
		public int CodigoFactura {
			get { return this.codigoFactura; }
			set { this.codigoFactura = value; }
		}
		
		[Property(NotNull=true)]
		public long NumeroFactura {
			get { return this.numeroFactura; }
			set { this.numeroFactura = value; }
		}
		
		[Property(NotNull=true)]
		public DateTime FechaVencimiento {
			get { return this.fechaVencimiento; }
			set { this.fechaVencimiento = value; }
		}
		
		[Property(NotNull=true)]
		public decimal ImporteTotal {
			get { return this.importeTotal; }
			set { this.importeTotal = value; }
		}
		
		[Property(NotNull=true)]
		public decimal ImportePendiente {
			get { return this.importePendiente; }
			set { this.importePendiente = value; }
		}
		#endregion
		
		#region Métodos
		protected static bool DocumentoDebitoAsociadoExiste(Factura factura)
		{
			return DocumentoDebito
				.Exists(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				        Restrictions.Eq("IdDocumento", factura.Id));
		}
		
		/// <summary>
		/// Crea un DocumentoDebito asociado a una Factura dada como argumento.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="Factura"/>
		/// </param>
		public static void CrearDebito(Factura factura)
		{
			if (DocumentoDebito.DocumentoDebitoAsociadoExiste(factura))
				return;
			
			DocumentoDebito debito = new DocumentoDebito();
			
			debito.Cliente = factura.Cliente;
			
			debito.IdDocumento = factura.Id;
			debito.TipoDocumento = TipoDocumento.Factura;
			debito.CodigoFactura = factura.Codigo;
			debito.NumeroFactura = factura.Numero;
			debito.FechaVencimiento = factura.Fecha.AddDays(factura.CondicionVenta.Plazo);
			
			debito.ImporteTotal = factura.ImporteTotal;
			debito.ImportePendiente = factura.ImporteTotal;
			
			debito.Create();
		}
		
		/// <summary>
		/// Elimina un DocumentoDebito asociado a una Factura dada como argumento.
		/// Si no hay débito asociado, no pasa nada.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="Factura"/>
		/// </param>
		public static void EliminarDebito(Factura factura)
		{
			// FIXME: Alternativa más eficiente, sin hacer select antes. Por alguna
			// razón dispara el evento de FacturaAnulada en Factura y este método se
			// llama nuevamente, lo que termina en un Stack Overflow.
//			DocumentoDebito
//				.DeleteAll(String.Format("TipoDocumento = {0} AND IdDocumento = {1}",
//				                         TipoDocumento.Factura, factura.Id));
			
			DocumentoDebito docDeb =
				DocumentoDebito.FindOne(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", factura.Id));
			
			if (docDeb != null)
				docDeb.Delete();
		}
		#endregion
	}
}
