// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Diagnostics;

using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using Castle.ActiveRecord.Queries;
using Castle.Components.Validator;

using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	public abstract class Persona<T> : ActiveRecordValidationBase<T>, IPersona
		where T : class
	{
		#region Campos
		private static bool migrandoDatos = false;
		
		private long codigo;
		private string razonSocial;
		
		private Domicilio domicilio;
		
		private TipoResponsable tipoResponsable;
		private CUIT cuit;
		
		private string telefono;
		private string correoElectronico;
		
		private string notas;
		
		private bool eliminado;
		private int version;
		
		private bool altaAutomatica;
		#endregion

		#region Constructores
		public Persona ()
		{
			this.domicilio = new Domicilio();
			this.cuit = new CUIT();
			this.notas = "";
			this.eliminado = false;
			
			this.altaAutomatica = false;
		}
		#endregion
		
		#region Propiedades
		public virtual long Codigo {
			get { return this.codigo; }
			set { this.codigo = value; }
		}
		
		[Property(NotNull=true)]
    	public bool AltaAutomatica {
    		get { return this.altaAutomatica; }
    		set { this.altaAutomatica = value; }
    	}

		// TODO: La razón social debería ser única, o se debería chequear que no haya
		// un cliente ya con la misma razón social, o muy parecida (esto es un algo
		// deseable nada más).
		[Property(NotNull=true)]
		[ValidateNonEmpty("La razón social del cliente debe ser asignada")]
		[ValidateLength(0, 50, "El código postal puede tener hasta 50 caracteres")]
    	public virtual string RazonSocial {
    		get { return this.razonSocial; }
    		set { this.razonSocial = value; }
    	}
		
		[Nested]
    	public virtual Domicilio Domicilio {
    		get {
				if (this.domicilio == null)
					this.domicilio = new Domicilio();
				
				return this.domicilio;
			}
    		set { this.domicilio = value; }
    	}
		
		[Property(NotNull=false)]
		[ValidateLength(0, 20, "El teléfono no debe superar los 20 caracteres")]
    	public virtual string Telefono {
    		get { return this.telefono; }
    		set { this.telefono = value; }
    	}
		
		[Property(NotNull=false)]
		[ValidateEmail("El correo electrónico es inválido")]
		public virtual string CorreoElectronico {
			get { return this.correoElectronico; }
			set { this.correoElectronico = value; }
		}
		
		[Property(NotNull=true)]
		[ValidateNonEmpty("El tipo de responsabilidad del cliente debe ser asignado")]
    	public virtual TipoResponsable TipoResponsable {
    		get { return this.tipoResponsable; }
    		set { this.tipoResponsable = value; }
    	}
		
		[Nested]
		public virtual CUIT Cuit {
    		get { return this.cuit; }
			set { this.cuit = value; }
    	}
		
		[Property(NotNull=false)]
		[ValidateLength(0, 300,
		                "Las notas del cliente no pueden tener mas de 300 caracteres")]
    	public virtual string Notas {
    		get { return this.notas; }
    		set { this.notas = value; }
    	}
		
		[Property(NotNull=true)]
		public virtual bool Eliminado {
			get { return this.eliminado; }
			set { this.eliminado = value; }
		}
		
		[Version("Version")]
		public virtual int Version {
			get { return this.version; }
			set { this.version = value; }
		}
		
		public abstract bool InstanciaValida {
			get;
		}
		
		private string NombreEntidad {
			get { return typeof(T).Name.ToLower(); }
		}
		#endregion
		
		#region Métodos de migración desde sistema viejo
		public static bool MigrandoDatos {
			get { return migrandoDatos; }
		}
		
		/// <summary>
		/// Este método activa el modo de migración de facturas del sistema viejo
		/// al sistema nuevo. Este modo hace que se pueda crear facturas con items donde el precio
		/// unitario del mismo difiere al precio del producto.
		/// Ver método OnNotValid()
		/// </summary>
		[Conditional("DEBUG")]
		public static void ModoMigracion()
		{
			migrandoDatos = true;
		}
		
		/// <summary>
		/// Vuelve al modo normal, y sí valida el precio unitario de los items de las facturas.
		/// </summary>
		[Conditional("DEBUG")]
		public static void ModoNormal()
		{
			migrandoDatos = false;
		}
		#endregion
		
		#region Métodos validadores
		[ValidateSelf]
		public void ValidarCuit(ErrorSummary errorSummary)
		{
			if (migrandoDatos)
				return;
			
			// Si la persona es inválida, entonces no valido el CUIT.
			if (!this.InstanciaValida)
				return;
			
			// Si el Cliente es Consumidor Final, no puede tener asignado un CUIT
			if (this.TipoResponsable == TipoResponsable.ConsumidorFinal &&
			    	this.Cuit != null && this.Cuit.CuitValido) {
				errorSummary
					.RegisterErrorMessage("Cuit",
						"Un Consumidor Final no puede tener asignado " +
						"un CUIT.");
				
				return;
			}
			
			// Si la Persona no es Consumidor Final, entonces debe tener asignado un CUIT
			else if (this.TipoResponsable != TipoResponsable.ConsumidorFinal && this.Cuit == null) {
				errorSummary
					.RegisterErrorMessage("Cuit",
						"Un Responsable Inscripto, Responsable No Inscripto, " +
						"Monotributo o Exento debe tener asignado un CUIT.");
				
				return;
			}
			
			else if (this.TipoResponsable != TipoResponsable.ConsumidorFinal && !this.Cuit.CuitValido) {
				errorSummary
					.RegisterErrorMessage("Cuit",
						"El Cuit ingresado es inválido.");
				
				return;
			}
			
			// Valido el cuit único, si este está establecido
			if (this.Cuit == null || !this.Cuit.ValorNumerico.HasValue)
				return;
			
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {				
				// Creo una consulta para ver si existe una persona con este cuit
				// ya ingresado en la base de datos.
				SimpleQuery<T> consulta =
					new SimpleQuery<T>("from " + typeof(T).FullName + " p " +
						"where p.Cuit.ValorNumerico = :cuit and " +
						"	p.Codigo != :codigo");
				
				// Seteo los parámetros de la consulta
				consulta.SetParameter("cuit", this.Cuit.ValorNumerico);
				consulta.SetParameter("codigo", this.Codigo);
				
				// Ejecuto la consulta
				Persona<T>[] resultado = consulta.Execute().Cast<Persona<T>>().ToArray();
				
				// Si hay resultados, entonces 
				if (resultado.Length > 0) {
					errorSummary
						.RegisterErrorMessage("Cuit",
							String.Format("El {0} '{1} - {2}' ya está ingresado al sistema " +
								"con el mismo CUIT ({3}).",
								this.NombreEntidad,
								resultado[0].Codigo,
								resultado[0].RazonSocial,
								this.Cuit.ValorFormateado));
					
					return;
				}
			}
		}
		#endregion
	}
}
