// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Castle.ActiveRecord;
using Castle.Components.Validator;

using Stratos.Common;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos
{
	[ActiveRecord]
	public class Producto : ActiveRecordValidationBase<Producto>, IVersionable
	{
		#region Campos
		private long id;
		private string codigo;
		private string descripcion;
		private bool valido;
		
		private bool productoAbstracto;
		private UnidadProducto unidad;
		
		private Rubro rubro; // (Comercio, Industria)
		private Rubro subRubro; // (Cerámico, mosaico, sanitario, grifería, teja, etc.)
		
		private decimal precioCosto; //VER: se podría actualizar cuando se ingresa una factura.
		private decimal importeSinIva;
		private decimal porcentajeIva = 21.0m; //(por defecto el 21%)
		
		private bool activarGestionStock = false;
		
		private bool eliminado = false;
		
		private int version;
		
		private string notas;
		#endregion
		
		#region Constructores
		public Producto()
		{
			this.valido = true;
			this.productoAbstracto = false;
			this.unidad = UnidadProducto.Unidad;
		}
		#endregion
		
		#region Propiedades
		[PrimaryKey(typeof(TableHiLoGeneratorWithMultipleColumns))]
		public long Id {
			get { return this.id; }
			set { this.id = value; }
		}
		
		[Property(NotNull=true,Unique=true)]
		[ValidateNonEmpty("El código del producto debe ser asignado")]
		[ValidateIsUnique("El código del producto debe ser único")]
		[ValidateLength(1, 15, "El código del producto debe tener entre 1 y 15 caracteres")]
    	public string Codigo {
    		get { return this.codigo; }
    		set { this.codigo = value; }
    	}

		[Property(NotNull=true,Unique=true)]
		[ValidateNonEmpty("La descripción del producto debe ser asignada")]
		[ValidateIsUnique("La descripción del producto debe ser única")]
    	public string Descripcion {
    		get { return this.descripcion; }
    		set { this.descripcion = value;}
    	}
		
		[Property(NotNull=true)]
    	public bool Valido {
    		get { return this.valido; }
    		set { this.valido = value; }
    	}
		
		[Property(NotNull=true)]
    	public bool ProductoAbstracto {
    		get { return this.productoAbstracto; }
    		set { this.productoAbstracto = value;}
    	}
		
		/* La propiedad Check intenta hacer lo siguiente: si el producto no es abstracto, entonces
		 * ImporteSinIva debe ser mayor a cero. Si se trata de un producto abstracto ImporteSinIva
		 * puede mayor o igual a 0. */
		[Property(NotNull=true)]
    	public decimal ImporteSinIva { 
    		get { return this.importeSinIva; }
    		set { this.importeSinIva = StratosUtil.RoundDecimal(value, 2); }
    	}
		
    	public decimal ImporteVenta {
    		get {
				return StratosUtil
					.RoundDecimal(this.importeSinIva * (1.00m + (this.porcentajeIva / 100.0m)), 2);
			}
    	}
		
		[Property(NotNull=true)]
    	public decimal PorcentajeIva {
    		get { return this.porcentajeIva; }
    		set { this.porcentajeIva = value; }
    	}
		
		/* La propiedad Check intenta hacer lo siguiente: si el producto no es abstracto, entonces
		 * PrecioCosto debe ser mayor a cero. Si se trata de un producto abstracto PrecioCosto
		 * puede mayor o igual a 0. */
		[Property(NotNull=true)]
    	public decimal PrecioCosto {
    		get { return this.precioCosto; }
    		set { this.precioCosto = value; }
    	}
		
		[BelongsTo("rubroId",NotNull=true)]
    	public Rubro Rubro {
    		get { return this.rubro; }
			set { this.rubro = value; }
    	}
		
		[BelongsTo("subRubroId",NotNull=false)]
    	public Rubro SubRubro {
    		get { return this.subRubro; }
			set {
				// Al setear el subrubro, también seteo el rubro padre.
				this.subRubro = value;
				
				if (this.subRubro != null)
					this.rubro = this.subRubro.RubroPadre;
			}
    	}
		
		[Property(NotNull=true)]
		[ValidateNonEmpty]
    	public UnidadProducto Unidad {
    		get { return this.unidad; }
    		set { this.unidad = value; }
    	}
		
		[Property(NotNull=true)]
		public bool ActivarGestionStock {
			get { return this.activarGestionStock; }
			set { this.activarGestionStock = value; }
		}
		
		[Property(NotNull=true)]
		public bool Eliminado {
			get { return this.eliminado; }
			set { this.eliminado = value; }
		}
		
		[Version("Version")]
		public int Version {
			get { return this.version; }
			set { this.version = value; }
		}
		
		[Property]
    	public string Notas {
    		get { return this.notas; }
    		set { this.notas = value;}
    	}
		#endregion
		
		#region Métodos validadores
		[ValidateSelf]
		public void ValidarImportes(ErrorSummary errorSummary)
		{
			/* Si el producto es un producto normal (no abstracto), entonces el precio de
			 * costo debe ser inferior al importe sin iva. En cambio si es abstracto,
			 * el precio de costo debe ser inferior o igual al importe sin iva. */
			if (this.ProductoAbstracto) {
				if (this.PrecioCosto > this.ImporteSinIva)
					errorSummary
						.RegisterErrorMessage("PrecioCosto",
						                      "El precio de costo del producto no puede ser igual " +
						                      "o mayor al precio sin IVA");
			}
			
			// Si es un producto NO Abstracto:
			else {
				// Valido que el precio de costo no sea menor o igual a cero
				if (this.PrecioCosto <= 0.00m)
					errorSummary
						.RegisterErrorMessage("PrecioCosto",
						                      "El precio de costo no puede ser menor o igual a cero");
				
				// Valido que el precio de costo no sea menor o igual a cero
				if (this.ImporteSinIva <= 0.00m)
					errorSummary
						.RegisterErrorMessage("ImporteSinIva",
						                      "El Importe sin IVA no puede ser menor o igual a cero");
				
				// Valido que el importe sin iva sea mayor al precio de costo
				if (this.PrecioCosto >= this.ImporteSinIva)
					errorSummary
						.RegisterErrorMessage("PrecioCosto",
						                      "El precio de costo del producto no puede ser igual " +
						                      "o mayor al precio sin IVA");
			}
		}
		#endregion
		
		#region Metodos sobreescritos
		public override void Delete ()
		{
			this.eliminado = true;
			this.Save();
		}

		public override void DeleteAndFlush ()
		{
			this.eliminado = true;
			this.SaveAndFlush();
		}
		#endregion
	}
}
