// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using NHibernate.Criterion;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Queries;
using Castle.Components.Validator;

namespace Stratos.ModeloObjetos
{
	public static class NotaCreditoValidador
	{
		#region Validaciones comunes
		/// <summary>
		/// Este método es común para validar tanto una NotaCredito como una PreNotaCredito.
		/// Contiene validaciones comunes.
		/// </summary>
		/// <param name="errorSummary">
		/// A <see cref="ErrorSummary"/>
		/// </param>
		/// <param name="notaCredito">
		/// A <see cref="INotaCredito"/>
		/// </param>
		private static void Validar(ErrorSummary errorSummary, INotaCredito notaCredito)
		{
			// Valido si la factura asociada fue seteada
			if (notaCredito.FacturaAsociada == null) {
				errorSummary.RegisterErrorMessage("FacturaAsociada",
				                                  "Una nota de crédito debe tener una factura asociada");
				return;
			}
			
			// Continuo con las demás validaciones
			using (SessionScope s = new SessionScope(FlushAction.Never)) {
				ValidarFacturaAsociada(errorSummary, notaCredito);
				ValidarImportes(errorSummary, notaCredito);
				ValidarItems(errorSummary, notaCredito);
			}
		}
		
		private static void ValidarFacturaAsociada(ErrorSummary errorSummary, INotaCredito notaCredito)
		{
			if (!notaCredito.FacturaAsociada.Valida)
				errorSummary
					.RegisterErrorMessage("FacturaAsociada",
					                      "Una nota de crédito no puede estar asociada a una " +
					                      "factura inválida");
			
			// Verifico que la factura no esté anulada
			if (notaCredito.FacturaAsociada.Anulada)
				errorSummary
					.RegisterErrorMessage("FacturaAsociada",
					                      "Una nota de crédito no puede estar asociada a una " +
					                      "factura anulada");
			
			// Verifico que el tipo de la nota de crédito sea el mismo de la factura
			if (notaCredito.FacturaAsociada.Tipo != notaCredito.Tipo)
				errorSummary
					.RegisterErrorMessage("Tipo",
					                      "Una nota de crédito no puede ser de un tipo distinto " +
					                      "al tipo de la factura asociada");
			
			// Verifico que el cliente de la nota de crédito sea el mismo cliente
			// de la factura asociada
			if (notaCredito.Cliente.Codigo != notaCredito.FacturaAsociada.Cliente.Codigo)
				errorSummary
					.RegisterErrorMessage("Cliente",
					                      "Una nota de crédito debe tener el mismo cliente que " +
					                      "la factura asociada");
		}
		
		private static void ValidarImportes(ErrorSummary errorSummary, INotaCredito notaCredito)
		{
			/* Verifico que la suma de todos los importes totales de todas las notas de crédito
			 * asociadas a la misma factura más el importe de esta nueva nota de crédito no superen
			 * el importe total de la factura asociada */
			
			/* Obtengo todas las notas de crédito asociadas a la factura asociada que no están
			 * anuladas */
			List<ICriterion> criteriosBusqueda = new List<ICriterion>();
			criteriosBusqueda.Add(Restrictions.Eq("FacturaAsociada", notaCredito.FacturaAsociada));
			criteriosBusqueda.Add(Restrictions.Eq("Anulada", false));
			
			/* Si estoy validando una NotaCredito (no una PreNotaCredito) excluyo de la búsqueda
			 * a la misma (sino nunca podría, por ejemplo, anular una nota de crédito) */
			if (notaCredito is NotaCredito)
				criteriosBusqueda.Add(Restrictions.Not(Restrictions.Eq("Id", notaCredito.Id)));
			
			NotaCredito[] notasCredito =
				NotaCredito.FindAll(criteriosBusqueda.ToArray());
			
			decimal sumaImportesNotasCredito = 0.00m;
			foreach (NotaCredito nc in notasCredito)
				sumaImportesNotasCredito += nc.ImporteTotal;
			
			// Ahora sumo el total de la nueva nota de crédito
			sumaImportesNotasCredito += notaCredito.ImporteTotal;
			
			// Si la suma excede el importe total de la Factura Asociada, hay un error
			if (sumaImportesNotasCredito > notaCredito.FacturaAsociada.ImporteTotal)
				errorSummary
					.RegisterErrorMessage("ImporteTotal",
					                      "La suma de los importes totales de todas las notas de " +
					                      "crédito asociadas a la Factura más esta nueva, exceden " +
					                      "el importe total de la factura asociada.");
		}
		
		private static void ValidarItems(ErrorSummary errorSummary, INotaCredito notaCredito)
		{
			/* Verifico que todos los items de la nota de crédito correspondan a un producto que
			 * esté también en la factura asociada. A esta validación la realizo sobre todos los
			 * productos, abstractos y no abstractos, cuidándome de no incluir el item "NOTA CREDITO"
			 * que puede ser usado sin ser correspondido por uno en la factura asociada. */
			var productosConcretos =
				from item in notaCredito.ItemsGenericos.Cast<IItemProducto>()
				where !item.Descripcion.ToLower().Contains("nota ") &&
					(!item.Descripcion.ToLower().Contains(" cred") ||
					 	!item.Descripcion.ToLower().Contains(" créd"))
				select item;
			
			// Busco todos los productos de la nota de crédito que existen en la factura asociada
			var productosCoincidentes =
				from item in notaCredito.ItemsGenericos.Cast<IItemProducto>()
				where notaCredito
					.FacturaAsociada.Items.Cast<ItemFactura>()
						.Any(x => x.Descripcion.ToLower() == item.Descripcion.ToLower())
				select item;
			
			/* Si todos los productos de la nota de crédito (exceptuando el item "NOTA CREDITO")
			 * existen en la factura asociada, entonces esta validación pasa */
			if (productosCoincidentes.Count() != productosConcretos.Count()) {
				errorSummary
					.RegisterErrorMessage("Items",
					                      "Los items de la nota de crédito deben estar en la factura " +
					                      "asociada");
				return;
			}
			
			/* Si omitiendo el producto específico de nota de crédito ("NOTA CREDITO") no hay productos
			 * concretos, entonces no continúo con la validación de las cantidades. */
			if (productosConcretos.Count() == 0)
				return;
			
			/* Verifico que la suma de las cantidades de cada producto en todas las notas de crédito
			 * asociadas a la factura, más la cantidad establecida por esta nota de crédito, no supere
			 * la cantidad establecida en la factura asociada. */
			
			// Creo la consulta para obtener todas las notas de crédito asociadas a la misma
			// factura
			DetachedCriteria consulta =
				DetachedCriteria.For<NotaCredito>()
					.CreateAlias("Items", "it")
					.Add(Restrictions.Eq("FacturaAsociada", notaCredito.FacturaAsociada))
					.Add(Restrictions.Eq("Anulada", false));
			
			/* Si estoy validando una NotaCredito (por ejemplo, si esta fue anulada o desanulada),
			 * la excluyo de la lista de notas de crédito */
			if (notaCredito is NotaCredito)
				consulta.Add(Restrictions.Not(Restrictions.Eq("Id", notaCredito.Id)));
			
			/* A la consulta anterior le hago una Projection para obtener los items
			 * agrupados por descripción, con la suma de sus cantidades */
			ProjectionQuery<NotaCredito,DescripcionYCantidad> projQuery =
				new ProjectionQuery<NotaCredito,DescripcionYCantidad>(consulta,
				                                 Projections.ProjectionList()
				                                 	.Add(Projections.GroupProperty("it.Descripcion"))
					    							.Add(Projections.Sum("it.Cantidad")));
			
			// Realizo la consulta a la base de datos
			IList<DescripcionYCantidad> resultado = projQuery.Execute();
			
			/* Ahora que tengo los resultados agrupados por producto, sumo las cantidades de
			 * la nueva nota de crédito */
			var cantidadesMasNuevaNotaCredito =
				from it in notaCredito.ItemsGenericos.Cast<IItemProducto>()
				let itemExistente =
					resultado.SingleOrDefault(x => x.Descripcion.ToLower() == it.Descripcion.ToLower())
				let cantidadExistente = (itemExistente != null) ? itemExistente.Cantidad : 0.00m
				select new {
					Descripcion = it.Descripcion,
					CantidadExistente = cantidadExistente,
					CantidadSumada = it.Cantidad + cantidadExistente
				};
			
			// Ahora valido contra la factura asociada
			decimal cantidadItem;
			StringBuilder mensajesError = new StringBuilder();
			
			// Itero por la lista de descripciones de productos agrupadas con la suma de cantidades
			foreach (var res in cantidadesMasNuevaNotaCredito) {
				/* Busco todos los items de la factura asociada que coinciden con la descripción
				 * agrupada y sumo sus cantidades. */
				cantidadItem =
					notaCredito.FacturaAsociada.Items.Cast<ItemFactura>()
						.Where(
							itemFactura => itemFactura.Descripcion.ToLower() == res.Descripcion.ToLower()
						)
						.Sum(
						     itemFactura => itemFactura.Cantidad
						);
				
				/* Comparo la cantidad sumada del producto con la cantidad que tiene la factura
				 * asociada. Si es mayor a la misma, registro un error en la propiedad 'Items' */
				if (res.CantidadSumada > cantidadItem)
					mensajesError.AppendFormat("  '{0}' -- {1}\n",
					                           res.Descripcion,
					                           Stratos.Common.StratosUtil.DecimalToString(cantidadItem - res.CantidadExistente));
			}
			
			if (mensajesError.Length > 0)
					errorSummary
						.RegisterErrorMessage("Items",
				                              "Las cantidades de algunos items superan a las cantidades " +
				                              "especificadas en la factura asociada. Los valores máximos posibles " +
				                              "para cada item son los siguientes:\n" + mensajesError.ToString());
		}
		#endregion
		
		#region Validaciones específicas
		public static void ValidarNotaCredito(ErrorSummary errorSummary, NotaCredito notaCredito)
		{
			// Primero realizo las validaciones comunes entre NotaCredito y PreNotaCredito
			Validar(errorSummary, notaCredito);
			
			// Para el resto de las verificaciones necesito que la factura asociada esté seteada
			if (notaCredito.FacturaAsociada == null)
				return;
		}
		
		public static void ValidarPreNotaCredito(ErrorSummary errorSummary, PreNotaCredito notaCredito)
		{
			// Primero realizo las validaciones comunes entre NotaCredito y PreNotaCredito
			Validar(errorSummary, notaCredito);
		}
		#endregion
	}
}
