// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Stratos.Common;
using StratosClient.Common;
using StratosClient.Core.Gui.ImpresionBorrador;
using Stratos.ModeloDto;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;

namespace StratosClient.Clientes.CuentaCorriente
{
	public class CuentaCorrientePresenter
	{
		#region Campos
		protected ICuentaCorrienteVista cuentaCorrienteVista;
		protected ControladorClientesPrx controladorClientes;
		
		private readonly string DIRECTORIO_ADDIN;
		private readonly string ARCHIVO_TEMPLATE_CUENTACORRIENTE;
		#endregion
		
		#region Constructores
		public CuentaCorrientePresenter (ICuentaCorrienteVista cuentaCorrienteVista,
		                        ControladorClientesPrx controladorClientes)
		{
			DIRECTORIO_ADDIN =
				"Addins" + System.IO.Path.DirectorySeparatorChar +
				"Clientes.CuentaCorriente" + System.IO.Path.DirectorySeparatorChar;
			
			ARCHIVO_TEMPLATE_CUENTACORRIENTE =
				DIRECTORIO_ADDIN + "templateCuentaCorriente.vm";
			
			this.cuentaCorrienteVista = cuentaCorrienteVista;
			this.controladorClientes = controladorClientes;
			
			// Razón social
			this.cuentaCorrienteVista.RazonSocialVista.PersonaCargada += this.OnClienteCargado;
			this.cuentaCorrienteVista.RazonSocialVista.PersonaQuitada += this.OnClienteQuitado;
			
			this.Vaciar();
		}
		#endregion
		
		#region Propiedades
		public ControladorClientesPrx ControladorClientes {
			get { return this.controladorClientes; }
			set { this.controladorClientes = value; }
		}
		#endregion
		
		#region Eventos
		protected void OnClienteCargado(ClienteDto clienteCargado)
		{
			this.cuentaCorrienteVista.EsPosibleActualizar = true;
			this.cuentaCorrienteVista.EsPosibleImprimir = true;
			
			this.ActualizarDatos();
		}
		
		protected void OnClienteQuitado()
		{
			this.cuentaCorrienteVista.EsPosibleActualizar = false;
			this.cuentaCorrienteVista.EsPosibleImprimir = false;
			
			this.Vaciar();
		}
		#endregion
		
		#region Métodos / Acciones
		/// <summary>
		/// Vacía el formulario.
		/// </summary>
		protected void Vaciar()
		{
			// Vacío el formulario
			this.cuentaCorrienteVista.DocumentosCredito.Vaciar();
			this.cuentaCorrienteVista.ImporteFavor = 0.00;
			
			this.cuentaCorrienteVista.DocumentosDebito.Vaciar();
			this.cuentaCorrienteVista.ImporteContra = 0.00;
			
			this.cuentaCorrienteVista.SaldoTotal = 0.00;
		}
		
		/// <summary>
		/// Actualiza los datos según el cliente cargado. Pide los recibos con importe
		/// disponible y las facturas por pagar.
		/// </summary>
		public void ActualizarDatos()
		{
			// Vacío las listas de recibos y facturas
			this.cuentaCorrienteVista.DocumentosCredito.Vaciar();
			this.cuentaCorrienteVista.DocumentosDebito.Vaciar();
			
			ClienteDto clienteCargado = this.cuentaCorrienteVista.RazonSocialVista.Persona;
			
			DocumentoDebitoDto[] documentosDebito =
				this.controladorClientes.ObtenerDocumentosDebito(clienteCargado.codigo);
			
			this.cuentaCorrienteVista
				.DocumentosDebito
					.AgregarElementos(documentosDebito
					                  	.OrderByDescending(docDeb => docDeb.fechaVencimientoTicks)
					                  		.ToArray());
			
			// Ahora busco las facturas por pagar del cliente y las agrego en la caché de facturas
			DocumentoCreditoDto[] documentosCredito =
				this.controladorClientes.ObtenerDocumentosCredito(clienteCargado.codigo);
			
			this.cuentaCorrienteVista.DocumentosCredito.AgregarElementos(documentosCredito);
			
			// Ahora calculos los importes a favor, en contra y saldo total
			double importeFavor = documentosCredito.Sum(docCred => docCred.importePendiente);
			this.cuentaCorrienteVista.ImporteFavor = StratosUtil.RoundDouble(importeFavor, 2);
			
			double importeContra = documentosDebito.Sum(docDeb => docDeb.importePendiente);
			this.cuentaCorrienteVista.ImporteContra = StratosUtil.RoundDouble(importeContra, 2);
			
			this.cuentaCorrienteVista.SaldoTotal = StratosUtil.RoundDouble(importeFavor - importeContra, 2);
		}
		
		/// <summary>
		/// Imprime el estado de cuenta corriente para el cliente seleccionado.
		/// </summary>
		public void Imprimir()
		{
			// Desactivo el botón de imprimir hasta que no haya aparecido el diálogo de impresión
			this.cuentaCorrienteVista.EsPosibleImprimir = false;
			
			// Creo el objeto ImpresionBorrador y agrego los objetos para el template
			ImpresionDocumento impresionBorrador =
				new ImpresionDocumento(ARCHIVO_TEMPLATE_CUENTACORRIENTE,
				                       DIRECTORIO_ADDIN);
			
			impresionBorrador
				.AgregarObjeto("cuentaCorriente",
				               new CuentaCorrienteWrapper(this.cuentaCorrienteVista.RazonSocialVista.Persona,
                                                          this.cuentaCorrienteVista.DocumentosCredito.ObtenerTodosElementosPrimarios(),
                                                          this.cuentaCorrienteVista.ImporteFavor,
                                                          this.cuentaCorrienteVista.DocumentosDebito.ObtenerTodosElementosPrimarios(),
                                                          this.cuentaCorrienteVista.ImporteContra,
                                                          this.cuentaCorrienteVista.SaldoTotal));
			
			impresionBorrador.ImpresionEjecutada += delegate {
				this.cuentaCorrienteVista.EsPosibleImprimir = true;
			};
			
			try {
				impresionBorrador.Imprimir();
			}
			catch {
				// Si hay algún problema, vuelvo a habilitar el botón para imprimir
				// y relanzo la excepción.
				this.cuentaCorrienteVista.EsPosibleImprimir = true;
				throw;
			}
		}
		#endregion
	}
}
