// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Stratos.ModeloDto;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;

namespace StratosClient.Clientes.CuentaCorriente
{
	/// <summary>
	/// Presenter para ImputacionesVista.
	/// </summary>
	public class ImputacionesPresenter
	{
		#region Campos
		protected IImputacionesVista imputacionesVista;
		protected ControladorClientesPrx controladorClientes;
		#endregion
		
		#region Constructores
		public ImputacionesPresenter (IImputacionesVista imputacionesVista,
		                              ControladorClientesPrx controladorClientes)
		{
			this.imputacionesVista = imputacionesVista;
			this.controladorClientes = controladorClientes;
			
			// Razón social
			this.imputacionesVista
				.RazonSocialVista.PersonaCargada += this.OnClienteCargado;
			this.imputacionesVista
				.RazonSocialVista.PersonaQuitada += this.OnClienteQuitado;
			
			// Documentos crédito - Permitir selección
			this.imputacionesVista.DocumentosCredito
				.PermitirSeleccion = this.PermitirSeleccionDocumentoCredito;
			
			// Documentos débito - Permitir selección
			this.imputacionesVista.DocumentosDebito
				.PermitirSeleccion = this.PermitirSeleccionDocumentoDebito;
			
			this.imputacionesVista.DocumentosCredito.Vaciar();
			this.imputacionesVista.DocumentosDebito.Vaciar();
			
			this.imputacionesVista.EsPosibleActualizarRecibosYFacturas = false;
		}
		#endregion
		
		#region Propiedades
		public ControladorClientesPrx ControladorClientes {
			get { return this.controladorClientes; }
			set { this.controladorClientes = value; }
		}
		#endregion
		
		#region Eventos
		protected void OnClienteCargado(ClienteDto clienteCargado)
		{
			this.imputacionesVista.EsPosibleActualizarRecibosYFacturas = true;
			this.ActualizarDocumentos();
		}
		
		protected void OnClienteQuitado()
		{
			this.imputacionesVista.EsPosibleActualizarRecibosYFacturas = false;
			
			this.imputacionesVista.DocumentosCredito.Vaciar();
			this.imputacionesVista.DocumentosDebito.Vaciar();
		}
		#endregion
		
		#region Métodos
		/// <summary>
		/// Imputa las facturas seleccionadas al recibo seleccionado.
		/// </summary>
		public void Imputar()
		{
			try {
				// Verifico que se haya cargado un cliente
				if (this.imputacionesVista.RazonSocialVista.Persona == null) {
					this.imputacionesVista
						.NotificarAdvertencia("Seleccione un cliente antes de imputar");
					return;
				}
				
				// Obtengo los documentos de crédito seleccionados y verifico
				// que se haya seleccionado algo
				List<DocumentoCreditoDto> documentosCreditoSeleccionados =
					this.imputacionesVista.DocumentosCredito.ElementosSeleccionados;
				
				if (documentosCreditoSeleccionados.Count == 0) {
					this.imputacionesVista
						.NotificarAdvertencia("Seleccione un documento de crédito antes de imputar");
					return;
				}
				
				/* Obtengo los documentos de débito seleccionados y también verifico que
				 * se haya seleccionado alguno. */
				List<DocumentoDebitoDto> documentosDebitoSeleccionados =
					this.imputacionesVista.DocumentosDebito.ElementosSeleccionados;
				
				if (documentosDebitoSeleccionados.Count == 0) {
					this.imputacionesVista
						.NotificarAdvertencia("Seleccione algún documento de débito antes de imputar");
					return;
				}
				
				// Realizo la operación de imputación
				this.controladorClientes.Imputar(documentosCreditoSeleccionados.ToArray(),
				                                 documentosDebitoSeleccionados.ToArray());
				
				this.imputacionesVista.NotificarInformacion("Imputación realizada correctamente.");
				
				// Actualizo los datos del cliente (recibos con importes disponibles y facturas por pagar)
				this.ActualizarDocumentos();
			}
			catch (ObjetoInvalidoException ex) {
				this.imputacionesVista
					.NotificarError("El recibo tiene datos inválidos:\n" +
					                ex.mensaje);
			}
			catch (ObjetoVersionIncorrectaException ex) {
			  	this.imputacionesVista
					.NotificarError("<b>El recibo que intenta modificar fue " +
						"modificado ya por otra terminal</b>.\nVuelva a cargar el " +
						"recibo actual y realice nuevamente las modificaciones");
				return;
			}
			catch (ObjetoInexistenteException ex) {
				this.imputacionesVista
					.NotificarError("El recibo que se quiere modificar no existe " +
					                "en la base de datos.", ex);
			}
		}
		
		/// <summary>
		/// Actualiza los recibos con importes disponibles y las facturas por
		/// pagar, según el cliente cargado.
		/// </summary>
		public void ActualizarDocumentos()
		{
			// Vacío las listas de recibos y facturas
			this.imputacionesVista.DocumentosCredito.Vaciar();
			this.imputacionesVista.DocumentosDebito.Vaciar();
			
			ClienteDto clienteCargado = this.imputacionesVista.RazonSocialVista.Persona;
			
			// Busco todos los recibos con importe disponible del cliente cargado
			DocumentoCreditoDto[] documentosCredito =
				this.controladorClientes.ObtenerDocumentosCredito(clienteCargado.codigo);
			
			this.imputacionesVista
				.DocumentosCredito.AgregarElementos(documentosCredito);
			
			// Ahora busco las facturas por pagar del cliente y las agrego en la caché de facturas
			DocumentoDebitoDto[] documentosDebito =
				this.controladorClientes.ObtenerDocumentosDebito(clienteCargado.codigo);
			
			this.imputacionesVista
				.DocumentosDebito
					.AgregarElementos(documentosDebito
						              	.OrderByDescending(docDeb => docDeb.fechaVencimientoTicks)
						                	.ToArray());
		}
		
		protected bool PermitirSeleccionDocumentoCredito(DocumentoCreditoDto documentoCreditoSeleccionado)
		{
			return true;
		}
		
		protected bool PermitirSeleccionDocumentoDebito(DocumentoDebitoDto documentoDebitoSeleccionado)
		{
			List<DocumentoCreditoDto> documentosCreditoSeleccionados =
				this.imputacionesVista.DocumentosCredito.ElementosSeleccionados;
			
			List<DocumentoDebitoDto> documentosDebitoSeleccionados =
				this.imputacionesVista.DocumentosDebito.ElementosSeleccionados;
			
			// Calculo la suma de los importes totales de todas las facturas ya seleccionadas
			double sumaDocumentosDebito =
				documentosDebitoSeleccionados.Sum(docDeb => docDeb.importePendiente);
			
			double sumaDocumentosCredito =
				documentosCreditoSeleccionados.Sum(docCred => docCred.importePendiente);
			
			// Si el resto del importe total del recibo no fue consumido por las facturas
			// ya seleccionadas, permito seleccionar esta nueva
			if ((sumaDocumentosCredito - sumaDocumentosDebito) <= 0.00)
				return false;
			
			return true;
		}
		#endregion
	}
}
