// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using StratosClient.Common;
using StratosClient.Core.Gui.ImpresionBorrador;
using Stratos.ModeloDto;
using StratosServer.Controladores;

namespace StratosClient.Clientes.Movimientos
{
	public class MovimientosPresenter
	{
		#region Campos
		private IMovimientosVista movimientosVista;
		private ControladorClientesPrx controladorClientes;
		
		private readonly string DIRECTORIO_ADDIN;
		private readonly string ARCHIVO_TEMPLATE_MOVIMIENTOS;
		#endregion
		
		#region Constructores
		public MovimientosPresenter (IMovimientosVista movimientosVista,
		                             ControladorClientesPrx controladorClientes)
		{
			DIRECTORIO_ADDIN =
				"Addins" + System.IO.Path.DirectorySeparatorChar +
				"Clientes.Movimientos" + System.IO.Path.DirectorySeparatorChar;
			
			ARCHIVO_TEMPLATE_MOVIMIENTOS =
				DIRECTORIO_ADDIN + "templateMovimientos.vm";
			
			this.movimientosVista = movimientosVista;
			this.controladorClientes = controladorClientes;
			
			this.movimientosVista.FechaDesde.Fecha = DateTime.Today;
			this.movimientosVista.FechaHasta.Fecha = DateTime.Today;
			
			this.movimientosVista.RazonSocialVista.PersonaQuitada += this.OnClienteQuitado;
			this.movimientosVista.FechaDesde.FechaCambiada += this.OnFechaCambiada;
			this.movimientosVista.FechaHasta.FechaCambiada += this.OnFechaCambiada;
		}
		#endregion
		
		#region Propiedades
		public ControladorClientesPrx ControladorClientes {
			get { return this.controladorClientes; }
			set { this.controladorClientes = value; }
		}
		#endregion
		
		#region Eventos
		private void OnClienteQuitado()
		{
			this.movimientosVista.ListaMovimientosVista.Vaciar();
			this.movimientosVista.EsPosibleImprimir = false;
		}
		
		private void OnFechaCambiada(DateTime nuevaFecha)
		{
			this.movimientosVista.ListaMovimientosVista.Vaciar();
			this.movimientosVista.EsPosibleImprimir = false;
		}
		#endregion
		
		#region Métodos
		/// <summary>
		/// Carga los movimientos del cliente seleccionado, en el rango de fecha
		/// indicado.
		/// </summary>
		public void CargarMovimientos()
		{
			// Verifico que haya un cliente cargado
			if (this.movimientosVista.RazonSocialVista.Persona == null) {
				this.movimientosVista
					.NotificarError("No se ha seleccionado un cliente");
				return;
			}
			
			// Verifico el rango de fechas, que se correcto.
			if (this.movimientosVista.FechaDesde.Fecha > this.movimientosVista.FechaHasta.Fecha) {
				this.movimientosVista
					.NotificarError("El rango de fechas es inconsistente");
				return;
			}
			
			// Vacío la lista de movimientos
			this.movimientosVista.ListaMovimientosVista.Vaciar();
			
			ClienteDto clienteCargado = this.movimientosVista.RazonSocialVista.Persona;
			
			bool resultadoRecortado;
			
			DocumentoGenericoDto[] movimientos =
				this.controladorClientes.ObtenerMovimientos(clienteCargado.codigo,
				                                            this.movimientosVista.FechaDesde.Fecha.Ticks,
				                                            this.movimientosVista.FechaHasta.Fecha.Ticks,
				                                            out resultadoRecortado);
			
			if (movimientos.Length == 0) {
				this.movimientosVista
					.NotificarInformacion("El cliente no tiene movimientos hechos " +
						"en el rango de fechas seleccionado");
				return;
			}
			
			this.movimientosVista.ListaMovimientosVista.AgregarElementos(movimientos);
			this.movimientosVista.EsPosibleImprimir = true;
			
			if (resultadoRecortado)
				this.movimientosVista
					.NotificarAdvertencia("<b>El resultado de la búsqueda ha sobrepasado el máximo permitido.</b>\n\n " +
						"Particione el rango seleccionado para incluir todos los resultados.");
		}
		
		/// <summary>
		/// Imprime un listado con los movimientos del cliente que se están visualizando.
		/// </summary>
		public void ImprimirMovimientos()
		{
			// Desactivo el botón de imprimir hasta que no haya aparecido el diálogo de impresión
			this.movimientosVista.EsPosibleImprimir = false;
			
			// Creo el objeto ImpresionBorrador y agrego los objetos para el template
			ImpresionDocumento impresionBorrador =
				new ImpresionDocumento(ARCHIVO_TEMPLATE_MOVIMIENTOS,
				                       DIRECTORIO_ADDIN);
			
			impresionBorrador
				.AgregarObjeto("movimientos",
				               new MovimientosWrapper(this.movimientosVista.RazonSocialVista.Persona,
				                                      this.movimientosVista.FechaDesde.Fecha,
				                                      this.movimientosVista.FechaHasta.Fecha,
				                                      this.movimientosVista.ListaMovimientosVista.ObtenerTodosGenerico().ToArray()));
			
			impresionBorrador.ImpresionEjecutada += delegate {
				this.movimientosVista.EsPosibleImprimir = true;
			};
			
			try {
				impresionBorrador.Imprimir();
			}
			catch {
				// Si hay algún problema, vuelvo a habilitar el botón para imprimir
				// y relanzo la excepción.
				this.movimientosVista.EsPosibleImprimir = true;
				throw;
			}
		}
		#endregion
	}
}
