// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;

using StratosClient.Core.Gui;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;
using Stratos.ModeloDto;
using Stratos.Common;

namespace StratosClient.Compras
{
	public class BusquedasComprasPresenter
	{
		#region Campos
		private const int TAMAÑO_PAGINA = 25;
		
		protected static BusquedasComprasPresenter instancia;
		
		protected readonly IBusquedasComprasVista busquedaVista;
		
		protected ControladorComprasPrx controladorCompras;
		
		protected CriterioOrdenacion criterioDeOrdenacion;
		protected string consultaActual;
		
		protected FacturasProveedoresPresenter imputacionPresenter;
		#endregion
		
		#region Constructores
		public BusquedasComprasPresenter(IBusquedasComprasVista busquedaVista,
		                                 ControladorComprasPrx controladorCompras)
		{
			this.busquedaVista = busquedaVista;
			this.controladorCompras = controladorCompras;
			
			this.consultaActual = String.Empty;
			
			instancia = this;
		}
		#endregion
		
		#region Propiedades
		public static BusquedasComprasPresenter Instancia {
			get { return instancia; }
		}
		
		public ControladorComprasPrx ControladorCompras {
			get { return this.controladorCompras; }
			set { this.controladorCompras = value; }
		}
		
		public FacturasProveedoresPresenter SolapaFacturasProveedores {
			set { this.imputacionPresenter = value; }
		}
		
		public CriterioOrdenacion CriterioDeOrdenacion {
			get { 
				if (this.criterioDeOrdenacion == null)
					this.criterioDeOrdenacion = 
						new CriterioOrdenacion("FechaImputacion", SentidoOrdenacion.Desc);
				return this.criterioDeOrdenacion;
			}
		}
		
		public int TamañoPagina {
			get { return TAMAÑO_PAGINA; }
		}
		#endregion
			
		#region Métodos públicos
		public void Buscar() {
			
			// Valido la razón social del proveedor
			if (this.busquedaVista.BusquedaPorRazonSocialActiva &&
			    (this.busquedaVista.RazonSocial.Trim() == String.Empty)) {
				
				this.busquedaVista.NotificarAdvertencia("El campo Razón Social se encuentra vacío. " +
					"Por favor, especifique un criterio para este campo o desactivelo.");
				
				return;
			}
			
			// Valido el número de factura
			if (this.busquedaVista.BusquedaPorNumeroActiva &&
			    	!StratosUtil.IsLong(this.busquedaVista.NumeroFactura)) {
				
				this.busquedaVista
					.NotificarAdvertencia("El número de factura especificado no " +
						"es correcto. Solo se pueden ingresar valores numéricos.");
				
				return;
			}
			
			try {
				this.ValidarRangoFechas();
			}
			catch (DatosIngresadosInvalidosException ex) {
				this.busquedaVista.NotificarError(ex.Message);
				return;
			}
			
			this.consultaActual = this.CrearConsultaHql();
			
			// Creo una consulta paginable
			ConsultaPaginable cp = new ConsultaPaginable();
			cp.pagina = 0;
			cp.tamanioPagina = TAMAÑO_PAGINA;
			cp.criteriosOrdenacion = new List<CriterioOrdenacion> { this.CriterioDeOrdenacion };
			
			FacturaCompraDto[] facturasEncontradas = this.EjecutarConsulta(cp);
			
			/* Si no hay facturas que cumplan con los criterios de busqueda,
			 * notifico al usuario y retorno */
			if (facturasEncontradas.Length <= 0) {
				this.busquedaVista.NotificarAdvertencia("No hay facturas de compras que cumplan" +
					" con los criterios de busqueda. Reformule la busqueda e intente nuevamente.");
				return;
			}
					
			// Activo el Navegador de páginas
			this.busquedaVista.NavegadorPaginasVista.Editable = true;
			
			// Vacío la lista de facturas
			this.busquedaVista.ListaFacturasComprasVista.Vaciar();
			
			// Cargo los clientes encontrados en el treeview
			this.busquedaVista.ListaFacturasComprasVista.AgregarElementosPrimeraPagina(facturasEncontradas);
		}
		
		public FacturaCompraDto ObtenerObjetoRefrescado(FacturaCompraDto factura)
		{
			FacturaCompraDto facturaRefrescada;
			
			// Obtengo la factura refrescada del servidor
			try {
				facturaRefrescada = (FacturaCompraDto)this.controladorCompras.Refrescar(factura);
			}
			catch(StratosException e) {
				this.busquedaVista.NotificarError("Ha ocurrido un error al tratar de refrescar el objeto", e);
				return null;
			}
			
			return facturaRefrescada;
		}
		
		/// <summary>
		/// Carga la factura de compra seleccionada en la lista
		/// </summary>
		public void CargarFacturaSeleccionada()
		{
			// Obtengo las facturas seleccionadas
			FacturaCompraDto[] facturasSeleccionadas =
				this.busquedaVista.ListaFacturasComprasVista.ObtenerSeleccionadosGenericos();
			
			// Chequeo que haya al menos una factura seleccionada. Si no hay ninguna, retorno.
			if (facturasSeleccionadas.Length <= 0)
				return;
			
			// Refresco la primer factura de la selección, ya que solo se puede
			// modificar una factura a la vez.
			FacturaCompraDto facturaRefrescada =
				this.ObtenerObjetoRefrescado(facturasSeleccionadas[0]);
			
			this.busquedaVista.ListaFacturasComprasVista
				.ActualizarObjetoSeleccionado(facturaRefrescada);
			
			// Muestro la solapa de imputación
			SolapaCompras.Instancia.MostrarSolapaImputacion();
			
			// Cargo la factura en la solapa de imputación
			this.imputacionPresenter.Cargar(facturaRefrescada);
		}
		
		public void EliminarFacturasSeleccionadas()
		{
			// Obtengo las facturas seleccionadas
			FacturaCompraDto[] facturasSeleccionadas =
				this.busquedaVista.ListaFacturasComprasVista.ObtenerSeleccionadosGenericos();
			
			// Obtengo los ids de las facturas seleccinadas
			IEnumerable<long> idsFacturasSeleccionadasNoHistoricas =
				from f in facturasSeleccionadas
					where f.historica == false
					select f.id;
			
			// Chequeo que haya al menos una factura seleccionada que no sea histórica.
			// Si no hay ninguna, retorno.
			if (idsFacturasSeleccionadasNoHistoricas.Count() <= 0)
				return;
			
			Gtk.ResponseType respuesta =
				this.busquedaVista.RealizarPregunta("<b>¿Está seguro de eliminar las facturas " +
					"seleccionadas?.</b> Una vez eliminada, una Factura de Compra no puede " +
					"recuperarse.");
			
			if (respuesta != Gtk.ResponseType.Yes)
				return;
			
			// Llamo al método EliminarFacturas del controlador de compras
			this.controladorCompras.EliminarFacturas(idsFacturasSeleccionadasNoHistoricas.ToArray());
			
			// Recargo la página actual
			this.busquedaVista.NavegadorPaginasVista.RecargarPaginaActual();
		}
		
		public void Vaciar()
		{
			
		}
		
		public object[] CargadorFacturas(ConsultaPaginable cp)
		{
			return this.EjecutarConsulta(cp);
		}
		#endregion
		
		#region Métodos privados
		private FacturaCompraDto[] EjecutarConsulta(ConsultaPaginable cp)
		{
			Ice.Object[] resultadoBusqueda = 
					this.ControladorCompras.BuscarHQL(this.consultaActual, cp);
			
			FacturaCompraDto[] facturasEncontradas =
				resultadoBusqueda.Cast<FacturaCompraDto>().ToArray();
			
			return facturasEncontradas;
		}
		
		private string CrearConsultaHql()
		{
			// Creo una lista de criterios y los voy agregando de acuerdo a si
			// la opción de búsqueda fue seleccionada.
			List<string> criterios = new List<string>();
			
			if (this.busquedaVista.BusquedaPorRazonSocialActiva)
				criterios
					.AddRange(StratosClientUtils
					          .GenerarCriteriosCadenaEnHQL("f.ProveedorRazonSocial",
					                                       this.busquedaVista.RazonSocial));
			
			if (this.busquedaVista.BusquedaPorNumeroActiva)
				criterios.Add("f.Numero = " + this.busquedaVista.NumeroFactura);
			
			if (this.busquedaVista.BusquedaPorFechaInicialActiva)
				criterios.Add("f.Fecha >= '" +
				              this.busquedaVista.DateComboFechaInicialVista.Fecha.ToShortDateString() +
				              "'");
				
			if (this.busquedaVista.BusquedaPorFechaFinalActiva)
				criterios.Add("f.Fecha <= '" +
				              this.busquedaVista.DateComboFechaFinalVista.Fecha.ToShortDateString() +
				              "'");
			
			// Genero la consulta final
			return "from FacturaCompra f " +
				StratosClientUtils.GenerarClausulaWhere(criterios);
		}
		
		private void ValidarRangoFechas()
		{
			// Veo que la fecha inicial sea menor que la fecha final
			if ((this.busquedaVista.BusquedaPorFechaInicialActiva &&
			    this.busquedaVista.BusquedaPorFechaFinalActiva) &&
				(this.busquedaVista.DateComboFechaInicialVista.Fecha.Date >
			     this.busquedaVista.DateComboFechaFinalVista.Fecha.Date))
				
				throw new DatosIngresadosInvalidosException("La fecha inicial no puede" +
					"ser posterior a la fecha final");
			
			// Veo que la fecha inicial no sea mayor a hoy
			if (this.busquedaVista.BusquedaPorFechaInicialActiva &&
			   (this.busquedaVista.DateComboFechaInicialVista.Fecha.Date > DateTime.Today))
				
				throw new DatosIngresadosInvalidosException("La fecha inicial no puede " +
					"ser posterior a la fecha actual.");
		}
		#endregion
	}
}