// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;

using Stratos.ModeloDto;
using StratosClient.Core.Gui;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;

namespace StratosClient.Compras
{
	public class FacturasProveedoresPresenter
	{
		public Action FacturaNueva;
		public Action FacturaCargada;
		public Action FacturaGuardada;
		public Action FacturaModificada;
		
		#region Campos
		protected static FacturasProveedoresPresenter instancia;
		protected readonly IFacturasProveedoresVista imputacionVista;
		
		protected ControladorComprasPrx controladorCompras;
		
		protected AdministradorEstadoObjeto administradorEstado;
		
		protected FacturaCompraDto facturaCargada;
		#endregion
		
		public FacturasProveedoresPresenter(IFacturasProveedoresVista vista,
		                                    ControladorComprasPrx controlador)
		{
			this.imputacionVista = vista;
			this.controladorCompras = controlador;
			
			#region Administrador de estado
			// Creo el administrador de estado de los recibos
			this.administradorEstado =
				new AdministradorEstadoObjeto(estado => this.imputacionVista.EstadoFactura = estado,
				                              activarGuardar => this.imputacionVista.EsPosibleGuardarFactura = activarGuardar);
			#endregion
			
			#region Suscripción a eventos
			// Widget de razón social
			this.imputacionVista.RazonSocialVista.PersonaCargada += this.OnProveedorCargado;
			this.imputacionVista.RazonSocialVista.PersonaQuitada += this.OnProveedorQuitado;
			
			#region Eventos de cambio de estado de la factura
			// Cuando se va a ingresar una factura nueva, cambio el estado a 'Nuevo',
			// deshabilito el navegador de elementos
			this.FacturaNueva += delegate { this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Nuevo); };
			this.FacturaNueva += delegate { this.imputacionVista.NavegadorElementos.Sensitive = false; };
			
			// Cuando se carga una factura, cambio el estado a 'Cargado', activo el navegador de elementos
			this.FacturaCargada += delegate { this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Cargado); };
			this.FacturaCargada += delegate { this.imputacionVista.NavegadorElementos.Sensitive = true; };
			
			// Cuando se modifica un campo de la factura, cambio el estado a 'Modificado'
			this.FacturaModificada += delegate { this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Modificado); };

			// Cuando se guarda una factura, cambio el estado a 'Guardado'
			this.FacturaGuardada += delegate { this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Guardado); };
			#endregion
			#endregion
			
			#region Navegador de elementos
			// Seteo el proveedor de la instancia que implementa
			// la interface IListaNavegable
			this.imputacionVista.NavegadorElementos.ProveedorListaNavegable = delegate {
				return BusquedasComprasVista.Instancia;
			};
			
			// Seteo el cargador
			this.imputacionVista.NavegadorElementos.Cargador = delegate(object o) {
				this.Cargar((FacturaCompraDto) o);
				
				// Si despues de cargar el estado es 'Cargado', retorno true
				if (this.administradorEstado.EstaCargado)
					return true;
				
				return false;
			};
			#endregion
			
			this.VaciarVista();
			
			instancia = this;
		}
		
		#region Propiedades
		public static FacturasProveedoresPresenter Instancia {
			get { return instancia; }
		}
		
		public ControladorComprasPrx ControladorCompras {
			get { return this.controladorCompras; }
			set { this.controladorCompras = value; }
		}
		#endregion
		
		#region Eventos
		protected void OnProveedorCargado(ProveedorDto proveedor)
		{
			// Cargo los datos del proveedor
			this.CargarDatosExtraProveedor(proveedor);
			
			// Marco los widgets de edición de datos del proveedor como no editables
			this.DatosProveedorEditables(false);
		}
		
		protected void OnProveedorQuitado()
		{
			// Cargo un proveedor vacío para limpiar los campos
			ProveedorDto proveedorVacio = new ProveedorDto();
			proveedorVacio.codigo = -1;
			proveedorVacio.domicilio = new DomicilioDto();
			
			this.CargarDatosExtraProveedor(proveedorVacio);
			
			// Habilito los widgets
			this.DatosProveedorEditables(true);
			
			// Según el tipo de factura, coloco el tipo de responsable por defecto
			if (this.imputacionVista.ComboBoxTipoFacturaVista.ObjetoActivo == FacturaTipoCompra.FacturaA)
				this.imputacionVista.ComboBoxTipoResponsableVista.ObjetoActivo = TipoResponsable.ResponsableInscripto;
			else
				this.imputacionVista.ComboBoxTipoResponsableVista.ObjetoActivo = TipoResponsable.Monotributo;
		}
		#endregion
		
		#region Métodos publicos
		public void Cargar(FacturaCompraDto factura)
		{
			// Primero me fijo si no hay modificaciones sin guardar
			if (this.administradorEstado.EstaModificado &&
			    	!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones)
				return;
			
			// Cargo la factura
			this.CargarFacturaInterno(factura);
			
			// Lanzo el evento 'FacturaCargada'
			if (this.FacturaCargada != null)
				this.FacturaCargada();
		}
		
		public void NuevaFactura()
		{
			// Si hay modificaciones hechas, pregunto al usuario si está seguro de perderlas
			if (this.administradorEstado.EstaModificado &&
			    	!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones)
				return;
			
			this.imputacionVista.SoloLectura = false;
			
			// Si está seguro, continúo y vacío el formulario
			this.VaciarVista();
			
			// Lanzo el evento 'FacturaNueva'
			if (this.FacturaNueva != null)
				this.FacturaNueva();
		}
		
		public void VaciarVista()
		{
			// Fechas
			this.imputacionVista.DateComboFechaImputacionVista.Fecha = DateTime.Today;
			this.imputacionVista.DateComboFechaFacturaVista.Fecha = DateTime.Today;
			
			// Vacío los datos de la cabecera
			this.imputacionVista.ComboBoxTipoFacturaVista.ObjetoActivo = FacturaTipoCompra.FacturaA;
			this.imputacionVista.RegExEntryCodigoVista.Text = string.Empty;
			this.imputacionVista.RegExEntryNumeroVista.Text = string.Empty;
			
			// Vacío los datos del proveedor
			this.imputacionVista.RazonSocialVista.Vaciar();
			
			// Vacío los items
			this.imputacionVista.ListaItems.Vaciar();
			
			// Cargo los importes
			this.imputacionVista.ImporteGravadoVista.ImporteDouble = 0.00;
			this.imputacionVista.ImporteExentoVista.ImporteDouble = 0.00;
			this.imputacionVista.ImporteIvaVista.ImporteDouble = 0.00;
			this.imputacionVista.PercepcionIBVista.ImporteDouble = 0.00;
			this.imputacionVista.PercepcionIvaVista.ImporteDouble = 0.00;
			this.imputacionVista.ImpCombustibleVista.ImporteDouble = 0.00;
			
			this.imputacionVista.ImporteTotalVista.ImporteDouble = 0.00;
			
			// Observaciones
			this.imputacionVista.ObservacionesVista = string.Empty;
			
			// Seteo la propiedad 'facturaCargada' en null
			this.facturaCargada = null;
		}
		
		/// <summary>
		/// Método que se ejecuta desde la vista cuando un campo
		/// de la interfaz es modificado. Si el parámetro es true
		/// se recalcula el importe total.
		/// </summary>
		/// <param name="recalcularImportes">
		/// A <see cref="System.Boolean"/>
		/// </param>
		public void DatosModificados(bool recalcularImportes)
		{
			if (recalcularImportes)
				this.CalcularImportes();
			
			if (this.FacturaModificada != null)
				this.FacturaModificada();
		}
		
		/// <summary>
		/// Guarda la factura ingresada en el formulario, ya sea esta esté
		/// en modificación o sea una factura nueva.
		/// </summary>
		public void Guardar()
		{
			string mensajeError;
			
			// Valido los datos ingresados
			if (!this.DatosIngresadosValidos(out mensajeError)) {
				
				// Si hay datos inválidos, muestro el mensaje de error y retorno
				this.imputacionVista.NotificarError(mensajeError);
				
				return;
			}
			
			// Creo una nueva factura
			FacturaCompraDto facturaIngresada = new FacturaCompraDto();
			
			// Si hay una factura cargada en el formulario actualmente,
			// se la asigno a la nueva factura que va a contener los datos ingresados
			if (this.facturaCargada != null)
				facturaIngresada = this.facturaCargada;
			
			// Leo los datos ingresados
			this.LeerDatosIngresados(ref facturaIngresada);
			
			FacturaCompraDto facturaPersistida;
			
			try {
				// Si la factura estaba en modificación, o sea, ya existe en la base de datos
				// llamo al metodo 'Modificar'
				if (facturaIngresada.id > 0) {
					facturaPersistida =
						(FacturaCompraDto)this.controladorCompras.Modificar(facturaIngresada);
					
					// Cargo la factura persistida a la interfaz
					this.CargarFacturaInterno(facturaPersistida);
					
					this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Guardado);
				}
				
				// Sino, la factura es nueva, por lo que llamo al método 'Guardar'
				else {
					facturaPersistida =
						(FacturaCompraDto)this.controladorCompras.Guardar(facturaIngresada);
					
					// Cargo la factura persistida a la interfaz
					this.CargarFacturaInterno(facturaPersistida);
					
					this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Guardado);
					
					// Notifico al usuario
					this.imputacionVista.NotificarInformacion("La factura se creó correctamente.");
				}
				
				// Lanzo el evento 'FacturaGuardada'
				if (this.FacturaGuardada != null)
					this.FacturaGuardada();
			}
			catch (ObjetoInvalidoException ex) {
				SMessageDialog.MensajeError("La factura de compra tiene datos inválidos:\n" +
				                            ex.mensaje);
			}
			catch (ObjetoVersionIncorrectaException ex) {
			}
			catch (ObjetoInexistenteException ex) {
			}
		}
		#endregion
		
		#region Métodos privados
		/// <summary>
		/// Calcula el importe Total
		/// </summary>
		protected virtual void CalcularImportes()
		{
			decimal total = 0.00m;
			
			if (this.imputacionVista.ImporteGravadoVista.TextoValido)
				total += this.imputacionVista.ImporteGravadoVista.Importe;
		
			if(this.imputacionVista.ImporteExentoVista.TextoValido)
				total += this.imputacionVista.ImporteExentoVista.Importe;
			
			if (this.imputacionVista.ImporteIvaVista.TextoValido)
			    total += this.imputacionVista.ImporteIvaVista.Importe;
			
			if (this.imputacionVista.PercepcionIvaVista.TextoValido)
				total += this.imputacionVista.PercepcionIvaVista.Importe;
			
			if (this.imputacionVista.PercepcionIBVista.TextoValido)
				total += this.imputacionVista.PercepcionIBVista.Importe;
			
			if (this.imputacionVista.ImpCombustibleVista.TextoValido)
				total += this.imputacionVista.ImpCombustibleVista.Importe;
			
			this.imputacionVista.ImporteTotalVista.Importe = total;
		}
		
		/// <summary>
		/// Marca todos los widgets de entrada de datos del proveedor como editables o no,
		/// de acuerdo al parámetro recibido.
		/// </summary>
		/// <param name="editable">
		/// A <see cref="System.Boolean"/>
		/// </param>
		private void DatosProveedorEditables(bool editable)
		{
			// Habilito/deshabilito los Entry relacionados con el proveedor, según corresponda
			
			// Razón Social
			this.imputacionVista.RazonSocialVista.Editable = editable;
			
			// Tipo de responsable
			this.imputacionVista.ComboBoxTipoResponsableVista.Editable = editable;
			
			// Cuit
			this.imputacionVista.RegExEntryCuitVista.Editable = editable;
			
			// Domicilio
			this.imputacionVista.DomicilioEntryVista.Editable = editable;
		}
		
		/// <summary>
		/// Permite cargar los datos restantes del proveedor en la interfaz gráfica, a partir
		/// del evento ProveedorCargado del widget RazonSocial.
		/// </summary>
		/// <param name="proveedor">
		/// A <see cref="ProveedorDto"/>
		/// </param>
		protected void CargarDatosExtraProveedor(ProveedorDto proveedor)
		{
			// Cargo los datos del domicilio del proveedor
			this.imputacionVista.DomicilioEntryVista.Domicilio = proveedor.domicilio;
			
			// FIXME: Ver esta comprobación
			// Cargo el tipo de responsabilidad
			if (proveedor.domicilio.direccion != null ||
			    proveedor.domicilio.localidad != null ||
			    proveedor.domicilio.codigoPostal != null)
				
				this.imputacionVista.ComboBoxTipoResponsableVista.ObjetoActivo = proveedor.tipoRes;
			
			// Cargo el cuit
			this.imputacionVista.RegExEntryCuitVista.Text = proveedor.cuit == null ? "" : proveedor.cuit;
		}
		
		/// <summary>
		/// Carga los datos de la factura pasada como parámetro en el
		/// formulario de imputación.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="FacturaCompraDto"/>
		/// </param>
		protected void CargarFacturaInterno(FacturaCompraDto factura)
		{
			// Cargo las fechas de imputación y la fecha de la factura
			this.imputacionVista.DateComboFechaImputacionVista.Fecha = new DateTime(factura.fechaImputacionTicks);
			this.imputacionVista.DateComboFechaFacturaVista.Fecha = new DateTime(factura.fechaTicks);
			
			// Cargo los datos de la cabecera de la factura
			this.imputacionVista.ComboBoxTipoFacturaVista.ObjetoActivo = factura.tipo;
			this.imputacionVista.RegExEntryCodigoVista.Text = factura.codigo.ToString("D4");
			this.imputacionVista.RegExEntryNumeroVista.Text = factura.numero.ToString("D8");
			
			// Cargo los datos del proveedor
			this.imputacionVista.RazonSocialVista.Persona = factura.proveedor;
			this.CargarDatosExtraProveedor(factura.proveedor);
			
			// Si el proveedor está ingresado en el sistema, deshabilito los campos
			// de edición del proveedor
			if (factura.proveedor.codigo > 0)
				this.DatosProveedorEditables(false);
			
			// Cargo los items
			this.imputacionVista.ListaItems.Vaciar(false);
			
			if (factura.items != null)
				this.imputacionVista.ListaItems.AgregarItems(factura.items.ToArray());
			else
				this.imputacionVista.ListaItems.AgregarItems();
			
			// Cargo los importes
			this.imputacionVista.ImporteGravadoVista.ImporteDouble = factura.importeGravado;
			this.imputacionVista.ImporteExentoVista.ImporteDouble = factura.importeExento;
			this.imputacionVista.ImporteIvaVista.ImporteDouble = factura.importeIva;
			this.imputacionVista.PercepcionIBVista.ImporteDouble = factura.percepcionIB;
			this.imputacionVista.PercepcionIvaVista.ImporteDouble = factura.percepcionIva;
			this.imputacionVista.ImpCombustibleVista.ImporteDouble = factura.impuestoCombustible;
			
			this.imputacionVista.ImporteTotalVista.ImporteDouble = factura.importeTotal;
			
			// Seto la propiedad 'facturaCargada'
			this.facturaCargada = factura;
			
			// Observaciones
			this.imputacionVista.ObservacionesVista = factura.observaciones;
			
			this.imputacionVista.SoloLectura = factura.historica;
		}
		
		#region Métodos de validación
		/// <summary>
		/// Valida los datos ingresados en el formulario, devolviendo
		/// true si son válidos y false en caso contrario. Setea en el
		/// parámetro out el mensaje de error, si es que existe alguno.
		/// </summary>
		/// <param name="mensajeError">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected bool DatosIngresadosValidos(out string mensajeError)
		{
			mensajeError = "";
			
			if (!this.imputacionVista.RegExEntryCodigoVista.TextoValido) {
				mensajeError = this.imputacionVista.RegExEntryCodigoVista.MensajeError;
				return false;
			}
			
			if (!this.imputacionVista.RegExEntryNumeroVista.TextoValido) {
				mensajeError = this.imputacionVista.RegExEntryNumeroVista.MensajeError;
				return false;
			}
			
			if (!this.ValidarProveedor(out mensajeError))
				return false;
			
			if (this.imputacionVista.ListaItems.Vacio) {
				mensajeError = "<b>No ha ingresado items.</b> Una factura de compra " +
					"debe tener al menos un item asociado.";
				return false;
			}
			
			if (!this.imputacionVista.ListaItems.ItemsValidos()) {
				mensajeError = "<b>Uno o más items son inválidos.</b>" +
					"Verifique que todos los items ingresados tengan un producto cargado " +
					"y un precio unitario y cantidad mayor a cero.";
				return false;
			}
			
			// Valido los importes
			if (!this.imputacionVista.ImporteGravadoVista.TextoValido) {
				mensajeError = this.imputacionVista.ImporteGravadoVista.MensajeError;
				return false;
			}
			
			if (!this.imputacionVista.ImporteExentoVista.TextoValido) {
				mensajeError = this.imputacionVista.ImporteExentoVista.MensajeError;
				return false;
			}
			
			if (!this.imputacionVista.ImporteIvaVista.TextoValido) {
				mensajeError = this.imputacionVista.ImporteIvaVista.MensajeError;
				return false;
			}
			
			if (!this.imputacionVista.PercepcionIBVista.TextoValido) {
				mensajeError = this.imputacionVista.PercepcionIBVista.MensajeError;
				return false;
			}
			
			if (!this.imputacionVista.PercepcionIvaVista.TextoValido) {
				mensajeError = this.imputacionVista.PercepcionIvaVista.MensajeError;
				return false;
			}
			
			if (!this.imputacionVista.ImpCombustibleVista.TextoValido) {
				mensajeError = this.imputacionVista.ImpCombustibleVista.MensajeError;
				return false;
			}
			
			return true;
		}
		
		protected bool ValidarProveedor(out string mensajeError)
		{
			mensajeError = "";
			
			if (this.imputacionVista.RazonSocialVista.Persona != null)
				return true;
			
			if (this.imputacionVista.RazonSocialProveedor.Text == null ||
			    	this.imputacionVista.RazonSocialProveedor.Text.Trim() == String.Empty) {
				mensajeError = "La razón social del proveedor no puede ser vacía";
				return false;
			}
			
			if (!this.imputacionVista.RegExEntryCuitVista.TextoValido) {
				mensajeError = this.imputacionVista.RegExEntryCuitVista.MensajeError;
				return false;
			}
			
			return true;
		}
		#endregion
		
		/// <summary>
		/// Lee los datos ingresado en el formulario y los
		/// carga en la factura pasada por referencia
		/// </summary>
		/// <param name="factura">
		/// A <see cref="FacturaCompraDto"/>
		/// </param>
		private void LeerDatosIngresados(ref FacturaCompraDto factura)
		{
			// Fechas
			factura.fechaImputacionTicks =
				this.imputacionVista.DateComboFechaImputacionVista.Fecha.Ticks;
			factura.fechaTicks =
				this.imputacionVista.DateComboFechaFacturaVista.Fecha.Ticks;
			
			// Datos de la cabecera
			factura.tipo = this.imputacionVista.ComboBoxTipoFacturaVista.ObjetoActivo;
			factura.codigo = int.Parse(this.imputacionVista.RegExEntryCodigoVista.Text);
			factura.numero = long.Parse(this.imputacionVista.RegExEntryNumeroVista.Text);
			
			// Datos del proveedor
			// Si el proveedor es nulo (o sea, no se cargó un proveedor de la base de datos)
			// entonces leo los datos mínimos
			if (this.imputacionVista.RazonSocialVista.Persona == null) {
				factura.proveedor = new ProveedorDto();
				factura.proveedor.razonSocial = this.imputacionVista.RazonSocialProveedor.Text;
				factura.proveedor.tipoRes = this.imputacionVista.ComboBoxTipoResponsableVista.ObjetoActivo;
				factura.proveedor.cuit = this.imputacionVista.RegExEntryCuitVista.Text;
			}
			else
				factura.proveedor = this.imputacionVista.RazonSocialVista.Persona;
			
			// FIXME: Creo que no tiene sentido setearle el domicilio acá al proveedor
			//factura.proveedor.domicilio = this.imputacionVista.DomicilioEntryVista.Domicilio;
			
			// Items
			factura.items =
				this.imputacionVista.ListaItems.ListaItems.Cast<ItemGenericoDto>().ToList();
			
			// Importes
			factura.importeGravado = this.imputacionVista.ImporteGravadoVista.ImporteDouble;
			factura.importeExento = this.imputacionVista.ImporteExentoVista.ImporteDouble;
			factura.importeIva = this.imputacionVista.ImporteIvaVista.ImporteDouble;
			factura.percepcionIB = this.imputacionVista.PercepcionIBVista.ImporteDouble;
			factura.percepcionIva = this.imputacionVista.PercepcionIvaVista.ImporteDouble;
			factura.impuestoCombustible = this.imputacionVista.ImpCombustibleVista.ImporteDouble;
			
			factura.importeTotal = this.imputacionVista.ImporteTotalVista.ImporteDouble;
			
			// Observaciones
			factura.observaciones = this.imputacionVista.ObservacionesVista;
		}

		#endregion
	}
}
