// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Gtk;

using Stratos.ModeloDto;
using Stratos.Common;
using Stratos.Common.Excepciones;
using StratosClient.Core.Gui;
using StratosServer.Controladores;
using StratosClient.Widgets.Level3;
using StratosClient.Widgets.Level1;

namespace StratosClient.Proveedores
{
	public partial class SolapaAgregarModificar : Gtk.Bin, IVisualizadorNavegable
	{
		private ComboBoxEnumeracion<TipoResponsable> cmbTipoResponsable;
		
		private AdministradorEstadoObjeto administradorEstado;
		
		private ProveedorDto proveedorCargado;

		public SolapaAgregarModificar()
		{
			this.Build();
			
			#region Combo TipoResponsable
			this.cmbTipoResponsable = new ComboBoxEnumeracion<TipoResponsable>();
			this.cmbTipoResponsable.Renderizador = delegate (TipoResponsable tipoResponsable) {
				return StratosUtil.SplitCamelString(tipoResponsable.ToString());
			};
			this.hbTipoIva.Add(this.cmbTipoResponsable);
			#endregion
			
			// Desactivo el textbox que muestra el código
			this.txtCodigo.Sensitive = false;
			
			#region Suscripción a eventos
			this.cmbTipoResponsable.Changed += OnCmbTipoResponsableChanged;
			#endregion
			
			#region Administrador del estado del presupuesto
			// Creo el objeto administrador del estado del presupuesto
			this.administradorEstado =
				new AdministradorEstadoObjeto(estado => this.lblEstadoProveedor.Markup = estado,
				                              x => this.btnGuardar.Sensitive = x);
			
			// Me suscribo a los eventos que cambian el estado de los datos
			this.txtRazonSocial.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			this.cmbTipoResponsable.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			this.rtxtCuit.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			
			this.domicilio.Cambiado += delegate { this.administradorEstado.DatosHanCambiado(); };
			
			this.rtxtTelefono.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			this.rtxtEmail.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			
			this.txtObservaciones.Buffer.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			#endregion
			
			#region NavegadorElementos
			// Seteo el proveedor de la instancia que implementa
			// la interface IListaNavegable
			this.navegadorElementos.ProveedorListaNavegable = delegate {
				return SolapaProveedores.Instancia;
			};
			
			// Seteo el cargador
			this.navegadorElementos.Cargador = delegate(object o) {
				this.CargarProveedor((ProveedorDto) o);
				
				// Si despues de cargar el estado es 'Cargado', retorno true
				if (this.administradorEstado.EstaCargado)
					return true;
				
				return false;
			};
			#endregion
			
			this.VaciarFormulario();
		}
		
		#region Propiedades
		public bool ProveedorEnModificacion {
			get {
				if (this.proveedorCargado != null)
					return true;
				
				return false;
			}
		}
		
		public AdministradorEstadoObjeto AdministradorEstado {
			get { return this.administradorEstado; }
		}
		
		public INavegadorElementosVista NavegadorElementos {
			get { return this.navegadorElementos; }
		}
		#endregion
		
		#region Eventos
		protected virtual void OnBtnGuardarClicked(object sender, System.EventArgs e)
		{
			/* Valido los datos ingresados en el formulario. Si se lanza alguna execepción
			 * es porque hay campos inválidos. Muestro los mensajes y retorno */
			try {
				this.ValidarDatosIngresados();
			}
			catch(DatosIngresadosInvalidosException ex) {
				SMessageDialog.MensajeError(ex.Message);
				return;
			}
			
			// Si no hay un proveedor cargado, creo uno nuevo
			if (this.proveedorCargado == null)
				this.proveedorCargado = new ProveedorDto();
			
			ProveedorDto proveedorPersistido;
			
			// Cargo el objeto con los datos ingresado en el formulario
			this.LeerDatosIngresados(ref this.proveedorCargado);
			
			// Si el proveedor es nuevo, llamo al metodo Agregarproveedor
			if (this.proveedorCargado.codigo <= 0) {
					
				try {
					proveedorPersistido = (ProveedorDto)
						((ControladorProveedoresPrx)MainIce.GetControlador("ControladorProveedores"))
							.Guardar(this.proveedorCargado);
					
					// Cargo el proveedor guardado en la interfaz gráfica y cambio el estado a guardado
					this.CargarProveedor(proveedorPersistido, true);
					
					// Desabilito el navegador de elementos, ya que este es un producto
					// nuevo y no tiene nada que ver con los posiblemente cargados en la lista
					this.navegadorElementos.Sensitive = false;
					
					// Cambio el estado del administrador a 'Guardado'
					this.administradorEstado
						.CambiarEstadoObjeto(EstadoObjeto.Guardado);
				}
				catch (StratosException ex) {
				  	SMessageDialog.MensajeError(ex.mensaje);
					return;
				}
				
				SMessageDialog.MensajeInformacion("<b>El proveedor se ha ingresado correctamente.</b> " +
				                                  " El código asignado a este proveedor es " +
				                                  proveedorPersistido.codigo.ToString() + ".");
			}
			
			// Si el proveedor ya existe, es porque está en modificación. Llamo al metodo 'Modificar'.
			else {
				try {
					proveedorPersistido = (ProveedorDto)
						((ControladorProveedoresPrx)MainIce.GetControlador("ControladorProveedores")).
						Modificar(this.proveedorCargado);
					
					// Cargo el proveedor guardado en la interfaz gráfica y cambio el estado a guardado
					this.CargarProveedor(proveedorPersistido, true);
					
					// Cambio el estado del administrador a 'Guardado'
					this.administradorEstado
						.CambiarEstadoObjeto(EstadoObjeto.Guardado);
				}
				catch (StratosException ex) {
				  	SMessageDialog.MensajeError(ex.mensaje);
					return;
				}
				
				SMessageDialog.MensajeInformacion("Las modificaciones hechas al proveedor se han realizado " +
				                                  "exitosamente.");
			}
		}
		
		
		protected virtual void OnBtnNuevoClicked(object sender, System.EventArgs e)
		{
			// Si hay modificaciones hechas, pregunto al usuario si está seguro de perderlas
			if (this.administradorEstado.EstaModificado &&
			    	!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones)
				return;
			
			// Vacío el formulario
			this.VaciarFormulario();
		}
		
		/// <summary>
		/// Este método se ejecuta cuando se dispara el evento Changed del ComboBox TipoResponsable
		/// y se encarga de activar o desactivar el campo que permite ingresar el cuit. Si el
		/// elemento seleccionado es 'Consumidor final' o 'Exento' el campo cuit se desactiva.
		/// </summary>
		/// <param name="sender">
		/// A <see cref="System.Object"/>
		/// </param>
		/// <param name="e">
		/// A <see cref="System.EventArgs"/>
		/// </param>
		protected virtual void OnCmbTipoResponsableChanged(object sender, System.EventArgs e)
		{
			if (this.cmbTipoResponsable.ObjetoActivo == TipoResponsable.ConsumidorFinal)
				this.rtxtCuit.Sensitive = false;
			else
				this.rtxtCuit.Sensitive = true;
		}
		
		#endregion
		
		#region Métodos Cargadores
		
		public void CargarProveedor(ProveedorDto proveedor, bool evitarPreguntaPorModificaciones)
		{
			if (!evitarPreguntaPorModificaciones &&
			    	(this.administradorEstado.EstaModificado &&
			     		!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones))
				return;
			
			// Cargo el proveedor
			this.CargarProveedorInterno(proveedor);
		}
		
		public void CargarProveedor(ProveedorDto proveedor)
		{
			// Primero me fijo si no hay modificaciones sin guardar
			if (this.administradorEstado.EstaModificado &&
			    	!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones)
				return;
			
			// Cargo el proveedor
			this.CargarProveedorInterno(proveedor);
		}
		
		public void CargarProveedorInterno(ProveedorDto proveedor)
		{
			// Seteo la instancia de proveedor que se está modificando
			this.proveedorCargado = proveedor;
			
			// Lleno los campos
			
			// Código
			this.txtCodigo.Text = proveedor.codigo.ToString();
			
			// Razón social
			this.txtRazonSocial.Text = proveedor.razonSocial;
			
			// Tipo responsable
			this.cmbTipoResponsable.Active = (int)proveedor.tipoRes;
			
			// Cuit
			this.rtxtCuit.Text = proveedor.cuit == null ? "" : proveedor.cuit;
			
			// Domicilio
			if (proveedor.domicilio == null)
				proveedor.domicilio = new DomicilioDto();
			
			this.domicilio.Domicilio = proveedor.domicilio;
			
			// Contacto
			this.rtxtTelefono.Text = proveedor.telefono;
			this.rtxtEmail.Text = proveedor.correoElectronico;
		
			// Observaciones
			this.txtObservaciones.Buffer.Text = proveedor.notas;
			
			// Activo el navegador de elementos
			this.navegadorElementos.Sensitive = true;
			
			// Cambio el estado de la interfaz gráfica
			this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Cargado);
		}
		
		#endregion
		
		#region Métodos Privados
		/// <summary>
		/// Crea un objeto ProveedorDto con los datos ingresados en el formulario
		/// </summary>
		/// <returns>
		/// A <see cref="ProveedorDto"/>
		/// </returns>
		protected void LeerDatosIngresados(ref ProveedorDto proveedor)
		{
			// Razon social y tipo de responsable
			proveedor.razonSocial = this.txtRazonSocial.Text.ToUpper();
			proveedor.tipoRes = this.cmbTipoResponsable.ObjetoActivo;

			// Si es Consumidor final o Exento, no corresponde un cuit
			if (proveedor.tipoRes != TipoResponsable.ConsumidorFinal)
				proveedor.cuit = this.rtxtCuit.Text;
			else
				proveedor.cuit = "";
			
			// Datos del domicilio
			proveedor.domicilio = this.domicilio.Domicilio;
			
			// Datos de contacto
			proveedor.telefono = this.rtxtTelefono.Text;
			proveedor.correoElectronico = this.rtxtEmail.Text;
					
			// Observaciones
			proveedor.notas = this.txtObservaciones.Buffer.Text;
		}
		
		protected void ValidarDatosIngresados()
		{
			// Razón social
			if (this.txtRazonSocial.Text.Equals(string.Empty))
			    throw new DatosIngresadosInvalidosException("Razón social invalida. No puede ser vacía.");
			
			// Tipo de responsable
			if (this.cmbTipoResponsable.TextoActivo.Equals(""))
				throw new DatosIngresadosInvalidosException("Tipo IVA no seleccionado. " +
							"Debe seleccionar uno.");
			
			// CUIT
			if (!this.rtxtCuit.TextoValido)
				throw new DatosIngresadosInvalidosException(this.rtxtCuit.MensajeError);
			
			// Domicilio
			string mensajeErrorDomicilio = String.Empty;
			
			if (!this.domicilio.DatosValidos(out mensajeErrorDomicilio))
				throw new DatosIngresadosInvalidosException(mensajeErrorDomicilio);
			
			// E-mail
			if (!this.rtxtEmail.TextoValido)
				throw new DatosIngresadosInvalidosException(this.rtxtEmail.MensajeError);
			
			// Telefono
			if (!this.rtxtTelefono.TextoValido)
				throw new DatosIngresadosInvalidosException(this.rtxtTelefono.MensajeError);
		}
		
		/// <summary>
		/// Vacía el formulario que permite Ingresar/Modificar los
		/// datos de un proveedor.
		/// </summary>
		protected void VaciarFormulario()
		{
			this.txtCodigo.Text = "";
			
			this.txtRazonSocial.Text = "";
			
			this.cmbTipoResponsable.Resetear();
			this.rtxtCuit.Text = "";
			
			this.domicilio.Vaciar();
			
			this.rtxtTelefono.Text = "";
			this.rtxtEmail.Text = "";
			
			this.txtObservaciones.Buffer.Text = "";
			
			// Reseteo el proveedor cargado
			this.proveedorCargado = null;
			
			// Desactivo el navegador de elementos
			this.navegadorElementos.Sensitive = false;
			
			// El estado de la interfaz es NUEVO
			this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Nuevo);
		}
		#endregion
	}
}
