// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.IO;

using Gtk;

using StratosClient.Reportes;
using StratosClient.Core.Gui;
using StratosClient.Common;
using Stratos.Common;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;

namespace StratosClient.Reportes
{
	/// <summary>
	/// Clase genérica que ejecuta las acciones para mostrar un diálogo (genérico, Gtk.Dialog),
	/// permite a las subclases leer los datos del mismo y se encarga de gestionar las cuestiones
	/// referidas al manejo del diálogo de espera.
	/// </summary>
	public abstract class ReporteGenerico : IComando
	{
		#region Campos
		protected ControladorReportesPrx controladorReportes;
		#endregion
		
		#region Constructores
		public ReporteGenerico()
		{
			// Obtengo el controlador de reportes
			this.controladorReportes =
				(ControladorReportesPrx) MainIce.GetControlador("ControladorReportes");
		}
		#endregion
		
		#region IComando implementation
		void IComando.Ejecutar ()
		{
			Gtk.Dialog dialogoReporte = this.DialogoReporte;
			
			// Muestro el diálogo y espero la respuesta
			ResponseType respuesta = (ResponseType) dialogoReporte.Run();
			
			// Leo los datos del diálogo antes de destruirlo
			this.LeerDatos(dialogoReporte);
			
			dialogoReporte.Destroy();
			
			// Si la respuesta es distinta de Ok retorno
			if (respuesta != ResponseType.Ok)
				return;
			
			// Muestro el diálogo de espera mientras genero el reporte remoto
			string cadenaHTML = String.Empty;
			
			DialogoEspera dialogoEspera =
				new DialogoEspera(delegate {
					cadenaHTML = this.GenerarReporteRemoto();
				});
			
			ResponseType respuestaEspera = (ResponseType) dialogoEspera.Run();
			dialogoEspera.Destroy();
			
			// Me fijo si hubo un error al ejecutar el método en el controlador
			// del servidor
			if (respuestaEspera == ResponseType.Cancel) {
				
				// Solo gestiono en esta instancia excepciones de cierto tipo
				// Si no lo son, la vuelvo a arrojar
				StratosException stratosException = null;
				
				if ((dialogoEspera.ExcepcionArrojada is ObjetoInexistenteException) ||
				    	(dialogoEspera.ExcepcionArrojada is ObjetoInvalidoException))
					stratosException = dialogoEspera.ExcepcionArrojada as StratosException;
				
				if (stratosException != null) {
					SMessageDialog.MensajeAdvertencia(stratosException.mensaje);
					return;
				}
				
				throw dialogoEspera.ExcepcionArrojada;
			}
			
			// Creo el objeto de impresión
			ImpresionDocumento impresionReporteIva = new ImpresionDocumento(cadenaHTML);
			
			impresionReporteIva.MargeIzquierda = 0.00;
			impresionReporteIva.MargenDerecha = 0.00;
			
			// Permito a las subclases configurar el objeto ImpresionDocumento antes
			// de imprimir.
			this.ConfigurarImpresionDocumento(impresionReporteIva);
			
			impresionReporteIva.Imprimir();
		}
		#endregion
		
		#region Propiedades abstractas
		/// <summary>
		/// Esta propiedad devuelve el diálogo correspondiente para generar el
		/// reporte.
		/// </summary>
		protected abstract Gtk.Dialog DialogoReporte {
			get;
		}
		#endregion
		
		#region Métodos abstractos
		/// <summary>
		/// Método utilizado por las subclases para leer los datos del diálogo pasado
		/// como argumento.
		/// </summary>
		/// <param name="dialogoReporte">
		/// A <see cref="Gtk.Dialog"/>
		/// </param>
		protected abstract void LeerDatos(Gtk.Dialog dialogoReporte);
		
		/// <summary>
		/// Genera el reporte remoto (utilizando los datos leídos desde LeerDatos, si es
		/// necesario), y retorna un string correspondiente al reporte.
		/// </summary>
		/// <param name="fechaInicio">
		/// A <see cref="DateTime"/>
		/// </param>
		/// <param name="fechaFinal">
		/// A <see cref="DateTime"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		protected abstract string GenerarReporteRemoto();
		
		/// <summary>
		/// Configura un objeto ImpresionDocumento antes de imprimirlo.
		/// </summary>
		/// <param name="impresionDocumento">
		/// A <see cref="ImpresionDocumento"/>
		/// </param>
		protected abstract void ConfigurarImpresionDocumento(ImpresionDocumento impresionDocumento);
		#endregion
	}
}
