// Copyright (C) 2009 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Gtk;

using StratosClient.Common;
using StratosClient.ServicioFacturacion.Configuracion;
using Stratos.ModeloDto;
using StratosClient.Core.Gui.ImpresionBorrador;

namespace StratosClient.ServicioFacturacion
{
	/// <summary>
	/// Esta clase es una ImpresionDocumento, con la diferencia de que en lugar
	/// de mostrar el diálogo de impresión, directamente imprime en la impresora
	/// seleccionada para imprimir.
	/// </summary>
	public class ImpresionFactura : ImpresionDocumento
	{
		#region Campos
		/// <summary>
		/// Evento que se dispara cuando el proceso de impresión finaliza. El mismo
		/// envía el estado final de la impresión.
		/// </summary>
		public event System.Action<PrintStatus> ImpresionFinalizada;
		
		private string nombreImpresora;
		private FacturaDto factura;
		#endregion
		
		#region Constructores
		public ImpresionFactura (string archivoTemplate, string directorio,
		                         string nombreFactura, FacturaDto factura,
		                         string nombreImpresora)
			: base(archivoTemplate, directorio, nombreFactura, new FacturaWrapper(factura))
		{
			this.nombreImpresora = nombreImpresora;
			this.factura = factura;
			
			// Agrego un objeto con el importe en letras
			NumeroLetras nl = new NumeroLetras();
			this.AgregarObjeto("importeLetras", nl.enletras(this.factura.importeTotal.ToString()));
			
			// FIXME: Esto debería estar en otro lado. Es el descuento por pago a término
			this.AgregarObjeto("descuentoPagoATermino", 0.10);
		}
		#endregion
		
		#region Métodos
		protected override PrintOperationAction PrepararPrintOperation (PrintOperation printOperation)
		{
			// Asigno un nombre al trabajo de impresión, de acuerdo a la
			// factura por imprimir.
			printOperation.JobName =
				this.factura.tipo.ToString() + "-" +
					(new DateTime(this.factura.fechaTicks)).ToShortDateString() + "-" +
					this.factura.cliente.razonSocial;
			
			// PageSetup de la impresión
			PageSetup pageSetup = new PageSetup();
			
			// Obtengo la sección de configuración de ServicioFacturacion
			ServicioFacturacionElement section = ServicioFacturacionElement.Instancia;
			
			// Obtengo de la configuración de la impresora para las facturas
			// los datos de la página
			Impresora impresora = section.Impresoras[this.nombreImpresora];
			
			PaperSize paperSize = null;
			
			if (impresora.ImpresionFacturas.Ancho == 0.0)
				paperSize = new PaperSize(impresora.ImpresionFacturas.NombrePapel);
			else {
				paperSize = 
					new PaperSize(impresora.ImpresionFacturas.NombrePapel,
					              impresora.ImpresionFacturas.NombrePapelDisplay,
				                  impresora.ImpresionFacturas.Ancho,
				                  impresora.ImpresionFacturas.Largo,
				                  Unit.Mm);
			}
			
			pageSetup.PaperSize = paperSize;
			pageSetup.Orientation = PageOrientation.Portrait;
			
			printOperation.DefaultPageSetup = pageSetup;
			
			// Asigno la impresora configurada para imprimir facturas
			printOperation.PrintSettings = new PrintSettings();
			printOperation.PrintSettings.Printer = this.nombreImpresora;
			printOperation.PrintSettings.UseColor = false;
			
			printOperation.UseFullPage = true;
			
//			printOperation.TrackPrintStatus = true;
			printOperation.StatusChanged += this.ImpresionStatusChanged;
			
			return PrintOperationAction.Print;
		}

		private void ImpresionStatusChanged (object sender, EventArgs e)
		{
			PrintOperation printOperation = sender as PrintOperation;
			
			if (printOperation == null)
				this.ImpresionFinalizada(PrintStatus.Finished);
			
			if (printOperation.IsFinished)
				this.ImpresionFinalizada(printOperation.Status);
		}
		#endregion
	}
}
