// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

using Gtk;

using StratosClient.Ventas.Facturacion.ImpresionFacturas.Configuracion;
using StratosClient.Ventas;
using Stratos.ModeloDto;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;
using StratosClient.Core.Gui;

namespace StratosClient.Ventas.Facturacion.ImpresionFacturas
{
	/// <summary>
	/// Centraliza las operaciones de impresión de facturas.
	/// </summary>
	public abstract class AccionadorPreFactura
	{
		#region Propiedades
		protected abstract FacturaGenericaDto FacturaSeleccionada {
			get;
		}
		#endregion
		
		#region Métodos de acciones
		protected virtual void OnImprimirActivated(object sender, EventArgs e)
		{
			// Obtengo la PreFacturaDto que se ha seleccionado para imprimir
			PreFacturaDto factura = this.ObtenerPreFactura();
			
			// Obtengo el tipo de documento para buscar la impresora configurada
			TipoDocumento tipoDocumento = factura.tipo == FacturaTipoVenta.FacturaA ?
				TipoDocumento.FacturaTipoA :TipoDocumento.FacturaTipoB;
			
			// Busco una impresora que soporte la impresión del tipo de documento seleccionado
			List<Impresora> impresoras = ImpresionFacturasElement
				.Instancia.BuscarImpresora(tipoDocumento);
			
			if (impresoras.Count == 0) {
				SMessageDialog.MensajeError("No hay impresoras configuradas para imprimir el tipo de " +
				                            "documento seleccionado.");
				
				return;
			}
			
			/* Me comunico con el servidor para conocer cuál es el siguiente número de factura
			 * del mismo tipo y código. Si hay un error (por ejemplo, la PreFactura es inválida)
			 * muestro el mensaje y salgo. */
			
			// Obtengo el controlador de facturación
			ControladorFacturacionPrx controladorFacturacion =
				((ControladorFacturacionPrx) MainIce.GetControlador("ControladorFacturacion"));
			
			long numeroFactura;
			
			try {
				// FIXME: ¿Qué hace exactamente este "Preparar..."? ¿Es realmente necesario?
				controladorFacturacion
						.PrepararPreFacturaParaImprimir(factura, out numeroFactura);
			}
			catch (ImpresionFacturaException ex) {
				SMessageDialog.MensajeError(ex.mensaje);
				
				return;
			}
			
			// Pregunto al usuario si está seguro de imprimir
			ResponseType respuestaImpresion =
				SMessageDialog.MensajePregunta("¿Seguro que desea imprimir la factura?");
			
			if (respuestaImpresion != ResponseType.Yes)
				return;
			
			// Realizo la impresión de la PreFactura seleccionada
			
			// Por defecto, al finalizar la impresión, doy de alta la Factura
			System.Action<PrintStatus> delegateImpresionFinalizada =
				delegate (PrintStatus printStatus) {
					if (printStatus == PrintStatus.Finished)
						this.PreFacturaImpresa(numeroFactura);
				};
			
			// Pregunto al usuario por el estado de la impresión (si la configuración
			// existe)
			ResultadoImpresion resultadoImpresionDialog = null;
			
			if (ImpresionFacturasElement.Instancia.PreguntarEstado)
			{
				resultadoImpresionDialog =
					new ResultadoImpresion(delegate (bool impresionCorrecta) {
						/* Si el usuario indicó que la Factura se imprimió correctamente, entonces
				 		 * le aviso al servidor y quito la PreFactura impresa de la lista. */
						
						if (impresionCorrecta)
							this.PreFacturaImpresa(numeroFactura);
					});
				
				Gtk.Application.Invoke(delegate {
					resultadoImpresionDialog.Run();
					resultadoImpresionDialog.Destroy();
				});
				
				// Cambio el delegate por defecto en este caso, habilitando los botones
				// para que el usuario indique si la impresión fue correcta o no.
				delegateImpresionFinalizada =
					delegate (PrintStatus printStatus) {
						resultadoImpresionDialog.HabilitarAcciones(printStatus);
				};
			}
			
			Impresora configImpresora = impresoras[0];
			
			ImpresionWebKit objetoImpresion =
				new ImpresionWebKit(factura, numeroFactura, tipoDocumento);
			
			try {
				objetoImpresion
					.EjecutarImpresion(configImpresora.Nombre,
					                   delegateImpresionFinalizada);
			}
			catch (NVelocity.Exception.ParseErrorException ex) {
				Gtk.Application.Invoke(delegate {
					// Cierro el diálogo de impresión (si fue abierto)
					if (resultadoImpresionDialog != null)
						resultadoImpresionDialog.Destroy();
					
					// Muestro el mensaje de error
					SMessageDialog
						.MensajeError("Hubo un error al imprimir la factura. Asegúrese que el " +
						              "archivo de template sea válido", ex);
				});
			}
		}
		
		protected virtual void OnYaImpresaActivated(object sender, EventArgs e)
		{
			// Obtengo la PreFacturaDto que se ha seleccionado para imprimir
			PreFacturaDto factura = this.ObtenerPreFactura();
			
			/* Primero obtengo el siguiente número de factura. */
			long numeroFactura;
			
			// Obtengo el controlador de facturación
			ControladorFacturacionPrx controladorFacturacion =
				((ControladorFacturacionPrx) MainIce.GetControlador("ControladorFacturacion"));
			
			try {
				controladorFacturacion
						.PrepararPreFacturaParaImprimir(factura, out numeroFactura);
			}
			catch (ImpresionFacturaException ex) {
				SMessageDialog.MensajeError(ex.mensaje);
				
				return;
			}
			
			// Pregunto al usuario si está seguro de continuar
			ResponseType respuesta =
				SMessageDialog.MensajePregunta("¿Esta seguro que la PreFactura seleccionada ya fue impresa?");
			
			if (respuesta != ResponseType.Yes)
				return;
			
			this.PreFacturaImpresa(numeroFactura);
		}
		#endregion
		
		#region Métodos
		protected virtual PreFacturaDto ObtenerPreFactura()
		{
			// Obtengo el controlador de facturación
			ControladorFacturacionPrx controladorFacturacion =
				((ControladorFacturacionPrx) MainIce.GetControlador("ControladorFacturacion"));
			
			// Obtengo la PreFacturaDto que se ha seleccionado para imprimir
			PreFacturaDto factura = (PreFacturaDto) this.FacturaSeleccionada;
			
			// Refresco la instancia de la pre factura
			factura = (PreFacturaDto) controladorFacturacion.Refrescar(factura);
			
			return factura;
		}
		
		protected virtual FacturaDto PreFacturaImpresa(long numeroFactura)
		{
			PreFacturaDto factura = this.ObtenerPreFactura();
			
			// Obtengo el controlador de facturación
			ControladorFacturacionPrx controladorFacturacion =
				((ControladorFacturacionPrx) MainIce.GetControlador("ControladorFacturacion"));
			
			FacturaDto facturaImpresa;
			
			try {
				facturaImpresa =
					controladorFacturacion
							.ImpresionPreFacturaFinalizadaCorrectamente(factura, numeroFactura);
				
				EventosGlobales.DispararElementoEliminado(factura);
				
				return facturaImpresa;
			}
			catch (ImpresionFacturaException ex) {
				SMessageDialog.MensajeError(ex.mensaje);
				
				return null;
			}
		}
		#endregion
	}
}
