// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Gtk;

using StratosClient.Core.Gui;
using Stratos.ModeloDto;
using Stratos.Common.Excepciones;
using StratosServer.Controladores;

namespace StratosClient.Ventas.Facturacion
{
	/// <summary>
	/// Se encarga de armar las acciones que se mostrarán en el listado
	/// de documentos de ventas cuando se esté consultando por facturas.
	/// </summary>
	public class AccionadorMenuFactura : IAccionadorSobreElemento
	{
		#region Campos
		private FacturaDto[] facturas;
		private TreeIter[] treeIters;
		private TreeModel modelo;
		
		private Gtk.Action visualizarAction;
		private Gtk.Action notaCreditoAction;
		private Gtk.Action anularAction;
		private Gtk.Action desanularAction;
		#endregion
		
		#region Constructores
		public AccionadorMenuFactura()
		{
			// Creo los Gtk.Action
			
			// VisualizarAction
			this.visualizarAction =
				new Gtk.Action("visualizar",
				               "Visualizar",
				               "Visualiza una Factura",
				               "gtk-jump-to");
			
			this.visualizarAction.Activated += this.OnVisualizarActivated;
			
			// NotaCreditoAction
			this.notaCreditoAction =
				new Gtk.Action("hacernotacredito",
				               "Hacer una nota de crédito",
				               "Crea una nota de crédito a partir de la factura seleccionada",
				               "gtk-edit");
			
			this.notaCreditoAction.Activated += this.OnHacerNotaDeCredito;
			
			// AnularAction
			this.anularAction =
				new Gtk.Action("anularfactura",
				               "Anular",
				               "Anula una factura",
				               "gtk-cancel");
			
			this.anularAction.Activated += this.OnAnularActivated;
			
			// DesanularAction
			this.desanularAction =
				new Gtk.Action("desanularfactura",
				               "Desanular",
				               "Desanula una factura anulada",
				               "gtk-apply");
			
			this.desanularAction.Activated += this.OnDesanularActivated;
		}
		#endregion
		
		#region IAccionadorMenu implementation
		public Gtk.Action[] Acciones {
			get {
				return new [] {this.visualizarAction, this.notaCreditoAction, this.anularAction, this.desanularAction};
			}
		}
		
		public void ActualizarAcciones (TreeIter[] iters, DocumentoGenericoDto[] documentos, TreeModel modelo)
		{
			this.treeIters = iters;
			this.facturas = documentos.Cast<FacturaDto>().ToArray();
			this.modelo = modelo;
			
			this.visualizarAction.Sensitive = this.facturas.Length == 1;
			
			// Se puede crear una nota de crédito únicamente desde facturas válidas
			// no anuladas
			this.notaCreditoAction.Sensitive =
				this.facturas.Length == 1 &&
					!this.facturas[0].anulada && this.facturas[0].valida &&
					!(this.facturas[0] is NotaCreditoDto);
			
			// Sólo se puede anular una factura no anulada
			this.anularAction.Sensitive =
				this.facturas.Length > 0 &&
					!this.facturas[0].anulada && !this.facturas[0].historica;
			
			this.desanularAction.Sensitive =
				this.facturas.Length > 0 &&
					!this.anularAction.Sensitive && !this.facturas[0].historica;
		}
		#endregion
		
		#region Acciones de los menúes
		private void OnVisualizarActivated(object sender, EventArgs e)
		{
			SolapaVentas.Instancia.EjecutarVisualizador(this.treeIters[0]);
		}
		
		private void OnHacerNotaDeCredito(object sender, EventArgs e)
		{
			// Refresco la instancia de la factura que estará asociada a la nota de crédito
			ControladorFacturacionPrx controladorFacturacion =
				(ControladorFacturacionPrx) MainIce.GetControlador("ControladorFacturacion");
			this.facturas[0] = (FacturaDto) controladorFacturacion.Refrescar(this.facturas[0]);
			
			// Creo una NotaCreditoDto a partir de la factura cargada
			PreNotaCreditoDto notaCredito =
				FacturacionUtil.PreNotaCreditoDesdeFactura(null, this.facturas[0]);
			
			SolapaVentas.Instancia.EjecutarVisualizador(notaCredito);
		}
		
		private void OnAnularActivated(object sender, EventArgs e)
		{
			string mensajePregunta;
			
			if (this.facturas[0] is NotaCreditoDto)
				mensajePregunta = "<b>¿Está seguro que desea anular la Nota de crédito?</b>";
			else
				mensajePregunta = "<b>¿Está seguro que desea anular la Factura?</b>";
			
			// Si se trata de una factura a cuenta corriente, entonces advierto que
			// si la misma tiene débito, éste será eliminado al anularse la factura.
			if (this.facturas[0].condVenta.condPago == CondicionPago.CuentaCorriente)
				mensajePregunta += "\n\n<u>Advertencia:</u> La factura que intenta anular " +
					"es de Cuenta Corriente. Si la misma tiene débito asociado en la cuenta " +
					"del cliente, éste será eliminado.";
			
			// Si se trata de una nota de crédito, advierto que si tiene crédito asociaco
			// éste será eliminado al anularse la nota de crédito.
			else if (this.facturas[0] is NotaCreditoDto)
				mensajePregunta += "\n\n<u>Advertencia:</u> Si la nota de crédito que intenta " +
					"anular tiene crédito asociado en la cuenta del cliente, éste será eliminado.";
			
			// Primero pregunto al usuario si está seguro en continuar con la anulación
			ResponseType respuesta = SMessageDialog.MensajePregunta(mensajePregunta);
			
			if (respuesta != ResponseType.Yes)
				return;
			
			// Si se trata de una Factura anulada, pregunto si quiere crear una
			// PreFactura correspondiente, pero hago esta pregunta solamente si la
			// factura no es una nota de crédito y no es una factura manual.
			ResponseType respuestaCrearPreFactura = ResponseType.No;
			
			if (!(this.facturas[0] is NotaCreditoDto) && !this.facturas[0].facturaManual) {
				respuestaCrearPreFactura =
					SMessageDialog.MensajePregunta("<b>¿Crear una PreFactura igual a la Factura " +
						"anulada?</b>\n\n" +
						"Creando una PreFactura igual a la Factura asociada le permite imprimir una nueva " +
						"con los mismos datos que la anulada.");
			}
			
			bool crearPrefactura = (respuestaCrearPreFactura == ResponseType.Yes);
			
			ControladorFacturacionPrx controladorFacturacion =
				(ControladorFacturacionPrx) MainIce.GetControlador("ControladorFacturacion");
			
			try {
				/* Realizo la operación de (des)anulación obteniendo la nueva versión de la
				 * Factura anulada. A ésta la ingreso en el modelo en reemplazo de la anterior.
				 * De esta forma puedo puedo seguir realizando operaciones sobre la misma mientras
				 * que no haya otra terminal que la modifique (en dicho caso tendré que refrescar
				 * la lista de facturas) */
				FacturaDto nuevaFactura;
				controladorFacturacion.AnularFactura(this.facturas[0], crearPrefactura, out nuevaFactura);
				
				// Cambio el objeto en el modelo
				this.modelo.SetValue(this.treeIters[0], 0, nuevaFactura);
				this.modelo.EmitRowChanged(this.modelo.GetPath(this.treeIters[0]), this.treeIters[0]);
			}
			catch (ObjetoInvalidoException ex) {
				SMessageDialog.MensajeError(ex.mensaje);
			}
		}
		
		private void OnDesanularActivated(object sender, EventArgs e)
		{
			string mensajeInformacion;
			
			if (this.facturas[0] is NotaCreditoDto)
				mensajeInformacion = "<b>Nota de crédito desanulada correctamente.</b>";
			else
				mensajeInformacion = "<b>Factura desanulada correctamente.</b>";
			
			// Informo al usuario que el débito o crédito (según corresponda) de la factura
			// o nota de crédito desanulada ha sido recreado.
			if (this.facturas[0].condVenta.condPago == CondicionPago.CuentaCorriente)
				mensajeInformacion += "\n\nLa factura que ha desanulado es de Cuenta Corriente. " +
					"Por lo tanto se ha creado débito asociado a la misma en la cuenta " +
					"del cliente.";
			else if (this.facturas[0] is NotaCreditoDto)
				mensajeInformacion += "\n\nSe ha creado crédito asociado a la misma " +
					"en la cuenta del cliente.";
			
			try {
				/* Realizo la operación de (des)anulación obteniendo la nueva versión de la
				 * Factura anulada. A ésta la ingreso en el modelo en reemplazo de la anterior.
				 * De esta forma puedo puedo seguir realizando operaciones sobre la misma mientras
				 * que no haya otra terminal que la modifique (en dicho caso tendré que refrescar
				 * la lista de facturas) */
				FacturaDto nuevaFactura;
				((ControladorFacturacionPrx) MainIce.GetControlador("ControladorFacturacion"))
					.DesanularFactura(this.facturas[0], out nuevaFactura);
				
				// Cambio el objeto en el modelo
				this.modelo.SetValue(this.treeIters[0], 0, nuevaFactura);
				this.modelo.EmitRowChanged(this.modelo.GetPath(this.treeIters[0]), this.treeIters[0]);
				
				// Informo al usuario de la operacion realizada
				SMessageDialog.MensajeInformacion(mensajeInformacion);
			}
			catch (ObjetoInvalidoException ex) {
				SMessageDialog.MensajeError(ex.mensaje);
			}
		}
		#endregion
	}
}
