// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

using Gtk;

using Stratos.Common;
using StratosClient.Common;
using StratosClient.Core.Gui;
using StratosClient.Core.Gui.ImpresionBorrador;
using Stratos.ModeloDto;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;
using StratosClient.Widgets.Level3;
using StratosClient.Widgets.Level2;
using StratosClient.Widgets.Level1;

namespace StratosClient.Ventas.Presupuestos
{
	public partial class SolapaPresupuestos : Gtk.Bin, IVisualizadorNavegable
	{
		#region Campos
		private static SolapaPresupuestos instanciaActual;
		
		private ItemsInteractivo itemsProducto;
		
		/// <summary>
		/// El valor de esta referencia es el mismo que this.itemsProducto.TreeView, pero
		/// está casteado para evitar tantas conversiones de tipo.
		/// </summary>
		private ListaItemsVenta tvItemsProducto;
		
		private readonly string DIRECTORIO_ADDIN;
		private readonly string ARCHIVO_TEMPLATE_PRESUPUESTO;
		
		private AdministradorEstadoObjeto administradorEstado;
		private PresupuestoDto presupuestoCargado;
		#endregion
		
		#region Constructores
		public SolapaPresupuestos()
		{
			DIRECTORIO_ADDIN =
				"Addins" + System.IO.Path.DirectorySeparatorChar +
				"Ventas.Presupuestos" + System.IO.Path.DirectorySeparatorChar;
			
			ARCHIVO_TEMPLATE_PRESUPUESTO =
				DIRECTORIO_ADDIN + "templatePresupuesto.vm";
			
			instanciaActual = this;
			
			this.Build();
			
			// Deshabilito el Entry de número de presupuesto, ya que nunca se va a modificar
			CustomWidgetsUtils.WidgetEditable(this.txtNumeroPresupuesto, false);
			
			// Agrego la lista de items
			this.tvItemsProducto = new ListaItemsVenta();
			this.itemsProducto = new ItemsInteractivo(this.tvItemsProducto);
			this.hbItems.PackStart(this.itemsProducto, true, true, 0);
			
			this.itemsProducto.Sensitive = true;
			
			this.tvItemsProducto.FacturaTipo =
				FacturaTipoVenta.FacturaB;
			
			#region Administrador del estado del presupuesto
			// Creo el objeto administrador del estado del presupuesto
			this.administradorEstado =
				new AdministradorEstadoObjeto(estado => this.lblEstado.Markup = estado,
				                              x => this.btnGuardar.Sensitive = x);
			
			this.dcFecha.FechaCambiada += delegate { this.administradorEstado.DatosHanCambiado(); };
			
			this.razonSocial.RazonSocialEntry.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			this.razonSocial.CodigoPersonaEntry.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			
			this.txtTelefono.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			this.txtMail.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			this.txtBonificacion.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			
			this.itemsProducto.TreeView.ItemsCambiados += delegate { this.administradorEstado.DatosHanCambiado(); };
			
			this.txtObservaciones.Buffer.Changed += delegate { this.administradorEstado.DatosHanCambiado(); };
			#endregion
			
			#region Suscripción a eventos básicos
			this.razonSocial.PersonaCargada += CargarDatosCliente;
			this.razonSocial.PersonaQuitada += delegate {
				// Limpio los datos extras
				this.VaciarDatosClienteExtra();
				
				this.CalcularImportes();
			};
			#endregion
			
			#region Eventos que requieren recalcular importes
			this.txtBonificacion.TextoValidoIngresado += CalcularImportes;
			this.itemsProducto.TreeView.ItemsCambiados += CalcularImportes;
			#endregion
			
			#region Navegador de elementos
			// Seteo el proveedor de la instancia que implementa
			// la interface IListaNavegable
			this.navegadorElementos.ProveedorListaNavegable = delegate {
				return SolapaVentas.Instancia;
			};
			
			// Seteo el cargador
			this.navegadorElementos.Cargador = delegate(object o) {
				this.CargarElementoNavegable((PresupuestoDto) o);
				
				// Si despues de cargar el estado es 'Cargado', retorno true
				if (this.administradorEstado.EstaCargado)
					return true;
				
				return false;
			};
			#endregion
			
			this.VaciarFormulario();
		}
		#endregion
		
		#region Propiedades
		public static SolapaPresupuestos InstanciaActual {
			get { return instanciaActual; }
		}
		
		public AdministradorEstadoObjeto AdministradorEstado {
			get { return this.administradorEstado; }
		}
		#endregion
		
		#region Eventos
		protected virtual void OnImprimirClicked (object sender, System.EventArgs e)
		{
			/* Antes de imprimir, verifico que el presupuesto esté guardado para tener
			 * acceso al código del mismo. */
			if (!this.administradorEstado.EstaCargado &&
			    	!this.administradorEstado.EstaGuardado) {
				SMessageDialog
					.MensajeAdvertencia("El presupuesto no está guardado. Guárdelo antes de " +
					                    "imprimirlo.");
				return;
			}
			
			// Valido los datos cargados en el Presupeusto
			try {
				this.ValidarDatosIngresados();
			}
			catch(DatosIngresadosInvalidosException ex) {
				SMessageDialog.MensajeError(ex.Message);
				return;
			}
			
			// Desactivo el botón de imprimir hasta que no haya aparecido el diálogo de impresión
			this.btnImprimir.Sensitive = false;
			
//			// Leo los datos del Presupuesto actualmente cargado en la interfaz gráfica
			this.LeerDatosPresupuesto(ref this.presupuestoCargado);
			
			ImpresionDocumento impresionBorrador =
				new ImpresionDocumento(ARCHIVO_TEMPLATE_PRESUPUESTO,
				                      DIRECTORIO_ADDIN,
				                      "presupuesto",
				                      new PresupuestoWrapper(this.presupuestoCargado));
			
			impresionBorrador.ImpresionEjecutada += delegate {
				this.btnImprimir.Sensitive = true;
			};
			
			try {
				impresionBorrador.Imprimir();
			}
			catch {
				// Si hay algún problema, vuelvo a habilitar el botón para imprimir
				// y relanzo la excepción.
				this.btnImprimir.Sensitive = true;
				throw;
			}
		}
		
		protected virtual void OnGuardarPresupuesto (object sender, System.EventArgs e)
		{
			// Valido los datos cargados en el Presupeusto
			try {
				this.ValidarDatosIngresados();
			}
			catch(DatosIngresadosInvalidosException ex) {
				SMessageDialog.MensajeError(ex.Message);
				return;
			}
			
			string mensajeError = null;
			
			try {
				ControladorPresupuestosPrx controlador = 
					((ControladorPresupuestosPrx)MainIce.GetControlador("ControladorPresupuestos"));
				
				// Si no hay un presupuesto guardado, entonces creo uno nuevo
				if (this.presupuestoCargado == null)
					this.presupuestoCargado = new PresupuestoDto();
				
				this.LeerDatosPresupuesto(ref this.presupuestoCargado);
				
				// Guardo el presupuesto si es uno nuevo
				if (this.presupuestoCargado.id <= 0) {
					PresupuestoDto presupuestoGuardado = (PresupuestoDto)
						controlador.Guardar(this.presupuestoCargado);
					
					// Vuelvo a cargar el presupuesto en la interfaz gráfica.
					this.CargarPresupuesto(presupuestoGuardado, true);
					
					this.administradorEstado
						.CambiarEstadoObjeto(EstadoObjeto.Guardado);
					
					SMessageDialog
						.MensajeInformacion("El presupuesto se ha creado correctamente.");
				}
				
				// O lo modifico si ya existe
				else {
					PresupuestoDto presupuestoGuardado = (PresupuestoDto)
						controlador.Modificar(this.presupuestoCargado);
					
					// Vuelvo a cargar el presupuesto en la interfaz gráfica.
					this.CargarPresupuesto(presupuestoGuardado, true);
					
					this.administradorEstado
						.CambiarEstadoObjeto(EstadoObjeto.Guardado);
				}
			}
			catch (ObjetoInvalidoException ex) {
			  	SMessageDialog.MensajeError("El Presupuesto tiene datos inválidos:\n" + ex.mensaje);
			}
			catch (ObjetoVersionIncorrectaException ex) {
			  	SMessageDialog
					.MensajeError("<b>El presupuesto que intenta modificar fue " +
						"modificado ya por otra terminal</b>.\nVuelva a cargar el " +
						"presupuesto actual y realice nuevamente las modificaciones");
				return;
			}
			catch (ObjetoInexistenteException ex) {
			  	SMessageDialog
					.MensajeError("El Presupuesto que se quiere modificar no existe en la base de datos.",
					              ex);
			}
		}
		
		protected virtual void OnNuevoClicked (object sender, System.EventArgs e)
		{
			// Si hay modificaciones hechas, pregunto al usuario si está seguro de perderlas
			if (this.administradorEstado.EstaModificado &&
			    	!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones)
				return;
			
			// Vacío el formulario
			this.VaciarFormulario();
			
			// Desactivo el navegador de elementos
			this.navegadorElementos.Sensitive = false;
			
			// Coloco el foco en el código de la razón social
			this.razonSocial.CodigoPersonaEntry.GrabFocus();
		}
		
		protected virtual void OnFacturarClicked (object sender, System.EventArgs e)
		{
			// Valido los datos cargados en el Presupuesto
			try {
				this.ValidarDatosIngresados();
			}
			catch(DatosIngresadosInvalidosException ex) {
				SMessageDialog.MensajeError(ex.Message);
				return;
			}
			
			// Leo el Presupuesto cargado actualmente
			PresupuestoDto presupuestoCargado = new PresupuestoDto();
			this.LeerDatosPresupuesto(ref presupuestoCargado);
			
			// Creo una PreFactura según los datos del Presupuesto
			PreFacturaDto preFacturaDto = new PreFacturaDto();
			
			// Le seteo el id en 0 para tratarla como una prefactura nueva
			preFacturaDto.id = 0;
			
			// Cuando facturamos un presupuesto, no colocamos la fecha del presupuesto, sino la
			// de hoy.
			preFacturaDto.fechaTicks = DateTime.Today.Date.Ticks;
			
			// Si el presupuesto tiene asignado un cliente, copio el objeto cliente
			if (presupuestoCargado.cliente != null)
				preFacturaDto.cliente = presupuestoCargado.cliente;
			
			// Si el presupuesto no tiene un cliente asignado, copio los campos
			// que tengo del cliente en el presupuesto y se los asigno a un cliente nuevo
			else {
				preFacturaDto.cliente = new ClienteDto();
				preFacturaDto.cliente.razonSocial = presupuestoCargado.razonSocial;
				preFacturaDto.cliente.tipoRes = TipoResponsable.ConsumidorFinal;
				preFacturaDto.cliente.telefono = presupuestoCargado.telefono;
				preFacturaDto.cliente.correoElectronico = presupuestoCargado.correoElectronico;
			}
			
			// Dependiendo del tipo de responsabilidad del cliente, cambio el tipo de PreFactura
			if (preFacturaDto.cliente.tipoRes == TipoResponsable.ResponsableInscripto ||
			    	preFacturaDto.cliente.tipoRes == TipoResponsable.ResponsableNoInscripto)
				preFacturaDto.tipo = FacturaTipoVenta.FacturaA;
			else
				preFacturaDto.tipo = FacturaTipoVenta.FacturaB;
			
			// Copio el contenido de los items
			preFacturaDto.items = new List<ItemGenericoDto>();
			ItemProductoDto itemNuevo;
			
			foreach (ItemProductoDto item in presupuestoCargado.items) {
				itemNuevo = new ItemProductoDto();
				itemNuevo.id = 0;
				
				itemNuevo.precioUnitario = item.precioUnitario;
				itemNuevo.importe = item.importe;
				
				itemNuevo.cantidad = item.cantidad;
				itemNuevo.porcentajeBonificacion = item.porcentajeBonificacion;
				
				itemNuevo.descripcion = item.descripcion;
				itemNuevo.orden = item.orden;
				itemNuevo.porcentajeBonificacion = item.porcentajeBonificacion;
				
				itemNuevo.productoId = item.productoId;
				itemNuevo.productoCodigo = item.productoCodigo;
				itemNuevo.productoDescripcion = item.productoDescripcion;
				itemNuevo.productoAbstracto = item.productoAbstracto;
				itemNuevo.productoImporteSinIva = item.productoImporteSinIva;
				itemNuevo.productoImporteVenta = item.productoImporteVenta;
				itemNuevo.productoUnidad = item.productoUnidad;
				
				preFacturaDto.items.Add(itemNuevo);
			}
			
			// Copio la bonificación
			preFacturaDto.porcentajeBonificacion = presupuestoCargado.porcentajeBonificacion;
			
			preFacturaDto.observaciones = presupuestoCargado.observaciones;
			
			// Muestro la PreFactura en la interfaz de facturación
			SolapaVentas.Instancia.EjecutarVisualizador(preFacturaDto);
		}
		#endregion
		
		#region IVisualizadorNavegable implementation
		public INavegadorElementosVista NavegadorElementos {
			get { return this.navegadorElementos; }
		}
		#endregion
		
		#region Métodos públicos
		/// <summary>
		/// Carga un presupuesto en el formulario. Brinda la posibilidad de especificar si se
		/// debe hacer la pregunta por las modificaciones o no.
		/// </summary>
		/// <param name="presupuesto">
		/// A <see cref="PresupuestoDto"/>
		/// </param>
		/// <param name="evitarPreguntaPorModificaciones">
		/// A <see cref="System.Boolean"/>
		/// </param>
		public void CargarPresupuesto(PresupuestoDto presupuesto, bool evitarPreguntaPorModificaciones)
		{
			if (!evitarPreguntaPorModificaciones &&
			    	(this.administradorEstado.EstaModificado &&
			     		!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones))
				return;
			
			this.CargarPresupuestoInterno(presupuesto);
		}
		
		/// <summary>
		/// Carga un presupuesto en el formulario. Si hay modificaciones hechas sin guardar
		/// pregunta si el usuario está seguro de perderlas.
		/// </summary>
		/// <param name="presupuesto">
		/// A <see cref="PresupuestoDto"/>
		/// </param>
		public void CargarPresupuesto(PresupuestoDto presupuesto)
		{
			// Primero me fijo si no hay modificaciones sin guardar
			if (this.administradorEstado.EstaModificado &&
			    	!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones)
				return;
			
			this.CargarPresupuestoInterno(presupuesto);
		}
		
		protected void CargarPresupuestoInterno(PresupuestoDto presupuesto)
		{
			// Antes de cargar un presupuesto, vacío el formulario
			this.VaciarFormulario();
			
			this.presupuestoCargado = presupuesto;
			
			// Fecha
			this.dcFecha.Fecha = new DateTime(this.presupuestoCargado.fechaTicks);
			
			// Número del presupuesto
			this.txtNumeroPresupuesto.Text = this.presupuestoCargado.id.ToString();
			
			// Datos del cliente
			if (presupuesto.cliente == null ||
			    	presupuesto.cliente.codigo <= 0)
				this.razonSocial.RazonSocialEntry.Text = presupuesto.razonSocial;
			else
				this.razonSocial.CargarCliente(presupuesto.cliente);
			
			this.txtTelefono.Text = presupuesto.telefono;
			this.txtMail.Text = presupuesto.correoElectronico;
			this.txtBonificacion.Text = presupuesto.porcentajeBonificacion.ToString();
			
			// Items
			this.itemsProducto.Vaciar(false);
			this.itemsProducto.AgregarItems(presupuesto.items.ToArray());
			
			// Otros datos
			this.txtObservaciones.Buffer.Text = presupuesto.observaciones;
			
			// Importes
			this.txtImporteBruto.Importe = (decimal) presupuesto.importeBruto;
			this.txtImporteBonificacion.Importe = (decimal) presupuesto.importeBonificacion;
			this.txtImporteTotal.Importe = (decimal) presupuesto.importeNeto;
			
			this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Cargado);
		}
		
		protected void CargarElementoNavegable (PresupuestoDto presupuesto)
		{
			this.CargarPresupuestoInterno(presupuesto);
		}
		#endregion
		
		#region Métodos internos
		protected virtual void LeerDatosPresupuesto(ref PresupuestoDto presupuestoDto)
		{
			presupuestoDto.fechaTicks = this.dcFecha.Fecha.Ticks;
			
			// Me fijo si el cliente es nulo o si solo se ingresó la razón social
			if (razonSocial.Cliente != null) {
				presupuestoDto.cliente = this.razonSocial.Cliente;
				presupuestoDto.razonSocial = this.razonSocial.Cliente.razonSocial;
			}
			else {
				presupuestoDto.cliente = null;
				presupuestoDto.razonSocial = this.razonSocial.RazonSocialEntry.Text;
			}
			
			presupuestoDto.telefono = this.txtTelefono.Text;
			presupuestoDto.correoElectronico = this.txtMail.Text;
			
			presupuestoDto.importeBruto = this.txtImporteBruto.ImporteDouble;
			presupuestoDto.porcentajeBonificacion =
				this.txtBonificacion.Text.Equals(String.Empty) ? 0.00 :
					StratosUtil.ParseDouble(this.txtBonificacion.Text, 2);
			presupuestoDto.importeBonificacion = this.txtImporteBonificacion.ImporteDouble;
			presupuestoDto.importeNeto = this.txtImporteTotal.ImporteDouble;
			presupuestoDto.importeFinal = presupuestoDto.importeNeto;
			presupuestoDto.observaciones = this.txtObservaciones.Buffer.Text;
			presupuestoDto.items = this.tvItemsProducto.Items.Cast<ItemGenericoDto>().ToList();
		}
		
		/// <summary>
		/// Quita los datos extra del cliente, como el teléfono y el
		/// mail.
		/// </summary>
		protected virtual void VaciarDatosClienteExtra()
		{
			this.txtTelefono.Text = String.Empty;
			this.txtMail.Text = String.Empty;
			this.txtBonificacion.Text = String.Empty;
		}
		
		/// <summary>
		/// Vacía los datos ingresado en el formulario
		/// </summary>
		protected virtual void VaciarFormulario()
		{
			this.presupuestoCargado = null;
			
			// Activo todos los botones
			this.btnGuardar.Sensitive = true;
			this.btnFacturar.Sensitive = true;
			this.btnImprimir.Sensitive = true;
			
			this.dcFecha.Fecha = DateTime.Today;
			
			this.txtNumeroPresupuesto.Text = String.Empty;
			
			this.razonSocial.Vaciar();
			
			this.VaciarDatosClienteExtra();
			
			this.itemsProducto.Vaciar();
			
			this.txtObservaciones.Buffer.Text = String.Empty;
			
			this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Nuevo);
			
			this.CalcularImportes();
		}
		
		/// <summary>
		/// Verifica los datos ingresados en SolapaFacturación y lanza una excepción del tipo
		/// DatosIngresadosInvalidosException en caso de haber errores.
		/// </summary>
		protected void ValidarDatosIngresados()
		{
			if (!this.txtTelefono.TextoValido)
				throw new DatosIngresadosInvalidosException(this.txtTelefono.MensajeError);
			
			if (!this.txtMail.TextoValido)
				throw new DatosIngresadosInvalidosException(this.txtMail.MensajeError);
			
			if (!this.txtBonificacion.TextoValido)
				throw new DatosIngresadosInvalidosException(this.txtBonificacion.MensajeError);
			
			// Chequeo que haya items agregados al treeview
			if (this.itemsProducto.TreeView.Vacio)
				throw new DatosIngresadosInvalidosException("No hay items ingresados. " + 
				          "Un presupuesto válido debe tener al menos un ítem ingresado.");
			
			// Chequeo que todos los items ingresado sean válidos
			if (!this.itemsProducto.TreeView.ItemsValidos())
				throw new DatosIngresadosInvalidosException("Uno o mas items son inválidos. " + 
				          "Verifique que todos los items ingresados sean válidos.");
		}
		
		protected virtual void CalcularImportes()
		{
			decimal importeBruto;
			decimal porcentajeBonificacion = 0.00m;
			decimal importeBonificacion;

			importeBruto = StratosUtil.RoundDecimal(this.tvItemsProducto.SumaImportesItemes(), 2);
			this.txtImporteBruto.Importe = importeBruto;
			
			if (!this.txtBonificacion.Text.Equals("") && this.txtBonificacion.TextoValido)
				porcentajeBonificacion = (StratosUtil.ParseDecimal(this.txtBonificacion.Text, 2));
			    
			importeBonificacion = StratosUtil.RoundDecimal(importeBruto * (porcentajeBonificacion / 100.0m), 2);
			this.txtImporteBonificacion.Importe = importeBonificacion;
			
			this.txtImporteTotal.Importe = importeBruto - importeBonificacion;
		}
		
		/// <summary>
		/// Permite cargar el cliente en la interfaz gráfica.
		/// </summary>
		/// <param name="cliente">
		/// A <see cref="ClienteDto"/>
		/// </param>
		protected void CargarDatosCliente(ClienteDto cliente)
		{
			this.txtMail.Text = cliente.correoElectronico == null ? "" : cliente.correoElectronico;
			this.txtTelefono.Text = cliente.telefono == null ? "" : cliente.telefono;
			this.txtBonificacion.Text = StratosUtil.DoubleToString(cliente.porcentajeBonificacion);
		}

		#endregion	
	}
}
