// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;

using StratosClient.Common;
using StratosClient.Core.Gui;
using StratosClient.Core.Gui.ImpresionBorrador;
using Stratos.ModeloDto;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;
using Stratos.Common;

namespace StratosClient.Ventas.Recibos
{
	/// <summary>
	/// Presenter para ReciboVista. Tiene toda la lógica de negocios para las operaciones
	/// de creación de recibos.
	/// </summary>
	public class ReciboPresenter
	{
		#region Campos
		protected static ReciboPresenter instancia;
		
		private readonly string DIRECTORIO_ADDIN;
		private readonly string ARCHIVO_TEMPLATE_RECIBO;
		
		protected event System.Action ReciboNuevo;
		protected event System.Action ReciboCargado;
		protected event System.Action ReciboGuardado;
		
		protected AdministradorEstadoObjeto administradorEstado;
		
		protected ReciboDto reciboCargado;
		
		protected readonly IReciboVista reciboVista;
		protected ControladorRecibosPrx controladorRecibos;
		protected ControladorClientesPrx controladorClientes;
		#endregion
		
		#region Constructores
		public ReciboPresenter(IReciboVista reciboVista,
		                       ControladorRecibosPrx controladorRecibos,
		                       ControladorClientesPrx controladorClientes)
		{
			this.reciboVista = reciboVista;
			this.controladorRecibos = controladorRecibos;
			this.controladorClientes = controladorClientes;
			
			DIRECTORIO_ADDIN =
				"Addins" + System.IO.Path.DirectorySeparatorChar +
				"Ventas.Recibos" + System.IO.Path.DirectorySeparatorChar;
			
			ARCHIVO_TEMPLATE_RECIBO =
				DIRECTORIO_ADDIN + "templateRecibos.vm";
			
			// Lista de recibos
			this.reciboVista.ListaItemsReciboVista.ItemsCambiados += this.CalcularImportes;
			
			// Creo el administrador de estado de los recibos
			this.administradorEstado =
				new AdministradorEstadoObjeto(estado => this.reciboVista.Estado = estado,
				                              activarGuardar => this.reciboVista.EsPosibleGuardarRecibo = activarGuardar);
			
			this.ReciboNuevo += delegate { this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Nuevo); };
			this.ReciboCargado += delegate { this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Cargado); };
			this.ReciboGuardado += delegate { this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Guardado); };
			
			// Cada vez que se carga un recibo recalculo los importes.
			this.ReciboCargado += delegate {
				 this.CalcularImportes();
			};
			
			#region Navegador de elementos
			// Seteo el proveedor de la instancia que implementa
			// la interface IListaNavegable
			this.reciboVista.NavegadorElementos.ProveedorListaNavegable = delegate {
				return SolapaVentas.Instancia;
			};
			
			// Seteo el cargador
			this.reciboVista.NavegadorElementos.Cargador = delegate(object o) {
				this.CargarElementoNavegable((ReciboDto) o);
				
				// Si despues de cargar el estado es 'Cargado', retorno true
				if (this.administradorEstado.EstaCargado)
					return true;
				
				return false;
			};
			#endregion
			
			this.NuevoRecibo();
			
			instancia = this;
		}
		#endregion
		
		#region Propiedades
		public static ReciboPresenter Instancia {
			get { return instancia; }
		}
		
		public IReciboVista ReciboVista {
			get { return this.reciboVista; }
		}
		
		public AdministradorEstadoObjeto AdministradorEstado {
			get { return this.administradorEstado; }
		}
		
		public ControladorRecibosPrx ControladorRecibos {
			get { return this.controladorRecibos; }
			set { this.controladorRecibos = value; }
		}
		
		public ControladorClientesPrx ControladorClientes {
			get { return this.controladorClientes; }
			set { this.controladorClientes = value; }
		}
		#endregion
		
		#region Eventos
		/// <summary>
		/// Este método es ejecutado por la Vista cuando los datos son cambiado en
		/// la interfaz gráfica.
		/// </summary>
		public void DatosModificados()
		{
			this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Modificado);
		}
		
		/// <summary>
		/// Prepara la interfaz para ingresar otro recibo.
		/// </summary>
		public void NuevoRecibo()
		{
			// Si hay modificaciones hechas, pregunto al usuario si está seguro de perderlas
			if (this.administradorEstado.EstaModificado &&
			    	!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones)
				return;
			
			// Si está seguro, continúo y vacío el formulario
			this.Vaciar();
			
			// Desactivo el navegador de elementos
			this.reciboVista.NavegadorElementos.Sensitive = false;
			
			// Disparo el evento si hay suscriptos
			if (this.ReciboNuevo != null)
				this.ReciboNuevo();
			
			// Coloco el foco en el código de la razón social
			this.reciboVista.DarElFoco_CodigoRazonSocial();
		}
		
		public void ImprimirRecibo()
		{
			if (!this.administradorEstado.EstaGuardado &&
			    	!this.administradorEstado.EstaCargado) {
				SMessageDialog
					.MensajeInformacion("<b>Solo se puede imprimir un recibo guardado o cargado</b>\n\n" +
						"Guarde el recibo antes de imprimirlo.");
				
				return;
			}
			
			// Desactivo el botón para imprimir
			this.reciboVista.EsPosibleImprimirRecibo = false;
			
			ImpresionDocumento impresionBorrador =
				new ImpresionDocumento(ARCHIVO_TEMPLATE_RECIBO,
				                      DIRECTORIO_ADDIN,
				                      "recibo",
				                      new ReciboWrapper(this.reciboCargado));
			
			impresionBorrador.ImpresionEjecutada += delegate {
				this.reciboVista.EsPosibleImprimirRecibo = true;
			};
			
			try {
				impresionBorrador.Imprimir();
			}
			catch {
				// Si hay algún problema, vuelvo a habilitar el botón para imprimir
				// y relanzo la excepción.
				this.reciboVista.EsPosibleImprimirRecibo = true;
				throw;
			}
		}
		#endregion
		
		#region Métodos
		protected void CargarElementoNavegable(ReciboDto recibo)
		{
			this.CargarRecibo(recibo, true);
		}
		
		/// <summary>
		/// Carga un recibo existente en la interfaz del usuario.
		/// </summary>
		/// <param name="recibo">
		/// A <see cref="ReciboDto"/>
		/// </param>
		public void CargarRecibo(ReciboDto recibo)
		{
			// Primero me fijo si no hay modificaciones sin guardar
			if (this.administradorEstado.EstaModificado &&
			    	!this.administradorEstado.UsuarioEstaSeguroDePerderModificaciones)
				return;
			
			this.CargarRecibo(recibo, true);
		}
		
		/// <summary>
		/// Carga un recibo existente en la interfaz del usuario.
		/// Permite especificar si se dispara o no el evento de recibo cargado.
		/// </summary>
		/// <param name="recibo">
		/// A <see cref="ReciboDto"/>
		/// </param>
		/// <param name="dispararEventoReciboCargado">
		/// A <see cref="System.Boolean"/>
		/// </param>
		public void CargarRecibo(ReciboDto recibo, bool dispararEventoReciboCargado)
		{
			this.Vaciar();
			
			this.reciboCargado = recibo;
			
			// Fecha
			this.reciboVista.DateComboVista.Fecha = new DateTime(this.reciboCargado.fechaTicks);
			this.reciboVista.EsPosibleModificarFecha = false;
			
			// Razón social
			this.reciboVista.RazonSocialVista.Persona = this.reciboCargado.cliente;
			this.reciboVista.RazonSocialVista.SoloLectura = true;
			
			// Concepto
			this.reciboVista.Concepto = this.reciboCargado.concepto;
			this.reciboVista.EsPosibleModificarConcepto = false;
			
			// Lista de recibos
			this.reciboVista.ListaItemsReciboVista.Vaciar(false);
			this.reciboVista.ListaItemsReciboVista.AgregarItems(this.reciboCargado.items.ToArray());
			this.reciboVista.EsPosibleModificarItemsRecibo = false;
			
			this.reciboVista.EsPosibleGuardarRecibo = false;
			
			// Disparo el evento de recibo cargado
			if (dispararEventoReciboCargado && this.ReciboCargado != null)
				this.ReciboCargado();
		}
		
		/// <summary>
		/// Guarda un recibo.
		/// </summary>
		public void GuardarRecibo ()
		{
			try {
				ReciboDto reciboNuevo = new ReciboDto();
				this.LeerDatosRecibo(ref reciboNuevo);
				
				// TODO: hay que volver a cargar el recibo recientemente guardado
				ReciboDto reciboGuardado = (ReciboDto)
					this.controladorRecibos.Guardar(reciboNuevo);
				
				this.reciboVista.NotificarInformacion("El recibo se ha guardado correctamente.");
				
				this.CargarRecibo(reciboGuardado, false);
				
				if (this.ReciboGuardado != null)
					this.ReciboGuardado();
			}
			catch (ObjetoInvalidoException ex) {
				this.reciboVista.NotificarError("El recibo tiene datos inválidos:\n" +
				                                ex.mensaje);
			}
			catch (ObjetoVersionIncorrectaException ex) {
			  	SMessageDialog
					.MensajeError("<b>El recibo que intenta modificar fue " +
						"modificado ya por otra terminal</b>.\nVuelva a cargar el " +
						"recibo actual y realice nuevamente las modificaciones");
				return;
			}
			catch (ObjetoInexistenteException ex) {
				this.reciboVista
					.NotificarError("El recibo que se quiere modificar no existe " +
					                "en la base de datos.", ex);
			}
		}
		
		protected void Vaciar()
		{
			this.reciboCargado = null;
			
			this.reciboVista.EsPosibleModificarFecha = true;
			this.reciboVista.DateComboVista.Fecha = DateTime.Today;
			
			this.reciboVista.RazonSocialVista.SoloLectura = false;
			this.reciboVista.RazonSocialVista.Vaciar();
			
			this.reciboVista.EsPosibleModificarConcepto = true;
			this.reciboVista.Concepto = String.Empty;
			
			this.reciboVista.EsPosibleModificarItemsRecibo = true;
			this.reciboVista.ListaItemsReciboVista.Vaciar();
			
			this.reciboVista.EsPosibleGuardarRecibo = true;
			
			this.administradorEstado.CambiarEstadoObjeto(EstadoObjeto.Nuevo);
			
			this.CalcularImportes();
		}
		
		/// <summary>
		/// Calcula el importe del recibo.
		/// </summary>
		protected void CalcularImportes()
		{
			this.reciboVista.Importe = this.reciboVista.ListaItemsReciboVista.SumaImportesItemes();
		}
		
		/// <summary>
		/// Lee los datos cargados en la interfaz gráfica y devuelve
		/// un objeto ReciboDto.
		/// </summary>
		/// <param name="recibo">
		/// A <see cref="ReciboDto"/>
		/// </param>
		protected void LeerDatosRecibo(ref ReciboDto recibo)
		{
			recibo.fechaTicks = this.reciboVista.DateComboVista.Fecha.Ticks;
			recibo.cliente = this.reciboVista.RazonSocialVista.Persona;
			recibo.concepto = this.reciboVista.Concepto;
			
			recibo.items = this.reciboVista.ListaItemsReciboVista.ListaItems.Cast<ItemGenericoDto>().ToList();
			recibo.importe = (double) this.reciboVista.Importe;
		}
		#endregion
	}
}
