// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Mono.Addins;
using log4net;

using Stratos.ModeloDto;
using Stratos.QueryObjects;
using StratosClient.Core.Gui;
using StratosServer.Controladores;
using StratosClient.Widgets.Level2;

namespace StratosClient.Ventas
{
	/// <summary>
	/// Esta clase define los campos que se pueden definir en el XML del Addin de un
	/// tipo de consulta.
	/// </summary>
	public class TipoConsultaNode : ExtensionNode
	{
		#region Campos
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);
		
		/// <value>
		/// Este es el nombre de la consulta que se mostrará en el ComboBox 
		/// </value>
		[NodeAttribute]
		private string nombreConsulta;
		
		/// <value>
		/// Este valor posee el nombre de la clase que implementa la interfaz IConsulta
		/// </value>
		[NodeAttribute]
		private string consulta;
		
		/// <value>
		/// Este es un valor opcional, y permite a los addins definir el nombre de una clase que
		/// sea Gtk.Bin e implemente la interfaz ICriteriosPersonalizados.
		/// </value>
		[NodeAttribute]
		private string criteriosPersonalizados;
		
		/// <value>
		/// Indica el objeto TreeView para mostrar la lista de elementos.
		/// </value>
		[NodeAttribute]
		private string treeView;
		
		private Dictionary<string, ColeccionElementos> treeViewsColeccionElementos;
		#endregion
		
		#region Constructores
		public TipoConsultaNode()
		{
			this.treeViewsColeccionElementos = new Dictionary<string, ColeccionElementos>();
		}
		#endregion
		
		#region Propiedades
		public string NombreConsulta {
			get { return this.nombreConsulta; }
		}
		
		public IConsulta Consulta {
			get {
				// Verifico que el objeto consulta sea de tipo IConsulta
				object temp = Addin.CreateInstance(this.consulta);
				IConsulta objetoConsulta = null;
				
				if (!(temp is IConsulta)) {
					logger.Debug("El objeto consulta especificado no es de tipo IConsulta");
					return null;
				}
				
				objetoConsulta = (IConsulta) temp;
				
				return objetoConsulta;
			}
		}
		
		public Gtk.Bin CriteriosPersonalizados {
			get {
				// Solo avanzo si los criterios personalizados fueron seteados
				if (this.criteriosPersonalizados == null ||
				    	this.criteriosPersonalizados.Trim() == String.Empty)
					return null;
				
				object temp = Addin.CreateInstance(this.criteriosPersonalizados);
				Gtk.Bin widgetCriteriosPersonalizados = null;
				
				// Me aseguro que sean del tipo correcto
				if (!(temp is Gtk.Bin) && !(temp is ICriteriosPersonalizados)) {
					logger.Debug("Los criterios personalizados especificados no son del tipo Gtk.Bin o no implementan la interfaz ICriteriosPersonalizados");
					return null;
				}
				
				widgetCriteriosPersonalizados = (Gtk.Bin) temp;
				widgetCriteriosPersonalizados.ShowAll();
				
				return widgetCriteriosPersonalizados;
			}
		}
		
		public ColeccionElementos TreeView {
			get {
				/* FIXME: Si es lento crear un TreeView nuevo cada vez que se cambia el tipo
				 * de documento, se podría cachearlo con un Diccionario. Ver también si esto
				 * ya no lo hace Mono.Addins. */
				
				// Verifico que se haya establecido un objeto ColeccionElementos
				if (this.treeView == null ||
				    	this.treeView.Trim() == String.Empty)
					return null;
				
				object temp = Addin.CreateInstance(this.treeView);
				ColeccionElementos coleccionElementos = null;
				
				// Me aseguro que el tipo establecido en 'treeView' sea derivado
				// de ColeccionElementos
				if (!(temp is ColeccionElementos)) {
					logger.Debug("El TreeView especificado en la propiedad 'treeView' no es de tipo CustomWidgets.ColeccionElementos");
					return null;
				}
				
				coleccionElementos = (ColeccionElementos) temp;
//				coleccionElementos.ShowAll();
				
				return coleccionElementos;
			}
		}
		#endregion
	}
}
