// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

using Gtk;
using Mono.Addins;
using log4net;

using StratosClient.Core.Gui;
using Stratos.ModeloDto;
using Stratos.QueryObjects;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;
using StratosClient.Widgets.Level3;
using StratosClient.Widgets.Level2;

namespace StratosClient.Ventas
{
	public partial class SolapaVentas : Gtk.Bin, IListaNavegable
	{
		#region Campos
		public event System.Action ListaElementosVaciada;
		
		private const int TAMAÑO_PAGINA = 25;
		
		private ColeccionElementos tvDocumentos;
		
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);
		
		/// <summary>
		/// Mapea el nombre del visualizador con el numero de solapa que le
		/// corresponde.
		/// </summary>
		private Dictionary<string, int> mapeoComponenteSolapa;
		
		/// <summary>
		/// Mapea los diferentes tipos de consultas (los nombres de las consultas)
		/// con el número de solapa en el que se encuentra el TreeView correspondiente
		/// para listar los elementos.
		/// </summary>
		private Dictionary<string, object[]> mapeosTipoConsultaSolapa;
		
		/// <summary>
		/// Mapeo el tipo de documento con el Gtk.Menu asociado que emergerá
		/// cuando se haga click derecho sobre algún documento de la lista.
		/// </summary>
		private Dictionary<string, Gtk.Menu> mapeosTipoDocumentoMenu;
		
		/// <summary>
		/// Mapeo un tipo de documento con un array de objetos de dos dimensiones:
		/// el primero tiene la toolbar correspondientes y el segundo el número de
		/// solapa donde se encuentra la toolbar de acciones para el tipo de documento.
		/// </summary>
		private Dictionary<string, object[]> mapeosTipoDocumentoToolbar;
		
		/// <summary>
		/// Este diccionario se utiliza para guardar los criterios de búsqueda que se especificaron
		/// en la interfaz de búsqueda. De esta forma se pasa luego este diccionario a los Addins
		/// que hayan suscripto alguna consulta, para que puedan realizar la búsqueda según dichos
		/// criterios.
		/// </summary>
		private Dictionary<string,object> criteriosBusquedaActuales;
		
		/// <summary>
		/// Además de los criterios estándar que provee StratosClient.Ventas, también un Addin
		/// puede agregar un criterio personalizado. Para esto puede agregar un Gtk.Bin a la
		/// GUI y además que el mismo implemente la interfaz ICriteriosPersonalizados. Esta
		/// interfaz tiene unos métodos especiales para realizar la búsqueda.
		/// </summary>
		private ICriteriosPersonalizados criteriosBusquedaPersonalizados;
		
		private TreeIter iterCargado;
        private int paginaIterCargado;
		private ServicioNavegadorElementos servicioNavegador;
		
		private static SolapaVentas instancia;
		#endregion
		
		#region Constructores
		public SolapaVentas()
		{
			this.Build();
			
			this.criteriosBusquedaActuales = new Dictionary<string, object>();
			
			#region Criterios de búsqueda
			this.chkRazonSocial.Toggled += delegate {
				this.txtRazonSocial.Sensitive = this.chkRazonSocial.Active;
			};
			this.chkRazonSocial.Active = false;
			
			this.dcFechaMinima.Fecha = DateTime.Today.Date;
			this.chkFechaMinima.Toggled += delegate {
				this.dcFechaMinima.Sensitive = this.chkFechaMinima.Active;
			};
			this.chkFechaMinima.Active = false;
			
			this.dcFechaMaxima.Fecha = DateTime.Today.Date;
			this.chkFechaMaxima.Toggled += delegate {
				this.dcFechaMaxima.Sensitive = this.chkFechaMaxima.Active;
			};
			this.chkFechaMaxima.Active = false;
			#endregion
			
			// Creo el servicio de navegacion de elementos
			this.servicioNavegador = new ServicioNavegadorElementos();
			
			// Este objeto se debe crear antes del método ConstruirSolapas
			this.mapeoComponenteSolapa = new Dictionary<string, int>();
			
			this.mapeosTipoConsultaSolapa = new Dictionary<string, object[]>();
			
			this.mapeosTipoDocumentoMenu = new Dictionary<string, Menu>();
			
			this.mapeosTipoDocumentoToolbar = new Dictionary<string, object[]>();
			
			this.ConstruirSolapas();
			this.CargarTiposConsultas();
			
			EventosGlobales.ElementoEliminado += delegate(DocumentoIdDto obj) {
				TipoConsultaNode tipoConsulta = this.GetTipoConsultaSeleccionado();
				
				if (!tipoConsulta.Consulta.TipoDevuelto.IsInstanceOfType(obj))
					return;
				
				this.navegadorPaginas.RecargarPaginaActual();
			};
			
			instancia = this;
		}
		#endregion
		
		#region Propiedades
		public ColeccionElementos TreeViewListado {
			get { return this.tvDocumentos; }
		}
		
		public static SolapaVentas Instancia {
			get { return instancia; }
		}
		
		protected IEnumerable<VentasTabNode> ListaTabs {
			get {
				var listaOrdenada =
					from tabNode in AddinManager.GetExtensionNodes("/StratosClient/Ventas/Tabs").Cast<VentasTabNode>()
					orderby tabNode.Posicion ascending
					select tabNode;
				
				return listaOrdenada;
			}
		}
		#endregion
		
		#region IListaNavegable Implementation
		public ServicioNavegadorElementos ServicioNavegador {
			get {
				this.servicioNavegador.ListaNavegable = this;
				return this.servicioNavegador;
			}
		}
		
		public TreeIter IterCargado {
			get { return this.iterCargado; }
			set { this.iterCargado = value; }
		}
		
		public int PaginaIterCargado {
			get { return this.paginaIterCargado; }
			set { this.paginaIterCargado = value; }
		}

		public ColeccionElementos ListaElementos {
			get { return this.tvDocumentos;	}
		}
		
		public NavegadorPaginas NavegadorDePaginas {
			get { return this.navegadorPaginas; }
		}
		
		public object ObtenerObjetoRefrescado(TreeIter iter)
		{
			return this.ObtenerDocumentoRefrescado(iter);
		}                                   
		#endregion
		
		#region Eventos
		[GLib.ConnectBefore]
		protected void OnButtonPressEvent(object o, ButtonPressEventArgs args)
		{
			bool retval = false;
			
			// Solo activo el menú si es el botón derecho del ratón el presionado
			if (args.Event.Type == Gdk.EventType.ButtonPress && args.Event.Button == 3) {
				this.MostrarPopUp();
				retval = true;
			}
			
			args.RetVal = retval;
		}
		
		private void OnDocumentosActivated(object o, RowActivatedArgs args)
		{
			TreeIter iter;
			this.tvDocumentos.Model.GetIter(out iter, args.Path);
			
			this.EjecutarVisualizador(iter);
		}
		
		protected virtual void OnBuscar (object sender, System.EventArgs e)
		{
			// Verifico si los criterios de búsqueda son válidos
			if (!this.dcFechaMinima.FechaValida() || !this.dcFechaMaxima.FechaValida()) {
				SMessageDialog.MensajeError("Las fechas son inválidas");
				return;
			}
			
			// Valido que las fechas máxima y mínima sean consistentes
			if ((this.chkFechaMaxima.Active && this.chkFechaMinima.Active) &&
			    	this.dcFechaMaxima.Fecha < this.dcFechaMinima.Fecha) {
				SMessageDialog.MensajeError("La fecha máxima no puede ser inferior a la fecha mínima");
				return;
			}
			
			// Valido los criterios personalizados
			string mensajeError;
			if (this.criteriosBusquedaPersonalizados != null &&
			    	!this.criteriosBusquedaPersonalizados.DatosValidos(out mensajeError)) {
				SMessageDialog.MensajeError(mensajeError);
				return;
			}
			
			// Creo la consulta paginable para cargar la primer página
			ConsultaPaginable consultaPaginable = new ConsultaPaginable();
			consultaPaginable.pagina = 0;
			consultaPaginable.tamanioPagina = TAMAÑO_PAGINA;
			
			// El criterio de ordenación será la fecha
			consultaPaginable.criteriosOrdenacion = new List<CriterioOrdenacion>();
			CriterioOrdenacion criterioOrdenacion = new CriterioOrdenacion();
			criterioOrdenacion.campoOrdenacion = "Fecha";
			criterioOrdenacion.sentidoOrd = SentidoOrdenacion.Desc;
			consultaPaginable.criteriosOrdenacion.Add(criterioOrdenacion);
			
			/* Preparo los criterios de búsqueda para que sean utilizados. Si los CheckBox
			 * no están activados no coloco los valores en los criterios. */
			this.criteriosBusquedaActuales.Clear();
			
			if (this.chkRazonSocial.Active && this.txtRazonSocial.Text != String.Empty)
				this.criteriosBusquedaActuales
					.Add("RazonSocial", this.txtRazonSocial.Text);
			
			if (this.chkFechaMinima.Active)
				this.criteriosBusquedaActuales
					.Add("FechaMinima", this.dcFechaMinima.Fecha);
			
			if (this.chkFechaMaxima.Active)
				this.criteriosBusquedaActuales
					.Add("FechaMaxima", this.dcFechaMaxima.Fecha);
			
			// Si fueron establecidos, agrego a los criterios de búsqueda de todos los
			// criterios personalizados
			if (this.criteriosBusquedaPersonalizados != null) {
				foreach(KeyValuePair<string,object> kvp
				        in this.criteriosBusquedaPersonalizados.CriteriosPersonalizados)
					this.criteriosBusquedaActuales.Add(kvp.Key, kvp.Value);
			}
			
			// Realizo la consulta y muestro los resultados
			object[] resultados = this.CargadorDocumentos(consultaPaginable);
			
			this.tvDocumentos.Vaciar();
			
			if (resultados.Length == 0)
				SMessageDialog.MensajeAdvertencia("No se encontraron resultados según los criterios de búsqueda");
			
			this.tvDocumentos.AgregarElementosPrimeraPagina(resultados);
		}
		
		protected virtual void OnTipoConsultaChanged (object sender, System.EventArgs e)
		{
			// Quito todos los criterios personalizados
			foreach(Gtk.Widget w in this.vbCriteriosPersonalizados.AllChildren)
				this.vbCriteriosPersonalizados.Remove(w);
			
			TipoConsultaNode tipoConsulta = this.GetTipoConsultaSeleccionado();
			
			// Si el tipo de consulta ha especificado criterios personalizados, los cargo
			Gtk.Bin criteriosPersonalizados = tipoConsulta.CriteriosPersonalizados;
			
			if (criteriosPersonalizados != null) {
				// Agrego los criterios a la GUI
				this.vbCriteriosPersonalizados.PackEnd(criteriosPersonalizados, false, false, 0);
				
				// Almaceno el objeto ICriteriosPersonalizados para luego armar los
				// criterios de búsqueda cuando el usuario presione el botón Buscar
				this.criteriosBusquedaPersonalizados = (ICriteriosPersonalizados) criteriosPersonalizados;
			}
			
			// Cargo las acciones para el tipo de documento en la toolbar
			object[] objetos = this.mapeosTipoDocumentoToolbar[tipoConsulta.Consulta.TipoDevuelto.FullName];
			this.nbAcciones.CurrentPage = (int) objetos[1];
			
			/* Obtengo los parámetros según el tipo de consulta seleccionado (el TreeView
			 * y el número de página dentro del Notebook). */
			objetos = this.mapeosTipoConsultaSolapa[tipoConsulta.NombreConsulta];
			
			this.tvDocumentos = (ColeccionElementos) objetos[0];
			
			this.nbLista.CurrentPage = (int) objetos[1];
			
			// Cuando el combo cambia, elimino todos los items del treeview y todos los
			// criterios personalizados de búsqueda
			this.tvDocumentos.Vaciar();
			
			this.ActualizarInformacionAccionadores();
			
			// Reseteo el iter cargado
			this.iterCargado = new TreeIter();
			
			// Desactivo el navegador de páginas
			this.navegadorPaginas.Sensitive = false;
			
			this.OnTreeViewChanged(this.tvDocumentos);
		}
		
		/// <summary>
		/// Este método se debe disparar cada vez que el TreeView cambia, ya que realiza
		/// una serie de operaciones de configuración sobre el mismo que son necesarias.
		/// </summary>
		/// <param name="coleccionElementos">
		/// A <see cref="ColeccionElementos"/>
		/// </param>
		private void OnTreeViewChanged(ColeccionElementos coleccionElementos)
		{
			// Navegador de páginas
			this.navegadorPaginas.TamañoPagina = TAMAÑO_PAGINA;
			this.navegadorPaginas.ColeccionElementos = this.tvDocumentos;
			this.navegadorPaginas.CargadorDatos = this.CargadorDocumentos;
		}
		
		protected virtual void OnCriteriosBusquedaKeyRelease (object o, Gtk.KeyReleaseEventArgs args)
		{
			if (args.Event.Key == Gdk.Key.Return || args.Event.Key == Gdk.Key.KP_Enter)
				this.OnBuscar(null, null);
		}
		
		protected virtual void OnTvDocumentosSelectionChanged(object o, System.EventArgs e)
		{
			this.ActualizarInformacionAccionadores();
		}
		#endregion
		
		#region Métodos
		protected void ActualizarInformacionAccionadores()
		{
			TreeIter[] iters;
			DocumentoGenericoDto[] documentos =
				this.tvDocumentos.ObtenerSeleccionados(out iters).Cast<DocumentoGenericoDto>().ToArray();
			
			TipoConsultaNode tipoConsultaActual = this.GetTipoConsultaSeleccionado();
			
			var accionadores =
				from accionMenu in
					AddinManager.GetExtensionNodes("/StratosClient/Ventas/AccionesSobreElemento")
						.Cast<AccionSobreElementoNode>()
				where accionMenu.TipoDocumento == tipoConsultaActual.Consulta.TipoDevuelto.FullName
				select accionMenu.Accionador;
			
			foreach (IAccionadorSobreElemento accionador in accionadores)
				accionador.ActualizarAcciones(iters, documentos, this.tvDocumentos.Model);
		}
		
		protected void MostrarPopUp()
		{
			TipoConsultaNode tipoConsultaActual = this.GetTipoConsultaSeleccionado();
			
			// Busco addins que hayan registrado una acción de menú para el tipo de documento
			// seleccionado.
			var accionadoresMenu =
				from accionMenu in
					AddinManager.GetExtensionNodes("/StratosClient/Ventas/AccionesSobreElemento")
						.Cast<AccionSobreElementoNode>()
				where accionMenu.TipoDocumento == tipoConsultaActual.Consulta.TipoDevuelto.FullName &&
					accionMenu.MostrarMenu
				select accionMenu.Accionador;
			
			if (accionadoresMenu.Count() == 0) {
				logger.Debug("No existen acciones de menú asociadas al tipo de documento " +
				             tipoConsultaActual.Consulta.TipoDevuelto.FullName);
				return;
			}
			
			// Creo el menu con las acciones (Gtk.Action) correspondientes
			Gtk.Menu menu = null;
			IAccionadorSobreElemento unAccionador;
			
			// Me fijo si el objeto menú ya no está creado.
			if (this.mapeosTipoDocumentoMenu.ContainsKey(tipoConsultaActual.Consulta.TipoDevuelto.FullName))
				menu = this.mapeosTipoDocumentoMenu[tipoConsultaActual.Consulta.TipoDevuelto.FullName];
			
			// Si no está, lo creo según las acciones definidas en los accionadores
			else {
				menu = new Gtk.Menu();
				this.mapeosTipoDocumentoMenu[tipoConsultaActual.Consulta.TipoDevuelto.FullName] = menu;
				
				for (int i=0; i<accionadoresMenu.Count(); i++) {
					unAccionador = accionadoresMenu.ElementAt(i);
					
					foreach (Gtk.Action accion in unAccionador.Acciones)
						menu.Append(accion.CreateMenuItem());
					
					// Siempre agrego un separador al final de las acciones definidas
					// por el accionador, siempre que no sea el último.
					if (i != accionadoresMenu.Count() - 1)
						menu.Append(new SeparatorMenuItem());
				}
			}
			
			menu.Popup();
			menu.ShowAll();
		}
		
		#region Método EjecutarVisualizador
		/// <summary>
		/// Dado un TreeIter correspondiente al listado de documento, obtiene el DocumentoGenericoDto
		/// y lo carga en la interfaz gráfica que corresponda. También se encarga de actualizar
		/// dicho documento (lo pide nuevamente al servidor antes de cargarlo, por si ha sufrido
		/// modificaciones desde otra terminal).
		/// </summary>
		/// <param name="iter">
		/// A <see cref="TreeIter"/>
		/// </param>
		public void EjecutarVisualizador(TreeIter iter)
		{
			DocumentoGenericoDto documento = this.ObtenerDocumentoRefrescado(iter);
			
			if (documento == null)
				return;
			
			// Almaceno la versión refrescada en el modelo
			ListStore modelo = (ListStore) this.tvDocumentos.Model;
			TreePath path = modelo.GetPath(iter);
			
			modelo.SetValue(iter, 0, documento);
			modelo.EmitRowChanged(path, iter);
			
			// Seteo el iter del elemento cargado
			this.iterCargado = iter;
			this.paginaIterCargado = this.navegadorPaginas.PaginaActual;
			this.servicioNavegador.PathActual = path.Copy();
			
			// Ejecuto el visualizador
			this.EjecutarVisualizador(documento, true);
		}
		
		public void EjecutarVisualizador(DocumentoGenericoDto documento, bool activarNavegadorElementos)
		{
			// Obtengo el visualizador para el tipo de documento correspondiente
			VisualizadorNode visualizador = this.ObtenerVisualizadorNode(documento);
			
			if (visualizador == null)
				return;
			
			// Le pregunto al componente cargador si se permite la carga del documento
			if (!visualizador.ComponenteCargador.PermitirCarga(documento))
				return;
			
			/* Ahora necesito activar la solapa en la que se encuentra el componente visualizador
			 * y enviarle el documento al que se quiere visualizar. */
			
			// Primero obtengo el numero de solapa según el nombre del componente visualizador
			int numeroSolapa = this.mapeoComponenteSolapa[visualizador.ComponenteVisualizador];
			
			// Activo la solapa que contiene el componente visualizador correspondiente
			this.nbSolapasVentas.CurrentPage = numeroSolapa;
			
			// Llamo al componente cargador para que cargue el documento por visualizar
			visualizador.ComponenteCargador.Cargar(documento, activarNavegadorElementos);
		}
		
		/// <summary>
		/// Dado un objeto DocumentoGenericoDto (Factura, PreFactura, etc), se encarga de cargarlo
		/// en la solapa correspondiente.
		/// </summary>
		/// <param name="documento">
		/// A <see cref="DocumentoGenericoDto"/>
		/// </param>
		public void EjecutarVisualizador(DocumentoGenericoDto documento)
		{
			this.EjecutarVisualizador(documento, false);
		}
		#endregion
		
		private DocumentoGenericoDto ObtenerDocumentoRefrescado(TreeIter iter)
		{
			// Obtengo el objeto al que se la hecho doble click en el TreeView
			DocumentoGenericoDto documento =
				(DocumentoGenericoDto) this.tvDocumentos.Model.GetValue(iter, 0);
			
			// Obtengo el visualizador para el tipo de documento correspondiente
			VisualizadorNode visualizador = this.ObtenerVisualizadorNode(documento);
			
			if (visualizador == null)
				return null;
			
			/* En este punto vuelvo a pedir el documento al servidor para refrescarlo. */
			try {
				documento = (DocumentoGenericoDto)
					visualizador.ComponenteCargador.ControladorRemoto
							.Refrescar(documento, null);
			}
			catch (ObjetoInexistenteException) {
				SMessageDialog.MensajeError("El documento que intenta visualizar ya no existe más en " +
				                            "la base de datos.");
				return null;
			}
			
			return documento;
		}
		
		/// <summary>
		/// Dado un tipo de DocumentoGenericoDto, devuelve el objeto VisualizadorNode correspondiente.
		/// Se enc
		/// </summary>
		/// <param name="documento">
		/// A <see cref="DocumentoGenericoDto"/>
		/// </param>
		/// <returns>
		/// A <see cref="VisualizadorNode"/>
		/// </returns>
		private VisualizadorNode ObtenerVisualizadorNode(DocumentoGenericoDto documento)
		{
			// Busco el visualizador adecuado para el documento activado
			var visualizadores =
				AddinManager.GetExtensionNodes("/StratosClient/Ventas/Visualizadores").Cast<VisualizadorNode>();
			var solapas =
				AddinManager.GetExtensionNodes("/StratosClient/Ventas/Tabs").Cast<TabNode>();
			
			var visualizadoresCompatibles =
				from vis in visualizadores
				from tabNode in solapas
				//join tabNode in solapas on tabNode.Componente equals vis.ComponenteVisualizador
				where vis.ComponenteVisualizador == tabNode.Componente &&
					vis.TipoDocumento == documento.GetType().FullName
				select vis;
			
			// Realizo una pequeña validación de los resultados obtenidos
			// Con respecto a los visualizadores compatibles encontrados:
			if (visualizadoresCompatibles.Count() == 0) {
				logger.Debug("No se encontraron visualizadores para el tipo de documento: " +
				             documento.GetType() + " o el componente visualizador asignado no existe.");
				return null;
			}
			else if (visualizadoresCompatibles.Count() > 1)
				logger.Debug("Se encontraron más de un visualizador para el tipo de documento: " +
				             documento.GetType());
			
			// Obtengo el primer visualizador de la lista
			VisualizadorNode visualizador = visualizadoresCompatibles.First();
			
			// Valido con respecto al componente de carga:
			if (visualizador.ComponenteCargador == null) {
				logger.Debug("No se ha especificado un componente cargador para el tipo de documento: " +
				             documento.GetType());
				return null;
			}
			
			return visualizador;
		}
		
		protected void ConstruirSolapas()
		{
			List<IVisualizadorNavegable> listaVisualizadoresNavegables =
				new List<IVisualizadorNavegable>();
			
			/* Agrego las solapas y almaceno el número en el diccionario, para luego poder obtener
			 * el número de solapa según el componente que almacena. */
			foreach (VentasTabNode node in this.ListaTabs) {
				Gtk.Bin widget = null;
				try {
					widget = node.GetWidget();
				}
				catch (System.Exception ex) {
					SMessageDialog
						.MensajeError(String.Format("No se pudo cargar la sub solapa '{0}'", node.EtiquetaCadena),
						              ex);
					
					continue;
				}
				
				// Si el widget implementa la interface IVisualizadorNavegable, lo agrego a la lista
				if (widget is IVisualizadorNavegable)
					listaVisualizadoresNavegables.Add(widget as IVisualizadorNavegable);
				
				this.mapeoComponenteSolapa[node.Componente] =
					nbSolapasVentas.AppendPage(widget, node.Etiqueta);
			}
			
			// Me suscribo al evento de ListaElementosVaciada, y cuando se dispare
			// siempre deshabilito todos los objetos NavegadorElementos de todos los
			// visualizadores registrados.
			this.ListaElementosVaciada += delegate {
				foreach (IVisualizadorNavegable vis in listaVisualizadoresNavegables)
					vis.NavegadorElementos.Sensitive = false;
			};
			
			nbSolapasVentas.ShowAll();
		}
		
		protected void CargarTiposConsultas()
		{
			// Ordeno las consultas por su nombre, así aparecen siempre en el mismo orden en el
			// ComboBox.
			logger.Debug("Obteniendo tipos de consulta");
			
			var listaOrdenada =
				from tipoConsulta in AddinManager
					.GetExtensionNodes("/StratosClient/Ventas/TiposConsulta").Cast<TipoConsultaNode>()
				orderby tipoConsulta.NombreConsulta ascending
				select tipoConsulta;
			
			// Creo el modelo para el combobox de tipos de consultas
			ListStore modelo = new ListStore(typeof(TipoConsultaNode));
			
			// Por cada tipo de consulta, la agrego en el combo box correspondiente
			// y además agrego el TreeView en el objeto Notebook para ir intercambiándolo
			// a medida que se cambia el tipo de consulta.
			
			// Borro la primer página del Notebook de TreeViews y acciones.
			this.nbLista.RemovePage(0);
			this.nbAcciones.RemovePage(0);
			
			ScrolledWindow scrolledWindow;
			
			// En esta variable almaceno, primero, el TreeView y luego el número
			// de página en el Notebook
			object[] objetos;
			ColeccionElementos listaElementos;
			
			IEnumerable<IAccionadorSobreElemento> accionadoresToolbar;
			Gtk.Toolbar toolbar;
			Gtk.ToolItem toolButton;
			
			foreach (TipoConsultaNode tipoConsulta in listaOrdenada) {
				logger.DebugFormat("Procesando tipo de consulta: {0}", tipoConsulta.NombreConsulta);
				
				modelo.AppendValues(tipoConsulta);
				
				#region TreeView de objetos
				objetos = new object[2];
				
				listaElementos = tipoConsulta.TreeView;
				
				// Evento para disparar el menú desplegable
				listaElementos.ButtonPressEvent += this.OnButtonPressEvent;
				
				// Evento para cargar un documento seleccionado
				listaElementos.RowActivated += this.OnDocumentosActivated;
				
				// Suscribo el evento 'ListaElementosVaciada' de IListaNavegable al
				// evento 'ListaVaciada' de treeview
				listaElementos.ListaVaciada += delegate {
					if (this.ListaElementosVaciada != null)
						this.ListaElementosVaciada();
				};
				
				listaElementos.Selection.Changed += this.OnTvDocumentosSelectionChanged;
				listaElementos.Model.RowChanged += delegate {
					this.ActualizarInformacionAccionadores();
				};
				
				listaElementos.ShowAll();
				objetos[0] = listaElementos;
				
				// Agrego el TreeView correspondiente al Notebook que tiene los
				// TreeView de listados
				scrolledWindow = new ScrolledWindow();
				scrolledWindow.Add(listaElementos);
				
				objetos[1] = this.nbLista.AppendPage(scrolledWindow, null);
				
				this.mapeosTipoConsultaSolapa[tipoConsulta.NombreConsulta] = objetos;
				#endregion
				
				#region Acciones personalizadas
				logger.DebugFormat("Obteniendos accionadores (toolbar) para el tipo {0}", tipoConsulta.Consulta.TipoDevuelto.FullName);
				
				accionadoresToolbar =
					from accionMenu in
						AddinManager.GetExtensionNodes("/StratosClient/Ventas/AccionesSobreElemento")
							.Cast<AccionSobreElementoNode>()
					where accionMenu.TipoDocumento == tipoConsulta.Consulta.TipoDevuelto.FullName &&
						accionMenu.MostrarToolbar
					select accionMenu.Accionador;
				
				logger.DebugFormat("Retornados {0} accionadores", accionadoresToolbar.Count());
				
				objetos = new object[2];
				
				// Creo la toolbar y la configuro
				logger.Debug("Creando la toolbar");
				
				toolbar = new Gtk.Toolbar();
				toolbar.Orientation = Orientation.Horizontal;
				toolbar.ToolbarStyle = ToolbarStyle.BothHoriz;
				toolbar.IconSize = IconSize.SmallToolbar;
				
				objetos[0] = toolbar;
				
				IAccionadorSobreElemento unAccionador;
				
				// Por cada accionador, agrego todas sus acciones a la toolbar
				logger.Debug("Iterando por cada accionador");
				
				for (int i=0; i<accionadoresToolbar.Count(); i++) {
					unAccionador = accionadoresToolbar.ElementAt(i);
					
					logger.DebugFormat("Procesando accionador {0} con {1} Gtk.Actions", i, unAccionador.Acciones.Length);
					
					foreach (Gtk.Action unaAccion in unAccionador.Acciones) {
						logger.DebugFormat("Agregando Action '{0}' a la toolbar", unaAccion.Name);
						
						toolButton = (Gtk.ToolItem) unaAccion.CreateToolItem();
						toolButton.IsImportant = true;
						
						toolbar.Add(toolButton);
					}
					
					// Si no es el último, agrego un separador
					if (i != accionadoresToolbar.Count() - 1)
						toolbar.Add(new SeparatorToolItem());
				}
				
				objetos[1] = this.nbAcciones.AppendPage(toolbar, null);
				toolbar.ShowAll();
				
				this.mapeosTipoDocumentoToolbar[tipoConsulta.Consulta.TipoDevuelto.FullName] = objetos;
				#endregion
			}
			
			// Asigno el modelo al combobox 
			this.cmbTipoConsulta.Model = modelo;
			
			// Creo y seteo el CellRenderer del combobox y el renderizador
			CellRendererText renderer = new CellRendererText();
			this.cmbTipoConsulta.PackStart(renderer, true);
			
			this.cmbTipoConsulta.SetCellDataFunc(renderer, this.RenderObjeto);
			
			/* Seteo el primer item por defecto para que no quede vacío */
			TreeIter iter;
			this.cmbTipoConsulta.Model.GetIterFirst(out iter);
			this.cmbTipoConsulta.SetActiveIter(iter);
		}
		
		/// <summary>
		/// Renderiza los tipos de consulta en el combobox.
		/// </summary>
		/// <param name="layout">
		/// A <see cref="CellLayout"/>
		/// </param>
		/// <param name="renderer">
		/// A <see cref="CellRenderer"/>
		/// </param>
		/// <param name="model">
		/// A <see cref="TreeModel"/>
		/// </param>
		/// <param name="iter">
		/// A <see cref="TreeIter"/>
		/// </param>
		protected void RenderObjeto(CellLayout layout,
		                            CellRenderer renderer,
		                            TreeModel model,
		                            TreeIter iter)
		{
			TipoConsultaNode tipoConsulta = (TipoConsultaNode) model.GetValue(iter, 0);
			(renderer as CellRendererText).Text = tipoConsulta.NombreConsulta;
		}
		
		private TipoConsultaNode GetTipoConsultaSeleccionado()
		{
			TreeIter iter;
			this.cmbTipoConsulta.GetActiveIter(out iter);
			
			TipoConsultaNode tipoConsulta =
				(TipoConsultaNode) this.cmbTipoConsulta.Model.GetValue(iter, 0);
			
			return tipoConsulta;
		}
		
		/// <summary>
		/// Este método se encarga de realizar la consulta. No realiza la consulta en sí,
		/// pero si se encarga de orquestarla (wow).
		/// Es utilizado por el navegador
		/// de páginas para recorrer las mismas.
		/// </summary>
		/// <param name="consultaPaginable">
		/// A <see cref="ConsultaPaginable"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Object"/>
		/// </returns>
		private object[] CargadorDocumentos(ConsultaPaginable consultaPaginable)
		{
			return
				this.GetTipoConsultaSeleccionado()
					.Consulta.RealizarConsulta(this.criteriosBusquedaActuales, consultaPaginable);
		}
		#endregion
	}
}
