// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;

using Gtk;

using Stratos.Common;
using Stratos.ModeloDto;

namespace CustomWidgets
{
	/// <summary>
	/// Permite crear ComboBoxes a partir de una enumeración (aunque no exista esta
	/// restricción de ser una enumeración, ese es el objetivo). Los items del ComboBox
	/// serán los campos de la enumeración. Permite fácilmente obtener y asignar los
	/// valores de una determinada enumeración en la interfaz gráfica.
	/// </summary>
	public class ComboBoxGenerico<T> : Gtk.ComboBox, IComboBoxGenericoVista<T>
	{
		#region Campos
		public delegate string RenderizadorNumerico(T obj);
		
		private RenderizadorNumerico renderizador;
		
		private Dictionary<T,TreeIter> iters;
		private TreeModelFilter modeloFiltro;
		#endregion
		
		#region Constructores
		public ComboBoxGenerico() : this(EqualityComparer<T>.Default)
		{
		}
		
		public ComboBoxGenerico(IEqualityComparer<T> equalityComparer)
		{
			this.renderizador = this.RenderizadorPorDefecto;
			this.iters = new Dictionary<T,TreeIter>(equalityComparer);
			
			/* Configuro el modelo con todos los campos existentes en la
			 * enumeración correspondiente. */
			ListStore modelo = new ListStore(typeof(T));
			this.Model = modelo;
			
			// Creo el TreeModelFilter
			this.modeloFiltro = new TreeModelFilter(modelo, null);
			
			// Creo y seteo el CellRenderer
			CellRendererText renderer = new CellRendererText();
			this.PackStart(renderer, true);
			this.SetCellDataFunc(renderer, this.RenderObjeto);
		}
		#endregion
		
		#region Propiedades
		public RenderizadorNumerico Renderizador {
			set { this.renderizador = value; }
		}
		
		public string TextoActivo {
			get {
				TreeIter iter;
		        this.GetActiveIter(out iter);
				T obj = (T) this.Model.GetValue(iter, 0);
				
				if (obj != null)	       
					return obj.ToString();
				
				return "";
			}
		}
		
		public T ObjetoActivo {
			get {
				TreeIter iter;
	        	this.GetActiveIter(out iter);
				T obj = (T) this.Model.GetValue(iter, 0);
				
				return obj;
			}
			
			set {
				if (this.iters.ContainsKey(value)) {
					TreeIter iter = this.iters[value];
					this.SetActiveIter(iter);
				}
			}
		}
		
		public TreeModelFilterVisibleFunc FuncionFiltrado {
			set {
				this.modeloFiltro.VisibleFunc = value;
				this.Model = this.modeloFiltro;
				this.modeloFiltro.Refilter();
			}
		}
		
		public bool Editable {
			set { CustomWidgetsUtils.WidgetEditable(this, value); }
		}
		#endregion
		
		#region Métodos
		protected void RenderObjeto(CellLayout layout,
		                               CellRenderer renderer,
		                               TreeModel model,
		                               TreeIter iter)
		{
			object o = model.GetValue(iter, 0);

			string valor = String.Empty;
			if (o != null)
				valor = this.renderizador((T) o);
			
			(renderer as CellRendererText).Text = valor;
		}
		
		public void Limpiar()
		{
			ListStore modelo = (ListStore) this.Model;
			modelo.Clear();
		}
		
		public void AgregarObjeto(T obj)
		{
			ListStore modelo = (ListStore) this.Model;
			TreeIter iter = modelo.AppendValues(obj);
			
			this.iters[obj] = iter;
		}
		
		public void AgregarObjetos(T[] objArray)
		{
			foreach (T obj in objArray)
				this.AgregarObjeto(obj);
		}
		
		protected string RenderizadorPorDefecto(T obj)
		{
			return obj.ToString();
		}
		
		/// <summary>
		/// Dado un valor específico, devuelve el TreeIter correspondiente en el modelo.
		/// Esta función es útil desde la interfaz gráfica. Al recibir un objeto,
		/// por ejemplo ClienteDto, es fácil seleccionar el TreeIter correcto del ComboBox
		/// con esta función.
		/// </summary>
		/// <param name="obj">
		/// A <see cref="T"/>
		/// </param>
		/// <returns>
		/// A <see cref="TreeIter"/>
		/// </returns>
		public TreeIter FindIter(T obj)
		{
			return this.iters[obj];
		}
		
		/// <summary>
		/// Hace que se muestre el primer elemento del combo.
		/// </summary>
		public void Resetear()
		{
			TreeIter iter;
			this.Model.GetIterFirst(out iter);
			this.SetActiveIter(iter);
		}
		#endregion
	}
}
