// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Gtk;

using Stratos.ModeloDto;
using Stratos.Common;

namespace CustomWidgets
{
	[System.ComponentModel.ToolboxItem(true)]
	public class EntryProvincias : Gtk.Entry
	{
		#region Campos
		private Dictionary<string, ProvinciaDto> codigoProvincias;
		private const string SEPARADOR = " - ";
		#endregion
		
		#region Constructores
		public EntryProvincias()
		{
			// Cargo las provincias y los códigos
			this.codigoProvincias = new Dictionary<string, ProvinciaDto>();
			
			this.codigoProvincias.Add("BS", ProvinciaDto.BuenosAires);
			this.codigoProvincias.Add("CAT", ProvinciaDto.Catamarca);
			this.codigoProvincias.Add("CHA", ProvinciaDto.Chaco);
			this.codigoProvincias.Add("CHU", ProvinciaDto.Chubut);
			this.codigoProvincias.Add("CBA", ProvinciaDto.Cordoba);
			this.codigoProvincias.Add("CTE", ProvinciaDto.Corrientes);
			this.codigoProvincias.Add("ERI", ProvinciaDto.EntreRios);
			this.codigoProvincias.Add("FOR", ProvinciaDto.Formosa);
			this.codigoProvincias.Add("JUY", ProvinciaDto.Jujuy);
			this.codigoProvincias.Add("LPA", ProvinciaDto.LaPampa);
			this.codigoProvincias.Add("LRJ", ProvinciaDto.LaRioja);
			this.codigoProvincias.Add("MZA", ProvinciaDto.Mendoza);
			this.codigoProvincias.Add("MIS", ProvinciaDto.Misiones);
			this.codigoProvincias.Add("NEU", ProvinciaDto.Neuquen);
			this.codigoProvincias.Add("RNG", ProvinciaDto.RioNegro);
			this.codigoProvincias.Add("STA", ProvinciaDto.Salta);
			this.codigoProvincias.Add("SJU", ProvinciaDto.SanJuan);
			
			this.codigoProvincias.Add("SLU", ProvinciaDto.SanLuis);
			this.codigoProvincias.Add("SCR", ProvinciaDto.SantaCruz);
			this.codigoProvincias.Add("SFE", ProvinciaDto.SantaFe);
			this.codigoProvincias.Add("SGO", ProvinciaDto.SantiagoDelEstero);
			this.codigoProvincias.Add("TDF", ProvinciaDto.TierraDelFuego);
			this.codigoProvincias.Add("TUC", ProvinciaDto.Tucuman);
			
			// Seteo el EntryCompletion
			EntryCompletion completion = new EntryCompletion();
			
			/* La primer columna del modelo es el código de la provincia y la segunda
			 * es el objeto ProvinciaDto. */
			ListStore completionModel = new ListStore(typeof(string), typeof(ProvinciaDto));
			
			// Cargo el modelo de provincias con los códigos correspondientes
			foreach (var kvp in this.codigoProvincias)
				completionModel.AppendValues(kvp.Key, kvp.Value);
			
			completion.Model = completionModel;
			
			// Agrego un CellRenderer al EntryCompletion
			CellRendererText cellRenderer = new CellRendererText();
			completion.PackStart(cellRenderer, true);
			
			// Seteo la función de renderizado para el EntryCompletion y otros eventos
			completion.SetCellDataFunc(cellRenderer, this.RenderEntryCompletion);
			completion.MatchSelected += this.OnEntryCompletionMatchSelected;
			completion.MatchFunc = this.MatchProvincia;
			
			// Asigno el EntryCompletion
			this.Completion = completion;
		}
		#endregion
		
		#region Propiedades
		/// <value>
		/// Devuelve true si el contenido actual del EntryProvincias es válido.
		/// </value>
		public bool Valido {
			get {
				if (this.Text.Length <= 6)
					return false;
				
				string supuestoCodigoProvincia =
					this.Text.Substring(0, 3).Trim();
				
				if (!this.codigoProvincias.ContainsKey(supuestoCodigoProvincia))
					return false;
				
				if (!this.Text.Equals(this.ObtenerCadenaCompleta(supuestoCodigoProvincia)))
					return false;
				
				return true;
			}
		}
		
		/// <value>
		/// Devuelve el código de la provincia actual, si el valor del Entry es válido.
		/// </value>
		private string CodigoProvincia {
			get {
				if (!this.Valido)
					return String.Empty;
				
				return this.Text.Substring(0, 3).Trim().ToUpper();
			}
		}
		
		/// <value>
		/// Propiedad nueva que carga con el valor correcto al EntryProvincias.
		/// Soporta entradas del tipo 'Santa Fe' o 'SantaFe' y carga el widget
		/// con el valor completo (código de provincia más el nombre).
		/// </value>
		public new string Text {
			get { return base.Text; }
			set {
				string codigoProvincia =
					this.ObtenerCodigoProvincia(value);
				
				if (codigoProvincia != null &&
				    	this.codigoProvincias.ContainsKey(codigoProvincia))
					base.Text = this.ObtenerCadenaCompleta(codigoProvincia);
				else
					base.Text = "";
			}
		}
		
		/// <value>
		/// Permite obtener la ProvinciaDto seteada o setear una.
		/// </value>
		public ProvinciaDto ProvinciaSeteada {
			get { return this.codigoProvincias[this.CodigoProvincia]; }
			set { this.Text = value.ToString(); }
		}
		#endregion
		
		#region Métodos relacionados con el EntryCompletion
		protected void RenderEntryCompletion(CellLayout cellLayout, CellRenderer renderer,
		                                     TreeModel model, TreeIter iter)
		{
			string codigoProvincia = (string) model.GetValue(iter, 0);
			
			// Renderizo en el EntryCompletion el código y el nombre de la provincia
			(renderer as CellRendererText).Text =
				this.ObtenerCadenaCompleta(codigoProvincia);
				
		}
		
		protected bool MatchProvincia(EntryCompletion entryCompletion, string key,
		                              TreeIter iter)
		{
			/* Los TreeIter matchean en el EntryCompletion si tanto el código como el
			 * nombre de la provincia comienzan con lo escrito por el usuario. */
			string codigoProvincia = ((string) entryCompletion.Model.GetValue(iter, 0))
				.ToLower();
			string provincia =
				StratosUtil.SplitCamelString(
					((ProvinciaDto) entryCompletion.Model.GetValue(iter, 1))
						.ToString()).ToLower();
			
			return codigoProvincia.StartsWith(key.ToLower()) ||
				provincia.StartsWith(key.ToLower());
		}
		
		[GLib.ConnectBefore]
		protected void OnEntryCompletionMatchSelected(object o, MatchSelectedArgs args)
		{
			string codigoProvincia = (string) args.Model.GetValue(args.Iter, 0);
			
			base.Text = this.ObtenerCadenaCompleta(codigoProvincia);
			
			args.RetVal = true;
		}
		#endregion
		
		#region Eventos
		protected override bool OnButtonReleaseEvent (Gdk.EventButton evnt)
		{
			/* Al presionar un botón del mouse sobre el Entry, selecciono todo
			 * el texto del mismo, dejándolo listo para reemplazar completamente
			 * el contenido de forma fácil. */
			bool retorno = base.OnButtonReleaseEvent(evnt);
			
			if (evnt.Type == Gdk.EventType.ButtonRelease)
				this.SelectRegion(0, -1);
			
			return retorno;
		}
		
		protected override bool OnFocusOutEvent (Gdk.EventFocus evnt)
		{
			/* Cuando el Entry pierde el foco, me aseguro que quede en el mismo un valor
			 * válido, o un string vacío en caso contrario. */
			
			// Si el valor actual ya es válido, no continúo.
			if (this.Valido)
				return base.OnFocusOutEvent (evnt);
			
			string valorActual = this.Text.ToUpper();
			
			// Si el valor actual es el código de una provincia, la cargo
			if (this.codigoProvincias.ContainsKey(valorActual))
				base.Text = this.ObtenerCadenaCompleta(valorActual);
			
			/* Si no es el código de una provincia, entonces quizá sea el nombre completo
			 * de la misma, como 'santa fe'. */
			else {
				/* Necesito obtener el código de la provincia a partir de lo cargado
				 * actualmente. */
				string codigoProvincia = this.ObtenerCodigoProvincia(this.Text);
				
				if (codigoProvincia != null && this.codigoProvincias.ContainsKey(codigoProvincia))
					base.Text = this.ObtenerCadenaCompleta(codigoProvincia);
				else
					base.Text = String.Empty;
			}
			
			return base.OnFocusOutEvent (evnt);
		}
		#endregion
		
		#region Métodos útiles
		/// <summary>
		/// Permite obtener el código de la provincia a partir de un objeto ProvinciaDto.
		/// </summary>
		/// <param name="provincia">
		/// A <see cref="ProvinciaDto"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		private string ObtenerCodigoProvincia(ProvinciaDto provincia)
		{
			return this.ObtenerCodigoProvincia(StratosUtil.SplitCamelString(provincia.ToString()));
		}
		
		/// <summary>
		/// A partir de una representación de una provincia, obtiene el código de la misma.
		/// Permite entradas del tipo 'SantaFe' (cuando se ejecuta un ToString sobre el
		/// objeto ProvinciaDto) o 'Santa Fe'. En este ejemplo devuelve 'SFE'.
		/// </summary>
		/// <param name="provincia">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		private string ObtenerCodigoProvincia(string provincia)
		{
			string codigoProvincia =
					(from k in this.codigoProvincias
					 where string.Compare(StratosUtil.SplitCamelString(k.Value.ToString()),
					                      provincia,
					                      true) == 0 ||
						 string.Compare(k.Value.ToString(), provincia, true) == 0
					 select k.Key).FirstOrDefault();
			
			return codigoProvincia;
		}
		
		/// <summary>
		/// Obtiene la cadena completa con el separador por defecto.
		/// </summary>
		/// <param name="codigoProvincia">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		private string ObtenerCadenaCompleta(string codigoProvincia)
		{
			return this.ObtenerCadenaCompleta(codigoProvincia, SEPARADOR);
		}
		
		/// <summary>
		/// A partir del código de una provincia, genera la cadena completa que
		/// será mostrada en el Entry. Por ejemplo, a partir de 'SFE' genera
		/// 'SFE - Santa Fe'. En este caso ' - ' es el separador, pero permite
		/// configurar uno.
		/// </summary>
		/// <param name="codigoProvincia">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="separador">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		private string ObtenerCadenaCompleta(string codigoProvincia, string separador)
		{
			ProvinciaDto provincia = this.codigoProvincias[codigoProvincia];
			
			string completo = codigoProvincia +
					separador +
					StratosUtil.SplitCamelString(provincia.ToString());
			
			return completo;
		}
		#endregion
	}
}
