// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;

using Gtk;

using StratosServer.Controladores;

namespace CustomWidgets
{
	/// <summary>
	/// Dada una fuente de datos y una función de carga, permite visualizar distintas páginas de
	/// una colección de elementos.
	/// </summary>
	[System.ComponentModel.ToolboxItem(true)]
	public partial class NavegadorPaginas : Gtk.Bin, INavegadorPaginasVista
	{
		public delegate object[] CargadorDatosHandler(ConsultaPaginable consultaPaginable);
		private int tamañoPagina;
		
		protected int paginaActual;
		
		protected CriterioOrdenacion criterioOrdenacion;
		
		protected ColeccionElementos coleccionElementos;
		protected CargadorDatosHandler cargadorDatos;
		
		protected bool ultimaPagina;
		
		public NavegadorPaginas()
		{
			this.Build();
			
			this.tamañoPagina = 10;
			
			this.ultimaPagina = true;
			
			/* Por defecto el criterio de ordenación es la columna Fecha. Esto va a fallar
			 * si el objeto no tiene esta columna. */
			this.criterioOrdenacion =
				new CriterioOrdenacion {
					campoOrdenacion = "Fecha",
					sentidoOrd = SentidoOrdenacion.Desc
				};
		}
		
		#region Propiedades
		public int PaginaActual {
			get { return this.paginaActual; }
			set {
				this.paginaActual = value;
				this.lblPaginaActual.Markup = "<b>" + (this.paginaActual + 1).ToString() + "</b>";
			}
		}
		
		public CriterioOrdenacion CriterioOrdenacion {
			get { return this.criterioOrdenacion; }
			set { this.criterioOrdenacion = value; }
		}
		
		public int TamañoPagina {
			get { return this.tamañoPagina; }
			set {
				if (value <= 0)
					this.tamañoPagina = 10;
				else
					this.tamañoPagina = value;
			}
		}
		
		public ColeccionElementos ColeccionElementos {
			set {
				this.coleccionElementos = value;
				
				this.coleccionElementos.PrimeraPaginaCargada += this.OnPrimeraPaginaCargada;
			}
		}
		
		public CargadorDatosHandler CargadorDatos {
			set { this.cargadorDatos = value; }
		}
		
		public bool UltimaPagina {
			get { return this.ultimaPagina; }
			private set {
				this.ultimaPagina = value;
				this.btnPaginaSiguiente.Sensitive = !value;
			}
		}
		
		public bool PrimeraPagina {
			get { return this.paginaActual == 0 ? true : false; }
		}
		
		public bool Editable {
			get { return this.Sensitive; }
			set { this.Sensitive = value; }
		}
		#endregion
		
		#region Métodos públicos
		public void RecargarPaginaActual()
		{
			this.coleccionElementos.Vaciar();
			
			// Utilizo la función de carga de datos
			object[] resultado = this.cargadorDatos(this.CrearConsultaPaginable());
			
			this.coleccionElementos.AgregarElementos(resultado);
		}
		
		public void CargarPaginaSiguiente()
		{
			if (!this.UltimaPagina)
				this.OnBtnPaginaSiguienteClicked(this, null);
		}
		
		public void CargarPaginaAnterior()
		{
			if (!this.PrimeraPagina)
				this.OnBtnPaginaAnteriorClicked(this, null);
		}
		#endregion
		
		#region Eventos
		protected void OnPrimeraPaginaCargada()
		{
			this.PaginaActual = 0;
			this.Editable = true;
			this.btnPrimeraPagina.Sensitive = false;
			
			// Siempre que refrescamos vamos a la primer página, por lo tanto el botón
			// para ir a la página anterior debe estar deshabilitado.
			this.btnPaginaAnterior.Sensitive = false;
			
			// Si la cantidad de elementos es menor al tamaño de página, entonces no hay una
			// página siguiente.
			if (this.coleccionElementos.CantidadElementos < this.tamañoPagina)
				this.UltimaPagina = true;
			else
				this.UltimaPagina = false;
		}
		
		protected virtual void OnBtnPaginaAnteriorClicked (object sender, System.EventArgs e)
		{
			// Quito todos los elementos de la lista
			this.coleccionElementos.Vaciar();
			
			this.PaginaActual--;
			
			// Seteo la ultima página en false, porque si fuí a una pagina anterior,
			// seguro entonces que esta no es la ultima
			this.UltimaPagina = false;
			
			if (this.PaginaActual == 0)
				this.btnPrimeraPagina.Sensitive = false;
			else
				this.btnPrimeraPagina.Sensitive = true;
			
			// Utilizo la función de carga de datos
			object[] resultado = this.cargadorDatos(this.CrearConsultaPaginable());
			
			this.coleccionElementos.AgregarElementos(resultado);
			
			// Si estoy en la primer página, no activo el botón anterior
			if (this.PaginaActual == 0)
				this.btnPaginaAnterior.Sensitive = false;
		}

		protected virtual void OnBtnPaginaSiguienteClicked (object sender, System.EventArgs e)
		{
			// Quito todos los elementos de la lista
			this.coleccionElementos.Vaciar();
						
			this.PaginaActual++;
			
			// Activo el botón para ir a la página anterior
			this.btnPaginaAnterior.Sensitive = true;
			this.btnPrimeraPagina.Sensitive = true;
			
			// Utilizo la función de carga de datos
			object[] resultado = this.cargadorDatos(this.CrearConsultaPaginable());
			
			/* Si los resultados son menores al tamaño de página, entonces no activo
			 * el botón para ir a la siguiente, ya que no habrá más datos. */
			if (resultado.Length < this.tamañoPagina)
				this.UltimaPagina = true;
			
			// Agrego los elementos resultado a la colección de elementos
			this.coleccionElementos.AgregarElementos(resultado);
		}
		
		protected virtual void OnBtnPrimeraPaginaClicked (object sender, System.EventArgs e)
		{
			// Quito todos los elementos de la lista
			this.coleccionElementos.Vaciar();
			
			this.PaginaActual = 0;
			
			// Utilizo la función de carga de datos
			object[] resultado = this.cargadorDatos(this.CrearConsultaPaginable());
			
			this.coleccionElementos.AgregarElementosPrimeraPagina(resultado);
		}
		#endregion
		
		#region Metodos internos
		private ConsultaPaginable CrearConsultaPaginable()
		{
			// Creo el objeto ConsultaPaginable
			ConsultaPaginable consultaPaginable = new ConsultaPaginable {
				pagina = this.PaginaActual,
				tamanioPagina = this.tamañoPagina,
				criteriosOrdenacion = new List<CriterioOrdenacion> {
					this.criterioOrdenacion
				}
			};
			
			return consultaPaginable;
		}
		#endregion
	}
}
