// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Gtk;

using StratosClient.Core.Gui;

namespace CustomWidgets
{
	/// <summary>
	/// Lista simple de objetos genéricos.
	/// </summary>
	public abstract class ListaElementos<T> : ColeccionElementos, IListaElementosVista<T>
	{
		#region Campos
		public event System.Action ModeloCambiado;
		public event System.Action<T[]> ElementosSeleccionados;
		#endregion
		
		#region Constructores
		public ListaElementos()
		{
			#region Propiedades generales
			this.Selection.Mode = SelectionMode.Multiple;
			#endregion
			
			this.Selection.Changed += this.OnSelectionChanged;
			
			this.ListaVaciada += this.OnModeloCambiado;
			
			#region Modelo del TreeView
			// Creo un modelo por defecto con el tipo T
			this.Model = new ListStore(typeof(T));
			#endregion
		}
		#endregion
		
		#region Propiedades
		protected ListStore Modelo {
			get { return (ListStore) base.Model; }
			set { this.Model = value; }
		}
		#endregion
		
		#region ColeccionElementos implementation
		public override void InsertarElemento(object elemento, int posicion)
		{
			this.InsertarElemento((T) elemento, posicion);
		}
		
		public override void AgregarElementos(params object[] elementos)
		{
			this.AgregarElementos(elementos.Cast<T>().ToArray());
		}
		
		public override object[] ObtenerSeleccionados(out TreeIter[] itersSeleccionado)
		{
			return this.ObtenerSeleccionadosGenericos(out itersSeleccionado).Cast<object>().ToArray();
		}
		
		public override object[] ObtenerSeleccionados()
		{
			return this.ObtenerSeleccionadosGenericos().Cast<object>().ToArray();
		}
		
		public override IList ObtenerTodos ()
		{
			return this.ObtenerTodosGenerico();
		}
		
		public override void BorrarItems(params TreeIter[] iters)
		{
			TreeIter it;
			
			for (int i=0; i<iters.Length; i++) {
				it = iters[i];
				
				if (!this.Modelo.IterIsValid(it))
					continue;
				
				this.Modelo.Remove(ref it);
			}
			
			this.OnModeloCambiado();
		}
		#endregion
		
		#region Métodos genéricos
		public virtual void InsertarElemento(T elemento, int posicion)
		{
			TreeIter iter;
			
			// Ingreso el documento. Si ya existe en el modelo, lo cambio.
			if (this.Contiene(elemento, out iter)) {
				this.Modelo.SetValue(iter, 0, elemento);
				this.Modelo.EmitRowChanged(this.Modelo.GetPath(iter), iter);
			}
			else
				this.Modelo.InsertWithValues(posicion, elemento);
			
			// Disparo el evento sólo si se agregaron documentos
			if (this.ModeloCambiado != null)
				this.ModeloCambiado();
		}
		
		/// <summary>
		/// Agrega todos los elementos al modelo de la lista.
		/// Devuelve un array de objetos TreeIter, correspondientes a los elementos
		/// agregados.
		/// </summary>
		/// <param name="elementos">
		/// A <see cref="T[]"/>
		/// </param>
		/// <returns>
		/// A <see cref="TreeIter[]"/>
		/// </returns>
		public virtual TreeIter[] AgregarElementosBase(params T[] elementos)
		{
			TreeIter iter;
			T elemento;
			TreeIter[] itersAgregados = new TreeIter[elementos.Length];
			
			// Ingreso todos los elementos. Si alguno ya existe en el modelo, lo cambio.
			for (int i=0; i<elementos.Length; i++) {
				elemento = elementos[i];
				
				if (this.Contiene(elemento, out iter)) {
					this.Modelo.SetValue(iter, 0, elemento);
					this.Modelo.EmitRowChanged(this.Modelo.GetPath(iter), iter);
					
					itersAgregados[i] = iter;
				}
				else
					itersAgregados[i] = this.Modelo.AppendValues(elemento);
			}
			
			// Disparo el evento sólo si se agregaron documentos
			if (this.ModeloCambiado != null && elementos.Length > 0)
				this.ModeloCambiado();
			
			return itersAgregados;
		}
		
		/// <summary>
		/// Agrega todos los elementos al modelo de la lista.
		/// Esta versión del método simplemente no retorna un array de
		/// TreeIters.
		/// </summary>
		/// <param name="elementos">
		/// A <see cref="T[]"/>
		/// </param>
		public virtual void AgregarElementos(params T[] elementos)
		{
			this.AgregarElementosBase(elementos);
		}
		
		/// <summary>
		/// Actualiza la primer fila seleccionada con la instancia pasada como
		/// argumento.
		/// </summary>
		/// <param name="objetoModificado">
		/// A <see cref="T"/>
		/// </param>
		public virtual void ActualizarObjetoSeleccionado(T objetoModificado)
		{
			TreePath[] pathsSeleccionados;
			TreeIter iter = new TreeIter();
			TreeModel modeloTemp;
			
			pathsSeleccionados = this.Selection.GetSelectedRows(out modeloTemp);
			modeloTemp.GetIter(out iter, pathsSeleccionados[0]);
			
			modeloTemp.SetValue(iter, 0, objetoModificado);
			modeloTemp.EmitRowChanged(modeloTemp.GetPath(iter), iter);
		}
		
		public void EliminarObjetosSeleccionados()
		{
			TreeIter[] iters;
			
			this.ObtenerSeleccionados(out iters);
			
			this.BorrarItems(iters);
		}
		
		protected virtual T ObtenerElemento(TreeIter iter)
		{
			T elemento = (T) this.Modelo.GetValue(iter, 0);
			
			return elemento;
		}
		
		private bool Contiene(T elemento, out TreeIter iterExistente)
		{
			T elementoExistente;
			
			iterExistente = TreeIter.Zero;
			
			/* Itero por todas las filas y me fijo si el documento ya está ingresado en el
			 * TreeView. Para esto me fijo en el tipo de documento y su id. */
			foreach (TreeIter iter in this.Modelo.Traverse()) {
				elementoExistente = this.ObtenerElemento(iter);
				
				if (elementoExistente.GetType().Equals(elemento.GetType()) &&
						this.Iguales(elementoExistente, elemento)) {
					iterExistente = iter;
					return true;
				}
			}
			
			return false;
		}
		
		public virtual T[] ObtenerSeleccionadosGenericos(out TreeIter[] itersSeleccionados)
		{
			// Obtengo todos los TreePath que corresponden a las filas seleccionadas
			TreeIter iter;
			TreePath[] filasSeleccionados = this.Selection.GetSelectedRows();
			
			// Cargo el array de TreeIters y elementos seleccionados
			itersSeleccionados = new TreeIter[filasSeleccionados.Length];
			T[] elementosSeleccionados = new T[filasSeleccionados.Length];
			
			for (int i=0; i<filasSeleccionados.Length; i++) {
				this.Model.GetIter(out iter, filasSeleccionados[i]);
				itersSeleccionados[i] = iter;
				
				elementosSeleccionados[i] = this.ObtenerElemento(iter);
			}
			
			return elementosSeleccionados;
		}
		
		public virtual T[] ObtenerSeleccionadosGenericos()
		{
			TreeIter[] iters;
			
			return this.ObtenerSeleccionadosGenericos(out iters);
		}
		
		public virtual List<T> ObtenerTodosGenerico()
		{
			List<T> todos = new List<T>();
			todos.AddRange(this.Model.TraverseColumn<T>(0));
			
			return todos;
		}

		public TreePath GetSelectedPath()
		{
			TreeIter iter;
			this.Selection.GetSelected(out iter);
			
			return this.Model.GetPath(iter);
		}
		#endregion
		
		#region Eventos		
		protected void OnSelectionChanged (object sender, EventArgs e)
		{
			// Me fijo si hay suscriptores al evento
			if (this.ElementosSeleccionados == null)
				return;
			
			this.ElementosSeleccionados(this.ObtenerSeleccionadosGenericos());
		}
		
		protected void OnModeloCambiado()
		{
			if (this.ModeloCambiado != null)
				this.ModeloCambiado();
		}
		#endregion
		
		#region Métodos abstractos
		protected abstract bool Iguales(T t1, T t2);
		#endregion
	}
}
