// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;

using Gtk;

using Stratos.ModeloDto;
using StratosClient.Core.Gui;

namespace CustomWidgets
{
	/// <summary>
	/// Clase base genérica para listados de elementos interactivos con el usuario.
	/// </summary>
	public abstract class ListaInteractiva<T> : ListaInteractivaBase, IListaInteractivaVista
		where T : ItemGenericoDto
	{
		#region Campos
		protected TreeIter filaEnModoEdicion;
		#endregion
		
		#region Constructores
		public ListaInteractiva() : base()
		{
			this.filaEnModoEdicion = TreeIter.Zero;
			
			#region Modelo del TreeView
			ListStore modeloItems = new ListStore(typeof(T));
			this.Model = modeloItems;
			
			this.AgregarItemVacio();
			#endregion
		}
		#endregion
		
		#region Propiedades
		public bool ForzarSetCursor {
			get { return this.forzarSetCursor; }
			set { this.forzarSetCursor = value; }
		}
		
		internal TreeIter FilaEnModoEdicion {
			get { return this.filaEnModoEdicion; }
			set { this.filaEnModoEdicion = value; }
		}
		
		public IList ListaItems {
			get { return this.Items; }
		}
		
		/// <summary>
		/// Devuelve una lista de los items del TreeView.
		/// </summary>
		public List<T> Items {
			get {
				List<T> items = new List<T>();
				
				int orden = 0;
				
				foreach (T item in this.Model.TraverseColumn<T>(0)) {
					if (!this.ItemVacio(item) && this.ItemValido(item)) {
						// Le asigno el orden al item
						item.orden = orden++;
						
						// Luego lo agrego a la lista
						items.Add(item);
					}
				}
				
				return items;
			}
		}
		#endregion
		
		#region Métodos abstractos
		/// <summary>
		/// Verifica si el item es un item vacío.
		/// </summary>
		/// <param name="item">
		/// A <see cref="T"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected abstract bool ItemVacioGenericoInterno(T item);
		
		/// <summary>
		/// Realiza la misma acción que ItemValido(T), pero además, si se trata
		/// de un item inválido, devuelve una columna en la cual haya algún valor inválido. Este
		/// método se utiliza en ItemModificado para dar el foco en modo edición a la columna
		/// que corresponda.
		/// la columna del 
		/// </summary>
		/// <param name="item">
		/// A <see cref="ItemDocumentoGenericoDto"/>
		/// </param>
		/// <param name="columnaInvalida">
		/// A <see cref="TreeViewColumn"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected abstract bool ItemValidoGenerico(T item, out TreeViewColumn columnaInvalida);
		
		/// <summary>
		/// Devuelve el objeto TreeViewColumn que va a aparecer en modo edición
		/// cuando se agregue un nuevo item al TreeView.
		/// </summary>
		protected abstract TreeViewColumn ColumnaEdicionSobreItemVacio {
			get;
		}
		
		/// <summary>
		/// Este método es ejecutado cuando un item ha sido modificado en el
		/// TreeView. La idea es que se lo utilice para hacer recálculos por ejemplo,
		/// en un Item de una factura se recalcularía el importe del item.
		/// </summary>
		/// <param name="iter">
		/// A <see cref="TreeIter"/>
		/// </param>
		internal abstract void ItemModificadoListener(T item);
		#endregion
		
		#region TreeViewInteractivoBase implementation
		public override bool ItemVacio (ItemGenericoDto item)
		{
			return this.ItemVacioGenerico((T) item);
		}
		
		public override bool ItemValido (ItemGenericoDto item)
		{
			return this.ItemValidoGenerico((T) item);
		}
		#endregion
		
		#region Métodos genéricos
		public T ObtenerItemSeleccionadoGenerico()
		{
			return (T) this.ObtenerItemSeleccionado();
		}
		#endregion
		
		#region Métodos
		/// <summary>
		/// Verifica si el item es un item vacío.
		/// </summary>
		/// <param name="item">
		/// A <see cref="T"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public bool ItemVacioGenerico(T item)
		{
			if (item == null)
				return true;
			
			return this.ItemVacioGenericoInterno(item);
		}
		
		protected bool EstaFilaEnModoEdicion(TreeIter iter)
		{
			if (this.FilaEnModoEdicion.Equals(TreeIter.Zero))
				return false;
			
			if (this.FilaEnModoEdicion.Equals(iter))
				return true;
			
			return false;
		}
		
		/// <summary>
		/// Dada una columna determinada (que corresponde a la columna de la fila que se acaba
		/// de editar), devuelve la siguiente columna a colocar en modo de edición.
		/// Las subclases pueden sobreescribir este método para definir su propio
		/// orden para colocar las columnas en modo edición.
		/// </summary>
		/// <param name="columna">
		/// A <see cref="TreeViewColumn"/>
		/// </param>
		/// <returns>
		/// A <see cref="TreeViewColumn"/>
		/// </returns>
		protected virtual TreeViewColumn SiguienteColumnaAEditar(TreeViewColumn columna)
		{
			/* Por defecto devuelvo null que proboca que se coloque en modo de edición
			 * a la siguiente columna inválida. */
			return null;
		}
		
		/// <summary>
		/// Método para validar un item.
		/// </summary>
		/// <param name="item">
		/// A <see cref="T"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected bool ItemValidoGenerico(T item)
		{
			TreeViewColumn temp;
			return this.ItemValidoGenerico(item, out temp);
		}
		
		/// <summary>
		/// Devuelve el TreePath de la fila actualmente seleccionada.
		/// </summary>
		/// <returns>
		/// A <see cref="TreePath"/>
		/// </returns>
		public TreePath GetSelectedPath()
		{
			return this.Model.GetPath(this.GetSelectedIter());
		}
		
		public TreeIter GetSelectedIter()
		{
			TreeIter iter;
			this.Selection.GetSelected(out iter);
			
			return iter;
		}
		
		#region Método ItemModificado
		/// <summary>
		/// Este método realiza algunas operaciones necesarias al modificarse algún item
		/// de producto, como recalcular el importe y agregar otros items. Permite especificar
		/// una columna para dejar en modo de edición.
		/// </summary>
		/// <param name="iter">
		/// TreeIter correspondiente a la fila que se modificó.
		/// </param>
		/// <param name="columnaADejarEnModoEdicion">
		/// TreeViewColumn que se dejará en modo edición luego de finalizar el método.
		/// </param>
		internal void ItemModificado(TreeIter iter, TreeViewColumn columnaADejarEnModoEdicion)
		{
			ListStore modelo = (ListStore) this.Model;
			
			/* Obtengo el objeto ItemProductoDto de la fila modificada */
			T item = (T) modelo.GetValue(iter, 0);
			
			/* Paso el item al listener por si hay que hacer algun recálculo u operación
			 * similar. */
			this.ItemModificadoListener(item);
			
			/* Emito una señal de que el item ha cambiado. Esto es necesario realizar siempre
			 * que se cambie un objeto del modelo. */
			modelo.EmitRowChanged(modelo.GetPath(iter), iter);
			
			/* El siguiente código ejecuta las tareas pendientes en el mainloop de Gtk. Esto
			 * se hace para que el TreeView termine de dibujarse (por ejemplo, si las columnas
			 * tienen que cambiar su tamaño por el producto recién cargado).
			 * Si esto no se realiza, sucede que al cargar algún tipo de producto que cambie
			 * el tamaño de las columnas (como el de código 'M-000'), el SetCursor de abajo
			 * hace que quede descolocado con respecto a la columna correspondiente. */
			while (Gtk.Application.EventsPending())
				Gtk.Application.RunIteration();
			
			/* En el código siguiente intento dar el foco en modo de edición a la columna
			 * del item modificado que tiene algún valor inválido. Si esto no es así, o sea
			 * si el item es válido, entonces ingreso uno nuevo si se dan dos condiciones:
			 * si el item modificado es el último y si no me paso del límite de items */
			TreeViewColumn columnaInvalida;
			
			if (!this.ItemValidoGenerico(item, out columnaInvalida) && this.HabilitarSetCursor()) {
				/* Si el item actual es inválido, la subclase puede configurar cómo desea
				 * que este método actúe: puede dejar que se coloque el foco en la siguiente
				 * columna inválida (comportamiento por defecto) o puede definir un orden
				 * de columnas a editar. */
				
				// Obtengo la columna y el path que se están editando actualmente.
				TreePath pathActual;
				TreeViewColumn columnaActual;
				this.GetCursor(out pathActual, out columnaActual);
				
				/* Si la subclase sobreescribió el método SiguienteColumnaAEditar, entonces
				 * pongo el modo de edición en la columna siguiente a la actual.
				 * Sino coloco en modo edición a la siguiente columna inválida. */
				TreeViewColumn siguienteColumnaAEditar =
					this.SiguienteColumnaAEditar(columnaActual);
				
				/* Si no se definió una siguiente columna a editar en el método
				 * SiguienteColumnaAEditar, entonces dejo en edición la columna inválida */
				if (siguienteColumnaAEditar == null)
					siguienteColumnaAEditar = columnaInvalida;
				
				if (columnaADejarEnModoEdicion != null)
					siguienteColumnaAEditar = columnaADejarEnModoEdicion;
				
				this.ColocarCeldaEnModoEdicion(modelo.GetPath(iter), siguienteColumnaAEditar);
			}
			else if (this.ItemValidoGenerico(item)) {
				// Agrego un nuevo item si el actual es el último y es válido
				TreeIter ultimoIter = this.ObtenerUltimoIter();
				
				if (iter.Equals(ultimoIter) && !this.LimiteItemsAlcanzado()) {
					TreeIter nuevoIter = this.AgregarItemVacio();
					
					/* Sólo activo en modo edición la columna indicada si el CellRenderer
					 * de la misma no es un CellRendererCombo y el evento fue
					 * presionar la tecla Enter. Si el usuario hizo click con el mouse
					 * en otra fila, no debo ejecutar SetCursor, de otra forma el TreeView
					 * queda en un estado incosistente. */
					this.ColocarCeldaEnModoEdicion(modelo.GetPath(nuevoIter),
					                               this.ColumnaEdicionSobreItemVacio);
				}
				else
					this.ColocarCeldaEnModoEdicion(modelo.GetPath(iter),
					                               columnaADejarEnModoEdicion);
			}
			
			// Lanzo el evento ItemsCambiados
			this.OnItemsCambiados();
		}
		
		/// <summary>
		/// Este método realiza algunas operaciones necesarias al modificarse algún item
		/// de producto, como recalcular el importe y agregar otros items. Esta versión
		/// deja en edición la columna por defecto.
		/// </summary>
		/// <param name="iter">
		/// A <see cref="TreeIter"/>
		/// </param>
		internal void ItemModificado(TreeIter iter)
		{
			this.ItemModificado(iter, null);
		}
		#endregion
		#endregion
	}
}
