// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Gtk;

using Stratos.ModeloDto;
using StratosClient.Core.Gui;
using StratosServer.Controladores;

namespace CustomWidgets
{
	public enum Estado {
		Busqueda,
		Seleccion
	};
	
	/// <summary>
	/// Esta clase abstrae a las subclases de implementar el código necesario para mostrar
	/// un menu desplegable.
	/// </summary>
	public abstract class CellRendererMenuDesplegable : CellRendererInteractivoBase<ItemProductoDto>
	{
		#region Campos
		/// <summary>
		/// Para gestionar la búsqueda y selección de productos en el EntryCompletion,
		/// es necesario llevar un estado del mismo. Cuando comienza la edición el estado
		/// es Busqueda. Cuando el usuario oprime la tecla Return se realiza una petición
		/// al servidor y se entra en estado de Seleccion, mostrando el EntryCompletion.
		/// El estado se utiliza para saber qué hacer cuando el usuario oprime Return.
		/// </summary>
		/// <param name="treeView">
		/// A <see cref="TvItemsProducto"/>
		/// </param>
		protected Estado estado;
		
		/// <summary>
		/// Esta es la misma referencia al TreeView que existe en CellRendererBase.
		/// Está acá para evitar casting cuando se acceda a los métodos exclusivos
		/// de TvItemsProducto.
		/// </summary>
		protected ListaItemsProducto tvItemsProducto;
		#endregion
		
		#region Constructores
		public CellRendererMenuDesplegable(ListaItemsProducto lista) : base(lista)
		{
			this.tvItemsProducto = lista;
		}
		#endregion
		
		#region CellRendererInteractivoBase - Implementación
		protected override void EdicionIniciada (Gtk.Entry entry, string path, ItemProductoDto item)
		{
			// Creo en EntryCompletion y el modelo.
			EntryCompletion completion = new EntryCompletion();
			ListStore completionModel = new ListStore(typeof(ProductoDto));
			completion.Model = completionModel;
			completion.PopupSetWidth = false;
			
			// CellRenderer para el código
			CellRendererText cellRendererCodigo = new CellRendererText();
			completion.PackStart(cellRendererCodigo, false);
			completion.SetCellDataFunc(cellRendererCodigo, this.RenderCodigoEntryCompletion);
			
			// CellRenderer para la descripción
			CellRendererText cellRendererDescripcion = new CellRendererText();
			completion.PackStart(cellRendererDescripcion, true);
			completion.SetCellDataFunc(cellRendererDescripcion, this.RenderDescripcionEntryCompletion);
			
			// CellRenderer para el precio de venta
			CellRendererText cellRendererPrecioVenta = new CellRendererText();
			cellRendererPrecioVenta.Xalign = 1.0f; // Centro a la derecha
			completion.PackStart(cellRendererPrecioVenta, false);
			completion.SetCellDataFunc(cellRendererPrecioVenta, this.RenderPrecioVentaEntryCompletion);
			
			// Este cell renderer hace que el EntryCompletion se expanda un poco más de lo normal.
			// No se muestra información útil, es sólo para ése propósito.
			CellRendererText temp = new CellRendererText();
			completion.PackStart(temp, false);
			completion.SetCellDataFunc(temp, this.RenderUltimaColumnaEntryCompletion);
			
			// Seteo algunos eventos del EntryCompletion
			completion.MatchSelected += this.OnEntryCompletionMatchSelected;
			completion.MatchFunc = delegate {
				return true;
			};
			
			// Asigno el EntryCompletion al widget Entry
			entry.Completion = completion;
			
			/* Me suscribo a algunos eventos importantes para gestionar la búsqueda
			 * de productos por descripción. */
			entry.KeyPressEvent += OnEntryKeyPress;
			entry.Changed += this.OnEntryChanged;
			
			// Al iniciarse la edición de la descripción del producto, el estado es de búsqueda
			this.estado = Estado.Busqueda;
		}
		#endregion
		
		#region Métodos Entry
		[GLib.ConnectBefore]
		protected void OnEntryKeyPress(object o, KeyPressEventArgs args)
		{
			/* Si el usuario presiona alguna tecla de flecha, entonces marcamos el estado
			 * como de Búsqueda, y así el próximo Return ejecutará una búsqueda al
			 * Servidor. */
			if (args.Event.Key == Gdk.Key.Right ||
			    	args.Event.Key == Gdk.Key.Left)
				this.estado = Estado.Busqueda;
			
			Entry entry = (Entry) o;
			
			// Proceso la consulta (le quito los espacios) y verifico que no sea vacía
			string consulta = entry.Text.Trim();
			if (consulta == String.Empty)
				return;
			
			while (Gtk.Application.EventsPending())
				Gtk.Application.RunIteration();
			
			/* Si el estado actual es de Seleccion, entonces el usuario ya realizó una búsqueda
			 * al servidor, y se encuentra observando la lista del EntryCompletion. Por lo tanto
			 * el próximo Return será para escoger un producto, salvo que el contenido del Entry
			 * cambie, en ese caso el usuario está intentando realizar otra búsqueda (ver el
			 * método OnEntryChanged).
			 * En este condicional también me fijo si no se cumplen otras condiciones, por ejemplo:
			 * - Si preciono Control+Enter, siempre realizo la búsqueda.
			 * - Si hay cargado un producto genérico y se presionó Control+Enter, siempre realizo
			 *   una búsqueda, es decir, no interpreto el Enter como edición, sino intensión de
			 *   búsqueda.
			 * - Etc...
			 * */
			ItemProductoDto itemProducto = this.treeView.ObtenerItemSeleccionadoGenerico();
			
			if (this.estado == Estado.Seleccion ||
			    	(args.Event.Key != Gdk.Key.Return && args.Event.Key != Gdk.Key.KP_Enter) ||
			    	this.NoRealizarConsulta(consulta, itemProducto, args))
				return;
			
			// Me aseguro que la consulta tenga un tamaño mínimo
			if (consulta.Length < this.TamañoMinimoConsulta) {
				SMessageDialog
					.MensajeAdvertencia("Los criterios de búsqueda son muy pocos. <b>Ingrese más " +
					                    "caracteres para buscar</b>.");
				
				// Dejo la misma celda en modo de edición
				this.treeView
					.ColocarCeldaEnModoEdicion(this.tvItemsProducto.GetSelectedPath(),
					                           this.Columna);
				
				return;
			}
			
			// Realizo la búsqueda de productos por descripción
			ProductoDto[] productos = this.RealizarBusqueda(consulta);
			
			// Quito todos los valores actuales del modelo del Entry
			ListStore modelo = (ListStore) entry.Completion.Model;
			modelo.Clear();
			
			// Verifico si se han devuelto resultados
			if (productos.Length == 0) {
				SMessageDialog
					.MensajeAdvertencia("<b>No se encontraron productos</b> según los " +
					                    "criterios de búsqueda: '" + consulta + "'");
				
				// Como no hubo resultado, vuelvo a colocar el foco sobre la celda de descripción
				this.treeView
					.ColocarCeldaEnModoEdicion(this.tvItemsProducto.GetSelectedPath(),
					                           this.Columna);
				
				return;
			}
			else if (productos.Length == 1 && !this.MostrarMenuConUnicoResultado(productos[0], consulta))
				return;
			
			// Agrego los nuevos valores al modelo del EntryCompletion
			foreach (ProductoDto p in productos)
				modelo.AppendValues(p);
			
			/* FIXME: El código a continuación intenta generar un evento Changed en el Entry para que
			 * se despliegue (popup) el EntryCompletion. No es lo más elegante pero funciona.
			 * Ver una forma más elegante de activar el EntryCompletion. */
			string valorAnterior = entry.Text;
			int posicionAnterior = entry.Position;
			entry.Text = String.Empty;
			entry.InsertText(valorAnterior);
			entry.Position = posicionAnterior;
			
			// Este bucle es parte de la intención anterior de mostrar el popup del EntryCompletion
			while (Gtk.Application.EventsPending())
				Gtk.Application.RunIteration();
			
			/* FIXME: Hay que verificar si la búsqueda fue exitosa antes de hacer lo siguiente.
			 * Habría que de alguna manera mostrarle al usuario que no ha habido resultados, pero
			 * sin mostrar un mensaje, porque es muy intrusivo. */
			this.estado = Estado.Seleccion;
			args.RetVal = true;
		}
		
		protected void OnEntryChanged(object sender, EventArgs e)
		{
			/* Si actualmente estoy en el estado de búsqueda, retorno.
			 * El objetivo de este método es cambiar al estado Busqueda cuando el
			 * usuario cambia el contenido del Entry, para así saber qué hacer con un
			 * Return precionado (si para realizar la petición al servidor o para seleccionar
			 * un item del EntryCompletion). */
			if (this.estado == Estado.Busqueda)
				return;
			
			// Vacío el modelo del EntryCompletion
			Entry entry = (Entry) sender;
			(entry.Completion.Model as ListStore).Clear();
			entry.Completion.Complete();
			
			// Cambio el estado a Busqueda
			this.estado = Estado.Busqueda;
		}
		#endregion
		
		#region Métodos EntryCompletion
		private ProductoDto RenderProductoAbstracto(CellLayout column,
		                                            CellRenderer cell,
		                                            TreeModel model,
		                                            TreeIter iter)
		{
			ProductoDto producto = (ProductoDto) model.GetValue(iter, 0);
			if (producto == null) return null;
			
			CellRendererText cellText = (CellRendererText) cell;
			
			if (producto != null && producto.productoAbstracto)
				cellText.Underline = Pango.Underline.Single;
			else
				cellText.Underline = Pango.Underline.None;
			
			return producto;
		}
		
		protected void RenderCodigoEntryCompletion(CellLayout cellLayout, CellRenderer renderer,
		                                           TreeModel model, TreeIter iter)
		{
			ProductoDto producto =
				this.RenderProductoAbstracto(cellLayout, renderer, model, iter);
			
			string valor = producto != null ? producto.codigo : String.Empty;
			
			(renderer as CellRendererText).Text = valor;
		}
		
		protected void RenderDescripcionEntryCompletion(CellLayout cellLayout, CellRenderer renderer,
		                                                TreeModel model, TreeIter iter)
		{
			ProductoDto producto =
				this.RenderProductoAbstracto(cellLayout, renderer, model, iter);
			
			string valor = producto != null ? producto.descripcion : String.Empty;
			
			(renderer as CellRendererText).Text = valor;
		}
		
		protected void RenderPrecioVentaEntryCompletion(CellLayout cellLayout, CellRenderer renderer,
		                                                TreeModel model, TreeIter iter)
		{
			ProductoDto producto =
				this.RenderProductoAbstracto(cellLayout, renderer, model, iter);
			
			string valor = producto != null ? producto.importeVenta.ToString("N2") : String.Empty;
			
			CellRendererText cellText = (CellRendererText) renderer;
			
			cellText.Text = valor;
		}
		
		protected void RenderUltimaColumnaEntryCompletion(CellLayout cellLayout, CellRenderer renderer,
		                                                  TreeModel model, TreeIter iter)
		{
			CellRendererText cellText = (CellRendererText) renderer;
			cellText.Text = "   ";
		}
		
		/// <summary>
		/// En este método nos encargamos de realizar las acciones necesarias cuando el
		/// usuario escoge un producto de la lista del EntryCompletion.
		/// </summary>
		/// <param name="o">
		/// A <see cref="System.Object"/>
		/// </param>
		/// <param name="args">
		/// A <see cref="MatchSelectedArgs"/>
		/// </param>
		[GLib.ConnectBefore]
		protected void OnEntryCompletionMatchSelected(object o, MatchSelectedArgs args)
		{
			ProductoDto producto = (ProductoDto) args.Model.GetValue(args.Iter, 0);
			
			/* Verifico que el código del producto sea único en todos los items */
			if (this.tvItemsProducto.ProductoEstaCargado(producto.codigo,
			                                             producto.descripcion)) {
				SMessageDialog.MensajeAdvertencia("El producto ya está cargado en los items");
				
				this.treeView
					.ColocarCeldaEnModoEdicion(this.tvItemsProducto.GetSelectedPath(),
					                           this.Columna);
				
				return;
			}
			
			/* Al seleccionar el producto de la lista, me aseguro que el TreeView asigne
			 * en modo de escritura a la siguiente columna, asignando
			 * ForzarSetCursor a 'true'. Esto es necesario porque cuando hay un click
			 * del ratón, se deshabilita. */
			this.tvItemsProducto.ForzarSetCursor = true;
			
			this.tvItemsProducto.CargarProducto(producto);
			
			args.RetVal = true;
		}
		#endregion
		
		#region Métodos y propiedades abstractas
		protected abstract TreeViewColumn Columna {
			get;
		}
		
		protected abstract int TamañoMinimoConsulta {
			get;
		}
		
		protected abstract bool NoRealizarConsulta(string consultaEscrita,
		                                           ItemProductoDto item,
		                                           KeyPressEventArgs args);
		
		protected abstract ProductoDto[] RealizarBusqueda(string consulta);
		
		protected abstract bool MostrarMenuConUnicoResultado(ProductoDto producto, string consulta);
		#endregion
	}
}
