// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Gtk;

using Stratos.ModeloDto;
using StratosClient.Core.Gui;
using StratosServer.Controladores;

namespace CustomWidgets
{
	public class DescripcionProductoCellRenderer : CellRendererMenuDesplegable
	{
		public DescripcionProductoCellRenderer(ListaItemsProducto treeView) : base(treeView)
		{
		}
		
		protected override bool ValidarDatos (ItemProductoDto item, string nuevoValor)
		{
			// Valido que el nuevo valor no sea vacío
			if (nuevoValor.Trim().Equals(String.Empty))
				return false;
			
			/* Si el item de produto cargado no es abstracto, entonces no
			 * se puede editar la descripción. */
			if (!item.productoAbstracto)
				return false;
			
			/* Verifico que la descripción del producto sea única en todos los items */
			if (this.tvItemsProducto.ProductoEstaCargado(item.productoCodigo, nuevoValor))
				return false;
			
			return true;
		}
		
		protected override TreeViewColumn ModificarItem (ItemProductoDto item,
		                                                 string nuevoValor)
		{
			/* Si se trata de un producto abstracto, entonces verifico que se
			 * haya modificado correctamente la descripción: un producto abstracto debe
			 * comenzar si o si con la descripción original del producto y un agregado
			 * opcional. */
			if (item.productoAbstracto && 
			    !(nuevoValor.StartsWith(item.productoDescripcion))) {
				
				SMessageDialog
					.MensajeAdvertencia("Cuando se trata de un producto genérico, " +
					                    "solamente se puede agregar texto a la descripción existente.\n\n" +
					                    "Si desea realizar <b>otra búsqueda</b> de produtos, escriba lo " +
					                    "que desea buscar, y presione <b>Control+Enter</b>");
				
				return this.tvItemsProducto.tvcDescripcionProducto;
			}
			
			item.descripcion = nuevoValor;
			
			return null;
		}
		
		protected override void EdicionIniciada (Gtk.Entry entry, string path, ItemProductoDto item)
		{
			base.EdicionIniciada(entry, path, item);
			
			entry.FocusGrabbed += delegate {
				/* Si se trata de un producto cargado y abstracto, coloco el cursor
				 * al final listo para editar la descripción. */
				if (item.productoAbstracto) {
					// Agrego un espacio al final si la descripción del producto
					// no ha sido cambiada
					if (!entry.Text.EndsWith(" "))
						entry.Text = entry.Text + " ";
					
					entry.Position = entry.Text.Length;
				}
			};
		}
		
		protected override bool ComenzarEdicion ()
		{
			return true;
		}
		
		#region CellRendererMenuDesplegable - Implementación
		protected override TreeViewColumn Columna {
			get { return this.tvItemsProducto.tvcDescripcionProducto; }
		}
		
		protected override int TamañoMinimoConsulta {
			get { return 3; }
		}
		
		protected override bool NoRealizarConsulta (string consultaEscrita,
		                                            ItemProductoDto item,
		                                            KeyPressEventArgs args)
		{
			/* Además de las condiciones de la clase base, no realizo la consulta por descripción si:
			 * - Hay actualmente cargado un producto abstracto, no se presionó la tecla Control y
			 *   se presionó la tecla Enter.
			 * - El nuevo valor ingresado es igual al actualmente asignado al item. */
			return
				(item.productoAbstracto &&
			     		(args.Event.State != Gdk.ModifierType.ControlMask &&
			      			(args.Event.Key == Gdk.Key.Return || args.Event.Key == Gdk.Key.KP_Enter))) ||
				(item.descripcion == consultaEscrita);
		}
		
		protected override ProductoDto[] RealizarBusqueda (string consulta)
		{
			// Realizo la búsqueda de productos por descripción
			ProductoDto[] productos =
				((ControladorProductosPrx)MainIce.GetControlador("ControladorProductos"))
					.BuscarPorDescripcion(consulta);
			
			return productos;
		}
		
		protected override bool MostrarMenuConUnicoResultado (ProductoDto producto, string consulta)
		{
			// Siempre muestro la lista de resultados cuando busco por descripción
			return true;
		}
		#endregion
	}
}
