// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

using Gtk;

using Stratos.Common;
using Stratos.ModeloDto;
using StratosClient.Core.Gui;

namespace CustomWidgets
{
	/// <summary>
	/// TreeView para seleccionar productos para la modificación de sus precios.
	/// </summary>
	public class ListaPreciosProductos : ListaSeleccionable<ProductoDto>
	{
		#region Campos
		public TreeViewColumn tvcCodigo;
		public TreeViewColumn tvcDescripcion;
		public TreeViewColumn tvcPrecioSinIva;
		public TreeViewColumn tvcPrecioConIva;
		#endregion
		
		#region Constructores
		public ListaPreciosProductos () : base()
		{
			this.Selection.Mode = SelectionMode.Single;
			
			#region Columnas
			TreeViewColumn tvcColumna;
			CellRendererText crtColumnaRender;
			
			tvcColumna = new TreeViewColumn();
			tvcColumna.Title = "Código";
			crtColumnaRender = new CellRendererText();
			tvcColumna.PackStart(crtColumnaRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaRender, this.RenderCodigo);
			this.AppendColumn(tvcColumna);
			this.tvcCodigo = tvcColumna;
			
			tvcColumna = new TreeViewColumn();
			tvcColumna.Title = "Descripción";
			tvcColumna.Expand = true;
			crtColumnaRender = new CellRendererText();
			tvcColumna.PackStart(crtColumnaRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaRender, this.RenderDescripcion);
			this.AppendColumn(tvcColumna);
			this.tvcDescripcion = tvcColumna;
			
			tvcColumna = new TreeViewColumn();
			tvcColumna.Title = "Precio sin IVA";
			crtColumnaRender = new PrecioSinIvaCellRenderer(this);
			crtColumnaRender.Xalign = 1.0f;
			crtColumnaRender.Alignment = Pango.Alignment.Right;
			tvcColumna.PackStart(crtColumnaRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaRender, this.RenderPrecioSinIva);
			this.AppendColumn(tvcColumna);
			this.tvcPrecioSinIva = tvcColumna;

			tvcColumna = new TreeViewColumn();
			tvcColumna.Title = "Precio con IVA";
			crtColumnaRender = new PrecioConIvaCellRenderer(this);
			crtColumnaRender.Xalign = 1.0f;
			crtColumnaRender.Alignment = Pango.Alignment.Right;
			tvcColumna.PackStart(crtColumnaRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaRender, this.RenderPrecioConIva);
			this.AppendColumn(tvcColumna);
			this.tvcPrecioConIva = tvcColumna;
			#endregion
		}
		#endregion
		
		#region Métodos de renderizado
		private void RenderCodigo(TreeViewColumn column,
		                          CellRenderer cell,
		                          TreeModel model,
		                          TreeIter iter)
		{
			ProductoDto producto = this.ObtenerElemento(iter).Elemento;
			string valor = producto != null ? producto.codigo.ToString() : String.Empty;
			
			(cell as CellRendererText).Text = valor;
		}

		private void RenderDescripcion(TreeViewColumn column,
		                               CellRenderer cell,
		                               TreeModel model,
		                               TreeIter iter)
		{
			ProductoDto producto = this.ObtenerElemento(iter).Elemento;
			string valor = producto != null ? producto.descripcion : String.Empty;
			
			(cell as CellRendererText).Text = valor;
		}
		
		private void RenderPrecioSinIva(TreeViewColumn column,
		                                   CellRenderer cell,
		                                   TreeModel model,
		                                   TreeIter iter)
		{
			ProductoDto producto = this.ObtenerElemento(iter).Elemento;
			string valor = producto != null ? producto.importeSinIva.ToString("N2") : String.Empty;
			
			(cell as CellRendererText).Text = valor;
		}

		private void RenderPrecioConIva(TreeViewColumn column,
		                                   CellRenderer cell,
		                                   TreeModel model,
		                                   TreeIter iter)
		{
			ProductoDto producto = this.ObtenerElemento(iter).Elemento;
			string valor = producto != null ? producto.importeVenta.ToString("N2") : String.Empty;
			
			(cell as CellRendererText).Text = valor;
		}
		#endregion
		
		#region Metodos públicos
		public long[] IdsProductosSeleccionados()
		{	
			List<long> productosSeleccionados = new List<long>();
			
			foreach (ProductoDto producto in this.ElementosSeleccionados)
				productosSeleccionados.Add(producto.id);
			
			return productosSeleccionados.ToArray();
		}
		
		public void ActualizarProductosSeleccionados(AccionModificacionPrecio accion)
		{
			double diferencia;
			ProductoDto temp;
			
			foreach (TreeIter iter in this.itersChequeados) {
				diferencia = 0.00d;
				
				// Obtengo el producto seleccionable del modelo
				temp = this.ObtenerElemento(iter).Elemento;
				
				// Calculo la diferencia a aplicar, segun el tipo de cambio
				if (accion.cambioTipo == TipoCambio.Porcentual)
					diferencia = temp.importeSinIva * (accion.cantidad/100.0);
				else if (accion.cambioTipo == TipoCambio.PrecioFijo)
					diferencia = accion.cantidad;
				
				// Si el tipo de acción es 'Disminuir' multiplico la diferencia por -1
				if (accion.accion == TipoAccion.Disminuir)
					diferencia = (-1) * diferencia;
				
				// Actualizo los importes del producto
				temp.importeSinIva = StratosUtil.RoundDouble(temp.importeSinIva + diferencia, 2);
				temp.importeVenta = StratosUtil.RoundDouble(temp.importeSinIva * 1.21, 2);
				
				// Actualizo el modelo del treeview
				this.Model.EmitRowChanged(this.Model.GetPath(iter), iter);
			}
		}
		#endregion
		
		#region ListaElementos implementation
		protected override bool ElementosIguales(ProductoDto p1, ProductoDto p2)
		{
			return (p1.id == p2.id);
		}
		#endregion
	}
}
