// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Gtk;

using StratosClient.Core.Gui;
using Stratos.ModeloDto;
using Stratos.Common;
using Stratos.Common.Excepciones;
using StratosServer.Controladores;

namespace CustomWidgets
{
	public abstract class PrecioCellRendererBase : CellRendererBase<ElementoSeleccionable<ProductoDto>>
	{
		#region Campos
		protected ListaPreciosProductos listaPreciosProducto;
		#endregion
		
		#region Constructores
		public PrecioCellRendererBase(ListaPreciosProductos treeView) : base(treeView)
		{
			this.listaPreciosProducto = treeView;
			this.Xalign = 1.0f;
		}
		#endregion
		
		#region Otros métodos
		protected double ObtenerPrecioRedondeado(string precioCadena)
		{
			return StratosUtil.ParseDouble(precioCadena, 2);
		}
		
		protected double ObtenerPrecioRedondeado(double precio)
		{
			return StratosUtil.RoundDouble(precio, 2);
		}
		#endregion
		
		#region CellRendererBase implementation
		protected override bool ValidarDatos (ElementoSeleccionable<ProductoDto> item, string nuevoValor)
		{
			double valorNumerico = 0;
		
			// Chequeo que la cantidad sea numérica (entera o decimal)
			if (StratosUtil.IsDouble(nuevoValor))
				valorNumerico = this.ObtenerPrecioRedondeado(nuevoValor);
			else
				return false;
			
			// Esa cantidad numérica debe ser mayor a cero
			if (valorNumerico <= 0)
				return false;
			
			return true;
		}
		
		protected override void ModificarItem (ElementoSeleccionable<ProductoDto> elementoSeleccionable,
		                                       string nuevoValor)
		{
			ProductoDto producto = elementoSeleccionable.Elemento;
			ProductoDto productoActualizado;
			
			double nuevoPrecio = this.ObtenerPrecioRedondeado(nuevoValor);
			
			/* Pregunto a la subclase si realmente es necesario realizar una modificación,
			 * que implica llamar al Controlador de Productos y actualizar el producto en
			 * el modelo del TreeView. Por ejemplo, si no se ha modificado el precio, no
			 * sería necesaria una modificacion. */
			if (this.ModificacionNecesaria(producto, nuevoPrecio)) {
				
				// Delego a la subclase la modificación del producto con el nuevo valor.
				this.ModificarPrecio(producto, nuevoPrecio);
				
				// Obtengo el controlador y actualizo el producto en la base de datos
				ControladorProductosPrx controladorProductos = (ControladorProductosPrx)
					MainIce.GetControlador("ControladorProductos");
				
				try {
					productoActualizado = (ProductoDto) controladorProductos.Modificar(producto);
					
					// Actualizo la lista con el nuevo producto actualizado
					ElementoSeleccionable<ProductoDto> elementoNuevo =
						new ElementoSeleccionable<ProductoDto>(productoActualizado);
					
					this.listaPreciosProducto.ActualizarObjetoSeleccionado(elementoNuevo);
				}
				catch (StratosException ex) {
					SMessageDialog.MensajeError(ex.mensaje, ex);
					return;
				}
			}
			
			// Obtengo la siguiente fila
			TreeIter iterSiguiente;
			TreePath[] seleccionadas = this.listaPreciosProducto.Selection.GetSelectedRows();
			
			if (seleccionadas.Length == 0)
				return;
			
			this.listaPreciosProducto.Model.GetIter(out iterSiguiente, seleccionadas[0]);
			
			/* Avanzo el iter al siguiente. Si llegué al final, no sigo editando */
			if (!this.listaPreciosProducto.Model.IterNext(ref iterSiguiente))
				return;
			
			// Pasamos a la siguiente fila para modificar el precio
			this.listaPreciosProducto
				.ColocarCeldaEnModoEdicion(this.listaPreciosProducto.Model.GetPath(iterSiguiente),
				                           this.Columna);
		}
		
		protected override bool ComenzarEdicion ()
		{
			return true;
		}
		#endregion
		
		#region Propiedades y métodos abstractos
		protected abstract bool ModificacionNecesaria(ProductoDto producto,
		                                                   double nuevoPrecio);
		
		protected abstract void ModificarPrecio(ProductoDto producto,
		                                        double nuevoPrecio);
		
		protected abstract TreeViewColumn Columna {
			get;
		}
		#endregion
	}
}
