// Copyright (C) 2009, 2010 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using Gtk;

using Stratos.ModeloDto;
using StratosClient.Core.Gui;

namespace CustomWidgets
{
	/// <summary>
	/// Permite crear un TreeView con una columna para seleccionar elementos, con un
	/// botón CheckButton. Esta clase es genérica y abstrae de todas las operaciones
	/// relacionadas con la selección de los elementos.
	/// </summary>
	public abstract class ListaSeleccionable<T> : ListaElementos<ElementoSeleccionable<T>>, IListaSeleccionable<T>
		where T : class
	{
		#region Campos
		public event System.Action SeleccionElementoCambiada;
		
		protected System.Func<T, bool> permitirSeleccion;
		protected string mensajeSeleccionNoPermitida;
		
		public TreeViewColumn tvcSeleccionado;
		
		protected List<TreeIter> itersChequeados;
		
		// Checkbox para el titulo de la columna
		private CheckButton chkSeleccionarTodo;
		#endregion
		
		#region Constructores
		public ListaSeleccionable () : base()
		{
			// Mensaje por defecto cuando no se permite la selección de un elemento
			this.mensajeSeleccionNoPermitida = "No se permite la selección del elemento";
			
			#region Columna con Checkboxes
			CellRendererToggle crtColumnaToggleRender = new CellRendererToggle();
			
			// Hago que los checkboxes de las celdas sean activables
			crtColumnaToggleRender.Activatable = true;
			
			// Conecto el evento togged de los checkboxes de las celdas
			crtColumnaToggleRender.Toggled +=  this.OnFilaSeleccionada;
			
			// Creo la columna
			TreeViewColumn tvcColumna = new TreeViewColumn("", crtColumnaToggleRender);
			
			// Hago que no se pueda modificar su tamaño
			tvcColumna.Resizable = false;
			
			// Creo el botón del titulo de la columna
			this.chkSeleccionarTodo = new CheckButton();
			this.chkSeleccionarTodo.TooltipText = "Seleccionar Todo/Nada";
			this.chkSeleccionarTodo.Active = false;
			this.chkSeleccionarTodo.Show();
			
			// Seteo el botón en el título de la columna
			tvcColumna.Widget = this.chkSeleccionarTodo;
			tvcColumna.Clickable = true;
			
			// Conecto el evento clicked de la columna
			tvcColumna.Clicked += this.OnColumnaSeleccionadoClicked;
			
			tvcColumna.PackStart(crtColumnaToggleRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaToggleRender, this.RenderSeleccion);
			this.AppendColumn(tvcColumna);
			this.tvcSeleccionado = tvcColumna;
			#endregion
			
			this.itersChequeados = new List<TreeIter>();
		}
		#endregion
		
		#region Propiedades
		public override bool SoloLectura {
			set {
				base.SoloLectura = value;
				
				this.tvcSeleccionado.Clickable = !value;
				
				ElementoSeleccionable<T> elemento;
				
				foreach (TreeIter iter in this.Model.Traverse()) {
					elemento = (ElementoSeleccionable<T>) this.Model.GetValue(iter, 0);
					elemento.SoloLectura = value;
					
					this.Model.EmitRowChanged(this.Model.GetPath(iter), iter);
				}
			}
		}
		
		public System.Func<T, bool> PermitirSeleccion {
			get { return this.permitirSeleccion; }
			set { this.permitirSeleccion = value; }
		}
		
		public string MensajeSeleccionNoPermitida {
			get { return this.mensajeSeleccionNoPermitida; }
			set { this.mensajeSeleccionNoPermitida = value; }
		}
		
		/// <summary>
		/// Devuelve los elementos primarios seleccionados.
		/// </summary>
		public List<T> ElementosSeleccionados {
			get {
				// Busco todos los elementos seleccionados.
				var elementosSeleccionablesSeleccionados =
					from elementoSeleccionable in this.ObtenerTodos().Cast<ElementoSeleccionable<T>>()
					where elementoSeleccionable.Seleccionado == true
					select elementoSeleccionable.Elemento;
				
				return elementosSeleccionablesSeleccionados.ToList();
			}
		}
		
		/// <summary>
		/// Devuelve los elementos primarios seleccionados, pero solo de aquellos
		/// que pueden ser seleccionados. Es decir, excluye los elementos de sólo
		/// lectura.
		/// </summary>
		public List<T> ElementosSeleccionablesSeleccionados {
			get {
				/* Busco solo los elementos seleccionados que son seleccionables
				 * (SoloLectura = false). */
				var elementosSeleccionablesSeleccionados =
					from elementoSeleccionable in this.ObtenerTodos().Cast<ElementoSeleccionable<T>>()
					where elementoSeleccionable.Seleccionado == true &&
						elementoSeleccionable.SoloLectura == false
					select elementoSeleccionable.Elemento;
				
				return elementosSeleccionablesSeleccionados.ToList();
			}
		}
		
		public IList ListaItems {
			get { return this.ObtenerTodosElementosPrimarios(); }
		}
		#endregion
		
		#region Métodos de renderizado
		private void RenderSeleccion(TreeViewColumn column,
		                          CellRenderer cell,
		                          TreeModel model,
		                          TreeIter iter)
		{
			ElementoSeleccionable<T> elemento = this.ObtenerElemento(iter);
			
			CellRendererToggle cellToggle = (CellRendererToggle) cell;
			
			cellToggle.Active = elemento.Seleccionado;
			cellToggle.Sensitive = !elemento.SoloLectura;
			cellToggle.Activatable = !elemento.SoloLectura;
		}
		#endregion
		
		#region Eventos
		protected virtual void OnFilaSeleccionada(object o, ToggledArgs args)
		{
			TreeIter iter;
			
			// Verifico que el path apuntado en args.Path sea válido
			if (!this.Modelo.GetIter(out iter, new TreePath(args.Path)))
				return;
			
			// Invierto el valor de seleccionado de la fila
			this.SeleccionarDeseleccionarElemento(iter);
		}
		
		protected virtual void OnColumnaSeleccionadoClicked(object o, EventArgs args)
		{
			// Si el treeview está vacío, retorno y no hago nada
			if (this.Modelo.IterNChildren() <= 0)
				return;
			
			// Cambio el estado del checkbox
			bool seleccionarTodo = !this.chkSeleccionarTodo.Active;
			
			// Si el checkbox no está activado...
			if (!seleccionarTodo) {
				// Itero y selecciono todos los items hasta que no se permita seleccionar más
				foreach (TreeIter iter in this.Modelo.Traverse())
					this.SeleccionarDeseleccionarElemento(iter, false);
			}
			
			/* Si el checkbox 'Seleccionar Todo' está activado, entonces tengo que activar los
			 * checkboxes de todas las filas y agrego a todos los elementos a la lista de
			 * itersChequeados.
			 * En esta iteración avanzo siempre que se haya permitido la seleccion. */
			else {
				bool seleccionNoPermitida;
				
				foreach (TreeIter iter in this.Modelo.Traverse()) {
					this.SeleccionarDeseleccionarElemento(iter, true, out seleccionNoPermitida);
					
					if (seleccionNoPermitida)
						break;
				}
			}
		}
		#endregion
		
		#region Métodos
		#region SeleccionarDeseleccionarElemento
		/// <summary>
		/// Dado un TreeIter, marca el elemento correspondiente como seleccionado
		/// o no, según el valor pasado como argumento.
		/// </summary>
		/// <param name="iter">
		/// A <see cref="TreeIter"/>
		/// </param>
		/// <param name="seleccionado">
		/// A <see cref="System.Boolean"/>
		/// </param>
		protected void SeleccionarDeseleccionarElemento(TreeIter iter,
		                                                bool seleccionado,
		                                                out bool noSePermiteSeleccion)
		{
			ElementoSeleccionable<T> elementoSeleccionable = this.ObtenerElemento(iter);
			
			noSePermiteSeleccion = false;
			
			// Si el elemento es de solo lectura, entonces no puedo cambiar su
			// estado de selección
			if (elementoSeleccionable.SoloLectura)
				return;
			
			// Verifico si se permite la selección del elemento
			if (seleccionado && this.permitirSeleccion != null &&
			    	elementoSeleccionable.Seleccionado == false &&
			    	!this.permitirSeleccion(elementoSeleccionable.Elemento)) {
				SMessageDialog.MensajeAdvertencia(this.mensajeSeleccionNoPermitida);
				noSePermiteSeleccion = true;
				return;
			}
			
			/* Actualizo el objeto */
			if (elementoSeleccionable.Seleccionado != seleccionado) {
				elementoSeleccionable.Seleccionado = seleccionado;
				this.Model.EmitRowChanged(this.Model.GetPath(iter), iter);
			}
			else {
				// Si el valor no cambia, entonces no tengo que seguir realizando operaciones
				// y salgo.
				return;
			}
			
			// Si la fila está ahora chequeada
			if (elementoSeleccionable.Seleccionado) {
				// Agrego el iter a la lista de itersChequeados
				this.itersChequeados.Add(iter);
				
				/* Comparo la cantidad de iters en la lista iterChequeados con la cantidad
				 * total de filas en el modelo. Si es igual quiere decir que todos los items
				 * están seleccionados, entonces activo el checkbox 'Seleccionar todo' */
				if (this.itersChequeados.Count == this.Modelo.IterNChildren())
					this.chkSeleccionarTodo.Active = true;
			}
			
			// Si la fila fue 'deschequeada'
			else {
				
				// Quito el iter de la lista de itersChequeados
				this.itersChequeados.Remove(iter);
				
				/* Si el checkbox 'Seleccionar todo' esta activo, ahora tengo que desactivarlo
				 * ya que una fila fue 'deschequeada' */
				if (this.chkSeleccionarTodo.Active)
					this.chkSeleccionarTodo.Active = false;
			}
			
			// Disparo el evento de cambio de selección de un elemento
			if (this.SeleccionElementoCambiada != null)
				this.SeleccionElementoCambiada();
		}
		
		protected void SeleccionarDeseleccionarElemento(TreeIter iter,
		                                                bool seleccionado)
		{
			bool temp;
			this.SeleccionarDeseleccionarElemento(iter, seleccionado, out temp);
		}
		
		
		protected void SeleccionarDeseleccionarElemento(TreeIter iter)
		{
			ElementoSeleccionable<T> elementoSeleccionable = this.ObtenerElemento(iter);
			
			/* Actualizo el objeto (si antes estaba seleccionado, no lo está y viceversa) */
			this.SeleccionarDeseleccionarElemento(iter, !elementoSeleccionable.Seleccionado);
		}
		#endregion
		
		/// <summary>
		/// Asigna a un TreeIter específico el elemento dado.
		/// </summary>
		/// <param name="iter">
		/// A <see cref="TreeIter"/>
		/// </param>
		/// <param name="elemento">
		/// A <see cref="T"/>
		/// </param>
		public void AsignarElemento(TreeIter iter, T elemento)
		{
			this.Modelo.SetValue(iter, 0, new ElementoSeleccionable<T>(elemento));
			this.Modelo.EmitRowChanged(this.Modelo.GetPath(iter), iter);
		}
		
		/// <summary>
		/// Agrega una lista de elementos, y permite especificar si van a aparecer como seleccionados
		/// o no, como solo lectura o no.
		/// Se utiliza de forma interna y centraliza el código común para agregar elementos.
		/// </summary>
		/// <param name="marcarComoSeleccionados">
		/// A <see cref="System.Boolean"/>
		/// </param>
		/// <param name="elementos">
		/// A <see cref="System.Object[]"/>
		/// </param>
		protected void AgregarElementosSeleccionables(bool marcarComoSeleccionados,
		                                              bool marcarComoSoloLectura,
		                                              params object[] elementos)
		{
			/* Este método se encuentra sobreescrito porque el modelo de un ListaSeleccionable
			 * siempre posee objetos del tipo ElementoSeleccionable<T>, por lo tanto debo tratar
			 * correctamente los métodos que agregan elementos. */
			
			// Creo un array de ElementoSeleccionable. Envuelvo cada objeto
			// agregado en un objeto ElementoSeleccionable.
			ElementoSeleccionable<T>[] elementosSeleccionables =
				new ElementoSeleccionable<T>[elementos.Length];
			
			T elementoGenerico;
			ElementoSeleccionable<T> elementoSeleccionable;
			
			for (int i=0; i<elementos.Length; i++) {
				elementoGenerico = (T) elementos[i];
				
				// Creo el objeto ElementoSeleccionable
				elementoSeleccionable = new ElementoSeleccionable<T>(elementoGenerico);
				elementoSeleccionable.Seleccionado = marcarComoSeleccionados;
				elementoSeleccionable.SoloLectura = marcarComoSoloLectura;
				
				// Lo agrego al array
				elementosSeleccionables[i] = elementoSeleccionable;
			}
			
			// Agrego los objetos ElementoSeleccionable<T> y obtengo los TreeIter
			// para poder gestionar correctamente los TreeIters chequeados
			// (ver campo this.itersChequeados).
			TreeIter[] itersAgregados = this.AgregarElementosBase(elementosSeleccionables);
			
			for (int i=0; i<itersAgregados.Length; i++) {
				if (elementosSeleccionables[i].Seleccionado)
					this.itersChequeados.Add(itersAgregados[i]);
			}
		}
		
		/// <summary>
		/// Agrega una lista de elementos como no seleccionados. Se utiliza de forma interactiva
		/// para que el usuario los seleccione.
		/// Esta versión no retorna un TreeIter[] correspondiente a los elementos agregados.
		/// </summary>
		/// <param name="elementos">
		/// A <see cref="System.Object[]"/>
		/// </param>
		public override void AgregarElementos(params object[] elementos)
		{
			this.AgregarElementosSeleccionables(false, false, elementos);
		}
		
		/// <summary>
		/// Agrega una lista de elementos del tipo ElementoSeleccionable.
		/// Esta versión retorna un TreeIter[] correspondiente a los elementos agregados.
		/// </summary>
		/// <param name="elementos">
		/// A <see cref="ElementoSeleccionable<T>[]"/>
		/// </param>
		public override TreeIter[] AgregarElementosBase(params ElementoSeleccionable<T>[] elementos)
		{
			// Seteo el estado del checkbox de la columna seleccionable
			this.chkSeleccionarTodo.Active = false;
			
			return base.AgregarElementosBase(elementos);
		}
		
		public void AgregarElementosComoSeleccionados (params T[] elementos)
		{
			this.AgregarElementosSeleccionables(true, false, elementos);
		}
		
		public List<T> ObtenerTodosElementosPrimarios()
		{
			List<T> todos = new List<T>();
			
			todos.AddRange(from ElementoSeleccionable<T> f
			               		in this.Model.TraverseColumn<ElementoSeleccionable<T>>(0)
			               select f.Elemento);
			
			return todos;
		}
		
		/// <summary>
		/// Marca todos los elementos de la lista como no seleccionados.
		/// </summary>
		public void DeseleccionarTodosLosElementos()
		{
			foreach (TreeIter iter in this.Model.Traverse())
				this.SeleccionarDeseleccionarElemento(iter, false);
		}
		
		protected override bool Iguales(ElementoSeleccionable<T> t1, ElementoSeleccionable<T> t2)
		{
			return this.ElementosIguales(t1.Elemento, t2.Elemento);
		}
		
		public override void Vaciar()
		{
			// Vacío la lista de iters chequeados
			this.itersChequeados.Clear();

			// Seteo el checkbox para seleccionar todo en false
			this.chkSeleccionarTodo.Active = false;
			
			// Llamo al método Limpiar de la superclase
			base.Vaciar();
		}
		#endregion
		
		#region Métodos abstractos
		protected abstract bool ElementosIguales(T t1, T t2);
		#endregion
	}
}
