// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;

using Gtk;
using WebKit;

using Stratos.Common;

namespace StratosClient.Common
{
	public class ImpresionDocumento : GeneradorDocumento
	{
		#region Campos
		/// <summary>
		/// Utilizo una sola instancia de WebView para renderizar e imprimir
		/// todos los documentos.
		/// </summary>
		protected static WebView webView;
		
		/// <summary>
		/// Evento que se ejecuta cuando la operación de impresión
		/// (PrintOperation) ha finalizado todas las tareas para enviar
		/// el trabajo a imprimir. Tener en cuenta que esto no significa
		/// que la impresora ya haya terminado de imprimir.
		/// </summary>
		public event System.Action ImpresionEjecutada;
		
		/// <summary>
		/// Evento que se dispara cuando el usuario ha escogido, del diálogo de
		/// impresión, la opción "Imprimir" (es decir, no lo canceló ni lo
		/// previsalizó, sino que lo va a imprimir en una impresora).
		/// </summary>
		public event System.Action AccionImprimirElegida;
		
		protected string contenidoTemplate;
		protected string directorio;
		
		// Márgenes en milímetros
		protected double margenArriba;
		protected double margenDerecha;
		protected double margenAbajo;
		protected double margeIzquierda;
		
		protected PageOrientation orientacionPagina;
		#endregion
		
		#region Constructores
		static ImpresionDocumento()
		{
			/* Creo el widget WebView para renderizar el archivo HTML.
			 * Al widget WebView no lo muestro en pantalla, sino que directamente muestro el
			 * diálogo para imprimir. Simplemente uso este widget imprimir el html. */
			webView = new WebView();
		}
		
		public ImpresionDocumento (string archivoTemplate, string directorio,
		                          string nombreObjeto, object objeto)
			: base (archivoTemplate, nombreObjeto, objeto)
		{
			this.directorio = directorio;
			
			/* Me suscribo al evento LoadFinished, de esta forma cuando termina de cargar el
			 * archivo HTML, recién ahí, envío la petición de impresión. */
			webView.LoadFinished += this.CargaWebViewFinalizada;
			
			// Márgenes
			this.margeIzquierda = 10.00;
			this.margenAbajo = 10.00;
			this.margenArriba = 10.00;
			this.margenDerecha = 10.00;
			
			this.orientacionPagina = PageOrientation.Portrait;
		}
		
		public ImpresionDocumento(string archivoTemplate, string directorio)
			: this (archivoTemplate, directorio, null, null)
		{
		}
		
		/// <summary>
		/// Esta variante del constructor, permite pasar directamente el contenido
		/// del template como cadena, y no un nombre de archivo.
		/// </summary>
		/// <param name="contenidoTemplate">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="directorio">
		/// A <see cref="System.String"/>
		/// </param>
		public ImpresionDocumento(string contenidoTemplate)
			: this(null, null, null, null)
		{
			this.contenidoTemplate = contenidoTemplate;
		}
		#endregion
		
		#region Propiedades
		public double MargenDerecha {
			get { return margenDerecha; }
			set { margenDerecha = value; }
		}
		
		public double MargenArriba {
			get { return margenArriba; }
			set { margenArriba = value; }
		}
		
		public double MargenAbajo {
			get { return margenAbajo; }
			set { margenAbajo = value; }
		}
		
		public double MargeIzquierda {
			get { return margeIzquierda; }
			set { margeIzquierda = value; }
		}
		
		public PageOrientation OrientacionPagina {
			get { return this.orientacionPagina; }
			set { this.orientacionPagina = value; }
		}
		#endregion
		
		#region Métodos
		public void Imprimir()
		{
			// Obtengo el contenido HTML, ya sea desde un template, o desde
			// el contenido mismo pasado en el constructor.
			string contenidoHtml;
			
			if (this.contenidoTemplate != null)
				contenidoHtml = this.contenidoTemplate;
			else
				contenidoHtml = this.CrearHTML();
			
			// Cargo el resultado del merge con NVelocity y lo muestro en el diálogo
			webView.LoadHtmlString(contenidoHtml,
			                       "file:///" + StratosUtil.DirectorioAplicacion() +
			                       		Path.DirectorySeparatorChar +
			                       		this.directorio != null ? this.directorio : String.Empty);
		}
		
		protected void CargaWebViewFinalizada(object o, LoadFinishedArgs args)
		{
			Gtk.Application.Invoke(delegate {
				while (Gtk.Application.EventsPending())
					Gtk.Application.RunIteration();
				
				// Creo un objeto PrintOperation y lo preparo
				PrintOperation printOperation = new PrintOperation();
				
				#region Código para disparar eventos del PrintOperation
				bool previewActivado = false;
				
				printOperation.Ready += delegate {
					previewActivado = true;
				};
				
				printOperation.Done += delegate(object doneObj, DoneArgs doneArgs) {
					// Si el usuario escogió imprimir el documento en una impresora
					// (no visualizarlo ni cancelarlo), disparo el evento correspondiente
					if (!previewActivado && doneArgs.Result == PrintOperationResult.Apply &&
					    	this.AccionImprimirElegida != null)
						this.AccionImprimirElegida();
				};
				#endregion
				
				// Preparo el PrintOperation y ejecuto la impresión con WebKit, según la configuración
				// del PrintOperation y la acción que ha devuelto.
				try {
					PrintOperationAction printOperationAction =
						this.PrepararPrintOperation(printOperation);
				
					webView.MainFrame.PrintFull(printOperation, printOperationAction);
				}
				catch (System.Exception ex) {
					throw new Stratos.Common.ExcepcionStratosException("Hubo un error al imprimir. Inténtenlo nuevamente", false);
				}
				finally {
					// Disparo el evento de ImpresionEjecutada
					if (this.ImpresionEjecutada != null)
						this.ImpresionEjecutada();
					
					/* Siempre al finalizar la impresión, desuscribo este método del
					 * evento LoadFinished del WebView. */
					webView.LoadFinished -= this.CargaWebViewFinalizada;
					
					printOperation.Dispose();
				}
			});
		}
		
		/// <summary>
		/// Este método permite a las subclases modificar el objeto PrintOperation,
		/// y cambiar la acción que ejecutará WebKit con el valor de retorno.
		/// Por defecto, simplemente retorno la acción de mostrar el diálogo.
		/// </summary>
		protected virtual PrintOperationAction PrepararPrintOperation(PrintOperation printOperation)
		{
			// Por defecto, asigno unos valores para los márgenes del documento
			PageSetup pageSetup = new PageSetup();
			
			pageSetup.SetTopMargin(this.MargenArriba, Unit.Mm);
			pageSetup.SetLeftMargin(this.margeIzquierda, Unit.Mm);
			pageSetup.SetRightMargin(this.MargenDerecha, Unit.Mm);
			pageSetup.SetBottomMargin(this.MargenAbajo, Unit.Mm);
			
			pageSetup.Orientation = this.OrientacionPagina;
			
			printOperation.DefaultPageSetup = pageSetup;
			
			// Siempre muestro el diálogo de impresión.
			return PrintOperationAction.PrintDialog;
		}
		#endregion
	}
}