// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;

using Gtk;

namespace StratosClient.Core.Gui
{
	public enum EstadoObjeto {
		Nuevo,
		Cargado,
		Modificado,
		Guardado
	};
	
	/// <summary>
	/// Clase para gestionar los estados de los objetos cargados en las
	/// interfaces gráficas.
	/// </summary>
	public class AdministradorEstadoObjeto
	{
		#region Campos
		private readonly string[] OBJETO_NUEVO = new string[] {"#000000", "NUEVO" };
		private readonly string[] OBJETO_CARGADO = new string[] {"#00CC00", "CARGADO" };
		private readonly string[] OBJETO_GUARDADO = new string[] {"#0000CC", "GUARDADO" };
		private readonly string[] OBJETO_MODIFICADO = new string[] {"#E6AC00", "MODIFICADO" };
		
		private EstadoObjeto estadoObjetoActual;
		
		private Dictionary<EstadoObjeto, string[]> mapeoEstadosEtiquetas;
		
		private System.Action<string> cambiarEstado;
		private System.Action<bool> activarGuardar;
		
		public event System.Action<EstadoObjeto> EstadoCambiado;
		#endregion
		
		#region Constructores
		public AdministradorEstadoObjeto(System.Action<string> cambiarEstado,
		                                 System.Action<bool> activarGuardar)
		{
			this.cambiarEstado = cambiarEstado;
			this.activarGuardar = activarGuardar;
			
			// Mapeo de objetos EstadoObjeto a arrays con colores y etiquetas
			this.mapeoEstadosEtiquetas = new Dictionary<EstadoObjeto, string[]>();
			this.mapeoEstadosEtiquetas.Add(EstadoObjeto.Nuevo, OBJETO_NUEVO);
			this.mapeoEstadosEtiquetas.Add(EstadoObjeto.Cargado, OBJETO_CARGADO);
			this.mapeoEstadosEtiquetas.Add(EstadoObjeto.Modificado, OBJETO_MODIFICADO);
			this.mapeoEstadosEtiquetas.Add(EstadoObjeto.Guardado, OBJETO_GUARDADO);
		}
		
		public AdministradorEstadoObjeto(System.Action<string> cambiarEstado)
			: this(cambiarEstado, x => {})
		{
		}
		#endregion
		
		#region Propiedades
		public EstadoObjeto EstadoObjeto {
			get { return this.estadoObjetoActual; }
		}
		
		public bool EsNuevo {
			get { return this.estadoObjetoActual == EstadoObjeto.Nuevo; }
		}
		
		public bool EstaCargado {
			get { return this.estadoObjetoActual == EstadoObjeto.Cargado; }
		}
		
		public bool EstaModificado {
			get { return this.estadoObjetoActual == EstadoObjeto.Modificado; }
		}
		
		public bool EstaGuardado {
			get { return this.estadoObjetoActual == EstadoObjeto.Guardado; }
		}
		
		/// <summary>
		/// Método para centralizar la pregunta al usuario de si está seguro pisar las modificaciones.
		/// Si no hay modificaciones o el usuario está seguro, retorna true. En caso contrario
		/// retorna false.
		/// </summary>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public bool UsuarioEstaSeguroDePerderModificaciones
		{
			get {
				if (!this.EstaModificado)
					return false;
				
				ResponseType respuesta =
					SMessageDialog
						.MensajePregunta("<b>Hay modificaciones no guardadas.</b>\n" +
							"Si continúa los cambios se perderán. " +
			                 "¿Está seguro que desea perder estos cambios?",
						     ResponseType.No);
				
				if (respuesta != ResponseType.Yes)
					return false;
				
				return true;
			}
		}
		#endregion
		
		#region Métodos
		/// <summary>
		/// Indica a AdministradorEstadoObjeto, que los datos han cambiado.
		/// </summary>
		public void DatosHanCambiado()
		{
			this.CambiarEstadoObjeto(EstadoObjeto.Modificado);
		}
		
		/// <summary>
		/// Cambia el estado del documento actual.
		/// </summary>
		/// <param name="nuevoEstado">
		/// A <see cref="EstadoObjeto"/>
		/// </param>
		public void CambiarEstadoObjeto(EstadoObjeto nuevoEstado)
		{
			string[] datos = this.mapeoEstadosEtiquetas[nuevoEstado];
			this.CambiarEstadoObjeto(nuevoEstado, datos[1], datos[0]);
		}
		
		public void CambiarEstadoObjeto(EstadoObjeto nuevoEstado, string cadenaEstado)
		{
			string[] datos = this.mapeoEstadosEtiquetas[nuevoEstado];
			this.CambiarEstadoObjeto(nuevoEstado, cadenaEstado, datos[0]);
		}
		
		/// <summary>
		/// Esta versión permite cargar un estado definido, pero con una etiqueta
		/// de estado personalizada y un color. 'nuevoEstado' debe ser consistente con la
		/// etiqueta utilizada.
		/// </summary>
		/// <param name="nuevoEstado">
		/// A <see cref="EstadoObjeto"/>
		/// </param>
		/// <param name="cadenaEstado">
		/// A <see cref="System.String"/>
		/// </param>
		public void CambiarEstadoObjeto(EstadoObjeto nuevoEstado, string cadenaEstado, string color)
		{
			this.estadoObjetoActual = nuevoEstado;
			
			string cadenaCompleta = this.FormarCadenaFormateada(color, cadenaEstado);
			
			switch (this.estadoObjetoActual) {
				case EstadoObjeto.Cargado:				
					this.cambiarEstado(cadenaCompleta);
					this.activarGuardar(false);
					break;
				case EstadoObjeto.Guardado:
					this.cambiarEstado(cadenaCompleta);
					this.activarGuardar(false);
					break;
				case EstadoObjeto.Modificado:
					this.cambiarEstado(cadenaCompleta);
					this.activarGuardar(true);
					break;
				case EstadoObjeto.Nuevo:
					this.cambiarEstado(cadenaCompleta);
					this.activarGuardar(true);
					break;
			}
			
			// Lanzo el evento 'EstadoCambiado'
			if (this.EstadoCambiado != null)
				this.EstadoCambiado(this.estadoObjetoActual);
		}
		
		private string FormarCadenaFormateada(string color, string etiqueta)
		{
			return String.Format("<b><span foreground=\"{0}\">{1}</span></b>", color, etiqueta);
		}
		#endregion
	}
}
