// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
//
// This file is part of Stratos.
//
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;
using System.Threading;
using System.Linq;

using Gtk;
using Mono.Addins;

using StratosClient.Core.Gui.Configuracion;

namespace StratosClient.Core.Gui
{
	public partial class MainWindow : Gtk.Window
	{
		private static MainWindow instancia;
		
		private bool mensajeConexionFallidaActivado = false;
		private bool mensajeDesconexionFallidaActivado = false;
		
		public MainWindow (): base (Gtk.WindowType.Toplevel)
		{
			Build ();

			instancia = this;
			
			// Me suscribo a eventos de conexión y desconexión
			MainIce.ConexionEstablecida += this.ConexionRealizada;
			MainIce.DesconexionRealizada += this.DesconexionRealizada;
			this.MensajeConexionFallida = true;
			this.MensajeDesconexionFallida = true;
			
			/* Al iniciar la ventana principal, es posible que ya estemos conectados
			 * al servidor. Por lo tanto realizo la comprobación aquí. */
			if (MainIce.ConectadoAlServidor)
				this.ConexionRealizada();
			else
				this.DesconexionRealizada();
			
			this.ConstruirToolbar();
			this.ConstruirMenu();
			this.ConstruirSolapas();

			// FIXME: Si se necesita más tamaño vertical, así se puede quitar la
			// toolbar
			//this.vbox1.Remove(this.barraHerramientas);
			
			if (VentanaPrincipalElement.Instancia.IniciarMaximizada)
				this.Maximize();
		}

		#region Propiedades
		public static MainWindow Instancia	{
			get { return instancia; }
		}
		
		public bool MensajeConexionFallida {
			get { return this.mensajeConexionFallidaActivado; }
			set {
				// Si el nuevo valor es igual al actual, no hago nada.
				if (this.mensajeConexionFallidaActivado == value)
					return;
				
				/* Si el nuevo valor es true, me suscribo. En caso contrario
				 * me desuscribo del evento */
				if (value)
					MainIce.ConexionFallida += this.ConexionFallida;
				else
					MainIce.ConexionFallida -= this.ConexionFallida;
				
				this.mensajeConexionFallidaActivado = value;
			}
		}
		
		public bool MensajeDesconexionFallida {
			get { return this.mensajeDesconexionFallidaActivado; }
			set {
				// Si el nuevo valor es igual al actual, no hago nada.
				if (this.mensajeDesconexionFallidaActivado == value)
					return;
				
				/* Si el nuevo valor es true, me suscribo. En caso contrario
				 * me desuscribo del evento */
				if (value)
					MainIce.DesconexionFallida += this.DesconexionFallida;
				else
					MainIce.DesconexionFallida -= this.DesconexionFallida;
					
				this.mensajeDesconexionFallidaActivado = value;
			}
		}
		#endregion
	
		#region Event handlers
		protected void OnDeleteEvent (object sender, DeleteEventArgs a)
		{
			this.OnSalirClicked(null, null);
			a.RetVal = true;
		}
	
		protected void OnExtensionChanged (object o, ExtensionEventArgs args)
		{
			// FIXME: Revisar por otros cambios, o directamente si esto se usa
			if (args.PathChanged ("/StratosClient/Core/MainToolbar"))
				this.ConstruirToolbar ();
			else if (args.PathChanged ("/StratosClient/Core/MainMenu"))
				this.ConstruirMenu ();
		}
		
		protected virtual void OnSalirClicked (object sender, System.EventArgs e)
		{
			bool evitarSalida = false;
			
			/* Al salir de la aplicación realizo unas operaciones de salida antes (como 
			 * desconectarme y guardar la configuración), pero si surgen errores, salgo
			 * de todas formas. */
			try {
				/* Primero me fijo si algún addin no requiere que se realicen ciertas
				 * acciones antes de salir, por ejemplo, un documento modificado que no ha
				 * sido guardado aún. Este tipo de acción puede prevenir que el cliente
				 * se cierre para que el usuario pueda terminar la tareas pendientes. */
				StringBuilder tareasPendientes = new StringBuilder();
				
				foreach (IVerificador verificador in
				         	AddinManager.GetExtensionObjects ("/StratosClient/Core/VerificarElementosSinGuardarAlSalir")) {
					if (!verificador.HayTareaPendiente)
						continue;
					
//					tareasPendientes.AppendFormat("<b>{0}</b>:\n", verificador.Seccion);
					
					foreach (string msj in verificador.Mensajes)
						tareasPendientes.AppendFormat("  - {0}\n", msj);
					
//					tareasPendientes.Append("\n");
				}
				
				// Si hay tareas pendientes, muestro el mensaje
				if (tareasPendientes.Length > 0) {
					string mensaje =
						String.Format("<b>Hay elementos sin guardar</b>:\n" +
							"{0}\n" +
							"<b>¿Está seguro que desea salir?</b>", tareasPendientes.ToString());
					
					ResponseType respuesta = SMessageDialog.MensajePregunta(mensaje);
					
					if (respuesta != ResponseType.Yes) {
						evitarSalida = true;
						
						return;
					}
				}
				
				this.MensajeDesconexionFallida = false;
				MainIce.DesconectarServidor();
				
				// Guardo la configuración
				this.OnGuardarConfiguracionActivated(null, null);
			}
			catch (System.Exception ex) {
				SMessageDialog.MensajeError(ex);
			}
			finally {
				if (!evitarSalida)
					Gtk.Application.Quit();
			}
		}

		protected virtual void OnConectarClicked (object sender, System.EventArgs e)
		{
			MainIce.ConectarServidor();
		}
		
		protected virtual void OnConectarAvanzadoClicked (object sender, System.EventArgs e)
		{
			/* Muestro el diálogo de conexión avanzada que permite ingresar la dirección
			 * del servidor. */
			ConexionAvanzada dialogoConexion = new ConexionAvanzada(this);
			ResponseType respuesta = (ResponseType) dialogoConexion.Run();
			
			if (respuesta == ResponseType.Ok) {
				/* Inicio un thread aparte que realice la conexión, para no detener al
				 * thread de Gtk+ (que no se congele la pantalla). */
				Thread threadConexion = new Thread(delegate() {
					try {
						MainIce.ConectarServidor(dialogoConexion.DireccionServidor);
					}
					catch (System.Exception ex) {
						GLib.ExceptionManager.RaiseUnhandledException(ex, false);
					}
				});
				threadConexion.Start();
				
				// Desactivo el diálogo
				dialogoConexion.Sensitive = false;
				
				/* Mientras el thread anterior esté corriendo, corro una iteración del
				 * thread de Gtk para mantener la interfaz gráfica sensible. */
				while (threadConexion.ThreadState == ThreadState.Running)
					Gtk.Application.RunIteration();
			}
			
			dialogoConexion.Destroy();
		}
		
		protected virtual void OnDesconectarClicked (object sender, System.EventArgs e)
		{
			MainIce.DesconectarServidor();
		}
		
		protected virtual void OnAcercaDeClicked (object sender, System.EventArgs e)
		{
			AcercaDeDialog ad = new AcercaDeDialog();
			ad.Run();
			ad.Destroy();
		}
		
		protected virtual void OnGestorComplementosClicked (object sender, System.EventArgs e)
		{
			Mono.Addins.Gui.AddinManagerWindow.Run(MainWindow.Instancia);
		}
		#endregion
	
		#region Métodos útiles
		protected void ConstruirToolbar ()
		{
			// FIXME: Hay que revisar este método.
			
			// Limpiar la barra de herramientas (toolbar)
//			foreach (Gtk.Widget w in barraHerramientas.Children)
//				barraHerramientas.Remove(w);
			
			// Agrego los botones de los distintos addins en la barra de herramientas
			foreach (WidgetNode node in AddinManager.GetExtensionNodes ("/StratosClient/Core/MainToolbar"))
				barraHerramientas.Insert ((Gtk.ToolItem)node.GetWidget(), -1);
			
			barraHerramientas.ShowAll ();
		}
	
		protected void ConstruirMenu ()
		{
			// Busco los add-ins suscriptos a MainMenu
			var nodosMenuPrincipal =
				from nodo in AddinManager.GetExtensionObjects("/StratosClient/Core/MainMenu").Cast<IMenuPrincipal>()
				select nodo;
			
			foreach (IMenuPrincipal menu in nodosMenuPrincipal)
				this.barraMenu.Insert(menu.Menu, menu.Orden);
			
			// Obtengo el menú de herramientas
			MenuItem menuHerramientas =
				(from menuItem in this.barraMenu.AllChildren.Cast<MenuItem>()
				where menuItem.Name == "Herramientas"
				select menuItem).First();
			
			var nodosMenuHerramientas =
				from nodo in AddinManager.GetExtensionNodes("/StratosClient/Core/MenuHerramientas").Cast<MenuBaseNode>()
				orderby nodo.etiqueta
				select nodo;
			
//			ExtensionNodeList nodosMenuHerramientas =
//				AddinManager.GetExtensionNodes("/StratosClient/Core/MenuHerramientas");
			
			// Me fijo si hay nodos registrados antes de continuar con el menú Herramientas
			if (nodosMenuHerramientas.Count() > 0) {
				MenuShell subMenu = (MenuShell) menuHerramientas.Submenu;
				
				// Agrego un separador y luego los subitems registrados
				subMenu.Insert(new Gtk.SeparatorMenuItem(), 0);
				
				foreach (MenuItemNode node in nodosMenuHerramientas) {
					subMenu.Insert(node.GetWidget(), 0);
				}
			}
			
			barraMenu.ShowAll ();
		}

		protected void ConstruirSolapas()
		{
			foreach(Gtk.Widget w in solapas.AllChildren)
				solapas.Remove(w);
			
			// Ordeno las solapas según la posición establecida
			var listaOrdenada =
				from tabNode in AddinManager.GetExtensionNodes("/StratosClient/Core/MainTabs").Cast<TabNode>()
				orderby tabNode.Posicion ascending
				select tabNode;
			
			foreach (TabNode node in listaOrdenada) {
				// Obtengo el widget del nodo
				Gtk.Bin widget = null;
				try {
					widget = node.GetWidget();
				}
				catch (System.Exception ex) {
					SMessageDialog
						.MensajeError(String.Format("No se pudo cargar la solapa '{0}'", node.EtiquetaCadena),
						              ex);
					
					continue;
				}
				
				widget.BorderWidth = ((uint)5);
				
				// Creo un ScrolledWindow
				ScrolledWindow scrolledWindow = new ScrolledWindow();
				scrolledWindow.HeightRequest = 450;
				scrolledWindow.WidthRequest = 700;
				scrolledWindow.ShadowType = ShadowType.None;
				scrolledWindow.SetPolicy(PolicyType.Automatic,
				                         PolicyType.Automatic);
				
				scrolledWindow.AddWithViewport(widget);
				
				solapas.Add(scrolledWindow);
				solapas.SetTabLabel(scrolledWindow, node.Etiqueta);
			}
			
			solapas.ShowAll();
		}
		
		/// <summary>
		/// Se ejecuta al realizarse la conexión. Este metodo se suscribe al evento de conexión
		/// realizada, y cuando ésta sucede, se encarga de activar los widgets en la ventana
		/// principal.
		/// </summary>
		protected void ConexionRealizada()
		{
			// Realizo los cambios correspondientes cuando se ha podido conectar con el servidor
			Gtk.Application.Invoke(delegate {
				this.btnConectar.Sensitive = false;
				this.mnuConectar.Sensitive = false;
				this.mnuConectarAvanzado.Sensitive = false;
				this.btnDesconectar.Sensitive = true;
				this.mnuDesconectar.Sensitive = true;
				this.solapas.Sensitive = true;
				this.GuardarConfiguracinAction.Sensitive = true;
			});
		}
		
		/// <summary>
		/// Este método se suscribe a un evento que se dispara cuando la conexión al servidor
		/// falla.
		/// </summary>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		protected void ConexionFallida(System.Exception excepcion)
		{
			string mensaje = "Error al conectar con el servidor. " +
				"Verifique que haya conectividad y que la PC Servidor esté encendida.";
			
			Gtk.Application.Invoke(delegate {
				SMessageDialog.MensajeError(this, mensaje, excepcion);
			});
		}
		
		/// <summary>
		/// Similar a ConexiónRealizada. Se encarga de desactivar los widgets de la ventana
		/// principal cuando se realiza la desconexión del servidor.
		/// </summary>
		protected void DesconexionRealizada()
		{
			Gtk.Application.Invoke(delegate {
				this.btnConectar.Sensitive = true;
				this.mnuConectar.Sensitive = true;
				this.mnuConectarAvanzado.Sensitive = true;
				this.btnDesconectar.Sensitive = false;
				this.mnuDesconectar.Sensitive = false;
				this.solapas.Sensitive = false;
				this.GuardarConfiguracinAction.Sensitive = false;
			});
		}
		
		/// <summary>
		/// Similar a ConexionFallida. Muestra un mensaje con un diálogo cuando la desconexión
		/// del servidor falla. El único lugar donde esto es útil es cuando el usuario activamente
		/// solicita la desconexión.
		/// </summary>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		protected void DesconexionFallida(System.Exception excepcion)
		{
			this.DesconexionRealizada();
			
			string mensaje = "Error al desconectar del servidor. " +
				"La terminal se encuentra ahora desconectada.";
			
			Gtk.Application.Invoke(delegate {
				SMessageDialog.MensajeError(this, mensaje, excepcion);
			});
		}

		#endregion

		protected virtual void OnGuardarConfiguracionActivated (object sender, System.EventArgs e)
		{
			StratosClientConfig.Save();
			
			// Actualizo la configuración en el servidor
			MainIce.ActualizarDatosTerminal();
		}
	}
}
