// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
//
// This file is part of Stratos.
//
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Threading;

using Mono.Addins;
using log4net;

using Stratos.Common;

namespace StratosClient.Core.Gui
{
	public class MainClient
	{
		/* El código "System.Reflection.MethodBase..." es igual que typeof(MainClient),
		 * pero es más portable, permitiendo copiar y pegar este código en otras clases
		 * sin cambiar nada
		 */
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);
		
		private static string[] argumentosIniciales;
		private static Thread threadGui;
		private static Thread threadIce;
		private static AutoResetEvent autoReset;
		
		public static void Iniciar(string[] args)
		{
			// Configuro el logger
			log4net.Config.XmlConfigurator.Configure();
			
			argumentosIniciales = args;
			
			/* Este objeto es para sincronizar threads diferentes. Lo utilizo para saber
			 * cuándo es posible continuar avanzando en este thread. */
			autoReset = new AutoResetEvent(false);
			
			// Muestro el Splash Screen
			Thread threadGui = new Thread(delegate() {
				MainGui.Instancia.IniciarSplashScreen(autoReset, ref args);
			});
			threadGui.SetApartmentState(ApartmentState.STA);
			threadGui.Start();
			
			// Espero a que el splash se muestre y Gtk se inicialice
			autoReset.WaitOne();
			
			// Inicializo la base de addins y busco por cambios en los mismos
			Log("Inicializando gestor de complementos");
			AddinManager.Initialize("addins_db");
			
			logger.Debug("Buscando addins nuevos o modificados");
			AddinManager.Registry.Update(null);
			
			// Me suscribo a los eventos de carga, descarga y error de los addins
			AddinManager.AddinLoadError += OnLoadError;
			AddinManager.AddinLoaded += OnLoad;
			AddinManager.AddinUnloaded += OnUnload;
			AddinManager.ExtensionChanged += OnExtensionChange;
			
			// Configuración
			logger.Debug("Iniciando sistema de configuración");
			Configuracion.StratosClientConfig.Inicializar();
			
			// Inicio el thread de Ice
			Log("Iniciando gestor de comunicaciones por red");
			threadIce = new Thread(delegate() {
				MainIce.Iniciar(autoReset, argumentosIniciales);
			});
			threadIce.Start();
			
			// Espero a que el thread de comunicaciones esté inicializado
			autoReset.WaitOne();
			
			// Ahora que todo está iniciado, ejecuto los comando suscriptos en
			// el inicio estandar del cliente
			Log("Ejecutando objetos suscriptos al inicio estandar");
			foreach (object suscripto in AddinManager.GetExtensionObjects("/StratosClient/Core/InicioEstandar")) {
				IComandoInicio comandoInicio = suscripto as IComandoInicio;
				
				// Verifico que implemente la interfaz IComandoInicio
				if (comandoInicio == null) {
					logger.ErrorFormat("El comando de inicio no implementa la interfaz '{0}': {1}",
					                   typeof(IComandoInicio).FullName, suscripto.ToString());
					continue;
				}
				
				try {
					comandoInicio.Ejecutar(args);
					logger.DebugFormat("Comando de inicio ejecutado correctamente: {0}", comandoInicio.ToString());
				}
				catch (System.Exception excepcion) {
					logger.ErrorFormat("Error al ejecutar el comando de inicio: {0}", comandoInicio.ToString());
				}
			}
			
			/* Ejecuto el código que muestra la ventana principal en un thread aparte, pero antes
			 * espero un segundo para que se pueda apreciar la belleza del Splash Screen :) */
			Log("Iniciando ventana principal");
			Thread.Sleep(1000);
			
			MainGui.Instancia.Iniciar(args);
			
			// Cierro el Splash Screen
			SplashScreen.Instancia.Cerrar();
			
			/* Espero a que finalice el thread de Gtk. En este punto este thread
			 * principal espera a que finalice el thread de la parte gráfica. */
			logger.Debug("Esperando a que finalice el thread de Gtk (GUI)");
			threadGui.Join();
			
			/* Interrumpo el thread de Ice (que se encarga de todas las comunicaciones)
			 * Luego salgo de la aplicación. */
			logger.Debug("Interrumpiendo el thread de Ice (comunicación por red)");
			threadIce.Interrupt();
			threadIce.Join();
		}
		
		#region Otros métodos
		/// <summary>
		/// Escribe un mensaje de log en el logger y en el SplashScreen.
		/// </summary>
		/// <param name="msj">
		/// A <see cref="System.String"/>
		/// </param>
		private static void Log(string msj)
		{
			logger.Debug(msj);
			SplashScreen.Instancia.ActualizarEstado(msj);
		}
		#endregion
		
		#region Eventos de Mono.Addins
		static void OnLoadError (object s, AddinErrorEventArgs args)
		{
			logger.Error("Error al cargar un add-in: (" + args.AddinId + ") " + args.Message);
		}

		static void OnLoad (object s, AddinEventArgs args)
		{
			logger.Debug("Add-in cargado: " + args.AddinId);
		}
		
		static void OnUnload (object s, AddinEventArgs args)
		{
			logger.Debug("Add-in descargado: " + args.AddinId);
		}

		static void OnExtensionChange (object s, ExtensionEventArgs args)
		{
			logger.Debug("Add-in cambiado: " + args.Path);
		}
		#endregion
	}
}
