// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
//
// This file is part of Stratos.
//
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Threading;

using Gtk;
using log4net;
using Mono.Addins;

using Stratos.Common;

namespace StratosClient.Core.Gui
{
	public class MainGui
	{
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);

		private static MainGui mainGui;
		private static Gtk.Window mainWindow;
		
		private static bool dialogoErrorMostrado;
		
		public void IniciarSplashScreen(AutoResetEvent autoReset, ref string[] args)
		{
			logger.Debug("Ejecutando Gtk.Application.Init()");
			Application.Init("StratosClient", ref args);
			
			logger.Debug("Creando SplashScreen");
			SplashScreen splashScreen = new SplashScreen();
			splashScreen.ActualizarEstado("Inicializando Stratos");
			splashScreen.ShowAll();
			
			autoReset.Set();
			
			/* Me suscribo a los eventos para capturar las excepciones no capturadas
			 * en la aplicación. Uno es para el thread de Gtk+ y el otro para el
			 * dominio de la aplicación. En el primero se puede evitar que la aplicación
			 * se cierre, pero no en el segundo (lo uso para hacer logging antes del
			 * cierre). */
			dialogoErrorMostrado = false;
			GLib.ExceptionManager.UnhandledException += GLibUnhandledException;
			AppDomain.CurrentDomain.UnhandledException += CurrentDomainUnhandledException;
			
			Application.Run();
		}
		
		static MainGui() {
			
		}
		
		private void CargarEstilos()
		{
			string archivoEstilos =
				StratosClientUtils.DirectorioAplicacion() +
					System.IO.Path.DirectorySeparatorChar +
					"gtkrc.stratos";
			
			if (System.IO.File.Exists(archivoEstilos))
			{
				string estilos = System.IO.File.ReadAllText(archivoEstilos);
				Gtk.Rc.ParseString(estilos);
				logger.Debug("Estilos personalizados cargados");
			}
		}
		
		public void Iniciar(string[] args)
		{
			CargarEstilos();
			
			Gtk.Application.Invoke(delegate {
				mainWindow = new MainWindow ();
				mainWindow.ShowAll();
			});
			
			logger.Debug("Ventana principal creada");
		}

		public static MainGui Instancia {
			get {
				if (mainGui == null)
					mainGui = new MainGui();
				
				return mainGui;
			}
		}

		public static Gtk.Window MainWindow {
			get { return mainWindow; }
		}
		
		#region Gestión de excepciones no capturadas (unhandled exceptions)
		/// <summary>
		/// Este método se ejecuta en el cliente cuando una excepción no ha sido capturada.
		/// Se deben colocar aquí todas las excepciones comunes y su tratamiento global.
		/// </summary>
		/// <param name="args">
		/// A <see cref="GLib.UnhandledExceptionArgs"/>
		/// </param>
		private static void GLibUnhandledException(GLib.UnhandledExceptionArgs args)
		{
			Exception exception = (Exception) args.ExceptionObject;
			
			logger.Error("Excepción no capturada en el thread de Gtk: " +
			             exception.ToString() + ": " + exception.Message + "\n" +
			             exception.StackTrace);
			
			GestionarException(exception, false);
		}
		
		private static void CurrentDomainUnhandledException(object sender, UnhandledExceptionEventArgs args)
		{
			Exception exception = (Exception) args.ExceptionObject;
			
			logger.Error("Excepción no capturada en el dominio de la aplicación: " +
			             exception.ToString() + ": " + exception.Message + "\n" +
			             exception.StackTrace);
			
			GestionarException(exception, true);
		}
		
		public static void GestionarException(System.Exception exception)
		{
			GestionarException(exception, false);
		}
		
		private static void GestionarException(System.Exception exception, bool appDomainUnhandledException)
		{
			bool dialogoCerrado = false;
			
			Gtk.Application.Invoke(delegate {
				// Si ya hay un diálogo de error mostrándose, no muestro otro más.
				if (dialogoErrorMostrado)
					return;
				
				dialogoErrorMostrado = true;

				string mensajeDialogo =
					"Hubo un error desconocido en la aplicación.";
				
				/* Si el tipo de excepción es TargetInvocationException, entonces me interesa
				 * la excepción interna. */
				if (exception is System.Reflection.TargetInvocationException)
					exception = exception.InnerException;
				
				// De acuerdo al tipo de excepción armo el mensaje a mostrar en el diálogo.
				// Agregar aquí otras excepciones comunes.
				if (exception is Ice.ConnectionRefusedException) {
					/* Para este caso, desconectamos el cliente, ya que el Servidor ha rechazado
					 * la operación. Seguramente se encuentra caído y el cliente deberá
					 * reconectarse. Antes de desconectar, desactivo los mensajes de diálogo de
					 * desconexión fallida (ya que la desconexión seguro va a fallar y no quiero
					 * mostrar otro diálogo al usuario). */
					StratosClient.Core.Gui.MainWindow.Instancia.MensajeDesconexionFallida = false;
					
					MainIce.DesconectarServidor();
				}

				/* Si se trata de una excepción crítica, interpretamos el mensaje de la misma como
				 * un mensaje para mostrar al usuario, y la InnerException como la excepción a
				 * mostrar en la sección avanzada. */
				if (exception is ErrorCriticoException) {
					mensajeDialogo = exception.Message;
					exception = exception.InnerException;
				}
				
				/* Creo el diálogo SMessageDialog y lo muestro. Si se trata de una excepción
				 * no gestionada del dominio de la aplicación, eso significa que la misma no
				 * podrá seguir ejecutándonse. En ese caso, le aviso al usuario que Stratos se
				 * cerrara. */
				if (appDomainUnhandledException) {
					SMessageDialog.MensajeError(mensajeDialogo + "\nLa aplicación de cerrará.",
					                            exception);
					
					Gtk.Application.Quit();
				}
				else
					SMessageDialog.MensajeError(exception);
				
				dialogoCerrado = true;
				
				dialogoErrorMostrado = false;
			});
			
			/* Si se trata de una excepción no capturada en el dominio de la aplicación, entonces
			 * espero a que el usuario cierre el diálogo. Debo hacer esto ya que en este punto
			 * estamos en un thread distinto al de Gtk+. */
			while (appDomainUnhandledException && !dialogoCerrado)
				Thread.Sleep(TimeSpan.FromSeconds(1));
		}
		#endregion
	}
}
