// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Threading;

using Ice;
using Mono.Addins;
using log4net;

using Stratos.Common;
using Stratos.Common.Excepciones;
using StratosClient.Core.Gui.Configuracion;
using Stratos.Servicios;
using StratosServer.Controladores;

namespace StratosClient.Core.Gui
{
	public class MainIce : Ice.Application
	{
		#region Campos
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);
		
		#region Eventos
		public static event System.Action ConexionEstablecida;
		public static event System.Action DesconexionRealizada;
		
		public static event System.Action<System.Exception> ConexionFallida;
		public static event System.Action<System.Exception> DesconexionFallida;
		#endregion
		
		private static bool conectadoAlServidor;

		private static MainIce mainIce;
		private static Ice.ObjectAdapter adapter;
		private static AutoResetEvent autoReset;
		private static Dictionary<string, Ice.ObjectPrx> controladores;
		
		private static Dictionary<string, Ice.ObjectPrx> proxiesServiciosServidores;
		
		private static DatosTerminal datosTerminal;
		
		private static ControladorTerminalesPrx controladorTerminales;
		#endregion
		
		#region Constructores
		static MainIce()
		{
			proxiesServiciosServidores = new Dictionary<string, Ice.ObjectPrx>();
			controladores = new Dictionary<string, Ice.ObjectPrx>();
			conectadoAlServidor = false;
			
			// Creo la estructura de datos para enviar al servidor
			datosTerminal = new DatosTerminal {
				identificador = ConexionElement.Instancia.NombreTerminal
			};
		}
		#endregion
		
		#region Propiedades
		public static Dictionary<string, Ice.ObjectPrx> ProxiesServiciosServidores {
			get { return proxiesServiciosServidores; }
		}
		
		public static ControladorTerminalesPrx ControladorTerminales {
			get { return controladorTerminales; }
		}
		
		public static bool ConectadoAlServidor {
			get { return conectadoAlServidor; }
		}
		
		public static DatosTerminal DatosTerminal {
			get { return datosTerminal; }
//			set {
//				datosTerminal = value;
//				
//				// Si estoy conectado, notifico al servidor que han cambiado los datos
//				// de la terminal
//				//ActualizarDatosTerminal();
//			}
		}
		
		public static Ice.ObjectAdapter IceAdapter {
			get {
				if (adapter == null) {
					int puertoLocal = ConexionElement.Instancia.PuertoLocal;
					
					adapter = communicator()
						.createObjectAdapterWithEndpoints("StratosClient",
						                                  "tcp -p " + puertoLocal.ToString());
				}
				
				return adapter;
			}
		}
		#endregion
		
		#region Métodos principales
		public static void Iniciar(AutoResetEvent autoResetEvent, string[] args)
		{
			logger.Debug("MainIce (Iniciar) ejecutado");
			
			autoReset = autoResetEvent;
			
			MainIce mainIce = new MainIce();
			mainIce.main(args);
		}
		
		private int RunInterno(string[] args)
		{
			logger.Debug("Ejecutando RunInterno");
			
			// Handler para bajar el communicator
			logger.Debug("Configurando handler: shutdownOnInterrupt");
			shutdownOnInterrupt();
			
			/* Accedo a la propiedad IceAdapter para que se cree el objeto Ice.ObjectAdapter. */
			try {
				object objTemp = MainIce.IceAdapter;
			}
			catch (Ice.SocketException ex) {
				logger.Error("Excepción al crear el objeto Ice.ObjectAdapter en la Terminal. Quizá la " +
				             "dirección está en uso a causa de otra instancia corriendo en la misma máquina.\n" +
				             "Detalles: " + ex.Message + "\n" +
				             "Stackstrace: " + ex.StackTrace);
				
				Gtk.Application.Invoke(delegate {
					SMessageDialog.MensajeError("La dirección utilizada por el cliente está en uso. " +
					                            "¿Hay otra instancia de la terminal corriendo? " +
					                            "Cierre la otra instancia e inténtelo nuevamente.",
					                            ex);
					
					// Salgo de la aplicación
					Gtk.Application.Quit();
				});
				
				// Le indico al thread principal que puede continuar ahora
				autoReset.Set();
				
				return 1;
			}
			catch (System.Exception ex) {
				logger.Error("Excepción al crear el objeto Ice.ObjectAdapter en la Terminal.\n" +
				             "Detalles: " + ex.Message + "\n" +
				             "Stackstrace: " + ex.StackTrace);
				
				Gtk.Application.Invoke(delegate {
					SMessageDialog.MensajeError("Hubo un error al configurar la conexión de red en la terminal.",
					                            ex);
					
					// Salgo de la aplicación
					Gtk.Application.Quit();
				});
				
				// Le indico al thread principal que puede continuar ahora
				autoReset.Set();
				
				return 1;
			}
			
			/* FIXME: Este código puede no funcionar correctamente por ahora. Hay que hacer un
			 * refactor de Stratos.Servicios para agregar otro diseño de comunicación entre el
			 * Servidor y la Terminal */
			
			/* Ejecuto todos los addins que van a brindar un objeto que actuará de servidor. */
			logger.Debug("Cargando todos los addins en modo Servicio Servidor");
			foreach (IServicioServidor servicioServidor in 
			         	AddinManager
			         		.GetExtensionObjects ("/StratosClient/Core/ServicioServidor", true)) {
				
				// Verifico que el tipo del servant sea Ice.ObjectImpl
				if (!(servicioServidor.Servant is Ice.ObjectImpl)) {
					logger.Error("El servant no es de tipo Ice.ObjectImpl: " +
					             servicioServidor.IceIdentity);
					continue;
				}
				
				// Obtengo el Servant
				logger.Debug("Creando el servant: " + servicioServidor.IceIdentity);
				Ice.Object obj = (Ice.Object)servicioServidor.Servant;
				
				logger.Debug("Agregando objeto al Ice.Adapter");
				proxiesServiciosServidores.Add(servicioServidor.IceIdentity,
				                               MainIce.IceAdapter.addWithUUID(obj));
			}
			
			// Cargo los addins que se registraron para modificar la estrutura DatosTerminal
			foreach(ICargaDatosTerminal cdt in
			        AddinManager.GetExtensionObjects("/StratosClient/Core/CargaDatosTerminal"))
			        cdt.CargarDatosTerminal();
			
			// Intento conectarme al servidor
			MainIce.ConectarServidor();
			
			// Activo el Ice.ObjectAdapter
			logger.Debug("Activando ObjectAdapter");
			IceAdapter.activate();
			
			// Le indico al thread principal que puede continuar ahora
			autoReset.Set();
			
			// Espero hasta que el communicator sea bajado. Esto hace
			// que quede esperando las conexiones
			try {
				logger.Debug("Componente Ice en espera");
				communicator().waitForShutdown();
			}
			catch (ThreadInterruptedException ex) {
				logger.Debug("Componente Ice apagado");
			}
			
			return 0;
		}
		
		public override int run (string[] args)
		{
			try {
				return this.RunInterno(args);
			}
			catch(System.Exception exception) {
				logger.ErrorFormat("Excepción al inicializar Ice: {0}", exception.ToString());
				
				Gtk.Application.Invoke(delegate {
					SMessageDialog
						.MensajeError("Hubo un error al conectar con el servidor. Posiblemente " +
						              "se trate de un problema de conectividad.\n" +
						              "La aplicación ahora se cerrará.",
						              exception);
					
					// Salgo de la aplicación
					Gtk.Application.Quit();
				});
				
				// Le indico al thread principal que puede continuar ahora
				autoReset.Set();
				
				return 1;
			}
		}
		#endregion
		
		#region Otros métodos
		public static Ice.ObjectPrx GetControlador(string idControlador)
		{
			if (controladores.ContainsKey(idControlador))
				return controladores[idControlador];
			else
				return null;
		}
		#endregion
		
		#region Métodos de Conexión
		private static ControladorTerminalesPrx
			ObtenerControladorTerminales(string direccionServidor)
		{
			// Obtengo el proxy al ControladorTerminales
			Ice.ObjectPrx proxyObj =
				MainIce.communicator()
					.stringToProxy("ControladorTerminales:tcp -h " +
					               direccionServidor +
					               " -p " + ConexionElement.Instancia.PuertoServidor +
					               " -t " + ConexionElement.Instancia.TimeoutControladorTerminales);
			
			controladorTerminales =
				ControladorTerminalesPrxHelper.checkedCast(proxyObj);
			
			return controladorTerminales;
		}
		
		/// <summary>
		/// Permite conectarse al servidor según la dirección especificad en el archivo
		/// de configuración
		/// </summary>
		public static void ConectarServidor()
		{
			ConectarServidor(ConexionElement.Instancia.DireccionServidor);
		}
		
		/// <summary>
		/// Permite conectarse al servidor con una dirección especificada
		/// </summary>
		/// <param name="direccionServidor">
		/// A <see cref="System.String"/>
		/// </param>
		public static void ConectarServidor(string direccionServidor)
		{
			ControladorTerminalesPrx controladorTerminalesproxy = null;
			try {
				/* FIXME: otras excepciones que puede tirar checkedCast son
				 * ConnectionTimeOut y ObjectNotExist. Revisar.
				 * FIXME: Ver también los timeouts y configurarlos para que sean
				 * coherentes en una red rápida. */
				logger.Debug("Intentando conectar con el servidor");
				
				logger.Debug("Intentando obtener el controlador de terminales");
				controladorTerminalesproxy = ObtenerControladorTerminales(direccionServidor);
			}
			catch (Ice.ConnectionRefusedException ex) {
				logger.Error("Conexión rechazada al intentar conectar con " +
				             "el Controlador de Terminales:\n" + ex.ToString());
				
				if (ConexionFallida != null)
						ConexionFallida(ex);
				
				return;
			}
			catch (Ice.DNSException ex) {
				logger.Debug("Error de DNS de Ice:\n" + ex.ToString());
				
				if (ConexionFallida != null)
					ConexionFallida(ex);
				
				return;
			}
			catch (System.Exception ex) {
				logger.Error("Error desconocido al intentar conectar con el " +
				             "Controlador de Terminales:\n" +
				             ex.ToString());
				
				if (ConexionFallida != null)
						ConexionFallida(ex);
				
				return;
			}
			
			// Verifico si el proxy para el Controlador de Terminales es válido
		    if (controladorTerminalesproxy == null) {
		        logger.Error("El proxy para el Controlador de Terminales es inválido");
				
				if (ConexionFallida != null)
						ConexionFallida(new System.Exception("El proxy para el Controlador " +
					                                     "de Terminales es inválido"));
				
				return;
			}
			
			// Me conecto al servidor
			// FIXME: podría arrojar excepciones
			try {
				logger.Debug("Conectando con el servidor");
				controladorTerminalesproxy.Conectar(datosTerminal);
			}
			catch (ConexionException ex) {
				logger.Error("Error al intentar conectar con el servidor: " + ex.mensaje);
				
				logger.Debug("Reintentando conexión al servidor");
				
				// FIXME: Podría haber un tipo de excepción especial para esto
				// Intento desconectarme del servidor (el error puede deberse a ello) y
				// reconectarme.
				
				try {
					/* Me desconecto del servidor que ha rechazado mi conexión */
					MainIce.DesconectarServidor(direccionServidor);
					
					controladorTerminalesproxy.Conectar(datosTerminal);
					
					logger.Debug("Reconexión exitosa.");
				}
				catch(System.Exception ex2) {
					logger.Debug("La reconexión ha fallado.");
					
					// Si vuelve a ocurrir una excepción, salgo del método
					if (ConexionFallida != null)
						ConexionFallida(ex2);
				
					return;
				}
			}
			catch (System.Exception ex) {
				logger.Error("Error desconocido al conectar con el servidor");
				
				if (ConexionFallida != null)
					ConexionFallida(ex);
				
				return;
			}
			
			ConexionElement.Instancia.DireccionServidor = direccionServidor;
			
			/* Ejecuto todos los addins disponibles que ofrecen información para crear un
			 * proxy, es decir todos aquellos que brindan información sobre cómo acceder remotamente
			 * a un controlador. */
			/* FIXME: En realidad, cada addin no debería proporcionar información sobre cómo obtener
			 * cada proxy, sino que el cliente debería tener la información mínima para acceder a un
			 * objeto gestor en el servidor, y obtener los proxies llamando a operaciones que los
			 * devuelvan. Ver el manual de Ice en la página 500 "Creating Proxies" */
			/* FIXME: Se debe verificar que los objetos realmente sean del tipo IServicioCliente.
			 * Sino pueden surgir excepciones de casting. */
			logger.Debug("Cargando todos los addins en modo Servicio Cliente");
			
			controladores.Clear();
			
			foreach (IServicioCliente servicioCliente in 
			         	AddinManager
			         		.GetExtensionObjects ("/StratosClient/Core/ServicioCliente", true)) {
				
				// Verifico que el servicio ya no esté asignado
				if (controladores.ContainsKey(servicioCliente.IceIdentity)) {
					logger.Error("El servicio cliente ya existe: " +
					             servicioCliente.IceIdentity);
					continue;
				}
				
				// Creo el proxy y lo valido
				Ice.ObjectPrx tempProxyObj =
					controladorTerminalesproxy.ObtenerProxyControlador(servicioCliente.IceIdentity);
				
				logger.Debug("Validando proxy contra el servidor: " +
				             servicioCliente.IceIdentity);
				
				/* Hago un checkedCast del proxy. Esto implica una conexión con el servidor
				 * y a este método lo brinda el addin. */
				Ice.ObjectPrx controladorProxy = null;
				try {
					/* FIXME: otras excepciones que puede tirar checkedCast son
					 * ConnectionTimeOut y ObjectNotExist. Revisar.
					 * FIXME: Ver también los timeouts y configurarlos para que sean
					 * coherentes en una red rápida. */
			    	controladorProxy =
						(servicioCliente.checkedCast(tempProxyObj) as Ice.ObjectPrx);
				}
				catch (Ice.ConnectionRefusedException ex) {
					logger.Error("Conexión rechazada al validar proxy: " +
					             servicioCliente.IceIdentity);
					
					continue;
				}
				
			    if (controladorProxy == null) {
			        logger.Error("El proxy para el Controlador de Terminales es inválido: " + 
					             servicioCliente.IceIdentity);
					
					continue;
				}
				
				// Agrego el proxy al diccionario
				controladores[servicioCliente.IceIdentity] = controladorProxy;
			}
			
			logger.Debug("Conexión establecida correctamente");
			
			conectadoAlServidor = true;
			
			// Disparo el evento de ConexionEstablecida
			if (ConexionEstablecida != null)
				ConexionEstablecida();
		}
		
		/// <summary>
		/// Actualiza los datos de la terminal en el servidor.
		/// </summary>
		public static void ActualizarDatosTerminal()
		{
			if (!conectadoAlServidor)
				return;
			
			// Envio al servidor los nuevos datos de la terminal
			// FIXME: Podría arrojar excepciones
			controladorTerminales.ActualizarDatosTerminal(datosTerminal);
		}
		
		/// <summary>
		/// Permite desconectarse del servidor según la dirección asignada en
		/// el archivo de configuración.
		/// </summary>
		public static void DesconectarServidor()
		{
			DesconectarServidor(ConexionElement.Instancia.DireccionServidor);
		}
		
		/// <summary>
		/// Permite desconectarse de un servidor indicando la dirección del mismo.
		/// Este método se puede utilizar cuando el servidor registra al cliente como
		/// conectado pero no lo está, no permitiendo este estado que el cliente se
		/// conecte. Por lo tanto se puede intentar una desconexión al servidor para luego
		/// conectarse (ver método ConectarAlServidor).
		/// </summary>
		/// <param name="direccionServidor">
		/// A <see cref="System.String"/>
		/// </param>
		public static void DesconectarServidor(string direccionServidor)
		{
			logger.Debug("Intentando desconectar con el servidor");
			
			conectadoAlServidor = false;
			
			// Hago un downcast del proxy y lo valido contra el servidor
			ControladorTerminalesPrx controladorTerminalesproxy = null;
			try {
				/* FIXME: otras excepciones que puede tirar checkedCast son
				 * ConnectionTimeOut y ObjectNotExist. Revisar.
				 * FIXME: Ver también los timeouts y configurarlos para que sean
				 * coherentes en una red rápida. */
				logger.Debug("Validando proxy de Servicio de Terminal contra el servidor");
		    	controladorTerminalesproxy =
					ObtenerControladorTerminales(direccionServidor);
				
				logger.Debug("Desconectando del servidor");
				// Intento desconectarme del servidor
				controladorTerminalesproxy.Desconectar(datosTerminal);
				logger.Debug("Desconexión realizada correctamente");
			}
			catch (Ice.ConnectionRefusedException ex) {
				logger.Error("Conexión rechazada al intentar conectar con el Controlador de Terminales");
				
				if (DesconexionFallida != null)
						DesconexionFallida(ex);
			}
			catch (System.Exception ex) {
				logger.Error("Error desconocido al intentar conectar con el Controlador de Terminales:\n" +
				             ex.ToString());
				
				if (DesconexionFallida != null)
						DesconexionFallida(ex);
			}
			finally {
				// Limpio la lista de controladores
				controladores.Clear();
				
					// Disparo el evento de ConexionEstablecida
				if (DesconexionRealizada != null)
					DesconexionRealizada();
				
				logger.Debug("Desconexión realizada");
			}
		}
		#endregion
	}
}
