// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Net;
using System.ComponentModel;
using System.IO;
using System.Reflection;
using System.Diagnostics;
using System.Xml;
using System.Threading;
using System.Linq;

using Gtk;
using NLog;
using Ionic.Zip;

namespace StratosClient.Updater
{
	/// <summary>
	/// Clase principal que orquesta todo el proceso de actualización.
	/// </summary>
	public class ClientUpdater
	{
		private static ClientUpdater instancia;

		private static readonly Logger logger = LogManager.GetCurrentClassLogger();
		
		private const double CANTIDAD_FASES_DEL_PROCESO_DE_ACTUALIZACION = 6.0;
		
		private const string PREFIX_ZIP_BACKUP = "StratosClient-backup-";
		
		private const string MENSAJE_BAJANDO_ACTUALIZACION = "Bajando actualización";
		
		private const string MENSAJE_DESCOMPRIMIENDO_ACTUALIZACION = "Descomprimiendo actualización";
		
		private const string MENSAJE_PREPARANDO_DIRECTORIO_INSTALACION = "Preparando actualización";
		
		private const string MENSAJE_INSTALANDO_ACTUALIZACION = "Instalando actualización";
		
		private System.Action<double, int, string> progresoActualizacionHandler;
		
		private bool ocupado = false;
		
		private bool actualizacionCorrecta = false;
		
		private System.Exception excepcionArrojada;
		
		private Version versionActual;

		private ClientUpdater()
		{
			// Inicializo la versión actual.
			#pragma warning disable 0219
			var temp = VersionActual;
			#pragma warning restore 0219
		}

		public static ClientUpdater Instancia
		{
			get
			{
				if (instancia == null)
					instancia = new ClientUpdater();
				
				return instancia;
			}
		}
		
		/// <summary>
		/// Indica si el Updater está ocupado (realizando la actualización en alguna
		/// de sus etapas).
		/// </summary>
		public bool Ocupado
		{
			get { return ocupado; }
		}
		
		/// <summary>
		/// Indica si la última actualización ejecutada fue exitosa o no.
		/// </summary>
		public bool ActualizacionCorrecta
		{
			get { return actualizacionCorrecta; }
		}
		
		/// <summary>
		/// En caso de que la actualización haya sido errónea, contiene la
		/// excepción arrojada.
		/// </summary>
		public System.Exception ExcepcionArrojada
		{
			get { return excepcionArrojada; }
		}
		
		/// <summary>
		/// Función que será llamada ante los distintos eventos que vayan ocurriendo
		/// en ClientUpdater. Se pasará como argumento el progreso de la operación,
		/// la fase actual y el mensaje de estado.
		/// El progreso es un número del 0 al 100. La fase comienza con 1.
		/// </summary>
		public System.Action<double, int, string> ProgresoActualizacionHandler
		{
			set { progresoActualizacionHandler = value; }
		}
		
		/// <summary>
		/// Devuelve la versión actual de StratosClient. Para esto toma un assembly de
		/// referencia, como StratosClient.Core.Gui.
		/// La versión actual es cacheada. Se lee un assembly de referencia y se mantiene
		/// durante la ejecución.
		/// </summary>
		public Version VersionActual
		{
			get
			{
				if (versionActual == null)
				{
					string pathAssembly =
						Path.Combine(FileUtils.DirectorioDelAssembly(),
						             Configuracion.ASSEMBLY_VERSION);
					
					if (!File.Exists(pathAssembly))
						throw new Exception("No se encuentra en assembly para leer la versión actual: " +
						                    pathAssembly);
					
					Assembly assembly = Assembly.LoadFrom(pathAssembly);
					versionActual = assembly.GetName().Version;
				}
				
				return versionActual;
			}
		}

		/// <summary>
		/// Devuelve la versión de StratosClient del repositorio. Depende de que antes
		/// se haya bajado el nuevo archivo de configuración.
		/// </summary>
		public Version VersionRepositorio
		{
			get { return new Version(Configuracion.VersionConfiguracion); }
		}
		
		public bool ActualizacionDisponible
		{
			get { return (VersionActual.CompareTo(VersionRepositorio) < 0); }
		}
		
		public string PathRealDeStratosClient { get; set; }

		public void ComenzarActualizacion()
		{
			if (!ActualizacionDisponible)
				throw new InvalidOperationException("No es necesario actualizar");
			
			logger.Info("Comenzando proceso de actualización");
			
			ocupado = true;
			
			Thread thread = new Thread(delegate()
			{
				try
				{
					BajarActualizacion();
					
					DescomprimirActualizacion();
					PrepararDirectorioActualizacion();
					InstalarActualizacion();
					
					actualizacionCorrecta = true;
				}
				catch (Exception ex)
				{
					logger.ErrorException("Hubo un error al actualizar Stratos", ex);
					
					actualizacionCorrecta = false;
					excepcionArrojada = ex;
				}
				finally
				{
					ocupado = false;
					
					// Elimino el archivo de actualización bajado
					if (File.Exists(Configuracion.ARCHIVO_ACTUALIZACION))
						File.Delete(Configuracion.ARCHIVO_ACTUALIZACION);
				}
			});
			
			thread.Start();
		}

		protected void BajarActualizacion()
		{
			Uri url = new Uri(Configuracion.UrlArchivoActualizacion);
			
			logger.Info("Bajando actualización desde: {0}", url);
			
			WebClient webClient = new WebClient();
			webClient.DownloadProgressChanged += delegate(object sender, DownloadProgressChangedEventArgs e)
			{
				progresoActualizacionHandler(e.ProgressPercentage / CANTIDAD_FASES_DEL_PROCESO_DE_ACTUALIZACION,
				                             1, MENSAJE_BAJANDO_ACTUALIZACION);
			};
			
			webClient.DownloadFileAsync(url, Configuracion.ARCHIVO_ACTUALIZACION);
			
			// FIXME: Agregar un timeout.
			while (webClient.IsBusy);
		}

		protected void DescomprimirActualizacion()
		{
			logger.Info("Descomprimiendo actualización: {0}",
			            Configuracion.ARCHIVO_ACTUALIZACION);
			
			if (Directory.Exists(Configuracion.PATH_DIRECTORIO_TEMPORAL_STRATOS))
				Directory.Delete(Configuracion.PATH_DIRECTORIO_TEMPORAL_STRATOS, true);
			
			if (!ZipFile.IsZipFile(Configuracion.ARCHIVO_ACTUALIZACION, true))
				throw new Exception("El archivo Zip descargado es inválido");
			
			using (ZipFile zipFile = new ZipFile(Configuracion.ARCHIVO_ACTUALIZACION))
			{
				zipFile.ExtractProgress += delegate(object sender, ExtractProgressEventArgs e)
				{
					if (e.EntriesTotal == 0)
						return;
					
					progresoActualizacionHandler(((e.EntriesExtracted / (double) e.EntriesTotal) * 100.0) / CANTIDAD_FASES_DEL_PROCESO_DE_ACTUALIZACION,
					                             2, MENSAJE_DESCOMPRIMIENDO_ACTUALIZACION);
				};
				
				zipFile.ExtractAll(Configuracion.PATH_DIRECTORIO_TEMPORAL_STRATOS,
				                   ExtractExistingFileAction.OverwriteSilently);
			}
		}

		protected void PrepararDirectorioActualizacion()
		{
			logger.Info("Preparando directorio de actualización: {0}",
			            Configuracion.PATH_DIRECTORIO_TEMPORAL_STRATOS);
			
			ICollection<string> archivosNoActualizables = Configuracion.ArchivosNoActualizables;
			
			double pasoProgreso = 100 / (double) archivosNoActualizables.Count / CANTIDAD_FASES_DEL_PROCESO_DE_ACTUALIZACION;
			double progreso = 0;
			
			// TODO Actualmente la estrategia para preparar el directorio de instalación es
			// eliminar los archivos que no hay que sobreescribir en el directorio temporal donde
			// se está preparando la instalación. Como el proceso de instalación funciona copiando
			// y sobreescribiendo los archivos nuevos, con esta estrategia de preparación se evita
			// pisar archivos que no hay que pisar.
			// Otra estrategia es que se copien los archivos que no deben sobreescribirse al directorio
			// temporal. Esto permitiría hacer una instalación limpia de la actualización (borrando
			// todos los archivos anteriores). Podría ser una opción en el archivo .config
			
			string pathDestino;
			foreach (string archivo in archivosNoActualizables)
			{
//				pathFuente = Path.Combine(PathRealDeStratosClient, archivo);
				pathDestino = Path.Combine(Configuracion.PATH_DIRECTORIO_TEMPORAL_STRATOS, archivo);
				
				logger.Info("Eliminando '{0}'", pathDestino);
				
				if (Directory.Exists(pathDestino))
				{
					Directory.Delete(pathDestino, true);
					
//					FileUtils.CopyAll(pathFuente, pathDestino);
				}
				else if (File.Exists(pathDestino))
				{
					File.Delete(pathDestino);
					
//					File.Copy(pathFuente, pathDestino, true);
				}
				
				progreso += pasoProgreso;
				progresoActualizacionHandler(progreso, 3, MENSAJE_PREPARANDO_DIRECTORIO_INSTALACION);
			}
		}
		
		/// <summary>
		/// Realiza el proceso de instalación de la actualización. Estos son los pasos ejecutados:
		/// 1. Hacer un backup del directorio actual de StratosClient. Se hace una copia del directorio
		/// actual y además un archivo comprimido del mismo. El primero es para una restauración rápida
		/// por si hay algún error en el paso de instalación. El segundo es para un backup a largo plazo.
		/// 	1.1. Si hay algún error en este paso, no se procede con la actualización.
		/// 2. Instalar la actualización en el directorio actual de StratosClient. Se sobreescriben
		/// todos los archivos nuevos en el directorio de instalación de StratosClient.
		/// 	2.1. Si hay algún error en este paso, automáticamente se vuelve a copiar todo el
		/// 	directorio de backup en el directorio de instalación.
		/// 	2.2. Si no hay errores, se elimina el directorio de backup, quedando solo el archivo
		/// 	comprimido conteniendo al mismo.
		/// </summary>
		protected void InstalarActualizacion()
		{
			CrearBackup();
			
//			if (Directory.Exists(PathRealDeStratosClient))
//				FileUtils.EmptyDirectory(PathRealDeStratosClient);
			
			// Realizo la instalación propiamente dicha de la actualización. Se sobreescriben
			// todos los archivos.
			try
			{
				logger.Info("Instalando actualización");
				
				FileUtils.CopyAll(Configuracion.PATH_DIRECTORIO_TEMPORAL_STRATOS,
				                  PathRealDeStratosClient,
				                  CrearHandler(5));
				
				logger.Info("La instalación de la actualización fue exitosa");
				
				EliminarDirectorioBackup();
			}
			catch (Exception ex)
			{
				logger.Error("Hubo un error al instalar la actualización", ex);
				
				RecuperarDirectorioBackup();
				
				throw new Exception("Hubo un error al instalar la actualización. " +
					"StratosClient no ha sido actualizado", ex);
			}
		}
		
		private void RecuperarDirectorioBackup()
		{
			logger.Info("Recuperando instalación anterior");
			
			FileUtils.CopyAll(Configuracion.PATH_DIRECTORIO_BACKUP_STRATOS,
			                  PathRealDeStratosClient);
			
			logger.Info("Recuperación exitosa");
		}
		
		private void EliminarDirectorioBackup()
		{
			if (!Directory.Exists(Configuracion.PATH_DIRECTORIO_BACKUP_STRATOS))
				return;
			
			logger.Info("Eliminando directorio de backup de StratosClient: {0}",
			            Configuracion.PATH_DIRECTORIO_BACKUP_STRATOS);
			
			Directory.Delete(Configuracion.PATH_DIRECTORIO_BACKUP_STRATOS, true);
		}
		
		private void CrearBackup()
		{
			// Obtengo algunas estadísticas sobre el directorio actual de StratosClient
			// para llevar correctamente el progreso de copia de archivos y compresión
			// de los mismos. De esta forma evito realizar la operación varias veces.
			long tamañoTotalDirectorioStratos = 0;
			long cantidadElementosDirectorioStratos = 0;
			
			FileUtils.EstadisticasDirectorio(PathRealDeStratosClient,
			                                 ref tamañoTotalDirectorioStratos,
			                                 ref cantidadElementosDirectorioStratos);
			
			// Creo un directorio (copia exacta de la instalación actual de StratosClient)
			// para una fácil recuperación en caso de problemas. Este directorio es eliminado
			// inmediatamente después de una instalación exitosa.
			CrearDirectorioBackup(tamañoTotalDirectorioStratos);
			
			// Creo un archivo comprimido con el directorio de backup. Lo almaceno un nivel
			// superior al directorio actual de StratosClient.
			string directorioBackupComprimido = Path.GetDirectoryName(PathRealDeStratosClient);
			
			string archivoBackupComprimido =
				CrearBackupComprimido(directorioBackupComprimido,
				                      cantidadElementosDirectorioStratos);
			
			EliminarBackupsComprimidosAnteriores(directorioBackupComprimido, archivoBackupComprimido);
		}
		
		private void CrearDirectorioBackup(long tamañoTotalDirectorioStratos)
		{
			logger.Info("Creando backup del directorio de StratosClient: {0}",
			            PathRealDeStratosClient);
			
			// Borro el directorio de backup si existe
			if (Directory.Exists(Configuracion.PATH_DIRECTORIO_BACKUP_STRATOS))
				Directory.Delete(Configuracion.PATH_DIRECTORIO_BACKUP_STRATOS, true);
			
			FileUtils.CopyAll(PathRealDeStratosClient,
			                  Configuracion.PATH_DIRECTORIO_BACKUP_STRATOS,
			                  tamañoTotalDirectorioStratos,
			                  CrearHandler(4));
		}
		
		private void EliminarBackupsComprimidosAnteriores(string directorioBackup, string archivoAOmitir)
		{
			logger.Info("Eliminando backups anteriores");
			
			IEnumerable<FileInfo> archivosBackupPorEliminar =
				from f in Directory.GetFiles(directorioBackup, "*.zip")
					let fi = new FileInfo(f)
					where fi.Name.StartsWith(PREFIX_ZIP_BACKUP) &&
						!fi.Name.Equals(archivoAOmitir)
					select fi;
			
			foreach (var fi in archivosBackupPorEliminar)
			{
				logger.Info("Borrando backup: {0}", fi.FullName);
				fi.Delete();
			}
		}
		
		private string CrearBackupComprimido(string directorioBackup, long cantidadElementosDirectorioStratos)
		{
			string nombreZip = PREFIX_ZIP_BACKUP + VersionActual.ToString(3) + ".zip";
			string pathZip = Path.Combine(directorioBackup, nombreZip);
			
			logger.Info("Creando archivo comprimido del directorio de backup {0}", pathZip);
			
			if (File.Exists(pathZip))
				File.Delete(pathZip);
			
			using (ZipFile zip = new ZipFile())
			{
				zip.AddProgress += delegate(object sender, AddProgressEventArgs e)
				{
					progresoActualizacionHandler(((e.EntriesTotal / (double) cantidadElementosDirectorioStratos) * 100.0) / CANTIDAD_FASES_DEL_PROCESO_DE_ACTUALIZACION,
					                             6, MENSAJE_INSTALANDO_ACTUALIZACION);
				};
				
				zip.AddDirectory(PathRealDeStratosClient);
				zip.Save(pathZip);
			}
			
			return nombreZip;
		}
		
		private System.Action<string, long, long> CrearHandler(int fase)
		{
			return delegate(string nombreArchivo, long tamañoCopiado, long tamañoTotal)
			{
				progresoActualizacionHandler((tamañoCopiado / (double) tamañoTotal) * 100.0 / CANTIDAD_FASES_DEL_PROCESO_DE_ACTUALIZACION,
				                             fase, MENSAJE_INSTALANDO_ACTUALIZACION);
			};
		}

		/// <summary>
		/// Baja el archivo nuevo de configuración desde el servidor y sobreescribe el
		/// actual. Se encarga de loguear todas las excepciones ocurridas.
		/// </summary>
		public void ActualizarArchivoConfiguracion()
		{
			string archivoConfiguracion =
				Path.Combine(FileUtils.DirectorioDelAssembly(),
				             Configuracion.ARCHIVO_CONFIGURACION_UPDATER);
			
			string archivoConfiguracionTemporal = archivoConfiguracion + ".new";
			
			try
			{
				logger.Info("Actualizando archivo de configuración desde: " +
				            Configuracion.UrlArchivoConfiguracionUpdater);
				
				using (WebClient webClient = new WebClient())
				{
					// FIXME: Agregar un timeout.
					webClient.DownloadFile(Configuracion.UrlArchivoConfiguracionUpdater,
					                       archivoConfiguracionTemporal);
				}
				
				// Verifico que se haya bajado correctamente el archivo de
				// configuración. Sino, arrojo una excepción.
				if (!File.Exists(archivoConfiguracionTemporal))
					throw new Exception("El archivo de configuración no se ha bajado correctamente.");
				
				Configuracion.CerrarConfiguracion();
				
				File.Delete(archivoConfiguracion);
				File.Move(archivoConfiguracionTemporal, archivoConfiguracion);
				
				Configuracion.RecargarConfiguracion();
			}
			catch (XmlException ex)
			{
				logger.ErrorException("Error al leer el archivo de configuración", ex);
				throw new UpdaterException(ex);
			}
			catch (Exception ex)
			{
				logger.ErrorException("Error al bajar el archivo de configuración", ex);
				throw new UpdaterException(ex);
			}
			finally
			{
				if (File.Exists(archivoConfiguracionTemporal))
					File.Delete(archivoConfiguracionTemporal);
			}
		}
	}
}
