// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Diagnostics;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;

using Mono.Unix.Native;

namespace StratosClient.Updater
{
	public static class FileUtils
	{
		/// <summary>
		/// Reemplaza texto en un archivo.
		/// </summary>
		/// <param name="filePath">Path of the text file.</param>
		/// <param name="searchText">Text to search for.</param>
		/// <param name="replaceText">Text to replace the search text.</param>
		static public void ReplaceInFile(string filePath, string searchText, string replaceText)
		{
			StreamReader reader = new StreamReader(filePath);
			string content = reader.ReadToEnd();
			reader.Close();
			
			content = content.Replace(searchText, replaceText);
			
			StreamWriter writer = new StreamWriter(filePath);
			writer.Write(content);
			writer.Close();
		}
		
		public static string DirectorioDelAssembly()
		{
			return Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
		}
		
		public static void EstadisticasDirectorio(string d, ref long tamañoEnBytes, ref long cantidadArchivos) 
		{
			EstadisticasDirectorio(new DirectoryInfo(d), ref tamañoEnBytes, ref cantidadArchivos);
		}
		
		public static void EstadisticasDirectorio(DirectoryInfo d, ref long tamañoEnBytes, ref long cantidadElementos) 
		{
			FileInfo[] fis = d.GetFiles();
			foreach (FileInfo fi in fis)
			{      
				tamañoEnBytes += fi.Length;
				cantidadElementos++;
			}
			
			DirectoryInfo[] dis = d.GetDirectories();
			foreach (DirectoryInfo di in dis)
			{
				cantidadElementos++;
				EstadisticasDirectorio(di, ref tamañoEnBytes, ref cantidadElementos);
			}
		}
		
		public static void CopyAll(string source, string target)
		{
			CopyAll(source, target, null);
		}
		
		/// <summary>
		/// Copia recursivamente un directorio a otro lugar. Crea el destino si no existe,
		/// incluyendo los directorios padres. Es decir, genera toda la estructura destino.
		/// También permite pasar como argumento una función que será llamada antes de copiar
		/// cada archivo. Útil si se desea seguir el proceso de copia.
		/// </summary>
		/// <param name="source">
		/// Directorio fuente.
		/// </param>
		/// <param name="target">
		/// Directorio destino.
		/// </param>
		/// <param name="handler">
		/// Una función que es llamada antes de copiar cada archivo. Recibe como argumento el
		/// nombre del archivo, su tamaño, y el tamaño total del directorio a copiar.
		/// </param>
		public static void CopyAll(string source, string target, System.Action<string, long, long> handler)
		{
			long temp = 0;
			CopyAll(new DirectoryInfo(source), new DirectoryInfo(target), handler, ref temp, -1);
		}
		
		public static void CopyAll(string source, string target, long tamañoTotal, System.Action<string, long, long> handler)
		{
			long temp = 0;
			CopyAll(new DirectoryInfo(source), new DirectoryInfo(target), handler, ref temp, tamañoTotal);
		}

		private static void CopyAll(DirectoryInfo source, DirectoryInfo target,
		                            System.Action<string, long, long> handler,
		                            ref long tamañoCopiado, long tamañoTotal)
		{
			// Creo el directorio destino si no existe
			if (Directory.Exists(target.FullName) == false)
			{
				Directory.CreateDirectory(target.FullName);
			}
			
			if (tamañoTotal < 0)
			{
				long temp = 0;
				EstadisticasDirectorio(source, ref tamañoTotal, ref temp);
			}
			
			if (handler == null)
				handler = delegate {};
			
			// Copio cada archivo del directorio a su correspondiente lugar en el destino
			foreach (FileInfo fi in source.GetFiles())
			{
				tamañoCopiado += fi.Length;
				
				fi.CopyTo(Path.Combine(target.ToString(), fi.Name), true);
				handler(fi.Name, tamañoCopiado, tamañoTotal);
			}
			
			// Copio recursivamente todos los subdirectorios.
			foreach (DirectoryInfo diSourceSubDir in source.GetDirectories())
			{
				DirectoryInfo nextTargetSubDir = target.CreateSubdirectory(diSourceSubDir.Name);
				CopyAll(diSourceSubDir, nextTargetSubDir, handler, ref tamañoTotal, tamañoTotal);
			}
		}
		
		public static void EmptyDirectory(string pathDirectorio)
		{
			DirectoryInfo dir = new DirectoryInfo(pathDirectorio);
			
			foreach (FileInfo fi in dir.GetFiles())
			{
				fi.Delete();
			}
			
			foreach (DirectoryInfo di in dir.GetDirectories())
			{
				di.Delete(true);
			}
		}
		
		public static IEnumerable<string> ObtenerPathsAbsolutos(string directorioBase, IEnumerable<string> archivos)
		{
			IList<string> resultado = new List<string>();
			
			foreach (var archivo in archivos)
			{
				resultado.Add(Path.Combine(directorioBase, archivo));
			}
			
			return resultado;
		}
		
		public static void EjecutarExe(string pathEjecutable)
		{
			EjecutarExe(pathEjecutable, string.Empty);
		}
		
		/// <summary>
		/// Ejecuta un archivo ".exe" pasándole los argumentos dados. Se asegura que el
		/// archivo tenga permisos de ejecusión para el usuario, grupo y otros.
		/// </summary>
		/// <param name="pathEjecutable">
		/// Path del archivo .exe a ejecutar.
		/// </param>
		/// <param name="argumentos">
		/// Argumentos a pasarle al archivo ejecutable.
		/// </param>
		public static void EjecutarExe(string pathEjecutable, string argumentos)
		{
			try
			{
				// Me aseguro que el exe sea ejecutable. Ignoro cualquier error.
				Syscall.chmod(pathEjecutable, FilePermissions.S_IRWXU);
			}
			catch
			{
			}
			
			Process proc = new Process();
			proc.StartInfo.FileName = pathEjecutable;
			proc.StartInfo.WorkingDirectory = Path.GetDirectoryName(pathEjecutable);
			
			// Al nuevo proceso le paso el directorio real de StratosClient
			// donde debe copiar los archivos actualizados.
			proc.StartInfo.Arguments = argumentos;
			
			proc.Start();
		}
	}
}
