// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Gtk;

namespace StratosClient.Updater.GUI
{
	/// <summary>
	/// Permite generar mensajes de error en forma de diálogos, pero de una manera más
	/// simplificada. Permite también mostrar los mensajes de las excepciones como
	/// "Detalles avanzados".
	/// </summary>
	public class SMessageDialog : Gtk.MessageDialog
	{
		public delegate void SMessageDialogHandler();
		
		// Este evento se dispara al cerrarse el diálogo
		public event SMessageDialogHandler DialogoCerrado;
		
		private TextView txtDetalles;
		private System.Exception excepcion;
		
		public SMessageDialog(Window parent_window,
		                      DialogFlags flags,
		                      MessageType type,
		                      ButtonsType bt,
		                      bool use_markup,
		                      string format,
		                      params object[] args)
			: this(parent_window, flags, type, bt, use_markup, format, null, args)
		{
		}
		
		public SMessageDialog(Window parent_window,
		                      DialogFlags flags,
		                      MessageType type,
		                      ButtonsType bt,
		                      bool use_markup,
		                      string format,
		                      System.Exception excepcion,
		                      params object[] args)
			: base(parent_window, flags, type, bt, use_markup, format.Trim(), args)
		{
			if (excepcion == null)
				return;
			
			this.excepcion = excepcion;
			
			/* En caso de que la excepción sea de tipo TargetInvocationException, la
			 * excepción real se encuentra en la propiedad InnerException. */
			if (this.excepcion is System.Reflection.TargetInvocationException)
				this.excepcion = this.excepcion.InnerException;
			
			// Agrego un Expander con un Entry
			this.txtDetalles = new Gtk.TextView();
			this.txtDetalles.WidthRequest = 500;
			this.txtDetalles.HeightRequest = 250;
			
			string detalles = this.excepcion.ToString();
			
			this.txtDetalles.Buffer.Text = detalles;
			
			// Agrego el Expander a una ScrolledWindow y lo muestro
			Gtk.ScrolledWindow scrolledWindow = new Gtk.ScrolledWindow();
			scrolledWindow.Add(this.txtDetalles);
			
			Expander expander = new Expander("Detalles (avanzado)");
			expander.Add(scrolledWindow);
			
			expander.ShowAll();
			
			this.VBox.PackStart(expander);
			
			/* En este punto al foco lo tiene el Expander. Le doy el foco al botón Aceptar. */
			this.ActionArea.Children[0].GrabFocus();
		}
		
		public System.Exception Excepcion {
			get; set;
		}
		
		/// <summary>
		/// Oculta el método Run de Gtk.MessageDialog agregandole a éste la ejecución
		/// del evento DialogoCerrado.
		/// </summary>
		/// <returns>
		/// A <see cref="System.Int32"/>
		/// </returns>
		public new int Run()
		{
			// Corro el diálogo
			int response = base.Run();
			
			/* Cuando el usuario oprime un botón en el mismo, ejecuto el evento
			 * y devuelvo la respuesta del diálogo. */
			if (this.DialogoCerrado != null)
				this.DialogoCerrado();
			
			return response;
		}
		
		private static string ObtenerMensaje(System.Exception excepcion)
		{
			/* Si se trata de una excepción de tipo TypeInitializationException, me
			 * fijo en la InnerException */
			System.Exception exTemp = excepcion;
			
			while (exTemp is TypeInitializationException &&
			       exTemp.InnerException != null)
				exTemp = exTemp.InnerException;
			
			return ObtenerMensajeInterno(exTemp);
		}
		
		/// <summary>
		/// Permite obtener el mensaje para el diálogo de acuerdo al tipo de excepción
		/// pasada como argumento. Este método es muy útil, y en lugar de utilizar mensajes
		/// distintos en todos los diálogos para un mismo tipo de error, se debe utilizar
		/// este método. Si surgen excepciones nuevas para tratar, agregarlas aquí.
		/// </summary>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		private static string ObtenerMensajeInterno(System.Exception excepcion)
		{
			// De acuerdo al tipo de excepción armo el mensaje a mostrar en el diálogo.
			// Agregar aquí otras excepciones comunes.
			
			return "Error. Poner algo como la gente acá";
		}
		
		#region Métodos estáticos
		public static void MensajeInformacion(string mensaje)
		{
			SMessageDialog dialog =
				GetSMessageDialog(MainWindow.Instancia, mensaje, DialogFlags.Modal,
				                  MessageType.Info, ButtonsType.Ok, null);
			
			dialog.Run();
			dialog.Destroy();
		}
		
		/// <summary>
		/// Muestra un mensaje de advertencia (Warning).
		/// </summary>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		public static void MensajeAdvertencia(string mensaje)
		{
			SMessageDialog dialog =
				GetSMessageDialog(MainWindow.Instancia, mensaje, DialogFlags.Modal,
				                  MessageType.Warning, ButtonsType.Ok, null);
			
			dialog.Run();
			dialog.Destroy();
		}
		
		/// <summary>
		/// Permite mostrar un diálogo para que el usuario responda "Si" o "No" ante
		/// un mensaje.
		/// </summary>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="ResponseType"/>
		/// </returns>
		public static ResponseType MensajePregunta(string mensaje)
		{
			return MensajePregunta(MainWindow.Instancia, mensaje, ResponseType.No);
		}
		
		public static ResponseType MensajePregunta(string mensaje, ResponseType respuestaPorDefecto)
		{
			return MensajePregunta(MainWindow.Instancia, mensaje, respuestaPorDefecto);
		}
		
		public static ResponseType MensajePregunta(Gtk.Window parent, string mensaje)
		{
			return MensajePregunta(MainWindow.Instancia, mensaje, ResponseType.No);
		}
		
		/// <summary>
		/// Permite mostrar un diálogo para que el usuario responda "Si" o "No" ante
		/// un mensaje. Esta sobrecarga permite especificar la ventana padre del diálogo.
		/// </summary>
		/// <param name="parent">
		/// A <see cref="Gtk.Window"/>
		/// </param>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="ResponseType"/>
		/// </returns>
		public static ResponseType MensajePregunta(Gtk.Window parent,
		                                           string mensaje,
		                                           ResponseType respuestaPorDefecto)
		{
			SMessageDialog dialog =
				GetSMessageDialog(parent, mensaje, DialogFlags.Modal, MessageType.Question,
				                  ButtonsType.YesNo);
			
			dialog.DefaultResponse = respuestaPorDefecto;
			
			ResponseType respuesta = (ResponseType) dialog.Run();
			dialog.Destroy();
			
			return respuesta;
		}
		
		/// <summary>
		/// Método para simplificar la ejecución del diálogo SMessageDialog.
		/// </summary>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		public static void MensajeError(string mensaje)
		{
			MensajeError(MainWindow.Instancia, mensaje, null);
		}
		
		public static void MensajeError(string mensaje, Gtk.Window parent)
		{
			MensajeError(parent, mensaje, null);
		}
		
		/// <summary>
		/// Método para simplificar la ejecución del diálogo SMessageDialog, pidiendo
		/// sólo la excepción y generándose automáticamente el mensaje del diálogo
		/// de acuerdo a ésta.
		/// </summary>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		public static void MensajeError(System.Exception excepcion)
		{
			SMessageDialog dialog = GetSMessageDialogError(excepcion);
			
			dialog.Run();
			dialog.Destroy();
		}
		
		/// <summary>
		/// Método para simplificar la ejecución del diálogo SMessageDialog.
		/// </summary>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		public static void MensajeError(string mensaje, System.Exception excepcion)
		{
			MensajeError(MainWindow.Instancia, mensaje, excepcion);
		}
		
		/// <summary>
		/// Método para simplificar la ejecución del diálogo SMessageDialog.
		/// </summary>
		/// <param name="parent_window">
		/// Ventana padre.
		/// </param>
		/// <param name="mensaje">
		/// Mensaje a mostrar en el diálogo.
		/// </param>
		/// <param name="excepcion">
		/// Excepción que se procesará y se mostrará en la parte inferior del diálogo,
		/// a modo de "Detalles avanzados".
		/// </param>
		public static void MensajeError(Window parent_window, string mensaje,
		                                System.Exception excepcion)
		{
			SMessageDialog dialog = GetSMessageDialogError(parent_window, mensaje, excepcion);
			
			dialog.Run();
			dialog.Destroy();
		}
		
		/// <summary>
		/// Permite obtener un objeto SMessageDialog (un diálogo). Esta versión simplificada
		/// pide solamente un mensaje.
		/// </summary>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="SMessageDialog"/>
		/// </returns>
		public static SMessageDialog GetSMessageDialogError(string mensaje)
		{
			return GetSMessageDialogError(MainWindow.Instancia, mensaje, null);
		}
		
		/// <summary>
		/// Permite obtener un SMessageDialog pasando únicamente la excepción, generándose
		/// automáticamente el mensaje del diálogo según el tipo de excepción.
		/// </summary>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		/// <returns>
		/// A <see cref="SMessageDialog"/>
		/// </returns>
		public static SMessageDialog GetSMessageDialogError(System.Exception excepcion)
		{
			return GetSMessageDialogError(ObtenerMensaje(excepcion), excepcion);
		}
		
		/// <summary>
		/// Permite obtener un objeto SMessageDialog (un diálogo). Esta versión simplificada
		/// pide solamente un mensaje y una excepción.
		/// </summary>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="exception">
		/// A <see cref="System.Exception"/>
		/// </param>
		/// <returns>
		/// A <see cref="SMessageDialog"/>
		/// </returns>
		public static SMessageDialog GetSMessageDialogError(string mensaje, System.Exception exception)
		{
			return GetSMessageDialogError(MainWindow.Instancia, mensaje, exception);
		}
		
		/// <summary>
		/// Es el método GetSMessageDialogError de más bajo nivel, que permite especificar
		/// todos los parámetros de un diálogo de error.
		/// </summary>
		/// <param name="parent_window">
		/// A <see cref="Window"/>
		/// </param>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		/// <returns>
		/// A <see cref="SMessageDialog"/>
		/// </returns>
		public static SMessageDialog
			GetSMessageDialogError(Window parent_window, string mensaje,
			                       System.Exception excepcion)
		{
			return GetSMessageDialog(parent_window, mensaje, DialogFlags.Modal,
			                         MessageType.Error, ButtonsType.Ok, excepcion);
		}
		
		/// <summary>
		/// Permite obtener un objeto SMessageDialog (un diálogo). Hay que especificar la
		/// ventana principal, el mensaje y la excepción.
		/// </summary>
		/// <param name="parent_window">
		/// A <see cref="Window"/>
		/// </param>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		/// <returns>
		/// A <see cref="SMessageDialog"/>
		/// </returns>
		public static SMessageDialog
			GetSMessageDialog(Window parent_window, string mensaje,
			                  DialogFlags dialogFlags, MessageType tipoMensaje,
			                  ButtonsType tipoBotones, System.Exception excepcion)
		{
			SMessageDialog dialog = new SMessageDialog(parent_window,
			                                           dialogFlags,
			                                           tipoMensaje,
			                                           tipoBotones,
			                                           true,
			                                           mensaje,
			                                           excepcion);
			
			string titulo = String.Empty;
			if (tipoMensaje == MessageType.Error)
				titulo = "Error";
			else if (tipoMensaje == MessageType.Warning)
				titulo = "Advertencia";
			else if (tipoMensaje == MessageType.Info)
				titulo = "Información";
			else if (tipoMensaje == MessageType.Question)
				titulo = "Pregunta";
			
			dialog.Title = titulo;
			
			return dialog;
		}
		
		/// <summary>
		/// Permite obtener un SMessageDialog sin especificar una excepción.
		/// </summary>
		/// <param name="parent_window">
		/// A <see cref="Window"/>
		/// </param>
		/// <param name="mensaje">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="dialogFlags">
		/// A <see cref="DialogFlags"/>
		/// </param>
		/// <param name="tipoMensaje">
		/// A <see cref="MessageType"/>
		/// </param>
		/// <param name="tipoBotones">
		/// A <see cref="ButtonsType"/>
		/// </param>
		/// <returns>
		/// A <see cref="SMessageDialog"/>
		/// </returns>
		public static SMessageDialog
			GetSMessageDialog(Window parent_window, string mensaje,
			                  DialogFlags dialogFlags, MessageType tipoMensaje,
			                  ButtonsType tipoBotones)
		{
			return GetSMessageDialog(parent_window, mensaje, dialogFlags, tipoMensaje,
			                         tipoBotones, null);
		}
		#endregion
	}
}
