// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.IO;
using System.Linq;
using System.Threading;
using System.Reflection;

using Gtk;
using NLog;

using StratosClient.Updater.GUI;

namespace StratosClient.Updater
{
	/// <summary>
	/// Se encarga de orquestar la actualización de StratosClient.
	/// <para />
	/// Utiliza la clase <c>ClientUpdater</c> para bajar las actualizaciones y conocer
	/// las versiones actuales y disponibles. Se encarga de ejecutar el binario de
	/// StratosClient cuando todo el proceso, exitoso o no, haya finalizado.
	/// </summary>
	public class UpdaterManager
	{
		private static readonly Logger logger = LogManager.GetCurrentClassLogger();
		
		private const string OPCION_NO_COPIAR_EJECUTABLE = "--noExeCopy";
		
		private string[] applicationArgs;
		
		private ClientUpdater stratosClientUpdater;
		
		public UpdaterManager(string[] args)
		{
			applicationArgs = args;
			
			stratosClientUpdater = ClientUpdater.Instancia;
			stratosClientUpdater.PathRealDeStratosClient = PathRealDeStratosClient;
		}
		
		/// <summary>
		/// Indica si es necesario realizar la copia de StratosClient.Updater a
		/// una ubicación temporal. Esto sucede si se encuentra en los argumentos
		/// de la aplicación la opción <see cref="OPCION_NO_COPIAR_EJECUTABLE"/>.
		/// </summary>
		/// <returns><c>true</c> si es necesario copiar StratosClient.Updater a una
		/// ubicación temporal. <c>false</c> en caso contrario.</returns>
		private bool EsNecesarioRealizarCopiaDelUpdater
		{
			get { return !applicationArgs.Contains(OPCION_NO_COPIAR_EJECUTABLE); }
		}
		
		/// <summary>
		/// Path real en donde se encuentran los archivos de StratosClient.
		/// <para>
		/// Cuando se ejecuta la copia en el directorio temporal, es necesario pasarle como
		/// argumento la ubicación real de StratosClient, para que pueda realizar la
		/// actualización correctamente, y además ejecutar el binario de StratosClient
		/// al finalizar.
		/// </para>
		/// </summary>
		/// <returns>El path real donde se encuentra StratosClient.</returns>
		private string PathRealDeStratosClient
		{
			get
			{
				string pathReal = applicationArgs
					.Where(el => !string.Equals(el, OPCION_NO_COPIAR_EJECUTABLE))
						.FirstOrDefault();
				
				if (string.IsNullOrEmpty(pathReal))
				{
					pathReal = FileUtils.DirectorioDelAssembly();
				}
				
				return Path.GetFullPath(pathReal) ?? string.Empty;
			}
		}
		
		/// <summary>
		/// Inicia el proceso de actualización. Es el punto de entrada para comenzar
		/// a actualizar el software de StratosClient.
		/// <para />
		/// Verifica la existencia de una nueva versión, y en caso de haberla lo notifica
		/// al usuario, dándole la posibilidad de realizarla o no. Siempre ejecuta el
		/// binario principal de StratosClient al finalizar, haya habido errores o no.
		/// </summary>
		public void Iniciar()
		{
			bool iniciarStratosClientAlFinal = true;
			
			try
			{
				logger.Info("Iniciando UpdaterManager con los siguientes argumentos: " +
				            string.Join(" ", applicationArgs));
				
				logger.Info("Versión actual de StratosClient: {0}", stratosClientUpdater.VersionActual);
				
				stratosClientUpdater.ActualizarArchivoConfiguracion();
				
				if (stratosClientUpdater.ActualizacionDisponible)
				{
					logger.Info("Hay una actualización disponible: " + stratosClientUpdater.VersionRepositorio);
					
					// Me fijo en las opciones para saber si tengo que copiar el
					// ejecutable en una carpeta temporal y ejecutarlo desde allí.
					// Cuando se ejecuta la copia, esta opción por línea de comandos
					// no se pasa, evitando una nueva copia.
					if (EsNecesarioRealizarCopiaDelUpdater)
					{
						logger.Info("Opción de copia en carpeta temporal activada");
						
						string nuevoEjecutable = CopiarEjecutableEnCarpetaTemporal();
						EjecutarCopia(nuevoEjecutable);
						
						iniciarStratosClientAlFinal = false;
						
						return;
					}
					
					IniciarGUI();
				}
				else
				{
					logger.Info("No hay actualizaciones disponibles");
				}
			}
			catch (UpdaterException)
			{
				// No logueamos nada. Este tipo de excepciones ya son manejadas en
				// la clase ClientUpdater.
			}
			catch (Exception ex)
			{
				logger.ErrorException("Hubo un error desconocido en el proceso de actualización", ex);
			}
			finally
			{
				if (iniciarStratosClientAlFinal)
					IniciarStratosClient();
			}
			
			logger.Info("Saliendo de UpdaterManager");
		}
		
		/// <summary>
		/// Copia los binarios y archivos necesarios de StratosClient.Updater a
		/// una ubicación temporal.
		/// </summary>
		/// <returns>
		/// La ruta del ejecutable principal de StratosClient.Updater, copiado en la
		/// carpeta temporal.
		/// </returns>
		private string CopiarEjecutableEnCarpetaTemporal()
		{
			string directorioTemporal = CrearDirectorioTemporal();
			
			string[] archivosPorCopiar = {
				"StratosClient.Core.Gui.dll",
				"StratosClient.Updater.exe",
				"StratosClient.Updater.exe.config",
				"Ionic.Zip.dll",
				"NLog.dll",
				"NLog.config"
			};
			
			foreach (string archivoPorCopiar in archivosPorCopiar)
			{
				string archivoDestino = Path.Combine(directorioTemporal, archivoPorCopiar);
				File.Copy(Path.Combine(FileUtils.DirectorioDelAssembly(), archivoPorCopiar),
				          archivoDestino, true);
			}
			
			// Modifico NLog.config para que realice logging en el directorio de instalación,
			// y no en el temporal.
			FileUtils.ReplaceInFile(Path.Combine(directorioTemporal, "NLog.config"),
			                        "${basedir}",
			                        PathRealDeStratosClient);
			
			return Path.Combine(directorioTemporal, "StratosClient.Updater.exe");
		}
		
		private string CrearDirectorioTemporal()
		{
			string path;
			
			do
			{
				path = Path.Combine(Path.GetTempPath(), Path.GetRandomFileName());
			}
			while (Directory.Exists(path));
			
			Directory.CreateDirectory(path);
			
			return path;
		}
		
		/// <summary>
		/// Ejecuta la copia de StratosClient.Updater dada como argumento.
		/// </summary>
		/// <param name="ejecutableCopiado">
		/// Ejecutable principal de StratosClient.Updater.
		/// </param>
		private void EjecutarCopia(string ejecutableCopiado)
		{
			try
			{
				logger.Info("Ejecutando nueva copia del Updater: {0}", ejecutableCopiado);
				
				// Al nuevo proceso le paso el directorio real de StratosClient
				// donde debe copiar los archivos actualizados.
				FileUtils.EjecutarExe(ejecutableCopiado,
				                      OPCION_NO_COPIAR_EJECUTABLE + " \"" +
				                      Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) +
				                      "\"");
			}
			catch (Exception ex)
			{
				logger.ErrorException("Error al iniciar nueva copia del Updater", ex);
				throw new UpdaterException(ex);
			}
		}
		
		/// <summary>
		/// Inicia StratosClient de acuerdo con la ruta como argumento al ejecutar
		/// StratosClient.Updater.
		/// </summary>
		public void IniciarStratosClient()
		{
			try
			{
				string rutaStratosClient =
					Path.Combine(PathRealDeStratosClient,
					             Configuracion.ASSEMBLY_PRINCIPAL_STRATOS_CLIENT);
				
				if (!File.Exists(rutaStratosClient))
				{
					logger.Error("No se encuentra el assembly principal de StratosClient: " +
					             rutaStratosClient);
				}
				else
				{
					logger.Info("Iniciando StratosClient: {0}", rutaStratosClient);
					
					FileUtils.EjecutarExe(rutaStratosClient);
				}
			}
			catch (Exception ex)
			{
				logger.ErrorException("Error al ejecutar StratosClient", ex);
			}
		}
		
		/// <summary>
		/// Inicia el componente gráfico de StratosClient.Updater para permitir al usuario
		/// realizar la actualización.
		/// </summary>
		private void IniciarGUI()
		{
			Thread gtkThread = new Thread(delegate ()
			{
				Gtk.Application.Init("StratosClient.Updater", ref applicationArgs);
				
				GLib.ExceptionManager.UnhandledException += delegate(GLib.UnhandledExceptionArgs args)
				{
					logger.Error("Excepción no capturada en el thread Gtk: {0}", args.ExceptionObject);
					Gtk.Application.Quit();
				};
				
				AppDomain.CurrentDomain.UnhandledException += delegate(object sender, UnhandledExceptionEventArgs e)
				{
					logger.Error("Excepción no capturada en el dominio de la aplicación: {0}", e.ExceptionObject);
				};
				
				MainWindow win = new MainWindow();
				win.Show();
				
				Gtk.Application.Run();
			});
			
			logger.Info("Iniciando GUI");
			gtkThread.Start();
			gtkThread.Join();
		}
	}
}

