// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Gtk;

using Stratos.ModeloDto;
using StratosClient.Widgets.Level1;

namespace StratosClient.Widgets.Level2
{
	/// <summary>
	/// Esta clase es utilizada para facilitar la inicialización de un Gtk.Entry
	/// que contenga la razón social de una persona, ya sea proveedor o cliente.
	/// </summary>
	//[System.ComponentModel.ToolboxItem(true)]
	public class EntryRazonSocial<T> : Gtk.Entry, IEntryRazonSocial
		where T : PersonaDto, new()
	{
		#region Constructores
		public EntryRazonSocial()
		{
			/* Creo el objeto EntryCompletion, su modelo (que contendrá objetos
			 * ClienteDto o ProveedorDto) y su CellRenderer. */
			EntryCompletion razonSocialCompletion = new EntryCompletion();
			razonSocialCompletion.PopupSetWidth = false;
			this.Completion = razonSocialCompletion;
			
			ListStore razonSocialModelo = new ListStore(typeof(T));
			razonSocialCompletion.Model = razonSocialModelo;
			
			CellRendererText codigoRenderer = new CellRendererText();
			razonSocialCompletion.PackStart(codigoRenderer, false);
			razonSocialCompletion.SetCellDataFunc(codigoRenderer, this.RenderCodigo);
			
			CellRendererText razonSocialRenderer = new CellRendererText();
			// HACK: Estas son posibles optimizaciones para el CellRenderer:
			//razonSocialRenderer.SetFixedSize(1, -1);
			//razonSocialRenderer.Height = 1;
			razonSocialCompletion.PackStart(razonSocialRenderer, true);
			razonSocialCompletion.SetCellDataFunc(razonSocialRenderer, this.RenderRazonSocial);
			
			CellRendererText cuitRenderer = new CellRendererText();
			razonSocialCompletion.PackStart(cuitRenderer, false);
			razonSocialCompletion.SetCellDataFunc(cuitRenderer, this.RenderCuit);
			
			// Este cell renderer hace que el EntryCompletion se expanda un poco más de lo normal.
			// No se muestra información útil, es sólo para ése propósito.
			CellRendererText temp = new CellRendererText();
			razonSocialCompletion.PackStart(temp, false);
			razonSocialCompletion.SetCellDataFunc(temp, this.RenderUltimaColumna);
			
			// Asigno la función de match
			razonSocialCompletion.MatchFunc = this.MatchRazonSocial;
		}
		#endregion
		
		#region Propiedades
		public new bool Sensitive {
			get { return this.Sensitive; }
			set { CustomWidgetsUtils.WidgetEditable(this, value); }
		}
		
		public ListStore ModeloEntryCompletion {
			get { return this.Completion.Model as ListStore; }
		}
		#endregion
		
		#region Eventos
		protected override bool OnKeyPressEvent (Gdk.EventKey evnt)
		{
			/* Hago un clear del modelo del EntryCompletion para forzar al usuario
			 * a realizar otra búsqueda al servidor, y que no se realice un filtrado
			 * local. */
			if (!(evnt.Key == Gdk.Key.Return)) {
				(this.Completion.Model as ListStore).Clear();
				this.Completion.Complete();
			}
			
			return base.OnKeyPressEvent(evnt);
		}
		#endregion
		
		#region Métodos de renderizado
		private void RenderCodigo(CellLayout cellLayout, CellRenderer renderer,
		                          TreeModel model, TreeIter iter)
		{
			T persona = (T) model.GetValue(iter, 0);
			string valor = persona != null ? persona.codigo.ToString() : String.Empty;
			
			(renderer as CellRendererText).Text = valor;
		}
		
		private void RenderRazonSocial(CellLayout cellLayout, CellRenderer renderer,
		                               TreeModel model, TreeIter iter)
		{
			T persona = (T) model.GetValue(iter, 0);
			string valor = (persona != null && persona.razonSocial != null) ? persona.razonSocial : String.Empty;
			
			(renderer as CellRendererText).Text = valor;
		}
		
		private void RenderCuit(CellLayout cellLayout, CellRenderer renderer,
		                        TreeModel model, TreeIter iter)
		{
			T persona = (T) model.GetValue(iter, 0);
			string valor = (persona != null && persona.cuit != null) ? persona.cuit : String.Empty;
			
			(renderer as CellRendererText).Text = valor;
		}
		
		protected void RenderUltimaColumna(CellLayout cellLayout, CellRenderer renderer,
		                                   TreeModel model, TreeIter iter)
		{
			CellRendererText cellText = (CellRendererText) renderer;
			cellText.Text = "   ";
		}
		#endregion
		
		#region Otros métodos
		private bool MatchRazonSocial(EntryCompletion entryCompletion, string key,
		                              TreeIter iter)
		{
			/* Se muestran siempre todas las personas, ya que al modificarse el Entry
			 * se fuerza al usuario a realizar una nueva búsqueda al servidor (no hay
			 * filtrado local). */
			return true;
		}
		#endregion
	}
}
