// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;

using Gtk;

using StratosClient.Core.Gui;
using Stratos.ModeloDto;
using Stratos.Common;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;

namespace StratosClient.Widgets.Level2
{
	/// <summary>
	/// Clase que tiene la lógica para gestionar la búsqueda y carga de personas (Clientes,
	/// Proveedores, etc).
	/// </summary>
	public abstract class RazonSocialPresenter<T> where T : PersonaDto, new()
	{
		#region Campos
		public event System.Action<T> PersonaCargada;
		public event System.Action PersonaQuitada;
		
		protected IRazonSocialVista<T> razonSocialVista;
		protected T persona;
		#endregion
		
		#region Constructores
		public RazonSocialPresenter (IRazonSocialVista<T> razonSocialVista)
		{
			this.razonSocialVista = razonSocialVista;
		}
		#endregion
		
		#region Propiedades
		public T Persona {
			get { return this.persona; }
		}
		#endregion
		
		#region Eventos
		internal void CodigoPersonaActivado()
		{
			// Si hay un cliente actualmente cargado, no realizo ninguna acción
			if (this.Persona != null && this.Persona.codigo > 0)
				return;
			
			// Verifico que el código sea válido y que posea un valor entero
			if (!this.razonSocialVista.EntryCodigoPersona.TextoValido ||
			    	!StratosUtil.IsInt(this.razonSocialVista.EntryCodigoPersona.Text)) {
				this.razonSocialVista.NotificarError("El código debe ser numérico.");
				return;
			}
			
			int codigoIngresado = Convert.ToInt32(this.razonSocialVista.EntryCodigoPersona.Text);
			
			/* FIXME: Ver de una forma de desactivar la interfaz gráfica o algo para
			 * que el usuario sepa que se está realizando la búsqueda. */
			
			// Buscamos la persona
			T[] personas =
				((ControladorPersonasPrx) MainIce.GetControlador(this.NombreControlador))
					.BusquedaExactaPorCodigo(codigoIngresado).Cast<T>().ToArray();
			
			if (personas.Length == 0) {
				this.razonSocialVista.NotificarError("No existe con el código '" +
				                                     codigoIngresado.ToString() +
				                                     "'");
				
				return;
			}
			
			this.CargarPersona(personas[0], true);
		}
		
		public void RazonSocialActivado ()
		{
			// Si hay una persona actualmente cargada, no realizo ninguna acción
			if (this.Persona != null && this.Persona.codigo > 0)
				return;
			
			// Verifico que se haya ingresado algún criterio
			if (this.razonSocialVista.EntryRazonSocialPersona.Text.Trim() == String.Empty) {
				this.razonSocialVista.NotificarAdvertencia("Ingrese algún criterio para buscar");
				return;
			}
			
			ListStore modelo =
				(ListStore) this.razonSocialVista
					.EntryRazonSocialPersona.ModeloEntryCompletion;
			
			// Realizo la búsqueda de las personas según el criterio ingresado
			T[] personasMatch = null;
			
			try {
				personasMatch =
					((ControladorPersonasPrx) MainIce.GetControlador(this.NombreControlador))
						.ConsultaPorRazonSocial(this.razonSocialVista.EntryRazonSocialPersona.Text)
							.Cast<T>().ToArray();
			}
			catch (ParametrosInvalidosException ex) {
				this.razonSocialVista
					.NotificarError("Hubo un error al realizar la consulta: " + ex.mensaje);
				return;
			}

			
			if (personasMatch.Length == 0) {
				this.razonSocialVista
					.NotificarError("No hubo resultados con los criterios " +
					                "de búsqueda usados:\n'" +
					                this.razonSocialVista.EntryRazonSocialPersona.Text + "'");
				
				return;
			}
			
			/* Una vez obtenidos los resultados, quito todos los que han sido agregados
			 * previamente al modelo del EntryCompletion en búsquedas anteriores */
			modelo.Clear();
			
			// Ingreso los nuevos
			foreach (T p in personasMatch)
				modelo.InsertWithValues(0, p);
			
			// FIXME: Ver de una forma más elegante de activar el EntryCompletion
			string valorAnterior = this.razonSocialVista.EntryRazonSocialPersona.Text;
			int posicionAnterior = this.razonSocialVista.EntryRazonSocialPersona.Position;
			this.razonSocialVista.EntryRazonSocialPersona.Text = String.Empty;
			this.razonSocialVista.EntryRazonSocialPersona.InsertText(valorAnterior);
			this.razonSocialVista.EntryRazonSocialPersona.Position = posicionAnterior;
		}
		#endregion
		
		#region Métodos de carga/vaciado
		/// <summary>
		/// Carga una persona.
		/// </summary>
		/// <param name="persona">
		/// A <see cref="T"/>
		/// </param>
		public void CargarPersona(T persona)
		{
			this.CargarPersona(persona, true);
		}
		
		/// <summary>
		/// Permite cargar una persona e indicar si se va a disparar o no el evento
		/// de PersonaCargada.
		/// </summary>
		/// <param name="persona">
		/// A <see cref="T"/>
		/// </param>
		/// <param name="dispararEventoPersonaCargada">
		/// A <see cref="System.Boolean"/>
		/// </param>
		protected void CargarPersona(T persona, bool dispararEventoPersonaCargada)
		{
			this.razonSocialVista.EntryCodigoPersona.Text =
				persona.codigo < 0 ? "" : persona.codigo.ToString();
			this.razonSocialVista.EntryRazonSocialPersona.Text =
				persona.razonSocial == null ? "" : persona.razonSocial;
			
			this.persona = persona;
			
			this.razonSocialVista.Editable = this.persona.codigo <= 0;
			
			// Lanzo el evento
			if (this.PersonaCargada != null && dispararEventoPersonaCargada)
				this.PersonaCargada(this.persona);
		}
		
		/// <summary>
		/// Quita la persona actualmente cargada en el componente.
		/// </summary>
		public void Vaciar()
		{
			// Cargo un cliente vacío, pero evitando que se dispare el evento PersonaCargada
			T personaVacia = new T();
			personaVacia.codigo = -1;
			personaVacia.domicilio = new DomicilioDto();
			
			this.CargarPersona(personaVacia, false);
			
			// Disparo el evento PersonaQuitada
			if (this.PersonaQuitada != null)
				this.PersonaQuitada();
			
			this.persona = null;
		}
		#endregion
		
		#region Métodos/Propiedades abstractas
		/// <summary>
		/// Valor que indica el nombre del controlador que será obtenido para realizar
		/// las consultas. Las subclases deben devolver el valor correspondiente.
		/// </summary>
		protected abstract string NombreControlador {
			get;
		}
		#endregion
	}
}
