// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Gtk;

using Stratos.ModeloDto;

namespace StratosClient.Widgets.Level2
{
	/// <summary>
	/// Renderizador base para todos los CellRendererText en un TreeView
	/// para mostrar los items de una factura.
	/// </summary>
	public abstract class CellRendererInteractivoBase<TItem> : CellRendererText
		where TItem : ItemGenericoDto
	{
		#region Campos
		/// <summary>
		/// TreeView al que pertenece el CellRendererBase
		/// </summary>
		/// <param name="treeView">
		/// A <see cref="TvItemsFactura"/>
		/// </param>
		protected ListaInteractiva<TItem> treeView;
		#endregion
		
		#region Constructores
		public CellRendererInteractivoBase(ListaInteractiva<TItem> treeView)
		{
			this.treeView = treeView;
			this.Editable = true;
		}
		#endregion
		
		#region Eventos
		/// <summary>
		/// Facilita la edición de un valor para todas las columnas, realizando en un
		/// solo lugar las operaciones repetitivas.
		/// </summary>
		/// <param name="path">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="new_text">
		/// A <see cref="System.String"/>
		/// </param>
		protected sealed override void OnEdited (string path, string new_text)
		{
			base.OnEdited(path, new_text);
			
			// Ahora la fila ya no está más en modo edición
			this.treeView.FilaEnModoEdicion = TreeIter.Zero;
			
			// Obtengo el TreePath y el TreeIter de la fila modificada
			TreePath treePath = new TreePath(path);
			
			TreeIter iter;
			treeView.Model.GetIter(out iter, treePath);
			
			/* Obtengo el objeto ItemDocumentoGenericoDto de la fila modificada. */
			TItem item = (TItem) treeView.Model.GetValue(iter, 0);
			
			// Valido los datos que se editaron
			// TODO: Quizá se debería emitir un mensaje de error al no validarse los datos.
			if (!this.ValidarDatos(item, new_text))
				return;
			
			/* Modifico el valor del ItemGeneridoDto. Si el método ModificarItem
			 * retorna true, es decir, si realmente hubo una modificación, ejecuto el método
			 * ItemModificado que se encarga de realizar algunas operaciones más. */
			TreeViewColumn columnaADejarEnModoEdicion = this.ModificarItem(item, new_text);
			
			this.treeView.ItemModificado(iter, columnaADejarEnModoEdicion);
		}
		
		/// <summary>
		/// Realiza unas operaciones antes de pasar el control, como verificar si el
		/// CellEditables es un Entry (recomendado por la documentación).
		/// </summary>
		/// <param name="editable">
		/// A <see cref="Gtk.CellEditable"/>
		/// </param>
		/// <param name="path">
		/// A <see cref="System.String"/>
		/// </param>
		protected sealed override void OnEditingStarted (Gtk.CellEditable editable, string path)
		{
			base.OnEditingStarted(editable, path);
			
			/* Verifico que el tipo sea un Entry (la documentación oficial recomienda
			 * realizar esta verificación) */
			if (!(editable is Entry))
				return;
			
			Entry entry = (Entry)editable;
			
			// Obtengo el TreePath y el TreeIter de la fila donde se comenzó la edición
			TreePath treePath = new TreePath(path);
			
			TreeIter iter;
			treeView.Model.GetIter(out iter, treePath);
			
			/* Obtengo el objeto TItem de la fila. */
			TItem item = (TItem) treeView.Model.GetValue(iter, 0);
			
			// Indico al TreeView padre que esta fila está en modo edición
			this.treeView.FilaEnModoEdicion = iter;
			
			this.EdicionIniciada(entry, path, item);
		}
		
		protected override void OnEditingCanceled ()
		{
			base.OnEditingCanceled ();
			
			// Ahora la fila ya no está más en modo edición
			this.treeView.FilaEnModoEdicion = TreeIter.Zero;
		}
		
		/// <summary>
		/// Este método se encarga de habilitar o no la edición en un CellRenderer.
		/// </summary>
		/// <param name="evnt">
		/// A <see cref="Gdk.Event"/>
		/// </param>
		/// <param name="widget">
		/// A <see cref="Gtk.Widget"/>
		/// </param>
		/// <param name="path">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="background_area">
		/// A <see cref="Gdk.Rectangle"/>
		/// </param>
		/// <param name="cell_area">
		/// A <see cref="Gdk.Rectangle"/>
		/// </param>
		/// <param name="flags">
		/// A <see cref="Gtk.CellRendererState"/>
		/// </param>
		/// <returns>
		/// A <see cref="CellEditable"/>
		/// </returns>
		public override CellEditable StartEditing (Gdk.Event evnt, Gtk.Widget widget, string path, Gdk.Rectangle background_area, Gdk.Rectangle cell_area, Gtk.CellRendererState flags)
		{
			CellEditable cellEditable;
			
			// Pregunto al CellRenderer si está permitido comenzar la edición.
			if (this.ComenzarEdicion())
				cellEditable =
					base.StartEditing(evnt, widget, path, background_area, cell_area, flags);
			else
				cellEditable = null;
			
			return cellEditable;
		}
		#endregion
		
		#region Métodos abstractos
		/// <summary>
		/// Este método recibe el ItemFaturaDto de la fila que se editó y el nuevo valor.
		/// La función es validar los datos, devolviendo true o false según corresponda.
		/// </summary>
		/// <param name="item">
		/// A <see cref="ItemDocumentoGenericoDto"/>
		/// </param>
		/// <param name="nuevoValor">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected abstract bool ValidarDatos(TItem item, string nuevoValor);
		
		/// <summary>
		/// Esta función realiza efectivamente las modificaciones en el item.
		/// </summary>
		/// <param name="item">
		/// A <see cref="ItemDocumentoGenericoDto"/>
		/// </param>
		/// <param name="nuevoValor">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected abstract TreeViewColumn ModificarItem(TItem item, string nuevoValor);
		
		/// <summary>
		/// Al comenzar a editarse una celda, en este método se brinda la posibilidad de
		/// realizar algún setup. Por ejemplo en la descripción del producto se suscribe a
		/// algunos eventos del Entry para poder gestionar la petición de productos al servidor
		/// y cargar el EntryCompletion.
		/// </summary>
		/// <param name="entry">
		/// A <see cref="Entry"/>
		/// </param>
		/// <param name="path">
		/// A <see cref="System.String"/>
		/// </param>
		protected abstract void EdicionIniciada(Entry entry, string path, TItem item);
		
		/// <summary>
		/// Este método debe ser implementado por todas las subclases. Sirve para indicar si
		/// se debe comenzar con la edición de la celda o no. Por ejemplo, PrecioUnitarioCellRenderer
		/// devuelve true cuando el producto cargado es abstracto. En otro caso, no es posible
		/// cambiar el precio unitario.
		/// </summary>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected abstract bool ComenzarEdicion();
		#endregion
	}
}
