// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;

using Gtk;

using Stratos.ModeloDto;
using StratosClient.Core.Gui;

namespace StratosClient.Widgets.Level2
{
	/// <summary>
	/// Clase base para la lista de elementos interactivos. No es genérica para poder ser
	/// utilizada como un componente en Stetic.
	/// </summary>
	public abstract class ListaInteractivaBase : Gtk.TreeView, IListaInteractivaBaseVista
	{
		#region Eventos
		public delegate void ItemSeleccionadoHandler(TreeIter iterSeleccionado, ItemGenericoDto item);
		public event ItemSeleccionadoHandler ItemSeleccionado;
		
		public delegate void TvVacioHandler();
		public event TvVacioHandler TvVacio;
		
		public event System.Action ItemsCambiados;
		
		public event System.Action ItemQuitado;
		
		protected int limiteItems;
		protected bool forzarSetCursor;
		#endregion
		
		#region Constructores
		public ListaInteractivaBase () : base()
		{
			this.forzarSetCursor = false;
			
			this.Selection.Mode = SelectionMode.Browse;
			
			// Limite por defecto para la lista de items
			this.limiteItems = 14;
			
			this.Selection.Changed += OnTvItemsSelectionChanged;
		}
		#endregion
		
		#region Propiedades
		public int CantidadItems {
			get {
				// Si el último item es el item vacío, resto uno a la cantidad
				int cantidad = this.Model.IterNChildren();
				
				if (this.Model.IterNChildren() > 0) {
					ItemGenericoDto ultimoItem = this.ObtenerUltimoItem();
					if (this.ItemVacio(ultimoItem))
						cantidad--;
				}
				
				return cantidad;
			}
		}
		
		/// <summary>
		/// Devuelve la cantidad máxima permitida de items.
		/// </summary>
		public int LimiteItems {
			get { return this.limiteItems; }
			set { this.limiteItems = value; }
		}
		
		/// <summary>
		/// Indica si el TreeView se encuentra vacío.
		/// </summary>
		public bool Vacio {
			get {
				if (this.Model.IterNChildren() > 1)
					return false;
				
				/* Si contiene un item, verifico si éste es el item vacio */
				else if (this.Model.IterNChildren() == 1) {
					
					TreeIter iter;
					this.Model.GetIterFirst(out iter);

					if (this.ItemVacio((ItemGenericoDto)this.Model.GetValue(iter, 0)))
						return true;
				}
				
				return false;
			}
		}
		#endregion
		
		#region Métodos gestores de eventos
		protected void OnItemQuitado()
		{
			if (this.ItemQuitado != null)
				this.ItemQuitado();
		}
		
		protected void OnItemSeleccionado(TreeIter iter, ItemGenericoDto item)
		{
			if (this.ItemSeleccionado != null)
				this.ItemSeleccionado(iter, item);
		}
		
		protected void OnTvVacio()
		{
			// Lanzo también aquí el evento avisando que los items han cambiado
			this.OnItemsCambiados();
			
			if (this.TvVacio != null)
				this.TvVacio();
		}
		
		protected void OnItemsCambiados()
		{
			if (this.ItemsCambiados != null)
				this.ItemsCambiados();
		}
		
		protected virtual void OnTvItemsSelectionChanged (object o, EventArgs args)
		{
			TreeSelection seleccion = (TreeSelection) o;
			TreeIter iter;
			
			// Obtengo el item seleccionado
			seleccion.GetSelected(out iter);
			ItemGenericoDto itemSeleccionado = (ItemGenericoDto) this.Model.GetValue(iter, 0);
			
			if (itemSeleccionado == null)
				return;
			
			// Lanzo el evento, si hay suscripciones
			this.OnItemSeleccionado(iter, itemSeleccionado);
		}
		#endregion
		
		#region Métodos abstractos
		/// <summary>
		/// Verifica si el item es un item vacío.
		/// </summary>
		/// <param name="item">
		/// A <see cref="T"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public abstract bool ItemVacio(ItemGenericoDto item);
		
		/// <summary>
		/// Método para validar un item.
		/// </summary>
		/// <param name="item">
		/// A <see cref="ItemGenericoDto"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public abstract bool ItemValido(ItemGenericoDto item);
		
		/// <summary>
		/// Agrega un item vacío al final del modelo, para que el
		/// usuario ingrese uno nuevo.
		/// </summary>
		/// <returns>
		/// A <see cref="TreeIter"/>
		/// </returns>
		internal abstract TreeIter AgregarItemVacio();
		#endregion
		
		#region Métodos
		#region Método HabilitarSetCursor
		protected bool HabilitarSetCursor()
		{
			bool resultado = this.HabilitarSetCursorInterno();
			
			// Siempre que ejecuto la función HabilitarSetCursor, deshabilito
			// el forzado de SetCursor para la próxima vez.
			this.forzarSetCursor = false;
			
			return resultado;
		}
		
		/// <summary>
		/// Esta función indica está permitida la ejecución de la función SetCursor
		/// sobre el TreeView. Hay casos (por ejemplo, cuando el usuario presiona un botón
		/// del mouse) donde esta función no debe ser llamada, ya que puede dejar al TreeView
		/// </summary>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected bool HabilitarSetCursorInterno()
		{
			Gdk.Event ev = Gtk.Application.CurrentEvent;
			
			if (ev == null)
				return true;
			
			// Verifico si el evento fue KeyPress y si se presionó la tecla return
			else if (ev.Type == Gdk.EventType.KeyPress) {
				Gdk.EventKey evKey = (Gdk.EventKey)ev;
				
				if (evKey.Key == Gdk.Key.Return || evKey.Key == Gdk.Key.KP_Enter)
					return true;
			}
			
			else if (this.forzarSetCursor) {
				/* Inhabilito el forzar SetCursor una vez que la he activado. Es
				 * decir, sólo permito una entrada. Si son necesarias más, es
				 * necesario que se vuelva a asignar. */
				this.forzarSetCursor = false;
				
				return true;
			}
			
			return false;
		}
		#endregion
		
		internal void ColocarCeldaEnModoEdicion(TreePath path, TreeViewColumn columna)
		{
			if (!this.HabilitarSetCursor() || columna == null)
				return;
			
			this.GrabFocus();
			this.SetCursor(path, columna, true);
		}
		
		/// <summary>
		/// Devuelve true si ya se ha alcanzado el límite de items permitido.
		/// </summary>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected bool LimiteItemsAlcanzado()
		{
			int cantidadItems = this.Model.IterNChildren();
			
			return (cantidadItems >= this.LimiteItems);
		}
		
		public void AgregarItems(params ItemGenericoDto[] items)
		{
			ListStore modelo = (ListStore) this.Model;
			
			// Ordeno los items
			var itemsOrdenados =
				from item in items
				orderby item.orden ascending
				select item;
			
			// Agrego todos los items
			foreach (ItemGenericoDto item in itemsOrdenados)
				modelo.AppendValues(item);
			
			// Al final agrego el item vacio si el límite de items me lo permite
			if (!this.LimiteItemsAlcanzado())
				this.AgregarItemVacio();
			
			this.OnItemsCambiados();
		}
		
		/// <summary>
		/// Devuelve true si todos los items ingresados actualmente son válidos.
		/// </summary>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public bool ItemsValidos()
		{
			//HACK: Ver si es necesario recorrer todos los items.
			foreach (ItemGenericoDto item in this.Model.TraverseColumn<ItemGenericoDto>(0)) {	
				
				if (this.ItemVacio(item))
					continue;
				
				if (!this.ItemValido(item))
					return false;
			}
			
			return true;
		}
		
		/// <summary>
		/// Obtiene el último elemento de la lista. Devuelve un TreeIter.
		/// </summary>
		/// <returns>
		/// A <see cref="ItemGenericoDto"/>
		/// </returns>
		public TreeIter ObtenerUltimoIter()
		{
			TreeIter ultimoIter;
			int cantidadItems = this.Model.IterNChildren();
			this.Model.IterNthChild(out ultimoIter, cantidadItems - 1);
			
			return ultimoIter;
		}
		
		/// <summary>
		/// Obtiene el último ItemGenericoDto de la lista.
		/// </summary>
		/// <returns>
		/// A <see cref="ItemGenericoDto"/>
		/// </returns>
		public ItemGenericoDto ObtenerUltimoItem()
		{
			TreeIter ultimoIter = this.ObtenerUltimoIter();
			
			return (ItemGenericoDto) this.Model.GetValue(ultimoIter, 0);
		}
		
		/// <summary>
		/// Devuelve el objeto ItemGenerico actualmente seleccionado.
		/// </summary>
		/// <returns>
		/// A <see cref="T"/>
		/// </returns>
		public ItemGenericoDto ObtenerItemSeleccionado()
		{
			TreeIter iter;
			this.Selection.GetSelected(out iter);
			
			/* Obtengo el objeto ItemDocumentoGenericoDto de la fila modificada. */
			ItemGenericoDto item = (ItemGenericoDto) this.Model.GetValue(iter, 0);
			
			return item;
		}
		
		public ItemGenericoDto ObtenerItemSeleccionado(out TreeIter iterSeleccionado)
		{
			this.Selection.GetSelected(out iterSeleccionado);
			
			/* Obtengo el objeto ItemDocumentoGenericoDto de la fila modificada. */
			ItemGenericoDto item = (ItemGenericoDto) this.Model.GetValue(iterSeleccionado, 0);
			
			return item;
		}
		
		/// <summary>
		/// Elimina el item seleccionado y selecciona el siguiente.
		/// </summary>
		public void QuitarItemsSeleccionados()
		{
			// Si la lista está vacía, no hay nada por borrar
			if (this.Model.IterNChildren() == 0)
				return;
			
			ListStore modeloListStore = (ListStore) this.Model;
			
			// Obtengo el item seleccionado
			TreeIter iter;
			ItemGenericoDto item = this.ObtenerItemSeleccionado(out iter);
			
			// Si el item seleccionado es el vacío, no se puede borrar
			if (this.ItemVacio(item))
				return;
			
			// Obtengo el TreePath del siguiente item para dejarlo seleccionado luego
			// de eliminar el actual.
			TreePath pathSiguienteItem = this.Model.GetPath(iter);
			
			// Borro el item seleccionado
			modeloListStore.Remove(ref iter);
			
			if (this.Model.IterNChildren() == 0) {
				// Si no hay items en la lista, me limito a agregar el item vacío
				this.AgregarItemVacio();
			}
			else {
				// Obtengo el último elemtno y me aseguro que siempre exista un elemento
				// para colocar un nuevo item.
				item = this.ObtenerUltimoItem();
				
				if (!this.ItemVacio(item) && this.ItemValido(item))
					this.AgregarItemVacio();
				
				// Dejo seleccionada la fila siguiente a la recientemente eliminada.
				this.Selection.SelectPath(pathSiguienteItem);
			}
			
			// Si la lista queda vacía, lanzo el evento correspondiente
			if (this.Vacio)
				this.OnTvVacio();
			
			// Si hay suscripciones, lanzo el evento.
			this.OnItemsCambiados();
			this.OnItemQuitado();
		}
		
		/// <summary>
		/// Vacia el TreeView.
		/// </summary>
		public void Vaciar()
		{
			this.Vaciar(true);
		}
		
		/// <summary>
		/// Vacia el TreeView. Se puede indicar si se debe agregar un item vacío o no.
		/// </summary>
		/// <param name="agregarItemVacio">
		/// A <see cref="System.Boolean"/>
		/// </param>
		public void Vaciar(bool agregarItemVacio)
		{
			((ListStore)this.Model).Clear();
			
			if (agregarItemVacio)
				this.AgregarItemVacio();
			
			// Lanzo los eventos correspondientes correspondiente
			this.OnTvVacio();
			this.OnItemsCambiados();
			this.OnItemQuitado();
		}
		#endregion
	}
}
