// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

using Gtk;

using Stratos.Common;
using StratosClient.Core.Gui;
using Stratos.ModeloDto;
using StratosServer.Controladores;

namespace StratosClient.Widgets.Level2
{
	/// <summary>
	/// Clase que permite simplificar la utilización del un Gtk.TreeView que
	/// almacene los items de una factura. Se encarga de inicializar el widget
	/// y prepararlo para el manejo de los items de una factura.
	/// </summary>
	public class ListaItemsVenta : ListaItemsProducto
	{
		#region Campos
		private FacturaTipoVenta facturaTipo;
		#endregion
		
		#region Constructores
		public ListaItemsVenta() : base()
		{
			this.facturaTipo = FacturaTipoVenta.FacturaB;
			
			// Por defecto, una lista de items de venta tiene 32 como máximo
			this.limiteItems = 32;
		}
		#endregion
		
		#region Propiedades
		public FacturaTipoVenta FacturaTipo {
			get { return this.facturaTipo; }
			set {
				this.facturaTipo = value;
				
				ItemProductoDto item;
				
				// Itero por cada item y recalculo el precio unitario
				foreach (TreeIter iter in this.Model.Traverse()) {
					item = (ItemProductoDto) this.Model.GetValue(iter, 0);
					
					if (this.ItemVacio(item))
						continue;
					
					// Solo cambio el precio unitario de los productos que no son abstractos
					if (!item.productoAbstracto) {
						if (this.facturaTipo == FacturaTipoVenta.FacturaA)
							item.precioUnitario = item.productoImporteSinIva;
						else if (this.facturaTipo == FacturaTipoVenta.FacturaB)
							item.precioUnitario = item.productoImporteVenta;
					}
					
					this.ItemModificado(iter);
				}
				
				this.OnItemsCambiados();
			}
		}
		#endregion
		
		#region Métodos
		protected override bool PermitirCargaProducto (ProductoDto producto, out string mensajeCargaProductoNoPermitida)
		{
			mensajeCargaProductoNoPermitida = String.Empty;
			
			// Siempre permito la carga de cualquier producto.
			return true;
		}
		
		protected override TreeViewColumn SiguienteColumnaAEditar (TreeViewColumn columna)
		{
			ItemProductoDto itemProductoActual = this.ObtenerItemSeleccionadoGenerico();
			
			#region Acciones para productos abstractos
			if (!itemProductoActual.productoAbstracto)
				return null;
			
			/* Coloco en modo de edición la columna de descripción si este valor no ha
			 * sido cambiado del original. */
			if ((columna == this.tvcCodigoProducto || columna == this.tvcDescripcionProducto) &&
			    	itemProductoActual.descripcion == itemProductoActual.productoDescripcion)
				return this.tvcDescripcionProducto;
			
			/* Si estoy en la columna de código o descripción y la misma ha sido cambiada
			 * ya, entonces coloco el foco en la columna de precio unitario. */
			if ((columna == this.tvcCodigoProducto || columna == this.tvcDescripcionProducto))
				return this.tvcPrecioUnitarioProducto;
			#endregion
			
			return null;
		}
		
		protected override bool ItemValidoGenerico(ItemProductoDto item, out TreeViewColumn columnaInvalida)
		{
			// Verifico si el código del producto es correcto
			if (item.productoCodigo == null ||
			    item.productoCodigo.Trim().Equals(String.Empty)) {
				
				columnaInvalida = this.tvcCodigoProducto;
				
				return false;
			}
			
			// Verifico si la descripción del producto es correcta
			if (item.descripcion == null ||
			    item.descripcion.Trim().Equals(String.Empty)) {
				
				columnaInvalida = this.tvcDescripcionProducto;
				
				return false;
			}
			
			if (item.productoAbstracto &&
			    item.precioUnitario == 0.00) {
				
				//this.tvcPrecioUnitarioProducto.CellRenderers[0]
				columnaInvalida = this.tvcPrecioUnitarioProducto;
				
				return false;
			}
			
			// Verifico si la cantidad es correcta
			if (item.cantidad <= 0.00) {
				
				columnaInvalida = this.tvcCantidad;
				
				return false;
			}
			
			columnaInvalida = null;
			
			return true;
		}
		
		#region Carga de productos
		protected override void CargarDatosExtraProducto(ProductoDto producto, ItemProductoDto item)
		{
			// Aquí el importe hay que seleccionarlo según el tipo de factura
			if (this.FacturaTipo == FacturaTipoVenta.FacturaA)
				item.precioUnitario = item.productoImporteSinIva;
			else if (this.FacturaTipo == FacturaTipoVenta.FacturaB)
				item.precioUnitario = item.productoImporteVenta;
		}
		#endregion
		#endregion
	}
}
