// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;

using Gtk;

using Stratos.Common;
using Stratos.ModeloDto;
using StratosClient.Core.Gui;

namespace StratosClient.Widgets.Level2
{
	/// <summary>
	/// Clase base para mostrar listas de Documentos Genéricos.
	/// </summary>
	public class TvDocumentosGenericos<T> : ListaElementos<T>
		where T : DocumentoGenericoDto
	{
		#region Campos
		public TreeViewColumn tvcTipoDocumento;
		public TreeViewColumn tvcFecha;
		public TreeViewColumn tvcObservaciones;
		#endregion
		
		#region Constructores
		public TvDocumentosGenericos()
		{
			#region Columnas
			TreeViewColumn tvcColumna;
			CellRendererText crtColumnaRender;
			
			tvcColumna = new TreeViewColumn();
			tvcColumna.Title = "Tipo";
			crtColumnaRender = new CellRendererText();
			tvcColumna.PackStart(crtColumnaRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaRender, this.RenderTipoDocumento);
			this.AppendColumn(tvcColumna);
			this.tvcTipoDocumento = tvcColumna;
			
			tvcColumna = new TreeViewColumn();
			tvcColumna.Title = "Fecha";
			crtColumnaRender = new CellRendererText();
			//crtColumnaRender.Xalign = 0.5f;
			tvcColumna.PackStart(crtColumnaRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaRender, this.RenderFecha);
			this.AppendColumn(tvcColumna);
			this.tvcFecha = tvcColumna;
			
			tvcColumna = new TreeViewColumn();
			tvcColumna.Title = "Observaciones";
			crtColumnaRender = new CellRendererText();
			//crtColumnaRender.Xalign = 0.5f;
			tvcColumna.PackStart(crtColumnaRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaRender, this.RenderObservaciones);
			this.AppendColumn(tvcColumna);
			this.tvcObservaciones = tvcColumna;
			#endregion
		}
		#endregion
		
		#region Funciones de renderizado del TreeView
		/// <summary>
		/// Este método realiza operaciones de renderizado básicas para todas las columnas.
		/// Por ejemplo, si una FacturaDto está anulada, tacha el texto en todas las columnas
		/// y cambia el color de fondo de la fila.
		/// </summary>
		/// <param name="column">
		/// A <see cref="TreeViewColumn"/>
		/// </param>
		/// <param name="cell">
		/// A <see cref="CellRenderer"/>
		/// </param>
		/// <param name="model">
		/// A <see cref="TreeModel"/>
		/// </param>
		/// <param name="iter">
		/// A <see cref="TreeIter"/>
		/// </param>
		/// <returns>
		/// A <see cref="DocumentoGenericoDto"/>
		/// </returns>
		protected virtual T RenderBase(TreeViewColumn column,
		                                                  CellRenderer cell,
		                                                  TreeModel model,
		                                                  TreeIter iter)
		{
			T doc = (T) model.GetValue(iter, 0);
			return doc;
		}
		
		protected virtual void RenderTipoDocumento(TreeViewColumn column,
		                                           CellRenderer cell,
		                                           TreeModel model,
		                                           TreeIter iter)
		{
			DocumentoGenericoDto doc = this.RenderBase(column, cell, model, iter);
			
			// Obtengo el tipo del documento de acuerdo a su tipo CLR, y quito
			// la parte Dto del final
			string valor = StratosUtil.SplitCamelString(doc.GetType().Name);
				
			if (valor.ToLower().EndsWith("dto"))
				valor = valor.Substring(0, valor.Length - 3).Trim();
			
			// Para los tipos de documentos específicos
			if (doc is FacturaGenericaDto) {
				FacturaGenericaDto factura = (FacturaGenericaDto) doc;
				valor += (factura.tipo == FacturaTipoVenta.FacturaA ? " A" : " B");
			}
			else if (doc is FacturaCompraDto) {
				FacturaCompraDto facturaCompra = (FacturaCompraDto) doc;
				valor += (facturaCompra.tipo == FacturaTipoCompra.FacturaA ? " A" : " C");
			}
			
			(cell as CellRendererText).Text = valor;
		}
		
		protected virtual void RenderFecha(TreeViewColumn column,
		                                    CellRenderer cell,
		                                    TreeModel model,
		                                    TreeIter iter)
		{
			DocumentoGenericoDto doc = this.RenderBase(column, cell, model, iter);
			string valor = doc != null ?
				(new DateTime(doc.fechaTicks)).ToShortDateString() :
				String.Empty;
			
			(cell as CellRendererText).Text = valor;
		}
		
		protected virtual void RenderObservaciones(TreeViewColumn column,
		                                           CellRenderer cell,
		                                           TreeModel model,
		                                           TreeIter iter)
		{
			DocumentoGenericoDto doc = this.RenderBase(column, cell, model, iter);
			string valor = doc != null ? doc.observaciones : String.Empty;
			
			// Solo muestro la primer linea de las observaciones
			int hasta = valor.IndexOf("\n");
			if (hasta >= 0)
				valor = valor.Substring(0, hasta) + ((hasta > 0) ? " ..." : String.Empty);
			
			(cell as CellRendererText).Text = valor;
		}
		#endregion
		
		#region ListaElementos implementation
		protected override bool Iguales (T t1, T t2)
		{
			return (t1.id == t2.id);
		}
		#endregion
	}
}
