// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Gtk;

using Stratos.ModeloDto;

namespace StratosClient.Widgets.Level2
{
	/// <summary>
	/// Clase para mostrar listados de elementos tipo FaturaDto.
	/// </summary>
	public class TvFacturas : TvDocumentosVenta<FacturaDto>
	{
		#region Campos
		public TreeViewColumn tvcCodigo;
		public TreeViewColumn tvcNumero;
		#endregion
		
		#region Constructores
		public TvFacturas() : base()
		{
			#region Columnas
			TreeViewColumn tvcColumna;
			CellRendererText crtColumnaRender;
			
			tvcColumna = new TreeViewColumn();
			tvcColumna.Title = "Número";
			crtColumnaRender = new CellRendererText();
			crtColumnaRender.Xalign = 1.0f;
			tvcColumna.PackStart(crtColumnaRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaRender, this.RenderNumero);
			this.InsertColumn(tvcColumna, 2);
			this.tvcNumero = tvcColumna;
			
			tvcColumna = new TreeViewColumn();
			tvcColumna.Title = "Código";
			crtColumnaRender = new CellRendererText();
			crtColumnaRender.Xalign = 1.0f;
			tvcColumna.PackStart(crtColumnaRender, true);
			tvcColumna.SetCellDataFunc(crtColumnaRender, this.RenderCodigo);
			this.InsertColumn(tvcColumna, 2);
			this.tvcCodigo = tvcColumna;
			#endregion
		}
		#endregion
		
		#region Métodos de renderizado
		/// <summary>
		/// Este método realiza operaciones de renderizado básicas para todas las columnas.
		/// Por ejemplo, si una FacturaDto está anulada, tacha el texto en todas las columnas
		/// y cambia el color de fondo de la fila.
		/// </summary>
		/// <param name="column">
		/// A <see cref="TreeViewColumn"/>
		/// </param>
		/// <param name="cell">
		/// A <see cref="CellRenderer"/>
		/// </param>
		/// <param name="model">
		/// A <see cref="TreeModel"/>
		/// </param>
		/// <param name="iter">
		/// A <see cref="TreeIter"/>
		/// </param>
		/// <returns>
		/// A <see cref="DocumentoGenericoDto"/>
		/// </returns>
		protected override FacturaDto RenderBase(TreeViewColumn column,
		                                                   CellRenderer cell,
		                                                   TreeModel model,
		                                                   TreeIter iter)
		{
			FacturaDto facturaDto = (FacturaDto) model.GetValue(iter, 0);
			if (facturaDto == null) return null;
			
			// Si la factura es histórica...
			if (facturaDto.historica) {
				CellRendererText cellText = cell as CellRendererText;
				
				cell.CellBackground = "lightgray";
				
				if (cellText != null)
					cellText.Strikethrough = facturaDto.anulada;
			}
			
			// Si la FacturaDto está anulada, entonces cambio el texto y el fondo de la fila
			else if (facturaDto.anulada) {
				CellRendererText cellText = cell as CellRendererText;
				
				cell.CellBackground = facturaDto.valida ? "wheat" : "red";
				
				if (cellText != null)
					cellText.Strikethrough = true;
			}
			
			// Si la FacturaDto no está anulada, entonces vuelvo
			// a colocar todos los valores por defecto.
			else {
				CellRendererText cellText = cell as CellRendererText;
				
				cell.CellBackground = null;
				
				if (cellText != null)
					cellText.Strikethrough = false;
			}
			
			return facturaDto;
		}
		
		protected override void RenderTipoDocumento (TreeViewColumn column,
		                                             CellRenderer cell,
		                                             TreeModel model,
		                                             TreeIter iter)
		{
			FacturaDto factura = this.RenderBase(column, cell, model, iter);
			
			string valor = factura != null ? "Factura" : String.Empty;
			
			// Me fijo si no se trata de una NotaCredito
			if (factura is NotaCreditoDto)
				valor = "Nota de crédito";
			
			if (factura != null) {
				// Me fijo en el tipo de Factura
				if (factura.tipo == FacturaTipoVenta.FacturaA)
					valor += " A";
				else
					valor += " B";
			}
			
			(cell as CellRendererText).Text = valor;
		}
		
		protected virtual void RenderCodigo(TreeViewColumn column,
		                                    CellRenderer cell,
		                                    TreeModel model,
		                                    TreeIter iter)
		{
			FacturaDto factura = this.RenderBase(column, cell, model, iter);
			string valor = factura != null ? factura.codigo.ToString("D4") : String.Empty;
			
			(cell as CellRendererText).Text = valor;
		}
		
		protected virtual void RenderNumero(TreeViewColumn column,
		                                    CellRenderer cell,
		                                    TreeModel model,
		                                    TreeIter iter)
		{
			FacturaDto factura = this.RenderBase(column, cell, model, iter);
			string valor = factura != null ? factura.numero.ToString() : String.Empty;
			
			(cell as CellRendererText).Text = valor;
		}
		
		protected override void RenderObservaciones (TreeViewColumn column,
		                                             CellRenderer cell,
		                                             TreeModel model,
		                                             TreeIter iter)
		{
			FacturaDto factura = this.RenderBase(column, cell, model, iter);
			
			// Si no se trata de una NotaCredito, entonces salgo
			NotaCreditoDto notaCredito = factura as NotaCreditoDto;
			if (notaCredito == null) {
				base.RenderObservaciones(column, cell, model, iter);
				return;
			}
			
			// Si es una NotaCredito, entonces pongo en las observaciones la factura asociada
			FacturaDto facturaAsociada = notaCredito.facturaAsociada;
			string valor = String.Empty;
			
			// Puede que la nota de crédito no tenga una factura asociada, por
			// las facturas de migración del sistema viejo.
			if (facturaAsociada != null)
				valor = "Factura asociada: " + facturaAsociada.codigo.ToString("D4") + "-" +
					facturaAsociada.numero.ToString();
			
			(cell as CellRendererText).Text = valor;
		}
		#endregion
	}
}
