// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;
using System.Linq;
using System.Reflection;
using System.Collections;
using System.Collections.Generic;

using Ice;
using log4net;
using NHibernate.Criterion;
using Castle.ActiveRecord.Framework;
using Castle.ActiveRecord.Queries;
using Castle.ActiveRecord;

using StratosServer.Core;
using Stratos.Common;
using Stratos.Common.Excepciones;
using StratosServer.Controladores;
using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using Stratos.QueryObjects;

namespace StratosServer.Clientes
{
	/// <summary>
	/// Clase encargada de gestionar las operaciones relacionadas con los clientes.
	/// </summary>
	public class ControladorClientesI : ControladorClientesDisp_
	{
		#region Campos
		public const string CONSUMIDOR_FINAL_RAZON_SOCIAL = "CONSUMIDOR FINAL - CONTADO";
		public const string CLIENTE_INVALIDO_RAZON_SOCIAL = "## CLIENTE INVALIDO ##";
		
		public static int CANTIDAD_MAXIMA_RESULTADOS_MOVIMIENTOS = 200;
		#endregion
		
		#region Métodos públicos a la Terminal
		public override Ice.Object[] BuscarHQL(string consulta, ConsultaPaginable consultaPaginable,
		                                       Current current__)
		{
			ClienteDto[] resultados =
				StratosServerUtil
					.BuscarHQLGenerico<ClienteDto,Cliente>(consulta, consultaPaginable);
			
			return resultados;
		}
		
		public override Ice.Object[] BuscarQueryObjects(QueryObject queryObject,
		                                                ConsultaPaginable consultaPaginable,
		                                                Current current__)
		{
			// Valido que el QueryObject sea de tipo ClienteDtoQuery
			if (!(queryObject is ClienteDtoQuery))
				throw new StratosException("El QueryObject no es de tipo ClienteDtoQuery");
			
			ClienteDto[] resultado =
				StratosServerUtil
					.BuscarQueryObjects<Cliente,ClienteDto>(queryObject,
					                                        consultaPaginable);
			
			return resultado;
		}
		
		public override Ice.Object Refrescar(Ice.Object objeto, Ice.Current current__)
		{
			// Verifico que el objeto por refrescar sea del tipo correcto
			if (!(objeto is ClienteDto)) {
				throw new StratosException("El objeto a refrescar no es de tipo ClienteDto");
			}
			
			ClienteDto clienteDto = objeto as ClienteDto;
			
			// Refresco los datos del cliente
			return StratosServerUtil
				.Refrescar<Cliente,ClienteDto>(clienteDto, clienteDto.codigo);
		}
		
		public override ClienteDto[] ObtenerClienteConsumidorFinalGenerico (Current current__)
		{
			Cliente clienteConsumidorFinal = this.ObtenerClienteConsumidorFinal();
			ClienteDto clienteDto = ConvertidorDto.ToDto(clienteConsumidorFinal);
			
			return new ClienteDto[] { clienteDto };
		}
		
		public override Ice.Object Guardar (Ice.Object cliente, Current current__)
		{
			// Verifico que la instancia se de tipo ClienteDto
			ClienteDto clienteDto =
				StratosServerUtil.TipoCorrecto<ClienteDto>(cliente);
			
			return this.Guardar(clienteDto, false);
		}

		public override PersonaDto[] BusquedaExactaPorCodigo (long codigoCliente, Current current__)
		{
			using (SessionScope s = new SessionScope(FlushAction.Never)) {
				Cliente cliente = null;
				try {
					cliente = Cliente.FindOne(Restrictions.Eq("Codigo", codigoCliente));
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				Cliente[] resultado;
				
				// Si hay un resultado, y este no está eliminado, creo un array con el cliente
				if (cliente != null && !cliente.Eliminado)
					resultado = new Cliente[] { cliente };
				else
					resultado = new Cliente[0];
				
				return (PersonaDto[]) ConvertidorDto.ToDto(resultado);
			}
		}
		
		public override PersonaDto[] BusquedaExactaPorRazonSocial (string razonSocial, Current current__)
		{
			// Verifico que la consulta sea válida
			if (razonSocial.Trim().Equals(String.Empty))
				throw new ParametrosInvalidosException("La cadena de búsqueda está vacía");
			
			if (razonSocial.Length < 3)
				throw new ParametrosInvalidosException("La cadena de búsqueda debe tener al menos 3 caracteres");
			
			using (SessionScope s = new SessionScope(FlushAction.Never)) {
				Cliente[] clientes = new Cliente[0];
				
				try {
					clientes = Cliente.FindAll(Expression.Eq("RazonSocial", razonSocial));
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				return (PersonaDto[]) ConvertidorDto.ToDto(clientes);
			}
		}
		
		public override PersonaDto[] ConsultaPorRazonSocial (string consulta, Current current__)
		{
			// Verifico que la consulta sea válida
			if (consulta.Trim().Equals(String.Empty))
				throw new StratosException("La cadena de búsqueda está vacía");
			
			if (consulta.Length < 3)
				throw new ParametrosInvalidosException("La cadena de búsqueda debe tener al menos 3 caracteres");
			
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				string[] palabras = consulta.Split(' ');
				
				// Creo la lista de criterios de búsqueda, para que incluya todas las palabras
				// de la consulta, sin importar el orden
				List<ICriterion> criteriosBusqueda = new List<ICriterion>();
				
				foreach (string p in palabras)
					criteriosBusqueda.Add(Expression.Like("RazonSocial", p, MatchMode.Anywhere)
					                      .IgnoreCase());
				
				// Agrego una condición para que no devuelva los clientes eliminados
				criteriosBusqueda.Add(Expression.Eq("Eliminado", false));
				
				Cliente[] clientes = null;
				try {
					clientes = Cliente.FindAll(criteriosBusqueda.ToArray());
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				return (PersonaDto[]) ConvertidorDto.ToDto(clientes);
			}
		}
		
		public override DocumentoGenericoDto[] ObtenerMovimientos(long idCliente,
		                                                          long fechaDesdeTicks,
		                                                          long fechaHastaTicks,
		                                                          out bool resultadoRecortado,
		                                                          Current current__)
		{
			resultadoRecortado = false;
			
			using (SessionScope session = new SessionScope(FlushAction.Never)) {
				
				List<DocumentoGenericoDto> listaMovimientos = new List<DocumentoGenericoDto>();
				
				// Me aseguro que el cliente especificado exista en la base de datos
				Cliente cliente = Cliente.TryFind(idCliente);
				
				if (cliente == null)
					throw new ObjetoInexistenteException("El cliente no existe en la base de datos");
				
				// Creo una lista de criterios de búsqueda
				ICriterion[] criterios = new ICriterion[] {
					Restrictions.Eq("Cliente", cliente),
					Restrictions.Ge("Fecha", new DateTime(fechaDesdeTicks)),
					Restrictions.Le("Fecha", new DateTime(fechaHastaTicks))
				};
				
				// Creo consultas separadas para cada tipo de documento
				DetachedCriteria consultaFacturas =
					DetachedCriteria.For<Factura>()
						.AddOrder(Order.Asc("Fecha"));
				
				DetachedCriteria consultaRecibos =
					DetachedCriteria.For<Recibo>()
						.AddOrder(Order.Asc("Fecha"));
				
				foreach (ICriterion unCriterio in criterios) {
					consultaFacturas.Add(unCriterio);
					consultaRecibos.Add(unCriterio);
				}
				
				/* Cuento la cantidad real de resultados. Si sobrepasa el máximo, entonces
				 * activo la variable 'resultadosRecortados' para avisar al usuario que se
				 * han quitado filas del resultado. */
				CountQuery cantidadFacturas = new CountQuery(typeof(Factura), criterios);
				CountQuery cantidadRecibos = new CountQuery(typeof(Recibo), criterios);
				
				resultadoRecortado =
					( ((int) ActiveRecordMediator.ExecuteQuery(cantidadFacturas)) +
						((int) ActiveRecordMediator.ExecuteQuery(cantidadRecibos)) )
							> CANTIDAD_MAXIMA_RESULTADOS_MOVIMIENTOS;
				
				/* Realizo las consultas y las agrego a la lista de movimientos. */
				Factura[] facturas =
					Factura.FindAll(consultaFacturas.SetMaxResults(CANTIDAD_MAXIMA_RESULTADOS_MOVIMIENTOS));
				
				Recibo[] recibos =
					Recibo.FindAll(consultaRecibos.SetMaxResults(CANTIDAD_MAXIMA_RESULTADOS_MOVIMIENTOS));
				
				listaMovimientos.AddRange(ConvertidorDto.ToDto<Factura,FacturaDto>(facturas));
				listaMovimientos.AddRange(ConvertidorDto.ToDto<Recibo,ReciboDto>(recibos));
				
				// Ordeno la lista de movimientos y recorto el resultado hasta el máximo permitido.
				var listaMovimientosOrdenada =
					(from doc in listaMovimientos
					 orderby doc.fechaTicks ascending
					 select doc).Take(CANTIDAD_MAXIMA_RESULTADOS_MOVIMIENTOS);
				
				return listaMovimientosOrdenada.ToArray();
			}
		}
		
		/// <summary>
		/// Dado un ClienteDto, lo marca como 'Eliminado' y persiste el cambio en la base de datos.
		/// Un cliente nunca se elimina físicamente de la base de datos, solo se lo marca como
		/// 'Eliminado'.
		/// </summary>
		/// <param name="clienteAEliminar">
		/// A <see cref="ClienteDto"/>
		/// </param>
		/// <param name="current__">
		/// A <see cref="Current"/>
		/// </param>
		public override void MarcarComoEliminado (ClienteDto clienteAEliminar, Current current__)
		{
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<Cliente>(clienteAEliminar.codigo,
					                          clienteAEliminar.version,
					                          transactionScope);
				
				// Obtengo el cliente de la base de datos
				Cliente cliente = Cliente.Find(clienteAEliminar.codigo);

				// Chequeo que el cliente no esté ya eliminado
				if (!cliente.Eliminado) {
					cliente.Eliminado = true;
					
					StratosServerUtil.ActualizarObjeto<Cliente>(cliente, transactionScope);
				}
				else {
					throw new ClienteEliminadoException("El cliente que intenta eliminar ya " +
					                                    "esta marcado como 'Eliminado'.");
				}
			}	
		}
		
		public override Ice.Object Modificar (Ice.Object cliente, Current current__)
		{
			// Verifico que la instancia se de tipo ClienteDto
			ClienteDto clienteDto =
				StratosServerUtil.TipoCorrecto<ClienteDto>(cliente);
			
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<Cliente>(clienteDto.codigo,
					                          clienteDto.version,
					                          transactionScope);
				
				// Obtengo el cliente de la base de datos
				Cliente clienteModificado = ConvertidorDto.FromDto(clienteDto);
				
				// Chequeo que el cliente no esté marcado como 'Eliminado'
				if (clienteModificado.Eliminado) {
					throw new ClienteEliminadoException("El cliente que intenta modificar " +
					                                    "esta marcado como 'Eliminado'.");
				}
				
				// Actualizo el cliente en la base de datos
				StratosServerUtil.ActualizarObjeto<Cliente>(clienteModificado, transactionScope);
				
				// Retorno una instancia Dto actualizada
				return ConvertidorDto.ToDto(clienteModificado);
			}	
		}
		
		#region Cuenta corriente
		public override DocumentoCreditoDto[] ObtenerDocumentosCredito (long idCliente, Current current__)
		{
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				DetachedCriteria consulta =
					DetachedCriteria.For<DocumentoCredito>()
						.CreateAlias("Cliente", "cli")
						.Add(Restrictions.Eq("cli.Codigo", idCliente));
				
				// Realizo la búsqueda
				DocumentoCredito[] resultados =
					DocumentoCredito.FindAll(consulta);
				
				return ConvertidorDto.ToDto<DocumentoCredito,DocumentoCreditoDto>(resultados);
			}
		}
		
		public override DocumentoDebitoDto[] ObtenerDocumentosDebito (long idCliente, Current current__)
		{
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				DetachedCriteria consulta =
					DetachedCriteria.For<DocumentoDebito>()
						.CreateAlias("Cliente", "cli")
						.Add(Restrictions.Eq("cli.Codigo", idCliente));
				
				// Realizo la búsqueda
				DocumentoDebito[] resultados =
					DocumentoDebito.FindAll(consulta);
				
				return ConvertidorDto.ToDto<DocumentoDebito,DocumentoDebitoDto>(resultados);
			}
		}
		
		public override void Imputar (DocumentoCreditoDto[] documentosCreditoDto, DocumentoDebitoDto[] documentosDebitoDto, Current current__)
		{
			// Verifico que se hayan pasado documentos de crédito y débito
			// para imputar.
			if (documentosCreditoDto.Length == 0 || documentosDebitoDto.Length == 0)
				throw new ParametrosInvalidosException("Se deben especificar documentos de crédito y débito para imputar");
			
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				DocumentoCredito[] documentosCredito =
					DocumentoCredito
						.FindAll(Order.Asc("Id"),
						         Restrictions.In("Id",
						                         (from dc in documentosCreditoDto
						                          select dc.id).ToArray()));
				
				DocumentoDebito[] documentosDebito =
					DocumentoDebito
						.FindAll(Order.Asc("Id"),
								 Restrictions.In("Id",
						                         (from dd in documentosDebitoDto
						                          select dd.id).ToArray()));
				
				DocumentoCredito docCred;
				DocumentoDebito docDeb;
				
				int indiceCredito = 0;
				int indiceDebito = 0;
				
				while (indiceCredito < documentosCredito.Length && indiceDebito < documentosDebito.Length) {
					docCred = documentosCredito[indiceCredito];
					docDeb = documentosDebito[indiceDebito];
					
					if (docCred.ImportePendiente > docDeb.ImportePendiente) {
						docCred.ImportePendiente -= docDeb.ImportePendiente;
						docDeb.Delete();
						indiceDebito++;
					}
					else if (docCred.ImportePendiente < docDeb.ImportePendiente) {
						docDeb.ImportePendiente -= docCred.ImportePendiente;
						docCred.Delete();
						indiceCredito++;
					}
					else {
						docCred.Delete();
						docDeb.Delete();
						indiceCredito++;
						indiceDebito++;
					}
				}
			}
		}
		
		public override void PasarFacturasACuentaCorriente (long[] idsFacturas, Current current__)
		{
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Levanto de la base de datos todas las facturas indicadas en el parámetro
				Factura[] facturas =
					Factura.FindAll(Restrictions.In("Id", idsFacturas));
				
				// Levanto también todos los documentos de débito asociados a esas facturas
				DocumentoDebito[] documentosDebitoAsociados =
					DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
					                        Restrictions.In("IdDocumento", idsFacturas));
				
				/* Antes de hacer algo verifico que ninguna factura tenga ya un documento 
				 * de débito asociado. Con que una lo tenga, ya cancelo la operación. */
				if (documentosDebitoAsociados.Length > 0) {
					
					StringBuilder mensaje =
						new StringBuilder("Las siguientes facturas ya están en cuenta corriente:");
					
					foreach (DocumentoDebito docDeb in documentosDebitoAsociados) {
						mensaje.AppendLine();
						
						mensaje.AppendFormat("{0}-{1}",
						                     docDeb.CodigoFactura.ToString("D4"),
						                     docDeb.NumeroFactura.ToString("D8"));
					}
					
					throw new ParametrosInvalidosException(mensaje.ToString());
				}
				
				/* Tambien chequeo que ninguno de los ids pasados como parametro correspondan a una 
				 * Nota de Credito. Si es asi, lanzo una excepcion y cancelo la operacion. */
				foreach (Factura fac in facturas)
				{
					if (fac is NotaCredito)
						throw new ParametrosInvalidosException("Solo se pueden pasar Facturas a Cuenta Corriente. " +
						 	"Asegurese de seleccionar solo Facturas.");
				}
				
				// Finalmente, por cada factura creo un documento de débito correspondiente.
				foreach (Factura fac in facturas)
				{
					DocumentoDebito.CrearDebito(fac);
				}
			}
		}
		
		public override ClienteDto RecuperarClienteEliminado (long codigo, Current current__)
		{
			using (TransactionScope transaccion = new TransactionScope(TransactionMode.Inherits)) {

				// Creo la lista de criterios de búsqueda para obtener el cliente por su código
				List<ICriterion> criteriosBusqueda = new List<ICriterion>();
				criteriosBusqueda.Add(Expression.Eq("Codigo", codigo));
				
				// Realizo la consulta
				Cliente[] clientesEncontrados = null;
				try {
					clientesEncontrados = Cliente.FindAll(criteriosBusqueda.ToArray());
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				// Si no hay clientes recuperados lanzo una excepción
				if (clientesEncontrados.Length == 0)
					throw new ObjetoInexistenteException("No existe un cliente con el codigo " +
						codigo + " ingresado al sistema.");
				
				Cliente cliente = clientesEncontrados[0];
				
				// Chequeo que el cliente esté marcado como Eliminado
				if (!cliente.Eliminado)
					throw new ObjetoInvalidoException("El cliente que intenta recuperar no " +
						"está marcado como eliminado.");
				
				// Actualizo la propiedad 'Eliminado'
				cliente.Eliminado = false;
				
				// Actualizo el objeto en la base de datos
				StratosServerUtil.ActualizarObjeto<Cliente>(cliente, transaccion);
				
				// Retorno el cliente actualizado
				return ConvertidorDto.ToDto(cliente);
			}
		}
		#endregion
		#endregion
		
		#region Métodos internos para el Servidor		
		/// <summary>
		/// Devuelve un Cliente para utilizar en una Factura inválida y anulada. Esto se utiliza
		/// para sincronizar los números de factura.
		/// </summary>
		/// <returns>
		/// A <see cref="ClienteDto"/>
		/// </returns>
		public Cliente ObtenerClienteInvalido(TipoResponsable tipoResponsable)
		{
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				Cliente clienteInvalido =
					Cliente.FindOne(Expression.Like("RazonSocial", CLIENTE_INVALIDO_RAZON_SOCIAL),
					                Expression.Eq("TipoResponsable", tipoResponsable));
				
				// Si no existe un cliente genérico, lo creamos
				if (clienteInvalido == null) {
					clienteInvalido = new Cliente();
					clienteInvalido.RazonSocial = CLIENTE_INVALIDO_RAZON_SOCIAL;
					clienteInvalido.Valido = false;
					
					clienteInvalido.Domicilio = new Domicilio();
					clienteInvalido.Domicilio.Provincia = ProvinciaDto.NoEspecificada;
					
					clienteInvalido.TipoResponsable = tipoResponsable;
					
					clienteInvalido.PorcentajeBonificacion = 0.00m;
					
					// Agrego el cliente genérico a la base de datos
					StratosServerUtil.CrearObjeto<Cliente>(clienteInvalido, transactionScope);
				}
				
				return clienteInvalido;
			}
		}
		
		public ClienteDto Guardar (ClienteDto cliente, bool altaAutomatica)
		{
			// Utilizo una transacción
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				Cliente clienteReal = ConvertidorDto.FromDto(cliente);
				clienteReal.AltaAutomatica = altaAutomatica;
				
				StratosServerUtil.CrearObjeto<Cliente>(clienteReal, transactionScope);
				
				return ConvertidorDto.ToDto(clienteReal);
			}
		}
		
		public Cliente ObtenerClienteConsumidorFinal()
		{
			Cliente clienteConsumidorFinal = null;
			
			using (SessionScope s = new SessionScope()) {
				// Realizo la búsqueda del cliente consumidor final, si es que existe.
				Cliente[] clientes =
					Cliente.FindAll(Restrictions.Eq("RazonSocial",
					                                CONSUMIDOR_FINAL_RAZON_SOCIAL));
				
				// Si no existe un cliente genérico, lo creamos
				// Para eso utilizo directamente el método Guardar con un ClienteDto
				if (clientes.Length == 0) {
					ClienteDto clienteGenerico = new ClienteDto();
					clienteGenerico.razonSocial = CONSUMIDOR_FINAL_RAZON_SOCIAL;
					
					clienteGenerico.domicilio = new DomicilioDto();
					clienteGenerico.domicilio.provincia = ProvinciaDto.NoEspecificada;
					
					clienteGenerico.tipoRes = TipoResponsable.ConsumidorFinal;
					clienteGenerico.porcentajeBonificacion = 0.00;
					
					// Agrego el cliente genérico a la base de datos
					ClienteDto clienteGuardado = (ClienteDto) this.Guardar(clienteGenerico);
					
					clienteConsumidorFinal = Cliente.Find(clienteGuardado.codigo);
				}
				else
					clienteConsumidorFinal = clientes[0];
				
				return clienteConsumidorFinal;
			}
		}
		#endregion
	}
}
