// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Text;
using System.Collections;
using System.Collections.Generic;

using Ice;
using log4net;
using NHibernate.Criterion;
using Castle.ActiveRecord.Framework;
using Castle.ActiveRecord.Queries;
using Castle.ActiveRecord;

using StratosServer.Core;
using StratosServer.Controladores;
using Stratos.QueryObjects;
using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using Stratos.Common;
using Stratos.Common.Excepciones;

namespace StratosServer.Productos
{
	/// <summary>
	/// Clase encargada de gestionar las operaciones relacionadas con los productos.
	/// </summary>
	public class ControladorProductosI : ControladorProductosDisp_
	{
		#region Campos
		public const string PRODUCTO_INVALIDO_CODIGO = "#PRO-AUTO#";
		public const string PRODUCTO_INVALIDO_DESCRIPCION = "## PRODUCTO AUTOMÁTICO ##";
		
		public const string RUBRO_INVALIDO_NOMBRE = "## RUBRO AUTOMÁTICO ##";
		public const string RUBRO_INVALIDO_DESCRIPCION = "Rubro automático de uso interno";
		#endregion
		
		#region Implementación ControladorProductosDisp_
		#region Consultas
		public override Ice.Object[] BuscarHQL(string consulta,
		                                       ConsultaPaginable consultaPaginable,
		                                       Current current__)
		{
			ProductoDto[] resultados =
				StratosServerUtil
					.BuscarHQLGenerico<ProductoDto,Producto>(consulta, consultaPaginable);
			
			return resultados;
		}
		
		public override Ice.Object[] BuscarQueryObjects(QueryObject queryObject,
		                                                ConsultaPaginable consultaPaginable,
		                                                Current current__)
		{
			// Valido que el QueryObject sea de tipo ProductoDto
			if (!(queryObject is ProductoDtoQuery))
				throw new StratosException("El QueryObject no es de tipo ProductoDto");
			
			ProductoDto[] resultado =
				StratosServerUtil
					.BuscarQueryObjects<Producto,ProductoDto>(queryObject, consultaPaginable);
			
			return resultado;
		}
		
		public override Ice.Object Refrescar(Ice.Object objeto, Ice.Current current__)
		{
			if (!(objeto is ProductoDto))
				throw new ObjetoInvalidoException("El objeto a refrescar debe ser del tipo " +
					"ProductoDto o PreFacturaDto");
				
			ProductoDto productoDto = objeto as ProductoDto;
			
			return StratosServerUtil
				.Refrescar<Producto,ProductoDto>(productoDto,
				                                 productoDto.id);				
		}
		
		/// <summary>
		/// Busca productos según la descripción.
		/// </summary>
		/// <param name="descripcion">
		/// Debe contener todas las cadenas buscadas. Todos los productos que contengan en su
		/// descripción todas las palabras que indica esta variable (sin importar el case) serán
		/// devueltos.
		/// </param>
		/// <param name="current__">
		/// A <see cref="Ice.Current"/>
		/// </param>
		/// <returns>
		/// A <see cref="ProductoDto"/>
		/// </returns>
		public override ProductoDto[] BuscarPorDescripcion (string descripcion, Ice.Current current__)
		{
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				// Verifico que la consulta sea válida
				if (descripcion.Trim().Equals(String.Empty))
					throw new StratosException("La cadena de búsqueda está vacía");
				
				/* FIXME: Tener cuidado de las búsquedas que retornan demasiados resultados, como por
				 * ejemplo con '': retorna todos los productos. Se podría colocar un limit cuando
				 * se hace este tipo de consulta. Abajo hay un ejemplo comentado. */
				string[] palabras = descripcion.Split(' ');
				
				// Creo la lista de criterios de búsqueda, para que incluya todas las palabras
				// de la consulta, sin importar el orden
				List<ICriterion> criteriosBusqueda = new List<ICriterion>();
				foreach (string p in palabras)
					criteriosBusqueda.Add(Expression.Like("Descripcion", p, MatchMode.Anywhere)
					                      .IgnoreCase());
				
				// Filtro los productos eliminados
				criteriosBusqueda.Add(this.CriterioEvitarProductosEliminados());
				
				// Realizo la consulta
				Producto[] productos = null;
				try {
					// Además ordeno los resultados por código de forma ascendente
					Order ordenPorCodigo = Order.Asc("Codigo");
					
					productos = Producto.FindAll(ordenPorCodigo, criteriosBusqueda.ToArray());
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				// Una forma de limitar los resultados:
	//			Producto[] productos = Producto.FindAll(
	//				DetachedCriteria.For<Producto>()
	//					.Add(Expression
	//							.Like("Descripcion", consulta, MatchMode.Anywhere)
	//							.IgnoreCase())
	//					.SetMaxResults(50)
	//			);
				
				return ConvertidorDto.ToDto(productos);
			}
		}
		
		/// <summary>
		/// Busca productos según el código.
		/// </summary>
		/// <param name="codigo">
		/// El código buscado debe ser exactamente el código del producto en la base
		/// de datos.
		/// </param>
		/// <param name="current__">
		/// A <see cref="Ice.Current"/>
		/// </param>
		/// <returns>
		/// A <see cref="ProductoDto"/>
		/// </returns>
		public override ProductoDto[] BuscarPorCodigo (string codigo, Ice.Current current__)
		{
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				
				// Verifico que la consulta sea válida
				if (codigo.Trim().Equals(String.Empty))
					throw new StratosException("La cadena de búsqueda está vacía");
				
				/* FIXME: Tener cuidado de las búsquedas que retornan demasiados resultados, como por
				 * ejemplo con '': retorna todos los productos. Se podría colocar un limit cuando
				 * se hace este tipo de consulta. Abajo hay un ejemplo comentado. */
				string[] palabras = codigo.Split(' ');
				
				// Creo la lista de criterios de búsqueda, para que incluya todas las palabras
				// de la consulta, sin importar el orden
				List<ICriterion> criteriosBusqueda = new List<ICriterion>();
				foreach (string p in palabras)
					criteriosBusqueda.Add(Expression.Like("Codigo", p, MatchMode.Anywhere)
					                      .IgnoreCase());
				
				// Filtro los productos eliminados
				criteriosBusqueda.Add(this.CriterioEvitarProductosEliminados());
				
				// Realizo la consulta
				Producto[] productos = null;
				try {
					// Además ordeno los resultados por código de forma ascendente
					Order ordenPorCodigo = Order.Asc("Codigo");
					
					productos = Producto.FindAll(criteriosBusqueda.ToArray());
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				// FIXME: Configurar AutoMapper para que al convertir una instancia que sea nula,
				// la deje nula. Si es null el resultado, sabemos que no encontró nada.
	//			if (producto == null)
	//				return null;
				
				return ConvertidorDto.ToDto(productos);
			}
		}
		
		public override RubroDto[] ObtenerRubros (Ice.Current current__)
		{
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				Rubro[] resultado = null;
				
				try {
					resultado = Rubro.FindAll(Restrictions.IsNull("RubroPadre"));
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
						
				return ConvertidorDto.ToDto(resultado);
			}
		}
		
		public override RubroDto[] ObtenerSubrubros (RubroDto rubroPadre, Ice.Current current__)
		{
			if (rubroPadre == null)
				return new RubroDto[0];
			
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				Rubro[] resultado = null;
				
				try {
					resultado = Rubro.FindAllByProperty("RubroPadre", ConvertidorDto.FromDto(rubroPadre));
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				return ConvertidorDto.ToDto(resultado);
			}
		}
		
		/// <summary>
		/// Devuelve 'true' si el código del producto pasado como parametro está disponible
		/// (no hay un producto ingresado en la base de datos con el mismo código). En caso
		/// de haber un producto con el mismo código ingresado, setea en el parametro out
		/// 'productoEliminado' si este está marcado como eliminado o no.
		/// </summary>
		/// <param name="codigo">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="productoEliminado">
		/// A <see cref="System.Boolean"/>
		/// </param>
		/// <param name="current__">
		/// A <see cref="Current"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public override bool CodigoDisponible (string codigo, out bool productoEliminado, Current current__)
		{
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				
				// Verifico que la consulta sea válida
				if (codigo.Trim().Equals(String.Empty))
					throw new ObjetoInvalidoException("El código del producto es vacío");
				
				// Creo la lista de criterios de búsqueda, para que incluya todas las palabras
				// de la consulta, sin importar el orden
				List<ICriterion> criteriosBusqueda = new List<ICriterion>();
				criteriosBusqueda.Add(Expression.Like("Codigo", codigo.Trim(), MatchMode.Exact).IgnoreCase());
				
				// Realizo la consulta
				Producto[] productosEncontrados = null;
				try {
					productosEncontrados = Producto.FindAll(criteriosBusqueda.ToArray());
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				// Si no hay productos recuperados retorno true
				if (productosEncontrados.Length == 0) {
					productoEliminado = false;
					return true;
				}
				
				Producto producto = productosEncontrados[0];
				
				// Seteo el parametro out para saber si el producto
				// encontrado está eliminado o no y retorno
				productoEliminado = producto.Eliminado;
				
				return false;
			}
		}
		
		/// <summary>
		/// Recupera el producto con el codigo pasado como parametro, si este está
		/// marcado como 'Eliminado'. Retorna el producto actualizado.
		/// </summary>
		/// <param name="codigo">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="current__">
		/// A <see cref="Current"/>
		/// </param>
		/// <returns>
		/// A <see cref="ProductoDto"/>
		/// </returns>
		public override ProductoDto RecuperarProductoEliminado (string codigo, Current current__)
		{
			using (TransactionScope transaccion = new TransactionScope(TransactionMode.Inherits)) {
				
				// Verifico que la consulta sea válida
				if (codigo.Trim().Equals(String.Empty))
					throw new ObjetoInvalidoException("El código del producto es vacío");
				
				// Creo la lista de criterios de búsqueda, para que incluya todas las palabras
				// de la consulta, sin importar el orden
				List<ICriterion> criteriosBusqueda = new List<ICriterion>();
				criteriosBusqueda.Add(Expression.Like("Codigo", codigo.Trim(), MatchMode.Exact).IgnoreCase());
				
				// Realizo la consulta
				Producto[] productosEncontrados = null;
				try {
					productosEncontrados = Producto.FindAll(criteriosBusqueda.ToArray());
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				// Si no hay productos recuperados lanzo una excepción
				if (productosEncontrados.Length == 0)
					throw new ObjetoInexistenteException("No existe un producto con el codigo " +
						codigo.ToUpper() + " ingresado al sistema.");
				
				Producto producto = productosEncontrados[0];
				
				// Chequeo que el producto esté marcado como Eliminado
				if (!producto.Eliminado)
					throw new ObjetoInvalidoException("El producto que intenta recuperar no " +
						"está marcado como eliminado.");
				
				// Actualizo la propiedad 'Eliminado'
				producto.Eliminado = false;
				
				// Actualizo el objeto en la base de datos
				StratosServerUtil.ActualizarObjeto<Producto>(producto, transaccion);
				
				// Retorno el producto actualizado
				return ConvertidorDto.ToDto(producto);
			}
		}
		#endregion
		
		#region Alta/Baja/Modificaciones
		public override void MarcarComoEliminado (ProductoDto productoAEliminar, Current current__)
		{
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<Producto>(productoAEliminar.id,
					                           productoAEliminar.version,
					                           transactionScope);
				
				// Obtengo el producto de la base de datos
				Producto producto = Producto.Find(productoAEliminar.id);

				// Chequeo que el producto no esté ya eliminado
				if (!producto.Eliminado) {
					producto.Eliminado = true;
					
					StratosServerUtil.ActualizarObjeto<Producto>(producto, transactionScope);
				}
				else {
					throw new ProductoEliminadoException("El producto que intenta eliminar ya " +
					                                     "esta marcado como 'Eliminado'.");
				}
			}	
		}
		
		public override Ice.Object Modificar (Ice.Object producto, Current current__)
		{
			// Verifico que la instancia se de tipo ClienteDto
			ProductoDto productoDto =
				StratosServerUtil.TipoCorrecto<ProductoDto>(producto);
			
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<Producto>(productoDto.id,
					                           productoDto.version,
					                           transactionScope);
				
				// Obtengo el producto de la base de datos
				Producto productoModificado = ConvertidorDto.FromDto(productoDto);

				// Chequeo que el producto no esté ya eliminado
				if (productoModificado.Eliminado) {
					throw new ProductoEliminadoException("El producto que intenta modificar está " +
					                                     "marcado como 'Eliminado'.");
				
				}
				
				// Actualizo el producto en la base de datos
				StratosServerUtil.ActualizarObjeto<Producto>(productoModificado, transactionScope);
				
				return ConvertidorDto.ToDto(productoModificado);
			}
		}
		
		public override Ice.Object Guardar (Ice.Object producto, Current current__)
		{
			// Verifico que la instancia se de tipo ClienteDto
			ProductoDto productoDto =
				StratosServerUtil.TipoCorrecto<ProductoDto>(producto);
			
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				// Convierto de ProductoDto a Producto
				Producto nuevoProducto = ConvertidorDto.FromDto(productoDto);	
				
				// Creo el objeto
				StratosServerUtil.CrearObjeto<Producto>(nuevoProducto, transactionScope);
				
				// Devuelvo una instancia Dto actualizada
				return ConvertidorDto.ToDto(nuevoProducto);
			}
		}
		
		public override void ModificarPrecios (long[] ids, AccionModificacionPrecio accion, Ice.Current current__)
		{
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				// Primero verifico que todos los productos indicados para modificar existan en
				// la base de datos.
				ScalarQuery<long> productosIndicadosCountQuery =
					new ScalarQuery<long>(typeof(Producto),
					                          @"select count(*)
												from Producto p
												where p.Id in (:ids)");
				
				productosIndicadosCountQuery.SetParameterList("ids", ids);
				
				long cantidadProductosIndicadosQueExisten =
					productosIndicadosCountQuery.Execute();
				
				// Si las cantidades no coinciden, le informo al usuario.
				if (ids.Length != cantidadProductosIndicadosQueExisten) {
					transactionScope.VoteRollBack();
					
					StratosException excepcion = new StratosException();
					excepcion.mensaje =
						"Uno o más de los productos seleccionados no fueron encontrados. " +
						"Actualice su lista de productos e inténtelo nuevamente";
					
					throw excepcion;
				}
				
				#region Creación de la consulta con StringBuilder
				/* Armo una consulta UPDATE con HQL para modificar todos los productos en batch. */
				
				// Creo la consulta inicial
				StringBuilder hqlUpdatePrecio = new StringBuilder();
				
				// Agrego el tipo de modificación, según la acción
				hqlUpdatePrecio.AppendFormat("update versioned Producto p set p.ImporteSinIva = round({0}, 2)",
				                             this.GenerarHQLModificacionPrecio(accion));
				
				hqlUpdatePrecio.AppendFormat(" where p.Id in (:ids)");
				#endregion
				
				// Creo la consulta personalizada para actualizar los productos
				CustomUpdateQuery<Producto> productoUpdate =
					new CustomUpdateQuery<Producto> (hqlUpdatePrecio.ToString());
				
				productoUpdate.SetearListaDeParametros("ids", ids);
				
				// Ejecuto la consulta
				Producto.ExecuteQuery(productoUpdate);
				
				// Ejecuto un commit de la transacción
				transactionScope.VoteCommit();
			}
		}
		#endregion
		#endregion
		
		#region Métodos internos para el Servidor
		protected ICriterion CriterioEvitarProductosEliminados()
		{
			return Restrictions.Eq("Eliminado", false);
		}
		
		/// <summary>
		/// Dada una acción de tipo AccionModificacionPrecio, genera código HQL
		/// para modificar el ImporteSinIva de los productos, de acuerdo al tipo
		/// de acción (porcentual o importe fijo).
		/// </summary>
		/// <param name="accion">
		/// A <see cref="AccionModificacionPrecio"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		private string GenerarHQLModificacionPrecio(AccionModificacionPrecio accion)
		{
			StringBuilder hql = new StringBuilder("p.ImporteSinIva ");
			
			if (accion.accion == TipoAccion.Aumentar)
				hql.Append(" + ");
			else if (accion.accion == TipoAccion.Disminuir)
				hql.Append(" - ");
			
			// Calculo la diferencia a aplicar, segun el tipo de cambio
			if (accion.cambioTipo == TipoCambio.Porcentual)
				hql.AppendFormat("cast(p.ImporteSinIva * {0} as decimal)",
				                 StratosUtil.DoubleToString(accion.cantidad / 100.0));
			
			else if (accion.cambioTipo == TipoCambio.PrecioFijo)
				hql.AppendFormat("{0}", StratosUtil.DoubleToString(accion.cantidad));
			
			// Retorno la diferencia
			return hql.ToString();
		}
		
		/// <summary>
		/// Devuelve un Producto inválido (inexistente para el usuario). Se utiliza para dar de
		/// alta una Factura inválida (para sincronizar los números de factura).
		/// </summary>
		/// <returns>
		/// A <see cref="Producto"/>
		/// </returns>
		public Producto ObtenerProductoInvalido()
		{
			using (SessionScope sessionScope = new SessionScope()) {
				
				// Primero busco el producto inválido, por si ya no existe
				Producto resultadoProductoInvalido =
					Producto.FindOne(Expression.Eq("Codigo", PRODUCTO_INVALIDO_CODIGO),
					                 Expression.Eq("Descripcion", PRODUCTO_INVALIDO_DESCRIPCION));
				
				// Si no hay un Producto inválido creado, lo creo
				if (resultadoProductoInvalido == null) {
					
					// Necesito algún rubro. Primero busco todos los rubros padres
					// existentes, y uso cualquiera de ellos, si es que hay alguno
					Rubro rubro;
					Rubro[] rubros = Rubro.FindAll(Restrictions.IsNull("RubroPadre"));
					
					if (rubros.Length > 0)
						rubro = rubros[0];
					// Si no existen rubros ingresados, creo un rubro inválido
					else {
						rubro =
							Rubro.FindOne(Expression.Eq("Nombre", RUBRO_INVALIDO_NOMBRE),
							              Expression.Eq("Descripcion", RUBRO_INVALIDO_DESCRIPCION));
						
						// Si no existe el rubro inválido lo creo
						if (rubro == null) {
							rubro = new Rubro {
								Nombre = RUBRO_INVALIDO_NOMBRE,
								Descripcion = RUBRO_INVALIDO_DESCRIPCION
							};
							
							rubro.Create();
						}
					}
					
					// Ahora creo el Producto inválido
					resultadoProductoInvalido = new Producto {
						Codigo = PRODUCTO_INVALIDO_CODIGO,
						Descripcion = PRODUCTO_INVALIDO_DESCRIPCION,
						ImporteSinIva = 1.00m,
						PorcentajeIva = 21m,
						PrecioCosto = 0.50m,
						Rubro = rubro,
						Unidad = UnidadProducto.Unidad,
						Notas = "Este es un Producto inválido que no debe ser utilizado"
					};
					
					resultadoProductoInvalido.Create();
				}
				
				return resultadoProductoInvalido;
			}
		}
		#endregion
	}
}
