// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

using log4net;
using NHibernate.Criterion;
using Castle.ActiveRecord;

using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using StratosServer.Core;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;
using Stratos.QueryObjects;


namespace StratosServer.Proveedores
{
	public class ControladorProveedoresI : ControladorProveedoresDisp_
	{
		public override Ice.Object[] BuscarHQL (string consultaHQL, ConsultaPaginable consPaginable, Ice.Current current__)
		{
			ProveedorDto[] resultados =
				StratosServerUtil
					.BuscarHQLGenerico<ProveedorDto, Proveedor>(consultaHQL, consPaginable);
			
			return resultados;
		}
		
		public override Ice.Object[] BuscarQueryObjects (Stratos.QueryObjects.QueryObject queryObject, ConsultaPaginable consPaginable, Ice.Current current__)
		{
			// Valido que el QueryObject sea de tipo ProveedorDtoQuery
			if (!(queryObject is ProveedorDtoQuery))
				throw new StratosException("El QueryObject no es de tipo ProveedorDtoQuery");
			
			ProveedorDto[] resultado =
				StratosServerUtil
					.BuscarQueryObjects<Proveedor,ProveedorDto>(queryObject,
					                                        	consPaginable);
			
			return resultado;
		}
		
		public override Ice.Object Refrescar (Ice.Object objeto, Ice.Current current__)
		{
			// Verifico que el objeto por refrescar sea del tipo correcto
			if (!(objeto is ProveedorDto)) {
				throw new StratosException("El objeto a refrescar no es de tipo ProveedorDto");
			}
			
			ProveedorDto proveedorDto = objeto as ProveedorDto;
			
			// Refresco los datos del proveedor
			return StratosServerUtil
				.Refrescar<Proveedor,ProveedorDto>(proveedorDto, proveedorDto.codigo);
		}
		
		public override Ice.Object Guardar (Ice.Object proveedor, Ice.Current current__)
		{
			// Utilizo una transacción
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Verifico que la instancia se de tipo ProveedorDto
				ProveedorDto proveedorDto =
					StratosServerUtil.TipoCorrecto<ProveedorDto>(proveedor);
				
				Proveedor nuevoProveedor = ConvertidorDto.FromDto(proveedorDto);
				
				StratosServerUtil.CrearObjeto<Proveedor>(nuevoProveedor, transactionScope);
				
				return ConvertidorDto.ToDto(nuevoProveedor);
			}
		}
		
		public override PersonaDto[] BusquedaExactaPorCodigo (long codigo, Ice.Current current__)
		{
			using (SessionScope s = new SessionScope(FlushAction.Never)) {
				Proveedor proveedor = null;
				try {
					
					proveedor = Proveedor.FindOne(Restrictions.Eq("Codigo", codigo));
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				Proveedor[] resultado;
				
				// Si hay un resultado, y este no está eliminado, creo un array con el proveedor
				if (proveedor != null && !proveedor.Eliminado)
					resultado = new Proveedor[] { proveedor };
				else
					resultado = new Proveedor[0];
				
				return (PersonaDto[]) ConvertidorDto.ToDto(resultado);
			}
		}
		
		public override PersonaDto[] BusquedaExactaPorRazonSocial (string razonSocial, Ice.Current current__)
		{
			// Verifico que la consulta sea válida
			if (razonSocial.Trim().Equals(String.Empty))
				throw new ParametrosInvalidosException("La cadena de búsqueda está vacía");
			
			if (razonSocial.Length < 3)
				throw new ParametrosInvalidosException("La cadena de búsqueda debe tener al menos 3 caracteres");
			
			using (SessionScope s = new SessionScope(FlushAction.Never)) {
				Proveedor[] proveedores = new Proveedor[0];
				
				try {
					proveedores = Proveedor.FindAll(Expression.Eq("RazonSocial", razonSocial));
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				return (PersonaDto[]) ConvertidorDto.ToDto(proveedores);
			}
		}
		
		public override PersonaDto[] ConsultaPorRazonSocial (string consulta, Ice.Current current__)
		{
			// Verifico que la consulta sea válida
			if (consulta.Trim().Equals(String.Empty))
				throw new ParametrosInvalidosException("La cadena de búsqueda está vacía");
			
			if (consulta.Length < 3)
				throw new ParametrosInvalidosException("La cadena de búsqueda debe tener al menos 3 caracteres");
			
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				string[] palabras = consulta.Split(' ');
				
				// Creo la lista de criterios de búsqueda, para que incluya todas las palabras
				// de la consulta, sin importar el orden
				List<ICriterion> criteriosBusqueda = new List<ICriterion>();
				
				foreach (string p in palabras)
					criteriosBusqueda.Add(Expression.Like("RazonSocial", p, MatchMode.Anywhere)
					                      .IgnoreCase());
				
				// Agrego una condición para que no devuelva los proveedores eliminados
				criteriosBusqueda.Add(Expression.Eq("Eliminado", false));
				
				Proveedor[] proveedores = null;
				try {
					proveedores = Proveedor.FindAll(criteriosBusqueda.ToArray());
				}
				catch (System.Exception ex) {
					StratosServerUtil.TratarExcepcion(ex);
				}
				
				return (PersonaDto[]) ConvertidorDto.ToDto(proveedores);
			}
		}
		
		public override Ice.Object Modificar (Ice.Object proveedor, Ice.Current current__)
		{
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Verifico que la instancia se de tipo ProveedorDto
				ProveedorDto proveedorDto =
					StratosServerUtil.TipoCorrecto<ProveedorDto>(proveedor);
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<Proveedor>(proveedorDto.codigo,
					                            proveedorDto.version,
					                            transactionScope);
				
				// Obtengo el proveedor de la base de datos
				Proveedor proveedorModificado = ConvertidorDto.FromDto(proveedorDto);
				
				// Chequeo que el proveedor no esté marcado como 'Eliminado'
				if (proveedorModificado.Eliminado) {
					throw new StratosException("El proveedor que intenta modificar " +
					                           "esta marcado como 'Eliminado'.");
				}
				
				// Actualizo el producto en la base de datos
				StratosServerUtil.ActualizarObjeto<Proveedor>(proveedorModificado, transactionScope);
				
				// Retorno una instancia Dto actualizada
				return ConvertidorDto.ToDto(proveedorModificado);
			}	
		}
	}
}
