// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;

using log4net;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Queries;
using NHibernate;
using NHibernate.Criterion;

using Stratos.ModeloObjetos;
using StratosServer.Controladores;
using StratosServer.Core;
using Stratos.Common;
using Stratos.Common.Excepciones;

namespace StratosServer.Reportes
{
	public class ControladorReportesI : ControladorReportesDisp_
	{
		#region Campos
		private const int IVA_VENTAS_FACTURAS_POR_PAGINA = 95;
		private const int IVA_COMPRAS_FACTURAS_POR_PAGINA = 50;
		
		private string archivoTemplateVentas;
		private string archivoTemplateCompras;
		
		protected delegate TWrapper[] ProcesadorFacturas<T,TWrapper>(T[] facturas, out decimal[] totales);
		#endregion
		
		#region Constructores
		public ControladorReportesI()
		{
			// Armo el path por defecto del template de IVA Ventas
			string directorioAddin = "Addins" + System.IO.Path.DirectorySeparatorChar +
				"Reportes" + System.IO.Path.DirectorySeparatorChar;
			
			this.archivoTemplateVentas = directorioAddin + "templateReporteIvaVentas.vm";
			
			// Idem Iva Compras
			this.archivoTemplateCompras = directorioAddin + "templateReporteIvaCompras.vm";
		}
		
		public ControladorReportesI(string archivoTemplateVentas, string archivoTemplateCompras)
		{
			this.archivoTemplateVentas = archivoTemplateVentas;
			this.archivoTemplateCompras = archivoTemplateCompras;
		}
		#endregion
		
		#region Ventas
		public override string GenerarReporteIvaVentas (long fechaInicialTicks,
		                                                long fechaFinalTicks,
		                                                string sistemaOperativo,
		                                                Ice.Current current__)
		{
			return
				this.GenerarReporte<Factura,FacturaIvaVentasWrapper>("IVA Ventas",
			                                 this.archivoTemplateVentas,
			                                 fechaInicialTicks,
			                                 fechaFinalTicks,
			                                 this.ConsultarFacturasVentas,
				                             this.ProcesarFacturasVenta,
				                             IVA_VENTAS_FACTURAS_POR_PAGINA,
				                             sistemaOperativo);
		}
		
		public override string GenerarResumenIvaVentas(long fechaInicialTicks, long fechaFinalTicks,
		                                      string sistemaOperativo, Ice.Current current__)
		{
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				DateTime fechaIncial = new DateTime(fechaInicialTicks);
				DateTime fechaFinal = new DateTime(fechaFinalTicks);
				
				// Chequeo que las fechas sean válidas
				this.ValidarFechas(fechaIncial, fechaFinal);
				
				// Realizo la consulta con el rango de fechas pasado como parámetro
				Factura[] resultado = this.RealizarConsultaResumenIvaVentas(fechaIncial, fechaFinal);
				
				if (resultado.Length == 0)
					throw new ObjetoInexistenteException("No existen facturas para el rango de fechas " +
						"especificado.");
	
				decimal[] totales;
				
				// Proceso las facturas que recuperé en la consulta
				ResumenSucursal[] resumenSucursales = this.ProcesarResumenIvaVentas(resultado, out totales);
				
				/**** Creación del string HTML ****/
				
				// Armo el path directorio donde se encuentra el Addin
				string directorioAddin = "Addins" + System.IO.Path.DirectorySeparatorChar +
					"Reportes" + System.IO.Path.DirectorySeparatorChar;
				
				// Armo el path al archivo del template
				string archivoTemplate = directorioAddin + "templateResumenIvaVentas.vm";
				
				// Creo el objeto ImpresionBorrador, que tiene un método para crear el HTML
				GeneradorDocumento impresion = new GeneradorDocumento(archivoTemplate, directorioAddin);
				
				// Agrego los objetos que necesito en el template
				impresion.AgregarObjeto("fechaDesde", fechaIncial);
				impresion.AgregarObjeto("fechaHasta", fechaFinal);
				impresion.AgregarObjeto("resumenSucursales", resumenSucursales);
				impresion.AgregarObjeto("totales", totales);
				impresion.AgregarObjeto("sistema_operativo", sistemaOperativo);
				
				// Retorno el string HTML
				return impresion.CrearHTML();
			}
		}
		
		public override void MarcarComoHistoricaFacturaVenta (long fechaInicialTicks, long fechaFinalTicks, Ice.Current current__)
		{
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				DateTime fechaInicial = new DateTime(fechaInicialTicks);
				DateTime fechaFinal = new DateTime(fechaFinalTicks);
				
				this.ValidarFechas(fechaInicial, fechaFinal);
				
				string cadenaConsulta = "update versioned Factura f set f.Historica = :historica" +
					" where f.Fecha between :fechaInicial and :fechaFinal" +
					" and f.Historica != :historica";
				
				CustomUpdateQuery<Factura> consulta = new CustomUpdateQuery<Factura>(cadenaConsulta);
				
				consulta.SetearParametros("historica", true);
				consulta.SetearParametros("fechaInicial", fechaInicial);
				consulta.SetearParametros("fechaFinal", fechaFinal);
				
				Factura.ExecuteQuery(consulta);
			}
		}
		
		#region IVA Ventas
		public Factura[] ConsultarFacturasVentas(DateTime fechaInicial, DateTime fechaFinal)
		{
			string cadenaConsulta =
				"from Factura f " +
				"	left join fetch f.Cliente c " +
				"where f.Fecha between :fechaInicial and :fechaFinal " +
				"and f.Valida = :valida " +
				"order by f.Fecha asc, f.Tipo asc, f.Codigo asc, f.Numero asc";
			
			SimpleQuery<Factura> consulta = new SimpleQuery<Factura> (cadenaConsulta);
			
			consulta.SetParameter("fechaInicial", fechaInicial);
			consulta.SetParameter("fechaFinal", fechaFinal);
			consulta.SetParameter("valida", true);
			
			return consulta.Execute();
		}
		
		public FacturaIvaVentasWrapper[] ProcesarFacturasVenta(Factura[] facturas, out decimal[] totales)
		{
			totales = new decimal[4];
			
			List<FacturaIvaVentasWrapper> facturasProcesadas = new List<FacturaIvaVentasWrapper>();
			
			decimal totalExento = 0.00m;
			decimal totalGravado = 0.00m;
			decimal totalIvaInsc = 0.00m;
			decimal totalTotal = 0.00m;
			
			FacturaIvaVentasWrapper temp;
			
			// Por cada factura pasada como parámetro
			foreach (Factura f in facturas) {
				
				// Creo un wrapper para la factura
				temp = new FacturaIvaVentasWrapper(f);
				
				// Sumo cada uno de los importes a los totales parciales
				// TODO: el exento retorna siempre 0.00, por eso no lo calculo.
				// totalExento += temp.ExentoNumerico;
				if (f is NotaCredito) {
					totalGravado -= temp.GravadoNumerico;
					totalIvaInsc -= temp.IvaInscNumerico;
					totalTotal -= temp.TotalNumerico;	
				}
				else {
					totalGravado += temp.GravadoNumerico;
					totalIvaInsc += temp.IvaInscNumerico;
					totalTotal += temp.TotalNumerico;	
				}
				
				// Agrego el wrapper a la lista de facturas procesadas
				facturasProcesadas.Add(temp);
			}
			
			// Seteo el array de totales con los resultados
			totales[0] = totalExento;
			totales[1] = totalGravado;
			totales[2] = totalIvaInsc;
			totales[3] = totalTotal;
		
			// Retorno las facturas procesadas
			return facturasProcesadas.ToArray();
		}
		#endregion
		
		#region Resumen IVA Ventas
		public Factura[] RealizarConsultaResumenIvaVentas (DateTime fechaInicial, DateTime fechaFinal)
		{
			string cadenaConsulta = "from Factura f " +
				"	left join fetch f.Items " +
				"where f.Fecha between :fechaInicial and :fechaFinal " +
				"and f.Valida = :valida " +
				"and f.Anulada = :anulada";
			
			SimpleQuery<Factura> consulta = new SimpleQuery<Factura>(cadenaConsulta);
			
			consulta.SetParameter("fechaInicial", fechaInicial);
			consulta.SetParameter("fechaFinal", fechaFinal);
			consulta.SetParameter("valida", true);
			consulta.SetParameter("anulada", false);
			
			// Seteo un transformer para que no haya resultados duplicados
			consulta.SetResultTransformer(new NHibernate.Transform.DistinctRootEntityResultTransformer());
			
			return consulta.Execute();
		}
		
		public ResumenSucursal[] ProcesarResumenIvaVentas(Factura[] facturas, out decimal[] totales)
		{
			// Sucursal Avellaneda
			ResumenSucursal sucursalAvellaneda = new ResumenSucursal();
			sucursalAvellaneda.Descripcion = "Avellaneda: códigos 0001 y 0004";
			
			// Sucursal Transporte
			ResumenSucursal sucursalTransporte = new ResumenSucursal();
			sucursalTransporte.Descripcion = "Transporte: códigos 0003, 0005, 0006, 0007 y 0008";
			
			// Sucursal Malabrigo
			ResumenSucursal sucursalMalabrigo = new ResumenSucursal();
			sucursalMalabrigo.Descripcion = "Malabrigo: código 0009";
			
			// Para cada factura veo, según su código, a la sucursal
			// que pertenece y la agrego.
			foreach (Factura f in facturas) {
				
				// Sucursal Avellaneda
				if ((f.Codigo == 1) || (f.Codigo == 4))
					sucursalAvellaneda.AgregarFactura(f);
				
				// Sucursal Malabrigo
				else if (f.Codigo == 9)
					sucursalMalabrigo.AgregarFactura(f);
				
				// Transporte
				else
					sucursalTransporte.AgregarFactura(f);
			}

			List<ResumenSucursal> listaResumenSucursales = new List<ResumenSucursal>();
			
			listaResumenSucursales.Add(sucursalAvellaneda);
			listaResumenSucursales.Add(sucursalTransporte);
			listaResumenSucursales.Add(sucursalMalabrigo);
			
			// Calculo los totales globales
			totales = new decimal[5] {0.00m, 0.00m, 0.00m, 0.00m, 0.00m};
			
			foreach (ResumenSucursal rs in listaResumenSucursales) {
				totales[0] += rs.TotalComercio;
				totales[1] += rs.TotalIndustria;
				totales[2] += rs.TotalNoDefinido;
				totales[3] += rs.TotalIva;
				totales[4] += rs.TotalTotal;
			}
			
			return listaResumenSucursales.ToArray();
		}
		#endregion
		#endregion
		
		#region Compras
		public override string GenerarReporteIvaCompras (long fechaInicialTicks, long fechaFinalTicks, string sistemaOperativo, Ice.Current current__)
		{
			return
				this.GenerarReporte<FacturaCompra,FacturaIvaComprasWrapper>("IVA Compras",
			                                 this.archivoTemplateCompras,
			                                 fechaInicialTicks,
			                                 fechaFinalTicks,
			                                 this.ConsultarFacturasCompras,
				                             this.ProcesarFacturasCompra,
				                             IVA_COMPRAS_FACTURAS_POR_PAGINA,
				                             sistemaOperativo);
		}

		public override void MarcarComoHistoricaFacturaCompra (long fechaInicialTicks, long fechaFinalTicks, Ice.Current current__)
		{
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				DateTime fechaInicial = new DateTime(fechaInicialTicks);
				DateTime fechaFinal = new DateTime(fechaFinalTicks);
				
				this.ValidarFechas(fechaInicial, fechaFinal);
				
				string cadenaConsulta = "update versioned FacturaCompra f set f.Historica = :historica" +
					" where f.FechaImputacion between :fechaInicial and :fechaFinal" +
					" and f.Historica != :historica";
				
				CustomUpdateQuery<FacturaCompra> consulta = new CustomUpdateQuery<FacturaCompra>(cadenaConsulta);
				
				consulta.SetearParametros("historica", true);
				consulta.SetearParametros("fechaInicial", fechaInicial);
				consulta.SetearParametros("fechaFinal", fechaFinal);
				
				FacturaCompra.ExecuteQuery(consulta);
			}
		}
		
		#region IVA Compras
		/// <summary>
		/// Realiza una consulta de las facturas de compras dadas las fechas iniciales y finales.
		/// </summary>
		/// <param name="fechaInicial">
		/// A <see cref="DateTime"/>
		/// </param>
		/// <param name="fechaFinal">
		/// A <see cref="DateTime"/>
		/// </param>
		/// <returns>
		/// A <see cref="FacturaCompra[]"/>
		/// </returns>
		public FacturaCompra[] ConsultarFacturasCompras(DateTime fechaInicial, DateTime fechaFinal)
		{
			return FacturaCompra
				.FindAll(new Order[] { Order.Asc("Fecha"), Order.Asc("Tipo"), Order.Asc("Codigo"), Order.Asc("Numero")},
						 Restrictions.Between("FechaImputacion", fechaInicial, fechaFinal));
		}
		
		public FacturaIvaComprasWrapper[] ProcesarFacturasCompra(FacturaCompra[] facturas, out decimal[] totales)
		{
			totales = new decimal[7];
			
			List<FacturaIvaComprasWrapper> facturasProcesadas = new List<FacturaIvaComprasWrapper>();
			
			decimal totalGravado = 0.00m;
			decimal totalExento = 0.00m;
			decimal totalIva = 0.00m;
			decimal totalPercIB = 0.00m;
			decimal totalPercIva = 0.00m;
			decimal totalImpComb = 0.00m;
			decimal totalTotal = 0.00m;
			
			FacturaIvaComprasWrapper temp;
			
			// Por cada factura pasada como parámetro
			foreach (FacturaCompra f in facturas) {
				
				// Creo un wrapper para la factura
				temp = new FacturaIvaComprasWrapper(f);
				
				totalGravado += f.ImporteGravado;
				totalExento += f.ImporteExento;
				totalIva += f.ImporteIva;
				totalPercIB += f.PercepcionIB;
				totalPercIva += f.PercepcionIva;
				totalImpComb += f.ImpuestoCombustible;
				totalTotal += f.ImporteTotal;
				
				// Agrego el wrapper a la lista de facturas procesadas
				facturasProcesadas.Add(temp);
			}
			
			// Seteo el array de totales con los resultados
			totales[0] = totalGravado;
			totales[1] = totalExento;
			totales[2] = totalIva;
			totales[3] = totalPercIB;
			totales[4] = totalPercIva;
			totales[5] = totalImpComb;
			totales[6] = totalTotal;
		
			// Retorno las facturas procesadas
			return facturasProcesadas.ToArray();
		}
		#endregion
		#endregion
		
		#region Otros métodos
		/// <summary>
		/// Método genérico para generar un reporte de IVA.
		/// </summary>
		/// <param name="nombreReporte">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="archivoTemplate">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="fechaInicialTicks">
		/// A <see cref="System.Int64"/>
		/// </param>
		/// <param name="fechaFinalTicks">
		/// A <see cref="System.Int64"/>
		/// </param>
		/// <param name="metodoConsulta">
		/// A <see cref="System.Func<DateTime, DateTime, T[]>"/>
		/// </param>
		/// <param name="metodoConsultaFacturasWrapper">
		/// A <see cref="ProcesadorFacturas<T, TWrapper>"/>
		/// </param>
		/// <param name="facturasPorPagina">
		/// A <see cref="System.Int32"/>
		/// </param>
		/// <param name="sistemaOperativo">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		protected string GenerarReporte<T,TWrapper>(string nombreReporte, string archivoTemplate, long fechaInicialTicks,
		                                            long fechaFinalTicks, System.Func<DateTime,DateTime,T[]> metodoConsulta,
		                                            ProcesadorFacturas<T,TWrapper> metodoConsultaFacturasWrapper,
		                                            int facturasPorPagina, string sistemaOperativo)
		{
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				
				DateTime fechaInicial = new DateTime(fechaInicialTicks);
				DateTime fechaFinal = new DateTime(fechaFinalTicks);
				
				// Chequeo que las fechas sean válidas
				this.ValidarFechas(fechaInicial, fechaFinal);
				
				// Realizo la consulta con el rango de fechas pasado como parámetro
				T[] resultado = metodoConsulta(fechaInicial, fechaFinal);
				
				if (resultado.Length == 0)
					throw new ObjetoInexistenteException("No existen facturas para el rango de fechas " +
						"especificado.");
				
				decimal[] totales;
				
				// Proceso las facturas que recuperé en la consulta
				TWrapper[] facturasProcesadas =
					metodoConsultaFacturasWrapper(resultado, out totales);
				
				/**** Creación del string HTML ****/
				
				// Creo el objeto ImpresionBorrador, que tiene un método para crear el HTML
				GeneradorDocumento impresion = new GeneradorDocumento(archivoTemplate, sistemaOperativo);
				
				// Agrego los objetos que necesito en el template
				impresion.AgregarObjeto("paginas", Convert.ToInt32(Math.Ceiling(facturasProcesadas.Length / (double)facturasPorPagina)));
				impresion.AgregarObjeto("facturasPorPagina", facturasPorPagina);
				impresion.AgregarObjeto("fechaDesde", fechaInicial);
				impresion.AgregarObjeto("fechaHasta", fechaFinal);
				impresion.AgregarObjeto("facturas", facturasProcesadas);
				impresion.AgregarObjeto("totales", totales);
				
				// Retorno el string HTML
				return impresion.CrearHTML();
			}
		}
		
		/// <summary>
		/// Valida el rango de fechas.
		/// </summary>
		/// <param name="fechaInicial">
		/// A <see cref="DateTime"/>
		/// </param>
		/// <param name="fechaFinal">
		/// A <see cref="DateTime"/>
		/// </param>
		public void ValidarFechas(DateTime fechaInicial, DateTime fechaFinal)
		{
			if (fechaInicial > fechaFinal)
				throw new ObjetoInvalidoException("Rango de fechas inválido. " +
					"La fecha final debe ser posterior a la inicial.");
			
			if (fechaFinal > DateTime.Today)
				throw new ObjetoInvalidoException("Rango de fechas inválido. " +
					"La fecha final no puede ser posterior a la fecha actual.");
		}
		#endregion
	}
}
