// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;

using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using Stratos.Common;

namespace StratosServer.Reportes
{
	public class FacturaIvaVentasWrapper
	{
		protected Factura facturaAsociada;
		
		protected decimal exento;
		protected decimal gravado;
		protected decimal ivaInsc;
		protected decimal total;
			
		public FacturaIvaVentasWrapper(Factura factura)
		{
			this.facturaAsociada = factura;
			
			this.exento = this.CalcularExento();
			this.gravado = this.CalcularGravado();
			this.ivaInsc = this.CalcularIvaIsc();
			this.total = this.CalcularTotal();
		}
		
		#region Propiedades
		
		#region Propiedades numericas para calculos
		public decimal ExentoNumerico {
			get { return this.exento; }
		}
		
		public decimal GravadoNumerico {
			get { return this.gravado; }
		}
		
		public decimal IvaInscNumerico {
			get { return this.ivaInsc; }
		}
		
		public decimal TotalNumerico {
			get { return this.total; }
		}
		#endregion
		
		#region Propiedades utilizadas para el reporte		
		public string Fecha {
			get { return this.facturaAsociada.Fecha.ToShortDateString(); }
		}
		
		/// <summary>
		/// Devuelve un string con un resumen del documento. El formato del resumen
		/// es, por ejemplo: "FA A  0004  00007634", para una Factura A cuyo código
		/// es 0004 y su número es 7634. 
		/// </summary>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		public string ResumenFactura {
			get {
				StringBuilder resumen = new StringBuilder();
				
				// Tipo de documento
				resumen.Append(this.facturaAsociada.TipoAbreviado);
				
				// Tipo de Factura
				if (this.facturaAsociada.Tipo == FacturaTipoVenta.FacturaA)
					resumen.Append(" A  ");
				else
					resumen.Append(" B  ");
				
				// Código en formato 'XXXX'
				resumen.Append(this.facturaAsociada.Codigo.ToString("D4") + "  ");
				
				// Número en formato 'XXXXXXXX'
				resumen.Append(this.facturaAsociada.Numero.ToString("D8"));
				
				return resumen.ToString();
			}
		}
		
		public string RazonSocial {
			get {
				if (this.facturaAsociada.Anulada)
					return "---ANULADA---";
				
				return this.facturaAsociada.ClienteRazonSocial;
			}
		}
		
		public string Iva {
			get {
				switch (this.facturaAsociada.ClienteTipoResponsable) {
					case TipoResponsable.ResponsableInscripto:
						return "I";
					case TipoResponsable.ConsumidorFinal:
						return "C";
					case TipoResponsable.Exento:
						return "E";
					case TipoResponsable.Monotributo:
						return "M";
					case TipoResponsable.ResponsableNoInscripto:
						return "N";
					default:
						return " ";
				}
			}
		}
		
		public string Cuit {
			get {
				if (!this.facturaAsociada.ClienteCuit.HasValue)
					return "";
				
				if (this.facturaAsociada.ClienteTipoResponsable == TipoResponsable.ConsumidorFinal)
					return "";
				
				if (this.facturaAsociada.Anulada)
					return "";
				
				CUIT cuit = new CUIT(this.facturaAsociada.ClienteCuit.Value);
				
				return cuit.ValorFormateado;					
			}
		}
		
		// TODO: Ver que representa este campo. Hay una consulta por hacer al contador.
		public string Exento {
			get {
				if (this.facturaAsociada.Anulada)
					return "";
				
				return this.facturaAsociada is NotaCredito ? "(0,00)" : "0,00";
			}
		}
		
		/// <summary>
		/// El importe gravado es el Neto, en una factura A y el importe Total menos
		/// el importe de IVA en una factura B. 
		/// </summary>
		public string Gravado {
			get {
				if (this.facturaAsociada.Anulada)
					return "";

				if (this.facturaAsociada is NotaCredito)
					return "(" + this.gravado.ToString("N2") + ")";
				
				return this.gravado.ToString("N2");
			}
		}
		
		public string IvaInsc {
			get {
				if (this.facturaAsociada.Anulada)
					return "";
		
				if (this.facturaAsociada is NotaCredito)
					return "(" + this.ivaInsc.ToString("N2") + ")";
				
				return this.ivaInsc.ToString("N2");
			}
		}
		
		public string Total {
			get {
				if (this.facturaAsociada.Anulada)
					return "";
				
				if (this.facturaAsociada is NotaCredito)
					return "(" + total.ToString("N2") + ")";
				
				return total.ToString("N2");
			}
		}
		#endregion
		
		#endregion
		
		#region Métodos que calculan importes
		// TODO: Ver en que casos el exento es distinto de 0.00
		protected decimal CalcularExento()
		{
			return 0.00m;
		}
		
		protected decimal CalcularGravado()
		{
			if (this.facturaAsociada.Anulada)
					return 0.00m;
			
			decimal gravado = 0.00m;
				
			// Si es Factura A, retorno el importe Neto
			if (this.facturaAsociada.Tipo == FacturaTipoVenta.FacturaA)
				gravado = this.facturaAsociada.ImporteNeto;
				
			// Si es una factura B, tengo que calcularlo, sacandole el Iva al importe Total
			else
				gravado = this.facturaAsociada.ImporteTotal/1.21m;
			
			return StratosUtil.RoundDecimal(gravado, 2);
		}
		
		protected decimal CalcularIvaIsc()
		{
			if (this.facturaAsociada.Anulada)
					return 0.00m;
			
			decimal ivaInsc = 0.00m;
			
			if (this.facturaAsociada.Tipo == FacturaTipoVenta.FacturaA)
				ivaInsc = this.facturaAsociada.ImporteIva;
			
			else
				ivaInsc = this.facturaAsociada.ImporteTotal * (1 - 1/1.21m);
			
			return StratosUtil.RoundDecimal(ivaInsc, 2);
		}
		
		protected decimal CalcularTotal()
		{
			if (this.facturaAsociada.Anulada)
					return 0.00m;
			
			return StratosUtil.RoundDecimal(this.facturaAsociada.ImporteTotal, 2);
		}
		#endregion
	}
}
