// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Collections.Generic;

using Stratos.ModeloDto;
using Stratos.Common.Excepciones;
using StratosServer.Ventas.Facturacion.Configuracion;

namespace StratosServer.Ventas.Facturacion
{
	public class CodigosFacturaHelper
	{
		private CodigoFacturaElementCollection coleccionCodigosDeFacturas;
		
		public CodigosFacturaHelper(CodigoFacturaElementCollection coleccion)
		{
			this.coleccionCodigosDeFacturas = coleccion;
		}
		
		/// <summary>
		/// Obtiene ls codigos de factura ingresados
		/// </summary>
		public CodigoFacturaDto[] ObtenerCodigosDeFactura()
		{
			List<CodigoFacturaDto> codigosFactura = new List<CodigoFacturaDto>();
			
			// Convierto los elementos de configuracion a objetos del tipo
			// CodigoFacturaDto, y los agrego a la lista.
			foreach (CodigoFacturaElement codigoElement in 
			         FacturacionElement.Instancia.CodigosDeFacturas)
			{
				codigosFactura.Add(this.CrearCodigoFacturaDto(codigoElement));
			}
			
			// Retorno un array con todos los codigos de factura ingresados
			return codigosFactura.ToArray();
		}
		
		/// <summary>
		/// Crea un objeto CodigoFacturaDto a partir de un objeto CodigoFacturaElement.
		/// </summary>
		/// <exception cref='StratosException'>
		/// Lanza una excepcion si alguno de los atributos de CodigoFacturaElemnt es invalido.
		/// </exception>
		public CodigoFacturaDto CrearCodigoFacturaDto(CodigoFacturaElement element)
		{
			#region Campo 'codigo'
			
			int codigo;
			
			// Chequeo que el codigo sea un entero
			if (!int.TryParse(element.Codigo, out codigo))
			{
				throw new StratosException("Error al obtener un codigo de factura desde el archivo " +
					"de configuracion. El atributo 'codigo' debe ser un numero entero.");
			}
			
			// Chequeo que no me devuelva el valor por defecto (-1). Si es asi, es porque no se seteo ningun codigo
			if (codigo == -1)
			{
				throw new StratosException("Error al obtener un codigo de factura desde el archivo " +
					"de configuracion. El atributo 'codigo' es obligatorio.");
			}
			
			// Chequeo que el codigo este entre 0 y 9999
			if (codigo < 0 || codigo > 9999)
			{
				throw new StratosException("Error al obtener un codigo de factura desde el archivo " +
					"de configuracion. El atributo 'codigo' debe ser un numero entero entre 0 y 9999.");
			}
			
			#endregion
			
			#region Campo 'descripcion'
			
			// Obtengo la descripcion
			string descripcion = element.Descripcion.Trim();
			
			// Chequeo que la descripcion no sea vacia
			if (string.IsNullOrEmpty(descripcion))
			{
				throw new StratosException("Error al obtener un codigo de factura desde el archivo " +
					"de configuracion. El atributo 'descripcion' no debe ser vacio.");
			}
			#endregion
			
			#region Campo 'facturamanual'
			
			bool facturaManual;
			
			// Chequeo que sea un valor valido
			if (!bool.TryParse(element.FacturaManual, out facturaManual))
			{
				throw new StratosException("Error al obtener un codigo de factura desde el archivo " +
					"de configuracion. El valor del atributo 'facturamanual' debe ser 'True' o 'False'.");
			}
			#endregion
			
			#region Campo 'solonotacredito'
			
			bool soloNotaDeCredito;
			
			// Chequeo que sea un valor valido
			if (!bool.TryParse(element.SoloNotaCredito, out soloNotaDeCredito))
			{
				throw new StratosException("Error al obtener un codigo de factura desde el archivo " +
					"de configuracion. El valor del atributo 'solonotacredito' debe ser 'True' o 'False'.");
			}
			#endregion
			
			// Una vez que hice todas las validaciones, creo un nuevo CodigoFacturaDto y retorno.
			return new CodigoFacturaDto(codigo, descripcion, facturaManual, soloNotaDeCredito);
		}
		
		/// <summary>
		/// Devuelve 'true' o 'false' según la configuración
		/// del código de factura pasado como parámetro, si este
		/// corresponde o no a una factura manual.
		/// </summary>
		/// <returns>
		/// </returns>
		/// <param name='codigo'>
		/// Código de factura.
		/// </param>
		public bool EsCodigoDeFacturaManual(int codigo)
		{
			foreach(CodigoFacturaDto codigoDto in this.ObtenerCodigosDeFactura())
			{
				if (codigoDto.codigo == codigo)
					return codigoDto.facturamanual;
			}
			
			throw new StratosException("El código de factura " + codigo.ToString("D4") + 
			                           " no está ingresado en el sistema.");
		}
		
		/// <summary>
		/// Devuelve 'true' o 'false' según la configuración
		/// del código de factura pasado como parámetro, si este
		/// es solo para notas de crédito o no.
		/// </summary>
		/// <returns>
		/// </returns>
		/// <param name='codigo'>
		/// Código de factura.
		/// </param>
		public bool EsCodigoSoloNotaCredito(int codigo)
		{
			foreach (CodigoFacturaDto codigoDto in this.ObtenerCodigosDeFactura())
			{
				if (codigoDto.codigo == codigo)
					return codigoDto.solonotacredito;
			}
			
			throw new StratosException("El código de factura " + codigo.ToString("D4") + 
			                           " no está ingresado en el sistema.");
		}
	}
}

