// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Reflection;
using System.Collections.Generic;

using Ice;
using log4net;
using NHibernate;
using NHibernate.Criterion;
using Castle.ActiveRecord.Framework;
using Castle.ActiveRecord.Queries;
using Castle.ActiveRecord;

using Stratos.QueryObjects;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;
using Stratos.Servicios;
using StratosServer.Controladores;
using StratosServer.Core;
using StratosServer.Clientes;
using StratosServer.Productos;
using Stratos.Common.Excepciones;
using StratosServer.Ventas.Facturacion.Configuracion;

namespace StratosServer.Ventas.Facturacion
{
	public class ControladorFacturacionI : ControladorFacturacionDisp_
	{
		#region Campos
		private ServidorStratos servidorStratos;
		private IBuscadorTerminalesImpresoras buscadorTerminalesImpresoras;
		private IControladorTerminales controladorTerminales;
		#endregion
		
		#region Constructores
		public ControladorFacturacionI()
		{
		}
		
		public ControladorFacturacionI(ServidorStratos servidorStratos,
		                               IBuscadorTerminalesImpresoras buscadorTerminalesImpresoras,
		                               IControladorTerminales controladorTerminales)
		{
			this.servidorStratos = servidorStratos;
			this.buscadorTerminalesImpresoras = buscadorTerminalesImpresoras;
			this.controladorTerminales = controladorTerminales;
		}
		#endregion
		
		#region Propiedades
		public ServidorStratos ServidorStratos {
			get { return this.servidorStratos; }
			set { this.servidorStratos = value; }
		}
		
		public IBuscadorTerminalesImpresoras BuscadorTerminalesImpresoras {
			get { return this.buscadorTerminalesImpresoras; }
			set { this.buscadorTerminalesImpresoras = value; }
		}
		
		public IControladorTerminales ControladorTerminales {
			get { return this.controladorTerminales; }
			set { this.controladorTerminales = value; }
		}
		#endregion
		
		#region Implementación de ControladorFacturacionDisp_
		#region Consulta
		#region Busquedas por HQL
		public override Ice.Object[] BuscarHQL(string consulta, ConsultaPaginable consultaPaginable,
		                                       Current current__)
		{
			Ice.Object[] resultado;
			
			/* En este caso, com */
			if (consulta.Contains(" PreFactura"))
				resultado = this.BuscarPreFacturaHQL(consulta, consultaPaginable, current__);
			else if (consulta.Contains(" Factura"))
				resultado = this.BuscarFacturaHQL(consulta, consultaPaginable, current__);
			else
				throw new StratosException("El código HQL es inválido. Este controlador sólo soporta " +
				                           "los tipos PreFactura y Factura");
			
			return resultado;
		}
		
		public override PreFacturaDto[] BuscarPreFacturaHQL(string consulta, ConsultaPaginable consultaPaginable,
		                                                    Current current__)
		{
			PreFacturaDto[] resultados =
				StratosServerUtil
					.BuscarHQLGenerico<PreFacturaDto,PreFactura>(consulta, consultaPaginable);
			
			return resultados;
		}
		
		public override FacturaDto[] BuscarFacturaHQL(string consulta, ConsultaPaginable consultaPaginable,
		                                              Current current__)
		{
			FacturaDto[] resultados =
				StratosServerUtil
					.BuscarHQLGenerico<FacturaDto,Factura>(consulta, consultaPaginable);
			
			return resultados;
		}
		#endregion
		
		public override PreFacturaDto[]
			ObtenerPreFacturasEncoladasParaImprimir(string identificadorTerminal,
		                                        	ConsultaPaginable consultaPaginable,
		                                            Current current__)
		{
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				DatosTerminal datosTerminal =
					this.controladorTerminales.ObtenerTerminal(identificadorTerminal, current__);
				
				// Cargo las condiciones dependiendo del tipo de documentos que soporta la terminal
				Disjunction condicionOr = Expression.Disjunction();
				FacturaTipoVenta facturaTipo = FacturaTipoVenta.FacturaA;
				
				foreach (TipoDocumento tipoDoc in datosTerminal.datosTermImpresora.tiposDocumento)
				{
					if (tipoDoc == TipoDocumento.FacturaTipoA)
						facturaTipo = FacturaTipoVenta.FacturaA;
					else if (tipoDoc == TipoDocumento.FacturaTipoB)
						facturaTipo = FacturaTipoVenta.FacturaB;
					
					condicionOr.Add(Expression.Eq("Tipo", facturaTipo));
				}
				
				// Creo la consulta con los criterios establecidos antes
				DetachedCriteria detachedCriteria = DetachedCriteria.For<PreFactura>();
				detachedCriteria.Add(condicionOr);
				
				/* Seteo un Transformer para evitar resultados duplicados cuando se hacen
				 * joins. */
				detachedCriteria.SetResultTransformer(new NHibernate.Transform.DistinctRootEntityResultTransformer());
				
				// Preparo la consulta para ser paginable si es que el usuario así lo quiso
				StratosServerUtil.PrepararConsultaPaginable(detachedCriteria, consultaPaginable);
				
				// Realizo la búsqueda
				PreFactura[] resultado = null;
				
				try {
					// Ordeno en forma descendente por la Fecha
					resultado = PreFactura.FindAll(detachedCriteria);
					
					// Transformo los resultados a DTO
					PreFacturaDto[] resultadosDto = ConvertidorDto.ToDto<PreFactura,PreFacturaDto>(resultado);
				
					return resultadosDto;
				}
				catch (ActiveRecordException ex) {
					StratosException exception = new StratosException();
					exception.mensaje = StratosServerUtil.ObtenerMensajesExcepcion(ex);
					throw exception;
				}
			}
		}
		
		public override Ice.Object[] BuscarQueryObjects(QueryObject queryObject,
		                                                ConsultaPaginable consultaPaginable,
		                                                Current current__)
		{
			
			Ice.Object[] resultado = null;
			
			// Me fijo el tipo del QueryObject, y en base a eso hago la consulta
			if (queryObject is FacturaDtoQuery)
				resultado = StratosServerUtil
					.BuscarQueryObjects<Factura,FacturaDto>(queryObject,consultaPaginable);
			else if (queryObject is PreFacturaDtoQuery)
				resultado = StratosServerUtil
					.BuscarQueryObjects<PreFactura,PreFacturaDto>(queryObject,consultaPaginable);
			else
				throw new StratosException("El QueryObject no es de tipo FacturaDtoQuery ni PreFacturaDtoQuery");
			
			return resultado;
		}
		#endregion
		
		#region Métodos Refrescar
		public override Ice.Object Refrescar(Ice.Object objeto, Ice.Current current__)
		{
			if (objeto is PreNotaCreditoDto) {
				PreNotaCreditoDto preNotaCreditoDto = objeto as PreNotaCreditoDto;
				
				return StratosServerUtil
					.Refrescar<PreNotaCredito,PreNotaCreditoDto>(preNotaCreditoDto,
					                                             preNotaCreditoDto.id);
			}
			else if (objeto is NotaCreditoDto) {
				NotaCreditoDto notaCreditoDto = objeto as NotaCreditoDto;
				
				return StratosServerUtil
					.Refrescar<NotaCredito,NotaCreditoDto>(notaCreditoDto,
					                                       notaCreditoDto.id);
			}
			if (objeto is PreFacturaDto) {
				PreFacturaDto preFacturaDto = objeto as PreFacturaDto;
				
				return StratosServerUtil
					.Refrescar<PreFactura,PreFacturaDto>(preFacturaDto,
					                                     preFacturaDto.id);
			}
			else if (objeto is FacturaDto) {
				FacturaDto facturaDto = objeto as FacturaDto;
				
				return StratosServerUtil
					.Refrescar<Factura,FacturaDto>(facturaDto,
					                               facturaDto.id);
			}
			else {
				throw new StratosException("El objeto a refrescar debe ser del tipo FacturaDto " +
				                           "o PreFacturaDto");
			}
		}
		#endregion
 		
		#region Guardar
		public override bool GuardarPreFactura (PreFacturaDto preFacturaDto, out string mensajeError, Current current__)
		{
			PreFacturaDto preFacturaGuardada;
			
			return this.GuardarPreFactura(preFacturaDto, out mensajeError, out preFacturaGuardada, current__);
		}
		
		/// <summary>
		/// Encola una prefactura para su posterior impresión.
		/// Devuelve true si la factura fue enviada a imprimir al cliente correspondiente y
		/// false en caso contrario, en el cual la prefactura quedara guardada en el servidor
		/// para su posterior recuperación e impresión.
		/// Esta versión del método además retorna una instancia Dto actualizada.
		/// </summary>
		/// <param name="preFacturaDto">
		/// A <see cref="PreFacturaDto"/>
		/// </param>
		/// <param name="current__">
		/// A <see cref="Ice.Current"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public bool GuardarPreFactura (PreFacturaDto preFacturaDto,
                                       out string mensajeError,
                                       out PreFacturaDto preFacturaDtoGuardada,
                                       Ice.Current current__)
		{
			PreFactura nuevaPreFactura = null;
			preFacturaDtoGuardada = null;
			
			// Utilizo una transacción para realizar todas las operaciones en la base de datos
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<PreFactura>(preFacturaDto.id,
					                             preFacturaDto.version,
					                             transactionScope);
				
				/* Si se trata de una Factura B, el Cliente es Consumidor Final y éste no se encuentra
				 * persistido, entonces puedo darle de alta automáticamente aquí. Me fijo si el mismo
				 * está persistido o no y lo hago en caso de ser necesario. */
				if (preFacturaDto.tipo == FacturaTipoVenta.FacturaB &&
				    	preFacturaDto.cliente.tipoRes ==
				    		TipoResponsable.ConsumidorFinal &&
				    	preFacturaDto.cliente.codigo < 0) {
					
					ControladorClientesI controladorClientes = new ControladorClientesI();
					
					// Agrego el cliente. Si hay una excepción la devuelvo
					try {
						ClienteDto clientePersistido =
							controladorClientes.Guardar(preFacturaDto.cliente,
							                            true);
						
						// Una vez persistido, reasigno el cliente guardado a la PreFacturaDto,
						// para que luego quede disponible
						preFacturaDto.cliente = clientePersistido;
					}
					catch (System.Exception ex) {
						StratosServerUtil.TratarExcepcion(ex);
					}
				}
				
				/* La PreFacturaDto puede ser una PreNotaCreditoDto. Por lo tanto debo utilizar
				 * el tipo correcto al guardar. */
				
				PreNotaCreditoDto preNotaCreditoDto = null;
				PreNotaCredito nuevaPreNotaCredito = null;
				
				// Convierto el objeto Dto a uno presistente de acuerdo al tipo del mismo
				if (preFacturaDto is PreNotaCreditoDto) {
					preNotaCreditoDto = (PreNotaCreditoDto) preFacturaDto;
					
					nuevaPreNotaCredito =
						ConvertidorDto
							.FromDto<PreNotaCreditoDto,PreNotaCredito>(preNotaCreditoDto);
					
					/* También asigno la nota de crédito persistente a la variable nuevaPreFactura,
					 * ya que esta se utiliza abajo para avisar a la terminarl impresora. */
					nuevaPreFactura = nuevaPreNotaCredito;
				}
				else {
					nuevaPreFactura =
						ConvertidorDto
							.FromDto<PreFacturaDto,PreFactura>(preFacturaDto);
				}
				
				/* Persisto la PreFactura o PreNotaCredito (según corresponda), fijándome si debo
				 * crear una o actualizar la existente */
				// Si es una PreNotaCredito...
				if (preNotaCreditoDto != null) {
					if (preFacturaDto.id <= 0)
						StratosServerUtil
							.CrearObjeto<PreFactura>(nuevaPreNotaCredito, transactionScope);
					else
						StratosServerUtil
							.ActualizarObjeto<PreFactura>(nuevaPreNotaCredito, transactionScope);
					
					// Almaceno la versión Dto actualizada
					preFacturaDtoGuardada = ConvertidorDto.ToDto(nuevaPreNotaCredito);
				}
				
				// Si es una PreFactura...
				else {
					if (preFacturaDto.id <= 0)
						StratosServerUtil
							.CrearObjeto<PreFactura>(nuevaPreFactura, transactionScope);
					else
						StratosServerUtil
							.ActualizarObjeto<PreFactura>(nuevaPreFactura, transactionScope);
					
					// Almaceno la versión Dto actualizada
					preFacturaDtoGuardada = ConvertidorDto.ToDto(nuevaPreFactura);
				}
			}
			
			// Aviso al Servicio de Facturacion que hay una nueva factura por imprimir
			TipoDocumento tipoDoc = preFacturaDto.tipo == FacturaTipoVenta.FacturaA ?
				TipoDocumento.FacturaTipoA : TipoDocumento.FacturaTipoB;
			
			// Aviso al Servicio de Facturacion que hay una nueva factura por imprimir
			List<DatosTerminal> terminalesImpresoras =
				this.buscadorTerminalesImpresoras.BuscarTerminalesImpresora(tipoDoc);
			
			bool enviadaCorrectamente;
			mensajeError = "";
			
			if (terminalesImpresoras.Count > 0) {
				try {
					// FIXME: Se debería seleccionar la terminal, y no elegir la primera siempre
					terminalesImpresoras[0].datosTermImpresora
						.servFacturacion[0]
							.ImprimirPreFactura(ConvertidorDto.ToDto(nuevaPreFactura));
					
					enviadaCorrectamente = true;
				}
				catch (System.Exception ex) {
					mensajeError = StratosServerUtil.ObtenerMensajesExcepcion(ex);
					enviadaCorrectamente = false;
				}
			}
			else {
				mensajeError =
					"El documento se ha guardado correctamente en el sistema, pero " +
					"no existen terminales conectadas que soporten la impresión del tipo " +
					"de documento solicitado";
				
				enviadaCorrectamente = false;
			}
			
			return enviadaCorrectamente;
		}
		
		public override FacturaDto RegistrarFacturaManual(FacturaDto facturaDto, Ice.Current current__)
		{
			/* Chequeo que el código de la factura corresponda a una factura manual y
			 * en caso de que sea factura, chequeo que el código no sea solo para Nota de Crédito */
			
			// Creo un helper
			CodigosFacturaHelper helper = new CodigosFacturaHelper(FacturacionElement.Instancia.CodigosDeFacturas);
			
			// Chequeo que el código sea de una factura manual
			if (!helper.EsCodigoDeFacturaManual(facturaDto.codigo))
				throw new StratosException("No se pueden ingresar facturas manuales con código " + facturaDto.codigo.ToString("D4") + 
				                           ".\nVerifique el archivo de configuración.");
			
			// Si es una factura, chequeo que no se utilize un código de solo nota de crédito
			if ((facturaDto is NotaCreditoDto) && helper.EsCodigoSoloNotaCredito(facturaDto.codigo))
				throw new StratosException("El código " + facturaDto.codigo.ToString("D4") + " solo se puede utilizar para Notas de Crédito.\n" +
				                           "Verifique el archivo de configuración.");
			
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				// Paso la FacturaDto a Factura
				Factura factura = ConvertidorDto.FromDto<FacturaDto,Factura>(facturaDto);
				
				Factura facturaIngresada =
					this.RegistrarFacturaManual(factura, transactionScope);
				
				return ConvertidorDto.ToDto(facturaIngresada);
			}
		}
		
		/// <summary>
		/// Guarda un PreFactura.
		/// </summary>
		/// <param name="objeto">
		/// A <see cref="Ice.Object"/>
		/// </param>
		/// <param name="current__">
		/// A <see cref="Current"/>
		/// </param>
		/// <returns>
		/// A <see cref="Ice.Object"/>
		/// </returns>
		public override Ice.Object Guardar (Ice.Object preFactura, Current current__)
		{
			PreFacturaDto preFacturaGuardada;
			string mensaje;
			
			// Verifico el tipo
			PreFacturaDto preFacturaDto =
				StratosServerUtil.TipoCorrecto<PreFacturaDto>(preFactura);
			
			this.GuardarPreFactura(preFacturaDto, out mensaje, out preFacturaGuardada, current__);
			
			return preFacturaGuardada;
		}
		
		public override Ice.Object Modificar (Ice.Object preFactura, Current current__)
		{
			/* FIXME: Actualmente, en este controlador, la operación de crear una nueva
			 * PreFactura o modificar una existente se realiza en el mismo método:
			 * GuardarPreFactura. Por lo tanto este método Modificar llama directamente
			 * al método Guardar, con el mismo resultado.
			 * Esta tarea de realizar lo mismo en el mismo método podría estar separada. */
			return this.Guardar(preFactura, current__);
		}
		
		public override FacturaDto GuardarFacturaManual (FacturaDto facturaManualDto, Current current__)
		{
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<Factura>(facturaManualDto.id,
					                          facturaManualDto.version,
					                          transactionScope);
				
				// Me aseguro que la factura por guardar sea manual
				Factura facturaGuardada = Factura.Find(facturaManualDto.id);
				
				if (!facturaGuardada.FacturaManual)
					throw new ObjetoInvalidoException("No se pueden modificar las facturas que no son manuales");
				
				// Paso los datos de la FacturaDto
				Factura facturaManual = ConvertidorDto.FromDto<FacturaDto,Factura>(facturaManualDto);
				
				// Guardo los cambios hechos
				StratosServerUtil.ActualizarObjeto<Factura>(facturaManual, transactionScope);
				
				// Paso a Dto la factura guardada y retorno
				return ConvertidorDto.ToDto(facturaManual);
			}
		}
		#endregion
		
		#region Impresion de facturas
		public override void PrepararPreFacturaParaImprimir (PreFacturaDto preFacturaDto,
		                                                     out long numeroFactura,
		                                                     Ice.Current current__)
		{
			numeroFactura = -1;
			
			// Utilizo una sesión
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<PreFactura>(preFacturaDto.id,
					                             preFacturaDto.version);
				
				/* Consulto el último número de Factura según el tipo y código de la PreFactura. */
				ScalarQuery<Nullable<long>> consultaUltimoNumero =
					new ScalarQuery<Nullable<long>>(typeof(Factura),
					                          @"select max(f.Numero)
												from Factura f
												where f.Tipo = :tipo
													and f.Codigo = :codigo");
				
				consultaUltimoNumero.SetParameter("tipo", preFacturaDto.tipo);
				consultaUltimoNumero.SetParameter("codigo", preFacturaDto.codigo);
				
				long? resultado = consultaUltimoNumero.Execute();
				
				// Si hay resultados, entonces el numero de factura es el siguiente del último
				if (resultado.HasValue)
					numeroFactura = resultado.Value + 1;
				else
					// FIXME: Ver si acá no es 0 en realidad
					numeroFactura = 1;
				
				// Ahora valido que la fecha de la factura por imprimir sea válida
				PreFactura preFactura = PreFactura.Find(preFacturaDto.id);
				
				if (!preFactura.IsValid()) {
					if (preFactura.PropertiesValidationErrorMessages
					    	.Keys.Cast<PropertyInfo>().Any<PropertyInfo>(x => x.Name == "Fecha")) {
						
						throw new ImpresionFacturaException(
							"La fecha de la PreFactura por imprimir es inválida. Asegúrese que ésta no sea " +
							"una fecha anterior a la última Factura impresa del mismo tipo y código."
						);
					}
					
					else {
						string mensajeValidacion =
							StratosServerUtil.ObtenerMensajesValidacion(preFactura);
						
						throw new ImpresionFacturaException(
							"La PreFactura que se envió a imprimir tiene propiedades inválidas:\n" +
							mensajeValidacion
						);
					}
				}
			}
		}
		
		public override FacturaDto ImpresionPreFacturaFinalizadaCorrectamente (PreFacturaDto preFacturaDto,
		                                                                 long numeroFactura,
		                                                                 Ice.Current current__)
		{
			// Utilizo una transacción
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<PreFactura>(preFacturaDto.id,
					                             preFacturaDto.version,
					                             transactionScope);
				
				// TODO: Quizá aquí se debería chequear nuevamente si el número de factura es
				// realmente correcto, el siguiente.
				
				// preFacturaDto puede ser una PreNotaCreditoDto. De acuerdo al tipo real realizo las operciones
				// correctas.
				
				PreFactura preFactura = null;
				Factura factura = null;
				
				PreNotaCredito preNotaCredito = null;
				NotaCredito notaCredito = null;
				
				// En caso de que sea una PreNotaCreditoDto...
				if (preFacturaDto is PreNotaCreditoDto) {
					preNotaCredito = PreNotaCredito.Find(preFacturaDto.id);
					
					notaCredito = ConvertidorDto.ToNotaCredito(preNotaCredito);
				}
				
				// o realmente sea una PreFactura...
				else {
					preFactura = PreFactura.Find(preFacturaDto.id);
					
					factura = ConvertidorDto.ToFactura(preFactura);
				}
				
				// Ahora persisto la factura o la nota de crédito, según corresponda, y
				// elimino la PreFactura o PreNotaCredito
				if (notaCredito != null) {
					// FIXME: El codigo está hardcodeado. Debería ser parte de la PreFactura
					notaCredito.Codigo = preNotaCredito.Codigo;
					notaCredito.Numero = numeroFactura;
					
					// Creo la nota de crédito
					StratosServerUtil.CrearObjeto<Factura>(notaCredito, transactionScope);
					
					// Una vez que la NotaCredito ha sido correctamente almacenada, elimino la
					// PreNotaCredito de la base de datos
					StratosServerUtil.EliminarObjeto<PreFactura>(preNotaCredito, transactionScope);
					
					return ConvertidorDto.ToDto(notaCredito);
				}
				else {
					factura.Codigo = preFactura.Codigo;
					factura.Numero = numeroFactura;
					
					// Creo la factura
					StratosServerUtil.CrearObjeto<Factura>(factura, transactionScope);
					
					// Una vez que la Factura ha sido correctamente almacenada, elimino la PreFactura de
					// la base de datos
					StratosServerUtil.EliminarObjeto<PreFactura>(preFactura, transactionScope);
					
					return ConvertidorDto.ToDto(factura);
				}
			}
		}
		#endregion
		
		#region Eliminación
		public override void EliminarFactura(FacturaDto facturaDto, Ice.Current current__)
		{
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				Factura factura = Factura.TryFind(facturaDto.id);
				
				// Si la factura por eliminar es una factura histórica, entonces no permito
				// la operación y devuelvo una excepción acorde
				if (factura.Historica)
					throw new FacturaImposibleDeEliminarException("Una factura histórica no puede eliminarse");
				
				/* Consulto el último número de Factura según el tipo y código de la Factura.
				 * Si la Factura por eliminar no posee este número, entonces la eliminación no
				 * es posible. */
				ScalarQuery<Nullable<long>> consultaUltimoNumero =
					new ScalarQuery<Nullable<long>>(typeof(Factura),
					                          @"select max(f.Numero)
												from Factura f
												where f.Tipo = :tipo
													and f.Codigo = :codigo");
				
				consultaUltimoNumero.SetParameter("tipo", facturaDto.tipo);
				consultaUltimoNumero.SetParameter("codigo", facturaDto.codigo);
				
				long? resultado = consultaUltimoNumero.Execute();
				
				if (!resultado.HasValue) {
					transactionScope.VoteRollBack();
					
					FacturaImposibleDeEliminarException ex = new FacturaImposibleDeEliminarException();
					ex.mensaje = "La Factura que se desea eliminar no existe en la base de datos";
					
					throw ex;
				}
				else if (resultado.Value != facturaDto.numero) {
					transactionScope.VoteRollBack();
					
					FacturaImposibleDeEliminarException ex = new FacturaImposibleDeEliminarException();
					ex.mensaje = "Sólo es posible eliminar la última Factura";
					
					throw ex;
				}
				
				// Creo una PreFactura correspondiente a la Factura, por si necesitan
				// reimprimir. Pero sólo si no es una factura inválida de sincronización
				// o una factura manual.
				if (factura.Valida && !factura.FacturaManual)
					this.CrearPreFacturaDesdeFactura(factura, transactionScope);
				
				// Realizo la operación de eliminación
				StratosServerUtil.EliminarObjeto<Factura>(factura, transactionScope);
			}
		}
		
		public override void EliminarPreFactura(PreFacturaDto preFacturaDto, Ice.Current current__)
		{
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<PreFactura>(preFacturaDto.id,
					                             preFacturaDto.version,
					                             transactionScope);
				
				// PreFacturaDto -> PreFactura
				PreFactura preFactura = PreFactura.TryFind(preFacturaDto.id);
				
				StratosServerUtil.EliminarObjeto<PreFactura>(preFactura, transactionScope);
			}
		}
		#endregion
		
		#region Anulación de facturas
		public override void AnularFactura(FacturaDto facturaDto, bool crearPreFactura, out FacturaDto facturaDtoNuevaVersion,
		                                   Ice.Current current__)
		{
			/* Por defecto asigno a la factura de salida la misma versión que la de entrada.
			 * Si la operación se lleva acabo sin problemas, se actualizará a la versión real
			 * modificada. */
			facturaDtoNuevaVersion = facturaDto;
			
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<Factura>(facturaDto.id,
					                          facturaDto.version,
					                          transactionScope);
				
				/* En este caso no utilizo ConvertidorDto para levantar la factura, porque no me
				 * interesa tenerla completa. */
				Factura factura = Factura.Find(facturaDto.id);
				
				/* Si se trata de una Factura (no una NotaCredito), primero anulo todas las notas
				 * de crédito asociadas */
				if (factura.GetType().Equals(typeof(Factura))) {
					NotaCredito[] notasCreditoAsociadasNoAnuladas =
						NotaCredito.FindAll(Restrictions.Eq("FacturaAsociada", factura),
						                    Restrictions.Eq("Anulada", false));
					
					foreach (NotaCredito nc in notasCreditoAsociadasNoAnuladas) {
						nc.Anulada = true;
						
						try {
							nc.UpdateAndFlush();
						}
						catch (System.Exception ex) {
							transactionScope.VoteRollBack();
							StratosServerUtil.TratarExcepcion(ex);
						}
					}
				}
				
				// Anulo la factura
				if (!factura.Anulada) {
					factura.Anulada = true;
					
					StratosServerUtil.ActualizarObjeto<Factura>(factura, transactionScope);
					
					facturaDtoNuevaVersion = ConvertidorDto.ToDto<Factura,FacturaDto>(factura);
					
					// Si se especificó, creo una PreFactura asociada
					if (!(factura is NotaCredito) && crearPreFactura)
						this.CrearPreFacturaDesdeFactura(factura, transactionScope);
				}
			}
		}
		
		public override void DesanularFactura(FacturaDto facturaDto,
		                                      out FacturaDto facturaDtoNuevaVersion,
		                                      Ice.Current current__)
		{
			/* Por defecto asigno a la factura de salida la misma versión que la de entrada.
			 * Si la operación se lleva acabo sin problemas, se actualizará a la versión real
			 * modificada. */
			facturaDtoNuevaVersion = facturaDto;
			
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<Factura>(facturaDto.id,
					                          facturaDto.version,
					                          transactionScope);
				
				/* En este caso no utilizo ConvertidorDto para levantar la factura, porque no me
				 * interesa tenerla completa. */
				Factura factura = Factura.Find(facturaDto.id);
				
				// Valido que no se esté desanulando una Factura inválida
				if (!factura.Valida)
					throw new ObjetoInvalidoException("No se puede desanular una Factura inválida." +
					                                  "La Factura que intenta desanular fue creada " +
					                                  "para sincronizar los números de Factura. No es " +
					                                  "una Factura válida que pueda desanularse.");
				
				if (factura.Anulada) {
					factura.Anulada = false;
					
					StratosServerUtil.ActualizarObjeto<Factura>(factura, transactionScope);
					
					facturaDtoNuevaVersion = ConvertidorDto.ToDto<Factura,FacturaDto>(factura);
				}
			}
		}
		#endregion
		
		#region Sincronización de facturas
		public override FacturaDto IngresarFacturaDeSincronizacion(FacturaTipoVenta facturaTipo, int codigo,
		                                                           long numero, Ice.Current current__)
		{
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				/* Busco la última factura anterior al tipo y código indicados. */
				DetachedCriteria numeroConsulta = DetachedCriteria.For<Factura>();
				numeroConsulta.Add(Expression.Eq("Tipo", facturaTipo));
				numeroConsulta.Add(Expression.Eq("Codigo", codigo));
				numeroConsulta.SetProjection(Projections.Max("Numero"));
				
				DetachedCriteria consultaUltimaFactura = DetachedCriteria.For<Factura>();
				consultaUltimaFactura.Add(Expression.Eq("Tipo", facturaTipo));
				consultaUltimaFactura.Add(Expression.Eq("Codigo", codigo));
				consultaUltimaFactura.Add(Subqueries.PropertyEq("Numero", numeroConsulta));
				
				Factura ultimaFactura = Factura.FindOne(consultaUltimaFactura);
				
				// Si la última factura no existe, entonces el siguiente número debe ser el 1.
				if (ultimaFactura == null) {
					if (numero != 1) {
						transactionScope.VoteRollBack();
						
						throw new ParametrosInvalidosException("El número es incorrecto.");
					}
				}
				
				/* Si existe una última factura, verifico que el número indicado para la factura
				 * de sincronización no sea menor o igual al de ésta. */
				else if (ultimaFactura.Numero >= numero) {
					transactionScope.VoteRollBack();
					
					throw new ParametrosInvalidosException("El número indicado debe ser superior al último " +
						"número de factura del mismo tipo y código (" + ultimaFactura.Numero + ")");
				}
				
				// FIXME: Se podría crear un tipo especial FacturaDeSincronizacion que
				// tenga toda esta lógica, y que evite algunas validaciones (como por ejemplo
				// Continuo creando la Factura inválida
				Factura facturaInvalida = new Factura();
				
				/* Si encontré una última factura creada del mismo tipo y código, entonces le asigno a ésta
				 * la misma fecha de aquella. */
				facturaInvalida.Fecha = (ultimaFactura == null ? DateTime.Today : ultimaFactura.Fecha);
				
				facturaInvalida.Tipo = facturaTipo;
				facturaInvalida.Codigo = codigo;
				facturaInvalida.Numero = numero;
				
				// Asigno un cliente inválido a la factura de sincronización, según el tipo de ésta
				ControladorClientesI controladorClientes = new ControladorClientesI();
				if (facturaTipo == FacturaTipoVenta.FacturaA)
					facturaInvalida.Cliente =
						controladorClientes.ObtenerClienteInvalido(TipoResponsable.ResponsableInscripto);
				else
					facturaInvalida.Cliente =
						controladorClientes.ObtenerClienteInvalido(TipoResponsable.ConsumidorFinal);
				
				// Asigno un item de producto
				ControladorProductosI controladorProductos = new ControladorProductosI();
				ItemFactura itemFactura = new ItemFactura(controladorProductos.ObtenerProductoInvalido(), 1);
				itemFactura.PrecioUnitario = 1.00m;
				itemFactura.Importe = 1.00m;
				facturaInvalida.Items.Add(itemFactura);
				
				// La nueva factura será inválida y no estará anulada
				facturaInvalida.Anulada = false;
				facturaInvalida.Valida = false;
				
				// Asigno los valores de los importes según corresponda
				if (facturaTipo == FacturaTipoVenta.FacturaA) {
					facturaInvalida.ImporteBruto = 1.00m;
					facturaInvalida.ImporteBonificacion = 0.00m;
					facturaInvalida.ImporteNeto = 1.00m;
					facturaInvalida.ImporteIva = 0.21m;
					facturaInvalida.ImporteTotal = 1.21m;
				}
				else {
					// Si es una Factura B, el Item debe tener el precio unitario más el IVA
					itemFactura.PrecioUnitario = itemFactura.ProductoImporteVenta;
					itemFactura.Importe = itemFactura.PrecioUnitario;
					
					facturaInvalida.ImporteBruto = 1.21m;
					facturaInvalida.ImporteBonificacion = 0.00m;
					facturaInvalida.ImporteNeto = 1.21m;
					facturaInvalida.ImporteIva = 0.00m;
					facturaInvalida.ImporteTotal = 1.21m;
				}
				
				this.RegistrarFacturaManual(facturaInvalida, transactionScope);
				
				return ConvertidorDto.ToDto(facturaInvalida);
			}
		}
		#endregion
		
		#region Codigos de Facturas
		public override Stratos.ModeloDto.CodigoFacturaDto[] ObtenerCodigosDeFactura (Ice.Current current__)
		{
			// Creo un helper
			CodigosFacturaHelper helper = new CodigosFacturaHelper(FacturacionElement.Instancia.CodigosDeFacturas);
			
			// Obtengo un array de tipo CodigoFacturaDto
			return helper.ObtenerCodigosDeFactura();
		}
		
		#endregion
		
		#endregion
		
		#region Métodos internos para el Servidor
		protected void CrearPreFacturaDesdeFactura(Factura factura, TransactionScope transactionScope)
		{
			PreFactura preFactura = ConvertidorDto.ToPreFactura(factura);
			
			try {
				preFactura.Create();
			}
			catch (System.Exception ex) {
				// FIXME: ¿Qué pasa acá? Arreglar.
			}
		}
		
		public Factura RegistrarFacturaManual(Factura factura, TransactionScope transactionScope)
		{
			// Marco la factura como Factura manual
			factura.FacturaManual = true;
			
			// Persisto la Factura
			StratosServerUtil.CrearObjeto<Factura>(factura, transactionScope);
			
			return Factura.Find(factura.Id);
		}

		#endregion
	}
}
