// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using Ice;
using log4net;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;

using StratosServer.Core;
using Stratos.Common.Excepciones;
using StratosServer.Controladores;
using Stratos.ModeloDto;
using Stratos.QueryObjects;
using Stratos.ModeloObjetos;

namespace StratosServer.Ventas.Presupuestos
{
	/// <summary>
	/// Clase encargada de gestionar las operaciones relacionadas con los presupuestos.
	/// </summary>
	public class ControladorPresupuestosI : ControladorPresupuestosDisp_
	{
		public override Ice.Object[] BuscarHQL(string consulta,
		                                       ConsultaPaginable consultaPaginable,
		                                       Current current__)
		{
			PresupuestoDto[] resultados =
				StratosServerUtil
					.BuscarHQLGenerico<PresupuestoDto,Presupuesto>(consulta, consultaPaginable);
			
			return resultados;
		}
		
		public override Ice.Object[] BuscarQueryObjects(QueryObject queryObject,
		                                                ConsultaPaginable consultaPaginable,
		                                                Current current__)
		{
			// Valido que el QueryObject sea de tipo PresupuestoDtoQuery
			if (!(queryObject is PresupuestoDtoQuery))
				throw new StratosException("El QueryObject no es de tipo PresupuestoDtoQuery");
			
			PresupuestoDto[] resultado =
				StratosServerUtil
					.BuscarQueryObjects<Presupuesto,PresupuestoDto>(queryObject,
					                                                consultaPaginable);
			
			return resultado;
		}
		
		public override Ice.Object Refrescar(Ice.Object objeto, Ice.Current current__)
		{
			PresupuestoDto presupuestoDto = objeto as PresupuestoDto;
			
			return StratosServerUtil
				.Refrescar<Presupuesto,PresupuestoDto>(presupuestoDto,
				                                       presupuestoDto.id);
		}
		
		// FIXME: Este método sería innecesario. Utilizar BuscarHQL
		public override PresupuestoDto[] ObtenerTodos (Current current__)
		{
			Presupuesto[] presupuestos;
			try {
				presupuestos = Presupuesto.FindAll();
			}
			catch (ActiveRecordException ex) {
				StratosException excepcion =
					new StratosException(StratosServerUtil.ObtenerMensajesExcepcion(ex));
				
				throw excepcion;
			}
			
			return ConvertidorDto.ToDto(presupuestos);
		}
		
		public override Ice.Object Guardar (Ice.Object presupuesto, Current current__)
		{
			// Utilizo una transacción para realizar todas las operaciones en la base de datos
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Verifico el tipo
				PresupuestoDto presupuestoDto =
					StratosServerUtil.TipoCorrecto<PresupuestoDto>(presupuesto);
				
				/* Si estoy modificando un Presupuesto existente, entonces me aseguro que los cambios
				 * que hizo el usuario hayan sido aplicados sobre la misma versión que existe en la
				 * base de datos. */
				if (presupuestoDto.id > 0)
					StratosServerUtil
						.VersionCorrecta<Presupuesto>(presupuestoDto.id,
						                              presupuestoDto.version,
						                              transactionScope);
				
				// Persisto el Presupuesto
				Presupuesto nuevoPresupuesto =
					ConvertidorDto.FromDto<PresupuestoDto,Presupuesto>(presupuestoDto);
				
				// FIXME: Este código podría estar separado en métodos para guardar (crear) un
				// presupuesto nuevo y modificar un presupuesto existente.
				if (presupuestoDto.id <= 0) {
					StratosServerUtil
						.CrearObjeto<Presupuesto>(nuevoPresupuesto, transactionScope);
				}
				else {
					StratosServerUtil
						.ActualizarObjeto<Presupuesto>(nuevoPresupuesto, transactionScope);
				}
				
				return ConvertidorDto.ToDto(nuevoPresupuesto);
			}
		}
		
		public override Ice.Object Modificar (Ice.Object presupuesto, Current current__)
		{
			// FIXME: En este controlador las operaciones de crear un nuevo presupuesto
			// y modificar uno existente están bajo el método Guardar. Se podría separar
			// dicha funcionalidad.
			return this.Guardar(presupuesto, current__);
		}
		
		public override void EliminarPresupuesto(PresupuestoDto presupuestoDto, Ice.Current current__)
		{
			// Utilizo una transaccion
			using (TransactionScope transactionScope = new TransactionScope(TransactionMode.Inherits)) {
				
				// Chequeo que la versión que se intenta eliminar sea la actual
				StratosServerUtil
					.VersionCorrecta<Presupuesto>(presupuestoDto.id,
					                              presupuestoDto.version,
					                              transactionScope);
				
				Presupuesto presupuesto = Presupuesto.Find(presupuestoDto.id);
				
				StratosServerUtil.EliminarObjeto<Presupuesto>(presupuesto, transactionScope);
			}
		}
		
	}
}
