// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;

using log4net;
using Ice;
using Mono.Addins;

using Stratos.Servicios;
using Stratos.Common;
using Stratos.Common.Excepciones;
using StratosServer.Controladores;

namespace StratosServer.Core
{
	public class ControladorTerminalesI : ControladorTerminalesDisp_, IControladorTerminales
	{
		#region Campos
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);
		
		private Regex regexDireccionRemota;
		
		private static ControladorTerminalesI instancia;
		
		private Dictionary<string, DatosTerminal> terminalesConectadas;
		
		private Dictionary<string, Ice.ObjectPrx> controladoresDisponibles;
		#endregion
		
		#region Constructores
		private ControladorTerminalesI()
		{
			this.regexDireccionRemota =
				new Regex(@"remote address[ ]+=[ ]+(?<direccion>[0-9\.]+)",
				          RegexOptions.Compiled);
			
			this.terminalesConectadas = new Dictionary<string, DatosTerminal>();
			this.controladoresDisponibles = new Dictionary<string, Ice.ObjectPrx>();
		}
		#endregion
		
		#region Propiedades
		public static ControladorTerminalesI Instancia {
			get {
				if (instancia == null)
					instancia = new ControladorTerminalesI();
				
				return instancia;
			}
		}
		
		public Dictionary<string, Ice.ObjectPrx> ControladoresDisponibles {
			get { return this.controladoresDisponibles; }
		}
		
		public Dictionary<string, DatosTerminal> TerminalesConectadas {
			get { return this.terminalesConectadas; }
		}
		#endregion
		
		/// <summary>
		/// Este método es utilizado en los Tests. Elimina la instancia actual para recomenzar
		/// un nuevo test.
		/// </summary>
		public static void EliminarInstancia()
		{
			instancia = null;
		}
		
		/// <summary>
		/// Agrega un proxy a un ControladorBase a la lista de controladores disponibles.
		/// </summary>
		/// <param name="controladorProxy">
		/// A <see cref="ControladorBasePrx"/>
		/// </param>
		public void AgregarControlador(Ice.ObjectPrx controladorProxy)
		{
			logger.Debug("Agregando proxy del controlador: " + controladorProxy.ToString());
			
			if (this.controladoresDisponibles.ContainsKey(controladorProxy.ice_getIdentity().name)) {
				logger.Debug("El controlador ya existe en la lista de controladores disponibles");
				return;
			}
			
			this.controladoresDisponibles[controladorProxy.ice_getIdentity().name] =
				controladorProxy;
		}
		
		/// <summary>
		/// Extrae la IP de una representación de string de un Ice.Connection.
		/// </summary>
		/// <param name="iceConnection">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		public string ObtenerDireccionRemota(string iceConnection)
		{
			Match match = this.regexDireccionRemota.Match(iceConnection);
			
			return match.Groups["direccion"].ToString();
		}
		
		/// <summary>
		/// Genera un identificador interno de una terminal, dado su identificador externo (por
		/// ejemplo 'Terminal de Caja') y su Ice.Current. Actualmente lo que hace es concatenar
		/// el primero a la dirección IP de la terminal.
		/// </summary>
		/// <param name="identificador">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="iceCurrent">
		/// A <see cref="Ice.Current"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		public string GenerarIdentificadorTerminal(string identificador, Ice.Current iceCurrent)
		{
			string identificadorGenerado = identificador +
				" - " +
					this.ObtenerDireccionRemota(iceCurrent.con.ice_toString_());
			
			return identificadorGenerado;
		}
		
		/// <summary>
		/// Dado el identificador externo y el Ice.Current de una terminal, devuelve el objeto
		/// DatosTerminal.
		/// </summary>
		/// <param name="identificadorNoGenerado">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="iceCurrent">
		/// A <see cref="Ice.Current"/>
		/// </param>
		/// <returns>
		/// A <see cref="DatosTerminal"/>
		/// </returns>
		public DatosTerminal ObtenerTerminal(string identificadorNoGenerado, Ice.Current iceCurrent)
		{
			string identificadorReal = this.GenerarIdentificadorTerminal(identificadorNoGenerado,
			                                                             iceCurrent);
			
			if (!this.terminalesConectadas.ContainsKey(identificadorReal))
				throw new ConexionException("Terminal no conectada");
			
			return this.terminalesConectadas[identificadorReal];
		}
		
		#region ControladorTerminalesDisp_ implementation
		public override void Conectar(DatosTerminal ter, Ice.Current current__)
		{
			logger.Debug("Nuevo pedido de conexión");
			
			// Verifico que el identificador dado en DatosTerminal sea válido
			if (ter.identificador == null || ter.identificador.Trim().Equals(String.Empty)) {
				logger.Error("El identificador dado por la terminal es inválido");
				
				throw new ConexionException("El identificador dado por la terminal es inválido.");
			}
			
			// Genero el identificador de la terminal que quiere conectarse
			string identificador = this.GenerarIdentificadorTerminal(ter.identificador, current__);
			
			logger.Debug("Identificador de la terminal: " +
			             identificador);
			
			// Verifico que la terminal no se encuentre conectada ya
			logger.Debug("Verificando que la terminal no se encuentre conectada");
			if (this.terminalesConectadas.ContainsKey(identificador)) {
				logger.Error("La terminal ya se encuentra conectada");
				
				throw new ConexionException("La terminal con identificador '" + identificador +
				                            "' ya se encuentra conectada al servidor");
			}
			
			// Agrego la terminal a la lista de terminales conectadas
			this.terminalesConectadas[identificador] = ter;
			
			// Los addin suscriptos pueden hacer algo cuando la terminal se conecte
			foreach (ITerminalConectada itc in
			         AddinManager.GetExtensionObjects("/StratosServer/Core/TerminalConectada"))
				itc.TerminalConectada(identificador, ter);
			
			logger.Debug("Conexión de terminal establecida correctamente: '" +
			             identificador + "'");
		}
		
		public override void Desconectar(DatosTerminal ter, Ice.Current current__)
		{
			// Genero el identificador de la terminal que quiere conectarse
			string identificador = this.GenerarIdentificadorTerminal(ter.identificador, current__);
			
			logger.Debug("Pedido de desconexión de una terminal: " + identificador);
			
			logger.Debug("Verificando que la terminal se encuentre conectada");
			if (!this.terminalesConectadas.ContainsKey(identificador)) {
				logger.Error("La terminal no se encuentra conectada");
				
				throw new ConexionException("La terminal no se encuentra conectada al servidor");
			}
			
			// Agrego la terminal a la lista de terminales conectadas
			this.terminalesConectadas.Remove(identificador);
			
			// Los addin suscriptos pueden hacer algo cuando la terminal se desconecte
			foreach (ITerminalDesconectada itd in
			         AddinManager.GetExtensionObjects("/StratosServer/Core/TerminalDesconectada"))
				itd.TerminalDesconectada(identificador, ter);
			
			logger.Debug("Desconexión de terminal realizada correctamente: '" +
			             identificador + "'");
		}
		
		public override void ActualizarDatosTerminal(DatosTerminal ter, Ice.Current current__)
		{
			// Genero el identificador de la terminal que quiere conectarse
			string identificador = this.GenerarIdentificadorTerminal(ter.identificador, current__);
			
			logger.Debug("Pedido de actualización de datos de una terminal: '" +
			             identificador +
			             "'");
			
			logger.Debug("Verificando que la terminal se encuentre conectada");
			if (!this.terminalesConectadas.ContainsKey(identificador)) {
				logger.Error("La terminal no se encuentra conectada");
				
				throw new ConexionException("La terminal con identificador '" + identificador +
				                            "' no se encuentra conectada al servidor");
			}
			
			// Sobre escribo los datos anteriores de la terminal
			this.terminalesConectadas[identificador] = ter;
			
			// Los addin suscriptos pueden hacer algo cuando la terminal se conecte
			foreach (IDatosTerminalActualizados idta in
			         AddinManager.GetExtensionObjects("/StratosServer/Core/DatosTerminalActualizados"))
				idta.DatosTerminalActualizados(identificador, ter);
			
			logger.Debug("Actualización de datos de terminal realizada correctamente: '" +
			             identificador + "'");
		}

		public override Ice.ObjectPrx ObtenerProxyControlador(string idControlador, Ice.Current current__)
		{
			if (!this.controladoresDisponibles.ContainsKey(idControlador))
				throw new StratosException("El controlador no existe");
			
			return this.controladoresDisponibles[idControlador];
		}
		#endregion
	}
}
