// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
//
// This file is part of Stratos.
//
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using System.Threading;

using Mono.Addins;
using log4net;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using Castle.ActiveRecord.Framework.Config;

using Stratos.Common;

namespace StratosServer.Core
{
	public enum EstadoServidor {
		NoIniciado,
		LoggingIniciado,
		AddinManagerIniciado,
		ConfiguracionIniciado,
		ActiveRecordIniciado,
		MainIceIniciado,
		ServidorIniciado,
		ServidorDetenido
	};
	
	public class ServidorStratos
	{
		#region Campos
		/* El código "System.Reflection.MethodBase..." es igual que typeof(MainServer),
		 * pero es más portable, permitiendo copiar y pegar este código en otras clases
		 * sin cambiar nada
		 */
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);
		
		private string carpetaStratos;
		private string configuracionInterfaces;
		private bool esperarThreadIce;
		private EstadoServidor estadoServer;
		private StratosIce componenteIce;
		private IConfigurationSource configurationSource;
		
		private Thread threadIce;
		
		private static ServidorStratos instancia;
		#endregion
		
		#region Constructores
		public ServidorStratos() : this(null, null, true)
		{
		}
		
		public ServidorStratos(bool esperarThreadIce) :
			this(null, null, esperarThreadIce)
		{
		}
		
		public ServidorStratos(string configuracionInterfaces) : this(null, configuracionInterfaces, true)
		{
		}
		
		public ServidorStratos(IConfigurationSource configurationSource, string configuracionInterfaces)
			: this(configurationSource, configuracionInterfaces, true)
		{
		}
		
		public ServidorStratos(IConfigurationSource configurationSource,
		                       string configuracionInterfaces, bool esperarThreadIce)
		{
			this.esperarThreadIce = esperarThreadIce;
			this.configuracionInterfaces = configuracionInterfaces;
			this.configurationSource = configurationSource;
			this.estadoServer = EstadoServidor.NoIniciado;
			instancia = this;
			
			logger.Debug("Instancia de ServidorStratos creada");
		}
		#endregion
		
		#region Propiedades
		public static ServidorStratos Instancia {
			get { return instancia; }
		}
		
		public EstadoServidor Estado {
			get { return this.estadoServer; }
		}
		
		internal EstadoServidor NuevoEstado {
			set { this.estadoServer = value; }
		}
		
		public StratosIce ComponenteIce {
			get { return this.componenteIce; }
		}
		#endregion
		
		#region Métodos
		public void Iniciar(string[] args)
		{
			Iniciar(args, null);
		}
		
		public void Iniciar(string[] args, string carpetaStratos)
		{
			this.carpetaStratos = carpetaStratos;
			
//			// Configuro el logger
//			this.InicializarLog4Net();
//			this.estadoServer = EstadoServidor.LoggingIniciado;
			
			// Inicializo Mono.Addins
			this.InicializarMonoAddins();
			this.estadoServer = EstadoServidor.AddinManagerIniciado;
			
			// Inicializo la configuración
			this.InicializarConfiguracion();
			this.estadoServer = EstadoServidor.ConfiguracionIniciado;
			
			// Inicializo ActiveRecord
			this.InicializarActiveRecord(args);
			this.estadoServer = EstadoServidor.ActiveRecordIniciado;
		
			// Ejecuto el código que maneja las conexiones Ice en un thread aparte.
			// Este thread es el que se encarga de poner en 'estadoServer'
			// el valor ServidorIniciado.
			logger.Debug("Iniciando Ice (comunicación por red)");
			this.threadIce = new Thread(delegate() {
				this.componenteIce = new StratosIce(this, this.configuracionInterfaces);
				this.componenteIce.Iniciar(args);
			});
			this.threadIce.Start();
			
			logger.Debug("Esperando a que finalice el thread de Ice (comunicación por red)");
			
			if (esperarThreadIce)
			{
				threadIce.Join();
				logger.Debug("Thread de Ice (comunicación por red) finalizado");
			}
		}

		public void Detener()
		{
			// Guardo la configuración al salir
			logger.Debug("Guardando configuración");
			Configuracion.StratosServerConfig.Save();
			
			// Detengo el thread de Ice y espero a que el thread finalice
			this.componenteIce.Detener();
			this.threadIce.Join();
		}
		#endregion
		
		#region Métodos inicializadores
		private void InicializarLog4Net()
		{
			log4net.Config.XmlConfigurator.Configure();
		}
		
		private void InicializarMonoAddins()
		{
			// Inicializo la base de addins y busco por cambios en los mismos
			logger.Debug("Inicializando Mono.Addins");
			AddinManager.Initialize(CrearPathLocal("addins_db"));

			logger.Debug("Buscando addins nuevos o modificados");
			AddinManager.Registry.Update(null);
			
			// Me suscribo a los eventos de carga, descarga y error de los addins
			AddinManager.AddinLoadError += OnLoadError;
			AddinManager.AddinLoaded += OnLoad;
			AddinManager.AddinUnloaded += OnUnload;
			AddinManager.ExtensionChanged += OnExtensionChange;
		}
		
		private void InicializarConfiguracion()
		{
			logger.Debug("Iniciando sistema de configuración");
			Configuracion.StratosServerConfig.Inicializar();
		}
		
		private void InicializarActiveRecord(string[] args)
		{
			logger.Debug("Inicializando ActiveRecord");
			
			IConfigurationSource source = null;
			
			/* Me fijo si un objeto IConfigurationSource fue seteado (por ejemplo, para las pruebas.
			 * de unidad). */
			if (this.configurationSource == null)
				source = new XmlConfigurationSource(CrearPathLocal("ActiveRecord.config.xml"));
			else
				source = this.configurationSource;
			
			Assembly assemblyModeloObjetos =
				Assembly.GetAssembly(typeof(Stratos.ModeloObjetos.Cliente));
			
			if (!ActiveRecordStarter.IsInitialized)
				ActiveRecordStarter.Initialize(assemblyModeloObjetos, source);
			
			if (args.Contains("--create-schema"))
				ActiveRecordStarter.CreateSchema();
		}
		#endregion
		
		#region Eventos de Mono.Addins
		private void OnLoadError (object s, AddinErrorEventArgs args)
		{
			logger.Error("Error al cargar un add-in: (" + args.AddinId + ") " + args.Message);
		}
		
		private void OnLoad (object s, AddinEventArgs args)
		{
			logger.Debug("Add-in cargado: " + args.AddinId);
		}
		
		private void OnUnload (object s, AddinEventArgs args)
		{
			logger.Debug("Add-in descargado: " + args.AddinId);
		}

		private void OnExtensionChange (object s, ExtensionEventArgs args)
		{
			logger.Debug("Add-in cambiado: " + args.Path);
		}
		#endregion
		
		#region Otros métodos
		private string CrearPathLocal(string path)
		{
			if (carpetaStratos != null)
				return System.IO.Path.Combine(carpetaStratos, path);
			else
				return path;
		}
		#endregion
	}
}
