// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Net.NetworkInformation;
using System.Threading;

using log4net;
using Mono.Addins;
using Castle.DynamicProxy;

using Stratos.Servicios;
using Stratos.Common;
using StratosServer.Controladores;
using StratosServer.Core.Interceptores;

namespace StratosServer.Core
{
	public class StratosIce : Ice.Application
	{
		#region Campos
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);
		
		private static readonly ProxyGenerator generator = new ProxyGenerator();
		
		private Ice.ObjectAdapter iceObjectAdapter;
		private ServidorStratos servidorStratos;
		private string configuracionInterfaces;
		#endregion
		
		#region Constructores
		public StratosIce(ServidorStratos servidorStratos) :
			this(servidorStratos, Configuracion.ConexionElement.Instancia.InterfacesConfiguracion)
		{
		}
		
		public StratosIce(ServidorStratos servidorStratos, string configuracionInterfaces)
		{
			this.servidorStratos = servidorStratos;
			
			if (String.IsNullOrEmpty(configuracionInterfaces))
				this.configuracionInterfaces = Configuracion.ConexionElement.Instancia.InterfacesConfiguracion;
			else
				this.configuracionInterfaces = configuracionInterfaces;
			
			logger.Debug("Suscribiendo a evento NetworkAddressChanged");
			
			// Cuando una dirección de alguna red cambia o cambia la disponibilidad
			// del dispositivo de red, actualizo la información de los endpoints
			// FIXME: Estos eventos no funcionan en Mono todavía.
			NetworkChange.NetworkAddressChanged += delegate {
				logger.Debug("Alguna dirección en una interfaz de red cambió. Actulizando los endpoints publicados");
				this.IceObjectAdapter.refreshPublishedEndpoints();
			};
			
			NetworkChange.NetworkAvailabilityChanged += delegate {
				logger.Debug("Alguna dirección en una interfaz de red cambió. Actulizando los endpoints publicados");
				this.IceObjectAdapter.refreshPublishedEndpoints();
			};
			
			logger.Debug("Instancia de StratosIce creada");
		}
		#endregion
		
		public void Iniciar(string[] args)
		{
			logger.Debug("Ejecutando Ice.Application.main(args)");
			this.main(args);
		}
		
		public void Detener()
		{
			// Si el communicator está activo, lo bajo.
			Ice.Communicator com = StratosIce.communicator();
			
			if (com != null)
			{
				com.shutdown();
				com.waitForShutdown();
			}
			
			this.servidorStratos.NuevoEstado = EstadoServidor.ServidorDetenido;
		}
		
		private int RunInterno(string[] args)
		{
			logger.Debug("Ejecutando RunInterno");
			
			// Handler para bajar el communicator
			logger.Debug("Configurando handler: shutdownOnInterrupt");
			shutdownOnInterrupt();
			
			// Sirvo el Controlador de Terminales
			logger.Debug("Creando instancia de Controlador de Terminales");
			Ice.Object obj = ControladorTerminalesI.Instancia;
			
			logger.Debug("Agregando el Controlador de Terminales al Ice.ObjectAdapter");
			this.IceObjectAdapter.add(
				obj,
				StratosIce.communicator().stringToIdentity("ControladorTerminales")
			);
			
			/* Ejecuto todos los addin que van a brindar un objeto que actuará de servidor. */
			logger.Debug("Cargando todos los addins en modo Servicio Servidor");
			foreach (IServicioServidor servicioServidor in 
			         	AddinManager
			         		.GetExtensionObjects ("/StratosServer/Core/ServicioServidor", true)) {
				
				servicioServidor.InstanciaServidorStratos = this.servidorStratos;
				
				object servant = servicioServidor.Servant;
				
				// Verifico que el tipo del servant sea Ice.ObjectImpl
				if (!(servant is Ice.ObjectImpl)) {
					logger.Error("El servant no es de tipo Ice.ObjectImpl: " +
					             servicioServidor.IceIdentity);
					continue;
				}
				
				// Creo un proxy para el servant. Le agrego algunos interceptores básicos
				var loggerInterceptor = new LoggerInterceptor(servant.GetType());
				
				var proxyOptions = new ProxyGenerationOptions();
				proxyOptions.BaseTypeForInterfaceProxy = servant.GetType();
				
				var proxy = generator
					.CreateInterfaceProxyWithTarget(servicioServidor.InterfaceServicio,
					                                servant,
					                                proxyOptions,
					                                loggerInterceptor);
				
				/* Obtengo un proxy para el servant */
				Ice.ObjectPrx servantProxy = servicioServidor.uncheckedCast(
					this.IceObjectAdapter.add(
						(proxy as Ice.Object),
						StratosIce.communicator().stringToIdentity(servicioServidor.IceIdentity)
					)
				) as Ice.ObjectPrx;
				
				// Verifico que el proxy se haya podido crear
				if (servantProxy == null) {
					logger.Error("No se pudo obtener el proxy para el servicio '" +
					             servicioServidor.IceIdentity + "'. El objeto posiblmente no es " +
					             	"un Ice.ObjectPrx");
					continue;
				}
				
				// Agrego el proxy al Controlador de Terminales
				ControladorTerminalesI.Instancia.AgregarControlador(servantProxy);
			}
			
			// Activo el Ice.ObjectAdapter
			logger.Debug("Activando Ice.ObjectAdapter");
			this.IceObjectAdapter.activate();
			
			// Espero hasta que el communicator sea bajado. Esto hace
			// que quede esperando las conexiones
			
			logger.Debug("Componente Ice en espera");
			this.servidorStratos.NuevoEstado = EstadoServidor.MainIceIniciado;
			this.servidorStratos.NuevoEstado = EstadoServidor.ServidorIniciado;
			communicator().waitForShutdown();
			
			if (interrupted())
				logger.Debug("Componente Ice interrumpido");
			
			logger.Debug("Cerrando componente Ice");
			
			return 0;
		}
		
		public override int run (string[] args)
		{
			try {
				return this.RunInterno(args);
			}
			catch(System.Exception exception) {
				logger.ErrorFormat("Excepción al inicializar Ice: {0}", exception.ToString());
				
				return 1;
			}
		}
		
		private Ice.ObjectAdapter IceObjectAdapter {
			get {
				if (this.iceObjectAdapter == null) {
					logger.Debug("ObjectAdapter no creado. Creando uno nuevo");
					
					logger.DebugFormat("Configuracion de interfaces: {0}", this.configuracionInterfaces);
					
					try {
						this.iceObjectAdapter = communicator()
							.createObjectAdapterWithEndpoints("StratosServer",
							                                  this.configuracionInterfaces);
					}
					catch (Ice.SocketException ex) {
						logger
							.ErrorFormat("Error al crear el ObjectAdapter. Verificar configuración de red " +
								"de Stratos:\n{0}", ex.ToString());
						throw;
					}
					catch (System.Exception ex) {
						logger.ErrorFormat("Error al crear el ObjectAdapter:\n{0}", ex.ToString());
						throw;
					}
				}
				
				return this.iceObjectAdapter;
			}
		}
		
		public ControladorTerminalesPrx ControladorTerminalesProxy {
			get {
				Ice.ObjectPrx proxy =
					communicator().stringToProxy("ControladorTerminales");
				
				return ControladorTerminalesPrxHelper.checkedCast(proxy);
			}
		}
	}
}
