// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;

using NHibernate.Criterion;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using Castle.ActiveRecord.Queries;
using log4net;

using Stratos.Common;
using StratosServer.Controladores;
using Stratos.Common.Excepciones;
using Stratos.QueryObjects;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace StratosServer.Core
{
	public static class StratosServerUtil
	{
		private static readonly ILog logger =
			LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod()
			                     .DeclaringType);
		
		#region Métodos de generación de mensajes a partir de excepciones
		/// <summary>
		/// Dada una excepción del tipo System.Exception, genera un mensaje de error
		/// buscando también las Inner Exceptions.
		/// </summary>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		public static string ObtenerMensajesExcepcion(System.Exception excepcion)
		{
			StringBuilder mensaje = new StringBuilder();
			System.Exception exTemp = excepcion;
			int i = 1;
			
			// Proceso el mensaje de la excepción y todas sus excepciones internas
			do {
				if (exTemp.Message != null && !exTemp.Message.Equals(String.Empty)) {
					mensaje.Append(i.ToString() + ": " + exTemp.ToString());
					mensaje.Append("\nStacktrace:\n" + exTemp.StackTrace);
					
					mensaje.Append("---\n\n");
				}
				
				exTemp = exTemp.InnerException;
				i++;
			}
			while (exTemp != null);
			
			return mensaje.ToString();
		}
		
		/// <summary>
		/// Genera un mensaje de error dada una excepción del tipo ActiveRecordValidationException.
		/// </summary>
		/// <param name="excepcion">
		/// A <see cref="ActiveRecordValidationException"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		public static string ObtenerMensajesExcepcion(ActiveRecordValidationException excepcion)
		{
			return ObtenerMensajesValidacion(excepcion.ValidationErrorMessages);
		}
		
		/// <summary>
		/// Dado un objeto ActiveRecordValidationBase, genera un mensaje de error.
		/// </summary>
		/// <param name="objeto">
		/// A <see cref="ActiveRecordValidationBase"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		public static string ObtenerMensajesValidacion(ActiveRecordValidationBase objeto)
		{
			return ObtenerMensajesValidacion(objeto.ValidationErrorMessages);
		}
		
		public static string ObtenerMensajesValidacion<T>(ActiveRecordValidationBase<T> objeto)
			where T : class
		{
			return ObtenerMensajesValidacion(objeto.ValidationErrorMessages);
		}
		
		/// <summary>
		/// Dado un array de mensajes de error, crea un mensaje único con todos ellos
		/// en forma de lista.
		/// </summary>
		/// <param name="mensajesError">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		private static string ObtenerMensajesValidacion(string[] mensajesError)
		{
			StringBuilder mensaje = new StringBuilder();
			
			foreach (string errStr in mensajesError) {
				mensaje.Append(" - ");
				mensaje.Append(errStr);
				mensaje.Append("\n");
			}
			
			return mensaje.ToString();
		}
		#endregion
		
		#region Métodos comunes para consultas a la Base de Datos
		#region Consultas paginables
		/// <summary>
		/// Verifica que una ConsultaPaginable sea válida (que haya sido establecida en la
		/// terminal).
		/// </summary>
		/// <param name="consultaPaginable">
		/// A <see cref="ConsultaPaginable"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		private static bool ConsultaPaginableValida(ConsultaPaginable consultaPaginable)
		{
			if (consultaPaginable == null ||
			    	consultaPaginable.criteriosOrdenacion == null ||
			    	consultaPaginable.criteriosOrdenacion.Count == 0)
			    return false;
			
			return true;
		}
		
		/// <summary>
		/// Prepara una consulta del tipo DetachedCriteria para ser paginable.
		/// </summary>
		/// <param name="consulta">
		/// A <see cref="DetachedCriteria"/>
		/// </param>
		/// <param name="consultaPaginable">
		/// A <see cref="ConsultaPaginable"/>
		/// </param>
		public static void PrepararConsultaPaginable(DetachedCriteria consulta,
		                                             ConsultaPaginable consultaPaginable)
		{
			logger.Debug("Preparando ConsultaPaginable");
			
			// Antes de continuar verifico que la consulta paginable sea válida
			if (!ConsultaPaginableValida(consultaPaginable))
				return;
			
			// Agrego a la consulta el criterio de ordenación
			
			foreach (CriterioOrdenacion criterioOrdenacion in consultaPaginable.criteriosOrdenacion) {
				Order order = null;
				
				if (criterioOrdenacion.sentidoOrd == SentidoOrdenacion.Asc)
					order = Order.Asc(criterioOrdenacion.campoOrdenacion);
				else
					order = Order.Desc(criterioOrdenacion.campoOrdenacion);
				
				consulta.AddOrder(order);
			}
			
			// Establezo el OFFSET y LIMIT de acuerdo a los datos de la consulta paginable
			consulta.SetFirstResult(consultaPaginable.tamanioPagina * consultaPaginable.pagina);
			consulta.SetMaxResults(consultaPaginable.tamanioPagina);
			
			logger.Debug("ConsultaPaginable preparada correctamente");
		}
		
		/// <summary>
		/// Prepara una consulta del tipo HqlBasedQuery para ser paginable.
		/// </summary>
		/// <param name="consulta">
		/// A <see cref="HqlBasedQuery"/>
		/// </param>
		/// <param name="consultaPaginable">
		/// A <see cref="ConsultaPaginable"/>
		/// </param>
		public static void PrepararConsultaPaginable(HqlBasedQuery consulta,
		                                             ConsultaPaginable consultaPaginable)
		{
			// Antes de continuar verifico que la consulta paginable sea válida
			if (!ConsultaPaginableValida(consultaPaginable))
				return;
			
			// Agrego a la consulta el criterio de ordenación
			string criterioOrdenacion = String.Empty;
			
			if (!consulta.Query.ToLower().Contains("order by"))
				criterioOrdenacion = " order by ";
			else
				criterioOrdenacion = ", ";
			
			criterioOrdenacion +=
				consultaPaginable.criteriosOrdenacion[0].campoOrdenacion + " " +
				consultaPaginable.criteriosOrdenacion[0].sentidoOrd.ToString().ToLower();
			
			for (int i=1; i<consultaPaginable.criteriosOrdenacion.Count; i++) {
				criterioOrdenacion += ", " +
					consultaPaginable.criteriosOrdenacion[i].campoOrdenacion + " " +
						consultaPaginable.criteriosOrdenacion[i].sentidoOrd.ToString().ToLower();
			}
			
			consulta.Query += criterioOrdenacion;
			
			// Establezo el OFFSET y LIMIT de acuerdo a los datos de la consulta paginable
			consulta.SetQueryRange(consultaPaginable.tamanioPagina * consultaPaginable.pagina,
			                       consultaPaginable.tamanioPagina);
		}
		#endregion
		
		#region Consultas HQL
		/// <summary>
		/// Método para centralizar las consultas utilizando HQL.
		/// </summary>
		/// <param name="consulta">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="consultaPaginable">
		/// A <see cref="ConsultaPaginable"/>
		/// </param>
		/// <returns>
		/// A <see cref="R"/>
		/// </returns>
		public static R[] BuscarHQLGenerico<R,T>(string consulta, ConsultaPaginable consultaPaginable)
		{
			logger.Debug("Consulta HQL recibida: " + consulta);
			
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				T[] resultados = null;
				
				try {
					SimpleQuery<T> simpleQuery = new SimpleQuery<T>(consulta);
					
					// Preparo la consulta para ser paginable si es que el usuario así lo quiso
					PrepararConsultaPaginable(simpleQuery, consultaPaginable);
					
					/* Seteo un Transformer para evitar resultados duplicados cuando se hacen
					 * joins. */
					simpleQuery.SetResultTransformer(new NHibernate.Transform.DistinctRootEntityResultTransformer());
					
					logger.Debug("Ejecutando consulta HQL");
					resultados = simpleQuery.Execute();
					logger.Debug("Consulta HQL ejecutada correctamente");
					
					// Transformo los resultados a DTO
					logger.Debug("Convirtiendo resultados a Dto");
					R[] resultadosDto = ConvertidorDto.ToDto<T,R>(resultados);
					logger.Debug("Resultados convertidos a Dto correctamente");
				
					return resultadosDto;
				}
				catch (ActiveRecordException ex) {
					logger.Debug("Excepción arrojada: " + ex.ToString());
					
					StratosException exception = new StratosException();
					exception.mensaje = StratosServerUtil.ObtenerMensajesExcepcion(ex);
					throw exception;
				}
			}
		}
		#endregion
		
		#region Consultas tipo QueryObjects
		/// <summary>
		/// Este método centraliza la búsqueda con QueryObjects. Especificando los tipos de
		/// objetos correctos, se puede procesar cualquier consulta de cualquier tipo.
		/// </summary>
		/// <param name="queryObject">
		/// El QueryObject correspondiente (PreFacturaQueryObject, ClienteQueryObject, etc).
		/// </param>
		/// <param name="consultaPaginable">
		/// La consulta paginable.
		/// </param>
		/// <returns>
		/// Devuelve un array con los objetos DTO correspondientes.
		/// </returns>
		public static TDto[] BuscarQueryObjects<T,TDto>(QueryObject queryObject,
		                                                ConsultaPaginable consultaPaginable)
			where TDto : Ice.Object
		{
			logger.Debug("Ejecutando BuscarQueryObjects (" + queryObject.GetType().FullName + ")");
			
			// Verifico que haya un objeto en la lista de criterios
			if (queryObject.listaCriteriosComplejos == null)
				queryObject.listaCriteriosComplejos = new List<CriterioComplejo>();
			
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				// Preparo la lista de ICriterion con todos los criterios del QueryObject
				List<ICriterion> criteriosBusqueda = new List<ICriterion>();
				
				foreach (CriterioComplejo criterio in queryObject.listaCriteriosComplejos)
					criteriosBusqueda.Add(ProcesarCriterio(criterio));
				
				// Creo la consulta con los criterios establecidos antes
				DetachedCriteria detachedCriteria = DetachedCriteria.For<T>();
				foreach (ICriterion c in criteriosBusqueda)
					detachedCriteria.Add(c);
				
				// Preparo la consulta para ser paginable si es que el usuario así lo quiso
				PrepararConsultaPaginable(detachedCriteria, consultaPaginable);
				
				// Realizo la búsqueda concretamente
				T[] resultado = null;
				TDto[] resultadosDto = null;
				
				try {
					logger.Debug("Realizando consulta QueryObjects del tipo '" + typeof(T).FullName + "'");
					resultado = ActiveRecordBase<T>.FindAll(detachedCriteria);
					logger.Debug("Consulta realizada correctamente");
					
					resultadosDto = ConvertidorDto.ToDto<T,TDto>(resultado);
				}
				catch (System.Exception excepcion) {
					TratarExcepcion(excepcion);
				}
				
				return resultadosDto;
			}
		}
		
		/// <summary>
		/// Procesa un CriterioComplejo y devuelve un ICriterion para agregar a una consulta
		/// en ActiveRecord.
		/// </summary>
		/// <param name="criterioComplejo">
		/// A <see cref="CriterioComplejo"/>
		/// </param>
		/// <returns>
		/// A <see cref="ICriterion"/>
		/// </returns>
		public static ICriterion ProcesarCriterio(CriterioComplejo criterioComplejo)
		{
			Junction criterio;
			
			// Me fijo si es una condición AND u OR
			if (criterioComplejo is CriterioAnd)
				criterio = Expression.Conjunction();
			else
				criterio = Expression.Disjunction();
			
			foreach (CriterioGenerico criterioGenerico in criterioComplejo.criterios) {
				if (criterioGenerico is CriterioSimple)
					criterio.Add(ProcesarCriterio(criterioGenerico as CriterioSimple));
				else
					criterio.Add(ProcesarCriterio(criterioGenerico as CriterioComplejo));
			}
			
			return criterio;
		}
		
		/// <summary>
		/// Procesa un CriterioSimple.
		/// </summary>
		/// <param name="criterioSimple">
		/// A <see cref="CriterioSimple"/>
		/// </param>
		/// <returns>
		/// A <see cref="ICriterion"/>
		/// </returns>
		public static ICriterion ProcesarCriterio(CriterioSimple criterioSimple)
		{
			SimpleExpression criterio = null;
			
			// Proceso el valor del criterio
			object valorFinal = null;
			
			// Me fijo si es un integer
			if (StratosUtil.IsInt(criterioSimple.valor))
				valorFinal = Int32.Parse(criterioSimple.valor);
			
			// Me fijo si es un double
			else if (StratosUtil.IsDouble(criterioSimple.valor))
				valorFinal = Double.Parse(criterioSimple.valor);
			
			// Me fijo si es un bool
			else if (StratosUtil.IsBool(criterioSimple.valor))
				valorFinal = bool.Parse(criterioSimple.valor);
			
			// Me fijo si es una enumeración
			else if (criterioSimple.tipoValor != null && criterioSimple.tipoValor != String.Empty) {
				Type tipo;
				
				/* Para obtener la enumeración, cargo el assembly Stratos.ModeloDto.dll. En dicho
				 * assembly se encuentran todas las enumeraciones que pueden ser utilizadas en una
				 * consulta (por ej, FacturaTipo, TipoResponsable). Busco allí el tipo que fue
				 * especificado en el criterio simple. */
				Assembly modeloDto = Assembly.LoadFrom("Stratos.ModeloDto.dll");
				tipo = modeloDto.GetType("Stratos.ModeloDto." + criterioSimple.tipoValor);
				
				// Si no lo encuentro allí, lo busco en los assembly base del framework .NET
				if (tipo == null)
					tipo = Type.GetType(criterioSimple.tipoValor);
				
				/* FIXME: En este caso, hay que ver qué pasa si no lo encuentra aquí tampoco. Eso
				 * puede suceder si está mal especificado el tipo y el valor. */
				valorFinal = Enum.Parse(tipo, criterioSimple.valor);
			}
			
			/* En caso de que no sea ninguno de los anteriores, es posible aún que sea una
			 * representación especial de algún otro tipo, como por ejemplo un long. */
			else {
				object valorEspecial;
				
				// Me fijo si el tipo es un long.
				if (EsLong(criterioSimple.valor, out valorEspecial))
					valorFinal = valorEspecial;
				
				// En el caso de que no sea ningún tipo anterio, lo trato como string
				else
					valorFinal = criterioSimple.valor;
			}
			
			/* Proceso el operador del criterio y lo transformo a su equivalente de
			 * NHibernate */
			if (criterioSimple.operador == TipoOperador.Igual)
				criterio = Expression.Eq(criterioSimple.propiedad, valorFinal);
			
			else if (criterioSimple.operador == TipoOperador.LikeMatchAnywhere)
				criterio = Expression.Like(criterioSimple.propiedad, valorFinal.ToString(),
				                           MatchMode.Anywhere);
			
			return criterio;
		}
		
		/// <summary>
		/// Esta función revisa si una cadena es un long, con formato interno (uno utilizado
		/// internamente por Stratos). Un long, para una Query Object, se _debe_ especificar
		/// como un string finalizando con una "L". Por ej: 12L
		/// </summary>
		/// <param name="cadena">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="valorLong">
		/// Variable donde se colocará el valor tipo long en caso de que cadena lo sea.
		/// </param>
		/// <returns>
		/// Retorna true en caso de que cadena se refiera a una representación interna de long.
		/// </returns>
		private static bool EsLong(string cadena, out object valorLong)
		{
			valorLong = null;
			
			string numero = cadena.Substring(0, cadena.Length - 1);
			
			if (cadena.EndsWith("L") && StratosUtil.IsLong(numero)) {
				valorLong = long.Parse(numero);
				return true;
			}
			
			return false;
		}
		#endregion
		
		#region Otros tipos de consultas
		/// <summary>
		/// Dado un objeto del tipo Ice.Object (ProductoDto, ClienteDto, etc), y el id
		/// correspondiente a su clave primaria, refresca los datos del objeto consultando
		/// a la base de datos por la última versión del mismo, y lo devuelve.
		/// </summary>
		/// <param name="objeto">
		/// A <see cref="Ice.Object"/>
		/// </param>
		/// <param name="id">
		/// A <see cref="System.Int64"/>
		/// </param>
		/// <returns>
		/// A <see cref="TDto"/>
		/// </returns>
		public static TDto Refrescar<T,TDto>(Ice.Object objeto, long id)
			where T : ActiveRecordBase
		{
			logger.Debug("Ejecutando Refrescar");
			
			// Verifico que el tipo de objeto sea correcto
			if (objeto == null || !(objeto is TDto))
				throw new StratosException("El objeto a refrescar debe ser de tipo '" +
				                           typeof(TDto).FullName + "'");
			
			// Vuelvo a traer el objeto de la base de datos
			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {
				
				T objetoPersistente = ActiveRecordBase<T>.TryFind(id);
				
				if (objetoPersistente == null) {
					logger.Debug("No se ha encontrado el objeto '" + typeof(TDto).FullName +
					             "' por refrescar: " + id);
					
					ObjetoInexistenteException ex = new ObjetoInexistenteException();
					ex.mensaje = "El objeto de tipo '" + typeof(T).FullName +
						"' no existe en la base de datos";
					
					throw ex;
				}
				
				/* Si se trata de un documento de tipo IDocumentoItems, inicializo los
				 * items del mismo. */
				if (objetoPersistente is IDocumentoItems) {
					IDocumentoItems documentoItems = objetoPersistente as IDocumentoItems;
					NHibernate.NHibernateUtil.Initialize(documentoItems.ItemsGenericos);
				}
				
				logger.Debug("Refrescar ejecutado correctamente");
				
				return ConvertidorDto.ToDto<T,TDto>(objetoPersistente);
			}
		}
		#endregion
		
		#region Métodos comunes de ABM
		/// <summary>
		/// Verifica que el objeto Ice.Object sea del tipo especificado T.
		/// En caso de ser así, devuelve ese tipo casteado.
		/// </summary>
		/// <param name="objeto">
		/// A <see cref="Ice.Object"/>
		/// </param>
		/// <returns>
		/// A <see cref="T"/>
		/// </returns>
		public static T TipoCorrecto<T>(Ice.Object objeto) where T : Ice.Object
		{
			if (!(objeto is T)) {
				string mensaje = "La instancia no es de tipo '" + typeof(T).FullName + "'";
				
				logger.Error(mensaje);
				throw new ObjetoInvalidoException(mensaje);
			}
			
			return ((T) objeto);
		}
		
		/// <summary>
		/// Dado el id y la versión de un objeto IVersionable, se encarga de gestionar
		/// el chequeo de versión del objeto. Si el objeto tiene un id mayor a cero
		/// (o sea, si debería existir en la base de datos), verifica que exista y que
		/// la versión coincida. Si no coincide, hace un rollback de la transacción y
		/// arroja una excepción.
		/// </summary>
		/// <param name="id">
		/// A <see cref="System.Int64"/>
		/// </param>
		/// <param name="version">
		/// A <see cref="System.Int32"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public static void VersionCorrecta<T>(long id,
		                                      int version,
		                                      TransactionScope transactionScope) where T : IVersionable
		{
			// Si el id del objeto es menor o igual a cero, entonces todavía no está guardado.
			// En ese caso retorno.
			if (id <= 0)
				return;
			
			// Obtengo el objeto
			T elementoGuardado = ActiveRecordBase<T>.TryFind(id);
			
			// Verifico que existe, y si es asi, que tenga la misma versión
			if (elementoGuardado == null ||
			    	(elementoGuardado != null && elementoGuardado.Version != version)) {
				
				logger.ErrorFormat("El objeto de tipo {0} no existe o ya fue modificado " +
					"por otra terminal en la base de datos.", typeof(T).Name);
				
				// Hago un rollback y arrojo una excepción
				if (transactionScope != null)
					transactionScope.VoteRollBack();
				
				string mensaje =
					String.Format("El objeto de tipo {0} no existe o ya fue modificado " +
						"por otra terminal en la base de datos.", typeof(T).Name);
				
				ObjetoVersionIncorrectaException excepcion =
					new ObjetoVersionIncorrectaException();
				excepcion.mensaje = mensaje;
				
				throw excepcion;
			}
		}
		
		public static void VersionCorrecta<T>(long id, int version) where T : IVersionable
		{
			VersionCorrecta<T>(id, version, null);
		}
		
		/// <summary>
		/// Dada una excepción del tipo System.Exception, revisa su tipo y de acuerdo al mismo
		/// genera una excepción StratosExcepcion, según la excepción dada.
		/// El objetivo del método es centralizar el tratamiento de las excepciones.
		/// </summary>
		/// <param name="excepcion">
		/// A <see cref="System.Exception"/>
		/// </param>
		public static void TratarExcepcion(System.Exception excepcion)
		{
			if (excepcion is ActiveRecordValidationException) {
				ActiveRecordValidationException e =
					(ActiveRecordValidationException) excepcion;
				
				string errorExcepcion = StratosServerUtil.ObtenerMensajesExcepcion(e);
				logger.Error("Error de validación: " + errorExcepcion);
				
				ObjetoInvalidoException ex = new ObjetoInvalidoException();
				ex.mensaje = errorExcepcion;
				throw ex;
			}
			else if (excepcion is ActiveRecordException) {
				ActiveRecordException e = (ActiveRecordException) excepcion;
				
				string errorExcepcion = StratosServerUtil.ObtenerMensajesExcepcion(e);
				logger.Error("Error de ActiveRecord: " + errorExcepcion);
				
				StratosException ex = new StratosException();
				ex.mensaje = errorExcepcion;
				throw ex;
			}
			else {
				string errorExcepcion = StratosServerUtil.ObtenerMensajesExcepcion(excepcion);
				logger.Error("Error desconocido: " + errorExcepcion);
				
				StratosException ex = new StratosException();
				ex.mensaje = errorExcepcion;
				throw ex;
			}
		}
		
		/// <summary>
		/// Dado un objeto del tipo ActiveRecordValidationBase (Cliente, Factura, etc) y
		/// una transacción, se encarga de guardar el objeto en la base de datos y tratar
		/// todas las excepciones posibles, además de gestiionar correctamente la transacción
		/// utilizada.
		/// El objetivo es centralizar esta operación común de creación, tratando las excepciones
		/// posibles y el estado de la transacción.
		/// </summary>
		/// <param name="objeto">
		/// A <see cref="ActiveRecordValidationBase"/>
		/// </param>
		/// <param name="transactionScope">
		/// A <see cref="TransactionScope"/>
		/// </param>
		public static void CrearObjeto<T>(ActiveRecordValidationBase<T> objeto,
		                                  TransactionScope transactionScope)
			where T : class
		{
			try {
				logger.Debug("Creando objeto '" + typeof(T).FullName + "'");
				
				objeto.Create();
				
				logger.Debug("Objeto '" + typeof(T).FullName + "' creado correctamente");
			}
			catch (System.Exception ex) {
				transactionScope.VoteRollBack();
				
				// Llamo a TratarExcepcion para que lance una excepción correcta según el tipo
				// de la misma
				TratarExcepcion(ex);
			}
		}
		
		/// <summary>
		/// Método para centralizar la actualización de un objeto tipo ActiveRecordValidationBase
		/// Idem a CrearObjeto.
		/// </summary>
		/// <param name="objeto">
		/// A <see cref="ActiveRecordValidationBase"/>
		/// </param>
		/// <param name="transactionScope">
		/// A <see cref="TransactionScope"/>
		/// </param>
		public static void ActualizarObjeto<T>(ActiveRecordValidationBase<T> objeto,
		                                       TransactionScope transactionScope)
			where T : class
		{
			try {
				logger.Debug("Actualizando objeto '" + typeof(T).FullName + "'");
				
				/* Utilizo el método UpdateAndFlush ya que si hay un error de validación al
				 * actualizar, salta aca. Si uso Update salta cuando se hace un Dispose de la
				 * transacción. */
				objeto.UpdateAndFlush();
				
				logger.Debug("Objeto '" + typeof(T).FullName + "' actualizado correctamente");
			}
			catch (System.Exception ex) {
				transactionScope.VoteRollBack();
				
				TratarExcepcion(ex);
			}
		}
		
		/// <summary>
		/// Método para centralizar la eliminación de un objeto tipo ActiveRecordValidationBase
		/// (Cliente, Producto...).
		/// Idem a CrearObjeto.
		/// </summary>
		/// <param name="objeto">
		/// A <see cref="ActiveRecordValidationBase"/>
		/// </param>
		/// <param name="transactionScope">
		/// A <see cref="TransactionScope"/>
		/// </param>
		public static void EliminarObjeto<T>(ActiveRecordValidationBase<T> objeto,
		                                     TransactionScope transactionScope)
			where T : class
		{
			try {
				logger.Debug("Eliminando objeto '" + typeof(T).FullName + "'");
				
				objeto.Delete();
				
				logger.Debug("Objeto '" + typeof(T).FullName + "' eliminado correctamente");
			}
			catch (System.Exception ex) {
				transactionScope.VoteRollBack();
				
				TratarExcepcion(ex);
			}
		}
//		
//		/// <summary>
//		/// Verifica si ya existe una Persona (Cliente o Proveedor) ingresada con
//		/// el cuit pasado como parámetro. Devuelve 'true' si el cuit está disponible
//		/// y 'false' si no lo está. Si se dá este ultimo caso, se setea en el parámetro
//		/// out la persona ingresada con el cuit.
//		/// </summary>
//		/// <param name="cuit">
//		/// A <see cref="CUIT"/>
//		/// </param>
//		/// <param name="personaConCuit">
//		/// A <see cref="T"/>
//		/// </param>
//		/// <returns>
//		/// A <see cref="System.Boolean"/>
//		/// </returns>
//		public static bool CuitDisponible<T>(CUIT cuit, out T personaConCuit)
//			where T : Persona<T>
//		{
//			using (SessionScope sessionScope = new SessionScope(FlushAction.Never)) {				
//				personaConCuit = null;
//				
//				// Creo una consulta para ver si existe una persona con este cuit
//				// ya ingresado en la base de datos.
//				SimpleQuery<T> consulta =
//					new SimpleQuery<T>("from " + typeof(T).FullName +
//						" p where p.Cuit.ValorNumerico = :cuit");
//				
//				// Seteo el cuit en la consulta
//				consulta.SetParameter("cuit", cuit.ValorNumerico);
//				
//				// Ejecuto la consulta
//				T[] resultado = consulta.Execute();
//				
//				// Si no hay resultados, retorno true
//				if (resultado.Length <= 0)
//					return true;
//				
//				// Si hay resultados, seteo el parametro out y retorno false
//				personaConCuit = resultado[0];
//				
//				return false;
//			}
//		}
		#endregion
		#endregion
	}
}
