// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using NUnit.Framework;

using Stratos.Common;

namespace Stratos.Common.Tests
{
	[TestFixture]
	public class StratosUtilTests
	{
		#region Métodos de soporte para los tests
		private void AssertParseDoubleAndParseDecimal(string valorAParsear,
		                                              int decimales,
		                                              double valorEsperado)
		{
			// Testeo la función ParseDouble
			try {
				double resultadoDouble = StratosUtil.ParseDouble(valorAParsear, decimales);
				Assert.AreEqual(valorEsperado, resultadoDouble, "ParseDouble");
				
				// Testeo la función IsDouble
				Assert.IsTrue(StratosUtil.IsDouble(valorAParsear));
			}
			catch (NumeroInvalidoException) {
				// Testeo la función IsDouble si el valor es inválido
				Assert.IsFalse(StratosUtil.IsDouble(valorAParsear));
				throw;
			}
			
			// Testeo la función ParseDecimal
			try {
				decimal resultadoDecimal = StratosUtil.ParseDecimal(valorAParsear, decimales);
				Assert.AreEqual((decimal)valorEsperado, resultadoDecimal, "ParseDecimal");
				
				// Testeo la función IsDecimal
				Assert.IsTrue(StratosUtil.IsDecimal(valorAParsear));
			}
			catch (NumeroInvalidoException) {
				// Testeo la función IsDecimal si el valor es inválido
				Assert.IsFalse(StratosUtil.IsDecimal(valorAParsear));
				throw;
			}
		}
		
		private void AssertRoundDoubleAndRoundDecimal(double valorARedondear,
		                                              int decimales,
		                                              double valorEsperado)
		{
			// Testeo la función RoundDecimal
			decimal resultadoDecimal = StratosUtil.RoundDecimal((decimal)valorARedondear, decimales);
			Assert.AreEqual((decimal)valorEsperado, resultadoDecimal, "RoundDecimal");
			
			// Testeo la función RoundDouble
			double resultadoDouble = StratosUtil.RoundDouble(valorARedondear, decimales);
			Assert.AreEqual(valorEsperado, resultadoDouble, "RoundDouble");
		}
		
		private void AssertDoubleToStringAndDecimalToString(double valorAConvertir,
		                                                    string valorEsperado)
		{
			string resultadoDouble = StratosUtil.DoubleToString(valorAConvertir);
			Assert.AreEqual(valorEsperado, resultadoDouble, "DoubleToString");
			
			string resultadoDecimal = StratosUtil.DecimalToString((decimal)valorAConvertir);
			Assert.AreEqual(valorEsperado, resultadoDecimal, "DecimalToString");
		}
		#endregion
		
		#region ParseDouble y ParseDecimal
		[Test]
		[ExpectedException(typeof(NumeroInvalidoException))]
		public void ValorInvalido()
		{
			AssertParseDoubleAndParseDecimal("noEsUnNumero", 2, 0.0);
		}
		
		[Test]
		public void SinDecimales()
		{
			AssertParseDoubleAndParseDecimal("123", 2, 123.0);
		}
		
		[Test]
		public void ConDecimales_ConComa()
		{
			AssertParseDoubleAndParseDecimal("123,43", 2, 123.43);
		}
		
		[Test]
		public void ConDecimales_ConPunto()
		{
			AssertParseDoubleAndParseDecimal("123.43", 2, 123.43);
		}
		
		[Test]
		public void Negativo_ConDecimales_ConComa()
		{
			AssertParseDoubleAndParseDecimal("-123,43", 2, -123.43);
		}
		
		[Test]
		public void Negativo_ConDecimales_ConPunto()
		{
			AssertParseDoubleAndParseDecimal("-123.43", 2, -123.43);
		}
		
		#region Con dos decimales
		[Test]
		public void ConDecimales_ConComa_MasDeDosDecimales_RedondeoA_Dos_RedondeaParaAbajo()
		{
			AssertParseDoubleAndParseDecimal("123,43321", 2, 123.43);
		}
		
		[Test]
		public void ConDecimales_ConComa_MasDeDosDecimales_RedondeoA_DosDecimales_RedondeaParaArriba_ConCinco()
		{
			AssertParseDoubleAndParseDecimal("123,43521", 2, 123.44);
		}
		
		[Test]
		public void ConDecimales_ConComa_MasDeDosDecimales_RedondeoA_DosDecimales_RedondeaParaArriba_ConMayorACinco()
		{
			AssertParseDoubleAndParseDecimal("123,43721", 2, 123.44);
		}
		#endregion
		
		#region Con tres decimales
		[Test]
		public void ConDecimales_ConComa_MasDeDosDecimales_RedondeoA_TresDecimales_RedondeaParaAbajo()
		{
			AssertParseDoubleAndParseDecimal("123,43321", 3, 123.433);
		}
		
		[Test]
		public void ConDecimales_ConComa_MasDeDosDecimales_RedondeoA_TresDecimales_RedondeaParaArriba_ConCinco()
		{
			AssertParseDoubleAndParseDecimal("123,43351", 3, 123.434);
		}
		
		[Test]
		public void ConDecimales_ConComa_MasDeDosDecimales_RedondeoA_TresDecimales_RedondeaParaArriba_ConMayorACinco()
		{
			AssertParseDoubleAndParseDecimal("123,43371", 3, 123.434);
		}
		#endregion
		
		#region Con separador de miles
		[Test]
		[ExpectedException(typeof(NumeroInvalidoException))]
		public void Positivo_ConSeparador_De_Miles()
		{
			// Por ahora no se soportan los separadores de miles
			AssertParseDoubleAndParseDecimal("1.323,43", 2, 1323.43);
		}
		
		[Test]
		[ExpectedException(typeof(NumeroInvalidoException))]
		public void Negativo_ConSeparador_De_Miles()
		{
			// Por ahora no se soportan los separadores de miles
			AssertParseDoubleAndParseDecimal("-1.323,43", 2, -1323.43);
		}
		#endregion
		#endregion
		
		#region RoundDouble y RoundDecimal
		[Test]
		public void Positivo_RedondearParaAbajo()
		{
			AssertRoundDoubleAndRoundDecimal(123.343, 2, 123.34);
		}
		
		[Test]
		public void Positivo_RedondearParaArriba_EnLaMitad()
		{
			AssertRoundDoubleAndRoundDecimal(123.345, 2, 123.35);
		}
		
		[Test]
		public void Positivo_RedondearParaArriba_EnLaMitad_ValorCalculado()
		{
			AssertRoundDoubleAndRoundDecimal(20.99 * 7.50, 2, 157.43);
		}
		
		[Test]
		public void Positivo_RedondearParaArriba()
		{
			AssertRoundDoubleAndRoundDecimal(123.347, 2, 123.35);
		}
		
		[Test]
		public void Negativo_RedondearParaAbajo()
		{
			AssertRoundDoubleAndRoundDecimal(-123.343, 2, -123.34);
		}
		
		[Test]
		public void Negativo_RedondearParaArriba_EnLaMitad()
		{
			AssertRoundDoubleAndRoundDecimal(-123.345, 2, -123.35);
		}
		
		[Test]
		public void Negativo_RedondearParaArriba()
		{
			AssertRoundDoubleAndRoundDecimal(-123.347, 2, -123.35);
		}
		#endregion
		
		#region DoubleToString y DecimalToString
		[Test]
		public void Positivo()
		{
			AssertDoubleToStringAndDecimalToString(123.23, "123.23");
		}
		
		[Test]
		public void Negativo()
		{
			AssertDoubleToStringAndDecimalToString(-123.23, "-123.23");
		}
		#endregion
	}
}
