// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;

using NUnit.Framework;
using Castle.ActiveRecord;

using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloObjetos.MetaCampos;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture()]
	public class ClienteTests : TestBase
	{
		#region Campos
		private Cliente clienteCompleto;
		private Cliente clienteVacio;
		#endregion

		#region Métodos de inicialización y finalización de la prueba
		[SetUp]
		public override void TestInit ()
		{
			base.TestInit();

			this.clienteCompleto = TestBase.CrearClienteMilton();

			this.clienteVacio = new Cliente();
		}
		#endregion

		#region Pruebas de persistencia de datos
		#region Creación
		[Test()]
		public void Creacion()
		{
			// Persisto el cliente creado en el método Init, ya que ese posee todos
			// los campos completos.
			
			try {
				this.clienteCompleto.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(ex.ValidationErrorMessages[0]);
			}
			
			base.FlushAndRecreateScope();
			
			Cliente[] clientes = Cliente.FindAll();
			
			Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
			
			Cliente cliente = clientes[0];
			Assert.AreEqual("Milton Pividori", cliente.RazonSocial, "Razón social");
			Assert.IsTrue(cliente.Valido, "Valido");
			Assert.AreEqual("Bv. Galvez 956", cliente.Domicilio.Direccion, "Direccion");
			Assert.AreEqual("Santa Fe de la Veracruz", cliente.Domicilio.Localidad, "Localidad");
			Assert.AreEqual("3000", cliente.Domicilio.CodigoPostal, "Codigo Postal");
			Assert.AreEqual(ProvinciaDto.SantaFe, cliente.Domicilio.Provincia, "Provincia");
			Assert.AreEqual("0342 156654321", cliente.Telefono, "Teléfono");
			Assert.AreEqual("miltondp@gmail.com", cliente.CorreoElectronico, "Correo electrónico");
			Assert.AreEqual(TipoResponsable.Monotributo, cliente.TipoResponsable, "TipoResponsable");
			Assert.AreEqual(30674393683, cliente.Cuit.ValorNumerico, "CUIT");
			Assert.AreEqual(15m, cliente.PorcentajeBonificacion, "Bonificacion");
			Assert.AreEqual("Estudiante de ISI", cliente.Notas, "Notas");
			Assert.IsFalse(cliente.Eliminado, "Eliminado");
			Assert.AreEqual(cliente.Version, 2, "Versión");
		}
		
		[Test]
		public void ActualizarClienteExistente()
		{
			this.clienteCompleto.Create();
			this.FlushAndRecreateScope();
			
			// Actualizo alguno de sus campos
			this.clienteCompleto = Cliente.Find(this.clienteCompleto.Codigo);
			this.clienteCompleto.Notas += " agregado";
			
			try {
				this.clienteCompleto.UpdateAndFlush();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinDatos()
		{
			this.clienteVacio.Create();
		}
		
		[Test()]
		public void SinCodigo()
		{
			this.clienteCompleto.Codigo = this.clienteVacio.Codigo;
			
			this.clienteCompleto.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinRazonSocial()
		{
			this.clienteCompleto.RazonSocial = this.clienteVacio.RazonSocial;
			
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinAltaAutomatica()
		{
			this.clienteCompleto.AltaAutomatica = this.clienteVacio.AltaAutomatica;
			
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinDomicilio()
		{
			this.clienteCompleto.Domicilio = this.clienteVacio.Domicilio;
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinDireccion()
		{
			this.clienteCompleto.Domicilio.Direccion =
				this.clienteVacio.Domicilio.Direccion;
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinLocalidad()
		{
			this.clienteCompleto.Domicilio.Localidad =
				this.clienteVacio.Domicilio.Localidad;
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinCodigoPostal()
		{
			this.clienteCompleto.Domicilio.CodigoPostal =
				this.clienteVacio.Domicilio.CodigoPostal;
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinProvincia()
		{
			this.clienteCompleto.Domicilio.Provincia =
				this.clienteVacio.Domicilio.Provincia;
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinTelefono()
		{
			this.clienteCompleto.Telefono = this.clienteVacio.Telefono;
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinCorreoElectronico()
		{
			this.clienteCompleto.CorreoElectronico = this.clienteVacio.CorreoElectronico;
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinTipoResponsable()
		{
			this.clienteCompleto.TipoResponsable = this.clienteVacio.TipoResponsable;
			
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinCuit_ClienteResponsableInscripto()
		{
			this.clienteCompleto.TipoResponsable = TipoResponsable.ResponsableInscripto;
			this.clienteCompleto.Cuit = this.clienteVacio.Cuit;
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		
		[Test()]
		public void SinCuit_CuitNulo_ClienteResponsableInscripto()
		{
			this.clienteCompleto.TipoResponsable = TipoResponsable.ResponsableInscripto;
			this.clienteCompleto.Cuit = null;
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		
		[Test()]
		public void SinCuit_ClienteResponsableNoInscripto()
		{
			this.clienteCompleto.TipoResponsable = TipoResponsable.ResponsableNoInscripto;
			this.clienteCompleto.Cuit = this.clienteVacio.Cuit;
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		
		[Test()]
		public void SinCuit_ClienteMonotributo()
		{
			this.clienteCompleto.TipoResponsable = TipoResponsable.Monotributo;
			this.clienteCompleto.Cuit = this.clienteVacio.Cuit;
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		
		[Test()]
		public void SinCuit_ClienteExento()
		{
			this.clienteCompleto.TipoResponsable = TipoResponsable.Exento;
			this.clienteCompleto.Cuit = this.clienteVacio.Cuit;
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		
		[Test()]
		public void SinCuit_ClienteConsumidorFinal()
		{
			this.clienteCompleto.TipoResponsable = TipoResponsable.ConsumidorFinal;
			this.clienteCompleto.Cuit = this.clienteVacio.Cuit;
			
			try {
				this.clienteCompleto.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("No pudo crearse: " + this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void SinCuit_CuitNulo_ClienteConsumidorFinal()
		{
			this.clienteCompleto.TipoResponsable = TipoResponsable.ConsumidorFinal;
			this.clienteCompleto.Cuit = null;
			
			try {
				this.clienteCompleto.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("No pudo crearse: " + this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void SinBonificacion()
		{
			this.clienteCompleto.PorcentajeBonificacion = this.clienteVacio.PorcentajeBonificacion;
			this.clienteCompleto.Create();
		}
		
		[Test()]
		public void SinNotas()
		{
			this.clienteCompleto.Notas = this.clienteVacio.Notas;
			this.clienteCompleto.Create();
		}
		#endregion
		#endregion
		
		#region Pruebas de alto nivel (código repetido, bonificación excesiva, etc)
		[Test]
		public void EmailInvalido1()
		{
			this.clienteCompleto.CorreoElectronico = "miltondp";
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepción");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("CorreoElectronico", ex),
				              "Propiedad CorreoElectronico errónea");
			}
		}
		
		[Test]
		public void EmailInvalido2()
		{
			this.clienteCompleto.CorreoElectronico = "miltondp@";
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepción");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("CorreoElectronico", ex),
				              "Propiedad CorreoElectronico errónea");
			}
		}
		
		[Test]
		public void EmailInvalido3()
		{
			this.clienteCompleto.CorreoElectronico = "@";
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepción");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("CorreoElectronico", ex),
				              "Propiedad CorreoElectronico errónea");
			}
		}
		
		[Test]
		public void EmailInvalido4()
		{
			this.clienteCompleto.CorreoElectronico = "@gmail";
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepción");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("CorreoElectronico", ex),
				              "Propiedad CorreoElectronico errónea");
			}
		}
		
		[Test]
		public void CuitRepetido()
		{
			Cliente c1 = new Cliente {
				Codigo = 1,
				RazonSocial = "Milton Pividori",
				Domicilio = new Domicilio {
					Direccion = "Bv. Galvez 956",
					Localidad = "Santa Fe de la Veracruz",
					Provincia = ProvinciaDto.SantaFe
				},
				TipoResponsable = TipoResponsable.ResponsableInscripto,
				Cuit = new CUIT(30674393683),
				PorcentajeBonificacion = 15m,
				Notas = "Estudiante de ISI"
			};

			// Este cliente debe estar persistido antes de intentar crear el otro
			c1.Create();
			
			this.FlushAndRecreateScope();
			
			Cliente c2 = new Cliente {
				Codigo = 2,
				RazonSocial = "César Sandrigo",
				Domicilio = new Domicilio {
					Direccion = "Laprida 5054",
					Localidad = "Santa Fe de la Veracruz",
					Provincia = ProvinciaDto.SantaFe
				},
				TipoResponsable = TipoResponsable.ResponsableInscripto,
				Cuit = new CUIT(30674393683),
				PorcentajeBonificacion = 15m,
				Notas = "Estudiante de ISI"
			};
			
			try {
				c2.Create();
				Assert.Fail("Se pudo crear el cliente");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex), "Propiedad fallida");
			}
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void BonificacionExcesiva()
		{
			this.clienteCompleto.PorcentajeBonificacion = 75m;
			
			this.clienteCompleto.Create();	
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void BonificacionMenorACero()
		{
			this.clienteCompleto.PorcentajeBonificacion = -1.0m;
			
			this.clienteCompleto.Create();	
		}
		
		[Test()]
		public void ConsumidorFinalConCuit()
		{
			this.clienteCompleto.Cuit = new CUIT(30674393683);
			this.clienteCompleto.TipoResponsable = TipoResponsable.ConsumidorFinal;
			
			try {
				this.clienteCompleto.Create();
				Assert.Fail("No hubo excepción");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		#endregion
		
		#region Metacampos
		
		[Test()]
		public void CreacionClienteConMetaCampo()
		{
			try {
				this.clienteCompleto.SetearMetaCampo("SobreNombre", "Pepe");
				this.clienteCompleto.Create();
				
				base.FlushAndRecreateScope();
			
				Cliente[] clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}
		
		[Test()]
		public void RecuperacionClienteConMetaCampo()
		{
			try {
				this.clienteCompleto.SetearMetaCampo("SobreNombre", "Pepe");
				this.clienteCompleto.Create();
				
				base.FlushAndRecreateScope();
			
				Cliente[] clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Assert.AreEqual("Pepe", clientes[0].ObtenerMetaCampo("SobreNombre"), "Obtención de metacampo");
				
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}
		
		[Test()]
		public void ActualizacionMetaCampoSinObtenerCliente()
		{
			try {
				this.clienteCompleto.SetearMetaCampo("SobreNombre", "Pepe");
				this.clienteCompleto.Create();
				
				this.clienteCompleto.SetearMetaCampo("SobreNombre", "Juancho");
				this.clienteCompleto.Save();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}
		
		[Test()]
		public void ActualizacionMetaCampoObteniendoElCliente()
		{
			try {
				this.clienteCompleto.SetearMetaCampo("SobreNombre", "Pepe");
				this.clienteCompleto.Create();
				
				base.FlushAndRecreateScope();
			
				Cliente[] clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Cliente clienteRecuperado = clientes[0];
				
				Assert.AreEqual(2, clienteRecuperado.MetaCampos.Count, "Cantidad de metacampos");
				Assert.AreEqual("Pepe", clienteRecuperado.ObtenerMetaCampo("SobreNombre"), "Recuperación metacampo");
				
				clienteRecuperado.SetearMetaCampo("SobreNombre", "Juancho");
				clienteRecuperado.Save();
				
				base.FlushAndRecreateScope();
			
				clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}
		
		[Test()]
		public void RecuperacionMetaCampo()
		{
			try {
				this.clienteCompleto.SetearMetaCampo("SobreNombre", "Pepe");
				this.clienteCompleto.Create();
				
				base.FlushAndRecreateScope();
			
				Cliente[] clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Cliente clienteRecuperado = clientes[0];
				
				Assert.AreEqual("Pepe", clienteRecuperado.ObtenerMetaCampo("SobreNombre"));
				
				clienteRecuperado.SetearMetaCampo("SobreNombre", "Juancho");
				clienteRecuperado.Save();
				
				base.FlushAndRecreateScope();
				
				clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Cliente otroCliente = clientes[0];
				
				Assert.AreEqual("Juancho", otroCliente.ObtenerMetaCampo("SobreNombre"));
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}
		
		[Test()]
		public void RecuperacionMetaCampoNoExistente()
		{
			try {
				Assert.IsNull(this.clienteCompleto.ObtenerMetaCampo("SobreNombre"));
				
				this.clienteCompleto.Create();
				base.FlushAndRecreateScope();
				
				Cliente[] clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Cliente clienteRecuperado = clientes[0];
				
				Assert.AreEqual(1, clienteRecuperado.MetaCampos.Count, "Cantidad de metacampos");
				Assert.IsNull(clienteRecuperado.ObtenerMetaCampo("SobreNombre"), "Valor de metacampo inexistente");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}
		
		[Test()]
		public void MetaCampoSobreescritoAntesDeGuardar()
		{
			try {
				this.clienteCompleto.Create();
				
				this.clienteCompleto.SetearMetaCampo("SobreNombre", "Pepe");
				this.clienteCompleto.SetearMetaCampo("SobreNombre", "Juancho");
				
				this.clienteCompleto.Save();
				
				base.FlushAndRecreateScope();
				
				Cliente[] clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Cliente clienteRecuperado = clientes[0];
				
				Assert.AreEqual(2, clienteRecuperado.MetaCampos.Count, "Cantidad de metacampos");
				Assert.AreEqual("Juancho", clienteRecuperado.ObtenerMetaCampo("SobreNombre"), "Valor del metacampo");
				
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}
		
		[Test()]
		public void ClienteNuevoConMuchosMetacampos()
		{
			try {
				this.clienteCompleto.SetearMetaCampo("SobreNombre", "Pepe");
				this.clienteCompleto.SetearMetaCampo("ProductoPreferido", "Mesada");
				this.clienteCompleto.SetearMetaCampo("FechaDeNacimiento", "01/01/1980");
				this.clienteCompleto.SetearMetaCampo("EstadoCivil", "Soltero");
				
				this.clienteCompleto.Create();
				
				base.FlushAndRecreateScope();
				
				Cliente[] clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Cliente clienteRecuperado = clientes[0];
				
				Assert.AreEqual(5, clienteRecuperado.MetaCampos.Count, "Cantidad de metacampos");
				Assert.AreEqual("Pepe", clienteRecuperado.ObtenerMetaCampo("SobreNombre"), "Valor para SobreNombre");
				Assert.AreEqual("Mesada", clienteRecuperado.ObtenerMetaCampo("ProductoPreferido"), "Valor para ProductoPreferido");
				Assert.AreEqual("01/01/1980", clienteRecuperado.ObtenerMetaCampo("FechaDeNacimiento"), "Valor para FechaDeNacimiento");
				Assert.AreEqual("Soltero", clienteRecuperado.ObtenerMetaCampo("EstadoCivil"), "Valor para EstadoCivil");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}
		
		[Test()]
		public void ClienteExistenteConMuchosMetacampos()
		{
			try {
				this.clienteCompleto.Create();
				base.FlushAndRecreateScope();
				
				Cliente[] clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Cliente clienteExistente = clientes[0];
				
				clienteExistente.SetearMetaCampo("SobreNombre", "Pepe");
				clienteExistente.SetearMetaCampo("ProductoPreferido", "Mesada");
				clienteExistente.SetearMetaCampo("FechaDeNacimiento", "01/01/1980");
				clienteExistente.SetearMetaCampo("EstadoCivil", "Soltero");
				
				clienteExistente.Save();
				
				base.FlushAndRecreateScope();
				
				clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Cliente clienteRecuperado = clientes[0];
				
				Assert.AreEqual(5, clienteRecuperado.MetaCampos.Count, "Cantidad de metacampos");
				Assert.AreEqual("Pepe", clienteRecuperado.ObtenerMetaCampo("SobreNombre"), "Valor para SobreNombre");
				Assert.AreEqual("Mesada", clienteRecuperado.ObtenerMetaCampo("ProductoPreferido"), "Valor para ProductoPreferido");
				Assert.AreEqual("01/01/1980", clienteRecuperado.ObtenerMetaCampo("FechaDeNacimiento"), "Valor para FechaDeNacimiento");
				Assert.AreEqual("Soltero", clienteRecuperado.ObtenerMetaCampo("EstadoCivil"), "Valor para EstadoCivil");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}
		
		[Test()]
		public void EliminacionMetacampos()
		{
			try {
				this.clienteCompleto.Create();
				base.FlushAndRecreateScope();
				
				Cliente[] clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				Cliente clienteExistente = clientes[0];
				
				// Creo un metacampo para el cliente
				clienteExistente.SetearMetaCampo("SobreNombre", "Pepe");
				clienteExistente.Save();
				
				base.FlushAndRecreateScope();
				
				clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				// Me aseguro de que el metacampo esté persistido
				Cliente clienteRecuperado = clientes[0];
				Assert.AreEqual("Pepe", clienteRecuperado.ObtenerMetaCampo("SobreNombre"), "Valor para SobreNombre");
				
				// Elimino el metacampo y guardo el cliente
				clienteRecuperado.EliminarMetaCampo("SobreNombre");
				clienteRecuperado.Save();
				
				base.FlushAndRecreateScope();
				
				clientes = Cliente.FindAll();
				Assert.AreEqual(1, clientes.Length, "Cantidad de clientes insertados");
				
				// Intento obtener el metacampo eliminado
				clienteRecuperado = clientes[0];
				Assert.IsNull(clienteRecuperado.ObtenerMetaCampo("SobreNombre"), "Obtención de metacampo eliminado");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Ocurrió una excepción: " + ex.Message);
			}
		}

		#endregion
	}
}
