// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

using NUnit.Framework;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using NHibernate.Criterion;
using NHibernate.Exceptions;

using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture()]
	public class FacturaCompraTests : DocumentoGenericoTestBase
	{
		#region Campos
		private FacturaCompra factura1;
		private FacturaCompra factura2;
		private FacturaCompra facturaVacia;
		#endregion
		
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			this.factura1 = CrearFacturaCompra();
			this.factura1.Proveedor = this.ProveedorFerrum;
			this.factura1.Items.Clear();
			this.factura1.Items.Add(CrearItemFacturaCompra(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFacturaCompra(this.factura1, true);
			
			this.factura2 = CrearFacturaCompra();
			this.factura2.Proveedor = this.ProveedorFV;
			this.factura2.Items.Clear();
			this.factura2.Items.Add(CrearItemFacturaCompra(this.ProductoGasto, 2));
			TestBase.ProcesarItemsYCalcularImportesFacturaCompra(this.factura2, true);
			
			this.facturaVacia = new FacturaCompra();
		}
		
		#region Pruebas de persistencia de datos
		[Test()]
		public void Creacion()
		{
			try {
				this.factura1.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			base.FlushAndRecreateScope();
			
			FacturaCompra[] facturas = FacturaCompra.FindAll();
			
			Assert.AreEqual(1, facturas.Length, "Cantidad de facturas insertadas");
			
			FacturaCompra fc = facturas[0];
			
			// Cabecera factura
			Assert.AreEqual(FacturaTipoCompra.FacturaA, fc.Tipo, "Tipo");
			Assert.AreEqual(1, fc.Codigo, "Código");
			Assert.AreEqual(1234, fc.Numero, "Número");
			
			// Fechas
			Assert.AreEqual(DateTime.Today.Date - TimeSpan.FromDays(10), fc.Fecha.Date, "Fecha");
			Assert.AreEqual(DateTime.Today.Date, fc.FechaImputacion.Date, "Fecha de imputación");
			
			// Proveedor
			Assert.AreEqual(this.ProveedorFerrum.Codigo, fc.Proveedor.Codigo, "Codigo proveedor");
			Assert.AreEqual(this.ProveedorFerrum.RazonSocial, fc.ProveedorRazonSocial, "Proveedor razón social");
			Assert.AreEqual(this.ProveedorFerrum.TipoResponsable, fc.ProveedorTipoResponsable, "Proveedor tipo responsable");
			Assert.AreEqual(this.ProveedorFerrum.Cuit.ValorNumerico.Value, fc.ProveedorCuit, "Proveedor cuit");
			
			// Importes
			Assert.AreEqual(54.22m, fc.ImporteGravado, "Gravado");
			Assert.AreEqual(2.77m, fc.ImporteExento, "Exento");
			Assert.AreEqual(11.97m, fc.ImporteIva, "Importe iva");
			Assert.AreEqual(1.45m, fc.PercepcionIB, "Percepción IB");
			Assert.AreEqual(2.33m, fc.PercepcionIva, "Percepción IVA");
			Assert.AreEqual(19.22m, fc.ImpuestoCombustible, "Impuesto combustible");
			Assert.AreEqual(91.96m, fc.ImporteTotal, "Importe total");
			Assert.AreEqual(91.96m, fc.ImporteFinal, "Importe final");
			
			// Otros
			Assert.AreEqual("Estas son las observaciones", fc.Observaciones, "Observaciones");
			Assert.IsFalse(fc.Historica, "Historica");
		}

		[Test()]
		public void SinDatos()
		{
			try {
				this.facturaVacia.Create();
				Assert.Fail("La factura se creo correctamente");
			}
			catch (ActiveRecordValidationException ex) {
			}
		}
		
		[Test()]
		public void SinDatosMinimosProveedor()
		{
			// La factura no tiene asignado un proveedor ni los datos mínimos del mismo
			this.factura1.Proveedor = this.facturaVacia.Proveedor;
			this.factura1.ProveedorRazonSocial = this.facturaVacia.ProveedorRazonSocial;
			this.factura1.ProveedorTipoResponsable = TipoResponsable.ConsumidorFinal; // Coloco un tipo responsable inválido
			this.factura1.ProveedorCuit = this.facturaVacia.ProveedorCuit;
			
			try {
				this.factura1.Create();
				Assert.Fail("La factura se creo correctamente");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("ProveedorRazonSocial", ex), "ProveedorRazonSocial");
				Assert.IsTrue(this.PropiedadFallida("ProveedorTipoResponsable", ex), "ProveedorTipoResponsable");
				Assert.IsTrue(this.PropiedadFallida("ProveedorCuit", ex), "ProveedorCuit");
			}
		}
		
		[Test()]
		public void ConDatosMinimosProveedor_PeroSinProveedorExistenteAsociado()
		{
			// La factura no tiene asginado un proveedor, pero sí los datos mínimos
			// del mismo.
			this.factura1.Proveedor = this.facturaVacia.Proveedor;
			
			this.factura1.ProveedorRazonSocial = this.ProveedorFerrum.RazonSocial;
			this.factura1.ProveedorTipoResponsable = this.ProveedorFerrum.TipoResponsable;
			this.factura1.ProveedorCuit = this.ProveedorFerrum.Cuit.ValorNumerico.Value;
			
			try {
				this.factura1.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			
			Assert.IsNull(this.factura1.Proveedor, "Proveedor");
		}

		[Test()]
		public void SinNumero()
		{
			this.factura1.Numero = this.facturaVacia.Numero;
			
			try {
				this.factura1.Create();
				Assert.Fail("La factura se creo correctamente");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Numero", ex), "Propiedad fallida");
			}
		}
		
		[Test()]
		public void SinItems()
		{
			this.factura1.Items = this.facturaVacia.Items;
			
			try {
				this.factura1.Create();
				Assert.Fail("La factura se creo correctamente");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida");
			}
		}
		
		[Test()]
		public void SinImporteGravado()
		{
			this.factura1.ImporteGravado = this.facturaVacia.ImporteGravado;
			
			try {
				this.factura1.Create();
				Assert.Fail("La factura se creo correctamente");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("ImporteGravado", ex), "Propiedad fallida");
			}
		}
		
		[Test()]
		public void SinImporteExento()
		{
			this.factura1.ImporteTotal -= this.factura1.ImporteExento;
			this.factura1.ImporteExento = this.facturaVacia.ImporteExento;
			
			try {
				this.factura1.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void SinPercepcionIva()
		{
			this.factura1.ImporteTotal -= this.factura1.PercepcionIva;
			this.factura1.PercepcionIva = this.facturaVacia.PercepcionIva;
			
			try {
				this.factura1.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void SinPercepcionIB()
		{
			this.factura1.ImporteTotal -= this.factura1.PercepcionIB;
			this.factura1.PercepcionIB = this.facturaVacia.PercepcionIva;
			
			try {
				this.factura1.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void SinImpuestoCombustible()
		{
			this.factura1.ImporteTotal -= this.factura1.ImpuestoCombustible;
			this.factura1.ImpuestoCombustible = this.facturaVacia.ImpuestoCombustible;
			
			try {
				this.factura1.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void SinImporteIva()
		{
			this.factura1.ImporteIva = this.facturaVacia.ImporteIva;
			
			try {
				this.factura1.Create();
				Assert.Fail("La factura se creo correctamente");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("ImporteIva", ex), "Propiedad fallida");
			}
		}
		
		[Test()]
		public void SinImporteTotal()
		{
			this.factura1.ImporteTotal = this.facturaVacia.ImporteTotal;
			
			try {
				this.factura1.Create();
				Assert.Fail("La factura se creo correctamente");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("ImporteTotal", ex), "Propiedad fallida");
			}
		}
		#endregion
		
		#region Pruebas de alto nivel (código repetido, bonificación excesiva, etc)
		[Test()]
		public void NumeroRepetidoDistintoProveedor()
		{
			this.factura1.Numero = 1111;
			this.factura1.Proveedor = this.ProveedorFerrum;
			this.factura1.Create();
			
			this.FlushAndRecreateScope();
			
			this.factura2.Numero = 1111;
			this.factura2.Proveedor = this.ProveedorFV;
			
			try {
				this.factura2.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void ProveedorRepetidoDistintoNumero()
		{
			this.factura1.Numero = 1111;
			this.factura1.Proveedor = this.ProveedorFerrum;
			this.factura1.Create();
			
			this.FlushAndRecreateScope();
			
			this.factura2.Numero = 1112;
			this.factura2.Proveedor = this.ProveedorFerrum;
			
			try {
				this.factura2.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void ClaveUnicaRepetida_ProveedorAsignado()
		{
			this.factura1.Tipo = FacturaTipoCompra.FacturaA;
			this.factura1.Codigo = 1;
			this.factura1.Numero = 1111;
			this.factura1.Proveedor = this.ProveedorFerrum;
			this.factura1.Create();
			
			this.FlushAndRecreateScope();
			
			this.factura2.Tipo = FacturaTipoCompra.FacturaA;
			this.factura2.Codigo = 1;
			this.factura2.Numero = 1111;
			this.factura2.Proveedor = this.ProveedorFerrum;
			
			try {
				this.factura2.Create();
				Assert.Fail("Se pudo crear la factura con clave repetida");
			}
			catch(ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("Numero", e), "Propiedad fallida");
				return;
			}
		}
		
		#region Fecha de comprobante/factura
		[Test()]
		public void FechaFacturaMayorAHoy()
		{
			this.factura1.Fecha = DateTime.Today + TimeSpan.FromDays(5);
			
			try {
				this.factura1.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Fecha", ex), "Propiedad Fecha errónea");
			}
		}
		
		[Test()]
		[Ignore("Ver cuál es la fecha mínima posible para una fatura de compra")]
		public void FechaFacturaMenorA()
		{
//			this.factura1.Fecha = DateTime.Today + TimeSpan.FromDays(5);
//			
//			try {
//				this.factura1.Create();
//				Assert.Fail("No hubo excepcion");
//			}
//			catch (ActiveRecordValidationException ex) {
//				// Me aseguro que haya un error en la propiedad correspondiente
//				Assert.IsTrue(this.PropiedadFallida("Fecha", ex), "Propiedad Fecha errónea");
//			}
		}
		#endregion
		
		#region Fecha de imputación
		[Test()]
		public void FechaImputacionMayorAHoy()
		{
			this.factura1.FechaImputacion = DateTime.Today + TimeSpan.FromDays(4);

			try {
				this.factura1.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("FechaImputacion", ex),
				              "Propiedad Fecha errónea");
			}
		}
		
//		[Test()]
//		public void FechaImputacionMenorA60Dias()
//		{
//			this.factura1.FechaImputacion = DateTime.Today.AddDays(-65);
//
//			try {
//				this.factura1.Create();
//				Assert.Fail("No hubo excepcion");
//			}
//			catch (ActiveRecordValidationException ex) {
//				// Me aseguro que haya un error en la propiedad correspondiente
//				Assert.IsTrue(this.PropiedadFallida("FechaImputacion", ex),
//				              "Propiedad Fecha errónea");
//			}
//		}
		
		[Test()]
		public void FechaImputacionMenorAFechaFactura()
		{
			this.factura1.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			this.factura1.FechaImputacion = DateTime.Today - TimeSpan.FromDays(5);
			
			try {
				this.factura1.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("FechaImputacion", ex),
				              "Propiedad FechaImputacion errónea");
			}
		}
		
		[Test()]
		public void FechaImputacionMayorAFechaFactura()
		{
			this.factura1.Fecha = DateTime.Today - TimeSpan.FromDays(5);
			this.factura1.FechaImputacion = DateTime.Today;
			
			try {
				this.factura1.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		#endregion
		
		#region Factura histórica
		[Test]
		public void FacturaHistorica_NoPuedeEliminarse()
		{
			this.factura1.Historica = true;
			this.factura1.Create();
			
			this.FlushAndRecreateScope();
			
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			
			try {
				this.factura1.DeleteAndFlush();
				Assert.Fail("Se pudo eliminar");
			}
			catch (ActiveRecordException) {
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			Assert.AreEqual(1,
			                FacturaCompra.FindAll(Restrictions.Eq("Id", this.factura1.Id)).Length,
			                "Factura no eliminada");
		}
		
		[Test]
		public void FacturaHistorica_NoPuedeModificarse()
		{
			this.factura1.Historica = true;
			this.factura1.Observaciones = null;
			this.factura1.Create();
			
			this.FlushAndRecreateScope();
			
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			this.factura1.Observaciones += "Obs";
			
			try {
				this.factura1.Update();
				this.FlushAndRecreateScope();
				Assert.Fail("Se pudo modificar");
			}
			catch (ActiveRecordValidationException) {
			}
			
			try {
				this.FlushAndRecreateReadOnlyScope();
			} catch (ActiveRecordValidationException ex) {
				Assert.Fail("2 " + this.ObtenerMensaje(ex));
			}
			
			Assert.AreEqual(1, FacturaCompra.FindAll().Length, "Factura no eliminada");
			Assert.IsNull(FacturaCompra.FindAll()[0].Observaciones, "Factura observaciones nulas");
		}
		#endregion
		
//		[Test]
//		public void ItemsGastos_NoSonDeRubroGastos()
//		{
//			this.factura1.Items.Clear();
//			this.factura1.Items.Add(new ItemFacturaGasto(this.ProductoGriferia, 5));
//			TestBase.ProcesarItemsYCalcularImportesFacturaCompra(this.factura1);
//			
//			try {
//				this.factura1.Create();
//				Assert.Fail("Se pudo crear la factura");
//			}
//			catch (ActiveRecordValidationException ex) {
//				Assert.IsTrue(this.PropiedadFallida("ProductoRubroId", ex), "Propiedad fallida");
//			}
//		}
		
		[Test()]
		public void ImporteFinalCorrecto()
		{
			ItemFacturaCompra item1 = TestBase.CrearItemFacturaCompra(this.ProductoGriferia, 2);
			ItemFacturaCompra item2 = TestBase.CrearItemFacturaCompra(this.ProductoMosaico, 12.34m);
			
			this.factura1.Items.Clear();
			this.factura1.Items.Add(item1);
			this.factura1.Items.Add(item2);
			
			TestBase.ProcesarItemsYCalcularImportesFacturaCompra(this.factura1, true);
			
			this.factura1.ImporteGravado = 312.22m;
			this.factura1.ImporteExento = 10.00m;
			this.factura1.ImporteIva = 78.18m;
			this.factura1.PercepcionIB = 33.11m;
			this.factura1.PercepcionIva = 0.66m;
			this.factura1.ImpuestoCombustible = 1.22m;
			
			this.factura1.ImporteTotal = 435.39m;
			
			try {
				this.factura1.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void ImporteTotalInvalido()
		{
			ItemFacturaCompra item1 = TestBase.CrearItemFacturaCompra(this.ProductoGriferia, 2);
			ItemFacturaCompra item2 = TestBase.CrearItemFacturaCompra(this.ProductoMosaico, 12.34m);
			
			this.factura1.Items.Clear();
			this.factura1.Items.Add(item1);
			this.factura1.Items.Add(item2);
			
			TestBase.ProcesarItemsYCalcularImportesFacturaCompra(this.factura1, true);
			
			this.factura1.ImporteGravado = 312.22m;
			this.factura1.ImporteExento = 10.00m;
			this.factura1.ImporteIva = 78.18m;
			this.factura1.PercepcionIB = 33.11m;
			this.factura1.PercepcionIva = 0.66m;
			this.factura1.ImpuestoCombustible = 1.22m;
			
			this.factura1.ImporteTotal = 500.40m; // El importe total es incorrecto
			
			try {
				this.factura1.Create();
				Assert.Fail("Se pudo crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("ImporteTotal", ex), "Propiedad fallida");
			}
		}
		
		#region Facturas con tipos de proveedores
		[Test()]
		public void FacturaA_ProveedorMonotributo() {
			// Ahora el proveedor es monotributista
			this.ProveedorFerrum.TipoResponsable = TipoResponsable.Monotributo;
			this.ProveedorFerrum.Update();
			
			this.FlushAndRecreateScope();
			
			// La factura queda como factura A pero con proveedor monotributista
			this.factura1.Tipo = FacturaTipoCompra.FacturaA;
			this.factura1.Proveedor = this.ProveedorFerrum;
			
			try {
				this.factura1.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ProveedorTipoResponsable", ex),
				              "Propiedad ProveedorTipoResponsable errónea");
			}
		}
		
		[Test()]
		public void FacturaA_ProveedorResponsableInscripto() {
			// Ahora el proveedor es monotributista
			this.ProveedorFerrum.TipoResponsable = TipoResponsable.ResponsableInscripto;
			this.ProveedorFerrum.Update();
			
			this.FlushAndRecreateScope();
			
			// La factura queda como factura A pero con proveedor monotributista
			this.factura1.Tipo = FacturaTipoCompra.FacturaA;
			this.factura1.Proveedor = this.ProveedorFerrum;
			
			try {
				this.factura1.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void FacturaC_ProveedorResponsableInscripto()
		{
			this.ProveedorFerrum.TipoResponsable = TipoResponsable.ResponsableInscripto;
					
			this.factura1.Tipo = FacturaTipoCompra.FacturaC;
			this.factura1.Proveedor = this.ProveedorFerrum;
			
			try {
				this.factura1.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ProveedorTipoResponsable", ex),
				              "Propiedad ProveedorTipoResponsable errónea");
			}
		}
		
		[Test()]
		public void FacturaC_ProveedorMonotributo() {
			// Ahora el proveedor es monotributista
			this.ProveedorFerrum.TipoResponsable = TipoResponsable.Monotributo;
			this.ProveedorFerrum.Update();
			
			this.FlushAndRecreateScope();
			
			// La factura queda como factura A pero con proveedor monotributista
			this.factura1.Tipo = FacturaTipoCompra.FacturaC;
			this.factura1.Proveedor = this.ProveedorFerrum;
			
			try {
				this.factura1.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		#endregion
		#endregion
	}
}
