// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

using NUnit.Framework;
using NUnit.Framework.Constraints;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using NHibernate.Criterion;

using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture()]
	public class FacturaTests : DocumentoGenericoTestBase
	{
		#region Campos
		private Factura facturaCompleta;
		private Factura facturaCompleta2;
		private Factura facturaVacia;
		#endregion

		#region Métodos de inicialización y finalización de la prueba
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();

			// Creo la factura con los datos completos para realizar
			// las pruebas de persistencia
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			this.facturaCompleta2 = TestBase.CrearFactura();
			this.facturaCompleta2.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta2.Items.Clear();
			this.facturaCompleta2.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta2);

			this.facturaVacia = new Factura();
		}
		#endregion

		#region Pruebas de persistencia de datos
		[Test()]
		public void Creacion()
		{
			try {
				this.facturaCompleta.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}

			base.FlushAndRecreateScope();
			
			Factura[] facturas = Factura.FindAll();
			
			Assert.AreEqual(1, facturas.Length, "Cantidad de facturas insertadas");
			
			Factura fr = facturas[0];
			
			Assert.AreEqual(this.ClienteSinBonificacion.Codigo, fr.Cliente.Codigo, "Id cliente");
			Assert.AreEqual(this.ClienteSinBonificacion.RazonSocial, fr.ClienteRazonSocial, "ClienteRazonSocial");
			Assert.AreEqual(this.ClienteSinBonificacion.TipoResponsable, fr.ClienteTipoResponsable, "ClienteTipoResponsable");
			
			Assert.IsTrue(this.ClienteSinBonificacion.Cuit.ValorNumerico.HasValue, "Tiene cuit el cliente");
			Assert.AreEqual(this.ClienteSinBonificacion.Cuit.ValorNumerico.Value, fr.ClienteCuit, "ClienteCuit");
			
			Assert.AreEqual(4, fr.Codigo, "Código");
			Assert.AreEqual(60.00m, fr.ImporteBruto, "Importe bruto");
			Assert.AreEqual(0.00m, fr.PorcentajeBonificacion, "Porcentaje bonificación");
			Assert.AreEqual(0.00m, fr.ImporteBonificacion, "Importe bonificación");
			Assert.AreEqual(60.00m, fr.ImporteNeto, "Importe Neto");
			Assert.AreEqual(12.60m, fr.ImporteIva, "Importe iva");
			Assert.AreEqual(72.60m, fr.ImporteTotal, "Importe total");
			Assert.AreEqual(72.60m, fr.ImporteFinal, "Importe final");
			Assert.AreEqual(23423, fr.Numero, "Numero");
			Assert.IsFalse(fr.Anulada, "Anulada");
			Assert.IsFalse(fr.FacturaManual, "Factura manual");
			Assert.IsTrue(fr.Valida, "Válida");
			Assert.IsFalse(fr.Historica, "Historica");
			Assert.AreEqual("Estas son las observaciones", fr.Observaciones, "Observaciones");
			Assert.AreEqual(FacturaTipoVenta.FacturaA, fr.Tipo, "Tipo factura");
			Assert.AreEqual(DateTime.Today, fr.Fecha.Date, "Fecha");
			Assert.AreEqual(CondicionPago.Contado, fr.CondicionVenta.Condicion,
			                "Condicion de venta");
			Assert.AreEqual("FA", fr.TipoAbreviado, "Tipo abreviado");
			
			// Items
			Assert.AreEqual(1, fr.Items.Count, "Cantidad de items");
			ItemFactura item = fr.Items[0];
			
			Assert.AreEqual(this.ProductoGriferia.Descripcion, item.Descripcion, "Item descripción");
			Assert.AreEqual(1, item.Cantidad, "Item cantidad");
			Assert.AreEqual(0.00m, item.PorcentajeBonificacion, "Item porcentaje bonificación");
			Assert.AreEqual(this.ProductoGriferia.ImporteSinIva, item.PrecioUnitario, "Item precio unitario");
			
			Assert.AreEqual(this.ProductoGriferia.Id, item.ProductoId, "Item producto id");
			Assert.AreEqual(this.ProductoGriferia.ProductoAbstracto, item.ProductoAbstracto, "Item producto abstracto");
			Assert.AreEqual(this.ProductoGriferia.Codigo, item.ProductoCodigo, "Item producto codigo");
			Assert.IsNull(item.ProductoDescripcion, "Item producto descripcion");
			Assert.AreEqual(this.ProductoGriferia.ImporteSinIva, item.ProductoImporteSinIva, "Item producto importe sin iva");
			Assert.AreEqual(this.ProductoGriferia.ImporteVenta, item.ProductoImporteVenta, "Item producto importe venta");
			Assert.AreEqual(this.ProductoGriferia.PrecioCosto, item.ProductoPrecioCosto, "Item producto precio costo");
			Assert.AreEqual(this.ProductoGriferia.Unidad, item.ProductoUnidad, "Item producto id");
		}
		
		[Test()]
		public void ClienteFacturaActualizadoLuegoDeCrearFactura_FacturaMantieneDatosFijos()
		{
			try {
				this.facturaCompleta.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			// Copio los datos del cliente antes de actualizarlos
			string razonSocialVieja = this.ClienteSinBonificacion.RazonSocial;
			TipoResponsable tipoResponsableViejo = this.ClienteSinBonificacion.TipoResponsable;
			
			base.FlushAndRecreateScope();
			
			// Actulizo el cliente asociado a la factura, modifico su razón social
			this.ClienteSinBonificacion.RazonSocial += "agregado";
			this.ClienteSinBonificacion.TipoResponsable = TipoResponsable.ResponsableNoInscripto;
			this.ClienteSinBonificacion.Update();
			
			this.FlushAndRecreateScope();
			
			// Me aseguro que la factura mantenga los datos fijos
			Factura[] facturas = Factura.FindAll();
			
			Assert.AreEqual(1, facturas.Length, "Cantidad de facturas insertadas");
			
			Factura fr = facturas[0];
			
			Assert.AreEqual(this.ClienteSinBonificacion.Codigo, fr.Cliente.Codigo, "Id cliente");
			Assert.AreEqual(razonSocialVieja, fr.ClienteRazonSocial, "ClienteRazonSocial");
			Assert.AreEqual(tipoResponsableViejo, fr.ClienteTipoResponsable, "ClienteTipoResponsable");
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinDatos()
		{
			this.facturaVacia.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinFecha()
		{
			this.facturaCompleta.Fecha = this.facturaVacia.Fecha;
			this.facturaCompleta.Create();
		}

		[Test()]
		public void SinTipo()
		{
			// Por defecto tiene asignado un tipo, por eso este test pasa.
			this.facturaCompleta.Tipo = this.facturaVacia.Tipo;
			this.facturaCompleta.Create();
		}

		[Test()]
		public void SinCondicionVenta()
		{
			// La condición de venta por defecto es contado
			this.facturaCompleta.CondicionVenta = this.facturaVacia.CondicionVenta;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinCliente()
		{
			// TODO: Hacer un tests que verifique que el cliente exista.
			this.facturaCompleta.Cliente = this.facturaVacia.Cliente;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinCodigo()
		{
			this.facturaCompleta.Codigo = this.facturaVacia.Codigo;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinNumero()
		{
			this.facturaCompleta.Numero = this.facturaVacia.Numero;
			this.facturaCompleta.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinItems()
		{
			this.facturaCompleta.Items = this.facturaVacia.Items;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteBruto()
		{
			this.facturaCompleta.ImporteBruto = this.facturaVacia.ImporteBruto;
			this.facturaCompleta.Create();
		}

		[Test()]
		public void SinImporteBonificacion()
		{
			this.facturaCompleta.ImporteBonificacion =
				this.facturaVacia.ImporteBonificacion;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteNeto()
		{
			this.facturaCompleta.ImporteNeto = this.facturaVacia.ImporteNeto;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteIva()
		{
			// Como facturaCompleta es una FacturaA, debe tener un importe de iva.
			this.facturaCompleta.ImporteIva = this.facturaVacia.ImporteIva;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteTotal()
		{
			this.facturaCompleta.ImporteTotal = this.facturaVacia.ImporteTotal;
			this.facturaCompleta.Create();
		}

		[Test()]
		public void SinObservaciones()
		{
			this.facturaCompleta.Observaciones = this.facturaVacia.Observaciones;
			this.facturaCompleta.Create();
		}
		#endregion

		#region Pruebas de alto nivel (código repetido, bonificación excesiva, etc)
		#region Fechas
		[Test()]
		public void FechaMinimaCorrecta()
		{
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 1;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void FechaMinimaCorrecta_FacturaAnuladaNoSeTieneEnCuenta()
		{
			// Creo una factura normal
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			/* Ahora creo una factura anulada con una fecha posterior a la factura
			 * anterior. Esta factura no se debería tener en cuenta cuando se valida
			 * la fecha de facturas posteriores. */
			Factura facturaNueva1 = TestBase.CrearFactura();
			facturaNueva1.Anulada = true;
			facturaNueva1.Cliente = this.facturaCompleta.Cliente;
			facturaNueva1.Fecha = DateTime.Today - TimeSpan.FromDays(3);
			facturaNueva1.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva1.Codigo = this.facturaCompleta.Codigo;
			facturaNueva1.Numero = this.facturaCompleta.Numero + 1;
			
			facturaNueva1.Items.Clear();
			facturaNueva1.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva1);
			
			facturaNueva1.Create();
			
			this.FlushAndRecreateScope();
			
			// Esta factura se debería poder crear correctamente (no se tiene en cuenta
			// la fecha de la factura anulada).
			Factura facturaNueva2 = TestBase.CrearFactura();
			facturaNueva2.Cliente = this.facturaCompleta.Cliente;
			facturaNueva2.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			facturaNueva2.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva2.Codigo = this.facturaCompleta.Codigo;
			facturaNueva2.Numero = facturaNueva1.Numero + 1;
			
			facturaNueva2.Items.Clear();
			facturaNueva2.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 5));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva2);
			
			try {
				facturaNueva2.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void FechaMinimaCorrecta_GuardandoFactura()
		{
			/* Este tipo de test guarda cambios en una factura existente. Para eso
			 * la factura debe ser manual en la práctica. */
			
			#region Código de preparación para el test
			// Creo dos facturas manuales
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Codigo = 8;
			this.facturaCompleta.Numero = 1;
			this.facturaCompleta.FacturaManual = true;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(3);
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 1;
			facturaNueva.FacturaManual = true;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			#endregion
			
			this.FlushAndRecreateScope();
			
			// Ahora levanto la última factura, y le cambio la fecha
			facturaNueva = Factura.Find(facturaNueva.Id);
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			
			try {
				facturaNueva.UpdateAndFlush();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("No se pudo guardar factura: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Me aseguro que la fecha esté bien
			facturaNueva = Factura.Find(facturaNueva.Id);
			Assert.AreEqual(DateTime.Today - TimeSpan.FromDays(4), facturaNueva.Fecha, "Fecha factura guardada");
		}
		
		[Test()]
		public void FechaMinimaCorrectaFacturasDistintoCodigo()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Codigo = 3;
			this.facturaCompleta.Numero = 1;
			this.facturaCompleta.Fecha = DateTime.Today;
			this.facturaCompleta.Create();
			
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = 4;
			facturaNueva.Numero = 1;
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}

		[Test()]
		public void FechaMinimaIncorrecta()
		{
			this.facturaCompleta.Codigo = 4;
			this.facturaCompleta.Numero = 1;
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(5);
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 1;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Fecha", ex),
				              "Propiedad Fecha errónea");
			}
		}
		
		[Test()]
		public void FechaMinimaIncorrecta_GuardandoFactura()
		{
			/* Este tipo de test guarda cambios en una factura existente. Para eso
			 * la factura debe ser manual en la práctica. */
			
			#region Código de preparación para el test
			// Creo dos facturas manuales
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Codigo = 8;
			this.facturaCompleta.Numero = 1;
			this.facturaCompleta.FacturaManual = true;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 1;
			facturaNueva.FacturaManual = true;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			#endregion
			
			this.FlushAndRecreateScope();
			
			// Ahora levanto la última factura, y le cambio la fecha por una inválida
			facturaNueva = Factura.Find(facturaNueva.Id);
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(5);
			
			try {
				facturaNueva.UpdateAndFlush();
				Assert.Fail("Se pudo guardar la factura con fecha inválida");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Fecha", ex),
				              "Propiedad Fecha errónea");
			}
		}
		
		[Test()]
		public void FechaMinimaIncorrecta_NumeroAnteriorNoConcecutivo()
		{
			// Creo una factura.
			this.facturaCompleta.Codigo = 4;
			this.facturaCompleta.Numero = 1;
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo una factura con un número vacío entre medio con la anterior,
			// pero con la fecha más vieja que esa factura.
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(5);
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 2;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Fecha", ex),
				              "Propiedad Fecha errónea");
			}
		}
		
		[Test()]
		public void FechaMaximaCorrecta()
		{
			this.facturaCompleta.Fecha = DateTime.Today;
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today;
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 1;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void FechaMaximaCorrecta_GuardandoFactura_FacturaAnuladaNoSeTieneEnCuenta()
		{
			// Creo una factura normal
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			/* Ahora creo una factura anulada con una fecha posterior a la factura
			 * anterior. Esta factura no se debería tener en cuenta cuando se valida
			 * la fecha de facturas posteriores. */
			Factura facturaNueva1 = TestBase.CrearFactura();
			facturaNueva1.Anulada = true;
			facturaNueva1.Cliente = this.facturaCompleta.Cliente;
			facturaNueva1.Fecha = DateTime.Today - TimeSpan.FromDays(3);
			facturaNueva1.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva1.Codigo = this.facturaCompleta.Codigo;
			facturaNueva1.Numero = this.facturaCompleta.Numero + 1;
			
			facturaNueva1.Items.Clear();
			facturaNueva1.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva1);
			
			facturaNueva1.Create();
			
			this.FlushAndRecreateScope();
			
			// Ahora creo otra factura normal posterior a la anulada
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(2);
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 2;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);
			
			facturaNueva.Create();
			
			this.FlushAndRecreateScope();
			
			// Ahora levanto la primer factura creada factura, y le cambio la fecha,
			// le asigno un valor igual a la factura siguiente sin anular.
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(2);
			
			try {
				this.facturaCompleta.UpdateAndFlush();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("No se pudo guardar factura: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Me aseguro que la fecha esté bien
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			Assert.AreEqual(DateTime.Today - TimeSpan.FromDays(2), this.facturaCompleta.Fecha, "Fecha factura guardada");
		}
		
		[Test()]
		public void FechaMaximaCorrecta_GuardandoFactura()
		{
			/* Este tipo de test guarda cambios en una factura existente. Para eso
			 * la factura debe ser manual en la práctica. */
			
			#region Código de preparación para el test
			// Creo dos facturas manuales
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(1);
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Codigo = 8;
			this.facturaCompleta.Numero = 1;
			this.facturaCompleta.FacturaManual = true;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today; // Fecha de ayer
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 1;
			facturaNueva.FacturaManual = true;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			#endregion
			
			this.FlushAndRecreateScope();
			
			// Ahora levanto la última factura, y le cambio la fecha
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			this.facturaCompleta.Fecha = DateTime.Today;
			
			try {
				this.facturaCompleta.UpdateAndFlush();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("No se pudo guardar factura: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Me aseguro que la fecha esté bien
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			Assert.AreEqual(DateTime.Today, this.facturaCompleta.Fecha, "Fecha factura guardada");
		}
		
		[Test()]
		public void FechaMaximaIncorrecta()
		{
			this.facturaCompleta.Fecha = DateTime.Today;
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today + TimeSpan.FromDays(1);
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 1;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));

			try {
				facturaNueva.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Fecha", ex),
				              "Propiedad Fecha errónea");
			}
		}
		
		[Test()]
		public void FechaMaximaIncorrecta_GuardandoFactura()
		{
			/* Este tipo de test guarda cambios en una factura existente. Para eso
			 * la factura debe ser manual en la práctica. */
			
			#region Código de preparación para el test
			// Creo dos facturas manuales
			this.facturaCompleta.Fecha = DateTime.Today - TimeSpan.FromDays(1);
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Codigo = 8;
			this.facturaCompleta.Numero = 1;
			this.facturaCompleta.FacturaManual = true;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			Factura facturaNueva = TestBase.CrearFactura();
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			facturaNueva.Fecha = DateTime.Today;
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Codigo = this.facturaCompleta.Codigo;
			facturaNueva.Numero = this.facturaCompleta.Numero + 1;
			facturaNueva.FacturaManual = true;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			#endregion
			
			this.FlushAndRecreateScope();
			
			// Ahora levanto la última factura, y le cambio la fecha por una inválida
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			this.facturaCompleta.Fecha = DateTime.Today + TimeSpan.FromDays(1);
			
			try {
				this.facturaCompleta.UpdateAndFlush();
				Assert.Fail("Se pudo guardar la factura con fecha inválida");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Fecha", ex),
				              "Propiedad Fecha errónea");
			}
		}
		#endregion
		
		#region Cantidades
		[Test]
		public void CantidadDemasiadoGrande()
		{
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 10000));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Se pudo crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Cantidad", ex), "Propiedad fallida");
			}
		}
		#endregion
		
		#region Importes
		[Test]
		public void ImporteTotalDemasiadoGrande()
		{
			this.ProductoGriferia.ImporteSinIva = 1000.0m;
			this.ProductoGriferia.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			// Agrego un item que hace que el importe del mismo sea mayor a 999.999,99
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1000.0m));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Se pudo crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("ImporteBruto", ex), "Propiedad fallida");
			}
		}
		
		[Test()]
		public void ImportesValidosConBonificacion()
		{
			// Modifico el IVA del cliente sin bonificación
			this.ClienteConBonificacion.TipoResponsable = TipoResponsable.ResponsableInscripto;
			this.ClienteConBonificacion.Update();
			
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteBruto, "Importe bruto");
				Assert.AreEqual(278.89m, this.facturaCompleta.ImporteBonificacion, "Importe bonificación");
				Assert.AreEqual(2510.00m, this.facturaCompleta.ImporteNeto, "Importe neto");
				Assert.AreEqual(527.10m, this.facturaCompleta.ImporteIva, "Importe iva");
				Assert.AreEqual(3037.10, this.facturaCompleta.ImporteTotal, "Importe total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void ImportesValidosSinBonificacion()
		{
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteBruto, "Importe bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe bonificación");
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteNeto, "Importe neto");
				Assert.AreEqual(585.67m, this.facturaCompleta.ImporteIva, "Importe iva");
				Assert.AreEqual(3374.56m, this.facturaCompleta.ImporteTotal, "Importe total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}

		[Test()]
		public void ImporteBrutoInvalido()
		{
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			this.facturaCompleta.ImporteBruto = 1133.23m; // Incorrecto
			this.facturaCompleta.ImporteBonificacion = 278.89m;
			this.facturaCompleta.ImporteNeto = 2510.00m;			
			this.facturaCompleta.ImporteIva = 527.10m;
			this.facturaCompleta.ImporteTotal = 3037.10m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteBruto", ex),
				              "Propiedad ImporteBruto errónea");
			}
		}

		[Test()]
		public void ImporteBonificacionInvalido()
		{
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			this.facturaCompleta.ImporteBruto = 2788.89m;
			this.facturaCompleta.ImporteBonificacion = 345.00m; // Incorrecto
			this.facturaCompleta.ImporteNeto = 2510.00m;			
			this.facturaCompleta.ImporteIva = 527.10m;
			this.facturaCompleta.ImporteTotal = 3037.10m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteBonificacion", ex),
				              "Propiedad ImporteBonificacion errónea");
			}
		}
		
		[Test()]
		public void ImporteNetoInvalido()
		{
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			this.facturaCompleta.ImporteBruto = 2788.89m;
			this.facturaCompleta.ImporteBonificacion = 278.89m;
			this.facturaCompleta.ImporteNeto = 2900.22m; // Incorrecto
			this.facturaCompleta.ImporteIva = 527.10m;
			this.facturaCompleta.ImporteTotal = 3037.10m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteNeto", ex),
				              "Propiedad ImporteNeto errónea");
			}
		}

		[Test()]
		public void ImporteIvaInvalido()
		{
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			this.facturaCompleta.ImporteBruto = 2788.89m;
			this.facturaCompleta.ImporteBonificacion = 278.89m;
			this.facturaCompleta.ImporteNeto = 2510.00m;
			this.facturaCompleta.ImporteIva = 523.14m; // Incorrecto
			this.facturaCompleta.ImporteTotal = 3037.10m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteIva", ex),
				              "Propiedad ImporteIva errónea");
			}
		}

		[Test()]
		public void ImporteTotalInvalido()
		{
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			this.facturaCompleta.ImporteBruto = 2788.89m;
			this.facturaCompleta.ImporteBonificacion = 278.89m;
			this.facturaCompleta.ImporteNeto = 2510.00m;
			this.facturaCompleta.ImporteIva = 527.10m;
			this.facturaCompleta.ImporteTotal = 3034.99m; // Incorrecto
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteTotal", ex),
				              "Propiedad ImporteTotal errónea");
			}
		}

		[Test()]
		public void ImportesValidosFacturaB()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
//			this.facturaCompleta.ImporteBruto = 3374.56m;
//			this.facturaCompleta.ImporteBonificacion = 337.46m;
//			this.facturaCompleta.ImporteNeto = 3037.10m;
//			this.facturaCompleta.ImporteIva = 0.00m;
//			this.facturaCompleta.ImporteTotal = 3037.10m;
			
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void ImporteIvaInvalidoFacturaB()
		{
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			this.facturaCompleta.ImporteBruto = 3374.56m;
			this.facturaCompleta.ImporteBonificacion = 337.46m;
			this.facturaCompleta.ImporteNeto = 3037.10m;
			this.facturaCompleta.ImporteIva = 637.79m; // no puede tener valor en factura B
			this.facturaCompleta.ImporteTotal = 3037.10m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteIva", ex),
				              "Propiedad ImporteIva errónea");
			}
		}

		[Test()]
		public void ImportesClienteSinBonificacion()
		{
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		#endregion
		
		#region Clave (código y número)
		[Test()]
		public void ClaveRepetida()
		{
			// Creo la factura completa
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Completo la factura vacía con la misma clave
			Factura facturaRepetida = TestBase.CrearFactura();
			facturaRepetida.Tipo = FacturaTipoVenta.FacturaA;
			facturaRepetida.Codigo = this.facturaCompleta.Codigo;
			facturaRepetida.Numero = this.facturaCompleta.Numero;
			
			facturaRepetida.Items.Clear();
			facturaRepetida.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaRepetida);
			
			try {
				facturaRepetida.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Numero", ex));
			}
		}
		
		[Test()]
		public void ClaveNoRepetida()
		{
			// Creo la factura completa
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Create();
			
			// Completo la factura vacía con la misma clave
			Factura facturaRepetida = TestBase.CrearFactura();
			facturaRepetida.Tipo = this.facturaCompleta.Tipo;
			facturaRepetida.Codigo = this.facturaCompleta.Codigo;
			facturaRepetida.Numero = this.facturaCompleta.Numero + 1;
			
			facturaRepetida.Cliente = this.facturaCompleta.Cliente;
			
			facturaRepetida.Items.Clear();
			facturaRepetida.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaRepetida);
			
			try {
				facturaRepetida.Create();
			}
			catch (Exception e) {
				Assert.Fail("No se pudo crear la factura: " + e.Message);
			}
		}
		
		[Test()]
		public void NumeroFacturaConsecutivoConMismoCodigo()
		{
			try {
				this.facturaCompleta.Codigo = 4;
				this.facturaCompleta.Numero = 100;
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
			
			// Creo la siguiente factura del mismo código con un número consecutivo.
			try {
				this.facturaCompleta2.Codigo = 4;
				this.facturaCompleta2.Numero = 101;
				this.facturaCompleta2.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void NumeroFacturaNoConsecutivoConCodigoDistinto()
		{
			try {
				this.facturaCompleta.Codigo = 3;
				this.facturaCompleta.Numero = 100;
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
			
			// Creo una factura con código distinto a la anterior con un número no consecutivo
			try {
				this.facturaCompleta2.Codigo = 4;
				this.facturaCompleta2.Numero = 105;
				this.facturaCompleta2.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void NumeroFacturaNoConsecutivo()
		{
			try {
				this.facturaCompleta.Codigo = 4;
				this.facturaCompleta.Numero = 100;
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
			
			this.FlushAndRecreateScope();
			
			// Creo la siguiente factura con un número NO consecutivo.
			// Debería arrojar una excepción.
			try {
				this.facturaCompleta2.Codigo = 4;
				this.facturaCompleta2.Numero = 102;
				this.facturaCompleta2.Create();
				
				Assert.Fail("La segunda factura no debió haberse podido crear");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("Numero", e));
			}
		}
		#endregion
		
		#region Clientes
		[Test()]
		public void FacturaAConClienteResponsableInscripto()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.ResponsableInscripto;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Los totales de la factura
//			this.facturaCompleta.ImporteBruto = 2788.89m;
//			this.facturaCompleta.ImporteBonificacion = 0.00m;
//			this.facturaCompleta.ImporteNeto = 2788.89m;
//			this.facturaCompleta.ImporteIva = 585.67m;
//			this.facturaCompleta.ImporteTotal = 3374.56m;
			
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaAConClienteResponsableNoInscripto()
		{
			// Cliente Reponsable Inscripto
			Cliente cliente = this.ClienteSinBonificacion;
			cliente.TipoResponsable = TipoResponsable.ResponsableNoInscripto;
			cliente.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaAConClienteConsumidorFinal()
		{
			// Cliente Consumidor final
			Cliente cliente = this.ClienteSinBonificacion;
			cliente.TipoResponsable = TipoResponsable.ConsumidorFinal;
			cliente.Cuit = new CUIT();
			
			cliente.UpdateAndFlush();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			// Los totales de la factura
			this.facturaCompleta.ImporteBruto = 2788.89m;
			this.facturaCompleta.ImporteBonificacion = 0.00m;
			this.facturaCompleta.ImporteNeto = 2788.89m;
			this.facturaCompleta.ImporteIva = 585.67m;
			this.facturaCompleta.ImporteTotal = 3374.56m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura A con Consumidor Final no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		
		[Test()]
		public void FacturaAConClienteExento()
		{
			// Cliente Exento
			Cliente cliente = this.ClienteSinBonificacion;
			cliente.TipoResponsable = TipoResponsable.Exento;
			
			try {
				cliente.UpdateAndFlush();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			// Los totales de la factura
			this.facturaCompleta.ImporteBruto = 2788.89m;
			this.facturaCompleta.ImporteBonificacion = 0.00m;
			this.facturaCompleta.ImporteNeto = 2788.89m;
			this.facturaCompleta.ImporteIva = 585.67m;
			this.facturaCompleta.ImporteTotal = 3374.56m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura A con un Excento no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		
		[Test()]
		public void FacturaAConClienteMonotributo()
		{
			// Cliente Monotributo
			Cliente cliente = this.ClienteSinBonificacion;
			cliente.TipoResponsable = TipoResponsable.Monotributo;
			cliente.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			// Los totales de la factura
			this.facturaCompleta.ImporteBruto = 2788.89m;
			this.facturaCompleta.ImporteBonificacion = 0.00m;
			this.facturaCompleta.ImporteNeto = 2788.89m;
			this.facturaCompleta.ImporteIva = 585.67m;
			this.facturaCompleta.ImporteTotal = 3374.56m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura A con Monotributo no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		
		[Test()]
		public void FacturaBConClienteConsumidorFinal()
		{
			// Cliente Consumidor Final
			Cliente cliente = this.ClienteSinBonificacion;
			cliente.TipoResponsable = TipoResponsable.ConsumidorFinal;
			cliente.Cuit = new CUIT();
			cliente.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaBConClienteExento()
		{
			// Cliente Reponsable Inscripto
			Cliente cliente = this.ClienteSinBonificacion;
			cliente.TipoResponsable = TipoResponsable.Exento;
			cliente.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaBConClienteMonotributo()
		{
			// Cliente Monotributo
			Cliente cliente = this.ClienteSinBonificacion;
			cliente.TipoResponsable = TipoResponsable.Monotributo;
			cliente.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Los totales de la factura
//			this.facturaCompleta.ImporteBruto = 3374.56m;
//			this.facturaCompleta.ImporteBonificacion = 0.00m;
//			this.facturaCompleta.ImporteNeto = 3374.56m;
//			this.facturaCompleta.ImporteIva = 0.00m;
//			this.facturaCompleta.ImporteTotal = 3374.56m;
			
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaBConClienteResponsableInscripto()
		{
			// Cliente Monotributo
			Cliente cliente = this.ClienteSinBonificacion;
			cliente.TipoResponsable = TipoResponsable.ResponsableInscripto;
			cliente.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			// Los totales de la factura
			this.facturaCompleta.ImporteBruto = 3374.56m;
			this.facturaCompleta.ImporteBonificacion = 0.00m;
			this.facturaCompleta.ImporteNeto = 3374.56m;
			this.facturaCompleta.ImporteIva = 0.00m;
			this.facturaCompleta.ImporteTotal = 3374.56m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura B con Responsable Inscripto no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		
		[Test()]
		public void FacturaBConClienteResponsableNoInscripto()
		{
			// Cliente Responsable no inscripto
			Cliente cliente = this.ClienteSinBonificacion;
			cliente.TipoResponsable = TipoResponsable.ResponsableNoInscripto;
			cliente.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			ItemFactura item2 = TestBase.CrearItemFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			// Los totales de la factura
			this.facturaCompleta.ImporteBruto = 3374.56m;
			this.facturaCompleta.ImporteBonificacion = 0.00m;
			this.facturaCompleta.ImporteNeto = 3374.56m;
			this.facturaCompleta.ImporteIva = 0.00m;
			this.facturaCompleta.ImporteTotal = 3374.56m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura B con Responsable No Inscripto no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		#endregion
		
		#region Productos abstractos
		[Test()]
		public void FacturaAConProductoAbstracto()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Modifico los productos para que sean productos abstractos
			this.ProductoGriferia.ImporteSinIva = 0;
			this.ProductoGriferia.PrecioCosto = 0;
			this.ProductoGriferia.ProductoAbstracto = true;
			this.ProductoGriferia.Save();
			
			this.ProductoMosaico.ImporteSinIva = 0;
			this.ProductoMosaico.PrecioCosto = 0;
			this.ProductoMosaico.ProductoAbstracto = true;
			this.ProductoMosaico.Save();
			
			ItemFactura item1 = TestBase.CrearItemFactura(ProductoGriferia, 1);
			item1.PrecioUnitario = 12;
			ItemFactura item2 = TestBase.CrearItemFactura(ProductoMosaico, 125.3m);
			item2.PrecioUnitario = 23;
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido que los importes sean válidos
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(607.72m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3501.62m, this.facturaCompleta.ImporteTotal, "Importe Total");
				
				// Verifico que cuando el producto es abstracto, se copie la descripción del mismo
				Assert.AreEqual(this.ProductoGriferia.Descripcion, this.facturaCompleta.Items[0].ProductoDescripcion, "Item producto descripción grifería");
				Assert.AreEqual(this.ProductoMosaico.Descripcion, this.facturaCompleta.Items[1].ProductoDescripcion, "Item producto descripción mosaico");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaBConProductoAbstracto()
		{
			// Modifico el IVA del cliente sin bonificación
			this.ClienteSinBonificacion.TipoResponsable = TipoResponsable.ConsumidorFinal;
			this.ClienteSinBonificacion.Cuit = new CUIT();
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Modifico los productos para que sean productos abstractos
			this.ProductoGriferia.ImporteSinIva = 0;
			this.ProductoGriferia.PrecioCosto = 0;
			this.ProductoGriferia.ProductoAbstracto = true;
			this.ProductoGriferia.Save();
			
			this.ProductoMosaico.ImporteSinIva = 0;
			this.ProductoMosaico.PrecioCosto = 0;
			this.ProductoMosaico.ProductoAbstracto = true;
			this.ProductoMosaico.Save();
			
			ItemFactura item1 = TestBase.CrearItemFactura(ProductoGriferia, 1);
			item1.PrecioUnitario = 12;
			ItemFactura item2 = TestBase.CrearItemFactura(ProductoMosaico, 125.3m);
			item2.PrecioUnitario = 23;
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido que los importes sean válidos
				
				/* NOTA: Los importes son distintos al método FacturaAConProductoAbstracto porque
				 * el precio unitario de los items con producto abstracto se matiene igual (en este
				 * caso de Factura B se debería agregarle el IVA a los mismos para que de lo mismo). */
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		#endregion
		
		#region Documentos de débito
		[Test]
		public void FacturaContado_NoCrea_DocumentoDebito()
		{
			this.facturaCompleta.CondicionVenta.Condicion = CondicionPago.Contado;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Busco los documentos de débito asociados a la factura
			DocumentoDebito[] documentosDebito =
				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
			
			Assert.AreEqual(0, documentosDebito.Length, "Cantidad documentos débito");
		}
		
		[Test]
		public void FacturaCredito_SiCrea_DocumentoDebito()
		{
			this.facturaCompleta.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.facturaCompleta.CondicionVenta.Plazo = 10; // Plazo de diez días
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Busco los documentos de débito asociados a la factura
			DocumentoDebito[] documentosDebito =
				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
			
			Assert.AreEqual(1, documentosDebito.Length, "Cantidad documentos débito");
			
			DocumentoDebito dd = documentosDebito[0];
			Assert.IsNotNull(dd.Cliente, "Documento débito cliente no nulo");
			Assert.AreEqual(this.facturaCompleta.Cliente.Codigo, dd.Cliente.Codigo, "Documento débito código cliente");
			
			Assert.AreEqual(TipoDocumento.Factura, dd.TipoDocumento, "Documento débito tipo documento");
			Assert.AreEqual(this.facturaCompleta.Id, dd.IdDocumento, "Documento débito id documento");
			Assert.AreEqual(this.facturaCompleta.Codigo, dd.CodigoFactura, "Documento débito codigo factura");
			Assert.AreEqual(this.facturaCompleta.Numero, dd.NumeroFactura, "Documento débito numero factura");
			Assert.AreEqual(this.facturaCompleta.Fecha + TimeSpan.FromDays(this.facturaCompleta.CondicionVenta.Plazo),
			                dd.FechaVencimiento, "Documento débito fecha vencimiento");
			
			Assert.AreEqual(this.facturaCompleta.ImporteTotal, dd.ImporteTotal, "Documento débito importe total");
			Assert.AreEqual(this.facturaCompleta.ImporteTotal, dd.ImportePendiente, "Documento débito importe pendiente");
		}
		
		[Test]
		public void FacturaCreditoEliminada_EliminaDebitoCliente()
		{
			this.facturaCompleta.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.facturaCompleta.CondicionVenta.Plazo = 10; // Plazo de diez días
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			
			this.facturaCompleta2.Cliente = this.facturaCompleta.Cliente;
			this.facturaCompleta2.Codigo = this.facturaCompleta.Codigo;
			this.facturaCompleta2.Numero = this.facturaCompleta.Numero + 1;
			this.facturaCompleta2.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.facturaCompleta2.CondicionVenta.Plazo = 10; // Plazo de diez días
			this.facturaCompleta2.Create();
			
			this.FlushAndRecreateScope();
			
			// Busco los documentos de débito asociados a la factura
			DocumentoDebito[] documentosDebito = DocumentoDebito.FindAll();
			
			Assert.AreEqual(2, documentosDebito.Length, "Cantidad documentos débito");
			
			// Ahora elimino la factura a crédito creada recién
			this.facturaCompleta.Delete();
			
			this.FlushAndRecreateReadOnlyScope();
			
			// No debe haber más documentos de débito para esa factura
			// Busco los documentos de débito asociados a la factura
			documentosDebito = DocumentoDebito.FindAll();
			
			Assert.AreEqual(1, documentosDebito.Length, "Cantidad documentos débito 2");
			
			// Solo hay para la factura no eliminada
			DocumentoDebito dd = documentosDebito[0];
			
			Assert.AreEqual(TipoDocumento.Factura, dd.TipoDocumento, "Documento débito tipo documento");
			Assert.AreEqual(this.facturaCompleta2.Id, dd.IdDocumento, "Documento débito id documento");
		}
		
		[Test]
		public void FacturaCreditoActualizada_NoVuelve_A_CrearDocumentoDebito()
		{
			this.facturaCompleta.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.facturaCompleta.CondicionVenta.Plazo = 10; // Plazo de diez días
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Busco los documentos de débito asociados a la factura
			DocumentoDebito[] documentosDebito =
				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
			
			Assert.AreEqual(1, documentosDebito.Length, "Cantidad documentos débito");
			
			// Actualizo la factura
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			this.facturaCompleta.Observaciones += "agregado";
			
			try {
				this.facturaCompleta.Update();
				this.FlushAndRecreateReadOnlyScope();
			}
			catch (Exception ex) {
				Assert.Fail("Error al actualizar factura:\n" + ex.ToString());
			}
			
			// Me aseguro que siga habiendo un solo documento de débito
			documentosDebito =
				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
			
			Assert.AreEqual(1, documentosDebito.Length, "Cantidad documentos débito");
		}
		
		[Test]
		public void FacturaCreditoAnulada_EliminaDebitoAsociado()
		{
			this.facturaCompleta.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.facturaCompleta.CondicionVenta.Plazo = 10; // Plazo de diez días
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Me aseguro que se haya creado el DocumentoDebito correspondiente
			DocumentoDebito[] documentosDebito =
				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
			
			Assert.AreEqual(1, documentosDebito.Length, "Cantidad documentos débito");
			
			this.FlushAndRecreateScope();
			
			// Ahora anulo la factura y verifico que se haya eliminado el DocumentoDebito
			// asociado
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			this.facturaCompleta.Anulada = true;
			
			this.FlushAndRecreateReadOnlyScope();
			
			documentosDebito =
				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
			
			Assert.AreEqual(0, documentosDebito.Length, "Cantidad documentos débito");
		}
		
		[Test]
		public void FacturaCreditoDesanulada_RecreaDebitoAsociado()
		{
			#region Código de preparación para el test
			this.facturaCompleta.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.facturaCompleta.CondicionVenta.Plazo = 10; // Plazo de diez días
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Me aseguro que se haya creado el DocumentoDebito correspondiente
			DocumentoDebito[] documentosDebito =
				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
			
			Assert.AreEqual(1, documentosDebito.Length, "Preparación 1: Cantidad documentos débito");
			
			this.FlushAndRecreateScope();
			
			// Ahora anulo la factura, y me aseguro que se haya eliminado el DocumentoDebito
			// correspondiente.
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			this.facturaCompleta.Anulada = true;
			
			this.FlushAndRecreateReadOnlyScope();
			
			documentosDebito =
				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
			
			Assert.AreEqual(0, documentosDebito.Length, "Preparación 2: Cantidad documentos débito");
			#endregion
			
			this.FlushAndRecreateScope();
			
			// Ahora desanulo la Factura: se debería recrear el DocumentoDebito
			// correspondiente.
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			this.facturaCompleta.Anulada = false;
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Me aseguro que se haya creado el DocumentoDebito correspondiente
			documentosDebito =
				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
			
			Assert.AreEqual(1, documentosDebito.Length, "Cantidad documentos débito");
		}
		#endregion
		
		#region Anulaciones
		[Test]
		public void FacturaNormal_PuedeAnularse()
		{
			this.facturaCompleta.Historica = false;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			this.facturaCompleta.Anulada = true;
			this.facturaCompleta.Update();
			
			this.FlushAndRecreateReadOnlyScope();
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			Assert.IsTrue(this.facturaCompleta.Anulada, "Factura anulada");
		}
		
		[Test]
		public void FacturaHistorica_NoPuedeAnularse()
		{
			this.facturaCompleta.Historica = true;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			Assert.IsTrue(this.facturaCompleta.Historica, "Factura histórica");
			
			this.facturaCompleta.Anulada = true;
			
			try {
				this.facturaCompleta.UpdateAndFlush();
				Assert.Fail("Se pudo anular");
			}
			catch (ActiveRecordValidationException) {
			}
		}
		#endregion
		
		#region Eliminaciones
		[Test]
		public void FacturaHistorica_NoPuedeEliminarse()
		{
			this.facturaCompleta.Historica = true;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			
			try {
				this.facturaCompleta.DeleteAndFlush();
				Assert.Fail("Se pudo eliminar");
			}
			catch (ActiveRecordException) {
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			Assert.AreEqual(1,
			                Factura.FindAll(Restrictions.Eq("Id", this.facturaCompleta.Id)).Length,
			                "Factura no eliminada");
		}
		
		[Test]
		public void EliminarFactura_ExistePreNotaCredito_ConEsaFacturaAsociada()
		{
			// Creo la factura que más adelante voy a intentar eliminar
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			
			// Creo una PreNotaCredito con esta factura asociada
			PreNotaCredito preNotaCredito = TestBase.CrearPreNotaCredito();
			preNotaCredito.FacturaAsociada = this.facturaCompleta;
			preNotaCredito.Cliente = this.facturaCompleta.Cliente;
			preNotaCredito.Items.Clear();
			preNotaCredito.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(preNotaCredito);
			
			preNotaCredito.Create();
			
			this.FlushAndRecreateScope();
			
			// Ahora intento eliminar la factura
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			
			try {
				this.facturaCompleta.Delete();
				this.FlushAndRecreateReadOnlyScope();
			}
			catch (Exception ex) {
				Assert.Fail("Excepción arrojada: " + ex.ToString());
			}
			
			// Se debió haber eliminado la Factura y la PreFactura correspondiente
			Assert.IsTrue(Factura.TryFind(this.facturaCompleta.Id) == null, "La factura fue eliminada");
//			Assert.IsTrue(PreNotaCredito.TryFind(preNotaCredito.Id) == null, "La PreNotaCredito fue eliminada");
		}
		#endregion
		#endregion
	}
}
