// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

using NUnit.Framework;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using NHibernate.Criterion;

using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture()]
	public class NotaCreditoTests : DocumentoGenericoTestBase
	{
		#region Campos
		private Factura facturaCompleta;
		
		private NotaCredito notaCreditoCompleta;
		private NotaCredito notaCreditoVacia;
		#endregion

		#region Métodos de inicialización y finalización de la prueba
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();

			// Creo la factura con los datos completos para realizar
			// las pruebas de persistencia
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo;
			this.notaCreditoCompleta.Numero = this.facturaCompleta.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = this.facturaCompleta;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			this.notaCreditoVacia = new NotaCredito();
			
			this.FlushAndRecreateScope();
		}
		#endregion

		#region Pruebas de persistencia de datos
		[Test()]
		public void Creacion()
		{
			try {
				this.notaCreditoCompleta.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}

			base.FlushAndRecreateScope();
			
			NotaCredito[] notasCredito = NotaCredito.FindAll();
			
			Assert.AreEqual(1, notasCredito.Length, "Cantidad de facturas insertadas");
			
			NotaCredito fr = notasCredito[0];
			Assert.IsNotNull(fr.FacturaAsociada, "Factura asociada no nula");
			Assert.AreEqual(this.facturaCompleta.Id, fr.FacturaAsociada.Id, "Id factura asociada");
			
			Assert.AreEqual(this.facturaCompleta.Cliente.Codigo, fr.Cliente.Codigo, "Id cliente");
			Assert.AreEqual(4, fr.Codigo, "Código");
			Assert.AreEqual(23424, fr.Numero, "Numero");
			Assert.AreEqual(60.00m, fr.ImporteBruto, "Importe bruto");
			Assert.AreEqual(0.00m, fr.PorcentajeBonificacion, "Porcentaje bonificación");
			Assert.AreEqual(0.00m, fr.ImporteBonificacion, "Importe bonificación");
			Assert.AreEqual(60.00m, fr.ImporteNeto, "Importe Neto");
			Assert.AreEqual(12.60m, fr.ImporteIva, "Importe iva");
			Assert.AreEqual(72.60m, fr.ImporteTotal, "Importe total");
			Assert.AreEqual(72.60m, fr.ImporteFinal, "Importe final");
			Assert.IsFalse(fr.Anulada, "Anulada");
			Assert.IsFalse(fr.FacturaManual, "Factura manual");
			Assert.IsTrue(fr.Valida, "Válida");
			Assert.IsFalse(fr.Historica, "Válida");
			Assert.AreEqual("Estas son las observaciones", fr.Observaciones, "Observaciones");
			Assert.AreEqual(FacturaTipoVenta.FacturaA, fr.Tipo, "Tipo factura");
			Assert.AreEqual(DateTime.Today, fr.Fecha.Date, "Fecha");
			Assert.AreEqual(CondicionPago.Contado, fr.CondicionVenta.Condicion,
			                "Condicion de venta");
			Assert.AreEqual("NC", fr.TipoAbreviado, "Tipo abreviado");
		}
		
		[Test]
		[Ignore("En realidad, nunca se crea un nota de crédito con la condición de pago a Crédito")]
		public void NotaCredito_A_Credito()
		{
			this.notaCreditoCompleta.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("No se debió haber podido crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("CondicionVenta", ex));
			}
		}
		
		[Test]
		public void SinFacturaAsociada()
		{
			this.notaCreditoCompleta.FacturaAsociada = this.notaCreditoVacia.FacturaAsociada;
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("No se debió haber podido crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("FacturaAsociada", ex));
			}
		}
		#endregion
		
		#region Pruebas de alto nivel (código repetido, bonificación excesiva, etc)
		#region Validaciones específicas de NotaCredito
		[Test]
		public void CodigoDistintoAFacturaAsociada()
		{
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo + 1;
			
			this.notaCreditoCompleta.Create();
		}
		#endregion
		
		#region Documentos de crédito
		[Test]
		public void NotaCredito_SiempreCrea_DocumentoCredito()
		{
			this.notaCreditoCompleta.Create();
			this.FlushAndRecreateScope();
			
			// Busco los documentos de débito asociados a la factura
			DocumentoCredito[] documentosCredito =
				DocumentoCredito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.NotaCredito),
				                         Restrictions.Eq("IdDocumento", this.notaCreditoCompleta.Id));
			
			Assert.AreEqual(1, documentosCredito.Length, "Cantidad documentos crédito");
			
			DocumentoCredito dc = documentosCredito[0];
			Assert.IsNotNull(dc.Cliente, "Documento crédito cliente no nulo");
			Assert.AreEqual(this.notaCreditoCompleta.Cliente.Codigo, dc.Cliente.Codigo, "Documento crédito código cliente");
			
			Assert.AreEqual(TipoDocumento.NotaCredito, dc.TipoDocumento, "Documento crédito tipo documento");
			Assert.AreEqual(this.notaCreditoCompleta.Id, dc.IdDocumento, "Documento crédito id documento");
			Assert.AreEqual(this.notaCreditoCompleta.Codigo, dc.CodigoNotaCredito, "Documento crédito código");
			Assert.AreEqual(this.notaCreditoCompleta.Numero, dc.NumeroNotaCredito, "Documento crédito número");
			
			Assert.AreEqual(this.notaCreditoCompleta.ImporteTotal, dc.ImporteTotal, "Documento crédito importe total");
			Assert.AreEqual(this.notaCreditoCompleta.ImporteTotal, dc.ImportePendiente, "Documento crédito importe pendiente");
		}
		
		[Test]
		public void NotaCreditoActualizada_NoVuelve_A_Crear_DocumentoCredito()
		{
			this.notaCreditoCompleta.Create();
			this.FlushAndRecreateScope();
			
			// Busco los documentos de débito asociados a la factura
			DocumentoCredito[] documentosCredito =
				DocumentoCredito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.NotaCredito),
				                         Restrictions.Eq("IdDocumento", this.notaCreditoCompleta.Id));
			
			Assert.AreEqual(1, documentosCredito.Length, "Cantidad documentos crédito");
			
			// Actualizo la nota de crédito
			this.notaCreditoCompleta = NotaCredito.Find(this.notaCreditoCompleta.Id);
			this.notaCreditoCompleta.Observaciones += "agregado";
			
			try {
				this.notaCreditoCompleta.Update();
				this.FlushAndRecreateReadOnlyScope();
			}
			catch (Exception ex) {
				Assert.Fail("Excepción al actualizar:\n" + ex.ToString());
			}
			
			// Me aseguro que no se hayan creado más documentos de crédito
			documentosCredito =
				DocumentoCredito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.NotaCredito),
				                         Restrictions.Eq("IdDocumento", this.notaCreditoCompleta.Id));
			
			Assert.AreEqual(1, documentosCredito.Length, "Cantidad documentos crédito");
		}
		
		[Test]
		public void NotaCredito_AlEliminarse_SeEliminaElDebitoAsociado()
		{
			this.notaCreditoCompleta.Create();
			this.FlushAndRecreateScope();
			
			// Busco los documentos de débito asociados a la factura
			DocumentoCredito[] documentosCredito = DocumentoCredito.FindAll();
			
			Assert.AreEqual(1, documentosCredito.Length, "Cantidad documentos crédito");
			
			// Ahora elimino la factura a crédito creada recién
			this.notaCreditoCompleta.Delete();
			
			this.FlushAndRecreateReadOnlyScope();
			
			// No debe haber más documentos de crédito para esa nota de crédito
			documentosCredito =
				DocumentoCredito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.NotaCredito),
				                         Restrictions.Eq("IdDocumento", this.notaCreditoCompleta.Id));
			
			Assert.AreEqual(0, documentosCredito.Length, "Cantidad documentos crédito 2");
		}
		
		[Test]
		public void NotaCreditoAnulada_EliminaCreditoAsociado()
		{
			#region Preparación para el test
			this.notaCreditoCompleta.Create();
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Me aseguro que se haya creado el DocumentoCredito asociado
			DocumentoCredito[] documentosCredito =
				DocumentoCredito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.NotaCredito),
				                         Restrictions.Eq("IdDocumento", this.notaCreditoCompleta.Id));
			
			Assert.AreEqual(1, documentosCredito.Length, "Preparación 1: Cantidad documentos crédito");
			#endregion
			
			this.FlushAndRecreateScope();
			
			// Ahora anulo la NotaCredito
			this.notaCreditoCompleta = NotaCredito.Find(this.notaCreditoCompleta.Id);
			this.notaCreditoCompleta.Anulada = true;
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Me aseguro que ya no haya más crédito asociado a la nota de crédito
			documentosCredito =
				DocumentoCredito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.NotaCredito),
				                         Restrictions.Eq("IdDocumento", this.notaCreditoCompleta.Id));
			
			Assert.AreEqual(0, documentosCredito.Length, "Cantidad documentos crédito");
		}
		
		[Test]
		public void FacturaCreditoDesanulada_RecreaDebitoAsociado()
		{
//			#region Código de preparación para el test
//			this.facturaCompleta.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
//			this.facturaCompleta.CondicionVenta.Plazo = 10; // Plazo de diez días
//			this.facturaCompleta.Create();
//			
//			this.FlushAndRecreateReadOnlyScope();
//			
//			// Me aseguro que se haya creado el DocumentoDebito correspondiente
//			DocumentoDebito[] documentosDebito =
//				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
//				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
//			
//			Assert.AreEqual(1, documentosDebito.Length, "Preparación 1: Cantidad documentos débito");
//			
//			this.FlushAndRecreateScope();
//			
//			// Ahora anulo la factura, y me aseguro que se haya eliminado el DocumentoDebito
//			// correspondiente.
//			
//			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
//			this.facturaCompleta.Anulada = true;
//			
//			this.FlushAndRecreateReadOnlyScope();
//			
//			documentosDebito =
//				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
//				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
//			
//			Assert.AreEqual(0, documentosDebito.Length, "Preparación 2: Cantidad documentos débito");
//			#endregion
//			
//			this.FlushAndRecreateScope();
//			
//			// Ahora desanulo la Factura: se debería recrear el DocumentoDebito
//			// correspondiente.
//			
//			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
//			this.facturaCompleta.Anulada = false;
//			
//			this.FlushAndRecreateReadOnlyScope();
//			
//			// Me aseguro que se haya creado el DocumentoDebito correspondiente
//			documentosDebito =
//				DocumentoDebito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Factura),
//				                        Restrictions.Eq("IdDocumento", this.facturaCompleta.Id));
//			
//			Assert.AreEqual(1, documentosDebito.Length, "Cantidad documentos débito");
		}
		#endregion

		#region Validaciones sobre la Factura asociada
		[Test]
		public void ConFacturaAsociada_FacturaAnulada()
		{
			// Borro todas las facturas
			ActiveRecordValidationBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a crear la factura pero como anulada
			this.MarcarComoNoPersistido(ref this.facturaCompleta);
			
			this.facturaCompleta.Anulada = true;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("FacturaAsociada", ex));
			}
		}
		
		[Test]
		public void ConFacturaAsociada_FacturaInvalida()
		{
			// Borro todas las facturas
			ActiveRecordValidationBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a crear la factura pero como anulada
			this.MarcarComoNoPersistido(ref this.facturaCompleta);
			
			this.facturaCompleta.Valida = false;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("FacturaAsociada", ex));
			}
		}
		#endregion
		
		#region Validaciones sobre el cliente
		[Test]
		public void ClienteNotaCredito_IgualA_ClienteFacturaAsociada()
		{
			// Borro todas las facturas
			ActiveRecordValidationBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a crear la factura pero como anulada
			this.MarcarComoNoPersistido(ref this.facturaCompleta);
			
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex), "No se pudo crear la nota de crédito");
			}
		}
		
		[Test]
		public void ClienteNotaCredito_DistintoA_ClienteFacturaAsociada()
		{
			// Borro todas las facturas
			ActiveRecordValidationBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a crear la factura pero como anulada
			this.MarcarComoNoPersistido(ref this.facturaCompleta);
			
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteConBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex));
			}
		}
		#endregion
		
		#region Validaciones de tipos (Factura A con Nota Credito B, etc...)
		[Test]
		public void FacturaAsociadaTipoA_NotaCreditoTipoA()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Update();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void FacturaAsociadaTipoB_NotaCreditoTipoB()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion2;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Update();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion2;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void FacturaAsociadaTipoA_NotaCreditoTipoB()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Update();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion2;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo + 1;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Tipo", ex), "Tipo - Propiedad fallida");
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex), "Cliente - Propiedad fallida");
			}
		}
		
		[Test]
		public void FacturaAsociadaTipoB_NotaCreditoTipoA()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion2;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Update();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo + 1;
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Tipo", ex), "Tipo - Propiedad fallida");
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex), "Cliente - Propiedad fallida");
			}
		}
		#endregion
		
		#region Validaciones sobre los importes
		[Test]
		public void ImporteNotaCredito_MenorA_ImporteFacturaAsociada()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito con un importe inferior a la factura
			// asociada
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo;
			this.notaCreditoCompleta.Numero = this.facturaCompleta.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void ImporteNotaCredito_IgualA_ImporteFacturaAsociada()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito con un importe inferior a la factura
			// asociada
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo;
			this.notaCreditoCompleta.Numero = this.facturaCompleta.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void ImporteNotaCredito_MayorA_ImporteFacturaAsociada()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito con un importe inferior a la factura
			// asociada
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo;
			this.notaCreditoCompleta.Numero = this.facturaCompleta.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("ImporteTotal", ex), "Propiedad fallida ImporteTotal");
			}
		}
		#endregion
		
		#region Validaciones de tipos de productos
		[Test]
		public void ItemsNotaCredito_NoExistenEn_ItemsFacturaAsociada()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito con un importe inferior a la factura
			// asociada
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo;
			this.notaCreditoCompleta.Numero = this.facturaCompleta.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		
		[Test]
		public void ItemsNotaCredito_NoExistenEn_ItemsFacturaAsociada_ProductosAbstractos()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			// Agrego los productos abstractos con un precio unitario determinado
			ItemFactura itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 2);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " PA 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 7);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " PA 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			
			ItemFactura itemNotaCreditoAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 1);
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " PA 1";
			itemNotaCreditoAbstracto.PrecioUnitario = 500.0m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			itemNotaCreditoAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 4);
			// Si bien se trata del mismo producto abstracto, la descirpción difiere, y por lo tanto físicamente
			// se trata de otro producto. Entonces este item es inválido, porque no existe en la factura
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " PA OTRO 2";
			itemNotaCreditoAbstracto.PrecioUnitario = 275.90m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		#endregion
		
		#region Validaciones de cantidad de productos con varias notas de crédito
		[Test]
		public void VariasNotasCredito_AsociadasA_Una_Factura_Cantidades_E_ImportesCorrectos()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la NotaCredito
			NotaCredito notaCredito = TestBase.CrearNotaCredito();
			notaCredito.Tipo = this.facturaCompleta.Tipo;
			notaCredito.Codigo = this.facturaCompleta.Codigo;
			notaCredito.Numero = this.facturaCompleta.Numero + 1;
			notaCredito.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			notaCredito.Cliente = this.ClienteSinBonificacion;
			notaCredito.Items.Clear();
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(notaCredito);
			
			try {
				notaCredito.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la primer nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			/* Creo ahora una PreNotaCredito con las cantidades correctas en los productos */
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = notaCredito.Codigo;
			this.notaCreditoCompleta.Numero = notaCredito.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la segunda nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void VariasNotasCredito_AsociadasA_Una_Factura_CantidadesIncorrectas()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la NotaCredito
			NotaCredito notaCredito = TestBase.CrearNotaCredito();
			notaCredito.Tipo = this.facturaCompleta.Tipo;
			notaCredito.Codigo = this.facturaCompleta.Codigo;
			notaCredito.Numero = this.facturaCompleta.Numero + 1;
			notaCredito.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			notaCredito.Cliente = this.ClienteSinBonificacion;
			notaCredito.Items.Clear();
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(notaCredito);
			
			try {
				notaCredito.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la primer nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			/* Creo ahora una PreNotaCredito con una cantidad incorrecta de productos en el item */
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = notaCredito.Codigo;
			this.notaCreditoCompleta.Numero = notaCredito.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la PreNotaCredito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		
		[Test]
		public void VariasNotasCredito_AsociadasA_Una_Factura_ConProductosAbstractos_Cantidades_E_ImportesCorrectos()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			
			// Agrego los productos abstractos con un precio unitario determinado
			ItemFactura itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 2);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 7);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la NotaCredito
			NotaCredito notaCredito = TestBase.CrearNotaCredito();
			notaCredito.Tipo = this.facturaCompleta.Tipo;
			notaCredito.Codigo = this.facturaCompleta.Codigo;
			notaCredito.Numero = this.facturaCompleta.Numero + 1;
			notaCredito.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			notaCredito.Cliente = this.ClienteSinBonificacion;
			notaCredito.Items.Clear();
			
			// Agrego solo algunos de los productos abstractos de la factura
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 1);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			notaCredito.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 4);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			notaCredito.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(notaCredito);
			
			try {
				notaCredito.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la primer nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			// Creo la segunda nota de crédito, que también va a tener un producto grifería, pero
			// sin superar la cantidad que hay en la factura
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = notaCredito.Codigo;
			this.notaCreditoCompleta.Numero = notaCredito.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			
			// Agrego y completo todos los productos abstractos de la factura
			ItemFactura itemNotaCreditoAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 1);
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " MOD 1";
			itemNotaCreditoAbstracto.PrecioUnitario = 500.0m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			itemNotaCreditoAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 3);
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " MOD 2";
			itemNotaCreditoAbstracto.PrecioUnitario = 275.90m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la segunda nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void VariasNotasCredito_AsociadasA_Una_Factura_ConProductosAbstractos_Cantidades_E_ImportesIncorrectos()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			
			// Agrego los productos abstractos con un precio unitario determinado
			ItemFactura itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 2);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 7);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la NotaCredito
			NotaCredito notaCredito = TestBase.CrearNotaCredito();
			notaCredito.Tipo = this.facturaCompleta.Tipo;
			notaCredito.Codigo = this.facturaCompleta.Codigo;
			notaCredito.Numero = this.facturaCompleta.Numero + 1;
			notaCredito.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			notaCredito.Cliente = this.ClienteSinBonificacion;
			notaCredito.Items.Clear();
			
			// Agrego solo algunos de los productos abstractos de la factura
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 1);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			notaCredito.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 4);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			notaCredito.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(notaCredito);
			
			try {
				notaCredito.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la primer nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			// Creo la segunda nota de crédito, que también va a tener un producto grifería, pero
			// sin superar la cantidad que hay en la factura
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = notaCredito.Codigo;
			this.notaCreditoCompleta.Numero = notaCredito.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			
			// Agrego y completo todos los productos abstractos de la factura. Uno de los
			// productos abstractos tiene una cantidad superior a la suma de la nota de crédito
			// con la factura asociada
			ItemFactura itemNotaCreditoAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 2);
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " MOD 1";
			itemNotaCreditoAbstracto.PrecioUnitario = 500.0m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			itemNotaCreditoAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 3);
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " MOD 2";
			itemNotaCreditoAbstracto.PrecioUnitario = 275.90m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de credito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		#endregion
		
		#region Validaciones de cantidad de productos con una sola nota de crédito
		[Test]
		public void NotasCredito_CantidadDeUnProductoIncorrecta()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 10));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 20));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			/* Ahora creo la PreNotaCredito, que va a tener 8 productos mosaico más, pero que
			 * en total con la nota de crédito anterior suman 16, y esto es INCORRECTO */
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 12));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la PreNotaCredito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		
		[Test]
		public void NotasCredito_CantidadDeUnProductoIncorrecta_ConProductoAbstracto()
		{
			// Borro todas las facturas y notas de crédito
			ActiveRecordValidationBase<Factura>.DeleteAll();
			ActiveRecordValidationBase<NotaCredito>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			// Agrego los productos abstractos con un precio unitario determinado
			ItemFactura itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 2);
			itemAbstracto.PrecioUnitario = 500.0m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			/* Ahora creo la PreNotaCredito, que va a tener 8 productos mosaico más, pero que
			 * en total con la nota de crédito anterior suman 16, y esto es INCORRECTO */
			this.notaCreditoCompleta = TestBase.CrearNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.Codigo = this.facturaCompleta.Codigo;
			this.notaCreditoCompleta.Numero = this.facturaCompleta.Numero + 1;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			
			// Agrego el item pero con una cantidad que, sumándola a la nota de crédito anterior,
			// supera a la factura asociada.
			ItemFactura itemNotaCreditoAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 3);
			itemNotaCreditoAbstracto.PrecioUnitario = 500.0m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la PreNotaCredito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		#endregion
		#endregion
	}
}
