// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections.Generic;

using NUnit.Framework;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using NHibernate.Criterion;

using Stratos.Common;
using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture()]
	public class PreFacturaTests : DocumentoGenericoTestBase
	{
		#region Campos
		private PreFactura facturaCompleta;
		private PreFactura facturaVacia;
		#endregion

		#region Métodos de inicialización y finalización de la prueba
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			// Creo la factura con los datos completos para realizar
			// las pruebas de persistencia
			this.facturaCompleta = TestBase.CrearPreFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);

			this.facturaVacia = new PreFactura();
		}
		#endregion

		#region Pruebas de persistencia de datos
		[Test()]
		public void Creacion()
		{
			try {
				this.facturaCompleta.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}

			base.FlushAndRecreateScope();
			
			PreFactura[] facturas = PreFactura.FindAll();
			
			Assert.AreEqual(1, facturas.Length, "Cantidad de facturas insertadas");
			
			PreFactura fr = facturas[0];
			Assert.AreEqual(this.ClienteSinBonificacion.Codigo, fr.Cliente.Codigo, "Id cliente");
			
			Assert.AreEqual(1, fr.Items.Count, "Cantidad de items");
			Assert.AreEqual(this.ProductoGriferia.Descripcion, fr.Items[0].Descripcion, "Item Descripcion");
			Assert.AreEqual(1, fr.Items[0].Cantidad, "Item Cantidad");
			Assert.AreEqual(0.00m, fr.Items[0].PorcentajeBonificacion, "Item Porcentaje bonificación");
			Assert.AreEqual(60.00, fr.Items[0].PrecioUnitario, "Item Precio unitario");
			Assert.AreEqual(60.00, fr.Items[0].Importe, "Item Importe");
			
			Assert.AreEqual(60.00m, fr.ImporteBruto, "Importe bruto");
			Assert.AreEqual(0.00m, fr.PorcentajeBonificacion, "Porcentaje bonificación");
			Assert.AreEqual(0.00m, fr.ImporteBonificacion, "Importe bonificación");
			Assert.AreEqual(60.00m, fr.ImporteNeto, "Importe Neto");
			Assert.AreEqual(12.60m, fr.ImporteIva, "Importe iva");
			Assert.AreEqual(72.60m, fr.ImporteTotal, "Importe total");
			Assert.AreEqual(72.60m, fr.ImporteFinal, "Importe final");
			Assert.AreEqual("Estas son las observaciones", fr.Observaciones, "Observaciones");
			Assert.AreEqual(FacturaTipoVenta.FacturaA, fr.Tipo, "Tipo factura");
			Assert.AreEqual(DateTime.Today, fr.Fecha.Date, "Fecha");
			Assert.AreEqual(CondicionPago.Contado, fr.CondicionVenta.Condicion,
			                "Condicion de venta");
		}
		
		[Test()]
		public void ActualizacionDeItems()
		{
			// Primero creo la factura
			try {
				this.facturaCompleta.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}

			base.FlushAndRecreateScope();
			
			// Luego intento modificarla y persistir las modificaciones
			PreFactura facturaCargada = PreFactura.FindAll()[0];
			
			facturaCargada.Items.Clear();
			facturaCargada.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 7));
			facturaCargada.Items.Add(TestBase.CrearItemPreFactura(this.ProductoMosaico, 10));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaCargada);
			
			// Importes
//			facturaCargada.ImporteBruto = 633.0m;
//			facturaCargada.PorcentajeBonificacion = 0.00m;
//			facturaCargada.ImporteBonificacion = 0.00m;
//			facturaCargada.ImporteNeto = 633.0m;
//			facturaCargada.ImporteIva = 132.93m;
//			facturaCargada.ImporteTotal = 765.93m;
			
			try {
				facturaCargada.Update();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			Assert.AreEqual(2, facturaCargada.Items.Count, "Cantidad de items");
			foreach (ItemPreFactura item in facturaCargada.Items) {
				if (item.ProductoCodigo == this.ProductoGriferia.Codigo)
					Assert.AreEqual(7, item.Cantidad, "Cantidad item producto 1");
				else
					Assert.AreEqual(10, item.Cantidad, "Cantidad item producto 2");
			}
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinDatos()
		{
			this.facturaVacia.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinFecha()
		{
			this.facturaCompleta.Fecha = this.facturaVacia.Fecha;
			this.facturaCompleta.Create();
		}

		[Test()]
		public void SinTipo()
		{
			// Por defecto tiene asignado un tipo, por eso este test pasa.
			this.facturaCompleta.Tipo = this.facturaVacia.Tipo;
			this.facturaCompleta.Create();
		}

		[Test()]
		public void SinCondicionVenta()
		{
			// La condición de venta por defecto es contado
			this.facturaCompleta.CondicionVenta = this.facturaVacia.CondicionVenta;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinCliente()
		{
			// TODO: Hacer un tests que verifique que el cliente exista.
			this.facturaCompleta.Cliente = this.facturaVacia.Cliente;
			this.facturaCompleta.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinItems()
		{
			this.facturaCompleta.Items = this.facturaVacia.Items;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteBruto()
		{
			this.facturaCompleta.ImporteBruto = this.facturaVacia.ImporteBruto;
			this.facturaCompleta.Create();
		}

		[Test()]
		public void SinImporteBonificacion()
		{
			this.facturaCompleta.ImporteBonificacion =
				this.facturaVacia.ImporteBonificacion;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteNeto()
		{
			this.facturaCompleta.ImporteNeto = this.facturaVacia.ImporteNeto;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteIva()
		{
			// Como facturaCompleta es una FacturaA, debe tener un importe de iva.
			this.facturaCompleta.ImporteIva = this.facturaVacia.ImporteIva;
			this.facturaCompleta.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteTotal()
		{
			this.facturaCompleta.ImporteTotal = this.facturaVacia.ImporteTotal;
			this.facturaCompleta.Create();
		}

		[Test()]
		public void SinObservaciones()
		{
			this.facturaCompleta.Observaciones = this.facturaVacia.Observaciones;
			this.facturaCompleta.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void CodigoCero()
		{
			this.facturaCompleta.Codigo = 0;
			this.facturaCompleta.Create();
		}
		#endregion

		#region Pruebas de alto nivel (código repetido, bonificación excesiva, etc)
		[Test()]
		public void DosFacturas()
		{
			// Creo la factura completa
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			// Completo la factura vacía con la misma clave
			PreFactura facturaRepetida = TestBase.CrearPreFactura();
			facturaRepetida.Tipo = this.facturaCompleta.Tipo;
			facturaRepetida.Cliente = this.facturaCompleta.Cliente;
			
			facturaRepetida.Items.Clear();
			facturaRepetida.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaRepetida);
			
			try {
				facturaRepetida.Create();
			}
			catch (ActiveRecordException e) {
				Assert.Fail("No se pudo crear la factura: " + e.Message);
			}
		}
		
		[Test()]
		public void FacturaA_ItemsPrecioUnitarioCorrecto()
		{
			// Modifico el IVA del cliente sin bonificación
			this.ClienteSinBonificacion.TipoResponsable = TipoResponsable.ResponsableInscripto;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			// Al realizar el cálculo siguiente, se coloca automáticamente el precio unitario
			// sin IVA.
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Verifico que los items tengan el precio unitario correcto
				item1 = ItemPreFactura.Find(item1.Id);
				Assert.AreEqual(60.0m, item1.PrecioUnitario, "Item 1 Precio unitario");
				
				item2 = ItemPreFactura.Find(item2.Id);
				Assert.AreEqual(21.3m, item2.PrecioUnitario, "Item 2 Precio unitario");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaA_ItemsPrecioUnitarioIncorrecto()
		{
			// Modifico el IVA del cliente sin bonificación
			this.ClienteSinBonificacion.TipoResponsable = TipoResponsable.ResponsableInscripto;
			this.ClienteSinBonificacion.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			// Al realizar el cálculo siguiente, se coloca automáticamente el precio unitario
			// sin IVA. Modifico los items para que tengan un precio unitario incorrecto
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			item1.PrecioUnitario = item1.PrecioUnitario + 1.00m;
			item2.PrecioUnitario = item2.PrecioUnitario + 1.00m;
			
			this.facturaCompleta.ImporteBruto += 2.00m + 125.3m;
			this.facturaCompleta.ImporteNeto += 2.00m + 125.3m;
			this.facturaCompleta.ImporteIva = StratosUtil.RoundDecimal(this.facturaCompleta.ImporteNeto * 0.21m, 2);
			this.facturaCompleta.ImporteTotal = this.facturaCompleta.ImporteNeto + this.facturaCompleta.ImporteIva;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No se debió haber podido crear la factura");
			}
			catch (ActiveRecordValidationException e) {
				// Me aseguro que la propiedad fallida sea la que espero
				Assert.IsTrue(this.PropiedadFallida("Items", e), "Propiedad correcta");
			}
		}
		
		[Test()]
		public void FacturaB_ItemsPrecioUnitarioCorrecto()
		{
			// Modifico el IVA del cliente sin bonificación
			this.ClienteSinBonificacion.TipoResponsable = TipoResponsable.ConsumidorFinal;
			this.ClienteSinBonificacion.Cuit = new CUIT();
			this.ClienteSinBonificacion.Update();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			// Al realizar el cálculo siguiente, se coloca automáticamente el precio unitario
			// sin IVA.
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Verifico que los items tengan el precio unitario correcto
				item1 = ItemPreFactura.Find(item1.Id);
				Assert.AreEqual(StratosUtil.RoundDecimal(60.0m * 1.21m, 2), item1.PrecioUnitario, "Item 1 Precio unitario");
				
				item2 = ItemPreFactura.Find(item2.Id);
				Assert.AreEqual(StratosUtil.RoundDecimal(21.3m * 1.21m, 2), item2.PrecioUnitario, "Item 2 Precio unitario");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaB_ItemsPrecioUnitarioIncorrecto()
		{
			// Modifico el IVA del cliente sin bonificación
			this.ClienteSinBonificacion.TipoResponsable = TipoResponsable.ConsumidorFinal;
			this.ClienteSinBonificacion.Cuit = new CUIT();
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			
			// Al realizar el cálculo siguiente, se coloca automáticamente el precio unitario
			// sin IVA. Modifico los items para que tengan un precio unitario incorrecto
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			item1.PrecioUnitario = item1.PrecioUnitario + 1.00m;
			item2.PrecioUnitario = item2.PrecioUnitario + 1.00m;
			
			this.facturaCompleta.ImporteBruto += 2.00m + 125.3m;
			this.facturaCompleta.ImporteNeto += 2.00m + 125.3m;
			this.facturaCompleta.ImporteIva = 0.00m;
			this.facturaCompleta.ImporteTotal = this.facturaCompleta.ImporteNeto;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No se debió haber podido crear la factura");
			}
			catch (ActiveRecordValidationException e) {
				// Me aseguro que la propiedad fallida sea la que espero
				Assert.IsTrue(this.PropiedadFallida("Items", e), "Propiedad correcta");
			}
		}
		
		[Test()]
		public void FechaMinimaCorrecta()
		{
			/* Necesito crear una Factura, ya que la fecha de la PreFactura se valida
			 * con respecto a las Facturas. */
			Factura factura = TestBase.CrearFactura();
			factura.Codigo = 4;
			factura.Cliente = this.ClienteSinBonificacion;
			factura.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			factura.Tipo = FacturaTipoVenta.FacturaA;
			factura.Items.Clear();
			factura.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(factura);
			
			factura.Create();
			
			this.FlushAndRecreateScope();
			
			PreFactura facturaNueva = TestBase.CrearPreFactura();
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void FechaMinimaCorrecta_FacturaAnuladaNoSeTieneEnCuenta()
		{
			// Creo una factura normal
			Factura factura = TestBase.CrearFactura();
			factura.Codigo = 4;
			factura.Cliente = this.ClienteSinBonificacion;
			factura.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			factura.Tipo = FacturaTipoVenta.FacturaA;
			factura.Items.Clear();
			factura.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(factura);
			
			factura.Create();
			
			this.FlushAndRecreateScope();
			
			/* Ahora creo una factura anulada con una fecha posterior a la factura
			 * anterior. Esta factura no se debería tener en cuenta cuando se valida
			 * la fecha de facturas posteriores. */
			Factura facturaNueva1 = TestBase.CrearFactura();
			facturaNueva1.Anulada = true;
			facturaNueva1.Cliente = this.ClienteSinBonificacion;
			facturaNueva1.Fecha = DateTime.Today - TimeSpan.FromDays(3);
			facturaNueva1.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva1.Codigo = factura.Codigo;
			facturaNueva1.Numero = factura.Numero + 1;
			
			facturaNueva1.Items.Clear();
			facturaNueva1.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva1);
			
			facturaNueva1.Create();
			
			this.FlushAndRecreateScope();
			
			PreFactura facturaNueva = TestBase.CrearPreFactura();
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);
			
			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}

		[Test()]
		public void FechaMinimaIncorrecta()
		{
			/* Necesito crear una Factura, ya que la fecha de la PreFactura se valida
			 * con respecto a las Facturas. */
			Factura factura = TestBase.CrearFactura();
			factura.Cliente = this.ClienteSinBonificacion;
			factura.Fecha = DateTime.Today - TimeSpan.FromDays(4);
			factura.Tipo = FacturaTipoVenta.FacturaA;
			factura.Items.Clear();
			factura.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(factura);
			
			factura.Create();
			
			this.FlushAndRecreateScope();
			
			PreFactura facturaNueva = TestBase.CrearPreFactura();
			facturaNueva.Fecha = DateTime.Today - TimeSpan.FromDays(5);
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Fecha", ex),
				              "Propiedad Fecha errónea");
			}
		}

		[Test()]
		public void FechaMaximaCorrecta()
		{
			/* Necesito crear una Factura, ya que la fecha de la PreFactura se valida
			 * con respecto a las Facturas. */
			Factura factura = TestBase.CrearFactura();
			factura.Cliente = this.ClienteSinBonificacion;
			factura.Fecha = DateTime.Today;
			factura.Tipo = FacturaTipoVenta.FacturaA;
			factura.Items.Clear();
			factura.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(factura);
			
			try {
				factura.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			PreFactura facturaNueva = TestBase.CrearPreFactura();
			facturaNueva.Fecha = DateTime.Today;
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void FechaMaximaIncorrecta()
		{
			/* Necesito crear una Factura, ya que la fecha de la PreFactura se valida
			 * con respecto a las Facturas. */
			Factura factura = TestBase.CrearFactura();
			factura.Cliente = this.ClienteSinBonificacion;
			factura.Fecha = DateTime.Today;
			factura.Tipo = FacturaTipoVenta.FacturaA;
			factura.Items.Clear();
			factura.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(factura);
			
			factura.Create();
			
			this.FlushAndRecreateScope();
			
			PreFactura facturaNueva = TestBase.CrearPreFactura();
			facturaNueva.Fecha = DateTime.Today + TimeSpan.FromDays(1);
			facturaNueva.Tipo = FacturaTipoVenta.FacturaA;
			facturaNueva.Cliente = this.facturaCompleta.Cliente;
			
			facturaNueva.Items.Clear();
			facturaNueva.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(facturaNueva);

			try {
				facturaNueva.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Fecha", ex),
				              "Propiedad Fecha errónea");
			}
		}
		
		[Test()]
		public void ImportesValidosConBonificacion()
		{
			// Modifico el IVA del cliente sin bonificación
			this.ClienteConBonificacion.TipoResponsable = TipoResponsable.ResponsableInscripto;
			this.ClienteConBonificacion.Update();
			
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido que los importes sean válidos
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(278.89m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2510.00m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(527.10m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3037.10m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void ImportesValidosSinBonificacion()
		{
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido los importes
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(585.67m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3374.56m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}

		[Test()]
		public void ImporteBrutoInvalido()
		{
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe bruto a un valor incorrecto
			this.facturaCompleta.ImporteBruto = 1133.23m; // Incorrecto
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteBruto", ex),
				              "Propiedad ImporteBruto errónea");
			}
		}

		[Test()]
		public void ImporteBonificacionInvalido()
		{
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe de bonificación a un valor incorrecto
			this.facturaCompleta.ImporteBonificacion = 345.00m; // Incorrecto
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteBonificacion", ex),
				              "Propiedad ImporteBonificacion errónea");
			}
		}
		
		[Test()]
		public void ImporteNetoInvalido()
		{
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe neto a un valor incorrecto
			this.facturaCompleta.ImporteNeto = 2900.22m; // Incorrecto
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteNeto", ex),
				              "Propiedad ImporteNeto errónea");
			}
		}

		[Test()]
		public void ImporteIvaInvalido()
		{
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe de iva a un valor incorrecto
			this.facturaCompleta.ImporteIva = 523.14m; // Incorrecto
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteIva", ex),
				              "Propiedad ImporteIva errónea");
			}
		}

		[Test()]
		public void ImporteTotalInvalido()
		{
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe total a un valor incorrecto
			this.facturaCompleta.ImporteTotal = 3034.99m; // Incorrecto
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteTotal", ex),
				              "Propiedad ImporteTotal errónea");
			}
		}

		[Test()]
		public void ImportesValidosFacturaB()
		{
			this.ClienteConBonificacion.PorcentajeBonificacion = 10.0m;
			this.ClienteConBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido el item1
				item1 = ItemPreFactura.Find(item1.Id);
				
				Assert.AreEqual(0.00m, item1.PorcentajeBonificacion, "Item 1 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(60.0m * 1.21m, 2), item1.PrecioUnitario, "Item 1 Precio unitario");
				Assert.AreEqual(145.20m, item1.Importe, "Item 1 Importe");
				
				// Valido el item2
				item2 = ItemPreFactura.Find(item2.Id);
				
				Assert.AreEqual(0.00m, item2.PorcentajeBonificacion, "Item 2 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(21.3m * 1.21m, 2), item2.PrecioUnitario, "Item 2 Precio unitario");
				Assert.AreEqual(3228.98m, item2.Importe, "Item 2 Importe");
				
				// Valido los importes
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(337.42m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(3036.76m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3036.76m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void ImporteIvaInvalidoFacturaB()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe de iva a un valor inválido
			this.facturaCompleta.ImporteIva = 637.79m; // no puede tener valor en factura B
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteIva", ex),
				              "Propiedad ImporteIva errónea");
			}
		}

		[Test()]
		public void ImportesClienteSinBonificacion()
		{
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido los importes
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(585.67m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3374.56m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaAConProductoAbstracto()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Modifico los productos para que sean productos abstractos
			this.ProductoGriferia.ImporteSinIva = 0;
			this.ProductoGriferia.PrecioCosto = 0;
			this.ProductoGriferia.ProductoAbstracto = true;
			this.ProductoGriferia.Save();
			
			this.ProductoMosaico.ImporteSinIva = 0;
			this.ProductoMosaico.PrecioCosto = 0;
			this.ProductoMosaico.ProductoAbstracto = true;
			this.ProductoMosaico.Save();
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(ProductoGriferia, 1);
			item1.PrecioUnitario = 12;
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(ProductoMosaico, 125.3m);
			item2.PrecioUnitario = 23;
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido que los importes sean válidos
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(607.72m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3501.62m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaBConProductoAbstracto()
		{
			// Modifico el IVA del cliente sin bonificación
			this.ClienteSinBonificacion.TipoResponsable = TipoResponsable.ConsumidorFinal;
			this.ClienteSinBonificacion.Cuit = new CUIT();
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Modifico los productos para que sean productos abstractos
			this.ProductoGriferia.ImporteSinIva = 0;
			this.ProductoGriferia.PrecioCosto = 0;
			this.ProductoGriferia.ProductoAbstracto = true;
			this.ProductoGriferia.Save();
			
			this.ProductoMosaico.ImporteSinIva = 0;
			this.ProductoMosaico.PrecioCosto = 0;
			this.ProductoMosaico.ProductoAbstracto = true;
			this.ProductoMosaico.Save();
			
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(ProductoGriferia, 1);
			item1.PrecioUnitario = 12;
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(ProductoMosaico, 125.3m);
			item2.PrecioUnitario = 23;
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido que los importes sean válidos
				
				/* NOTA: Los importes son distintos al método FacturaAConProductoAbstracto porque
				 * el precio unitario de los items con producto abstracto se matiene igual (en este
				 * caso de Factura B se debería agregarle el IVA a los mismos para que de lo mismo). */
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(2893.90m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaAConClienteResponsableInscripto()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.ResponsableInscripto;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido los importes
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(585.67m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3374.56m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaAConClienteResponsableNoInscripto()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.ResponsableNoInscripto;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido los importes
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2788.89m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(585.67m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3374.56m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaAConClienteConsumidorFinal()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.Cuit = new CUIT();
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.ConsumidorFinal;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura A con Consumidor Final no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		
		[Test()]
		public void FacturaAConClienteExento()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.Exento;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura A con un Excento no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		
		[Test()]
		public void FacturaAConClienteMonotributo()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.Monotributo;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura A con Monotributo no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		
		[Test()]
		public void FacturaB_ConClienteConsumidorFinal()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.ConsumidorFinal;
			this.ClienteSinBonificacion.Cuit = new CUIT();
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido el item1
				item1 = ItemPreFactura.Find(item1.Id);
				
				Assert.AreEqual(0.00m, item1.PorcentajeBonificacion, "Item 1 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(60.0m * 1.21m, 2), item1.PrecioUnitario, "Item 1 Precio unitario");
				Assert.AreEqual(145.20m, item1.Importe, "Item 1 Importe");
				
				// Valido el item2
				item2 = ItemPreFactura.Find(item2.Id);
				
				Assert.AreEqual(0.00m, item2.PorcentajeBonificacion, "Item 2 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(21.3m * 1.21m, 2), item2.PrecioUnitario, "Item 2 Precio unitario");
				Assert.AreEqual(3228.98m, item2.Importe, "Item 2 Importe");
				
				// Valido los importes
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaB_ConClienteExento()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.Exento;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido el item1
				item1 = ItemPreFactura.Find(item1.Id);
				
				Assert.AreEqual(0.00m, item1.PorcentajeBonificacion, "Item 1 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(60.0m * 1.21m, 2), item1.PrecioUnitario, "Item 1 Precio unitario");
				Assert.AreEqual(145.20m, item1.Importe, "Item 1 Importe");
				
				// Valido el item2
				item2 = ItemPreFactura.Find(item2.Id);
				
				Assert.AreEqual(0.00m, item2.PorcentajeBonificacion, "Item 2 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(21.3m * 1.21m, 2), item2.PrecioUnitario, "Item 2 Precio unitario");
				Assert.AreEqual(3228.98m, item2.Importe, "Item 2 Importe");
				
				// Valido los importes
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaBConClienteMonotributo()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.Monotributo;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido el item1
				item1 = ItemPreFactura.Find(item1.Id);
				
				Assert.AreEqual(0.00m, item1.PorcentajeBonificacion, "Item 1 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(60.0m * 1.21m, 2), item1.PrecioUnitario, "Item 1 Precio unitario");
				Assert.AreEqual(145.20m, item1.Importe, "Item 1 Importe");
				
				// Valido el item2
				item2 = ItemPreFactura.Find(item2.Id);
				
				Assert.AreEqual(0.00m, item2.PorcentajeBonificacion, "Item 2 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(21.3m * 1.21m, 2), item2.PrecioUnitario, "Item 2 Precio unitario");
				Assert.AreEqual(3228.98m, item2.Importe, "Item 2 Importe");
				
				// Valido los importes
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3374.18m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaBConClienteResponsableInscripto()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.ResponsableInscripto;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura B con Responsable Inscripto no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		
		[Test()]
		public void FacturaBConClienteResponsableNoInscripto()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.ResponsableNoInscripto;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Los totales de la factura
//			this.facturaCompleta.ImporteBruto = 3374.56m;
//			this.facturaCompleta.ImporteBonificacion = 0.00m;
//			this.facturaCompleta.ImporteNeto = 3374.56m;
//			this.facturaCompleta.ImporteIva = 0.00m;
//			this.facturaCompleta.ImporteTotal = 3374.56m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Una factura B con Responsable No Inscripto no se puede crear");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex),
				              "Propiedad Cliente errónea");
			}
		}
		
		[Test()]
		public void FacturaB_ItemsConBonificacion()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.Monotributo;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			item1.PorcentajeBonificacion = 10.0m;
			
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido el item1
				item1 = ItemPreFactura.Find(item1.Id);
				
				Assert.AreEqual(10.0m, item1.PorcentajeBonificacion, "Item 1 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(60.0m * 1.21m, 2), item1.PrecioUnitario, "Item 1 Precio unitario");
				Assert.AreEqual(130.68m, item1.Importe, "Item 1 Importe");
				
				// Valido el item2
				item2 = ItemPreFactura.Find(item2.Id);
				
				Assert.AreEqual(0.00m, item2.PorcentajeBonificacion, "Item 2 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(21.3m * 1.21m, 2), item2.PrecioUnitario, "Item 2 Precio unitario");
				Assert.AreEqual(3228.98m, item2.Importe, "Item 2 Importe");
				
				// Valido los importes
				Assert.AreEqual(3359.66m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(3359.66m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3359.66m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void ItemsConImporte_UnCentavoMenorAlCorrecto()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.Monotributo;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe del item para que sea un centavo menor al correcto
			item1.Importe -= 0.01m;
			
			// Recalculo los importes
			this.facturaCompleta.ImporteBruto -= 0.01m;
			this.facturaCompleta.ImporteNeto -= 0.01m;
			this.facturaCompleta.ImporteTotal -= 0.01m;
			
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void ItemsConImporte_DosCentavosMenorAlCorrecto()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.Monotributo;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe del item para que sea un centavo menor al correcto
			item1.Importe -= 0.02m;
			
			// Recalculo los importes
			this.facturaCompleta.ImporteBruto -= 0.02m;
			this.facturaCompleta.ImporteNeto -= 0.02m;
			this.facturaCompleta.ImporteTotal -= 0.02m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Se pudo crear la factura con un importe erroneo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("Importe", e), "Propiedad fallida");
			}
		}
		
		[Test()]
		public void ItemsConImporte_UnCentavoMayorAlCorrecto()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.Monotributo;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe del item para que sea un centavo menor al correcto
			item1.Importe += 0.01m;
			
			// Recalculo los importes
			this.facturaCompleta.ImporteBruto += 0.01m;
			this.facturaCompleta.ImporteNeto += 0.01m;
			this.facturaCompleta.ImporteTotal += 0.01m;
			
			try {
				this.facturaCompleta.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void ItemsConImporte_DosCentavosMayorAlCorrecto()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.Monotributo;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			// Modifico el importe del item para que sea un centavo menor al correcto
			item1.Importe += 0.02m;
			
			// Recalculo los importes
			this.facturaCompleta.ImporteBruto += 0.02m;
			this.facturaCompleta.ImporteNeto += 0.02m;
			this.facturaCompleta.ImporteTotal += 0.02m;
			
			try {
				this.facturaCompleta.Create();
				Assert.Fail("Se pudo crear la factura con un importe erroneo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("Importe", e), "Propiedad fallida");
			}
		}
		
		[Test()]
		public void FacturaA_ItemsConBonificacion()
		{
			// Cliente Reponsable Inscripto
			this.ClienteSinBonificacion.TipoResponsable =
				TipoResponsable.ResponsableInscripto;
			this.ClienteSinBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteSinBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			item1.PorcentajeBonificacion = 10.0m;
			
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido el item1
				item1 = ItemPreFactura.Find(item1.Id);
				
				Assert.AreEqual(10.0m, item1.PorcentajeBonificacion, "Item 1 Porcentaje bonificacion");
				Assert.AreEqual(60.0m, item1.PrecioUnitario, "Item 1 Precio unitario");
				Assert.AreEqual(108.0m, item1.Importe, "Item 1 Importe");
				
				// Valido el item2
				item2 = ItemPreFactura.Find(item2.Id);
				
				Assert.AreEqual(0.00m, item2.PorcentajeBonificacion, "Item 2 Porcentaje bonificacion");
				Assert.AreEqual(21.3m, item2.PrecioUnitario, "Item 2 Precio unitario");
				Assert.AreEqual(2668.89m, item2.Importe, "Item 2 Importe");
				
				// Valido los importes
				Assert.AreEqual(2776.89m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(0.00m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2776.89m, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(583.15m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(3360.04m, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void FacturaA_ItemsConBonificacion_PorcentajeBonificacionRebuscado()
		{
			// Cliente Reponsable Inscripto
			this.ClienteConBonificacion.TipoResponsable =
				TipoResponsable.ResponsableInscripto;
			this.ClienteConBonificacion.PorcentajeBonificacion = 12.73m;
			this.ClienteConBonificacion.Update();
			
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteConBonificacion;
			this.facturaCompleta.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPreFactura item1 = TestBase.CrearItemPreFactura(this.ProductoGriferia, 2);
			item1.PorcentajeBonificacion = 7.78m;
			
			ItemPreFactura item2 = TestBase.CrearItemPreFactura(this.ProductoMosaico, 125.3m);
			item2.PorcentajeBonificacion = 12.91m;
			
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(item1);
			this.facturaCompleta.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
				
				// Valido el item1
				item1 = ItemPreFactura.Find(item1.Id);
				
				Assert.AreEqual(7.78m, item1.PorcentajeBonificacion, "Item 1 Porcentaje bonificacion");
				Assert.AreEqual(60.0m, item1.PrecioUnitario, "Item 1 Precio unitario");
				Assert.AreEqual(110.66m, item1.Importe, "Item 1 Importe");
				
				// Valido el item2
				item2 = ItemPreFactura.Find(item2.Id);
				
				Assert.AreEqual(12.91m, item2.PorcentajeBonificacion, "Item 2 Porcentaje bonificacion");
				Assert.AreEqual(21.3m, item2.PrecioUnitario, "Item 2 Precio unitario");
				Assert.AreEqual(2324.34m, item2.Importe, "Item 2 Importe");
				
				// Valido los importes
				Assert.AreEqual(2435.0m, this.facturaCompleta.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(309.98m, this.facturaCompleta.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2125.02, this.facturaCompleta.ImporteNeto, "Importe Neto");
				Assert.AreEqual(446.25m, this.facturaCompleta.ImporteIva, "Importe Iva");
				Assert.AreEqual(2571.27, this.facturaCompleta.ImporteTotal, "Importe Total");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		#region Items de producto cuando el producto cambia
		[Test]
		public void PreFacturaCreada_PrecioProductoAsociadoCambia_Pero_PreFacturaMantieneElPrecioViejo()
		{
			// Creo una prefactura
			PreFactura preFactura = TestBase.CrearPreFactura();
			preFactura.Tipo = FacturaTipoVenta.FacturaA;
			preFactura.Cliente = this.ClienteSinBonificacion;
			preFactura.Items.Clear();
			
			// Items
			ItemPreFactura itemGriferia = TestBase.CrearItemPreFactura(this.ProductoGriferia, 10);
			ItemPreFactura itemMosaico = TestBase.CrearItemPreFactura(this.ProductoMosaico, 7);
			
			preFactura.Items.Add(itemGriferia);
			preFactura.Items.Add(itemMosaico);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(preFactura);
			preFactura.Create();
			
			this.FlushAndRecreateScope();
			
			// Almaceno el precio del producto sin IVA
			decimal precioSinIvaProductoGriferia = this.ProductoGriferia.ImporteSinIva;
			decimal precioUnitarioProductoGriferia = this.ProductoGriferia.ImporteSinIva;
			
			decimal importeTotalPreFactura = preFactura.ImporteTotal;
			
			// Ahora modifico el precio del producto
			Producto producto = this.ProductoGriferia;
			producto.ImporteSinIva += 10.0m;
			producto.Update();
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Ahora verifico que la PreFactura siga siendo válida y que los importes
			// no se hayan modificado
			PreFactura preFacturaLeido = PreFactura.Find(preFactura.Id);
			
			Assert.IsTrue(preFacturaLeido.IsValid(), "PreFactura válida");
			
			// Busco el item correspondiente al producto modificado
			ItemPreFactura itemPreFacturaProductoModificado =
				(from item in preFacturaLeido.Items
				 where item.ProductoId == this.ProductoGriferia.Id
				 select item).First();
			
			Assert.AreEqual(itemGriferia.PrecioUnitario, itemPreFacturaProductoModificado.PrecioUnitario, "PreFactura precio unitario");
			Assert.AreEqual(itemGriferia.Cantidad, itemPreFacturaProductoModificado.Cantidad, "PreFactura cantidad");
			Assert.AreEqual(itemGriferia.Importe, itemPreFacturaProductoModificado.Importe, "PreFactura importe");
		}
		#endregion
		#endregion
	}
}
