// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

using NUnit.Framework;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using NHibernate.Criterion;

using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture()]
	public class PreNotaCreditoTests : DocumentoGenericoTestBase
	{
		#region Campos
		private Factura facturaCompleta;
		
		private PreNotaCredito notaCreditoCompleta;
		private PreNotaCredito notaCreditoVacia;
		#endregion

		#region Métodos de inicialización y finalización de la prueba
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Codigo = 4;
			this.facturaCompleta.Numero = 1;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.facturaCompleta = Factura.Find(this.facturaCompleta.Id);
			
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.FacturaAsociada = this.facturaCompleta;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);

			this.notaCreditoVacia = new PreNotaCredito();
		}
		#endregion

		#region Pruebas de persistencia de datos
		[Test()]
		public void Creacion()
		{
			try {
				this.notaCreditoCompleta.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}

			base.FlushAndRecreateScope();
			
			PreNotaCredito[] notasCredito = PreNotaCredito.FindAll();
			
			Assert.AreEqual(1, notasCredito.Length, "Cantidad de notas de crédito insertadas");
			
			PreNotaCredito fr = notasCredito[0];
			Assert.IsNotNull(fr.FacturaAsociada, "Factura asociada no nula");
			Assert.AreEqual(this.facturaCompleta.Id, fr.FacturaAsociada.Id, "Id factura asociada");
			
			Assert.AreEqual(this.ClienteSinBonificacion.Codigo, fr.Cliente.Codigo, "Id cliente");
			
			Assert.AreEqual(1, fr.Items.Count, "Cantidad de items");
			Assert.AreEqual(this.ProductoGriferia.Descripcion, fr.Items[0].Descripcion, "Item Descripcion");
			Assert.AreEqual(1, fr.Items[0].Cantidad, "Item Cantidad");
			Assert.AreEqual(0.00m, fr.Items[0].PorcentajeBonificacion, "Item Porcentaje bonificación");
			Assert.AreEqual(60.00, fr.Items[0].PrecioUnitario, "Item Precio unitario");
			Assert.AreEqual(60.00, fr.Items[0].Importe, "Item Importe");
			
			Assert.AreEqual(60.00m, fr.ImporteBruto, "Importe bruto");
			Assert.AreEqual(0.00m, fr.PorcentajeBonificacion, "Porcentaje bonificación");
			Assert.AreEqual(0.00m, fr.ImporteBonificacion, "Importe bonificación");
			Assert.AreEqual(60.00m, fr.ImporteNeto, "Importe Neto");
			Assert.AreEqual(12.60m, fr.ImporteIva, "Importe iva");
			Assert.AreEqual(72.60m, fr.ImporteTotal, "Importe total");
			Assert.AreEqual(72.60m, fr.ImporteFinal, "Importe final");
			Assert.AreEqual("Estas son las observaciones", fr.Observaciones, "Observaciones");
			Assert.AreEqual(FacturaTipoVenta.FacturaA, fr.Tipo, "Tipo factura");
			Assert.AreEqual(DateTime.Today, fr.Fecha.Date, "Fecha");
			Assert.AreEqual(CondicionPago.Contado, fr.CondicionVenta.Condicion,
			                "Condicion de venta");
		}
		
		[Test]
		public void SinFacturaAsociada()
		{
			this.notaCreditoCompleta.FacturaAsociada = this.notaCreditoVacia.FacturaAsociada;
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("No se debió haber podido crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("FacturaAsociada", ex));
			}
		}
		#endregion
		
		#region Pruebas de alto nivel (código repetido, bonificación excesiva, etc)
		#region Validaciones sobre la Factura asociada
		[Test]
		public void ConFacturaAsociada_FacturaAnulada()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a crear la factura pero como anulada
			this.MarcarComoNoPersistido(ref this.facturaCompleta);
			
			this.facturaCompleta.Anulada = true;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("FacturaAsociada", ex));
			}
		}
		
		[Test]
		public void ConFacturaAsociada_FacturaInvalida()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a crear la factura pero como anulada
			this.MarcarComoNoPersistido(ref this.facturaCompleta);
			
			this.facturaCompleta.Valida = false;
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("FacturaAsociada", ex));
			}
		}
		#endregion
		
		#region Validaciones sobre el cliente
		[Test]
		public void ClienteNotaCredito_IgualA_ClienteFacturaAsociada()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a crear la factura pero como anulada
			this.MarcarComoNoPersistido(ref this.facturaCompleta);
			
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex), "No se pudo crear la nota de crédito");
			}
		}
		
		[Test]
		public void ClienteNotaCredito_DistintoA_ClienteFacturaAsociada()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a crear la factura pero como anulada
			this.MarcarComoNoPersistido(ref this.facturaCompleta);
			
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			
			try {
				this.facturaCompleta.Create();
			} catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteConBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex));
			}
		}
		#endregion
		
		#region Validaciones de tipos (Factura A con Nota Credito B, etc...)
		[Test]
		public void FacturaAsociadaTipoA_NotaCreditoTipoA()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Update();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void FacturaAsociadaTipoB_NotaCreditoTipoB()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion2;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Update();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion2;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void FacturaAsociadaTipoA_NotaCreditoTipoB()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Update();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion2;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Tipo", ex), "Tipo - Propiedad fallida");
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex), "Cliente - Propiedad fallida");
			}
		}
		
		[Test]
		public void FacturaAsociadaTipoB_NotaCreditoTipoA()
		{
			this.facturaCompleta.Tipo = FacturaTipoVenta.FacturaB;
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion2;
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Update();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta.Tipo = FacturaTipoVenta.FacturaA;
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Tipo", ex), "Tipo - Propiedad fallida");
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex), "Cliente - Propiedad fallida");
			}
		}
		#endregion
		
		#region Validaciones sobre los importes
		[Test]
		public void ImporteNotaCredito_MenorA_ImporteFacturaAsociada()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito con un importe inferior a la factura
			// asociada
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void ImporteNotaCredito_IgualA_ImporteFacturaAsociada()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito con un importe inferior a la factura
			// asociada
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoMosaico, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void ImporteNotaCredito_MayorA_ImporteFacturaAsociada()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito con un importe inferior a la factura
			// asociada
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("ImporteTotal", ex), "Propiedad fallida ImporteTotal");
			}
		}
		#endregion
		
		#region Validaciones de tipos de productos
		[Test]
		public void ItemsNotaCredito_NoExistenEn_ItemsFacturaAsociada()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito con un importe inferior a la factura
			// asociada
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		
		[Test]
		public void ItemsNotaCredito_NoExistenEn_ItemsFacturaAsociada_ProductosAbstractos()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			// Agrego los productos abstractos con un precio unitario determinado
			ItemFactura itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 2);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " PA 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 7);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " PA 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la nota de crédito
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			
			ItemPreFactura itemPreNotaCreditoAbstracto = TestBase.CrearItemPreFactura(this.ProductoAbstracto, 1);
			itemPreNotaCreditoAbstracto.Descripcion = itemPreNotaCreditoAbstracto.ProductoDescripcion + " PA 1";
			itemPreNotaCreditoAbstracto.PrecioUnitario = 500.0m;
			this.notaCreditoCompleta.Items.Add(itemPreNotaCreditoAbstracto);
			
			itemPreNotaCreditoAbstracto = TestBase.CrearItemPreFactura(this.ProductoAbstracto, 4);
			// Si bien se trata del mismo producto abstracto, la descirpción difiere, y por lo tanto físicamente
			// se trata de otro producto. Entonces este item es inválido, porque no existe en la factura
			itemPreNotaCreditoAbstracto.Descripcion = itemPreNotaCreditoAbstracto.ProductoDescripcion + " PA OTRO 2";
			itemPreNotaCreditoAbstracto.PrecioUnitario = 275.90m;
			this.notaCreditoCompleta.Items.Add(itemPreNotaCreditoAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de crédito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		#endregion
		
		#region Validaciones de cantidad de productos con varias notas de crédito
		[Test]
		public void VariasNotasCredito_AsociadasA_Una_Factura_Cantidades_E_ImportesCorrectos()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la NotaCredito
			NotaCredito notaCredito = TestBase.CrearNotaCredito();
			notaCredito.Tipo = this.facturaCompleta.Tipo;
			notaCredito.Codigo = this.facturaCompleta.Codigo;
			notaCredito.Numero = this.facturaCompleta.Numero + 1;
			notaCredito.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			notaCredito.Cliente = this.ClienteSinBonificacion;
			notaCredito.Items.Clear();
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(notaCredito);
			
			try {
				notaCredito.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la primer nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			/* Creo ahora una PreNotaCredito con las cantidades correctas en los productos */
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la segunda nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void VariasNotasCredito_AsociadasA_Una_Factura_CantidadesIncorrectas()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la NotaCredito
			NotaCredito notaCredito = TestBase.CrearNotaCredito();
			notaCredito.Tipo = this.facturaCompleta.Tipo;
			notaCredito.Codigo = this.facturaCompleta.Codigo;
			notaCredito.Numero = this.facturaCompleta.Numero + 1;
			notaCredito.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			notaCredito.Cliente = this.ClienteSinBonificacion;
			notaCredito.Items.Clear();
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(notaCredito);
			
			try {
				notaCredito.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la primer nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			/* Creo ahora una PreNotaCredito con una cantidad incorrecta de productos en el item */
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la PreNotaCredito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		
		[Test]
		public void VariasNotasCredito_AsociadasA_Una_Factura_ConProductosAbstractos_Cantidades_E_ImportesCorrectos()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			
			// Agrego los productos abstractos con un precio unitario determinado
			ItemFactura itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 2);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 7);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la NotaCredito
			NotaCredito notaCredito = TestBase.CrearNotaCredito();
			notaCredito.Tipo = this.facturaCompleta.Tipo;
			notaCredito.Codigo = this.facturaCompleta.Codigo;
			notaCredito.Numero = this.facturaCompleta.Numero + 1;
			notaCredito.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			notaCredito.Cliente = this.ClienteSinBonificacion;
			notaCredito.Items.Clear();
			
			// Agrego solo algunos de los productos abstractos de la factura
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 1);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			notaCredito.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 4);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			notaCredito.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(notaCredito);
			
			try {
				notaCredito.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la primer nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			// Creo la segunda nota de crédito, que también va a tener un producto grifería, pero
			// sin superar la cantidad que hay en la factura
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			
			// Agrego y completo todos los productos abstractos de la factura
			ItemPreFactura itemNotaCreditoAbstracto = TestBase.CrearItemPreFactura(this.ProductoAbstracto, 1);
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " MOD 1";
			itemNotaCreditoAbstracto.PrecioUnitario = 500.0m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			itemNotaCreditoAbstracto = TestBase.CrearItemPreFactura(this.ProductoAbstracto, 3);
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " MOD 2";
			itemNotaCreditoAbstracto.PrecioUnitario = 275.90m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la segunda nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
		}
		
		[Test]
		public void VariasNotasCredito_AsociadasA_Una_Factura_ConProductosAbstractos_Cantidades_E_ImportesIncorrectos()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			
			// Agrego los productos abstractos con un precio unitario determinado
			ItemFactura itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 2);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 7);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la NotaCredito
			NotaCredito notaCredito = TestBase.CrearNotaCredito();
			notaCredito.Tipo = this.facturaCompleta.Tipo;
			notaCredito.Codigo = this.facturaCompleta.Codigo;
			notaCredito.Numero = this.facturaCompleta.Numero + 1;
			notaCredito.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			notaCredito.Cliente = this.ClienteSinBonificacion;
			notaCredito.Items.Clear();
			
			// Agrego solo algunos de los productos abstractos de la factura
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 1);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 1";
			itemAbstracto.PrecioUnitario = 500.0m;
			notaCredito.Items.Add(itemAbstracto);
			
			itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 4);
			itemAbstracto.Descripcion = itemAbstracto.ProductoDescripcion + " MOD 2";
			itemAbstracto.PrecioUnitario = 275.90m;
			notaCredito.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(notaCredito);
			
			try {
				notaCredito.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Falló la primer nota de crédito: {0}", this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			// Creo la segunda nota de crédito, que también va a tener un producto grifería, pero
			// sin superar la cantidad que hay en la factura
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			
			// Agrego y completo todos los productos abstractos de la factura. Uno de los
			// productos abstractos tiene una cantidad superior a la suma de la nota de crédito
			// con la factura asociada
			ItemPreFactura itemNotaCreditoAbstracto = TestBase.CrearItemPreFactura(this.ProductoAbstracto, 2);
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " MOD 1";
			itemNotaCreditoAbstracto.PrecioUnitario = 500.0m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			itemNotaCreditoAbstracto = TestBase.CrearItemPreFactura(this.ProductoAbstracto, 3);
			itemNotaCreditoAbstracto.Descripcion = itemNotaCreditoAbstracto.ProductoDescripcion + " MOD 2";
			itemNotaCreditoAbstracto.PrecioUnitario = 275.90m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la nota de credito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		#endregion
		
		#region Validaciones de cantidad de productos con una sola nota de crédito
		[Test]
		public void NotasCredito_CantidadDeUnProductoIncorrecta()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 10));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 20));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			/* Ahora creo la PreNotaCredito, que va a tener 8 productos mosaico más, pero que
			 * en total con la nota de crédito anterior suman 16, y esto es INCORRECTO */
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoMosaico, 22));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la PreNotaCredito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		
		[Test]
		public void NotasCredito_CantidadDeUnProductoIncorrecta_ConProductoAbstracto()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			// Actualizo el precio de los productos
			this.ProductoGriferia.PrecioCosto = 10.0m;
			this.ProductoGriferia.ImporteSinIva = 15.0m;
			this.ProductoGriferia.Update();
			
			this.ProductoMosaico.PrecioCosto = 20.0m;
			this.ProductoMosaico.ImporteSinIva = 25.0m;
			this.ProductoMosaico.Update();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			// Agrego los productos abstractos con un precio unitario determinado
			ItemFactura itemAbstracto = TestBase.CrearItemFactura(this.ProductoAbstracto, 2);
			itemAbstracto.PrecioUnitario = 500.0m;
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			/* Ahora creo la PreNotaCredito, que va a tener 8 productos mosaico más, pero que
			 * en total con la nota de crédito anterior suman 16, y esto es INCORRECTO */
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			
			// Agrego el item pero con una cantidad que, sumándola a la nota de crédito anterior,
			// supera a la factura asociada.
			ItemPreFactura itemNotaCreditoAbstracto = TestBase.CrearItemPreFactura(this.ProductoAbstracto, 3);
			itemNotaCreditoAbstracto.PrecioUnitario = 500.0m;
			this.notaCreditoCompleta.Items.Add(itemNotaCreditoAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				Assert.Fail("Se pudo crear la PreNotaCredito");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex), "Propiedad fallida Items");
			}
		}
		
		[Test]
		public void NotaCredito_ConItemsProductoRepetidos()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 10));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 5));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoMosaico, 5));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			PreNotaCredito preNotaCreditoLeida = PreNotaCredito.FindAll()[0];
			
			Assert.IsNotNull(preNotaCreditoLeida.FacturaAsociada, "Factura asociada no es nula");
			Assert.AreEqual(this.facturaCompleta.Id, preNotaCreditoLeida.FacturaAsociada.Id, "Factura asociada id");
			
			// Items de la pre nota de crédito
			Assert.AreEqual(1, preNotaCreditoLeida.Items.Count, "Cantidad items");
			Assert.AreEqual(this.ProductoMosaico.Id, preNotaCreditoLeida.Items[0].ProductoId, "Item producto id");
			Assert.AreEqual(5, preNotaCreditoLeida.Items[0].Cantidad, "Item cantidad");
		}
		#endregion
		
		#region Otras validaciones
		[Test]
		public void NotaCredito_ProductoDescuentoParaNotaCredito()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			
			// Agrego los productos abstractos con un precio unitario determinado
			ItemFactura itemAbstracto = TestBase.CrearItemFactura(this.ProductoGriferia, 2);
			this.facturaCompleta.Items.Add(itemAbstracto);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Ahora creo la pre nota de crédito
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			
			this.notaCreditoCompleta.Items.Clear();
			
			ItemPreFactura itemPreNotaCredito = TestBase.CrearItemPreFactura(this.ProductoNotaCredito, 1);
			itemPreNotaCredito.Cantidad = 1.00m;
			itemPreNotaCredito.PrecioUnitario = this.facturaCompleta.ImporteTotal / 2.0m;
			
			this.notaCreditoCompleta.Items.Add(itemPreNotaCredito);
			
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			Assert.IsNotNull(PreNotaCredito.TryFind(this.notaCreditoCompleta.Id), "PreNotaCredito creada");
		}
		
		[Test]
		public void NotaDeCredito_Sobre_Factura_ConNotasDeCredito_Anuladas()
		{
			// Borro todas las PreNotaCredito y las Facturas (debe ser en ese orden porque las
			// PreNotaCredito están relacionadas con las Facturas).
			ActiveRecordBase<PreNotaCredito>.DeleteAll();
			ActiveRecordBase<Factura>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			// Creo la factura
			this.facturaCompleta = TestBase.CrearFactura();
			this.facturaCompleta.Cliente = this.ClienteSinBonificacion;
			this.facturaCompleta.Items.Clear();
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			this.facturaCompleta.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaCompleta);
			this.facturaCompleta.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la NotaCredito pero anulada
			NotaCredito notaCredito = TestBase.CrearNotaCredito();
			notaCredito.Anulada = true;
			notaCredito.Tipo = this.facturaCompleta.Tipo;
			notaCredito.Codigo = this.facturaCompleta.Codigo;
			notaCredito.Numero = this.facturaCompleta.Numero + 1;
			notaCredito.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			notaCredito.Cliente = this.ClienteSinBonificacion;
			notaCredito.Items.Clear();
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			notaCredito.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(notaCredito);
			notaCredito.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo la PreNotaCredito, que debería poder crearse porque la nota de crédito anterior
			// está anulada
			this.notaCreditoCompleta = TestBase.CrearPreNotaCredito();
			this.notaCreditoCompleta.Tipo = this.facturaCompleta.Tipo;
			this.notaCreditoCompleta.FacturaAsociada = Factura.Find(this.facturaCompleta.Id);
			this.notaCreditoCompleta.Cliente = this.ClienteSinBonificacion;
			this.notaCreditoCompleta.Items.Clear();
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoGriferia, 1));
			this.notaCreditoCompleta.Items.Add(TestBase.CrearItemPreFactura(this.ProductoMosaico, 2));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.notaCreditoCompleta);
			
			try {
				this.notaCreditoCompleta.Create();
				
				PreNotaCredito preNotaCreditoLeida = PreNotaCredito.Find(this.notaCreditoCompleta.Id);
				
				Assert.IsNotNull(preNotaCreditoLeida.FacturaAsociada, "Factura asociada no nula");
				Assert.AreEqual(this.facturaCompleta.Id, preNotaCreditoLeida.FacturaAsociada.Id, "Factura asociada");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("No se pudo crear la PreNotaCredito: " + this.ObtenerMensaje(ex));
			}
		}
		#endregion
		#endregion
	}
}
