// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

using NUnit.Framework;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using NHibernate.Criterion;

using Stratos.Common;
using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture()]
	public class PresupuestoTests : DocumentoGenericoTestBase
	{
		#region Campos
		private Presupuesto presupuestoCompleto;
		private Presupuesto presupuestoVacio;
		#endregion

		#region Métodos de inicialización y finalización de la prueba
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			this.presupuestoCompleto = TestBase.CrearPresupuesto();
			this.presupuestoCompleto.Cliente = this.ClienteSinBonificacion2;
			this.presupuestoCompleto.Items.Clear();
			this.presupuestoCompleto.Items.Add(TestBase.CrearItemPresupuesto(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesPresupuesto(this.presupuestoCompleto);
			
			this.presupuestoVacio = new Presupuesto();
		}
		#endregion

		#region Pruebas de persistencia de datos
		[Test()]
		public void Creacion()
		{
			try {
				this.presupuestoCompleto.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			base.FlushAndRecreateReadOnlyScope();
			
			Presupuesto[] presupuestos = Presupuesto.FindAll();
			
			Assert.AreEqual(1, presupuestos.Length, "Cantidad de presupuestos insertadas");
			
			Presupuesto pr = presupuestos[0];
			Assert.AreEqual(this.ClienteSinBonificacion2.Codigo, pr.Cliente.Codigo, "Id cliente");
			Assert.AreEqual(this.ClienteSinBonificacion2.RazonSocial, pr.Cliente.RazonSocial, "Razon social cliente");
			Assert.AreEqual(pr.Cliente.RazonSocial, pr.RazonSocial, "Razon social cliente 2");
			Assert.AreEqual("123456", pr.Telefono, "Telefono cliente");
			Assert.AreEqual("alguncorreo@algundominio.com", pr.CorreoElectronico, "Correo electrónico cliente");
			Assert.AreEqual(72.60m, pr.ImporteBruto, "Importe bruto");
			Assert.AreEqual(0.00m, pr.PorcentajeBonificacion, "Porcentaje bonificación");
			Assert.AreEqual(0.00m, pr.ImporteBonificacion, "Importe bonificación");
			Assert.AreEqual(72.60m, pr.ImporteNeto, "Importe Neto");
			Assert.AreEqual(72.60m, pr.ImporteFinal, "Importe final");
			Assert.AreEqual("Estas son las observaciones", pr.Observaciones, "Observaciones");
			Assert.AreEqual(DateTime.Today, pr.Fecha.Date, "Fecha");
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinDatos()
		{
			this.presupuestoVacio.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinFecha()
		{
			this.presupuestoCompleto.Fecha = this.presupuestoVacio.Fecha;
			this.presupuestoCompleto.Create();
		}

		[Test()]
		public void SinCliente()
		{
			this.presupuestoCompleto.Cliente = this.presupuestoVacio.Cliente;
			this.presupuestoCompleto.Create();
		}
		
		[Test()]
		public void SinCliente_SinRazonSocial()
		{
			this.presupuestoCompleto.Cliente = this.presupuestoVacio.Cliente;
			this.presupuestoCompleto.RazonSocial = this.presupuestoVacio.RazonSocial;
			this.presupuestoCompleto.Create();
		}
		
		[Test()]
		public void SinTelefono()
		{
			this.presupuestoCompleto.Telefono = this.presupuestoVacio.Telefono;
			this.presupuestoCompleto.Create();
		}
		
		[Test()]
		public void SinCorreoElectronico()
		{
			this.presupuestoCompleto.CorreoElectronico = this.presupuestoVacio.CorreoElectronico;
			this.presupuestoCompleto.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinItems()
		{
			this.presupuestoCompleto.Items = this.presupuestoVacio.Items;
			this.presupuestoCompleto.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteBruto()
		{
			this.presupuestoCompleto.ImporteBruto = this.presupuestoVacio.ImporteBruto;
			this.presupuestoCompleto.Create();
		}

		[Test()]
		public void SinImporteBonificacion()
		{
			this.presupuestoCompleto.ImporteBonificacion =
				this.presupuestoVacio.ImporteBonificacion;
			this.presupuestoCompleto.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinImporteNeto()
		{
			this.presupuestoCompleto.ImporteNeto = this.presupuestoVacio.ImporteNeto;
			this.presupuestoCompleto.Create();
		}

		[Test()]
		public void SinObservaciones()
		{
			this.presupuestoCompleto.Observaciones = this.presupuestoVacio.Observaciones;
			this.presupuestoCompleto.Create();
		}
		
		[Test]
		public void Eliminacion()
		{
			this.presupuestoCompleto.Create();
			this.FlushAndRecreateScope();
			
			ItemPresupuesto[] itemsPresupuesto = ItemPresupuesto.FindAll();
			Assert.AreEqual(1, itemsPresupuesto.Length, "Cantidad de items después de borrar");
			
			// Ahora elimino el presupuesto y me aseguro que los items se borren
			this.presupuestoCompleto = Presupuesto.Find(this.presupuestoCompleto.Id);
			this.presupuestoCompleto.Delete();
			this.FlushAndRecreateReadOnlyScope();
			
			itemsPresupuesto = ItemPresupuesto.FindAll();
			Assert.AreEqual(0, itemsPresupuesto.Length, "Cantidad de items después de borrar");
		}
		#endregion

		#region Pruebas de alto nivel (código repetido, bonificación excesiva, etc)
		[Test()]
		public void FechaMinimaCorrecta()
		{
			this.presupuestoCompleto.Fecha = DateTime.Today.Date - TimeSpan.FromDays(90);
			this.presupuestoCompleto.Create();
		}
		
		[Test()]
		public void FechaMinimaCorrecta2()
		{
			this.presupuestoCompleto.Fecha = DateTime.Today.Date - TimeSpan.FromDays(30);
			this.presupuestoCompleto.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void FechaMinimaIncorrecta()
		{
			this.presupuestoCompleto.Fecha = DateTime.Today - TimeSpan.FromDays(91);
			this.presupuestoCompleto.Create();
		}

		[Test()]
		public void FechaMaximaCorrecta()
		{
			this.presupuestoCompleto.Fecha = DateTime.Today;
			this.presupuestoCompleto.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void FechaMaximaIncorrecta()
		{
			this.presupuestoCompleto.Fecha = DateTime.Today + TimeSpan.FromDays(1);
			this.presupuestoCompleto.Create();
		}
		
		[Test()]
		public void ImportesValidosConBonificacion()
		{
			this.presupuestoCompleto.Cliente = this.ClienteConBonificacion;
			this.presupuestoCompleto.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPresupuesto item1 = TestBase.CrearItemPresupuesto(this.ProductoGriferia, 2);
			ItemPresupuesto item2 = TestBase.CrearItemPresupuesto(this.ProductoMosaico, 125.3m);
			this.presupuestoCompleto.Items.Clear();
			this.presupuestoCompleto.Items.Add(item1);
			this.presupuestoCompleto.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesPresupuesto(this.presupuestoCompleto);
			
//			this.presupuestoCompleto.ImporteBruto = 3374.56m;
//			this.presupuestoCompleto.ImporteBonificacion = 337.46m;
//			this.presupuestoCompleto.ImporteNeto = 3037.10m;
			
			try {
				this.presupuestoCompleto.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void ImportesValidosSinBonificacion()
		{
			this.presupuestoCompleto.Cliente = this.ClienteSinBonificacion2;
			this.presupuestoCompleto.PorcentajeBonificacion =
				this.ClienteSinBonificacion2.PorcentajeBonificacion;
			
			ItemPresupuesto item1 = TestBase.CrearItemPresupuesto(this.ProductoGriferia, 2);
			ItemPresupuesto item2 = TestBase.CrearItemPresupuesto(this.ProductoMosaico, 125.3m);
			this.presupuestoCompleto.Items.Clear();
			this.presupuestoCompleto.Items.Add(item1);
			this.presupuestoCompleto.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesPresupuesto(this.presupuestoCompleto);
			
//			this.presupuestoCompleto.ImporteBruto = 3374.56m;
//			this.presupuestoCompleto.ImporteBonificacion = 0.00m;
//			this.presupuestoCompleto.ImporteNeto = 3374.56m;
			
			try {
				this.presupuestoCompleto.Create();
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}

		[Test()]
		public void ImporteBrutoInvalido()
		{
			this.presupuestoCompleto.Cliente = this.ClienteConBonificacion;
			this.presupuestoCompleto.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPresupuesto item1 = TestBase.CrearItemPresupuesto(this.ProductoGriferia, 2);
			ItemPresupuesto item2 = TestBase.CrearItemPresupuesto(this.ProductoMosaico, 125.3m);
			this.presupuestoCompleto.Items.Clear();
			this.presupuestoCompleto.Items.Add(item1);
			this.presupuestoCompleto.Items.Add(item2);
			
			this.presupuestoCompleto.ImporteBruto = 3074.56m; // Incorrecto
			this.presupuestoCompleto.ImporteBonificacion = 337.46m;
			this.presupuestoCompleto.ImporteNeto = 3037.10m;
			
			try {
				this.presupuestoCompleto.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteBruto", ex),
				              "Propiedad ImporteBruto errónea");
			}
		}

		[Test()]
		public void ImporteBonificacionInvalido()
		{
			this.presupuestoCompleto.Cliente = this.ClienteConBonificacion;
			this.presupuestoCompleto.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPresupuesto item1 = TestBase.CrearItemPresupuesto(this.ProductoGriferia, 2);
			ItemPresupuesto item2 = TestBase.CrearItemPresupuesto(this.ProductoMosaico, 125.3m);
			this.presupuestoCompleto.Items.Clear();
			this.presupuestoCompleto.Items.Add(item1);
			this.presupuestoCompleto.Items.Add(item2);
			
			this.presupuestoCompleto.ImporteBruto = 2788.89m;
			this.presupuestoCompleto.ImporteBonificacion = 345.00m; // Incorrecto
			this.presupuestoCompleto.ImporteNeto = 2510.00m;			
			
			try {
				this.presupuestoCompleto.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteBonificacion", ex),
				              "Propiedad ImporteBonificacion errónea");
			}
		}
		
		[Test()]
		public void ImporteNetoInvalido()
		{
			this.presupuestoCompleto.Cliente = this.ClienteConBonificacion;
			this.presupuestoCompleto.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			ItemPresupuesto item1 = TestBase.CrearItemPresupuesto(this.ProductoGriferia, 2);
			ItemPresupuesto item2 = TestBase.CrearItemPresupuesto(this.ProductoMosaico, 125.3m);
			this.presupuestoCompleto.Items.Clear();
			this.presupuestoCompleto.Items.Add(item1);
			this.presupuestoCompleto.Items.Add(item2);
			
			this.presupuestoCompleto.ImporteBruto = 3374.56m;
			this.presupuestoCompleto.ImporteBonificacion = 337.46m;
			this.presupuestoCompleto.ImporteNeto = 2037.10m; // Incorrecto
			
			try {
				this.presupuestoCompleto.Create();
				Assert.Fail("No debió haberse podido crear la factura");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("ImporteNeto", ex),
				              "Propiedad ImporteNeto errónea");
			}
		}
		
		[Test()]
		public void SinCliente_ConRazonSocial()
		{
			this.presupuestoCompleto.Cliente = null;
			this.presupuestoCompleto.RazonSocial = "Pepe Biondi";
			this.presupuestoCompleto.Telefono = "123456";
			this.presupuestoCompleto.CorreoElectronico = "algo@algo.com";
			
			try {
				this.presupuestoCompleto.Create();
				
				this.FlushAndRecreateReadOnlyScope();
				
				Presupuesto[] resultado = Presupuesto.FindAll();
				Assert.AreEqual(1, resultado.Length, "Cantidad presupuestos");
				
				Presupuesto pr = resultado[0];
				
				Assert.IsNull(pr.Cliente, "Cliente");
				Assert.AreEqual("Pepe Biondi", pr.RazonSocial, "Razón social");
				Assert.AreEqual("123456", pr.Telefono, "Teléfono");
				Assert.AreEqual("algo@algo.com", pr.CorreoElectronico, "Correo electrónico");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test()]
		public void ItemsConBonificacion()
		{
			// Cliente Reponsable Inscripto
			this.ClienteConBonificacion.TipoResponsable =
				TipoResponsable.Monotributo;
			this.ClienteConBonificacion.PorcentajeBonificacion = 12.78m;
			this.ClienteConBonificacion.Update();
			
			this.presupuestoCompleto.Cliente = this.ClienteConBonificacion;
			this.presupuestoCompleto.PorcentajeBonificacion =
				this.ClienteConBonificacion.PorcentajeBonificacion;
			
			// Agrego los items
			ItemPresupuesto item1 = TestBase.CrearItemPresupuesto(this.ProductoGriferia, 2);
			item1.PorcentajeBonificacion = 10.0m;
			
			ItemPresupuesto item2 = TestBase.CrearItemPresupuesto(this.ProductoMosaico, 125.3m);
			item2.PorcentajeBonificacion = 10.5m;
			
			this.presupuestoCompleto.Items.Clear();
			this.presupuestoCompleto.Items.Add(item1);
			this.presupuestoCompleto.Items.Add(item2);
			TestBase.ProcesarItemsYCalcularImportesPresupuesto(this.presupuestoCompleto);
			
			try {
				this.presupuestoCompleto.Create();
				
				// Valido el item1
				item1 = ItemPresupuesto.Find(item1.Id);
				
				Assert.AreEqual(10.0m, item1.PorcentajeBonificacion, "Item 1 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(60.0m * 1.21m, 2), item1.PrecioUnitario, "Item 1 Precio unitario");
				Assert.AreEqual(130.68m, item1.Importe, "Item 1 Importe");
				
				// Valido el item2
				item2 = ItemPresupuesto.Find(item2.Id);
				
				Assert.AreEqual(10.5m, item2.PorcentajeBonificacion, "Item 2 Porcentaje bonificacion");
				Assert.AreEqual(StratosUtil.RoundDecimal(21.3m * 1.21m, 2), item2.PrecioUnitario, "Item 2 Precio unitario");
				Assert.AreEqual(2889.94m, item2.Importe, "Item 2 Importe");
				
				// Valido los importes
				Assert.AreEqual(3020.62m, this.presupuestoCompleto.ImporteBruto, "Importe Bruto");
				Assert.AreEqual(386.04m, this.presupuestoCompleto.ImporteBonificacion, "Importe Bonificacion");
				Assert.AreEqual(2634.58m, this.presupuestoCompleto.ImporteNeto, "Importe Neto");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		#endregion
	}
}
