// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using NUnit.Framework;
using Castle.ActiveRecord;

using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

using Stratos.Common;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture()]
	public class ProductoTests : TestBase
	{
		#region Campos
		private Rubro rubroIndustria;
		private Rubro rubroMosaicos;
		
		private Producto productoCompleto;
		private Producto productoVacio;
		#endregion
		
		#region Métodos de inicialización y finalización de la prueba
		[SetUp]
		public override void TestInit ()
		{
			base.TestInit ();
			
			// Rubro industria
			this.rubroIndustria = new Rubro {
				Codigo = 30,
				Nombre = "Industria",
				Descripcion = "Industria",
				RubroPadre = null
			};
			this.rubroIndustria.Create();
			
			this.rubroMosaicos = TestBase.CrearRubroMosaico();
			this.rubroMosaicos.RubroPadre = this.rubroIndustria;
			this.rubroMosaicos.Create();
			
			this.productoCompleto = TestBase.CrearProductoMosaico();
			this.productoCompleto.SubRubro = this.rubroMosaicos;
			
			this.productoVacio = new Producto();
		}
		#endregion

		#region Pruebas de persistencia de datos
		[Test()]
		public void Creacion()
		{
			try {
				this.productoCompleto.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(ex.ValidationErrorMessages[0]);
			}
			
			base.FlushAndRecreateScope();
			
			Producto[] productos = Producto.FindAll();
			
			Assert.AreEqual(1, productos.Length, "Cantidad de productos insertados");
			
			Producto pr = productos[0];
			
			Assert.AreEqual(this.productoCompleto.Codigo, pr.Codigo, "Codigo");
			Assert.AreEqual(this.productoCompleto.Descripcion, pr.Descripcion, "Descripción");
			Assert.IsTrue(pr.Valido, "Valido");
			Assert.AreEqual(this.productoCompleto.ProductoAbstracto, pr.ProductoAbstracto, "ProductoAbstracto");
			Assert.AreEqual(this.productoCompleto.Unidad, pr.Unidad, "Unidad");
			
			Assert.IsNotNull(this.productoCompleto.Rubro, "Rubro no nulo");
			Assert.IsNotNull(this.productoCompleto.SubRubro, "SubRubro no nulo");
			Assert.AreEqual(this.productoCompleto.Rubro.Descripcion, pr.Rubro.Descripcion, "Rubro descripción");
			Assert.AreEqual(this.productoCompleto.SubRubro.Descripcion, pr.SubRubro.Descripcion, "SubRubro descripción");
			
			Assert.AreEqual(this.productoCompleto.PrecioCosto, pr.PrecioCosto, "Precio de costo");
			Assert.AreEqual(this.productoCompleto.ImporteSinIva, pr.ImporteSinIva, "Importe sin IVA");
			Assert.AreEqual(this.productoCompleto.PorcentajeIva, pr.PorcentajeIva, "Porcentaje IVA");
			Assert.AreEqual(StratosUtil.RoundDecimal(this.productoCompleto.ImporteSinIva *
			                 	(1 + (21m / 100m)), 2),
			                pr.ImporteVenta, "Importe venta");
			Assert.AreEqual(this.productoCompleto.Notas, pr.Notas, "Notas");
			Assert.IsFalse(pr.Eliminado, "Eliminado");
			Assert.IsFalse(pr.ActivarGestionStock, "ActivarGestionStock");
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinDatos()
		{
			this.productoVacio.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinCodigo()
		{
			this.productoCompleto.Codigo = this.productoVacio.Codigo;
			this.productoCompleto.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinDescripcion()
		{
			this.productoCompleto.Descripcion = this.productoVacio.Descripcion;
			this.productoCompleto.Create();
		}
		
		[Test()]
		public void SinProductoAbstracto()
		{
			this.productoCompleto.ProductoAbstracto = this.productoVacio.ProductoAbstracto;
			this.productoCompleto.Create();
		}
		
		[Test()]
		public void SinUnidad()
		{
			this.productoCompleto.Unidad = this.productoVacio.Unidad;
			this.productoCompleto.Create();
		}
		
		[Test()]
		public void Borrado()
		{
			this.productoCompleto.Create();
			
			base.FlushAndRecreateScope();
			
			Producto[] productos = Producto.FindAll();			
			Producto pr = productos[0];
			
			Assert.AreEqual(1, productos.Length, "Cantidad de productos sin eliminar");
			Assert.IsFalse(pr.Eliminado, "Producto sin eliminar");
			pr.Delete();
			
			base.FlushAndRecreateScope();
			
			Producto[] productos2 = Producto.FindAll();			
			Producto pr2 = productos2[0];
			
			Assert.AreEqual(1, productos2.Length, "Cantidad de productos eliminado");
			Assert.IsTrue(pr2.Eliminado, "Producto ya eliminado");
		}
		#endregion

		#region Pruebas de alto nivel
		[Test()]
		public void ProductoAbstractoPrecioCosto_e_ImporteSinIvaCero()
		{
			this.productoCompleto.ProductoAbstracto = true;
			this.productoCompleto.PrecioCosto = 0m;
			this.productoCompleto.ImporteSinIva = 0m;
			
			try {
				this.productoCompleto.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		public void ProductoAbstractoPrecioCostoIgualImporteSinIva()
		{
			this.productoCompleto.ProductoAbstracto = true;
			this.productoCompleto.PrecioCosto = this.productoCompleto.ImporteSinIva = 10m;
			
			try {
				this.productoCompleto.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void ProductoAbstractoPrecioCostoMayorQueImporteSinIva()
		{
			this.rubroMosaicos.Create();
			
			this.productoCompleto.ProductoAbstracto = true;
			this.productoCompleto.PrecioCosto = this.productoCompleto.ImporteSinIva + 1;
			
			this.productoCompleto.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void PrecioCostoIgulaCero()
		{
			this.rubroMosaicos.Create();
			
			this.productoCompleto.PrecioCosto = 0m;
			this.productoCompleto.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void PrecioCostoMenor_a_Cero()
		{
			this.rubroMosaicos.Create();
			
			this.productoCompleto.PrecioCosto = -1m;
			this.productoCompleto.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void PrecioCostoMayorQuePrecioSinIva()
		{
			this.rubroMosaicos.Create();
			
			this.productoCompleto.PrecioCosto = this.productoCompleto.ImporteSinIva + 1;
			this.productoCompleto.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void PrecioCostoIgualQuePrecioSinIva()
		{
			this.rubroMosaicos.Create();
			
			this.productoCompleto.PrecioCosto = this.productoCompleto.ImporteSinIva;
			this.productoCompleto.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void ImporteSinIvaIgualaCero()
		{
			this.rubroMosaicos.Create();
			
			this.productoCompleto.ImporteSinIva = 0m;
			this.productoCompleto.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void ImporteSinIvaMenor_a_Cero()
		{
			this.rubroMosaicos.Create();
			
			this.productoCompleto.ImporteSinIva = -1m;
			this.productoCompleto.Create();
		}

		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void CodigoRepetido()
		{
			this.productoCompleto.CreateAndFlush();
			
			Producto p2 = TestBase.CrearProductoMosaico();
			p2.Codigo = this.productoCompleto.Codigo;
			p2.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void DescripcionRepetida()
		{
			this.rubroMosaicos.Create();
			
			this.productoCompleto.CreateAndFlush();
			
			Producto p2 = new Producto {
				Codigo = "M-0555",
				Descripcion = this.productoCompleto.Descripcion,
				ImporteSinIva = 21.3m,
				PrecioCosto = 25.12m,
				SubRubro = this.rubroMosaicos,
				Unidad = UnidadProducto.MetroCuadrado
			};
			
			p2.Create();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void CodigoInvalido()
		{
			this.rubroMosaicos.Create();
			
			this.productoCompleto.Codigo = "";
			this.productoCompleto.Create();
		}
		#endregion
	}
}
