// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Text;

using NUnit.Framework;
using Castle.ActiveRecord;

using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture()]
	public class ProveedorTests : TestBase
	{
		#region Campos
		private Proveedor proveedorCompleto;
		private Proveedor proveedorVacio;
		#endregion
		
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			this.proveedorCompleto = CrearProveedorFV();
			this.proveedorVacio = new Proveedor();
		}
		
		#region Pruebas de persistencia de datos
		[Test()]
		public void Creacion()
		{
			try {
				this.proveedorCompleto.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(ex.ValidationErrorMessages[0]);
			}
			
			base.FlushAndRecreateScope();
			
			Proveedor[] proveedores = Proveedor.FindAll();
			
			Assert.AreEqual(1, proveedores.Length, "Cantidad de proveedores insertados");
			
			Proveedor proveedor = proveedores[0];
			Assert.AreEqual("Proveedor de FV", proveedor.RazonSocial, "Razón social");
			Assert.AreEqual("Bv. Galvez 956", proveedor.Domicilio.Direccion, "Direccion");
			Assert.AreEqual("Santa Fe de la Veracruz", proveedor.Domicilio.Localidad, "Localidad");
			Assert.AreEqual("3000", proveedor.Domicilio.CodigoPostal, "Codigo Postal");
			Assert.AreEqual(ProvinciaDto.SantaFe, proveedor.Domicilio.Provincia, "Provincia");
			Assert.AreEqual("0342 22222", proveedor.Telefono, "Teléfono");
			Assert.AreEqual("ventas@fv.com", proveedor.CorreoElectronico, "Correo electrónico");
			Assert.AreEqual(TipoResponsable.ResponsableInscripto, proveedor.TipoResponsable, "TipoResponsable");
			Assert.AreEqual(20062192578, proveedor.Cuit.ValorNumerico, "CUIT");
			Assert.AreEqual("Otra nota", proveedor.Notas, "Notas");
			Assert.IsFalse(proveedor.Eliminado, "Eliminado");
			Assert.AreEqual(4, proveedor.Version, "Versión");
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinDatos()
		{
			this.proveedorVacio.Create();
		}
		
		[Test()]
		public void SinCodigo()
		{
			this.proveedorCompleto.Codigo = this.proveedorVacio.Codigo;
			
			this.proveedorCompleto.Create();
		}
				
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinRazonSocial()
		{
			this.proveedorCompleto.RazonSocial = this.proveedorVacio.RazonSocial;
			
			this.proveedorCompleto.Create();
		}
		
		[Test()]
		public void SinDomicilio()
		{
			this.proveedorCompleto.Domicilio = this.proveedorVacio.Domicilio;
			this.proveedorCompleto.Create();
		}
		
		[Test()]
		public void SinDireccion()
		{
			this.proveedorCompleto.Domicilio.Direccion =
				this.proveedorVacio.Domicilio.Direccion;
			this.proveedorCompleto.Create();
		}
		
		[Test()]
		public void SinLocalidad()
		{
			this.proveedorCompleto.Domicilio.Localidad =
				this.proveedorVacio.Domicilio.Localidad;
			this.proveedorCompleto.Create();
		}
		
		[Test()]
		public void SinCodigoPostal()
		{
			this.proveedorCompleto.Domicilio.CodigoPostal =
				this.proveedorVacio.Domicilio.CodigoPostal;
			this.proveedorCompleto.Create();
		}
		
		[Test()]
		public void SinProvincia()
		{
			this.proveedorCompleto.Domicilio.Provincia =
				this.proveedorVacio.Domicilio.Provincia;
			this.proveedorCompleto.Create();
		}
		
		[Test()]
		public void SinTelefono()
		{
			this.proveedorCompleto.Telefono = this.proveedorVacio.Telefono;
			this.proveedorCompleto.Create();
		}
		
		[Test()]
		public void SinCorreoElectronico()
		{
			this.proveedorCompleto.CorreoElectronico = this.proveedorVacio.CorreoElectronico;
			this.proveedorCompleto.Create();
		}
		
		[Test()]
		public void SinTipoResponsable()
		{
			this.proveedorCompleto.TipoResponsable = this.proveedorVacio.TipoResponsable;
			
			this.proveedorCompleto.Create();
		}
		
		[Test()]
		public void SinCuit_ClienteResponsableInscripto()
		{
			this.proveedorCompleto.TipoResponsable = TipoResponsable.ResponsableInscripto;
			this.proveedorCompleto.Cuit = this.proveedorVacio.Cuit;
			
			try {
				this.proveedorCompleto.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		
		[Test()]
		public void SinCuit_ClienteResponsableNoInscripto()
		{
			this.proveedorCompleto.TipoResponsable = TipoResponsable.ResponsableNoInscripto;
			this.proveedorCompleto.Cuit = this.proveedorVacio.Cuit;
			
			try {
				this.proveedorCompleto.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		
		[Test()]
		public void SinCuit_ClienteMonotributo()
		{
			this.proveedorCompleto.TipoResponsable = TipoResponsable.Monotributo;
			this.proveedorCompleto.Cuit = this.proveedorVacio.Cuit;
			
			try {
				this.proveedorCompleto.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		
		[Test()]
		public void SinCuit_ClienteExento()
		{
			this.proveedorCompleto.TipoResponsable = TipoResponsable.Exento;
			this.proveedorCompleto.Cuit = this.proveedorVacio.Cuit;
			
			try {
				this.proveedorCompleto.Create();
				Assert.Fail("No hubo excepcion");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex),
				              "Propiedad Cuit errónea");
			}
		}
		
		[Test()]
		public void SinCuit_ClienteConsumidorFinal()
		{
			this.proveedorCompleto.TipoResponsable = TipoResponsable.ConsumidorFinal;
			this.proveedorCompleto.Cuit = this.proveedorVacio.Cuit;
			
			try {
				this.proveedorCompleto.Create();
			}
			catch (ActiveRecordValidationException ex) {
				Assert.Fail("Cliente Consumidor Final no pudo crearse por no tener cuit");
			}
		}
		
		[Test()]
		public void SinNotas()
		{
			this.proveedorCompleto.Notas = this.proveedorVacio.Notas;
			this.proveedorCompleto.Create();
		}
		#endregion
		
		#region Pruebas de alto nivel (código repetido, bonificación excesiva, etc)
		[Test]
		public void EmailInvalido1()
		{
			this.proveedorCompleto.CorreoElectronico = "ventas";
			
			try {
				this.proveedorCompleto.Create();
				Assert.Fail("No hubo excepción");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("CorreoElectronico", ex),
				              "Propiedad CorreoElectronico errónea");
			}
		}
		
		[Test]
		public void EmailInvalido2()
		{
			this.proveedorCompleto.CorreoElectronico = "ventas@";
			
			try {
				this.proveedorCompleto.Create();
				Assert.Fail("No hubo excepción");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("CorreoElectronico", ex),
				              "Propiedad CorreoElectronico errónea");
			}
		}
		
		[Test]
		public void EmailInvalido3()
		{
			this.proveedorCompleto.CorreoElectronico = "@";
			
			try {
				this.proveedorCompleto.Create();
				Assert.Fail("No hubo excepción");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("CorreoElectronico", ex),
				              "Propiedad CorreoElectronico errónea");
			}
		}
		
		[Test]
		public void EmailInvalido4()
		{
			this.proveedorCompleto.CorreoElectronico = "@fv";
			
			try {
				this.proveedorCompleto.Create();
				Assert.Fail("No hubo excepción");
			}
			catch (ActiveRecordValidationException ex) {
				// Me aseguro que haya un error en la propiedad correspondiente
				Assert.IsTrue(this.PropiedadFallida("CorreoElectronico", ex),
				              "Propiedad CorreoElectronico errónea");
			}
		}
		
		[Test()]
		public void CuitRepetido()
		{
			Proveedor p1 = CrearProveedorFerrum();
			p1.Create();
			
			FlushAndRecreateScope();

			Proveedor p2 = CrearProveedorFV();
			p2.Cuit.ValorNumerico = p1.Cuit.ValorNumerico;
			
			try {
				p2.Create();
				Assert.Fail("Se pudo crear el proveedor");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Cuit", ex), "Propiedad fallida");
			}
		}
		#endregion
	}
}
