// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using System.Collections.Generic;

using NUnit.Framework;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using NHibernate;
using NHibernate.Criterion;

using Stratos.Common;
using Stratos.Tests;
using Stratos.ModeloObjetos;
using Stratos.ModeloDto;

namespace Stratos.ModeloObjetos.Tests
{
	[TestFixture]
	public class ReciboTests  : DocumentoGenericoTestBase
	{
		#region Campos
		private Factura facturaAsociada;
		private Factura factura1;
		private Factura factura2;
		
		private Cheque cheque1;
		
		private Recibo reciboCompleto;
		private Recibo reciboVacio;
		#endregion
		
		#region Métodos de inicialización y finalización de la prueba
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			// Creo una factura asociada
			this.factura1 = TestBase.CrearFactura();
			this.factura1.Codigo = 4;
			this.factura1.Numero = 1;
			this.factura1.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.factura1.Cliente = this.ClienteSinBonificacion;
			this.factura1.Items.Clear();
			this.factura1.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 10));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.factura1);
			this.factura1.Create();
			
			// Creo una factura asociada
			this.factura2 = TestBase.CrearFactura();
			this.factura2.Codigo = 4;
			this.factura2.Numero = 2;
			this.factura2.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.factura2.Cliente = this.ClienteSinBonificacion;
			this.factura2.Items.Clear();
			this.factura2.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 7));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.factura2);
			this.factura2.Create();
			
			// Creo primero la factura que estará asociada al recibo
			this.facturaAsociada = TestBase.CrearFactura();
			this.facturaAsociada.Codigo = 3;
			this.facturaAsociada.Numero = 1;
			this.facturaAsociada.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.facturaAsociada.Cliente = this.ClienteSinBonificacion;
			this.facturaAsociada.Items.Clear();
			this.facturaAsociada.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.facturaAsociada);
			this.facturaAsociada.Create();
			
			// Creo un cheque
			this.cheque1 = new Cheque();
			this.cheque1.Banco = "BANCO XYZ";
			this.cheque1.Sucursal = "SUCURSAL DEL BANCO";
			this.cheque1.Numero = 123;
			this.cheque1.FechaEmision = DateTime.Today.Date;
			this.cheque1.FechaVencimiento = DateTime.Today.Date + TimeSpan.FromDays(30);
			this.cheque1.Importe = this.factura1.ImporteTotal;
			this.cheque1.Observacion = "OBSERVACION DEL CHEQUE";
			
			this.FlushAndRecreateScope();
			
			// Creo el recibo completo
			this.reciboCompleto = TestBase.CrearRecibo();
			this.reciboCompleto.Cliente = this.ClienteSinBonificacion;
			this.reciboCompleto.Items.Clear();
			this.reciboCompleto.Items.Add(new ItemRecibo(this.facturaAsociada.ImporteTotal));
			
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			// Creo el recibo vacío
			this.reciboVacio = new Recibo();
		}
		#endregion
		
		#region Pruebas de persistencia de datos
		[Test()]
		public void Creacion()
		{
			try {
				this.reciboCompleto.Create();
			}
			catch(ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.FlushAndRecreateScope();
			
			Recibo[] recibos = Recibo.FindAll();
			
			Assert.AreEqual(1, recibos.Length, "Cantidad de recibos");
			
			Recibo rec = recibos[0];
			
			// Fecha
			Assert.AreEqual(this.reciboCompleto.Fecha, rec.Fecha, "Fecha");
			
			// Cliente
			Assert.IsNotNull(rec.Cliente, "Cliente no nulo");
			Assert.AreEqual(this.reciboCompleto.Cliente.Codigo, rec.Cliente.Codigo, "Cliente código");
			
			// Items
			Assert.AreEqual(1, rec.Items.Count, "Cantidad de items");
			Assert.AreEqual(MedioPago.Efectivo, rec.Items[0].MedioPago, "Item medio pago");
			Assert.AreEqual(this.facturaAsociada.ImporteTotal, rec.Items[0].Importe, "Item importe");
			
			Assert.AreEqual("Concepto del recibo", rec.Concepto, "Recibo concepto");
			Assert.AreEqual(this.facturaAsociada.ImporteTotal, rec.Importe, "Recibo importe");
		}
		
		[Test]
		public void CredoDosRecibos()
		{
			this.reciboCompleto.Create();
			this.FlushAndRecreateScope();
			
			// Creo el recibo completo
			this.reciboCompleto = TestBase.CrearRecibo();
			this.reciboCompleto.Cliente = this.ClienteSinBonificacion;
			this.reciboCompleto.Items.Clear();
			this.reciboCompleto.Items.Add(new ItemRecibo(this.facturaAsociada.ImporteTotal));
			
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			this.reciboCompleto.Create();
			
			this.FlushAndRecreateScope();
		}
		
		[Test()]
		[ExpectedException(typeof(ActiveRecordValidationException))]
		public void SinDatos()
		{
			this.reciboVacio.Create();
		}
		
		[Test]
		public void SinCliente()
		{
			try {
				this.reciboCompleto.Cliente = this.reciboVacio.Cliente;
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Cliente", ex));
			}
		}
		
		[Test]
		public void SinItems()
		{
			try {
				this.reciboCompleto.Items = this.reciboVacio.Items;
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Items", ex));
			}
		}
		
		[Test]
		public void SinConcepto()
		{
			try {
				// Aquí el concepto es nulo
				this.reciboCompleto.Concepto = this.reciboVacio.Concepto;
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Concepto", ex));
			}
		}
		
		[Test]
		public void SinImporte()
		{
			try {
				this.reciboCompleto.Importe = this.reciboVacio.Importe;
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException ex) {
				Assert.IsTrue(this.PropiedadFallida("Importe", ex));
			}
		}
		
		[Test]
		public void Eliminacion()
		{
			// Creo el recibo
			this.reciboCompleto.Create();
			this.FlushAndRecreateScope();
			
			// Lo elimino
			this.reciboCompleto.Delete();
			this.FlushAndRecreateReadOnlyScope();
			
			// Los items del recibo no deben existir más
			ItemRecibo[] itemsRecibo = ItemRecibo.FindAll();
			Assert.AreEqual(0, itemsRecibo.Length, "Cantidad items recibo");
		}
		#endregion
		
		#region Pruebas de alto nivel (código repetido, bonificación excesiva, etc)
		#region Documentos de crédito
		[Test]
		public void Recibo_SiempreCrea_DocumentoCredito()
		{
			this.reciboCompleto.Create();
			this.FlushAndRecreateScope();
			
			// Busco los documentos de débito asociados a la factura
			DocumentoCredito[] documentosCredito =
				DocumentoCredito.FindAll(Restrictions.Eq("TipoDocumento", TipoDocumento.Recibo),
				                         Restrictions.Eq("IdDocumento", this.reciboCompleto.Id));
			
			Assert.AreEqual(1, documentosCredito.Length, "Cantidad documentos crédito");
			
			DocumentoCredito dc = documentosCredito[0];
			Assert.IsNotNull(dc.Cliente, "Documento crédito cliente no nulo");
			Assert.AreEqual(this.reciboCompleto.Cliente.Codigo, dc.Cliente.Codigo, "Documento crédito código cliente");
			
			Assert.AreEqual(TipoDocumento.Recibo, dc.TipoDocumento, "Documento crédito tipo documento");
			Assert.AreEqual(this.reciboCompleto.Id, dc.IdDocumento, "Documento crédito id documento");
			Assert.AreEqual(0, dc.CodigoNotaCredito, "Documento crédito código");
			Assert.AreEqual(0, dc.NumeroNotaCredito, "Documento crédito número");
			
			Assert.AreEqual(this.reciboCompleto.Importe, dc.ImporteTotal, "Documento crédito importe total");
			Assert.AreEqual(this.reciboCompleto.Importe, dc.ImportePendiente, "Documento crédito importe pendiente");
		}
		#endregion
		
		#region Items con cheques
		[Test]
		public void ItemCheque_DatosValidos()
		{
			// Creo el recibo completo
			this.reciboCompleto.Items.Clear();
			this.reciboCompleto.Items.Add(new ItemRecibo(this.cheque1));
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			try {
				this.reciboCompleto.Create();
				
				this.FlushAndRecreateScope();
				
				Assert.AreEqual(this.factura1.ImporteTotal,
				                this.reciboCompleto.Importe, "Importe del recibo");
				
				// Verifico que el Cheque se haya guardado y coincida con el item del recibo
				Cheque[] chequesRecuperado = Cheque.FindAll();
				Recibo reciboRecuperado = Recibo.Find(this.reciboCompleto.Id);
				
				// Datos del cheque
				Assert.AreEqual(1, chequesRecuperado.Length, "Cantidad cheques");
				Assert.AreEqual("BANCO XYZ", chequesRecuperado[0].Banco, "Banco cheque");
				Assert.AreEqual("SUCURSAL DEL BANCO", chequesRecuperado[0].Sucursal, "Sucursal cheque");
				Assert.AreEqual(123, chequesRecuperado[0].Numero, "Numero cheque");
				Assert.AreEqual(DateTime.Today.Date, chequesRecuperado[0].FechaEmision, "Fecha emisión cheque");
				Assert.AreEqual(DateTime.Today.Date + TimeSpan.FromDays(30), chequesRecuperado[0].FechaVencimiento, "Fecha vencimiento cheque");
				Assert.AreEqual(this.factura1.ImporteTotal, chequesRecuperado[0].Importe, "Importe cheque");
				Assert.AreEqual("OBSERVACION DEL CHEQUE", chequesRecuperado[0].Observacion, "Observación cheque");
				
				// Cheque asociado al item
				Assert.IsNotNull(reciboRecuperado.Items[0].Cheque, "Cheque item no nulo");
				Assert.AreEqual(chequesRecuperado[0].Id, reciboCompleto.Items[0].Cheque.Id, "Id cheque asociado");
				Assert.AreEqual(chequesRecuperado[0].Importe, reciboCompleto.Items[0].Importe, "Importe item igual a importe cheque");
			}
			catch (ActiveRecordValidationException e) {
				Assert.Fail(this.ObtenerMensaje(e));
			}
		}
		
		[Test]
		public void ItemCheque_SinChequeAsignado()
		{
			// Creo el recibo completo
			this.reciboCompleto.Items.Clear();
			
			ItemRecibo itemRecibo = new ItemRecibo();
			itemRecibo.MedioPago = MedioPago.Cheque;
			itemRecibo.Importe = this.cheque1.Importe;
			itemRecibo.Cheque = null;
			this.reciboCompleto.Items.Add(itemRecibo);
			
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			try {
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("Items", e), "Propiedad fallida");
			}
		}
		
		[Test]
		public void ItemEfectivo_ChequeAsignado()
		{
			// Creo el recibo completo
			this.reciboCompleto.Items.Clear();
			
			// Creo un item con el modo de pago efectivo, pero le asigno un cheque,
			// lo cual es inválido
			ItemRecibo itemRecibo = new ItemRecibo(this.cheque1.Importe);
			itemRecibo.Cheque = this.cheque1;
			
			this.reciboCompleto.Items.Add(itemRecibo);
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			try {
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("Items", e), "Propiedad fallida");
			}
		}
		
		[Test]
		public void ItemCheque_DatosInvalidos_SinBanco()
		{
			// Creo el recibo completo
			this.reciboCompleto.Items.Clear();
			
			// Modifico el cheque con datos inválidos
			this.cheque1.Banco = String.Empty;
			
			this.reciboCompleto.Items.Add(new ItemRecibo(this.cheque1));
			
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			try {
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("Banco", e), "Propiedad fallida");
			}
		}
		
		[Test]
		public void ItemCheque_DatosInvalidos_SinSucursal()
		{
			// Creo el recibo completo
			this.reciboCompleto.Items.Clear();
			
			// Modifico el cheque con datos inválidos
			this.cheque1.Sucursal = String.Empty;
			
			this.reciboCompleto.Items.Add(new ItemRecibo(this.cheque1));
			
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			try {
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("Sucursal", e), "Propiedad fallida");
			}
		}
		
		[Test]
		public void ItemCheque_DatosInvalidos_SinNumero()
		{
			// Creo el recibo completo
			this.reciboCompleto.Items.Clear();
			
			// Modifico el cheque con datos inválidos
			this.cheque1.Numero = 0;
			
			this.reciboCompleto.Items.Add(new ItemRecibo(this.cheque1));
			
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			try {
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("Numero", e), "Propiedad fallida");
			}
		}
		
		[Test]
		public void ItemCheque_DatosInvalidos_SinFechaEmision()
		{
			// Creo el recibo completo
			this.reciboCompleto.Items.Clear();
			
			// Modifico el cheque con datos inválidos
			this.cheque1.FechaEmision = DateTime.MinValue;
			
			this.reciboCompleto.Items.Add(new ItemRecibo(this.cheque1));
			
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			try {
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("FechaEmision", e), "Propiedad fallida");
			}
		}
		
		[Test]
		public void ItemCheque_DatosInvalidos_SinFechaVencimiento()
		{
			// Creo el recibo completo
			this.reciboCompleto.Items.Clear();
			
			// Modifico el cheque con datos inválidos
			this.cheque1.FechaVencimiento = DateTime.MinValue;
			
			this.reciboCompleto.Items.Add(new ItemRecibo(this.cheque1));
			
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			try {
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("FechaVencimiento", e), "Propiedad fallida");
			}
		}
		
		[Test]
		public void ItemCheque_DatosInvalidos_FechaEmision_Mayor_FechaVencimiento()
		{
			// Creo el recibo completo
			this.reciboCompleto.Items.Clear();
			
			// Modifico el cheque con datos inválidos
			this.cheque1.FechaEmision = DateTime.Today.Date + TimeSpan.FromDays(30);
			this.cheque1.FechaVencimiento = DateTime.Today.Date;
			
			this.reciboCompleto.Items.Add(new ItemRecibo(this.cheque1));
			
			TestBase.ProcesarItemsRecibo(this.reciboCompleto);
			
			this.reciboCompleto.Concepto = "Concepto del recibo";
			
			try {
				this.reciboCompleto.Create();
				Assert.Fail("Se pudo crear el recibo");
			}
			catch (ActiveRecordValidationException e) {
				Assert.IsTrue(this.PropiedadFallida("FechaVencimiento", e), "Propiedad fallida");
			}
		}
		#endregion
		#endregion
	}
}
