// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Reflection;
using System.IO;
using NUnit.Framework;

using Castle.Core.Configuration;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using Castle.ActiveRecord.Framework.Config;
using Castle.ActiveRecord.Testing;

using Stratos.Common;
using Stratos.ModeloObjetos;
using Stratos.ModeloObjetos.MetaCampos;
using Stratos.ModeloDto;

namespace Stratos.Tests
{
	public abstract class TestBase : InMemoryTest
	{
		protected TransactionScope transaction;
		
		protected const string archivoConfiguracionActiveRecord =
			"ActiveRecordTests.config.xml";
		
		#region Métodos de inicialización y finalización de pruebas
		#region Métodos de InMemoryTest
		public override Assembly[] GetAssemblies()
		{
			return new Assembly[] { Assembly.GetAssembly(typeof(Stratos.ModeloObjetos.Cliente)) };
		}
		
		public override void Configure(InPlaceConfigurationSource config)
		{
			/* Si existe un archivo de configuración para ActiveRecord, llamado ActiveRecord.config.xml,
			 * entonces cambio las opciones por las que están allí.
			 * Esto se utiliza por ejemplo cuando queremos cambiar la base de datos SQLite por PostgreSQL. */
			if (!File.Exists(archivoConfiguracionActiveRecord))
				return;
			
			// Agrego las propiedades del archivo ActiveRecord.config.xml
			XmlConfigurationSource xmlConfig =
				new XmlConfigurationSource(archivoConfiguracionActiveRecord);
			
			config.Add(typeof(ActiveRecordBase), xmlConfig.GetConfiguration(typeof(ActiveRecordBase)));
		}
		#endregion
		
		[TestFixtureSetUp]
		public virtual void FixtureInit()
		{
			base.SetUp();
		}
		
		[TestFixtureTearDown]
		public virtual void FixterTearDown()
		{
			base.TearDown();
		}
		
		[SetUp]
		public virtual void TestInit()
		{
			ActiveRecordStarter.CreateSchema();
			this.transaction = new TransactionScope(TransactionMode.New, OnDispose.Commit);
		}
		
		[TearDown]
		public virtual void TestDispose()
		{
			if (this.transaction != null)
				this.transaction.Dispose();
		}
		#endregion
		
		#region Métodos útiles
		/// <summary>
		/// Dada una excepción de validación de ActiveRecord, devuelve el primer mensaje
		/// de error.
		/// </summary>
		/// <param name="ex">
		/// A <see cref="ActiveRecordValidationException"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/>
		/// </returns>
		protected string ObtenerMensaje(ActiveRecordValidationException ex)
		{
			string m = ex.Message;
			
			if (ex.ValidationErrorMessages.Length > 0)
				m = m + " : " + ex.ValidationErrorMessages[0];
			
			return m;
		}
		
		/// <summary>
		/// Dado un nombre de una propiedad y una excepción de validación de ActiveRecord,
		/// verifica si esa propiedad aparece como una propiedad en la que alguna validación
		/// ha fallado. Devuelve true si la encuentra, y false en caso contrario.
		/// </summary>
		/// <param name="nombrePropiedad">
		/// A <see cref="System.String"/>
		/// </param>
		/// <param name="excepcion">
		/// A <see cref="ActiveRecordValidationException"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		protected bool PropiedadFallida(string nombrePropiedad,
		                                ActiveRecordValidationException excepcion)
		{
			bool propiedadFallida = false;
			
			foreach (PropertyInfo pi in excepcion.FailedProperties) {
				if (pi.Name.Equals(nombrePropiedad)) {
					propiedadFallida = true;
					break;
				}
			}
			
			return propiedadFallida;
		}
		
		protected void RemoveTransaction()
		{
			this.FlushAndDisposeScope();
			this.transaction = null;
		}
		
		/// <summary>
		/// Hace un Dispose de la SessionScope actual y crea una que automáticamente guardará
		/// los cambios al terminar.
		/// </summary>
		protected void FlushAndRecreateScope()
		{
			this.FlushAndDisposeScope();
			this.transaction = new TransactionScope(TransactionMode.New, OnDispose.Commit);
		}
		
		/// <summary>
		/// Es similar a FlushAndRecreateScope, con la diferencia de que al crear la nueva SessionScope
		/// utiliza FlushAction.Never, lo que hace que no se guarden los cambios al hacerse un Dispose
		/// de la misma.
		/// Usar este método si se quiere hacer consultas de sólo lectura, ya que por defecto al hacer
		/// un Dispose a la SessionScope almacena todos los datos si se han modificado (y cuando se
		/// utilizan los mappings de AutoMapper para pasar de Dto a una Clase por ejemplo, la marca como
		/// modificada y al final intenta guardarla, lo que a veces ocasiona problemas).
		/// </summary>
		protected void FlushAndRecreateReadOnlyScope()
		{
			this.FlushAndDisposeScope();
			this.transaction = new TransactionScope(TransactionMode.New, OnDispose.Rollback);
		}
		
		private void FlushAndDisposeScope()
		{
			if (this.transaction != null) {
				
				try {
					this.transaction.Flush();
				}
				catch (Exception) {
					throw;
				}
				finally {
					try {
						this.transaction.Dispose();
					}
					catch {
						throw;
					}
					
					this.transaction = null;
				}
			}
		}
		#endregion
		
		#region Creación de objetos de negocio
		private void InicializarItems<T>(ref T objeto)
			where T : IDocumentoId, IDocumentoItems
		{
			using (SessionScope session = new SessionScope(FlushAction.Never)) {
				if (!ActiveRecordBase<T>.Exists(objeto.Id))
					return;
				
				objeto = ActiveRecordBase<T>.Find(objeto.Id);
				NHibernate.NHibernateUtil.Initialize(objeto.ItemsGenericos);
			}
		}
		
		/// <summary>
		/// Marca una PreFactura como no persistida.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="PreFactura"/>
		/// </param>
		public void MarcarComoNoPersistido(ref PreFactura factura)
		{
			// Primero inicializo el objeto completamente (con sus items), y luego
			// flusheo y recreo la sesión actual.
			this.InicializarItems<PreFactura>(ref factura);
			
			factura.Id = 0;
			factura.Cliente = Cliente.Find(factura.Cliente.Codigo);
			
			// Si la PreFactura es una PreNotaCredito, entonces recargo la factura asociada
			// desde la sesión actual.
			if (factura is PreNotaCredito)
				(factura as PreNotaCredito).FacturaAsociada =
					Factura.TryFind((factura as PreNotaCredito).FacturaAsociada.Id);
			
			IList<ItemPreFactura> itemsAnteriores = factura.Items;
			List<ItemPreFactura> nuevosItems = new List<ItemPreFactura>();
			
			foreach (ItemPreFactura item in itemsAnteriores) {
				item.Id = 0;
				
				nuevosItems.Add(item);
			}
			
			factura.Items = nuevosItems;
		}
		
		/// <summary>
		/// Marca una PreNotaCredito como no persistida.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="PreNotaCredito"/>
		/// </param>
		public void MarcarComoNoPersistido(ref PreNotaCredito factura)
		{
			// Primero inicializo el objeto completamente (con sus items), y luego
			// flusheo y recreo la sesión actual.
			this.InicializarItems<PreNotaCredito>(ref factura);
			
			factura.Id = 0;
			factura.Cliente = Cliente.Find(factura.Cliente.Codigo);
			
			factura.FacturaAsociada = Factura.TryFind(factura.FacturaAsociada.Id);
			
			IList<ItemPreFactura> itemsAnteriores = factura.Items;
			List<ItemPreFactura> nuevosItems = new List<ItemPreFactura>();
			
			foreach (ItemPreFactura item in itemsAnteriores) {
				item.Id = 0;
				
				nuevosItems.Add(item);
			}
			
			factura.Items = nuevosItems;
		}
		
		/// <summary>
		/// Marca una Factura como no persistida.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="Factura"/>
		/// </param>
		public void MarcarComoNoPersistido(ref Factura factura)
		{
			// Primero inicializo el objeto completamente (con sus items), y luego
			// flusheo y recreo la sesión actual.
			this.InicializarItems<Factura>(ref factura);
			
			factura.Id = 0;
			factura.Cliente = Cliente.Find(factura.Cliente.Codigo);
			
			// Items de la factura
			IList<ItemFactura> itemsAnteriores = factura.Items;
			List<ItemFactura> nuevosItems = new List<ItemFactura>();
			
			foreach (ItemFactura item in itemsAnteriores) {
				item.Id = 0;
				
				nuevosItems.Add(item);
			}
			
			factura.Items = nuevosItems;
		}
		
		/// <summary>
		/// Marca una NotaCredito como no persistida.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="NotaCredito"/>
		/// </param>
		public void MarcarComoNoPersistido(ref NotaCredito factura)
		{
			// Primero inicializo el objeto completamente (con sus items), y luego
			// flusheo y recreo la sesión actual.
			this.InicializarItems<NotaCredito>(ref factura);
			
			factura.Id = 0;
			factura.Cliente = Cliente.Find(factura.Cliente.Codigo);
			
			factura.FacturaAsociada = Factura.TryFind(factura.FacturaAsociada.Id);
			
			// Items de la factura
			IList<ItemFactura> itemsAnteriores = factura.Items;
			List<ItemFactura> nuevosItems = new List<ItemFactura>();
			
			foreach (ItemFactura item in itemsAnteriores) {
				item.Id = 0;
				
				nuevosItems.Add(item);
			}
			
			factura.Items = nuevosItems;
		}
		
		/// <summary>
		/// Marca una FacturaProveedor como no persistida.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="FacturaProveedor"/>
		/// </param>
		public void MarcarComoNoPersistido(ref FacturaCompra factura)
		{
			// Primero inicializo el objeto completamente (con sus items), y luego
			// flusheo y recreo la sesión actual.
			this.InicializarItems<FacturaCompra>(ref factura);
			
			factura.Id = 0;
			factura.Proveedor = Proveedor.Find(factura.Proveedor.Codigo);
			
			IList<ItemFacturaCompra> itemsAnteriores = factura.Items;
			List<ItemFacturaCompra> nuevosItems = new List<ItemFacturaCompra>();
			
			foreach (ItemFacturaCompra item in itemsAnteriores) {
				item.Id = 0;
				
				nuevosItems.Add(item);
			}
			
			factura.Items = nuevosItems;
		}
		
		/// <summary>
		/// Marca un presupuesto como no persistido.
		/// </summary>
		/// <param name="presupuesto">
		/// A <see cref="Presupuesto"/>
		/// </param>
		public void MarcarComoNoPersistido(ref Presupuesto presupuesto)
		{
			// Primero inicializo el objeto completamente (con sus items), y luego
			// flusheo y recreo la sesión actual.
			this.InicializarItems<Presupuesto>(ref presupuesto);
			
			presupuesto.Id = 0;
			
			if (presupuesto.Cliente != null)
				presupuesto.Cliente = Cliente.Find(presupuesto.Cliente.Codigo);
			
			IList<ItemPresupuesto> itemsAnteriores = presupuesto.Items;
			List<ItemPresupuesto> nuevosItems = new List<ItemPresupuesto>();
			
			foreach (ItemPresupuesto item in itemsAnteriores) {
				item.Id = 0;
				
				nuevosItems.Add(item);
			}
			
			presupuesto.Items = nuevosItems;
		}
		
		/// <summary>
		/// Marca un Recibo como no persistido.
		/// </summary>
		/// <param name="recibo">
		/// A <see cref="Recibo"/>
		/// </param>
		public void MarcarComoNoPersistido(ref Recibo recibo)
		{
			// Primero inicializo el objeto completamente (con sus items), y luego
			// flusheo y recreo la sesión actual.
			this.InicializarItems<Recibo>(ref recibo);
			
			recibo.Id = 0;
			recibo.Cliente = Cliente.Find(recibo.Cliente.Codigo);
			
			// Items del recibo
			IList<ItemRecibo> itemsAnteriores = recibo.Items;
			List<ItemRecibo> nuevosItems = new List<ItemRecibo>();
			
			foreach (ItemRecibo item in itemsAnteriores) {
				item.Id = 0;
				
				if (item.Cheque != null)
					item.Cheque.Id = 0;
				
				nuevosItems.Add(item);
			}
			
			recibo.Items = nuevosItems;
		}
		
		public static Domicilio CrearDomicilioMilton()
		{
			return new Domicilio {
				Direccion = "Bv. Galvez 956",
				Localidad = "Santa Fe de la Veracruz",
				CodigoPostal = "3000",
				Provincia = ProvinciaDto.SantaFe
			};
		}
		
		public static Domicilio CrearDomicilioCesar()
		{
			return new Domicilio {
				Direccion = "Laprida 5054, Piso 8, Depto D",
				Localidad = "Santa Fe de la Veracruz",
				Provincia = ProvinciaDto.Chaco
			};
		}
		
		public static Cliente CrearClienteMilton()
		{
			Cliente cliente = new Cliente {
				Codigo = 5,
				RazonSocial = "Milton Pividori",
				Domicilio = CrearDomicilioMilton(),
				Telefono = "0342 156654321",
				CorreoElectronico = "miltondp@gmail.com",
				TipoResponsable = TipoResponsable.Monotributo,
				Cuit = new CUIT(30674393683),
				PorcentajeBonificacion = 15m,
				Notas = "Estudiante de ISI",
				Eliminado = false,
				Version = 2
			};
			cliente.SetearMetaCampo(MetaCamposKeys.ClienteCondicionVentaPorDefecto, 
			                        ((int)CondicionPago.CuentaCorriente).ToString());
			
			return cliente;
		}
		
		public static Cliente CrearClienteCesar()
		{
			return new Cliente {
				Codigo = 6,
				RazonSocial = "César M. Sandrigo",
				Domicilio = CrearDomicilioCesar(),
				Telefono = "0342 154123456",
				CorreoElectronico = "miltondp@gmail.com",
				TipoResponsable = TipoResponsable.ResponsableInscripto,
				Cuit = new CUIT(20119260397),
				PorcentajeBonificacion = 10m,
				Notas = "Otro estudiante de ISI",
				Eliminado = false,
				Version = 4
			};
		}
		
		public static Proveedor CrearProveedorFerrum()
		{
			return new Proveedor {
				RazonSocial = "Proveedor de Ferrum",
				Domicilio = CrearDomicilioCesar(),
				Telefono = "0342 154123456",
				CorreoElectronico = "ferrum@gmail.com",
				TipoResponsable = TipoResponsable.ResponsableInscripto,
				Cuit = new CUIT(20119260397),
				Notas = "Una nota",
				Eliminado = false,
				Version = 4
			};
		}
		
		public static Proveedor CrearProveedorFV()
		{
			return new Proveedor {
				RazonSocial = "Proveedor de FV",
				Domicilio = CrearDomicilioMilton(),
				Telefono = "0342 22222",
				CorreoElectronico = "ventas@fv.com",
				TipoResponsable = TipoResponsable.ResponsableInscripto,
				Cuit = new CUIT(20062192578),
				Notas = "Otra nota",
				Eliminado = false,
				Version = 4
			};
		}
		
		public static Rubro CrearRubroGastos()
		{
			Rubro rubroGastos =  new Rubro {
											Codigo = 30,
											Nombre = "Gastos",
											Descripcion = "Gastos",
											RubroPadre = null
										};
			
			return rubroGastos;
		}
		
		public static Rubro CrearRubroGriferia()
		{
			Rubro rubroGriferia =  new Rubro {
											Codigo = 20,
											Nombre = "Grifería",
											Descripcion = "Griferías",
											RubroPadre = null
										};
			
			rubroGriferia.RubroPadre = new Rubro {
												Codigo = 1,
												Nombre = "Comercio",
												Descripcion = "Comercio",
												RubroPadre = null
											};
			return rubroGriferia;
		}
		
		public static Rubro CrearRubroMosaico()
		{
			Rubro rubroMosaico = new Rubro {
										Codigo = 30,
										Nombre = "Mosaico",
										Descripcion = "Mosaicos",
										RubroPadre = null
									};
			
			rubroMosaico.RubroPadre = new Rubro {
											Codigo = 2,
											Nombre = "Industria",
											Descripcion = "Industria",
											RubroPadre = null
			};
			
			return rubroMosaico;
		}
		
		public static Producto CrearProductoParaNotaCredito()
		{
			return new Producto {
				Codigo = "NC",
				Descripcion = "NOTA CREDITO SOBRE FACTURA",
				ImporteSinIva = 0.02m,
				PorcentajeIva = 21m,
				ProductoAbstracto = true,
				PrecioCosto = 0.01m,
				Unidad = UnidadProducto.Unidad,
				Notas = "Un producto para las notas de crédito"
			};
		}
		
		public static Producto CrearProductoGriferia()
		{
			return new Producto {
				Codigo = "7",
				Descripcion = "Griferia XYZ Modelo 8",
				ImporteSinIva = 60.00m,
				PorcentajeIva = 21m,
				PrecioCosto = 54.22m,
				SubRubro = CrearRubroGriferia(),
				Unidad = UnidadProducto.Unidad,
				Notas = "Un producto griferia"
			};
		}
		
		public static Producto CrearProductoMosaico()
		{
			return new Producto {
				Codigo = "5",
				Descripcion = "Mosaico granito blanco",
				ImporteSinIva = 21.3m,
				PorcentajeIva = 21m,
				PrecioCosto = 18.12m,
				SubRubro = CrearRubroMosaico(),
				Unidad = UnidadProducto.MetroCuadrado,
				Notas = "Un producto mosaico"
			};
		}
										
		public static Producto CrearProductoAbstracto()
		{
			return new Producto {
				Codigo = "13",
				Descripcion = "Producto de industria",
				ProductoAbstracto = true,
				ImporteSinIva = 1.00m,
				PorcentajeIva = 21m,
				PrecioCosto = 0.50m,
				Rubro = CrearRubroMosaico(),
				Unidad = UnidadProducto.Unidad,
				Notas = "Un producto abstracto de industria"
			};
		}
		
		public static Producto CrearProductoGasto()
		{
			return new Producto {
				Codigo = "GST-001",
				Descripcion = "GASTO: ",
				ImporteSinIva = 60.00m,
				PorcentajeIva = 21m,
				PrecioCosto = 54.22m,
				Rubro = CrearRubroGastos(),
				Unidad = UnidadProducto.Unidad,
				Notas = "UN PRODUCTO GASTO"
			};
		}
		
		public static ItemFactura CrearItemFactura(Producto producto, decimal cantidad)
		{
			return new ItemFactura(producto, cantidad);
		}
		
		public static ItemFacturaCompra CrearItemFacturaCompra(Producto producto, decimal cantidad)
		{
			return new ItemFacturaCompra(producto, cantidad);
		}
		
		public static ItemPreFactura CrearItemPreFactura(Producto producto, decimal cantidad)
		{
			return new ItemPreFactura(producto, cantidad);
		}
		
		public static ItemPresupuesto CrearItemPresupuesto(Producto producto, decimal cantidad)
		{
			return new ItemPresupuesto(producto, cantidad);
		}
		
		public static ItemRecibo CrearItemRecibo(decimal importe)
		{
			return new ItemRecibo(importe);
		}
		
		public static ItemRecibo CrearItemRecibo(Cheque cheque)
		{
			return new ItemRecibo(cheque);
		}
		
		public static Factura CrearFactura()
		{
			Factura factura = new Factura {
				Fecha = DateTime.Today,
				Tipo = FacturaTipoVenta.FacturaA,
				Codigo = 4,
				Numero = 23423,
				CondicionVenta = new CondicionVentaFacturaGenerica {
					Condicion = CondicionPago.Contado
				},
				Cliente = CrearClienteCesar(),
				Remitos = "12256666",
				ImporteBruto = 60.00m,
				PorcentajeBonificacion = 0.00m,
				ImporteBonificacion = 0.00m,
				ImporteNeto = 60.00m,
				ImporteIva = 12.60m,
				ImporteTotal = 72.60m,
				Observaciones = "Estas son las observaciones",
			};
			
			factura.ClienteRazonSocial = factura.Cliente.RazonSocial;
			factura.ClienteTipoResponsable = factura.Cliente.TipoResponsable;
			factura.ClienteCuit = factura.Cliente.Cuit.ValorNumerico;
			
			factura.Items.Add(TestBase.CrearItemFactura(TestBase.CrearProductoGriferia(), 1));
			
			return factura;
		}
		
		public static FacturaCompra CrearFacturaCompra()
		{
			FacturaCompra factura = new FacturaCompra {
				Tipo = FacturaTipoCompra.FacturaA,
				Codigo = 1,
				Numero = 1234,
				Proveedor = CrearProveedorFerrum(),
				ImporteGravado = 60.00m,
				ImporteIva = 12.60m,
				ImporteTotal = 72.60m,
				PercepcionIB = 0.00m,
				PercepcionIva = 0.00m,
				ImporteExento = 0.00m,
				ImpuestoCombustible = 0.00m,
				Observaciones = "Estas son las observaciones",
				Fecha = DateTime.Today - TimeSpan.FromDays(10),
				FechaImputacion = DateTime.Today,
				Historica = false
			};
			
			factura.ProveedorRazonSocial = factura.Proveedor.RazonSocial;
			factura.ProveedorTipoResponsable = factura.Proveedor.TipoResponsable;
			
			if (factura.Proveedor.Cuit.ValorNumerico.HasValue)
				factura.ProveedorCuit = factura.Proveedor.Cuit.ValorNumerico.Value;
			
			factura.Items.Add(CrearItemFacturaCompra(CrearProductoGriferia(), 1));
			
			return factura;
		}
		
		public static NotaCredito CrearNotaCredito()
		{
			NotaCredito notaCredito = new NotaCredito {
				Fecha = DateTime.Today,
				Tipo = FacturaTipoVenta.FacturaA,
				Codigo = 4,
				Numero = 23423,
				CondicionVenta = new CondicionVentaFacturaGenerica {
					Condicion = CondicionPago.Contado
				},
				Cliente = CrearClienteCesar(),
				Remitos = "12256666",
				ImporteBruto = 60.00m,
				PorcentajeBonificacion = 0.00m,
				ImporteBonificacion = 0.00m,
				ImporteNeto = 60.00m,
				ImporteIva = 12.60m,
				ImporteTotal = 72.60m,
				Observaciones = "Estas son las observaciones",
			};
			
			notaCredito.ClienteRazonSocial = notaCredito.Cliente.RazonSocial;
			notaCredito.ClienteTipoResponsable = notaCredito.Cliente.TipoResponsable;
			notaCredito.ClienteCuit = notaCredito.Cliente.Cuit.ValorNumerico;
			
			notaCredito.Items.Add(TestBase.CrearItemFactura(TestBase.CrearProductoGriferia(), 1));
			
			return notaCredito;
		}
		
		public static PreFactura CrearPreFactura()
		{
			PreFactura preFactura = new PreFactura {
				Fecha = DateTime.Today,
				Tipo = FacturaTipoVenta.FacturaA,
				Codigo = 4,
				CondicionVenta = new CondicionVentaFacturaGenerica {
					Condicion = CondicionPago.Contado
				},
				Cliente = CrearClienteMilton(),
				Remitos = "12256666,2343423",
				ImporteBruto = 60.00m,
				PorcentajeBonificacion = 0.00m,
				ImporteBonificacion = 0.00m,
				ImporteNeto = 60.00m,
				ImporteIva = 12.60m,
				ImporteTotal = 72.60m,
				Observaciones = "Estas son las observaciones",
			};
											
			preFactura.Items.Add(TestBase.CrearItemPreFactura(TestBase.CrearProductoGriferia(), 1));
			
			return preFactura;
		}
										
		public static PreFactura CrearPreFacturaTipoB()
		{
			PreFactura preFactura = new PreFactura {
				Fecha = DateTime.Today,
				Tipo = FacturaTipoVenta.FacturaB,
				Codigo = 4,
				CondicionVenta = new CondicionVentaFacturaGenerica {
					Condicion = CondicionPago.Contado
				},
				Cliente = CrearClienteMilton(),
				Remitos = "12256666,2343423",
				ImporteBruto = 72.60m,
				PorcentajeBonificacion = 0.00m,
				ImporteBonificacion = 0.00m,
				ImporteNeto = 72.60m,
				ImporteIva = 0.00m,
				ImporteTotal = 72.60m,
				Observaciones = "Estas son las observaciones"
			};
											
			preFactura.Items.Add(TestBase.CrearItemPreFactura(TestBase.CrearProductoGriferia(), 1));
			
			return preFactura;
		}
		
		public static PreNotaCredito CrearPreNotaCredito()
		{
			PreNotaCredito preNotaCredito = new PreNotaCredito {
				Fecha = DateTime.Today,
				Tipo = FacturaTipoVenta.FacturaA,
				Codigo = 4,
				CondicionVenta = new CondicionVentaFacturaGenerica {
					Condicion = CondicionPago.Contado
				},
				Cliente = CrearClienteMilton(),
				Remitos = "12256666,2343423",
				ImporteBruto = 60.00m,
				PorcentajeBonificacion = 0.00m,
				ImporteBonificacion = 0.00m,
				ImporteNeto = 60.00m,
				ImporteIva = 12.60m,
				ImporteTotal = 72.60m,
				Observaciones = "Estas son las observaciones",
			};
											
			preNotaCredito.Items.Add(TestBase.CrearItemPreFactura(TestBase.CrearProductoGriferia(), 1));
			
			return preNotaCredito;
		}
										
		public static PreNotaCredito CrearPreNotaCreditoB()
		{
			PreNotaCredito preNotaCredito = new PreNotaCredito {
				Fecha = DateTime.Today,
				Tipo = FacturaTipoVenta.FacturaB,
				CondicionVenta = new CondicionVentaFacturaGenerica {
					Condicion = CondicionPago.Contado
				},
				Cliente = CrearClienteMilton(),
				Remitos = "12256666,2343423",
				ImporteBruto = 72.60m,
				PorcentajeBonificacion = 0.00m,
				ImporteBonificacion = 0.00m,
				ImporteNeto = 72.60m,
				ImporteIva = 0.00m,
				ImporteTotal = 72.60m,
				Observaciones = "Estas son las observaciones"
			};
											
			preNotaCredito.Items.Add(TestBase.CrearItemPreFactura(TestBase.CrearProductoGriferia(), 1));
			
			return preNotaCredito;
		}
		
		public static Recibo CrearRecibo()
		{
			Recibo recibo = new Recibo {
				Cliente = CrearClienteCesar(),
				Concepto = "Algún concepto para el recibo",
				Fecha = DateTime.Today,
				Observaciones = "Observaciones del recibo",
			};
			
			recibo.Items.Add(TestBase.CrearItemRecibo(100.0m));
			
			return recibo;
		}
											
		public static void ProcesarItemsYCalcularImportesPresupuesto(Presupuesto presupuesto)
		{
			decimal sumaItems = 0.00m;
			
			// Proceso los items, asignándoles el precio unitario correcto
			if (presupuesto.Cliente != null &&
			    	(presupuesto.Cliente.TipoResponsable == TipoResponsable.ResponsableInscripto ||
			    	presupuesto.Cliente.TipoResponsable == TipoResponsable.ResponsableNoInscripto)) {
				
				foreach (ItemPresupuesto item in presupuesto.Items) {
					if (!item.ProductoAbstracto)
						item.PrecioUnitario = item.ProductoImporteSinIva;
					
					// Calculo el importe con precisión
					item.Importe =
						(item.PrecioUnitario - (item.PrecioUnitario * (item.PorcentajeBonificacion / 100.0m)))
							* (decimal)item.Cantidad;
					
					sumaItems += item.Importe;
				}
			}
			else {
				foreach (ItemPresupuesto item in presupuesto.Items) {
					if (!item.ProductoAbstracto)
						item.PrecioUnitario = item.ProductoImporteVenta;
					
					// Calculo el importe con precisión
					item.Importe =
						(item.PrecioUnitario - (item.PrecioUnitario * (item.PorcentajeBonificacion / 100.0m)))
							* (decimal)item.Cantidad;
					
					sumaItems += item.Importe;
				}
			}
			
			// Calculo los importes de la factura
			presupuesto.ImporteBruto = StratosUtil.RoundDecimal(sumaItems, 2);
			presupuesto.ImporteBonificacion =
				StratosUtil.RoundDecimal((presupuesto.ImporteBruto * (presupuesto.PorcentajeBonificacion / 100m)), 2);
			presupuesto.ImporteNeto = presupuesto.ImporteBruto - presupuesto.ImporteBonificacion;
		}
											
		public static void ProcesarItemsYCalcularImportesFactura(IFacturaGenerica factura)
		{
			decimal sumaItems = 0.00m;
			
			// Proceso los items de la factura, asignándoles el precio unitario correcto
			if (factura.Tipo == FacturaTipoVenta.FacturaA) {
				foreach (IItemProducto item in factura.ItemsGenericos) {
					
					if (!item.ProductoAbstracto)
						item.PrecioUnitario = item.ProductoImporteSinIva;
					
					// Calculo el importe con precisión
					item.Importe =
						(item.PrecioUnitario - (item.PrecioUnitario * (item.PorcentajeBonificacion / 100.0m)))
							* (decimal)item.Cantidad;
					
					sumaItems += item.Importe;
				}
			}
			else {
				foreach (IItemProducto item in factura.ItemsGenericos) {
					
					if (!item.ProductoAbstracto)
						item.PrecioUnitario = item.ProductoImporteVenta;
					
					// Calculo el importe con precisión
					item.Importe =
						(item.PrecioUnitario - (item.PrecioUnitario * (item.PorcentajeBonificacion / 100.0m)))
							* (decimal)item.Cantidad;
					
					sumaItems += item.Importe;
				}
			}
			
			// Calculo los importes de la factura
			factura.ImporteBruto = StratosUtil.RoundDecimal(sumaItems, 2);
			factura.ImporteBonificacion = StratosUtil.RoundDecimal(factura.ImporteBruto * (factura.PorcentajeBonificacion / 100m), 2);
			factura.ImporteNeto = factura.ImporteBruto - factura.ImporteBonificacion;
			
			if (factura.Tipo == FacturaTipoVenta.FacturaA)
				factura.ImporteIva = StratosUtil.RoundDecimal((factura.ImporteNeto * 0.21m), 2);
			else
				factura.ImporteIva = 0.00m;
			
			factura.ImporteTotal = (factura.ImporteIva + factura.ImporteNeto);
		}
		
		public static void ProcesarItemsYCalcularImportesFacturaCompra(FacturaCompra factura, bool colocarValoresPorDefecto)
		{
			decimal sumaItems = 0.00m;
			
			// Proceso los items de la factura, asignándoles el precio unitario correcto
			foreach (IItemProducto item in factura.Items) {
				
				item.PrecioUnitario = item.ProductoPrecioCosto;
				
				// Calculo el importe con precisión
				item.Importe =
					(item.PrecioUnitario - (item.PrecioUnitario * (item.PorcentajeBonificacion / 100.0m)))
						* (decimal)item.Cantidad;
				
				sumaItems += item.Importe;
			}
			
			// Solo asigno los importes por defecto si estos no fueron asignados.
			
			if (colocarValoresPorDefecto) {
				factura.ImporteGravado = StratosUtil.RoundDecimal(sumaItems, 2);
				factura.ImporteExento = 2.77m;
				factura.ImporteIva = StratosUtil.RoundDecimal((factura.ImporteGravado + factura.ImporteExento) * 0.21m, 2);
				factura.PercepcionIB = 1.45m;
				factura.PercepcionIva = 2.33m;
				factura.ImpuestoCombustible = 19.22m;
			}
			
			factura.ImporteTotal = factura.ImporteGravado + factura.ImporteExento + factura.ImporteIva +
				factura.PercepcionIB + factura.PercepcionIva + factura.ImpuestoCombustible;
		}
		
		public static void ProcesarItemsRecibo(Recibo recibo)
		{
			decimal sumaItems = 0.00m;
			
			// Proceso los items de la factura, asignándoles el precio unitario correcto
			foreach (ItemRecibo item in recibo.Items)
				sumaItems += item.Importe;
		
			recibo.Importe = StratosUtil.RoundDecimal(sumaItems, 2);
		}
		
		public static Presupuesto CrearPresupuesto()
		{
			Presupuesto presupuesto = new Presupuesto {
				Fecha = DateTime.Today,
				Cliente = CrearClienteMilton(),
				Telefono = "123456",
				CorreoElectronico = "alguncorreo@algundominio.com",
				ImporteBruto = 72.60m,
				PorcentajeBonificacion = 0.00m,
				ImporteBonificacion = 0.00m,
				ImporteNeto = 72.60m,
				Observaciones = "Estas son las observaciones",
			};
			
			presupuesto.Items.Add(TestBase.CrearItemPresupuesto(TestBase.CrearProductoGriferia(), 1));
			
			return presupuesto;
		}
		#endregion
	}
}
