// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;

using NUnit.Framework;
using Rhino.Mocks;
using Rhino.Mocks.Interfaces;

using Stratos.Common.Excepciones;
using StratosClient.Tests;
using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using StratosServer.Controladores;
using Stratos.Tests;
using StratosClient.Widgets.Level2;
using StratosClient.Widgets.Level3;
using StratosClient.Widgets.Level1;

namespace StratosClient.Ventas.Recibos.Tests
{
	[TestFixture]
	public class ReciboPresenterTest : StratosClientTestBase
	{
		#region Campos
		protected IReciboVista reciboVista;
		protected ControladorRecibosPrx controladorRecibos;
		protected ControladorClientesPrx controladorClientes;
		
		protected Factura factura1;
		protected Factura factura2;
		protected Factura factura3;
		protected Factura factura4;
		
		protected Recibo recibo;
		#endregion
		
		#region Métodos de inicialización
		[TestFixtureSetUp]
		public override void FixtureInit ()
		{
			base.FixtureInit();
			
			/* Obtengo los proxies a los controladores accediendo a la instancia del servidor
			 * Stratos que tengo. */
			this.controladorRecibos = ControladorRecibosPrxHelper.checkedCast(
				this.controladorTerminales
					.ObtenerProxyControlador("ControladorRecibos"));
			
			this.controladorClientes = ControladorClientesPrxHelper.checkedCast(
				this.controladorTerminales
					.ObtenerProxyControlador("ControladorClientes"));
		}
		
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			// Creo unas facturas
			this.factura1 = TestBase.CrearFactura();
			this.factura1.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.factura1.Codigo = 1;
			this.factura1.Numero = 1;
			this.factura1.Cliente = this.ClienteSinBonificacion;
			this.factura1.Items.Clear();
			this.factura1.Items.Add(TestBase.CrearItemFactura(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.factura1);
			this.factura1.Create();
			
			this.factura2 = TestBase.CrearFactura();
			this.factura2.CondicionVenta.Condicion = CondicionPago.CuentaCorriente;
			this.factura2.Codigo = 1;
			this.factura2.Numero = 2;
			this.factura2.Cliente = this.ClienteSinBonificacion;
			this.factura2.Items.Clear();
			this.factura2.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.factura2);
			this.factura2.Create();
			
			this.factura3 = TestBase.CrearFactura();
			this.factura3.CondicionVenta.Condicion = CondicionPago.Contado;
			this.factura3.Codigo = 1;
			this.factura3.Numero = 3;
			this.factura3.Cliente = this.ClienteSinBonificacion;
			this.factura3.Items.Clear();
			this.factura3.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.factura3);
			this.factura3.Create();
			
			this.factura4 = TestBase.CrearFactura();
			this.factura4.Tipo = FacturaTipoVenta.FacturaB;
			this.factura4.CondicionVenta.Condicion = CondicionPago.Contado;
			this.factura4.Codigo = 1;
			this.factura4.Numero = 4;
			this.factura4.Cliente = this.ClienteConBonificacion; // Aca pongo otro cliente
			this.factura4.Items.Clear();
			this.factura4.Items.Add(TestBase.CrearItemFactura(this.ProductoMosaico, 1));
			TestBase.ProcesarItemsYCalcularImportesFactura(this.factura4);
			this.factura4.Create();
			
			this.FlushAndRecreateScope();
			
			// Creo el recibo que cubre totalmente a las facturas asociadas
			this.factura1 = Factura.Find(this.factura1.Id);
			this.factura2 = Factura.Find(this.factura2.Id);
			
			this.recibo = TestBase.CrearRecibo();
			this.recibo.Cliente = this.factura1.Cliente;
			
			this.recibo.Items.Clear();
			this.recibo.Items.Add(new ItemRecibo(this.factura1.ImporteTotal));
			this.recibo.Items.Add(new ItemRecibo(this.factura2.ImporteTotal));
			
			TestBase.ProcesarItemsRecibo(this.recibo);
			
			this.recibo.Concepto = "Concepto del recibo";
		}
		#endregion
		
		#region Items cambiados
		[Test]
		public void ItemsCambiados_SeRecalculaElImporte()
		{
			// Vuelvo el MockRepository al modo Record
			this.mocks.BackToRecordAll();
			
			IListaItemsReciboVista listaItemsRecibo =
				this.mocks.Stub<IListaItemsReciboVista>();
			listaItemsRecibo.ItemsCambiados += null;
			IEventRaiser itemsRecibo_ItemsCambiados = LastCall.GetEventRaiser();
			
			SetupResult.For<double>(listaItemsRecibo.SumaImportesItemes())
				.Return(156.87);
			
			// Mock para IReciboVista
			this.reciboVista = this.mocks.DynamicMock<IReciboVista>();
			
			SetupResult.For(this.reciboVista.NavegadorElementos)
				.Return(this.mocks.Stub<INavegadorElementosVista>());
			SetupResult.For(this.reciboVista.ListaItemsReciboVista)
				.Return(listaItemsRecibo);
			SetupResult.For(this.reciboVista.RazonSocialVista)
				.Return(this.mocks.Stub<IRazonSocialVista<ClienteDto>>());
			SetupResult.For(this.reciboVista.DateComboVista)
				.Return(this.mocks.Stub<IDateComboVista>());
			
			Expect.Call(this.reciboVista.Importe = 156.87)
				.Message("Se asigna la propiedad Importe de ReciboVista con el valor correspondiente");
			
			this.mocks.ReplayAll();
			
			// Código a testear
			ReciboPresenter reciboPresenter =
				new ReciboPresenter(this.reciboVista,
				                    this.controladorRecibos,
				                    this.controladorClientes);
			
			// Disparo el evento de items cambiados
			itemsRecibo_ItemsCambiados.Raise();
			
			// Verifico las expectations
			this.mocks.VerifyAll();
		}
		#endregion
		
		#region Guardar recibo
		[Test]
		public void GuardarRecibo_ReciboValido()
		{
			// Paso el Recibo a ReciboDto para setear más fácilmente los mock objects
			ReciboDto reciboDto =
				ConvertidorDto.ToDto<Recibo,ReciboDto>(this.recibo);
			
			// Vuelvo el MockRepository al modo Record
			this.mocks.BackToRecordAll();
			
			// IDateComboVista
			IDateComboVista dateComboVista = this.mocks.Stub<IDateComboVista>();
			dateComboVista.Fecha = new DateTime(reciboDto.fechaTicks);
			
			// IRazonSocialVista
			IRazonSocialVista<ClienteDto> razonSocial = this.mocks.Stub<IRazonSocialVista<ClienteDto>>();
			razonSocial.Persona = reciboDto.cliente;
			
			// IListaItemsReciboVista
			IListaItemsReciboVista listaItemsRecibo = this.mocks.Stub<IListaItemsReciboVista>();
			
			SetupResult.For(listaItemsRecibo.ListaItems)
				.Return(reciboDto.items);
			
//			SetupResult.For(listaFacturas.ElementosSeleccionados)
//				.Return(reciboDto.facturasAsociadas);
			
			// Mock para IReciboVista
			this.reciboVista = this.mocks.Stub<IReciboVista>();
			
			SetupResult.For(this.reciboVista.NavegadorElementos)
				.Return(this.mocks.Stub<INavegadorElementosVista>());
			SetupResult.For(this.reciboVista.ListaItemsReciboVista).Return(listaItemsRecibo);
			SetupResult.For(this.reciboVista.RazonSocialVista).Return(razonSocial);
			SetupResult.For(this.reciboVista.DateComboVista).Return(dateComboVista);
			
			this.mocks.ReplayAll();
			
			// Creo el Presenter con los mocks y stubs creados antes
			ReciboPresenter reciboPresenter =
				new ReciboPresenter(this.reciboVista,
				                    this.controladorRecibos,
				                    this.controladorClientes);
			
			this.reciboVista.Concepto = reciboDto.concepto;
			this.reciboVista.Importe = reciboDto.importe;
			
			// Elimino la transacción actual para dejar trabajar tranquilo al controlador
			// de recibos
			this.RemoveTransaction();
			
			try {
				reciboPresenter.GuardarRecibo();
				
				this.mocks.VerifyAll();
			}
			catch (StratosException ex) {
				Assert.Fail("GuardarRecibo arrojó una excepción: " + ex.mensaje);
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Ahora verifico que el recibo se haya guardado
			ReciboDto[] recibos = this.controladorRecibos.BuscarHQL("from Recibo", null).Cast<ReciboDto>().ToArray();
			
			Assert.AreEqual(1, recibos.Length, "Cantidad de recibos");
			
			ReciboDto reciboDtoAlmacenado = recibos[0];
			
			Assert.Greater(reciboDtoAlmacenado.id, 0);
			
			Assert.IsNotNull(reciboDtoAlmacenado.cliente, "Recibo cliente no nulo");
			Assert.AreEqual(reciboDto.cliente.codigo, reciboDtoAlmacenado.cliente.codigo, "Recibo cliente codigo");
			Assert.AreEqual(reciboDto.concepto, reciboDtoAlmacenado.concepto, "Recibo concepto");
			
			Assert.IsNotNull(reciboDtoAlmacenado.items, "Recibo items no nulo");
			Assert.Greater(reciboDtoAlmacenado.items.Count, 0, "Recibo cantidad items mayor a 0");
			Assert.AreEqual(reciboDto.items.Count, reciboDtoAlmacenado.items.Count, "Recibo items cantidad");
			
//			Assert.IsNotNull(reciboDtoAlmacenado.facturasAsociadas, "Recibo facturas asociadas no nulo");
//			Assert.Greater(reciboDtoAlmacenado.facturasAsociadas.Count, 0, "Recibo cantidad facturas asociadas mayor a 0");
//			Assert.AreEqual(reciboDto.facturasAsociadas.Count, reciboDtoAlmacenado.facturasAsociadas.Count, "Recibo facturas asociadas cantidad");
			
			Assert.AreEqual(reciboDto.fechaTicks, reciboDtoAlmacenado.fechaTicks, "Recibo fecha");
			Assert.AreEqual(reciboDto.importe, reciboDtoAlmacenado.importe, "Recibo concepto");
		}
		
		[Test]
		public void GuardarRecibo_ReciboInvalido()
		{
			// Paso el Recibo a ReciboDto para setear más fácilmente los mock objects
			ReciboDto reciboDto =
				ConvertidorDto.ToDto<Recibo,ReciboDto>(this.recibo);
			
			// Vuelvo el MockRepository al modo Record
			this.mocks.BackToRecordAll();
			
			// IDateComboVista
			IDateComboVista dateComboVista = this.mocks.Stub<IDateComboVista>();
			dateComboVista.Fecha = new DateTime(reciboDto.fechaTicks);
			
			// IRazonSocialVista
			IRazonSocialVista<ClienteDto> razonSocial = this.mocks.Stub<IRazonSocialVista<ClienteDto>>();
			razonSocial.Persona = reciboDto.cliente;
			
			// IListaItemsReciboVista
			IListaItemsReciboVista listaItemsRecibo = this.mocks.Stub<IListaItemsReciboVista>();
			
			SetupResult.For(listaItemsRecibo.ListaItems)
				.Return(reciboDto.items);
			
//			SetupResult.For(listaFacturas.ElementosSeleccionados)
//				.Return(reciboDto.facturasAsociadas);
			
			// Mock para IReciboVista
			this.reciboVista = this.mocks.DynamicMock<IReciboVista>();
			
			SetupResult.For(this.reciboVista.NavegadorElementos)
				.Return(this.mocks.Stub<INavegadorElementosVista>());
			SetupResult.For(this.reciboVista.ListaItemsReciboVista).Return(listaItemsRecibo);
			SetupResult.For(this.reciboVista.RazonSocialVista).Return(razonSocial);
			SetupResult.For(this.reciboVista.DateComboVista).Return(dateComboVista);
			
			Expect.Call(delegate {
				this.reciboVista.NotificarError(String.Empty);
			}).IgnoreArguments()
				.Message("Se llama al método NotificarError por el recibo inválido");
			
			this.mocks.ReplayAll();
			
			// Creo el Presenter con los mocks y stubs creados antes
			ReciboPresenter reciboPresenter =
				new ReciboPresenter(this.reciboVista,
				                    this.controladorRecibos,
				                    this.controladorClientes);
			
			this.reciboVista.Concepto = reciboDto.concepto;
			
			// Modifico intencionalmente el importe para dejarlo incorrecto
			this.reciboVista.Importe = reciboDto.importe + 10.0;
			
			// Elimino la transacción actual para dejar trabajar tranquilo al controlador
			// de recibos
			this.RemoveTransaction();
			
			try {
				reciboPresenter.GuardarRecibo();
				
				this.mocks.VerifyAll();
			}
			catch (System.Exception ex) {
				Assert.Fail("Excepcion arrojada al guardar el recibo: " + ex.ToString());
			}
		}
		#endregion
	}
}
