// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;

using NUnit.Framework;
using NUnit.Framework.Constraints;
using Rhino.Mocks;
using NHibernate.Criterion;
using Castle.ActiveRecord;
using Castle.ActiveRecord.Queries;

using StratosServer.Controladores;
using StratosServer.Compras;
using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using Stratos.Tests;
using StratosServer.Tests;
using Stratos.Common.Excepciones;

namespace StratosServer.Compras.Tests
{
	[TestFixture]
	public class ControladorComprasTests : StratosServerDocumentoGenericoTestBase
	{
		#region Campos
		private ControladorComprasPrx controladorCompras;
		
		private FacturaCompra factura1;
		private FacturaCompra factura2;
		#endregion
		
		#region Métodos de inicialización
		[TestFixtureSetUp]
		public override void FixtureInit ()
		{
			base.FixtureInit ();
			
			try {
				this.controladorCompras = ControladorComprasPrxHelper.checkedCast(
					this.controladorTerminales
						.ObtenerProxyControlador("ControladorCompras"));
			}
			catch (StratosException ex) {
				Assert.Fail("Hubo una excepción al obtener el controlador: " + ex.mensaje);
			}
		}
		
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			this.factura1 = CrearFacturaCompra();
			this.factura1.Proveedor = this.ProveedorFerrum;
			this.factura1.Items.Clear();
			this.factura1.Items.Add(CrearItemFacturaCompra(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesFacturaCompra(this.factura1, true);
			
			this.factura2 = CrearFacturaCompra();
			this.factura2.Proveedor = this.ProveedorFV;
			this.factura2.Items.Clear();
			this.factura2.Items.Add(CrearItemFacturaCompra(this.ProductoGriferia, 2));
			TestBase.ProcesarItemsYCalcularImportesFacturaCompra(this.factura2, true);
		}
		#endregion
		
		#region BuscarHQL
		[Test]
		public void BuscarHQLTest()
		{
			
			try {
				this.factura1.Create();
			} catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.RemoveTransaction();
			
			FacturaCompraDto[] facturas = null;
			try {
				facturas = this.controladorCompras
					.BuscarHQL("from FacturaCompra", null).Cast<FacturaCompraDto>().ToArray();
			}
			catch (StratosException ex) {
				Assert.Fail("Hubo una excepción: " + ex.mensaje);
			}
			
			Assert.AreEqual(1, facturas.Length, "Cantidad de facturas");
			Assert.AreEqual(this.factura1.Id, facturas[0].id, "Factura compra id");
		}
		
		[Test]
		public void BuscarHQLTest_DosFacturas()
		{
			
			try {
				this.factura1.Create();
				this.factura2.Create();
			} catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.RemoveTransaction();
			
			FacturaCompraDto[] facturas = null;
			try {
				facturas = this.controladorCompras
					.BuscarHQL("from FacturaCompra", null).Cast<FacturaCompraDto>().ToArray();
			}
			catch (StratosException ex) {
				Assert.Fail("Hubo una excepción: " + ex.mensaje);
			}
			
			Assert.AreEqual(2, facturas.Length, "Cantidad de facturas");
		}
		#endregion
		
		#region BuscarQueryObjects
		[Test]
		[Ignore("Hacer")]
		public void BuscarQueryObjectsTests()
		{
			
		}
		#endregion
		
		#region Refrescar
		[Test]
		public void RefrescarTests()
		{
			this.factura1.Create();
			this.FlushAndRecreateScope();
			
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			FacturaCompraDto facturaDto =
				ConvertidorDto.ToDto<FacturaCompra,FacturaCompraDto>(this.factura1);
			
			// Modifico el objeto persistente
			this.factura1.Observaciones += " algun agregado a las observaciones";
			
			
			try {
				this.factura1.UpdateAndFlush();
			} catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.RemoveTransaction();
			
			// Refresco el objeto
			facturaDto = (FacturaCompraDto)
				this.controladorCompras.Refrescar(facturaDto);
			
			this.FlushAndRecreateReadOnlyScope();
			
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			
			Assert.IsTrue(this.factura1.Observaciones.EndsWith("algun agregado a las observaciones"), "Observaciones fueron modificadas");
			
			Assert.AreEqual(this.factura1.Id, facturaDto.id, "Factura id");
			Assert.AreEqual(this.factura1.Observaciones, facturaDto.observaciones, "Factura concepto");
		}
		#endregion
		
		#region Guardar
		[Test]
		public void GuardarFacturaCompra()
		{
			FacturaCompraDto facturaDto =
				ConvertidorDto.ToDto<FacturaCompra,FacturaCompraDto>(this.factura1);
			
			this.RemoveTransaction();
			
			try {
				this.controladorCompras.Guardar(facturaDto);
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			Assert.AreEqual(1, FacturaCompra.FindAll().Length, "Cantidad facturas");
			
			FacturaCompra facturaLeida = FacturaCompra.FindAll()[0];
			
			Assert.AreEqual(91.96m, facturaLeida.ImporteTotal, "Factura importe");
		}
		
		[Test]
		public void GuardarFacturaCompra_FacturaInvalida()
		{
			FacturaCompraDto facturaDto =
				ConvertidorDto.ToDto<FacturaCompra,FacturaCompraDto>(this.factura1);
			
			// Hago que la factura sea inválida
			facturaDto.importeTotal += 10.0;
			
			this.RemoveTransaction();
			
			try {
				this.controladorCompras.Guardar(facturaDto);
				Assert.Fail("Se pudo guardar la factura inválida");
			}
			catch(ObjetoInvalidoException) {
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			Assert.That(FacturaCompra.FindAll(), Is.Empty, "No hay facturas creadas");
		}
		
		[Test]
		public void GuardarFacturaCompra_ConDatosMinimosProveedor()
		{
			FacturaCompraDto facturaDto =
				ConvertidorDto.ToDto<FacturaCompra,FacturaCompraDto>(this.factura1);
			
			facturaDto.tipo = FacturaTipoCompra.FacturaC;
			
			facturaDto.proveedor = new ProveedorDto();
			facturaDto.proveedor.razonSocial = "Pepe Biondi";
			facturaDto.proveedor.tipoRes = TipoResponsable.Monotributo;
			facturaDto.proveedor.cuit = "20-31486846-4";
			
			this.RemoveTransaction();
			
			try {
				this.controladorCompras.Guardar(facturaDto);
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			Assert.AreEqual(1, FacturaCompra.FindAll().Length, "Cantidad facturas");
			
			FacturaCompra facturaLeida = FacturaCompra.FindAll()[0];
			
			Assert.AreEqual(91.96m, facturaLeida.ImporteTotal, "Factura importe");
			Assert.IsNull(facturaLeida.Proveedor, "Proveedor nulo");
			Assert.AreEqual("Pepe Biondi", facturaLeida.ProveedorRazonSocial, "Proveedor razon social");
			Assert.AreEqual(TipoResponsable.Monotributo, facturaLeida.ProveedorTipoResponsable, "Proveedor tipo responsable");
			Assert.AreEqual(20314868464, facturaLeida.ProveedorCuit, "Proveedor cuit");
		}
		#endregion
		
		#region Modificar
		[Test]
		public void ModificarFacturaCompra()
		{
			this.factura1.Create();
			this.FlushAndRecreateScope();
			
			// Vuelvo a levantar la factura y la paso a Dto
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			NHibernate.NHibernateUtil.Initialize(this.factura1.Items);
			
			FacturaCompraDto facturaDto =
				ConvertidorDto.ToDto<FacturaCompra,FacturaCompraDto>(this.factura1);
			
			// Realizo alguna modificación
			facturaDto.observaciones += "algo agregado a las observaciones";
			
			this.RemoveTransaction();
			
			try {
				this.controladorCompras.Modificar(facturaDto);
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			Assert.AreEqual(1, FacturaCompra.FindAll().Length, "Cantidad facturas");
			
			FacturaCompra facturaLeida = FacturaCompra.FindAll()[0];
			
			// Me aseguro que los cambios hayan sido guardados
			Assert.That(facturaLeida.Id, Is.EqualTo(this.factura1.Id), "Factura id");
			Assert.IsTrue(facturaLeida.Observaciones.EndsWith("algo agregado a las observaciones"), "Observaciones modificadas");
		}
		#endregion
		
		#region EliminarFacturas
		[Test]
		public void EliminarFacturas()
		{
			this.factura1.Create();
			this.factura2.Create();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a levantar las facturas
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			NHibernate.NHibernateUtil.Initialize(this.factura1.Items);
			
			this.factura2 = FacturaCompra.Find(this.factura2.Id);
			NHibernate.NHibernateUtil.Initialize(this.factura2.Items);
				
			this.RemoveTransaction();
			
			// Creo un array con los Ids de ambas facturas
			long[] ids = {this.factura1.Id, this.factura2.Id};
			
			// Ejecuto el método 'EliminarFacturas'
			try {
				this.controladorCompras.EliminarFacturas(ids);
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Me aseguro que ambas facturas se hayan eliminado de la base de datos
			Assert.AreEqual(0, FacturaCompra.FindAll().Length, "Cantidad facturas");
		}
		
		[Test]
		public void EliminarFacturas_UnaDeDos()
		{
			this.factura1.Create();
			this.factura2.Create();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a levantar las facturas
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			NHibernate.NHibernateUtil.Initialize(this.factura1.Items);
			
			this.factura2 = FacturaCompra.Find(this.factura2.Id);
			NHibernate.NHibernateUtil.Initialize(this.factura2.Items);
				
			this.RemoveTransaction();
			
			// Creo un array con el id solo de la factura1
			long[] ids = {this.factura1.Id};
			
			// Ejecuto el método 'EliminarFacturas'
			try {
				this.controladorCompras.EliminarFacturas(ids);
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Levanto todas las facturas de la base de datos
			FacturaCompra[] facturas = FacturaCompra.FindAll();
			
			// Me aseguro de que haya una sola factura
			Assert.AreEqual(1, facturas.Length, "Cantidad facturas");
			
			// Me aseguro que la factura que está en la base de datos sea la factura2
			Assert.AreEqual(this.factura2.Id, facturas[0].Id, "Codigo factura no eliminada");
		}
		
		[Test]
		public void EliminarFacturas_UnaHistorica()
		{
			// Marco a la factura2 como histórica, está no se debería eliminar
			// aunque pase su id como parámetro al método 'EliminarFacturas'
			this.factura2.Historica = true;
			
			// Creo las facturas
			this.factura1.Create();
			this.factura2.Create();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a levantar las facturas
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			NHibernate.NHibernateUtil.Initialize(this.factura1.Items);
			
			this.factura2 = FacturaCompra.Find(this.factura2.Id);
			NHibernate.NHibernateUtil.Initialize(this.factura2.Items);
				
			this.RemoveTransaction();
			
			// Creo un array con los ids de ambas facturas
			long[] ids = {this.factura1.Id, this.factura2.Id};
			
			// Ejecuto el método 'EliminarFacturas'
			try {
				this.controladorCompras.EliminarFacturas(ids);
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Levanto todas las facturas de la base de datos
			FacturaCompra[] facturas = FacturaCompra.FindAll();
			
			// Me aseguro de que haya una factura
			Assert.AreEqual(1, facturas.Length, "Cantidad facturas");
			
			// Me aseguro que la factura que está en la base de datos sea la factura2
			Assert.AreEqual(this.factura2.Id, facturas[0].Id, "Codigo factura no eliminada");
		}
		
		[Test]
		public void EliminarFacturas_FacturaInexistente()
		{
			// Creo las facturas
			this.factura1.Create();
			this.factura2.Create();
			
			this.FlushAndRecreateScope();
			
			// Vuelvo a levantar las facturas
			this.factura1 = FacturaCompra.Find(this.factura1.Id);
			NHibernate.NHibernateUtil.Initialize(this.factura1.Items);
			
			this.factura2 = FacturaCompra.Find(this.factura2.Id);
			NHibernate.NHibernateUtil.Initialize(this.factura2.Items);
				
			this.RemoveTransaction();
			
			// Creo un array con un Id distinto al de ambas facturas
			// Sumando ambos ids me aseguro que el nuevo id inventado sea distinto al de cualquiera
			// de las dos facturas ingresadas.
			long[] ids = {(this.factura1.Id + this.factura2.Id)};
			
			// Ejecuto el método 'EliminarFacturas'
			try {
				this.controladorCompras.EliminarFacturas(ids);
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			// Levanto todas las facturas de la base de datos
			FacturaCompra[] facturas = FacturaCompra.FindAll();
			
			// Me aseguro de que ninguna de las dos facturas haya sido eliminada
			Assert.AreEqual(2, facturas.Length, "Cantidad facturas");
		}
		
		#endregion
	}
}
