// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Threading;

using NUnit.Framework;
using Rhino.Mocks;

using Stratos.Tests;
using StratosServer.Tests;
using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using StratosServer.Productos;
using Stratos.Common.Excepciones;

namespace StratosServer.Productos.Tests
{
	[TestFixture]
	public class ControladorProductosITests : StratosServerDocumentoGenericoTestBase
	{
		#region Campos
		private ControladorProductosI controlador;
		
		private Rubro subRubroCanillas;
		private Rubro subRubroPicoCanillas;
		private Rubro subRubroMosaicoPulido;
		#endregion
		
		#region Métodos de inicialización
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			this.controlador =
				new ControladorProductosI();
			
			this.subRubroCanillas = new Rubro {
				Codigo = 5,
				Nombre = "Canillas",
				Descripcion = "Canillas de griferia",
				RubroPadre = this.RubroGriferia
			};
			this.subRubroCanillas.Create();
			
			this.subRubroPicoCanillas = new Rubro {
				Codigo = 6,
				Nombre = "Picos de canillas",
				Descripcion = "Picos de canillas de griferías",
				RubroPadre = this.RubroGriferia
			};
			this.subRubroPicoCanillas.Create();
			
			this.subRubroMosaicoPulido = new Rubro {
				Codigo = 10,
				Nombre = "Mosaico pulido",
				Descripcion = "Mosaicos pulidos",
				RubroPadre = this.RubroMosaico
			};
			this.subRubroMosaicoPulido.Create();
			
			this.FlushAndRecreateScope();
		}
		#endregion
		
		[Test]
		[Ignore("Test no implementado aún")]
		public void BuscarHQLTest()
		{
		}
		
		[Test]
		[Ignore("Test no implementado aún")]
		public void BuscarQueryObjectsTest()
		{
		}
		
		[Test]
		[Ignore("Test no implementado aún")]
		public void BuscarPorDescripcionTest()
		{
		}
		
		[Test]
		public void BuscarPorCodigoTest()
		{
			ProductoDto[] resultado = null;
			
			this.RemoveTransaction();
			
			try {
				resultado = this.controlador.BuscarPorCodigo(this.ProductoGriferia.Codigo);
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(1, resultado.Length, "Cantidad de productos encontrados");
			Assert.AreEqual(this.ProductoGriferia.Codigo, resultado[0].codigo, "Código");
			Assert.AreEqual(this.ProductoGriferia.Descripcion, resultado[0].descripcion, "Descripción");
		}
		
		[Test]
		public void BuscarPorCodigo_NoSeIncluyenProductosMarcadosComoEliminados()
		{
			// Marco el producto como eliminado
			Producto productoEliminado = this.ProductoGriferia;
			productoEliminado.Eliminado = true;
			productoEliminado.Update();
			
			this.FlushAndRecreateScope();
			
			ProductoDto[] resultado = null;
			
			this.RemoveTransaction();
			
			try {
				resultado = this.controlador.BuscarPorCodigo(this.ProductoGriferia.Codigo);
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(0, resultado.Length, "Cantidad de productos encontrados");
		}
		
		[Test]
		public void BuscarPorCodigo_CodigoCaseDistinto()
		{
			ProductoDto[] resultado = null;
			
			// Modifico el código del producto grifería
			this.productoGriferia.Codigo = "GFR-2882";
			this.productoGriferia.Update();
			
			this.RemoveTransaction();
			
			try {
				resultado = this.controlador.BuscarPorCodigo("gfr-2882");
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(1, resultado.Length, "Cantidad de productos encontrados");
			Assert.AreEqual(this.ProductoGriferia.Codigo, resultado[0].codigo, "Código");
			Assert.AreEqual(this.ProductoGriferia.Descripcion, resultado[0].descripcion, "Descripción");
		}
		
		[Test]
		public void BuscarPorCodigo_ParteDelCodigo_UnaPalabra()
		{
			ProductoDto[] resultado = null;
			
			// Modifico el código del producto grifería
			this.productoGriferia.Codigo = "GFR-2882";
			this.productoGriferia.Update();
			
			this.RemoveTransaction();
			
			try {
				resultado = this.controlador.BuscarPorCodigo("2882");
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(1, resultado.Length, "Cantidad de productos encontrados");
			Assert.AreEqual(this.ProductoGriferia.Codigo, resultado[0].codigo, "Código");
			Assert.AreEqual(this.ProductoGriferia.Descripcion, resultado[0].descripcion, "Descripción");
		}
		
		[Test]
		public void BuscarPorCodigo_ParteDelCodigo_DosPalabra()
		{
			ProductoDto[] resultado = null;
			
			// Modifico el código del producto grifería
			this.productoGriferia.Codigo = "GFR-2882";
			this.productoGriferia.Update();
			
			this.RemoveTransaction();
			
			try {
				resultado = this.controlador.BuscarPorCodigo("gfr 2882");
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(1, resultado.Length, "Cantidad de productos encontrados");
			Assert.AreEqual(this.ProductoGriferia.Codigo, resultado[0].codigo, "Código");
			Assert.AreEqual(this.ProductoGriferia.Descripcion, resultado[0].descripcion, "Descripción");
		}
		
		[Test]
		public void BuscarPorDescripcion_UnaPalabra()
		{
			ProductoDto[] resultado = null;
			
			this.RemoveTransaction();
			
			try {
				resultado = this.controlador.BuscarPorDescripcion("griferia");
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(1, resultado.Length, "Cantidad de productos encontrados");
			Assert.AreEqual(this.ProductoGriferia.Codigo, resultado[0].codigo, "Código");
			Assert.AreEqual(this.ProductoGriferia.Descripcion, resultado[0].descripcion, "Descripción");
		}
		
		[Test]
		public void BuscarPorDescripcion_NoSeIncluyenProductosMarcadosComoEliminados()
		{
			// Marco el producto como eliminado
			Producto productoEliminado = this.ProductoGriferia;
			productoEliminado.Eliminado = true;
			productoEliminado.Update();
			
			this.FlushAndRecreateScope();
			
			ProductoDto[] resultado = null;
			
			this.RemoveTransaction();
			
			try {
				resultado = this.controlador.BuscarPorDescripcion("griferia");
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(0, resultado.Length, "Cantidad de productos encontrados");
		}
		
		[Test]
		public void BuscarPorDescripcion_DosPalabra()
		{
			ProductoDto[] resultado = null;
			
			this.RemoveTransaction();
			
			try {
				resultado = this.controlador.BuscarPorDescripcion("griferia modelo");
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(1, resultado.Length, "Cantidad de productos encontrados");
			Assert.AreEqual(this.ProductoGriferia.Codigo, resultado[0].codigo, "Código");
			Assert.AreEqual(this.ProductoGriferia.Descripcion, resultado[0].descripcion, "Descripción");
		}
		
		[Test]
		public void BuscarPorDescripcion_DevuelveVariosProductos()
		{
			ProductoDto[] resultado = null;
			
			this.RemoveTransaction();
			
			try {
				resultado = this.controlador.BuscarPorDescripcion("g");
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(2, resultado.Length, "Cantidad de productos encontrados");
		}
		
		[Test]
		public void ObtenerRubrosTest()
		{
			this.RemoveTransaction();
			
			RubroDto[] rubros = null;
			try {
				rubros = this.controlador.ObtenerRubros();
			}
			catch (StratosException ex) {
				Assert.Fail(ex.mensaje);
			}
			
			Assert.AreEqual(2, rubros.Length, "Cantidad de Rubros");
			
			foreach(RubroDto rubro in rubros) {
				if ((rubro.codigo != this.RubroPadreComercio.Codigo) &&
				    	(rubro.codigo != this.RubroPadreIndustria.Codigo))
					Assert.Fail("Rubro inválido: " + rubro.nombre);
			}
		}
		
		[Test]
		public void ObtenerSubrubros()
		{
			this.RemoveTransaction();
			
			RubroDto[] subRubros = this.controlador.ObtenerSubrubros(ConvertidorDto.ToDto(this.RubroGriferia));
			
			Assert.AreEqual(2, subRubros.Length, "Cantidad de subrubros obtenidos inválida");
			
			foreach (RubroDto subrubro in subRubros) {
				if ((subrubro.codigo != this.subRubroCanillas.Codigo) &&
				    (subrubro.codigo != this.subRubroPicoCanillas.Codigo))
					Assert.Fail("Subrubro obtenido inválido: " + subrubro.nombre);
			}
		}
		
		[Test]
		public void ObtenerProductoInvalido()
		{
			this.RemoveTransaction();
			
			Producto productoInvalido =
				this.controlador.ObtenerProductoInvalido();
			
			Assert.AreEqual(ControladorProductosI.PRODUCTO_INVALIDO_CODIGO, productoInvalido.Codigo, "Codigo");
			Assert.AreEqual(ControladorProductosI.PRODUCTO_INVALIDO_DESCRIPCION, productoInvalido.Descripcion, "Descripción");
			
//			Assert.AreEqual(ControladorProductosI.RUBRO_INVALIDO_NOMBRE, productoInvalido.Rubro.Nombre, "Subrubro nombre");
//			Assert.AreEqual(ControladorProductosI.RUBRO_INVALIDO_DESCRIPCION, productoInvalido.Rubro.Descripcion, "Subrubro descripción");
			
			Assert.AreEqual(1.00m, productoInvalido.ImporteSinIva, "Importe sin iva");
		}
		
		[Test]
		public void ObtenerProductoInvalido_ProductoInvalidoYaExiste()
		{
			this.RemoveTransaction();
			
			Producto productoInvalido =
				this.controlador.ObtenerProductoInvalido();
			
			productoInvalido =
				this.controlador.ObtenerProductoInvalido();
			
			this.FlushAndRecreateReadOnlyScope();
			
			Producto[] resultado =
				Producto.FindAllByProperty("Descripcion",
				                           ControladorProductosI.PRODUCTO_INVALIDO_DESCRIPCION);
			
			Assert.AreEqual(1, resultado.Length, "Cantidad devuelta");
		}
	}
}
