// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;

using NUnit.Framework;

using Stratos.Tests;
using Stratos.ModeloObjetos;
using StratosServer.Reportes;
using Stratos.ModeloDto;
using Stratos.Common;

namespace StratosServer.Reportes.Tests
{
	[TestFixture()]
	public class FacturaIvaVentasWrapperTests
	{
		[Test()]
		public void FacturaA_NoAnulada ()
		{
			Factura f = TestBase.CrearFactura();
			FacturaIvaVentasWrapper fw = new FacturaIvaVentasWrapper(f);
			
			Assert.AreEqual(f.Cliente.RazonSocial, fw.RazonSocial, "Razón social");
			Assert.AreEqual(f.Cliente.Cuit.ValorFormateado, fw.Cuit, "Cuit");
			Assert.AreEqual(f.Fecha.ToShortDateString(), fw.Fecha, "Fecha");
			Assert.AreEqual("I", fw.Iva, "Iva");
			
			Assert.AreEqual("FA A  0004  00023423", fw.ResumenFactura, "Resumen");
			
			// Importes numéricos
			Assert.AreEqual(0.00m, fw.ExentoNumerico, "Exento numerico");
			Assert.AreEqual(f.ImporteNeto, fw.GravadoNumerico, "Gravado numerico");
			Assert.AreEqual(f.ImporteIva, fw.IvaInscNumerico, "IvaInsc numerico");
			Assert.AreEqual(f.ImporteTotal, fw.TotalNumerico, "Total numerico");
			
			// Importes string			
			Assert.AreEqual("0,00", fw.Exento, "Exento");
			Assert.AreEqual("60,00", fw.Gravado, "Gravado");
			Assert.AreEqual("12,60", fw.IvaInsc, "IvaInsc");
			Assert.AreEqual("72,60", fw.Total, "Total");
		}
		
		[Test()]
		public void FacturaA_Anulada ()
		{
			Factura f = TestBase.CrearFactura();
			f.Anulada = true;
			FacturaIvaVentasWrapper fw = new FacturaIvaVentasWrapper(f);
			
			Assert.AreEqual("---ANULADA---", fw.RazonSocial, "Razon social");
			Assert.AreEqual(String.Empty, fw.Cuit, "Cuit");
			Assert.AreEqual(f.Fecha.ToShortDateString(), fw.Fecha, "Fecha");
			Assert.AreEqual("I", fw.Iva, "Iva");
			
			Assert.AreEqual("FA A  0004  00023423", fw.ResumenFactura, "Resumen");
			
			// Importes numéricos
			Assert.AreEqual(0.00m, fw.ExentoNumerico, "Exento numerico");
			Assert.AreEqual(0.00m, fw.GravadoNumerico, "Gravado numerico");
			Assert.AreEqual(0.00m, fw.IvaInscNumerico, "IvaInsc numerico");
			Assert.AreEqual(0.00m, fw.TotalNumerico, "Total numerico");
			
			// Importes string			
			Assert.AreEqual("", fw.Exento, "Exento");
			Assert.AreEqual("", fw.Gravado, "Gravado");
			Assert.AreEqual("", fw.IvaInsc, "IvaInsc");
			Assert.AreEqual("", fw.Total, "Total");
		}
		
		[Test()]
		public void FacturaB_NoAnulada ()
		{
			// Cliente Consumidor Final
			Cliente cliente = TestBase.CrearClienteCesar();
			cliente.TipoResponsable = TipoResponsable.ConsumidorFinal;
			cliente.Cuit = new CUIT();
			cliente.PorcentajeBonificacion = 0.00m;

			Factura f = TestBase.CrearFactura();
			
			f.Tipo = FacturaTipoVenta.FacturaB;
			f.Cliente = cliente;
			f.ClienteRazonSocial = cliente.RazonSocial;
			f.ClienteCuit = cliente.Cuit.ValorNumerico;
			f.ClienteTipoResponsable = cliente.TipoResponsable;
			f.PorcentajeBonificacion = cliente.PorcentajeBonificacion;
			
			// Agrego los items
			ItemFactura item1 = TestBase.CrearItemFactura(TestBase.CrearProductoGriferia(), 2);
			ItemFactura item2 = TestBase.CrearItemFactura(TestBase.CrearProductoMosaico(), 125.3m);
			
			f.Items.Clear();
			f.Items.Add(item1);
			f.Items.Add(item2);
			ProcesarItemsYCalcularImportesFactura(f);
			
			FacturaIvaVentasWrapper fw = new FacturaIvaVentasWrapper(f);
			
			Assert.AreEqual(f.Cliente.RazonSocial, fw.RazonSocial, "Razón social");
			Assert.AreEqual(f.Cliente.Cuit.ValorFormateado, fw.Cuit, "Cuit");
			Assert.AreEqual(f.Fecha.ToShortDateString(), fw.Fecha, "Fecha");
			Assert.AreEqual("C", fw.Iva, "Iva");
			
			Assert.AreEqual("FA B  0004  00023423", fw.ResumenFactura, "Resumen");
			
			// Importes numéricos
			Assert.AreEqual(0.00m, fw.ExentoNumerico, "Exento numerico");
			Assert.AreEqual(StratosUtil.RoundDecimal((f.ImporteTotal/1.21m), 2), fw.GravadoNumerico, "Gravado numerico");
			Assert.AreEqual(StratosUtil.RoundDecimal((f.ImporteTotal * (1 - 1/1.21m)), 2), fw.IvaInscNumerico, "IvaInsc numerico");
			Assert.AreEqual(f.ImporteTotal, fw.TotalNumerico, "Total numerico");
			
			// Importes string			
			Assert.AreEqual("0,00", fw.Exento, "Exento");
			Assert.AreEqual("2.788,58", fw.Gravado, "Gravado");
			Assert.AreEqual("585,60", fw.IvaInsc, "IvaInsc");
			Assert.AreEqual("3.374,18", fw.Total, "Total");
		}
		
		[Test()]
		public void NotaCredito_NoAnulada ()
		{
			Factura f = TestBase.CrearNotaCredito();
			FacturaIvaVentasWrapper fw = new FacturaIvaVentasWrapper(f);
			
			Assert.AreEqual(f.Cliente.RazonSocial, fw.RazonSocial, "Razón social");
			Assert.AreEqual(f.Cliente.Cuit.ValorFormateado, fw.Cuit, "Cuit");
			Assert.AreEqual(f.Fecha.ToShortDateString(), fw.Fecha, "Fecha");
			Assert.AreEqual("I", fw.Iva, "Iva");
			
			Assert.AreEqual("NC A  0004  00023423", fw.ResumenFactura, "Resumen");
			
			// Importes numéricos
			Assert.AreEqual(0.00m, fw.ExentoNumerico, "Exento numerico");
			Assert.AreEqual(f.ImporteNeto, fw.GravadoNumerico, "Gravado numerico");
			Assert.AreEqual(f.ImporteIva, fw.IvaInscNumerico, "IvaInsc numerico");
			Assert.AreEqual(f.ImporteTotal, fw.TotalNumerico, "Total numerico");
			
			// Importes string			
			Assert.AreEqual("(0,00)", fw.Exento, "Exento");
			Assert.AreEqual("(60,00)", fw.Gravado, "Gravado");
			Assert.AreEqual("(12,60)", fw.IvaInsc, "IvaInsc");
			Assert.AreEqual("(72,60)", fw.Total, "Total");
		}
		
		/// <summary>
		/// Hace el mismo trabajo que el método con el mismo nombre que se encuentra
		/// en TestBase, con la diferencia de que no busca los productos en la base
		/// de datos. No tiene sentido hacerlo, ya que no es la finalidad del test
		/// comprobar esto.
		/// </summary>
		/// <param name="factura">
		/// A <see cref="IFacturaGenerica"/>
		/// </param>
		public static void ProcesarItemsYCalcularImportesFactura(IFacturaGenerica factura)
		{
			decimal sumaItems = 0.00m;
			
			// Proceso los items de la factura, asignándoles el precio unitario correcto
			if (factura.Tipo == FacturaTipoVenta.FacturaA) {
				foreach (IItemProducto item in factura.ItemsGenericos) {
					
					item.PrecioUnitario = item.ProductoImporteSinIva;
					
					// Calculo el importe con precisión
					item.Importe =
						(item.PrecioUnitario - (item.PrecioUnitario * (item.PorcentajeBonificacion / 100.0m)))
							* (decimal)item.Cantidad;
					
					sumaItems += item.Importe;
				}
			}
			else {
				foreach (IItemProducto item in factura.ItemsGenericos) {
					item.PrecioUnitario = item.ProductoImporteVenta;
					
					// Calculo el importe con precisión
					item.Importe =
						(item.PrecioUnitario - (item.PrecioUnitario * (item.PorcentajeBonificacion / 100.0m)))
							* (decimal)item.Cantidad;
					
					sumaItems += item.Importe;
				}
			}
			
			// Calculo los importes de la factura
			factura.ImporteBruto = StratosUtil.RoundDecimal(sumaItems, 2);
			factura.ImporteBonificacion = StratosUtil.RoundDecimal(factura.ImporteBruto * (factura.PorcentajeBonificacion / 100m), 2);
			factura.ImporteNeto = factura.ImporteBruto - factura.ImporteBonificacion;
			
			if (factura.Tipo == FacturaTipoVenta.FacturaA)
				factura.ImporteIva = StratosUtil.RoundDecimal((factura.ImporteNeto * 0.21m), 2);
			else
				factura.ImporteIva = 0.00m;
			
			factura.ImporteTotal = (factura.ImporteIva + factura.ImporteNeto);
		}
		
	}
}
