// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Threading;

using Castle.ActiveRecord;
using Castle.ActiveRecord.Framework;
using Castle.ActiveRecord.Framework.Config;
using Castle.ActiveRecord.Testing;
using NUnit.Framework;
using Rhino.Mocks;

using StratosServer.Core;
using StratosServer.Controladores;
using Stratos.Servicios;
using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using StratosServer.Tests;
using Stratos.Tests;

namespace StratosServer.Tests
{
	public abstract class StratosServerTestBase : TestBase
	{
		#region Campos
		protected IConfigurationSource configurationSource;

		private Thread threadMainServer;
		private Ice.Communicator iceCommunicator;
		protected ServidorStratos servidorStratos;
		protected ControladorTerminalesPrx controladorTerminales;
		
		protected MockRepository mocks;
		protected ControladorBasePrx controladorBase1;
		protected ControladorBasePrx controladorBase2;
		
		protected Ice.Current iceCurrentIp1Puerto1;
		protected Ice.Current iceCurrentIp1Puerto2;
		protected Ice.Current iceCurrentIp2Puerto1;
		
		protected ServicioFacturacionPrx servicioFacturacion1;
		protected ServicioFacturacionPrx servicioFacturacion2;
		
		protected DatosTerminal terminalNoImpresora1;
		protected DatosTerminal terminalImpresora1;
		protected DatosTerminal terminalImpresora2;
		#endregion
		
		#region Métodos de inicialización y finalización
		public override void Configure (InPlaceConfigurationSource config)
		{
			base.Configure(config);
			this.configurationSource = config;
		}
		
		[TestFixtureSetUp]
		public override void FixtureInit()
		{
			base.FixtureInit();
			
			// Configuro el servidor. Para las pruebas utilizo el puerto 20000
			this.servidorStratos = new ServidorStratos(this.configurationSource, "tcp -h 127.0.0.1 -p 20000");
			
			/* Inicio el servidor en un thread distinto. Necesito iniciarlo porque algunas
			 * clases utilizan, por ejemplo, Mono.Addins, y éste se inicializa con el servidor. */
			this.threadMainServer = new Thread(delegate() {
				try
				{
					this.servidorStratos.Iniciar(new string[0]);
				}
				catch (Exception ex)
				{
					Console.Error.WriteLine("Error iniciando StratosServer: {0}", ex);
				}
			});
			threadMainServer.Start();
			
			// Espero a que el thread del servidor termine de inicializarse
			while (this.servidorStratos.Estado != EstadoServidor.ServidorIniciado);

			/* Ahora obtengo un proxy para el Controlador de Terminales (o sea, actuo como cliente
			 * de la instancia del servidor que acabo de crear).
			 * Este proxy va a ser utilizado por otros tests de unidad, para obtener
			 * proxies a otros controladores. */
			this.iceCommunicator = Ice.Util.initialize();

			Ice.ObjectPrx proxyObj =
				this.iceCommunicator
					.stringToProxy("ControladorTerminales:tcp -h localhost" +
					               " -p 20000" +
					               " -t 10000");

			this.controladorTerminales =
				ControladorTerminalesPrxHelper.checkedCast(proxyObj);
		}
		
		[TestFixtureTearDown]
		public override void FixterTearDown()
		{
			this.iceCommunicator.destroy();
			
			this.servidorStratos.Detener();
			this.threadMainServer.Interrupt();

			base.FixterTearDown();
		}

		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			this.mocks = new MockRepository();
			this.mocks.Record();
			
			#region Primer mock para ControladorBasePrx
			this.controladorBase1 =
				this.mocks.Stub<ControladorBasePrx>();
			
			// Creo un stub para la clase Ice.Identity
			Ice.Identity iceIdentity1 = this.mocks.Stub<Ice.Identity>();
			iceIdentity1.name = "IdentidadControlador";
			
			/* El mock para ControladorBasePrx retorna iceIdentity1 al llamar al método
			 * ice_getIdentity() */
			SetupResult.For(this.controladorBase1.ice_getIdentity()).Return(iceIdentity1);
			#endregion
			
			#region Segundo mock para ControladorBasePrx
			this.controladorBase2 =
				this.mocks.Stub<ControladorBasePrx>();
			
			// Creo un stub para la clase Ice.Identity
			Ice.Identity iceIdentity2 = this.mocks.Stub<Ice.Identity>();
			iceIdentity2.name = "OtraIdentidadControlador";
			
			/* El mock para ControladorBasePrx retorna iceIdentity1 al llamar al método
			 * ice_getIdentity() */
			SetupResult.For(this.controladorBase2.ice_getIdentity()).Return(iceIdentity2);
			#endregion
			
			#region Ice.Current Ip 1 Puerto 1
			this.iceCurrentIp1Puerto1 = this.mocks.Stub<Ice.Current>();
			
			Ice.Connection iceConnection1 = this.mocks.Stub<Ice.Connection>();
			SetupResult.For(iceConnection1.ice_toString_())
				.Return("local address = 127.0.0.1:10000\n" +
				        "remote address = 127.0.0.1:3161");
			
			this.iceCurrentIp1Puerto1.con = iceConnection1;
			#endregion
			
			#region Ice.Current Ip 1 Puerto 2
			this.iceCurrentIp1Puerto2 = this.mocks.Stub<Ice.Current>();
			
			Ice.Connection iceConnection2 = this.mocks.Stub<Ice.Connection>();
			SetupResult.For(iceConnection2.ice_toString_())
				.Return("local address = 127.0.0.1:10000\n" +
				        "remote address = 127.0.0.1:3232");
			
			this.iceCurrentIp1Puerto2.con = iceConnection2;
			#endregion
			
			#region Ice.Current Ip 2 Puerto 1
			this.iceCurrentIp2Puerto1 = this.mocks.Stub<Ice.Current>();
			
			Ice.Connection iceConnection3 = this.mocks.Stub<Ice.Connection>();
			SetupResult.For(iceConnection3.ice_toString_())
				.Return("local address = 192.168.1.1:10000\n" +
				        "remote address = 192.168.1.102:9982");
			
			this.iceCurrentIp2Puerto1.con = iceConnection3;
			#endregion
			
			this.mocks.ReplayAll();
			
			// Terminal No Impresora 1
			this.terminalNoImpresora1 = new DatosTerminal {
				identificador= "IdentificadorTerminalNoImpresora1",
				datosTermImpresora = new DatosTerminalImpresora {
					tiposDocumento = new TipoDocumento[0],
					servFacturacion = new List<ServicioFacturacionPrx>()
				}
			};
			
			// Terminal Impresora 1
			this.servicioFacturacion1 =
				this.mocks.Stub<ServicioFacturacionPrx>();
			
			this.terminalImpresora1 = new DatosTerminal {
				identificador= "IdentificadorTerminal1",
				datosTermImpresora = new DatosTerminalImpresora {
					tiposDocumento = new TipoDocumento[] { TipoDocumento.FacturaTipoB },
					servFacturacion = new List<ServicioFacturacionPrx>() {
						this.servicioFacturacion1
					}
				}
			};
			
			// Terminal Impresora 2
			this.servicioFacturacion2 =
				this.mocks.Stub<ServicioFacturacionPrx>();
			
			this.terminalImpresora2 = new DatosTerminal {
				identificador= "IdentificadorTerminal2",
				datosTermImpresora = new DatosTerminalImpresora {
					tiposDocumento = new TipoDocumento[] {
						TipoDocumento.Presupuesto,
						TipoDocumento.FacturaTipoA
					},
					servFacturacion = new List<ServicioFacturacionPrx>() {
						this.servicioFacturacion2
					}
				}
			};
		}
		
		[TearDown]
		public override void TestDispose()
		{
			base.TestDispose();
			
			// Al finalizar cada test elimino la instancia de ControladorTerminalesI
			ControladorTerminalesI.EliminarInstancia();
		}
		#endregion
	}
}
