// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;

using NUnit.Framework;
using Castle.ActiveRecord;

using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using StratosServer.Tests;
using Stratos.Tests;
using Stratos.Common.Excepciones;

namespace StratosServer.Ventas.Presupuestos.Tests
{
	[TestFixture]
	public class ControladorPresupuestosITests : StratosServerDocumentoGenericoTestBase
	{
		#region Campos
		private ControladorPresupuestosI controladorPresupuestos;
		
		private Presupuesto presupuesto1;
		private Presupuesto presupuesto2;
		#endregion
		
		#region Métodos de inicialización
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			this.controladorPresupuestos = new ControladorPresupuestosI();
			
			this.presupuesto1 = TestBase.CrearPresupuesto();
			this.presupuesto1.Cliente = this.ClienteSinBonificacion_ConsumidorFinal;
			this.presupuesto1.Items.Clear();
			this.presupuesto1.Items.Add(TestBase.CrearItemPresupuesto(this.ProductoGriferia, 1));
			TestBase.ProcesarItemsYCalcularImportesPresupuesto(this.presupuesto1);
			this.presupuesto1.Create();
			
			this.FlushAndRecreateScope();
		}
		#endregion
		
		#region BuscarHQL
		[Test]
		public void BuscarHQLTest()
		{
			this.RemoveTransaction();
			
			PresupuestoDto[] presupuestos = (PresupuestoDto[])
				this.controladorPresupuestos.BuscarHQL("from Presupuesto", null);
			
			Assert.AreEqual(1, presupuestos.Length, "Cantidad de presupuestos");
			Assert.AreEqual(this.presupuesto1.Id, presupuestos[0].id, "Presupuesto id");
		}
		#endregion
		
		#region BuscarQueryObjects
		[Test]
		[Ignore("Hacer")]
		public void BuscarQueryObjectsTests()
		{
			
		}
		#endregion
		
		#region Refrescar
		[Test]
		public void RefrescarTests()
		{
			PresupuestoDto presupuestoDto =
				ConvertidorDto.ToDto<Presupuesto,PresupuestoDto>(this.presupuesto1);
			
			// Modifico el objeto persistente
			this.presupuesto1.CorreoElectronico = "algunadirecciondecorreo@algunproveedor.com";
			this.presupuesto1.Update();
			
			this.RemoveTransaction();
			
			// Refresco el objeto
			presupuestoDto = (PresupuestoDto)
				this.controladorPresupuestos.Refrescar(presupuestoDto);
			
			Assert.AreEqual(this.presupuesto1.Id, presupuestoDto.id, "Presupuesto id");
			Assert.AreEqual(this.presupuesto1.CorreoElectronico, presupuestoDto.correoElectronico, "Presupuesto mail");
		}
		#endregion
		
		#region GuardarPresupuesto
		[Test]
		public void GuardarPresupuesto_NuevoPresupuesto()
		{
			ActiveRecordBase<Presupuesto>.DeleteAll();
			
			this.FlushAndRecreateScope();
			
			this.MarcarComoNoPersistido(ref this.presupuesto1);
			PresupuestoDto presupuestoNuevo =
				ConvertidorDto.ToDto<Presupuesto,PresupuestoDto>(this.presupuesto1);
			
			this.RemoveTransaction();
			
			try {
				this.controladorPresupuestos.Guardar(presupuestoNuevo);
				
				// Verifico que se haya guardado
				PresupuestoDto[] presupuestos = (PresupuestoDto[])
					this.controladorPresupuestos.BuscarHQL("from Presupuesto", null);
				
				Assert.AreEqual(1, presupuestos.Length, "Cantidad presupuestos");
			}
			catch (StratosException ex) {
				Assert.Fail("Error al guardar el presupuesto: " + ex.mensaje);
			}
		}
		
		[Test]
		public void GuardarPresupuesto_ModificacionPresupuestoExistente()
		{
			// Elimino todos los presupuestos y vuelvo a crear presupuesto1
			ActiveRecordBase<Presupuesto>.DeleteAll();
			this.FlushAndRecreateScope();
			
			this.MarcarComoNoPersistido(ref this.presupuesto1);
			this.presupuesto1.Create();
			this.FlushAndRecreateScope();
			
			/* Agrego un nuevo item al presupuesto (no guardo los cambios en el objeto persistente,
			 * simplemente lo hago así para pasar más fácilmente al objeto DTO */
			this.presupuesto1.Items[0].Cantidad = 10;
			this.presupuesto1.Items.Add(TestBase.CrearItemPresupuesto(this.ProductoMosaico, 1));
			TestBase.ProcesarItemsYCalcularImportesPresupuesto(this.presupuesto1);
			
			PresupuestoDto presupuestoExistente =
				ConvertidorDto.ToDto<Presupuesto,PresupuestoDto>(this.presupuesto1);
			
			presupuestoExistente.correoElectronico = "midireccionde@correoelectronico.com";
			
			this.RemoveTransaction();
			
			try {
				this.controladorPresupuestos.Guardar(presupuestoExistente);
				
				// Verifico que se haya guardado
				PresupuestoDto[] presupuestos = (PresupuestoDto[])
					this.controladorPresupuestos
						.BuscarHQL("from Presupuesto p " +
						           "	join fetch p.Items", null);
				
				this.FlushAndRecreateReadOnlyScope();
				
				Assert.AreEqual(1, presupuestos.Length, "Cantidad presupuestos");
				Assert.AreEqual(presupuestoExistente.correoElectronico, presupuestos[0].correoElectronico, "Presupuesto mail");
				
				// Items en general (me aseguro que funcione el modo de cascade)
				ItemPresupuesto[] itemsTotales = ItemPresupuesto.FindAll();
				Assert.AreEqual(2, itemsTotales.Length, "Cantidad de items totales");
				
				// Items del presupuesto guardado
				Assert.AreEqual(2, presupuestos[0].items.Count, "Presupuesto cantidad items");
				
				ItemProductoDto[] items =
					presupuestos[0].items.Cast<ItemProductoDto>()
						.OrderByDescending(item => item.cantidad)
						.ToArray();
				
				Assert.AreEqual(10, items[0].cantidad, "Item cantidad");
				Assert.AreEqual(this.presupuesto1.Items[0].ProductoId, items[0].productoId, "Item producto id");
			}
			catch (StratosException ex) {
				Assert.Fail("Error al guardar el presupuesto: " + ex.mensaje);
			}
		}
		
		[Test]
		public void GuardarPresupuesto_VersionIncorrecta()
		{
			// Creo el objeto DTO
			PresupuestoDto presupuestoExistente =
				ConvertidorDto.ToDto<Presupuesto,PresupuestoDto>(this.presupuesto1);
			
			// Ahora modifico el objeto persistente para aumentar su versión en la base de datos
			this.presupuesto1.CorreoElectronico = "algunadireccion@algunprov.com";
			this.presupuesto1.Update();
			
			this.RemoveTransaction();
			
			try {
				this.controladorPresupuestos.Guardar(presupuestoExistente);
				Assert.Fail("Se pudo guardar el presupuesto");
			}
			catch (ObjetoInexistenteException exception) {
			}
		}
		#endregion
		
		#region EliminarPresupuesto
		[Test]
		public void EliminarPresupuesto()
		{
			ActiveRecordBase<Presupuesto>.DeleteAll();
			this.FlushAndRecreateScope();
			
			this.MarcarComoNoPersistido(ref this.presupuesto1);
			this.presupuesto1.Create();
			this.FlushAndRecreateScope();
			
			// Creo el objeto DTO
			PresupuestoDto presupuestoExistente =
				ConvertidorDto.ToDto<Presupuesto,PresupuestoDto>(this.presupuesto1);
			
			this.RemoveTransaction();
			
			try {
				this.controladorPresupuestos.EliminarPresupuesto(presupuestoExistente);
				
				this.FlushAndRecreateReadOnlyScope();
				
				// Me aseguro que esté eliminado
				Presupuesto[] presupuestos = Presupuesto.FindAll();
				
				Assert.AreEqual(0, presupuestos.Length, "Cantidad presupuestos");
			}
			catch (StratosException ex) {
				Assert.Fail("Error al eliminar presupuesto: " + ex.mensaje);
			}
		}
		
		[Test]
		public void EliminarPresupuesto_PresupuestoInexistente()
		{
			ActiveRecordBase<Presupuesto>.DeleteAll();
			this.FlushAndRecreateScope();
			
			this.MarcarComoNoPersistido(ref this.presupuesto1);
			this.presupuesto1.Create();
			this.FlushAndRecreateScope();
			
			// Creo el objeto DTO
			PresupuestoDto presupuestoExistente =
				ConvertidorDto.ToDto<Presupuesto,PresupuestoDto>(this.presupuesto1);
			
			// Ahora elimino el presupuesto
			this.presupuesto1.Delete();
			
			this.RemoveTransaction();
			
			try {
				this.controladorPresupuestos.EliminarPresupuesto(presupuestoExistente);
				Assert.Fail("Se pudo eliminar un presupuesto inexistente");
			}
			catch (ObjetoInexistenteException exception) {
			}
		}
		#endregion
	}
}
