// Copyright (C) 2009, 2011 Milton Pividori, César Sandrigo
// 
// This file is part of Stratos.
// 
// Stratos is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Stratos is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Stratos.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;

using NUnit.Framework;
using Castle.ActiveRecord;

using Stratos.Common;
using StratosServer.Controladores;
using Stratos.ModeloDto;
using Stratos.ModeloObjetos;
using StratosServer.Tests;
using Stratos.Tests;
using Stratos.Common.Excepciones;

namespace StratosServer.Ventas.Recibos.Tests
{
	[TestFixture]
	public class ControladorRecibosITests : StratosServerDocumentoGenericoTestBase
	{
		#region Campos
		private ControladorRecibosPrx controladorRecibos;
		
		protected Recibo recibo;
		#endregion
		
		#region Métodos de inicialización
		[TestFixtureSetUp]
		public override void FixtureInit ()
		{
			base.FixtureInit ();
			
			try {
				this.controladorRecibos = ControladorRecibosPrxHelper.checkedCast(
					this.controladorTerminales
						.ObtenerProxyControlador("ControladorRecibos"));
			}
			catch (StratosException ex) {
				Assert.Fail("Hubo una excepción al obtener el controlador: " + ex.mensaje);
			}
		}
		
		[SetUp]
		public override void TestInit()
		{
			base.TestInit();
			
			// Recibo
			this.recibo = TestBase.CrearRecibo();
			this.recibo.Cliente = this.ClienteSinBonificacion;
			
			this.recibo.Items.Clear();
			this.recibo.Items.Add(new ItemRecibo(552.88m));
			
			// Agrego un cheque a los items
			Cheque cheque = new Cheque();
			cheque.Banco = "El banco";
			cheque.Sucursal = "Avellaneda";
			cheque.FechaEmision = DateTime.Today.Date;
			cheque.FechaVencimiento = DateTime.Today + TimeSpan.FromDays(10);
			cheque.Importe = 123.12m;
			cheque.Numero = 123456;
			cheque.Observacion = "Observación";
			
			this.recibo.Items.Add(new ItemRecibo(cheque));
			
			TestBase.ProcesarItemsRecibo(this.recibo);
			
			this.recibo.Concepto = "Concepto del recibo";
		}
		#endregion
		
		#region BuscarHQL
		[Test]
		public void BuscarHQLTest()
		{
			
			try {
				this.recibo.Create();
			} catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.RemoveTransaction();
			
			ReciboDto[] recibos = null;
			try {
				recibos = this.controladorRecibos
					.BuscarHQL("from Recibo", null).Cast<ReciboDto>().ToArray();
			}
			catch (StratosException ex) {
				Assert.Fail("Hubo una excepción: " + ex.mensaje);
			}
			
			Assert.AreEqual(1, recibos.Length, "Cantidad de recibos");
			Assert.AreEqual(this.recibo.Id, recibos[0].id, "Recibo id");
		}
		#endregion
		
		#region BuscarQueryObjects
		[Test]
		[Ignore("Hacer")]
		public void BuscarQueryObjectsTests()
		{
			
		}
		#endregion
		
		#region Refrescar
		[Test]
		public void RefrescarTests()
		{
			this.recibo.Create();
			this.FlushAndRecreateScope();
			
			this.recibo = Recibo.Find(this.recibo.Id);
			ReciboDto reciboDto =
				ConvertidorDto.ToDto<Recibo,ReciboDto>(this.recibo);
			
			// Modifico el objeto persistente
			this.recibo.Concepto += " algun agregado al concepto";
			
			
			try {
				this.recibo.UpdateAndFlush();
			} catch (ActiveRecordValidationException ex) {
				Assert.Fail(this.ObtenerMensaje(ex));
			}
			
			this.RemoveTransaction();
			
			// Refresco el objeto
			reciboDto = (ReciboDto)
				this.controladorRecibos.Refrescar(reciboDto);
			
			this.FlushAndRecreateReadOnlyScope();
			this.recibo = Recibo.Find(this.recibo.Id);
			
			Assert.IsTrue(this.recibo.Concepto.EndsWith("algun agregado al concepto"), "Concepto fue modificado");
			
			Assert.AreEqual(this.recibo.Id, reciboDto.id, "Recibo id");
			Assert.AreEqual(this.recibo.Concepto, reciboDto.concepto, "Recibo concepto");
		}
		#endregion
		
		#region GuardarRecibo
		[Test]
		public void GuardarRecibo()
		{
			ReciboDto reciboDto =
				ConvertidorDto.ToDto<Recibo,ReciboDto>(this.recibo);
			
			this.RemoveTransaction();
			
			try {
				this.controladorRecibos.Guardar(reciboDto);
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			Assert.AreEqual(1, Recibo.FindAll().Length, "Cantidad recibos");
			
			Recibo reciboLeido = Recibo.FindAll()[0];
			
			Assert.AreEqual(552.88m + 123.12m, reciboLeido.Importe, "Recibo importe");
		}
		
		[Test]
		public void GuardarRecibo_ReciboInvalido()
		{
			ReciboDto reciboDto =
				ConvertidorDto.ToDto<Recibo,ReciboDto>(this.recibo);
			
			// Modifico el recibo para que el importe sea inválido
			reciboDto.importe += 10.0;
			
			ActiveRecordBase<Recibo>.DeleteAll();
			
			this.RemoveTransaction();
			
			try {
				this.controladorRecibos.Guardar(reciboDto);
				
				Assert.Fail("Se pudo crear el recibo inválido");
			}
			catch(ObjetoInvalidoException ex) {
			}
			catch(System.Exception ex) {
				Assert.Fail("Excepcion incorrecta: " + ex.ToString());
			}
			
			this.FlushAndRecreateReadOnlyScope();
			
			Assert.That(Recibo.FindAll(), Is.Empty, "No hay recibos creados");
		}
		#endregion
	}
}
